/*
 *   $RCSfile: BF_GetSmaller.c,v $  
 *   $Revision: 1.1.1.1 $  
 *   $Date: 1996/05/04 21:55:14 $      
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/


#include <stdio.h>
#include "ess.h"
#include "checking.h"
#include "io.h"
#include "page.h"
#include "list.h"
#include "object.h"
#include "tid.h"
#include "bf.h"
#include "chunk.h"
#include "bf_macro.h"
#include "error.h"
#include "trace.h"
#include "lock.h"
#include "bf_intfuncs.h"
#include "bf_extfuncs.h"
#include "bf_globals.h"

/*
 *============================================================================
 *
 * BF_FindFreeSlot:
 * 	Finds free region of contiguous buffer space
 *
 * RETURNS:
 *	information on the region
 *
 * ERROR RETURNS:
 *	none
 *
 * SIDE EFFECTS:
 *
 * BUGS:
 *
 *============================================================================
 */


 BUFINFO
*BF_FindFreeSlot(

	PAGE2SIZE			page2size  	/* power-of-2 sizeof chunk */
)
{

	register BUFINFO	*start;
	register BUFINFO	*end;
	register BUFINFO	*current;
	register BUFINFO	*limit;
	register int		length;
	register BUFINFO	*first;
	PAGEHASH			*pageHash;
	CHUNKDESC			*chunkDesc;
	int					foundSlots;


	TRPRINT(TR_BF, TR_LEVEL_1, ("looking for blockcount:%d", BLOCKCOUNT(page2size)));

	SM_ASSERT(LEVEL_3, page2size >= MIN_PAGE2SIZE)

	/*
	 *	get the actual number of blocks that are needed
	 */
	length = BLOCKCOUNT(page2size);

	/*
	 *	Set the starting location to be a multiple of the maximum page
	 *	size so that the buddy system allocation method is maintained.
	 *	This assures that we start at a place that has a header.
	 */
	start = first =
		&BufInfoTable[((CacheClock >> BLOCK2COUNT(page2size)) << BLOCK2COUNT(page2size))];
	end = &BufInfoTable[NumBuffers];
	TRPRINT(TR_BF, TR_LEVEL_2, ("start:%d end:%d", start->index, NumBuffers - 1));

	/*
	 *	check to see if we are past the end of the buffer table
	 *	which can happen if we don't have a power of 2 for
	 *	the number of buffers in the system;
	 */
	if (start >= end)	{

		/*
		 *	make start and end at the beginning of the table
		 */
		start = first = BufInfoTable;
		TRPRINT(TR_BF, TR_LEVEL_2, ("setting pointer to slot 0"));
	}

	/*
	 *	look through the entire cache looking for free slots
	 */
	for (;;)	{

		/*
		 *	set the start and end of the search
		 */
		current = start;
		limit   = start + length;
		TRPRINT(TR_BF, TR_LEVEL_2, ("current:%d limit:%d", current->index, limit->index));

		/*
		 *	look at each slot to see if it is free
		 */
		while ((current < limit) && (current < end))	{

			/*
			 *	Check to see if the slot is in use
			 */
			if (current->type == TYPE_PAGE)	{

				TRPRINT(TR_BF, TR_LEVEL_3, ("type page"));

				/*
				 *	get a pointer to the pageHash
				 */
				pageHash = (PAGEHASH *) current->blockInfo;
				CHECK_PAGEHASH_MAGIC(pageHash);

				/*
				 *	check to see if the page is in a buffer group
				 */
				if (PAGE_IS_USED(pageHash))	{

					/*
					 *	can't use this run
					 */
					TRPRINT(TR_BF, TR_LEVEL_3, ("page used:%d", current->index));
					break;
				}

			} else if (current->type == TYPE_CHUNK)	{

				TRPRINT(TR_BF, TR_LEVEL_3, ("type chunk"));

				/*
				 *	get a pointer to the chunkDesc
				 */
				chunkDesc = (CHUNKDESC *) current->blockInfo;
				CHECK_CHUNKDESC_MAGIC(chunkDesc);

				/*
				 *	check to see if the chunk is in a buffer group
				 */
				if (CHUNK_IS_USED(chunkDesc))	{

					/*
					 *	can't use this run
					 */
					TRPRINT(TR_BF, TR_LEVEL_3, ("chunk used:%d", current->index));
					break;					
				}
			}

			/*
			 *	jump to the slot occupied by the next chunk
			 */
			if (current->type == TYPE_PAGE)	{

				TRPRINT(TR_BF, TR_LEVEL_3, ("page2size:%d", pageHash->page2size));
				current += BLOCKCOUNT(pageHash->page2size);
				TRPRINT(TR_BF, TR_LEVEL_3, ("current:%d", current->index));
			
			} else {

				TRPRINT(TR_BF, TR_LEVEL_3, ("page2size:%d", chunkDesc->buf2size));
				current += BLOCKCOUNT(chunkDesc->buf2size);
				TRPRINT(TR_BF, TR_LEVEL_3, ("current:%d", current->index));
			}
		}

		/*
		 *	check to see if we have a slot of the correct length
		 */
		if ((foundSlots = (current - start)) >= length)	{

			/*
			 *	set the cache clock and return to user
			 */
			CacheClock = current - BufInfoTable;
			TRPRINT(TR_BF, TR_LEVEL_3, ("found:%d slots at:%d CacheClock:%d",
							foundSlots, start->index, CacheClock));
			return(start);

		} else {

			/*
			 *	increment the pointer
			 */
			start = (current >= end) ? BufInfoTable : start + length;
			TRPRINT(TR_BF, TR_LEVEL_3, ("start:%d", start->index));

			/*
			 *	check to see if we have searched the entire cache
			 */
			if (start == first)	{

				/*
				 *	there are no runs of sufficient length
				 */
				return(NULL);
			}
		}
	}	
}

/*============================================================================*/
/* 
 * BF_GetSmaller:
 * 	Routine to search the cache for fragments that can be coalesced
 *	into a larger buffer block
 *
 * RETURNS:
 *	a GROUPLINK pointer to buffer block if found
 *	NULL if no space is found
 *
 * SIDE EFFECTS:
 *	The page is entered into the page hash.
 *
 * SPECIAL NOTE:
 *	No check is made to see whether the page is already in the hash.
 * 	This check should be made by the calling routine.
 */
/*============================================================================*/


 char 
*BF_GetSmaller (

	PAGE2SIZE			block2size,		/* power-of-2 sizeof chunk */
	int					needType 
)
{

	register int		length;
	register PAGEHASH	*pageHash;
	register BUFINFO	*current;
	register BUFINFO	*start;
	register BUFINFO	*end;
	register GROUPLINK	*groupLink;
	register CHUNKDESC	*chunkDesc;
	register PAGEINFO	*pageInfo;


	TRPRINT(TR_BF, TR_LEVEL_1, ("looking for slot length:%d", BLOCKCOUNT(block2size)));

	SM_ASSERT(LEVEL_3, block2size >= MIN_PAGE2SIZE);

	/*
	 *	attempt to find a slot that is free
	 */
	if ((start = BF_FindFreeSlot(block2size)) == 0)	{

		/*
		 *	return the fact that we could not find a
		 *	free block of sufficient size
		 */
		return(NULL);
	}

	/*
	 *	get a pointer to the first slot in the group
	 */
	current = start;
	length = BLOCKCOUNT(block2size);
	end = current + length;

	/*
	 *	process the imbedded buffers in this block of free buffers
	 */
	while (current < end)	{

		TRPRINT(TR_BF, TR_LEVEL_2, ("current:%d", current->index));

		/*
		 *	check to see what type of buffer this is
		 */
		if (current->type == TYPE_PAGE)	{

			TRPRINT(TR_BF, TR_LEVEL_2, ("page buffer"));

			/*
			 *	get a pointer to the groupLink
			 */
			pageHash = (PAGEHASH *) current->blockInfo;
			CHECK_PAGEHASH_MAGIC(pageHash);

			/*
			 *	Remove page from the hash table and flush it if need be.
			 *  (remove==TRUE, force==TRUE)
			 */
			bf_InvalidatePage(pageHash, TRUE, TRUE);

			/*
			 *	Advance to the next buffer
			 */
			current += BLOCKCOUNT(pageHash->page2size);

			/*
			 *	get a pointer to the groupLink
			 */
			groupLink = (GROUPLINK *) pageHash->memberList.succ->item;
			CHECK_GROUPLINK_MAGIC(groupLink);

			/*
			 *	check to see if this is the first sub chunk
			 */
			if (groupLink->pageHash != (PAGEHASH *) start->blockInfo)	{

				TRPRINT(TR_BF, TR_LEVEL_2, ("this is not first link"));
				listMoveEnq( &BufStructPool, &(groupLink->groupList) );
				CHECK_PAGEHASH_MAGIC(groupLink->pageHash);
			}
		
		} else if (current->type == TYPE_CHUNK) {

			/*
			 *	get a pointer to the chunk descriptor
			 */
			chunkDesc = (CHUNKDESC *) current->blockInfo;
			CHECK_CHUNKDESC_MAGIC(chunkDesc);
			BF_AuditChunk(chunkDesc);
			TRPRINT(TR_BF, TR_LEVEL_2, ("chunk buffer size:%d", chunkDesc->bufCount));

			/*
			 *	Check to see if the buffer contains valid data
			 */
			if (CHUNK_IS_VALID(chunkDesc))	{

				BF_InvalidateChunk(chunkDesc);
			}

			/*
			 *	Advance to the next buffer
			 */
			current += BLOCKCOUNT(chunkDesc->buf2size);

			/*
			 *	take the grouplink off the free queue and
			 *	put it in the structure pool
			 */
			if (chunkDesc != (CHUNKDESC *) start->blockInfo)	{

				TRPRINT(TR_BF, TR_LEVEL_2, ("this is not the first chunk"));
				SM_ASSERT(LEVEL_1, (chunkDesc->chunkPage == NULL));
				listMoveEnq( &ChunkDescPool, &(chunkDesc->memberList) );
			}

		} else {

			/*
			 *	die
			 */
			SM_ERROR(TYPE_FATAL, esmINTERNAL);
		}
	}

	/*
	 *	check to see what type of buffer is needed
	 */
	if (needType == TYPE_PAGE)	{

		TRPRINT(TR_BF, TR_LEVEL_2, ("need page"));

		/*
		 *	check to see what kind of structure we have
		 */
		if (start->type == TYPE_PAGE)	{

			TRPRINT(TR_BF, TR_LEVEL_2, ("first is page"));

			/*
			 *	set the pageHash info
			 */
			pageHash = (PAGEHASH *) start->blockInfo;
			CHECK_PAGEHASH_MAGIC(pageHash);
			pageHash->page2size = block2size;

			/*
			 *	set the pageinfo information
			 */
			pageInfo = pageHash->pageInfo;
			pageInfo->blockCount = BLOCKCOUNT(block2size);

			/*
			 *	get a pointer to the groupLink
			 */
			groupLink = ((GROUPLINK *) pageHash->memberList.succ->item);
			CHECK_GROUPLINK_MAGIC(groupLink);

			/*
			 *	take the groupLink off the free list
			 */
			listRemove( &(groupLink->groupList) );

		} else {

			TRPRINT(TR_BF, TR_LEVEL_2, ("first is chunk"));

			/*
			 *	get a pointer to the chunk descriptor
			 */
			chunkDesc = (CHUNKDESC *) start->blockInfo;
			CHECK_CHUNKDESC_MAGIC(chunkDesc);

			/*
			 *	Get a pointer to the needed structure
			 */
			if ((groupLink = (GROUPLINK *) listDeq( &BufStructPool )) == NULL)	{

				return(NULL);
			}

			/*
			 *	setup the groupLink
			 */
			CHECK_GROUPLINK_MAGIC(groupLink);
			groupLink->bufFrame = chunkDesc->bufFrame;

			/*
			 *	setup the pageHash structure
			 */
			pageHash = groupLink->pageHash;	
			CHECK_PAGEHASH_MAGIC(pageHash);
			pageHash->page2size = block2size;

			/*
			 *	setup the pageinfo structure
			 */
			pageInfo = pageHash->pageInfo;
			pageInfo->bufIndex = chunkDesc->bufIndex;
			pageInfo->bufInfo = start;
			pageInfo->bufFrame = groupLink->bufFrame = chunkDesc->bufFrame;
			pageInfo->blockCount = BLOCKCOUNT(block2size);

			/*
			 *	give back the chunk descriptor
			 */
			SM_ASSERT(LEVEL_1, (chunkDesc->chunkPage == NULL));
			listMoveEnq( &ChunkDescPool, &(chunkDesc->memberList) );
		}

		/*
		 *	set the status of the slot
		 */
		BF_SetSlot(start, pageInfo->blockCount, (char *) pageHash, TYPE_PAGE);

		/*
		 *	return the saved link
		 */
		return((char *) groupLink);

	} else {

		TRPRINT(TR_BF, TR_LEVEL_2, ("need chunk"));

		/*
		 *	check to see what type first is
		 */
		if (start->type == TYPE_CHUNK)	{

			TRPRINT(TR_BF, TR_LEVEL_2, ("first is chunk"));

			/*
			 *	set the chunk descriptor info
			 */
			chunkDesc = (CHUNKDESC *) start->blockInfo;
			CHECK_CHUNKDESC_MAGIC(chunkDesc);
			chunkDesc->bufCount = BLOCKCOUNT(block2size);
			chunkDesc->buf2size = block2size;

			/*
			 *	take it off the free list
			 */
			listRemove( &(chunkDesc->memberList) );

		} else {

			TRPRINT(TR_BF, TR_LEVEL_2, ("first is page"));

			/*
			 *	get a chunk descriptor off the chunk descriptor pool
			 */
			if ((chunkDesc = (CHUNKDESC *) listDeq( &ChunkDescPool )) == NULL){
				return(NULL);
			}
			CHECK_CHUNKDESC_MAGIC(chunkDesc);
			SM_ASSERT(LEVEL_1, (chunkDesc->chunkPage == NULL));
			BF_AuditChunk(chunkDesc);

			/*
			 *	get a pointer to the pageHash struct
			 */
			pageHash = (PAGEHASH *) start->blockInfo;
			CHECK_PAGEHASH_MAGIC(pageHash);

			/*
			 *	initialize the chunk descriptor
			 */
			chunkDesc->bufIndex = pageHash->pageInfo->bufIndex;
			chunkDesc->bufCount = BLOCKCOUNT(block2size);
			chunkDesc->buf2size = block2size;
			chunkDesc->bufFrame = pageHash->pageInfo->bufFrame;
			chunkDesc->bufInfo  = start;

			/*
			 *	put the structure on the free list
			 */
			groupLink = ((GROUPLINK *) pageHash->memberList.succ->item);
			CHECK_GROUPLINK_MAGIC(groupLink);
			listMoveEnq( &BufStructPool, &(groupLink->groupList) );
		}

		/*
		 *	set the status of the slot
		 */
		BF_SetSlot(start, chunkDesc->bufCount, (char *) chunkDesc, TYPE_CHUNK);

		BF_AuditChunk(chunkDesc);

		/*
		 *	return the chunk descriptor
		 */
		return((char *) chunkDesc);
	}
}
