/*
 *   $RCSfile: BF_InitBuffers.c,v $  
 *   $Revision: 1.2 $  
 *   $Date: 1996/05/04 23:51:43 $      
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/


#include "ess.h"
#include "checking.h"
#include "sysdefs.h"
#include "list.h"
#include "io.h"
#include "page.h"
#include "object.h"
#include "tid.h"

#ifdef SERVER_MAKE
#include "lock.h"
#include "pool.h"
#include "bitvec.h"
#include "link.h"
#endif

#include "bf.h"
#include "bf_macro.h"
#ifdef SERVER_MAKE
#include "volume.h"
#include "lsn.h"
#include "trans.h"
#else SERVER_MAKE
#include "chunk.h"
#include "lock.h"
#endif SERVER_MAKE
#include "trace.h"
#include "error.h"
#include "bf_intfuncs.h"
#include "bf_extfuncs.h"
#include "util_funcs.h"
#include "bf_globals.h"

#ifdef SERVER_MAKE
#include "sharedmem_globals.h"
#endif SERVER_MAKE


 int
BF_InitBuffers ()
{

	TRACE(TR_BF, TR_LEVEL_1);

	/*
     *  Allocate the page chunk buffer pool.
	 *  Client & server do it a little differently.
	 */
#ifdef SERVER_MAKE
    /*
	 *  Buffer pool is in shared memory.
     *  Align pages for better performance with raw device io.
     */
#if defined(mips) || defined(vax) || defined(sparc) || defined(_AIX) || defined(hpux) || defined(linux)
    BufferPool = (char*) 
		( ((int) (ShmAddress+MIN_PAGESIZE)) & ~(MIN_PAGESIZE - 1));
    SM_ASSERT(LEVEL_1, sizeof(int) == sizeof(ShmAddress));
#else
	unsupported : compile-time error
#endif

#else SERVER_MAKE
	/*
	 *  CLIENT SIDE:
	 *	Allocate one extra page for page alignment.
	 *  MallocBufferPool is external so it can be freed on shutdown.
	 */
	SM_ASSERT(LEVEL_1, sizeof(int) == sizeof(BufferPool));

	/* can't use GET macro here */
	if ((MallocBufferPool = 
		(char *)malloc((size_t)((NumBuffers+1) * MIN_PAGESIZE))) == NULL)	{
		SM_ERROR(TYPE_SYS, esmMALLOCFAILED);
		return esmFAILURE;
	}

	BufferPool = (char*) 
		( ((int) (MallocBufferPool+MIN_PAGESIZE)) & ~(MIN_PAGESIZE - 1));
#endif SERVER_MAKE

	TRPRINT(TR_BF, TR_LEVEL_2, ("BufferPool:%x\n", BufferPool));

	GET(BufInfoTable, BUFINFO, NumBuffers);

	/*
	 *	Calculate the number of buffer sizes that are possible
	 */
	MaxBuf2Size = findLog2(NumBuffers);
	MaxPage2Size = MaxBuf2Size + MIN_PAGE2SIZE;
	NumBufferSizes = OFFSET(MaxPage2Size) + 1;
	TRPRINT(TR_BF, TR_LEVEL_2, 
		("MaxPage2Size:%d MaxBuf2Size:%d NumBufferSizes:%d",
					MaxPage2Size, MaxBuf2Size, NumBufferSizes));

	/*
	 *	Allocate space for the headers of all the buffer pools
	 */
	GET(FreeBufTable, LIST, NumBufferSizes);
	TRPRINT(TR_BF, TR_LEVEL_2, ("FreeBufTable:%x\n", FreeBufTable));

	{ 	int			i;
		for (i = 0; i < NumBufferSizes; i++)	{
			initializeList( &(FreeBufTable[i]) );
		}
	}
	{	int			i;
		for (i = 0; i < NumBuffers; i++)	{
			PAGEHASH	*pageHash;
			GROUPLINK	*groupLink;
#ifndef SERVER_MAKE
			PAGEINFO	*pageInfo;
#endif SERVER_MAKE

			pageHash = BF_NewPageHash();

#ifdef SERVER_MAKE
			if ((groupLink = (GROUPLINK *) poolDeq( &GroupLinkPool )) == NULL)	
#else SERVER_MAKE
			if ((groupLink = (GROUPLINK *) listDeq( &GroupLinkPool )) == NULL)	
#endif SERVER_MAKE
			{
				SM_ERROR(TYPE_FATAL, esmINTERNAL);
			}
			CHECK_GROUPLINK_MAGIC(groupLink);

#ifndef SERVER_MAKE
			/*
			 *  The client side has a pageInfo structure in the pageHash.
			 */
			if ((pageInfo = (PAGEINFO *) listDeq( &PageInfoPool )) == NULL)	{
				SM_ERROR(TYPE_FATAL, esmINTERNAL);
			}

			pageHash->pageInfo = pageInfo;
			pageInfo->bufFrame = &(BufferPool[i * MIN_PAGESIZE]);
			pageInfo->bufIndex = i;
			pageInfo->blockCount = 1;
			pageInfo->pageHash = pageHash;
			initializeList( &(pageInfo->userList) );
			pageInfo->bufInfo = &(BufInfoTable[i]);
#else SERVER_MAKE
			pageHash->bufFrame = &(BufferPool[i * MIN_PAGESIZE]);
			pageHash->bufIndex = i;
			pageHash->blockCount = 1;
			pageHash->bufInfo = &(BufInfoTable[i]);
#endif SERVER_MAKE

			/*
			 *	Set the group link pointer to the page hash
			 *	Set the group link pointer to the buffer
			 */
			groupLink->pageHash = pageHash;

#ifdef  SERVER_MAKE
			groupLink->bufFrame = pageHash->bufFrame;
#else  SERVER_MAKE
			groupLink->bufFrame = pageInfo->bufFrame;
#endif  SERVER_MAKE

			/*
			 *	Hang the group link off the page hash
			 */
			listEnq( &(pageHash->memberList), &(groupLink->hashList) );

#ifndef SERVER_MAKE
			BufInfoTable[i].type = TYPE_PAGE;
			BufInfoTable[i].blockInfo = (char *) pageHash;
#else SERVER_MAKE
			BufInfoTable[i].pageHash = pageHash;
#endif SERVER_MAKE

			BufInfoTable[i].index = i;

			/*
			 *	Hang the group link off the free list
			 */
			listEnq( FreeBufTable, &(groupLink->groupList) );
		}
	}
	initializeList( &BufStructPool );
	return esmNOERROR;
}
