/*
 *   $RCSfile: BF_ReadPageWithNeighbor.c,v $
 *   $Revision: 1.1.1.1 $  
 *   $Date: 1996/05/04 21:55:15 $      
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/


#include "sysdefs.h"
#include "ess.h"
#include "checking.h"
#include "io.h"
#include "page.h"
#include "list.h"
#include "object.h"
#include "tid.h"
#include "bf.h"
#include "chunk.h"
#include "bf_macro.h"
#include "sm_macro.h"
#include "trace.h"
#include "error.h"
#include "lock.h"
#include "volume.h"
#include "trans.h"
#include "lsn.h"
#include "loginfo.h"
#include "clog.h"
#include "lm.h"
#include "scan.h"
#include "fi_msg.h"
#include "bf_intfuncs.h"
#include "bf_extfuncs.h"
#include "io_extfuncs.h"
#include "lm_extfuncs.h"
#include "bf_globals.h"
#include "trans_globals.h"
#include "clog_globals.h"
#include "lock_globals.h"
#include "sm_globals.h"


 int 
BF_FindBufGroup (

	PAGEHASH			*pageHash,
	BUFGROUP			*bufGroup,
	GROUPLINK			**retLink 
)
{

	register GROUPLINK	*groupLink;


	TRACE(TR_BF, TR_LEVEL_1);

	/*
	 *	check the pagehash magic number
	 */
	CHECK_PAGEHASH_MAGIC(pageHash);

	/*
	 *	Check quick to see if the item is on the free list
	 */
	if (pageHash->memberCount == 0)	{

		/*
		 *	The page is on a free list
		 */
		TRPRINT(TR_BF, TR_LEVEL_2, ("page is on the free list%x", pageHash));
		return(BF_FREE);
	
	} else {

		TRPRINT(TR_BF, TR_LEVEL_2, ("memberCount:%d", pageHash->memberCount));

		/*
		 *	get a pointer to the first group link
		 */
		groupLink = (GROUPLINK *) FIRST_LIST_ELEMENT( &(pageHash->memberList) );

		/*
		 *	Traverse down the list looking for a match
		 */
		while (groupLink != NULL)	{

			/*
			 *	check magic numbers
			 */
			CHECK_GROUPLINK_MAGIC(groupLink);
			CHECK_BUFGROUP_MAGIC(groupLink->bufGroup);
			SM_ASSERT(LEVEL_3, groupLink->pageHash == pageHash);

			/*
			 *	Check for a match
			 */
			if (groupLink->bufGroup == bufGroup)	{

				/*
				 *	We have a match
				 */
				TRPRINT(TR_BF, TR_LEVEL_2, ("page is in the target group"));

				/*
				 *	Set the group link pointer for return
				 */
				*retLink = groupLink;

				/*
				 *	Check to see if the page is fixed
				 */
				if (groupLink->fixedCount != 0)	{

					/*
					 *	The page is fixed
					 */
					TRPRINT(TR_BF, TR_LEVEL_2, ("page is fixed:%d", groupLink->fixedCount));
					return(BF_FIXED);
				
				} else {

					/*
					 *	The page is unfixed
					 */
					TRPRINT(TR_BF, TR_LEVEL_2, ("page is unfixed"));
					return(BF_UNFIXED);
				}

			} else {	

				/*
				 *	Look at the next entry
				 */
				groupLink = (GROUPLINK *) NEXT_LIST_ELEMENT( &(groupLink->hashList) );
			}
		}

		/*
		 *	We did not find a match
		 */
		TRPRINT(TR_BF, TR_LEVEL_2, ("page is not in target group>"));
		return(BF_OTHER);
	}
}

/*============================================================================*/
/* 
 * This routine used to be bf_ReadPage.  
 * 	This routine will read a single page into a free buffer page unless
 *      the page is already in the buffer pool in which case it avoids an IO.
 *
 * RETURNS:
 *	esmNOERROR        - if all went well.
 *	
 * ERROR RETURNS:
 *	other errors returned by lower levels.
 *
 * SIDE EFFECTS:
 *	The requested page is read into the buffer pool if 
 *	necessary, added to the specified buffer group, and fixed
 *	The pid is entered in the page hash if it is not already there.
 *
 * SPECIAL NOTE:
 *	There are 5 valid states that a page can be in when a read request
 *	is received, namely:
 *
 *   not present
 *
 *   present, readers waiting
 *
 *   present, no readers waiting, not free
 *
 *   present, no readers waiting, free, being replaced
 *   present, no readers waiting, free, not being replaced
 *
 *	All other states are invalid because of the way the 
 *	read/write code has been structured and because of the rule in 
 *	our system which is stated as:
 *
 * 	READ-RULE:
 *		If a page is fixed, then all the pages that have a
 *		dependency on that page must be fixed, and similarly, if
 *		a page is unfixed, then all of that page's dependents
 *		must be unfixed.
 *		For hiearchical data structures this means that
 *		pages must be read and fixed  in root-to-leaf order,
 *		and must be unfixed in leaf-to-root order.
 *	
 *	The READ-RULE basically makes it impossible to run into a situation 
 *	where a fixed page needs to be flushed because of an IO dependency.
 *	It allows us to avoid ugly deadlock problems.
 *
 *	The code below has been set up to deal with the states mentioned 
 *	above, and only those states.
 *	If the the READ-RULE is violated, and a page enters a unexpected
 *	state then the code below will either panic or do something incorrect.
 *
 * SPECIAL NOTE ON MULTI-USER:
 *	Code which deals with multi-user states READERS_WAITING and
 *	BEING_REPLACED (see below) has been over-ridden with Panic's for now.
 *	Error recovery for these states still has to be written, too.
 *
 * NOTES ON LOCK CACHING:
 *	The control flow is as follows:
 *						|
 *						V
 *					bf_ReadPage
 *					   / \
 *					  /   \ (page is cached)
 *					 /	   \
 *			io_ReadPage		lm
 *		    with lock       / \
 *						   /   \ (lock is cached)
 *						  /     \
 *		 		rpc_LockPage()	 --. 
 *		 						  / \
 *								 /   \ (lock strong enough)
 *							    /     \
 *					rpc_LockPage()		return;
 *					to upgrade lock
 *
 */
/*============================================================================*/



 GROUPLINK 
*BF_ReadPageWithNeighbor (

	BUFGROUP			*bufGroup,
	const PID     		*pid,
	PAGETYPE			pageType,
	PAGE2SIZE			page2size,
	LOCKMODE			lockMode,
	NEIGHBORFLAGS		nflags,
	FID					*fid,
	PID 				*neighborPid
)
{
	register int		pageState;
	register PAGEHASH	*pageHash;
	register GROUPLINK	*groupLink = NULL;
	GROUPLINK			*retLink;
	SLOTTEDPAGE			*slottedPage;
	BOOL 				goToServer = FALSE;
	LOCKMODE 			pMode;
	BFFLAGS				bfflags=0;

	TRPRINT(TR_BF, TR_LEVEL_1, 
		("reading page:%d.%d pageType:%d pagesize:%d lockMode:%s",
				pid->page,  pid->volid,
				pageType, RAISETWO(page2size), GETMODE(lockMode)));

	/*
	 *	If the page is a slotted page and the default lock
	 *	mode is EX then get an EX lock.
	 *
	 *	Later, add support for sm_DefaultFileLock
	 */
	if ( (pageType == PAGE_SLOTTED) && (sm_DefaultPageLock == EX) ) {
		lockMode = EX;
	}

    /*
     *	Check to see if the page is in the cache and up-to-date.
     */
    pageHash = bf_FindPage(pid);
    if (pageHash == NULL) {
		TRPRINT(TR_BF, TR_LEVEL_1, ("page %d not in HT", pid->page));
		goToServer = TRUE;

		/*
		 *	Get a fixed size buffer off the free list
		 */
		if ((groupLink = 
			bf_GetPage(bufGroup, pid, pageType, NULL, page2size)) == NULL)	{
			return(NULL);
		}

		/*
		 *	get a pointer to the pagehash entry
		 */
		pageHash  = groupLink->pageHash;

		/*
		 * be sure the lrc for the page is impossibly low
		 * so the server sends us the page
		 */
		*bf_lrcLoc(pageHash) = LrcNone;

		/* end pageHash == NULL */  
	} else {
		/* pageHash != NULL */
		if(pageHash->flags & H_INVALID ) {
			TRPRINT(TR_BF, TR_LEVEL_1, 
				("page %d & H_INVALID (need server)", pid->page));
			goToServer = TRUE;

			/* the page cannot be new */
			pageHash->flags &= ~H_NEW;

#ifdef JUNK
			/* What follows is a kludge to make inter-tx page caching 
			 * work with logical files.  Some day we will eliminate 
			 * logical files, at which time, we can remove this (and lots 
			 * of other) hacks. NB: remove the logical link stuff in the
			 * lock table, the recovery hack (always redo logical link
			 * updates).
			 *
			 * The problem here is that the logical link is updated on the
			 * page by the server (which does NOT increment the lrc on the 
			 * page).
			 * Since it is updated exactly once in its lifetime, it is 
			 * sufficient to treat the page as invalid (even if it is
			 * up-to-date)if the logical link is null (because it is 
			 * possible that the server updated it thanks to another 
			 * intervening tx).
			 * Now, to "treat the page as invalid" means 2 things: we go to 
			 * the server for that page, as with all H_INVALID pages, 
			 * and we make sure the server treats it as invalid by 
			 * setting our lrc to LrcNone.
			 */
			if( PPAGE_TYPE(pageHash) == PAGE_SLOTTED) {
				slottedPage = (SLOTTEDPAGE *) pageHash->pageInfo->bufFrame;
				TRPRINT(TR_BF, TR_LEVEL_1, ("page:%d lrc %d.%d", 
					slottedPage->header.pid.page,
					slottedPage->header.lrc.wrapCount,
					slottedPage->header.lrc.count));
				if( slottedPage->header.nextLogicalPid == NULLPID ) {
					TRPRINT(TR_BF, TR_LEVEL_1, ("page:%d lrc zeroed",
										slottedPage->header.pid.page));
					slottedPage->header.lrc = LrcNone;
				}
			} /* slotted */
#endif JUNK
		} /* else page is valid */
	} /* end pageHash != NULL */

	if(nflags & NEIGHBOR_FLAGS) {
		/* if we are going to the server only to get the
		 * neighbor page, set NOREAD
		 */
		if(goToServer == FALSE)
			bfflags = BF_NOREAD;
		goToServer = TRUE;
	}
	if (goToServer) {
		/*
		 * we should get here ONLY if :
		 *	the page wasn't found in the HT, or 
		 *  it was there but was invalid, or
		 *  this is a scan, so we need the neighbor pid
		 */
		SM_ASSERT(LEVEL_1,  
			((pageHash->memberCount == 0) || 
			 (pageHash->memberCount == 1)) ||
			 (nflags & NEIGHBOR_FLAGS));

		TRPRINT(TR_BF, TR_LEVEL_1, 
			("asking server about page %d, memberCount %d", pid->page,
				pageHash->memberCount));

		SM_ASSERT(LEVEL_1, PPAGE_TYPE(pageHash) == pageType);
		SM_ASSERT(LEVEL_1, pageHash->page2size == page2size);


		/*
		 * Read the page in from disk.
		 */
		switch(io_ReadPageWithNeighbor(pid, pageHash->pageInfo->bufFrame, 
						&lockMode, pageHash, (FLAGS) (nflags | bfflags), 
						fid, neighborPid)){
		case esmAFTERLASTPID:
		case esmBEFOREFIRSTPID:
			if(nflags & NEIGHBOR_FLAGS)
				/* these errors are ok */
		case esmNOERROR:
				break;
		default:
			if(groupLink)
				bf_DiscardPage(bufGroup, groupLink);
			return(NULL);
		}

		if((bfflags & BF_NOREAD) == 0) { /* had to go to server */
			/*
			 *	initialize the firstLSN and firstLRC fields for this buffer
			 */
			pageHash->firstLSN.wrapCount = 0;
			pageHash->firstLSN.offset = 0;
			pageHash->firstLRC.wrapCount = 0;
			pageHash->firstLRC.count = 0;


			/*
			 *	mark the read lock and the write lock if necessary
			 */
			SET_PAGEHASHLOCK(pageHash, lockMode);
			pageHash->releaseCnt = 0;
			pageHash->releaseLimit = LM_RELEASELIMIT;

			/*
			 *	return the groupLink
			 *  In this case, we got groupLink from a getPage, which
			 *  means this page wasn't in the HT already.
			 */
			if(groupLink)
				return(groupLink);
			/* else have to add to the buffer group, etc. */ 
		}
	}

	/*
	 *	make sure we haven't found a chunk page
	 */
	SM_ASSERT(LEVEL_1, PPAGE_TYPE(pageHash)!= PAGE_LARGEDATA);

	/*
	 *	check to see if the type is correct
	 */
	if (!(PPAGE_TYPE(pageHash) == pageType))	{

		SM_ERROR(TYPE_WARNING, esmINTERNAL);
		return(NULL);
	}

	/*
	 *	check to see if this is a slotted page
	 */

	GET_PAGEHASHLOCK(pageHash, pMode);
	switch (pageType)  {
		int suprMode;
	
	case PAGE_SLOTTED: {
			slottedPage = (SLOTTEDPAGE *) pageHash->pageInfo->bufFrame;

			if (CHECK_SLOTTED_MAGIC(slottedPage))	{

				SM_ERROR(TYPE_FATAL, esmINTERNAL);
			}


			/*
			 *	Check to see if a lock is requested
			 */
			suprMode = LM_Supremum[pMode][lockMode];
			if (suprMode != pMode)  {
				if (lm_LockPage(pid, pageHash, suprMode, 
										&(slottedPage->header.fid)))  {
					return NULL;
				}

			}
		} break;
			
	case PAGE_INDEXDESC: {
			INDEXDESCPAGE* indexPage;
			/*
			 *	get a pointer to the page
			 */
			indexPage = (INDEXDESCPAGE*) pageHash->pageInfo->bufFrame;
			
			SM_ASSERT(LEVEL_1, indexPage->magic == NDXMAGIC);
			
			/*
			 *	Check to see if a lock is requested
			 */
			suprMode = LM_Supremum[pMode][lockMode];
			if (suprMode != pMode)  {
				if (lm_LockPage(pid, pageHash, suprMode, NULL))  {
					return NULL;
				}
			}
		} break;

	case PAGE_INDEX: {
			SM_ASSERT(LEVEL_3, (lockMode == SH || 
					lockMode == EX || lockMode == NL));

			suprMode = LM_Supremum[pMode][lockMode];
			if (suprMode != pMode)  {
				if (lm_LockPage(pid, pageHash, suprMode, NULL))  {
					return NULL;
				}

			}
		} break;

	} /* switch on pageType */

	/*
	 *	check to see where it resides in the cache
	 */
	pageState = BF_FindBufGroup(pageHash, bufGroup, &retLink);

	/*
	 *	take action based on status of page
	 */
    switch (pageState) {

		case (BF_FREE): {
			TRPRINT(TR_BF, TR_LEVEL_1, ("page is free"));

			/*
			 *	manage the buffer group space
			 */
			if (bf_Swap(bufGroup, BLOCKCOUNT(page2size)))	{

				return(NULL);
			}

			/*
			 *	get a register pointer to the groupLink
			 */
			groupLink = (GROUPLINK *) pageHash->memberList.succ->item;

			/*
			 *	Take it off the free list and
			 *	put the link onto the fixed list
			 */
			listMoveEnq( &(bufGroup->fixedList), &(groupLink->groupList) );
			groupLink->bufGroup = bufGroup;

			/*
			 *	Set the number of groups the page is in
			 *	and the fixed count for this group
			 */
			pageHash->memberCount = 1;
			groupLink->fixedCount = 1;

			/*
			 *	Increment the number of fixed charges for the buffer group
			 */
			bufGroup->fixedCharges += BLOCKCOUNT(page2size);

		} break;

		
		case (BF_OTHER): {

			TRPRINT(TR_BF, TR_LEVEL_1, ("page is in other group"));

			/*
			 *	manage the buffer group space
			 */
			if (bf_Swap(bufGroup, BLOCKCOUNT(page2size)))	{

				return(NULL);
			}

			/*
			 *	Get a free group link
			 */
			groupLink = (GROUPLINK *) listDeq( &GroupLinkPool );

			/*
			 *	set the fields of the link
			 */
			groupLink->pageHash	  = pageHash;
			groupLink->bufGroup	  = bufGroup;
			groupLink->fixedCount = 1;
			groupLink->bufFrame	  = pageHash->pageInfo->bufFrame;

			/*
			 *	Put the structure on the linked lists
			 */
			listEnq( &(pageHash->memberList), &(groupLink->hashList) );
			listEnq( &(bufGroup->fixedList), &(groupLink->groupList) );

			/*
			 *	Increment the group count in the hash structure
			 */
			pageHash->memberCount++;

			bufGroup->fixedCharges += BLOCKCOUNT(page2size);

		} break;


		case (BF_FIXED): {

			TRPRINT(TR_BF, TR_LEVEL_2, ("page is fixed in group"));

			CHECK_GROUPLINK_MAGIC(retLink);

			/*
			 *	The page is already fixed in the group.
			 *	Simply increment the fix count
			 */
			retLink->fixedCount++;

			/*
			 *	Set the pointer for return
			 */
			groupLink = retLink;

		} break;

		case (BF_UNFIXED): {

			TRPRINT(TR_BF, TR_LEVEL_2, ("page is unfixed in group"));

			groupLink = retLink;

			/*
			 *	Put it in the fixed partition
			 */
			listMoveEnq( &(bufGroup->fixedList), &(groupLink->groupList) );

			/*
			 *	Set the fixed count
			 */
			groupLink->fixedCount = 1;

			/*
			 *	Update the fixed and unfixed charges
			 */
			bufGroup->fixedCharges += BLOCKCOUNT(page2size);
			bufGroup->unfixedCharges -= BLOCKCOUNT(page2size);;

		}	break;
	} /* switch on pageState */

	TRPRINT(TR_BF, TR_LEVEL_2, ("fixed:%d unfixed:%d",
			bufGroup->fixedCharges, bufGroup->unfixedCharges));

	CHECK_GROUPLINK_MAGIC(groupLink);

	return(groupLink);
}
