/*
 *   $RCSfile: bf_DirtyChunk.c,v $  
 *   $Revision: 1.1.1.1 $  
 *   $Date: 1996/05/04 21:55:13 $      
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/


#include <stdio.h>
#include "ess.h"
#include "checking.h"
#include "io.h"
#include "page.h"
#include "list.h"
#include "object.h"
#include "tid.h"
#include "bf.h"
#include "chunk.h"
#include "lgobject.h"
#include "bf_macro.h"
#include "error.h"
#include "trace.h"
#include "lock.h"
#include "volume.h"
#include "bf_intfuncs.h"
#include "io_extfuncs.h"
#include "bf_extfuncs.h"
#include "bf_globals.h"


 void
BF_InvalidateChunkPage (

	register CHUNKPAGE	*chunkPage 
)
{

	register PAGEHASH	*pageHash;
	register CHUNKDESC	*chunkDesc;
	register int		i;


	TRPRINT(TR_BF, TR_LEVEL_1, ("invalidating page:%d", chunkPage->pid.page));

	/*
	 *	get a pointer to the page hash structure
	 */
	pageHash = chunkPage->pageHash;

	/*
	 *	check the pageHash magic
	 */
	CHECK_PAGEHASH_MAGIC(pageHash);

	/*
	 *	Remove the chunk from the list of chunks to which a page belongs.
	 * This will prevent the chunk from being destroyed since 
	 * we are operating	on its buffers.
	 */
	listRemove( &(chunkPage->hashList) );

	/*
	 *	check to see if we are the only chunk using the page
	 */
	if (--(pageHash->memberCount) == 0)	{

		TRPRINT(TR_BF, TR_LEVEL_2, ("single chunk using page"));


		/*
		 *	Check to see if the page is dirty
		 */
		if (PAGE_IS_DIRTY(pageHash))	{

			TRPRINT(TR_BF, TR_LEVEL_2, 
				("flushing page:%d", chunkPage->pid.page));

			/* 
			 * This is like bf_ForcePage(), but it uses
			 * a different bufFrame and an offset:
			 */
			if (io_WritePage( &(pageHash->hashList.hashPid),
			   	chunkPage->chunkDesc->bufFrame + chunkPage->offset,
			    chunkPage->pageHash)){

				SM_ERROR(TYPE_FATAL, esmINTERNAL);
			}

			bf_CleanPage(pageHash);
		}

		BF_DiscardPageHash(pageHash);

	} else {

		TRPRINT(TR_BF, TR_LEVEL_2, ("%d chunks using page", pageHash->memberCount));

		/*
		 *	get rid of all the chunks that reference these pages
		 *	except the chunk that is calling this routine.
		 */
		for (i = pageHash->memberCount; i > 0; i--)	{

			/*
			 *	Check to make sure that the list is not empty
			 */
			SM_ASSERT(LEVEL_3, LIST_NOT_EMPTY( &(pageHash->memberList) ));

			/*
			 *	extract the chunk descriptor pointer
			 */
			chunkDesc = ((CHUNKPAGE *) pageHash->memberList.succ->item)->chunkDesc;

			/*
			 *	Check the chunk descriptor magic
			 */
			CHECK_CHUNKDESC_MAGIC(chunkDesc);

			/*
			 *	if the chunk is fixed we cannot allow this to happen
			 */
			if (chunkDesc->flags & CHUNK_FIXED)	{

				SM_ERROR(TYPE_FATAL, esmINTERNAL);
			}

			/*
			 *	check to see if the block was used
			 */
			if (chunkDesc->flags & CHUNK_USED)	{
				
				TRPRINT(TR_BF, TR_LEVEL_2, ("chunk is used"));

				/*
				 *	invalidate the referenced chunk
				 */
				BF_InvalidateChunk(chunkDesc);

				/*
				 *	take the chunk off the bufgroup
				 */
				listRemove( &(chunkDesc->memberList) );

				/*
				 *	decrement the number of buffers in the group
				 */
				chunkDesc->bufGroup->unfixedCharges -= chunkDesc->bufCount;

				/*
				 *	we must free the block
				 */
				BF_FreeBlock(chunkDesc);
			
			} else {

				TRPRINT(TR_BF, TR_LEVEL_2, ("chunk is valid"));

				/*
				 *	just invalidate the chunk
				 */
				BF_InvalidateChunk(chunkDesc);
			}
		}
	}
}

 int
bf_DirtyChunk (

	int				pageCount,
	CHUNKPAGE		*pageList,
	CHUNKDESC		*chunkDesc 
)
{

	register CHUNKPAGE	*chunkPage;
	register CHUNKPAGE	*newPage;
	int					chunkLength;
	CHUNKPAGE			*lastPage;
	PAGEHASH			*pageHash;
	int					count;


	TRPRINT(TR_BF, TR_LEVEL_1, ("pageCount:%d", pageCount));

	/*
	 * Check the input paramters.
	 */
	SM_ASSERT(LEVEL_3, pageCount >= 0);

	/* 
	 *	setup loop conditions for processing new page list
	 */
	CHECK_CHUNKDESC_MAGIC(chunkDesc);
	chunkLength = 0;
	chunkPage   = chunkDesc->chunkPage;
	lastPage    = chunkPage + (chunkDesc->pageCount - 1);
	newPage     = pageList;
	count		= 0;
	chunkDesc->dirtyPages = pageCount;

	/*
	 *	Invalidate all the pages in the current chunk.  This assumes
	 *	that all pages will have been shadowed.  Create a new page
	 *	list that reflects the new pages that will makeup the chunk.
	 */
	while (chunkPage <= lastPage)	{

		/*
		 *	invalidate the old page if it is valid
		 */
		if (chunkPage->flags & PAGE_VALID)	{
			SM_ASSERT(LEVEL_1, LIST_MEMBER( &(chunkPage->hashList)));

			/*
			 *	mark it invalid in the cache
			 */
			BF_InvalidateChunkPage(chunkPage);
		}
		BF_AuditFreeList();

		/*
		 *	check to see if we have a valid entry
		 */
		if (count < pageCount)	{

			TRPRINT(TR_BF, TR_LEVEL_2, ("processing page:%d", newPage->pid.page));

			/*
			 *	copy the entry to the old list if it is valid
			 */
			*chunkPage = *newPage;

			/*
			 *	set the starting offset of the page in chunk
			 */
			chunkPage->offset = chunkLength;
			 
			/*
			 *	Accumulate the chunk size.
			 */
			chunkLength += chunkPage->size;
			TRPRINT(TR_BF, TR_LEVEL_2, ("length:%d offset:%d pid:%d",
					chunkPage->size, chunkPage->offset,
					chunkPage->pid.page));	

			/*
			 *	Check to see if the page is resident
			 */
			SM_ASSERT(LEVEL_3, (pageHash = bf_FindPage( &(chunkPage->pid) )) == NULL);

			if (( pageHash = BF_NewPageHash() ) == NULL ) {

				SM_ERROR(TYPE_WARNING, esmNOFREEPAGEHASH);
				return(esmFAILURE);
			}

			/*
			 *	Insert the hash into the list
			 */
			BF_InsertHashEntry(pageHash, &(chunkPage->pid), H_CHUNK);

			/*
			 *	dirty the page
			 */
			bf_DirtyPage(pageHash);

			/*
			 *	initialize the page list
			 */
			initializeListElement( &(chunkPage->hashList), (char *) chunkPage);

			/*
			 *	Hang the page list off the hash
			 */
			SM_ASSERT(LEVEL_1, NOT_LIST_MEMBER(&(chunkPage->hashList)));
			listEnq( &(pageHash->memberList), &(chunkPage->hashList) );

			/*
			 *	set the chunkPage fields
			 */
			chunkPage->pageHash  = pageHash;
			chunkPage->chunkDesc = chunkDesc;
			chunkPage->flags     = PAGE_VALID|PAGE_DIRTY;
			SM_ASSERT(LEVEL_1, LIST_MEMBER(&(chunkPage->hashList)));

			/*
			 *	Set the count field in the page hash
			 */
			pageHash->memberCount = 1;
		}

		/*
		 *	look at the next page
		 */
		newPage++;
		chunkPage++;
		count++;
	}
	TRPRINT(TR_BF, TR_LEVEL_2, ("total chunk length:%d", chunkLength));

	/*
	 *	mark the chunk as dirty
	 */
	DIRTY_CHUNK(chunkDesc);

	/*
	 *	update the number of pages in the chunk
	 */
	chunkDesc->pageCount = pageCount;
	BF_AuditChunk(chunkDesc);
	BF_AuditGroup(chunkDesc->bufGroup);

	/*
	 *	return success
	 */
	return(esmNOERROR);
}
