/*
 *   $RCSfile: bf_FlushAll.c,v $  
 *   $Revision: 1.1.1.1 $  
 *   $Date: 1996/05/04 21:55:14 $      
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/


#include <stdio.h>
#include "ess.h"
#include "checking.h"
#include "io.h"
#include "page.h"
#include "list.h"
#include "object.h"
#include "tid.h"
#include "bf.h"
#include "chunk.h"
#include "bf_macro.h"
#include "error.h"
#include "trace.h"
#include "lock.h"
#include "lsn.h"
#include "loginfo.h"
#include "clog.h"
#include "trans.h"
#include "bf_intfuncs.h"
#include "bf_extfuncs.h"
#include "clog_extfuncs.h"
#include "clog_globals.h"
#include "bf_globals.h"
#include "trans_states.h"


/* 
 * CALLED FROM: bf_FlushAll(on commit) and bf_ShutDown. 
 * Goes through free list and invalidates everything on it,
 * which causes dirty pages to be written & removes pages from
 * hash table.  No buffer group accounting goes on since it's
 * operating on free pages.
 */


void
BF_FlushFreeList(BOOL _abort, BOOL skipTemp)
{
    register PAGEHASH   *pageHash;
	register CHUNKDESC	*chunkDesc;
	register GROUPLINK	*groupLink;
	register int i;

	TRACE(TR_BF, TR_LEVEL_1);
	/*
	 *	loop through the free lists
	 */
	for (i = 0; i < NumBufferSizes; i++)	{

		TRPRINT(TR_BF, TR_LEVEL_3, ("processing list:%d", i));
		
		if (LIST_NOT_EMPTY( &FreeBufTable[i] ))	{

			/*
			 *	Flush all pages on the list
			 */
			groupLink = (GROUPLINK *) FIRST_LIST_ELEMENT( &(FreeBufTable[i]) );
			while (groupLink != NULL)	{

				if (groupLink->type == TYPE_PAGE)	{

					CHECK_GROUPLINK_MAGIC(groupLink);

					pageHash = groupLink->pageHash;
					CHECK_PAGEHASH_MAGIC(pageHash);

					TRPRINT(TR_BF, TR_LEVEL_3, ("checking page:%d %d",
							pageHash->hashList.hashPid.page,
							pageHash->hashList.hashPid.volid));

					if(	_abort || 
						(skipTemp && (pageHash->flags & H_TEMP) )) {

						bf_InvalidatePage(pageHash, TRUE, FALSE);
                        /* TRUE-> remove from hash table */
                        /* FALSE-> do not force to server       */
					} else {
						/*
						 * Invalidate but allow the page to remain in the
						 * hash table for reclaims (inter-tx caching).
						 */

						bf_InvalidatePage(pageHash, FALSE, TRUE);
						/* FALSE-> don't remove from hash table */
						/* TRUE-> force to server */
					}
					/*
					 *	Get a pointer to the next element
					 */
					groupLink = (GROUPLINK *) NEXT_LIST_ELEMENT( &(groupLink->groupList) );

				} else if (groupLink->type == TYPE_CHUNK)	{
					chunkDesc = (CHUNKDESC *) groupLink;
					CHECK_CHUNKDESC_MAGIC(chunkDesc);

					if (CHUNK_IS_VALID(chunkDesc))	{
						if(_abort) {
							BF_DiscardChunk(chunkDesc);
						} else {
							BF_InvalidateChunk(chunkDesc);
						}
					}
					groupLink = (GROUPLINK *) 
						NEXT_LIST_ELEMENT( &(chunkDesc->memberList) );
				
				} else {
					SM_ERROR(TYPE_FATAL, esmINTERNAL);
				}
			}
		}
	}
}


/* 
 * CALLED FROM: commitTrans.
 */

 void 
bf_FlushAll ( BOOL discard, BOOL discardTemp) 
{
	BUFGROUP			*bufGroup;
	
	TRACE(TR_BF, TR_LEVEL_1);

	/*
	 *	run through the open buffer group list
	 *	to see if there are any fixed pages
	 */
	bufGroup = (BUFGROUP *) FIRST_LIST_ELEMENT( &(UsedBufGroups) );
	while (bufGroup != NULL)	{

		CHECK_BUFGROUP_MAGIC(bufGroup);
		TRPRINT(TR_BF, TR_LEVEL_2, ("buffer group:%d", bufGroup->index));

		if(discard) {
			bf_UnfixGroup(bufGroup);
		} else {
			/* it's an error to have anything left fixed */
			if (bufGroup->fixedCharges > 0) {
#ifdef DEBUG
				bf_PrintFixedBG(bufGroup);
#endif DEBUG
				SM_ERROR(TYPE_FATAL, esmFIXEDPAGES);
			}
		}
		(void) bf_FreeUnfixedGroup(bufGroup, discard); 

		bufGroup = (BUFGROUP *) NEXT_LIST_ELEMENT( &(bufGroup->groupList) );
	}

	BF_FlushFreeList(discard, discardTemp); 

#ifdef DEBUG
	if(discard) {
		(void) BF_AssertPageHashTableEmpty();
	}
#endif DEBUG
}
