/*
 *   $RCSfile: bf_GetChunkBuf.c,v $  
 *   $Revision: 1.1.1.1 $  
 *   $Date: 1996/05/04 21:55:13 $      
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/


#include <stdio.h>
#include "ess.h"
#include "checking.h"
#include "io.h"
#include "page.h"
#include "list.h"
#include "object.h"
#include "tid.h"
#include "bf.h"
#include "chunk.h"
#include "bf_macro.h"
#include "error.h"
#include "trace.h"
#include "lock.h"
#include "volume.h"
#include "bf_intfuncs.h"
#include "bf_extfuncs.h"
#include "bf_globals.h"

 CHUNKDESC 
*BF_ConvertToChunk (

	register GROUPLINK	*groupLink 
)
{

	register PAGEHASH	*pageHash;
	register CHUNKDESC	*chunkDesc;


	TRACE(TR_BF, TR_LEVEL_1);

	/*
	 *	get a pointer to the pageHash
	 */
	pageHash = groupLink->pageHash;
	CHECK_PAGEHASH_MAGIC(pageHash);

	/* We got this page off the free list.
	 * Is the page in the hash table? If so, get rid of it.
	 * Force it to the server if it is dirty, and invalidate
	 * any locks.
	 * (removeFromHT== TRUE, forceToServer==TRUE)
	 */
	bf_InvalidatePage(pageHash, TRUE, TRUE);

	/*
	 *	get a chunk descriptor off the chunk descriptor pool
	 */
	if ((chunkDesc = (CHUNKDESC *) listDeq( &ChunkDescPool )) == NULL)	{
		return(NULL);
	}
	SM_ASSERT(LEVEL_1, (chunkDesc->chunkPage == NULL));

	/*
	 *	initialize the chunk descriptor
	 */
	CHECK_CHUNKDESC_MAGIC(chunkDesc);
	chunkDesc->bufIndex = pageHash->pageInfo->bufIndex;
	chunkDesc->bufCount = BLOCKCOUNT(pageHash->page2size);
	chunkDesc->buf2size = pageHash->page2size;
	chunkDesc->bufFrame = pageHash->pageInfo->bufFrame;
	chunkDesc->bufInfo  = pageHash->pageInfo->bufInfo;

	/*
	 *	take the grouplink off the free queue and
	 *	put it in the structure pool
	 */
	listEnq( &BufStructPool, &(groupLink->groupList) );

	/*
	 *	mark the new bufInfo information
	 */
	BF_SetSlot(chunkDesc->bufInfo, chunkDesc->bufCount, (char *) chunkDesc, TYPE_CHUNK);

	BF_AuditChunk(chunkDesc);
	/*
	 *	return the chunk descriptor
	 */
	return(chunkDesc);
}

 CHUNKDESC 
*bf_GetChunkBuf (

	BUFGROUP			*bufGroup,
	PAGE2SIZE			block2count 
)
{

	register GROUPLINK	*groupLink;
	register CHUNKDESC	*chunkDesc;


	TRPRINT(TR_BF, TR_LEVEL_1, ("need blocks:%d", RAISETWO(block2count)));

	/*
	 *	adjust the block2count for the pagesize
	 */
	block2count += MIN_PAGE2SIZE;

	/*
	 *	Attempt to get a buffer off of the free list
	 *	of the requested size
	 */
	if (groupLink = (GROUPLINK *) listDeq( &(FreeBufTable[OFFSET(block2count)]) ))	{

		TRPRINT(TR_BF, TR_LEVEL_2, ("got right size buffer"));

		/*
		 *	check to see if we have gotten a chunk
		 */
		if (groupLink->type == TYPE_CHUNK)	{

			/*
			 *	get a pointer to the chunk descriptor
			 */
			chunkDesc = (CHUNKDESC *) groupLink;
			CHECK_CHUNKDESC_MAGIC(chunkDesc);
			TRPRINT(TR_BF, TR_LEVEL_2, ("type chunk index:%d", chunkDesc->bufIndex));

			/*
			 *	flush and invalidate the chunk
			 */
			if (chunkDesc->flags & CHUNK_VALID)	{

				/*
				 *	Invalidate and flush if necessary
				 */
				BF_InvalidateChunk(chunkDesc);
			}

		} else if (groupLink->type == TYPE_PAGE)	{

			TRPRINT(TR_BF, TR_LEVEL_2, ("type page index:%d",
					groupLink->pageHash->pageInfo->bufIndex));

			/*
			 *	convert the group stuff to a chunk
			 */
			if ((chunkDesc = BF_ConvertToChunk(groupLink)) == NULL)	{

				SM_ERROR(TYPE_FATAL, esmINTERNAL);
			}
			BF_AuditChunk(chunkDesc);

		} else {

			/*
			 *	this cannot happen
			 */
			SM_ERROR(TYPE_FATAL, esmINTERNAL);
		}

	} else if (chunkDesc = (CHUNKDESC *) BF_GetLarger(block2count, TYPE_CHUNK))	{

		BF_AuditChunk(chunkDesc);
		/*
		 *	check the magic number
		 */
		CHECK_CHUNKDESC_MAGIC(chunkDesc);

	} else if (chunkDesc = (CHUNKDESC *) BF_GetSmaller(block2count, TYPE_CHUNK))	{

		BF_AuditChunk(chunkDesc);
		/*
		 *	check the magic number
		 */
		CHECK_CHUNKDESC_MAGIC(chunkDesc);

	} else {

		/*
		 *	can't find contiguous free space
		 */
		SM_ERROR(TYPE_WARNING, esmNOBUFSPACE);
		return(NULL);
	}

	/*
	 *	Hang the buffer off the fixed list
	 */
	listEnq( &(bufGroup->fixedList), &(chunkDesc->memberList) );

	/*
	 *	Mark the chunk as used and fixed
	 */
	chunkDesc->flags |= (CHUNK_VALID|CHUNK_USED|CHUNK_FIXED);
	chunkDesc->bufGroup = bufGroup;

	chunkDesc->pageCount = 0;
	chunkDesc->chunkPage = 0;

	/*
	 *	Increment the fixed charges
	 */
	bufGroup->fixedCharges += BLOCKCOUNT(block2count);
	TRPRINT(TR_BF, TR_LEVEL_2, ("fixedCharges:%d", bufGroup->fixedCharges));

	/*
	 * Set the return value and return.
	 */
	BF_AuditChunk(chunkDesc);
	return(chunkDesc);
}
