/*
 *   $RCSfile: bf_InsertChunk.c,v $  
 *   $Revision: 1.1.1.1 $  
 *   $Date: 1996/05/04 21:55:13 $      
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/


#include "sysdefs.h"
#include "ess.h"
#include "checking.h"
#include "io.h"
#include "page.h"
#include "list.h"
#include "object.h"
#include "tid.h"
#include "bf.h"
#include "chunk.h"
#include "bf_macro.h"
#include "error.h"
#include "trace.h"
#include "trans.h"
#include "lock.h"
#include "volume.h"
#include "lsn.h"
#include "loginfo.h"
#include "serverinfo.h"
#include "bf_intfuncs.h"
#include "bf_extfuncs.h"
#include "msg_funcs.h"
#include "bf_globals.h"
#include "clog_globals.h"
#include "sm_globals.h"


 void
BF_RenameChunkPage (

	register CHUNKPAGE	*homePage 
)
{

	register CHUNKPAGE	*chunkPage;
	register CHUNKPAGE	*nextPage;
	register PAGEHASH	*pageHash;
	register CHUNKDESC	*chunkDesc;


	TRPRINT(TR_BF, TR_LEVEL_1, ("page:%d", homePage->pid.page));

	/*
	 *	get a pointer to the page hash structure
	 */
	pageHash = homePage->pageHash;

	/*
	 *	check the pageHash magic
	 */
	CHECK_PAGEHASH_MAGIC(pageHash);

	/*
	 *	check to see if we are the only chunk using the page
	 */
	if (pageHash->memberCount > 1)	{

		TRPRINT(TR_BF, TR_LEVEL_2, ("%d chunks using page", pageHash->memberCount));

		/*
		 *	set the initial value of the chunkpage
		 */
		chunkPage = (CHUNKPAGE *) FIRST_LIST_ELEMENT( &(pageHash->memberList) );

		/*
		 *	get rid of all the chunks that reference these pages
		 */
		while (chunkPage != NULL)	{

			/*
			 *	Check to make sure that the list is not empty
			 */
			SM_ASSERT(LEVEL_3, LIST_NOT_EMPTY( &(pageHash->memberList) ));

			/*
			 *	Save a pointer to the next page
			 */
			nextPage = (CHUNKPAGE *) NEXT_LIST_ELEMENT( &(chunkPage->hashList) );

			/*
			 *	check to see if this is the home page
			 */
			if (chunkPage != homePage)	{

				TRPRINT(TR_BF, TR_LEVEL_2, ("not home page"));

				/*
				 *	extract the chunk descriptor pointer
				 */
				chunkDesc = chunkPage->chunkDesc;

				/*
				 *	Check the chunk descriptor magic
				 */
				CHECK_CHUNKDESC_MAGIC(chunkDesc);

				/*
				 *	if the chunk is fixed we cannot allow this to happen
				 */
				if (chunkDesc->flags & CHUNK_FIXED)	{

					SM_ERROR(TYPE_FATAL, esmINTERNAL);

				} else if (chunkDesc->flags & CHUNK_USED)	{
					
					TRPRINT(TR_BF, TR_LEVEL_2, ("chunk is used"));

					/*
					 *	invalidate the referenced chunk
					 */
					BF_InvalidateChunk(chunkDesc);

					/*
					 *	take the chunk off the bufgroup
					 */
					listRemove( &(chunkDesc->memberList) );

					/*
					 *	decrement the number of buffers in the group
					 */
					chunkDesc->bufGroup->unfixedCharges -= chunkDesc->bufCount;

					/*
					 *	we must free the block
					 */
					BF_FreeBlock(chunkDesc);
				
				} else {

					TRPRINT(TR_BF, TR_LEVEL_2, ("chunk is valid"));

					/*
					 *	just invalidate the chunk
					 */
					BF_InvalidateChunk(chunkDesc);
				}
			}

			/*
			 *	Increment the pointer to the next item
			 */
			chunkPage = nextPage;
		}
	}
}

 int
bf_InsertChunk (

	int					pageCount,
	CHUNKPAGE			*pageList, /* copies contents but does not
									* free this or make a copy of the
									* ptr
									*/
	CHUNKDESC			*chunkDesc,
	int					offset,
	int					length 
)
{

	register CHUNKPAGE	*chunkPage;
	register CHUNKPAGE	*newPage;
	int					chunkLength;
	CHUNKPAGE			*lastPage;
	PAGEHASH			*pageHash;
	char				*sourceChar;
	char				*destChar;
	char				*endChar;
	BOOL				freshPage;

	TRPRINT(TR_BF, TR_LEVEL_1, ("offset:%d length:%d pageCount:%d",
			offset, length, pageCount));

	/*
	 * Check the input paramters.
	 */
	SM_ASSERT(LEVEL_3, (pageCount >= 0) && (offset >= 0));

	/* 
	 *	setup loop conditions for processing new page list
	 */
	chunkLength = 0;
	newPage		= pageList;
	lastPage	= pageList + (pageCount - 1);
	chunkPage	= chunkDesc->chunkPage;
	chunkDesc->dirtyPages = pageCount;

	/*
	 *	Invalidate all the pages in the current chunk.  This assumes
	 *	that all pages will have been shadowed.  Create a new page
	 *	list that reflects the new pages that will makeup the chunk.
	 */
	while (newPage <= lastPage)	{

		/*
		 *	invalidate the old page if it is valid
		 */
		if (chunkPage->flags & PAGE_VALID)	{
			SM_ASSERT(LEVEL_1, LIST_MEMBER( &(chunkPage->hashList)));

			/*
			 *	check to see that pids are the same
			 */
			SM_ASSERT(LEVEL_3, newPage->pid.page == chunkPage->pid.page);

			/*
			 *	mark it invalid in the cache
			 */
			BF_RenameChunkPage(chunkPage);
			BF_AuditFreeList();

			/*
			 *	take the chunkpage off the list
			 */
			listRemove( &(chunkPage->hashList) );

			/*
			 *	set the reuse flag
			 */
			freshPage = FALSE;
			pageHash = chunkPage->pageHash;

		} else {

			/*
			 *	set newPage to be true
			 */
			freshPage = TRUE;
		}

		/*
		 *	copy the entry to the old list
		 */
		*chunkPage = *newPage;

		/*
		 *	set the starting offset of the page in chunk
		 */
		chunkPage->offset = chunkLength;
		 
		/*
		 *	Accumulate the chunk size.
		 */
		chunkLength += chunkPage->size;
		TRPRINT(TR_BF, TR_LEVEL_2, ("length:%d offset:%d pid:%d",
						chunkPage->size, chunkPage->offset, chunkPage->pid.page));	

		/*
		 *	allocate structures if this is a new page
		 */
		if (freshPage)	{

			TRPRINT(TR_BF, TR_LEVEL_2, ("fresh page"));

			/*
			 *	Check to see if the page is resident
			 */
			SM_ASSERT(LEVEL_3, (pageHash = bf_FindPage( &(chunkPage->pid) )) == NULL);

			if (( pageHash = BF_NewPageHash() ) == NULL ) {

				SM_ERROR(TYPE_WARNING, esmNOFREEPAGEHASH);
				return(esmFAILURE);
			}

#ifdef DEBUG
			pageHash->lrc = LrcNone;
#endif DEBUG
			BF_InsertHashEntry(pageHash, &(chunkPage->pid), H_CHUNK);

			SM_ASSERT(LEVEL_1, ((pageHash->lrc.wrapCount != LrcNone.wrapCount)
				|| (pageHash->lrc.count != LrcNone.count)));
			pageHash->page2size = LGDATA_PAGE2SIZE;
			pageHash->flags |= H_NEW;
		}

		/*
		 *	initialize the page list
		 */
		initializeListElement( &(chunkPage->hashList), (char *) chunkPage);

		/*
		 *	Hang the page list off the hash
		 */
		SM_ASSERT(LEVEL_1, NOT_LIST_MEMBER(&(chunkPage->hashList)));
		listEnq( &(pageHash->memberList), &(chunkPage->hashList) );

		/*
		 *	set the chunkPage fields
		 */
		chunkPage->pageHash  = pageHash;
		chunkPage->chunkDesc = chunkDesc;
		chunkPage->flags     = PAGE_VALID|PAGE_DIRTY;
		SM_ASSERT(LEVEL_1, LIST_MEMBER(&(chunkPage->hashList)));

		/*
		 *	Set the count field in the page hash
		 */
		pageHash->memberCount = 1;
		pageHash->page2size = LGDATA_PAGE2SIZE;

		/*
		 *	dirty the page
		 */
		bf_DirtyPage(pageHash);

		/*
		 *	look at the next page
		 */
		newPage++;
		chunkPage++;
	}
	TRPRINT(TR_BF, TR_LEVEL_2, ("total chunk length:%d", chunkLength));

	/*
	 *	move the data to make a hole for the insert if we are not
	 *	appending to the end of the data.
	 */
	if (offset != chunkDesc->totalLength)	{
		
		/*
		 *	setup pointers for copying
		 */
		sourceChar = chunkDesc->bufFrame + (chunkDesc->totalLength - 1);
		destChar   = chunkDesc->bufFrame + ((chunkDesc->totalLength + length) - 1);
		endChar    = chunkDesc->bufFrame + (offset - 1);
		TRPRINT(TR_BF, TR_LEVEL_2, ("source:%d dest:%d end:%d", chunkDesc->totalLength -1,
				(chunkDesc->totalLength + length) - 1, offset - 1));

		/*
		 *	create a hole for insert
		 */
		while (sourceChar > endChar)	{

			*destChar-- = *sourceChar--;
		}
	}

	/*
	 *	mark the chunk as dirty
	 */
	DIRTY_CHUNK(chunkDesc);

	/*
	 *	update the chunk length
	 */
	chunkDesc->totalLength = chunkLength;

	/*
	 *	update the number of pages in the chunk
	 */
	chunkDesc->pageCount = pageCount;

	BF_AuditChunk(chunkDesc);
	BF_AuditGroup(chunkDesc->bufGroup);
	/*
	 *	return success
	 */
	return(esmNOERROR);
}
