/*
 *   $RCSfile: bf_ReadChunk.c,v $  
 *   $Revision: 1.1.1.1 $  
 *   $Date: 1996/05/04 21:55:13 $      
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/


#include "ess.h"
#include "sysdefs.h"
#include "checking.h"
#include "trace.h"
#include "io.h"
#include "page.h"
#include "list.h"
#include "object.h"
#include "tid.h"
#include "bf.h"
#include "chunk.h"
#include "bf_macro.h"
#include "lgobject.h"
#include "error.h"
#include "lock.h"
#include "volume.h"
#include "lsn.h"
#include "loginfo.h"
#include "bf_intfuncs.h"
#include "bf_extfuncs.h"
#include "util_funcs.h"
#include "io_extfuncs.h"
#include "bf_globals.h"
#include "clog_globals.h"

 int
bf_FindChunkSize(

	int				pages 

)
{
	register unsigned	block2count;
	register int		shift;


	TRPRINT(TR_BF, TR_LEVEL_1, ("pages:%d", pages));

	/*
	 *	set the shift to zero
	 */
	shift = 0;
	block2count = BLOCK2COUNT(LGDATA_PAGE2SIZE);
	TRPRINT(TR_BF, TR_LEVEL_2, ("block2count:%d", block2count));

	/*
	 *	loop through until the number is zero
	 */
	while (pages != 0)	{

		/*
		 *	Check to see if there are any less powers
		 */
		if ((pages & 0x1) && (pages ^ 0x1))	{

			TRPRINT(TR_BF, TR_LEVEL_3, ("setting shift"));
			shift = 1;
		}

		/*
		 *	shift the counts
		 */
		pages >>= 1;
		block2count++;
	}

	/*
	 *	shift the number to account for the lesser powers
	 */
	block2count += (shift - 1);

	/*
	 *	return the block count
	 */
	TRPRINT(TR_BF, TR_LEVEL_2, ("block2count:%d", block2count));
	return(block2count);
}

/*============================================================================*/
/* 
 * bf_ReadChunk
 * 	Routine to read a chunk.
 *
 * RETURNS:
 *	esmNOERROR
 *	A user descriptor is set up if everything is sucessful.
 *
 * ERROR RETURNS:
 *	none
 *
 * SIDE EFFECTS:
 *	If the requested chunk is not resident, then a new chunk is
 *	created and the chunk is read in from disk.
 *
 */
/*============================================================================*/


 int
bf_ReadChunk (

	FID					*fid,
	BUFGROUP			*bufGroup,
	int					pageCount,
	int					descCount,
	CHUNKPAGE			*pageList, /* modifies contents but does not free 
									* or make anything else point to it
									* (therefore it can be freed by caller)
									*/
	int					start,	
	int					length,
	CHUNKDESC			**retDesc 
)
{
    register int		i;
    register CHUNKPAGE	*chunkPage;
    int					chunkLength;
    int					charge;
    CHUNKDESC			*chunkDesc;
	CHUNKPAGE			*newList = NULL;
	CHUNKPAGE			*lastPage;
	CHUNKPAGE			*sharePage;
    PAGEHASH			*pageHash;
	BOOL 				goToServer;
	LOCKMODE			lockMode = NL;


	*retDesc = NULL;

	TRPRINT(TR_BF, TR_LEVEL_1, ("pageCount:%d descCount:%d start:%d length:%d",
			pageCount, descCount, start, length));

	BF_AuditGroup(bufGroup); 

    /* 
     *	setup loop conditions for processing the page list.  The
	 *	passed page list contains a list of pids and counts for
	 *	each leaf page that is to be read.
     */
    chunkLength = 0;
    chunkPage = pageList;
	lastPage = pageList + (pageCount - 1);

	/*
	 *	calculate the total length of the chunk
	 */
	while (chunkPage <= lastPage)	{

		/*
		 *	set the starting offset of the page in chunk
		 */
		chunkPage->offset = chunkLength;
		 
		/*
		 * Accumulate the chunk size.
		 */
		chunkLength += chunkPage->size;

		/*
		 *	look at the next element
		 */
		TRPRINT(TR_BF, TR_LEVEL_2, ("length:%d offset:%d pid:%d",
				chunkPage->size, chunkPage->offset, chunkPage->pid.page));	
		chunkPage++;
    }
	TRPRINT(TR_BF, TR_LEVEL_2, ("total chunk length:%d", chunkLength));

	/* 
	 * Calculate the number of buffer frames in the request so
	 * we can figure out how much to charge the group.
	 */
	if (chunkLength == 0) {

		/*
		 *	We have to have one leaf page which is two PAGESIZE buffers
		 */
		charge = BLOCK2COUNT(LGDATA_PAGE2SIZE);	

	} else {

		/*
		 *	account for any hanging off the last page
		 */
		charge = (((lastPage->offset + LGDATA_PAGESIZE) - 1) / LGDATA_PAGESIZE) + 1;

		/*
		 *	convert to PAGESIZE pages
		 */
		charge = bf_FindChunkSize(charge);
	}
	TRPRINT(TR_BF, TR_LEVEL_2, ("charge:%d", charge));

	/*
	 *	if we were passed a positive length then allocate more
	 *	space if necessary.  This is done as a seperate step
	 *	because the amount calculated here might be less
	 *	than the original amount allocated.  This can happen because
	 *	of the above assumption about read overlaps.
	 */
	if (length > 0)	{

		int		newCharge;

		/*
		 *	check to see if the requested length is greater than
		 *	the calculated length
		 */
		SM_ASSERT(LEVEL_3, length > chunkLength);

		/*
		 *	the number of buffers needed
		 *	here we assume that any data that comes after the
		 *	data read from disk will NOT come from disk.  Thus
		 *	we don't have to allow extra slop space for disk
		 *	page clobbering on reads
		 */
		newCharge = (((length - 1) / LGDATA_PAGESIZE) + 1);
		TRPRINT(TR_BF, TR_LEVEL_2, ("new charge:%d", newCharge));
		newCharge = bf_FindChunkSize(newCharge);

		/*
		 *	Check to see if we need a new charge
		 */
		charge = charge > newCharge ? charge : newCharge;
		TRPRINT(TR_BF, TR_LEVEL_2, ("new charge:%d", charge));
	}

	{
		int tries = 0;
again:
		if (bf_Swap(bufGroup, RAISETWO(charge)))	{
			FAIL
		}

		/*
		 *	attempt to allocate the chunk
		 */
		if ((chunkDesc = 
			bf_GetChunkBuf(bufGroup, ((PAGE2SIZE)charge) )) == NULL) {

			tries++;
			/*
			 *	check to see if we can free some on unfixed
			 */
			if (bf_FreeUnfixedGroup(bufGroup, FALSE) == FALSE)	{
				/*  sm_errno was already set by bf_GetChunkBuf: 
				 *	SM_ERROR(TYPE_USER, esmNOBUFSPACE);
				 */
				FAIL
			} else {
				if(tries < 2) goto again;
				/*  sm_errno was already set by bf_Swap:: 
				 *	SM_ERROR(TYPE_USER, esmCANTSWAP);
				 */
				FAIL
			}
		}
	}

	/*
	 *	The chunk desc is now valid, used, and fixed.
	 */
	chunkDesc->chunkPage = NULL;
	chunkDesc->totalLength = chunkLength;
	chunkDesc->pageCount = 0; /* for audit */
	BF_AuditChunk(chunkDesc);
	BF_AuditGroup(bufGroup);

	/*
	 *	check to make sure allocate parameter is consistent
	 */
	SM_ASSERT(LEVEL_3, descCount >= pageCount);

	FAKEFAIL(esmMALLOCFAILED);

	GET(newList,CHUNKPAGE,descCount);

	/*
	 *	copy the page table
	 */
	bcopy((char *) pageList, (char *) newList, ((int)sizeof(CHUNKPAGE) * pageCount));

	/*
	 *	setup the chunk descriptor pointer to the list
	 */
	chunkDesc->chunkPage = newList;
	chunkDesc->pageCount = pageCount;
	chunkDesc->descCount = descCount;
	chunkDesc->fid       = *fid;

	/*
	 *	Initially assume the chunk has no shared pages
	 */
	chunkDesc->sharedPages = 0;

	/*
	 *	setup a new last page
	 */
	lastPage = newList + descCount;

	/*
	 *	process the page list twice: first time make it structurally
	 *  consistent, so if we have an error later on, we can back out.
	 */
   	for (i = 0, chunkPage = newList; chunkPage < lastPage; chunkPage++, i++){
		TRPRINT(TR_BF, TR_LEVEL_2, ("processing page:%d", chunkPage->pid.page));

		initializeListElement( &(chunkPage->hashList), (char *) chunkPage );
		chunkPage->chunkDesc = chunkDesc;
		chunkPage->flags = PAGE_INVALID; /* has no pagehash */
		chunkPage->pageHash = NULL;
	}
	BF_AuditInvalidChunk(chunkDesc);
   	for (i = 0, chunkPage = newList; chunkPage < lastPage; chunkPage++, i++){
		if (i < pageCount)	{
			/*
			 *	Check to see if the page is resident & valid(up-to-date)
			 *  It's valid iff we have a readLock for it.
			 *  If we don't have a read lock, we have to query the server
			 *  to get the page (or be told that our lrc indicates that
			 *  our copy is up-to-date).
			 */
			goToServer = FALSE;
			pageHash = bf_FindPage( &(chunkPage->pid) );

			if( (pageHash==NULL) || (pageHash->flags & H_INVALID)) {
				goToServer = TRUE;
			}
			if (pageHash == NULL) {

				TRPRINT(TR_BF, TR_LEVEL_2, ("hash entry not found"));

				FAKEFAIL(esmNOFREEPAGEHASH);
				if (( pageHash = BF_NewPageHash() ) == NULL ) {
					SM_ERROR(TYPE_USER, esmNOFREEPAGEHASH);
					FAIL
				}

				BF_InsertHashEntry(pageHash, &(chunkPage->pid), H_CHUNK);
				sharePage = NULL;

				pageHash->memberCount = 0; /* about to be incremented 
					when we add it to a memberList, below */
			} else {
				/* 
				 * Make sure the pageHash is what we expect.
				 * The data may still be out-of-date, so we
				 * might still have to go to the server
				 * before doing anything with the contents of the page.
				 */

				TRPRINT(TR_BF, TR_LEVEL_2, ("hash entry found"));

				SM_ASSERT(LEVEL_1, pageHash->flags & H_CHUNK);
				CHECK_PAGEHASH_MAGIC(pageHash);

				sharePage = (CHUNKPAGE *) pageHash->memberList.succ->item;
				SM_ASSERT(LEVEL_3, sharePage->pid.page == chunkPage->pid.page);
				SM_ASSERT(LEVEL_3, sharePage->size == chunkPage->size);

			}
			chunkPage->flags = PAGE_VALID;
			if (chunkPage->shared) chunkDesc->sharedPages++;

			/*
			 * include this page in the list of pages for this chunk
			 * need to assert that it's not already there 
			 */
			SM_ASSERT(LEVEL_1, NOT_LIST_MEMBER(&(chunkPage->hashList)));
			listEnq( &(pageHash->memberList), &(chunkPage->hashList) );
			SM_ASSERT(LEVEL_1, LIST_MEMBER(&(chunkPage->hashList)));

			pageHash->memberCount++;
			pageHash->page2size = LGDATA_PAGE2SIZE;
			TRPRINT(TR_BF, TR_LEVEL_2, 
				("memberCount:%d", pageHash->memberCount));

			/* 
			 *  Now the page may have been there from an old tx,
			 *  but the chunkPage still has no ptr to it, so make sure
			 *  that the chunkPage knows about this page.
			 */
			chunkPage->pageHash = pageHash;
			BF_AuditChunkPage(chunkPage);

			if (goToServer) {

				/*
				 *	The read must be relative to the chunkDesc->firstByte,
				 *	Only do the read if the data count is greater than
				 *	zero.  If the count is zero, there are no valid
				 *	data to retrieve.
				 */
				if (chunkPage->size > 0)	{

					TRPRINT(TR_BF, TR_LEVEL_2, 
						("getting page:%d in location:%x",
						chunkPage->pid.page, 
						chunkDesc->bufFrame + chunkPage->offset));

					/* Until we get inter-tx page caching working
					 * for chunks, we must set re-initialize the lrc
					 * so the server will always send us the page.
					 * The reason we are not getting caching is this:
					 * io_ReadPage is overloaded to do 2 things -- read the page if necessary,
					 * AND stash it into whatever location we give.  In the chunk-case, this
					 * is probably NOT a buffer, and probably doesn't match the location
					 * in which our up-to-date page resides.  In order to itx-caching
					 * work for chunks, we'd have to recognize this case and do a bcopy.
					 */
					pageHash->lrc = LrcNone;

					/* 
					 * the destination for this data might not be aligned
					 * but somewhere in the lg object code (prior to this)
					 * there was allocated a large enough contiguous block
					 * (set of pages) so that this will work- possibly
					 * writing into the next buffer.
					 */
					if (io_ReadPage( &(chunkPage->pid),
						 chunkDesc->bufFrame + chunkPage->offset,
						 &lockMode /* no locks */, pageHash))	{

					    SM_ERROR(TYPE_CRASH, esmINTERNAL);
					}

				} /* if (chunkPage->size > 0)  */
			} else {
				/* no need to go to the server */
				SM_ASSERT(LEVEL_1, sharePage != NULL);

				/*
				 *	copy the data out of the page
				 */
				TRPRINT(TR_BF, TR_LEVEL_2, 
					("copying data offset:%d size:%d from:%x to:%x bufIndex:%d",
						sharePage->offset, chunkPage->size,
						sharePage->chunkDesc->bufFrame + sharePage->offset,
					    chunkDesc->bufFrame + chunkPage->offset, 
						chunkDesc->bufIndex));
				bcopy(sharePage->chunkDesc->bufFrame + sharePage->offset,
					  chunkDesc->bufFrame + chunkPage->offset,
					  chunkPage->size);

			}
		} 
	}
	BF_AuditGroup(bufGroup); 
	BF_AuditChunk(chunkDesc);
	/*
	 *	increment the offset to point at start
	 */
	chunkDesc->basePtr = chunkDesc->bufFrame + start;	
	TRPRINT(TR_BF, TR_LEVEL_2, ("returning basePtr:%x", chunkDesc->basePtr));

	chunkDesc->firstOffset = start;

	/*
	 *	There cannot be more shared pages than pages in the chunk
	 */
	SM_ASSERT(LEVEL_3, chunkDesc->pageCount >= chunkDesc->sharedPages);
	/*
	 *	setup the return descriptor
	 */
	BF_AuditChunk(chunkDesc); 
	BF_AuditGroup(chunkDesc->bufGroup); 
	*retDesc = chunkDesc;
	return(esmNOERROR);

FailureReturn:
	/* when this chunkDesc gets invalidated, the chunkPage (newList)
	 * will be freed
	 */
	return esmFAILURE;
}
