/*
 *   $RCSfile: bf_ReadPage.c,v $  
 *   $Revision: 1.1.1.1 $  
 *   $Date: 1996/05/04 21:55:12 $      
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/


#include <stdio.h>
#include "ess.h"
#include "checking.h"
#include "io.h"
#include "page.h"
#include "list.h"
#include "object.h"
#include "tid.h"
#include "bf.h"
#include "chunk.h"
#include "bf_macro.h"
#include "sm_macro.h"
#include "trace.h"
#include "error.h"
#include "lock.h"
#include "volume.h"
#include "trans.h"
#include "lsn.h"
#include "loginfo.h"
#include "clog.h"
#include "lm.h"
#include "scan.h"
#include "bf_intfuncs.h"
#include "bf_extfuncs.h"
#include "io_extfuncs.h"
#include "lm_extfuncs.h"
#include "bf_globals.h"
#include "trans_globals.h"
#include "clog_globals.h"
#include "lock_globals.h"
#include "sm_globals.h"

 GROUPLINK 
*bf_ReadPage (

	BUFGROUP			*bufGroup,
	const	PID  	    *pid,
	PAGETYPE			pageType,
	PAGE2SIZE			page2size,
	LOCKMODE			lockMode 
)
{
	return
		BF_ReadPageWithNeighbor(bufGroup, pid, pageType, page2size, 
			lockMode, 0 /* flags */, NULL, NULL);
}


/*
 * Bf_CheckForSpace checks to see if buffer space is available
 * for an object operation.  If the operation could possibly
 * involve a large object, then "large" should be set to TRUE.
 */

 int
bf_CheckForSpace(
	BUFGROUP		*bufGroup,
	int				length,
	BOOL			large
)
{
	int				blocks = 0;
	int				lgPow2 = 0;
	int				lgBlocks = 0;
	int				size = 0;
	BOOL			spaceAvailable = FALSE;

	if (large || (length > LARGETHRESHOLD)) {
		length = bf_BreakdownChunk(bufGroup, length);
		/* operation could involve header, 2 lg obj interior nodes */
		blocks = BLOCKCOUNT(SLOTTED_PAGE2SIZE) + BLOCKCOUNT(LG_PAGE2SIZE);

		/* plus, contiguous space for the data */
		lgPow2 = (length / LGDATA_USABLESIZE) + 1;
		lgPow2 = bf_FindChunkSize(lgPow2);
		lgBlocks = RAISETWO(lgPow2);
	} else {
		/* operation could involve header and forwaded page */
		blocks = 2 * BLOCKCOUNT(SLOTTED_PAGE2SIZE);
	}

	if (bf_Swap(bufGroup,blocks+lgBlocks) != esmNOERROR) {
		return(esmFAILURE);
	}

	{
		int tries = 0;
again:

		/* make sure we have contiguous large object space if needed */
		if (lgBlocks > 0) {

			tries++;
			
			/* 
			 *	First see if there are any entries on the free spaces
			 *	lists for sizes large enough for the chunk
			 */
			spaceAvailable = FALSE;
			for (size = lgPow2; size <= OFFSET(MaxPage2Size); size++) {
				if (!LIST_EMPTY(&FreeBufTable[size])) {
					spaceAvailable = TRUE;
					break;
				}
			}

			/*
			 *	If contig space is not immediately available, then see
			 *	if it can be made by coalescing smaller fragments. 
			 */
			if (!spaceAvailable && (BF_FindFreeSlot(lgPow2+MIN_PAGE2SIZE) == NULL)) {
				/*
				 * can't find contiguous free space
				 * Try removing all unfixed pages from the group.
				 */
				if (bf_FreeUnfixedGroup(bufGroup, FALSE) == TRUE)  {
					/* freed something */
					/* see if we should try again */
					if(tries < 2) goto again; 
				}
				spaceAvailable = FALSE;
				SM_ERROR(TYPE_USER, esmNOBUFSPACE);
				return(esmFAILURE);
			}
		}
	}

	return(esmNOERROR);
}



/*
 * Bf_BreakdownChunk() returns a "proper" length for calls to
 * append, delete, and insert into large objects. The returned length
 * is the smaller of "length", maxLgOpChunk, and 1/4 of the buffer
 * group size.  1/4 of the buffer group was chosen to avoid using
 * too much of the group when it is small.
 */
 int
bf_BreakdownChunk(
	BUFGROUP*	bufGroup,
	int			length
)
{
	int chunkLen = MIN(MaxLgOpChunk,
					   (MIN_PAGESIZE * bufGroup->maxCharges) / 4);

	return MIN(chunkLen, length);
}
