/*
 *   $RCSfile: bf_ReplaceLeafChunk.c,v $  
 *   $Revision: 1.1.1.1 $  
 *   $Date: 1996/05/04 21:55:14 $      
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/


#include "sysdefs.h"
#include "ess.h"
#include "trace.h"
#include "checking.h"
#include "error.h"
#include "io.h"
#include "page.h"
#include "list.h"
#include "object.h"
#include "tid.h"
#include "bf.h"
#include "chunk.h"
#include "volume.h"
#include "lsn.h"
#include "loginfo.h"
#include "bf_macro.h"
#include "sm_macro.h"
#include "lock.h"
#include "serverinfo.h"
#include "bf_extfuncs.h"
#include "bf_intfuncs.h"
#include "msg_funcs.h"
#include "bf_globals.h"
#include "clog_globals.h"
#include "sm_globals.h"


/*
 *	Routine to replace a leaf node (data page) in a chunk.
 */


 int
bf_ReplaceLeafChunk (
	CHUNKDESC	*chunkDesc,
	CHUNKPAGE	*chunkPage,		/* page to replace				*/
	PID			*newPid,		/* new page						*/
	PID			*newParentPid 	/* parent of newPid (if parent new)	*/ 
)
{
	CHUNKPAGE	*sharedPage;	/* pages in other chunks for chunkPage*/
	CHUNKPAGE	*nextPage;		/* pages in other chunks for chunkPage*/
	PAGEHASH	*newPageHash;	/* new hash entry for newPid		*/
	PAGEHASH	*chunkPageHash;	/* hash entry for chunkPage			*/
#ifdef DEBUG
	int			totalPages;		/* number of copies of chunkPage	*/
#endif DEBUG

	TRPRINT(TR_LG, TR_LEVEL_1, ("replace page %d,%d", newPid->page,
								newPid->volid));

	chunkPageHash = chunkPage->pageHash;

#ifdef DEBUG
	/*
	 *	Remember the total page count for verification later
	 */
	totalPages = chunkPageHash->memberCount;
#endif DEBUG

	{
#ifdef DEBUG
		/*
		 *	New page should not already be in the hash table
		 */
		SM_ASSERT(LEVEL_1, bf_FindPage(newPid) == NULL);
#endif
		if (( newPageHash = BF_NewPageHash() ) == NULL ) {

			SM_ERROR(TYPE_WARNING, esmNOFREEPAGEHASH);
			return(esmFAILURE);
		}

		/*
		 *  Insert the hash into the table
		 */
		BF_InsertHashEntry(newPageHash, newPid, H_CHUNK);

		newPageHash->memberCount = 0;
		newPageHash->page2size = LGDATA_PAGE2SIZE;
	}

	/*
	 *	Get pointer to first page on list of chunks containing 
	 *	this page.
	 */
	sharedPage = (CHUNKPAGE *) chunkPageHash->memberList.succ->item;

	/*
	 *  check to make sure the pids match
	 */
	SM_ASSERT(LEVEL_3, sharedPage->pid.page == chunkPage->pid.page);

	/*
	 *  check to make sure the sizes are the same
	 */
	SM_ASSERT(LEVEL_3, sharedPage->size == chunkPage->size);

	/*
	 *	For each chunk page belonging to the object represented by
	 *	chunkPage->chunkDesc, rename the page and move it to the 
	 *	new hash entry.
	 */
	while (sharedPage != NULL) {
		/*
		 *	Save a pointer to the next page
		 */
		nextPage = (CHUNKPAGE *) NEXT_LIST_ELEMENT (&(sharedPage->hashList) );
		/*
		 *	Make sure object id for this page matches that of 
		 *	chunkPage.  This is true if either the chunk descriptors
		 *	are the same or if the root page id's are the same since 
		 *	they are unique to each object.  The chunk descs are 
		 *	checked first, since if they have root pid info only
		 *	if they were set up by lg_ReadObject()
		 */
		if ( (sharedPage->chunkDesc == chunkDesc) || 
			 (PIDEQ(sharedPage->chunkDesc->rootPid, 
				  chunkPage->chunkDesc->rootPid) &&
			 (sharedPage->chunkDesc->rootSlot == 
			  chunkPage->chunkDesc->rootSlot)) ) {

			/*
			 *	Remove the page from its current hash entry and decrement
			 *	the number of pages on this entry.
			 */
			listRemove ( &(sharedPage->hashList) );
			chunkPageHash->memberCount--;

			/*
			 *  Hang the page list off the hash
			 */
			listEnq( &(newPageHash->memberList), &(sharedPage->hashList) );

			/*
			 *  set the pointer in the list
			 *	Record the changes to the chunk entry
			 */
			sharedPage->pageHash = newPageHash;
			sharedPage->pid = *newPid;

			/*
			 *  Set the count field in the page hash
			 */
			newPageHash->memberCount++;

			/*
			 *	If there is a new parent page, record it
			 */
			if (newParentPid != NULL) {
				sharedPage->parentPid = newParentPid->page;
			}
		}

		/*
		 *	Advance to next page
		 */
		sharedPage = nextPage;

	}

	/*
	 *	If the chunk page hash entry is empty then remove it
	 */
	if (chunkPageHash->memberCount == 0) {
		/*
		 *	The entry has no members, so it should be empty
		 */
		SM_ASSERT( LEVEL_3, LIST_EMPTY(&(chunkPageHash->memberList)) );

		BF_DiscardPageHash(newPageHash);
	}

	/*
	 *	Make sure all pages are accounted for
	 */
#ifdef DEBUG
	SM_ASSERT(LEVEL_1, totalPages == chunkPageHash->memberCount +
									 newPageHash->memberCount) ;
#endif DEBUG

	BF_AuditChunk(chunkDesc);
	BF_AuditGroup(chunkDesc->bufGroup);
	/*
	 *	Finished
	 */
 	return (esmNOERROR);
}
