/*
 *   $RCSfile: bf_Swap.c,v $  
 *   $Revision: 1.1.1.1 $  
 *   $Date: 1996/05/04 21:55:12 $      
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/


#include <stdio.h>
#include "ess.h"
#include "checking.h"
#include "sysdefs.h"
#include "io.h"
#include "page.h"
#include "list.h"
#include "object.h"
#include "tid.h"
#include "bf.h"
#include "chunk.h"
#include "bf_macro.h"
#include "trace.h"
#include "error.h"
#include "lock.h"
#include "bf_intfuncs.h"
#include "bf_extfuncs.h"


/*============================================================================*/
/* 
 * bf_Swap:
 * 	This routine swaps out unfixed buffer regions from a group's
 *	unfixed list until the group size is decreased enough to
 *	handle a new buffer region of the specified size.
 *
 * RETURNS:
 *	esmNOERROR
 *
 * ERROR RETURNS:
 *	ebfCANTSWAP      - if there isn't enough room in the group
 *
 * SIDE EFFECTS:
 *	Unfixed regions are moved to the free list until the group's
 *	size is decreased enough to handle the new region.
 *
 *
 * The use of all the globals below should be clear except ReservedBufs.
 * ReservedBufs reflects the number of buffers that have been 
 * reserved by bf_OpenGroup() calls.
 * When a group is opened, a specified number of buffers is reserved 
 * for that group, so that when buffers are requested for the group,
 * there should be enough free buffers to satisfy the request.
 * If ReservedBufs is smaller than the number requested in an
 * bf_OpenGroup() call, then bf_OpenGroup() returns an error.
 */
/*============================================================================*/


 int
bf_Swap (
	register BUFGROUP	*bufGroup,
	int	   				charge 		/* # of min_size pages needed */
)
{

    register int		eventualFixedCharges;
    register int		maxCharges;	
	register LIST		*element;


	TRPRINT(TR_BF, TR_LEVEL_1, ("requested charge:%d", charge));

    /*
     *	Make sure the group is valid
     */
	CHECK_BUFGROUP_MAGIC(bufGroup);

	/*
	 *	get maximum charges into register
	 */
	maxCharges = bufGroup->maxCharges;

	/*
	 *	calculate the charges that will result from this request
	 */
    eventualFixedCharges = bufGroup->fixedCharges + charge;
	TRPRINT(TR_BF, TR_LEVEL_2, ("fixed:%d unfixed:%d max:%d eventual fixed:%d",
			bufGroup->fixedCharges, bufGroup->unfixedCharges,
			maxCharges, eventualFixedCharges));

#ifdef DEBUG
	if( (bufGroup->flags & LOG_GROUP)==0) {
		/* don't fake a failure on the log buffer group- none
		 * of the code can handle that anyway
		 */
		FAKEFAIL(esmCANTSWAP);
	}
#endif DEBUG

    /*
     * See if the group can handle the new region without a swap.
     */
    if (eventualFixedCharges + bufGroup->unfixedCharges <= maxCharges) {

		/*
		 *	Page can be accomodated without replacement
		 */
		TRPRINT(TR_BF, TR_LEVEL_2, ("can accomodate charge w/o replace"));
        return(esmNOERROR);
    }

    /*
     *	Looks like we have to do a swap.
     *	Check to see if there any hope of getting enough space.
     */
    if (eventualFixedCharges > maxCharges) {

		/*
		 *	print error message and return
		 */
		SM_ERROR(TYPE_WARNING, esmCANTSWAP);
		return(esmFAILURE);
    }

    /*
     *	At this point, we have determined that there are enough unfixed
     *	regions to swap out; we loop until we generate enough space.
     */
    while (eventualFixedCharges + bufGroup->unfixedCharges > maxCharges) {

		/*
		 *	take the next element off the list
		 */
		element = (LIST *) listDeq( &(bufGroup->unfixedList) );

		/*
		 *	check to make sure we got an element
		 */
		SM_ASSERT(LEVEL_3, element != NULL);

		/*
		 *	take action based on whether it is a
		 *	page or a chunk
		 */
		switch ( *((short *) element) )	{

			case TYPE_PAGE:

				/*
				 *	process the page
				 */
				BF_FreePage(bufGroup, (GROUPLINK *) element, FALSE);
				break;

			case TYPE_CHUNK:

				/*
				 *	process the chunk
				 */
				BF_FreeChunk(bufGroup, (CHUNKDESC *) element);
				break;

			default:

				/*
				 *	this cannot happen
				 */
				SM_ERROR(TYPE_FATAL, esmINTERNAL);
				break;
		}
    }
	BF_AuditGroup(bufGroup);

	/*
	 *	We can now get a buffer for the group
	 */
    return(esmNOERROR);
}
