/*
 * $RCSfile: BTREEPAGE.C,v $
 * $Revision: 1.1.1.1 $
 * $Date: 1996/05/04 21:55:16 $
 */
/**********************************************************************
 * EXODUS Database Toolkit Software
 * Copyright (c) 1991 Computer Sciences Department, University of
 *                    Wisconsin -- Madison
 * All Rights Reserved.
 *
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
 * MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
 * THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
 * WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * The EXODUS Project Group requests users of this software to return 
 * any improvements or extensions that they make to:
 *
 *   EXODUS Project Group 
 *     c/o David J. DeWitt and Michael J. Carey
 *   Computer Sciences Department
 *   University of Wisconsin -- Madison
 *   Madison, WI 53706
 *
 *	 or exodus@cs.wisc.edu
 *
 * In addition, the EXODUS Project Group requests that users grant the 
 * Computer Sciences Department rights to redistribute these changes.
 **********************************************************************/
#define BTREEPAGE_C

#include "BTREEPAGE.h"
#include "OVPAGE.h"
#include "TempBufSpace.h"


int BT_MAXKEYLEN = MAXKEYLEN;

void BTREEPAGE::Linkup(BTREEPAGE* rightPage, BTREEPAGE*
		       rightRightPage) 
{
    rightPage->btCtrl.prev = SelfID().page;
    rightPage->btCtrl.next = btCtrl.next;
    btCtrl.next = rightPage->SelfID().page;
    if (rightRightPage)  {
	rightRightPage->btCtrl.prev = 
	    rightPage->SelfID().page;
    }
}
    

void BTREEPAGE::Init(const PID& pid, int level, SMDATATYPE keyType) 
{
    btCtrl.endData = btCtrl.numOffsets = 0;
    btCtrl.numFree = BT_DATASIZE;
    btCtrl.selfID = pid;
    btCtrl.level = level;
    btCtrl.type = keyType;
    btCtrl.next = btCtrl.prev = btCtrl.pid0 = NULLPID;
}


void BTREEPAGE::CopyData(BTREEPAGE* target)	
{
    BCOPY(data, target->data, BT_DATASIZE);
    target->btCtrl.endData = btCtrl.endData;
    target->btCtrl.numOffsets = btCtrl.numOffsets;
    target->btCtrl.numFree = btCtrl.numFree;
    target->btCtrl.pid0 = btCtrl.pid0;
}

//
//	Class		: BTREEPAGE
//	Method		: int Search(int keyLen, void* keyVal, PFC compFunc
//				TWO& retSlot)
//	Description	: Searches the page for a key, and returns its slot in 
//				  retSlot if found.
//	Return		: TRUE if found, FALSE otherwise
//
BTREEPAGE::Search(int keyLen, const void* keyVal, PFC compFunc, TWO& retSlot)
{
    int low, high;
    register mid;
    for (low = mid = 0, high = NumSlots() - 1; low <= high;  )	{
	mid = (low + high) / 2;
	register diff = compFunc(keyLen, keyVal, Tuple(mid).KeyLen(), 
				 Tuple(mid).KeyValue());
	if ( ! diff )	{
	    // 	---- key found !
	    retSlot = mid;
	    return TRUE;
	}
	if ( diff > 0)
	    low = mid + 1;
	else
	    high = mid - 1;
    }
    
    retSlot = (low > mid) ? low : mid;
    
    return FALSE;
}




//
//	Class		: BTREEPAGE
//	Method		: void ShiftTuple(int slotNum, BTREEPAGE* destPage,
//						int destSlot, int numTuples)
//	Description	: Move numTuples from slotNum in this page to destSlot
//			  in destPage.
//
void BTREEPAGE::ShiftTuple(int slotNum, BTREEPAGE* destPage, 
			   int destSlot, int numTuples)
{
    //
    //	Assert that we shift at least 1 tuple
    //
    ASSERT3((numTuples > 0));
    
    //
    //	Assert that the 'destSlot' parameter is sane
    //
    ASSERT3((destSlot <= destPage->NumSlots()));
    
    //
    //	Calculate total # of bytes needed to move
    //
    int byteCnt = 0;
    for (register i = slotNum; i < slotNum + numTuples; i++)  {
	byteCnt += Tuple(i).Size() + sizeof(SLOT);
    }
    
    //
    //	Compress destination page if necessary
    //
    if (destPage->UsableSpace() < byteCnt) 
	destPage->Compress();
    
    ASSERT1(destPage->UsableSpace() >= byteCnt);
    
    //
    //	Make room in slot[] of destination page
    //
    register SLOT* p = &(destPage->slot[-(destPage->NumSlots()-1+numTuples)]);
    SLOT* ep = &(destPage->slot[-destSlot+1]);
    for ( ; p + numTuples != ep; p++) 		*p = *(p + numTuples);
    
    //
    //	Move the entries
    //
    for (i = 0; i < numTuples; i++)  {
	destPage->slot[-(i+destSlot)] = destPage->btCtrl.endData;
	BT_Tuple* tp = &Tuple(i+slotNum);
	int	entryLen = tp->Size();
	BCOPY((char*)tp, destPage->data + destPage->btCtrl.endData, entryLen);
	destPage->btCtrl.endData += entryLen;
    }
    
    //
    //	Close gap in slot[] of source page
    //
    p = & slot[-slotNum];
    ep = & slot[-NumSlots()];
    for ( ; p - numTuples != ep; p--)		*p = * (p - numTuples);
    
    //
    //	Update Control Info of both pages
    //
    destPage->btCtrl.numFree -= byteCnt;
    destPage->btCtrl.numOffsets += numTuples;
    
    btCtrl.numFree += byteCnt;
    btCtrl.numOffsets -= numTuples;
    
    ASSERT3((CheckContent()));
    ASSERT3((destPage->CheckContent()));
}




//
//	Class		: BTREEPAGE
//	Method		: void InsertTuple(BT_Tuple& tuple, int slotNum)
//	Description	: Insert 'tuple' into slot[slotNum] in this page
//
void BTREEPAGE::InsertTuple(BT_Tuple& tuple, int slotNum) 
{
    int 	length = tuple.Size();
    
    ASSERT3(slotNum <= NumSlots());
    
    //
    //	Assert right kind of tuple for right kind of page
    //
    ASSERT3((tuple.IsNodeTuple() && IsNode()) || 
	    (tuple.IsLeafTuple() && IsLeaf()));
    
    //
    //	Compress the page if there is not enough room at the end
    //
    if (UsableSpace() < length + sizeof(SLOT))  {
	Compress();
    }
    
    ASSERT1(UsableSpace() >= length + sizeof(SLOT));
    
    //
    //	Make room in slot array for the new entry
    //
    register SLOT* p = & slot[- btCtrl.numOffsets];
    SLOT* ep = & slot [- slotNum];
    for ( ; p != ep; p++)  *p = * (p+1);
    
    //
    //	Append the tuple to end of the data area and update control info
    //
    slot[-slotNum] = btCtrl.endData;
    BCOPY((char*) &tuple, data + btCtrl.endData, length);
    btCtrl.endData += length;
    btCtrl.numFree -= length + sizeof(SLOT);
    btCtrl.numOffsets++;
    
    ASSERT3(CheckContent());
}


//
//	Class		: BTREEPAGE
//	Method		: void RemoveTuple(int slotNum)
//	Description	: Remove the tuple in slot[slotNum] in this page
//
void BTREEPAGE::RemoveTuple(int slotNum) 
{
    ASSERT3(slotNum >= 0 && slotNum < NumSlots());
    
    //
    //	Increment numFree before we remove the tuple
    //
    btCtrl.numFree += Tuple(slotNum).Size() + sizeof(SLOT);	
    
    //
    // 	Close gap in the slot array 
    //
    register SLOT* p = & slot[-slotNum];
    SLOT* ep = & slot[- btCtrl.numOffsets];
    for ( ; p != ep; p--)		*p = *(p-1);
    
    btCtrl.numOffsets--;
    
    ASSERT3((CheckContent()));
}







//
//	Class		: BTREEPAGE
//	Method		: int AddElem(int sNum, void* const elem)
//	Description	: Add an element to the tuple in slot[sNum]. 
//				  This implementation
//				  should be improved further for efficiency.
//
int BTREEPAGE::AddElem(int sNum, const void* elem)
{
    ASSERT3(IsLeaf());
    ASSERT3(sNum < NumSlots());
    
    //
    //	Copy tuple into a scratch area to do the update
    //
#ifdef SERVER_MAKE
    // too big to put on stack
    TempBufSpace    tempSpace(sizeof(BT_Tuple));
    BT_Tuple&	    scratch = *(BT_Tuple*)tempSpace.location();
#else
    BT_Tuple scratch;   // scratch space
#endif /* SERVER_MAKE */
    int oldLength = Tuple(sNum).Size();	
    BCOPY((char*)&Tuple(sNum), (char*)&scratch, oldLength);
    
    ASSERT3( ! scratch.IsOverflow());
    ASSERT3( ! scratch.WillOverflow());
    
    if (scratch.AddElem(elem))
	return esmFAILURE;
    
    int newLength = scratch.Size();
    
    ASSERT3(newLength > oldLength);
    
    //
    //	See if we can put scratch at the end of data area
    //
    if (newLength < UsableSpace())   {
	slot[-sNum] = btCtrl.endData;
	BCOPY((char*)&scratch, data + slot[-sNum], newLength);
	
	btCtrl.numFree += oldLength;
    }
    else {
	//
	//	Compress to free up some space for scratch
	//
	slot[-sNum] = EMPTYSLOT;		// free old tuple but not its slot
	btCtrl.numFree += oldLength;
	
	Compress();
	
	ASSERT1(newLength <= UsableSpace());
	
	BCOPY((char*)&scratch, (char*)data + btCtrl.endData, newLength);
	slot[-sNum] = btCtrl.endData;
    }
    btCtrl.endData += newLength;
    btCtrl.numFree -= newLength;
    
    ASSERT3((CheckContent()));
    return esmNOERROR;
}






//
//	Class		: BTREEPAGE
//	Method		: int RemoveElem(int sNum, void* const elem)
//	Description	: Delete an element from the tuple at slot[sNum]
//
int BTREEPAGE::RemoveElem(int sNum, const void* elem)
{
    ASSERT3(IsLeaf());
    
    ASSERT3(sNum < NumSlots());
    
    BT_Tuple* p = & Tuple(sNum);
    int oldLength = p->Size();
    
    ASSERT3( ! p->IsOverflow());
    
    if (p->RemoveElem(elem))
	return esmFAILURE;
    
    btCtrl.numFree += oldLength - p->Size();
    
    ASSERT3((CheckContent()));
    
    return esmNOERROR;
}



//
//	Class		: BTREEPAGE
//	Method		: void Compress()
//	Description	: Compress the page to free up fragmented spaces
//
void BTREEPAGE::Compress()
{
    //
    //	Copy btPage onto scratch
    //
#ifdef SERVER_MAKE
    // too big to put on stack
    TempBufSpace    tempSpace(sizeof(BTREEPAGE));
    BTREEPAGE&	    scratch = *(BTREEPAGE*)tempSpace.location();
#else
    BTREEPAGE scratch;   // scratch space
#endif /* SERVER_MAKE */
    BCOPY((char*) this, (char*) &scratch, sizeof(BTREEPAGE));
    
    //
    //	Loop through tuples in scratch and copy them back to btPage
    //
    char* p = data;
    for (register i = 0; i < scratch.btCtrl.numOffsets; i++)  {
	if (scratch.slot[-i] == EMPTYSLOT)	continue;
	int length = scratch.Tuple(i).Size();
	BCOPY((char*)&scratch.Tuple(i), p, length);	;
	slot[-i] = p-data;
	
	p += length;
    }
    
    //
    // after compression, UsableSpace == FreeSpace 
    //
    btCtrl.endData = p - data;
    btCtrl.numFree = UsableSpace();	 
    
    ASSERT3((CheckContent()));
}


//
//	Class		: BTREEPAGE
//	Method		: int IsSafe(BT_OPER oper, int maxKeyLen)
//	Description	: Returns TRUE if this page is "safe", FALSE otherwise.
//
int BTREEPAGE::IsSafe(BT_OPER oper, int maxKeyLen)
{
    int retVal = FALSE;
    BT_Tuple* p = 0;
    
    switch (oper)  {
    case BT_READ: 		
	retVal = TRUE;
	break;
	
    case BT_INSERT:		
	retVal = FreeSpace() >= p->PredictSize(maxKeyLen);
	break;
	
    case BT_DELETE:		
	retVal = UsedSpace() - p->PredictSize(maxKeyLen) >= 
	    (BT_DATASIZE >> 1);
	break;
    case BT_INSERT_DELETE:
    default:
				// Illegal parameter
	ASSERT1(FALSE);
    }
    
    return retVal;
}




//
//	Class		: BTREEPAGE
//	Method		: void SetOverflow(int slotNum, PID& const ovPid)
//	Description	: Sets Tuple(slotNum) to overflow, account for spaces
//				  freed up by its transition.
//
void BTREEPAGE::SetOverflow(int slotNum, const PID& ovPid)
{
    BT_Tuple& tp = Tuple(slotNum);
    ASSERT3(! tp.IsOverflow());
    ASSERT3(tp.WillOverflow());
    int oldSize = tp.Size();
    tp.SetOverflow(ovPid);
    ASSERT3(oldSize > tp.Size());
    btCtrl.numFree += oldSize - tp.Size();
    ASSERT3(CheckContent());
}


//
//	Class		: BTREEPAGE
//	Method		: void SetUnderflow(int slotNum, int elCnt, void* const elList)
//	Description	: Sets Tuple(slotNum) to underflow. Insert elList
//				  into the tuple.
//
void BTREEPAGE::SetUnderflow(int slotNum, int elCnt, const void* elList)
{
    ASSERT3((IsLeaf()));
    ASSERT3(slotNum < NumSlots());
    
#ifdef SERVER_MAKE
    TempBufSpace    tempSpace(sizeof(BT_Tuple));
    BT_Tuple&	    scratch = *(BT_Tuple*)tempSpace.location();
#else
    BT_Tuple 	    scratch;   // scratch space
#endif /* SERVER_MAKE */

    //
    //	Copy tuple into a scratch area to do the update
    //
    int oldLength = Tuple(slotNum).Size();	
    BCOPY((char*)&Tuple(slotNum), (char*)&scratch, oldLength);
    
    scratch.ResetOverflow(elCnt, elList);
    
    int newLength = scratch.Size();
    
    ASSERT3(newLength > oldLength);
    
    //
    //	See if we can put scratch at the end of data area
    //
    if (newLength < UsableSpace())   {
	slot[-slotNum] = btCtrl.endData;
	BCOPY((char*)&scratch, data + slot[-slotNum], newLength);
	
	btCtrl.numFree += oldLength;
    }
    else {
	//
	//	Compress to free up some space for scratch
	//
	slot[-slotNum] = EMPTYSLOT;		// free old tuple but not its slot
	btCtrl.numFree += oldLength;
	
	Compress();
	
	ASSERT1(newLength <= UsableSpace());
	
	BCOPY((char*)&scratch, data + btCtrl.endData, newLength);
	slot[-slotNum] = btCtrl.endData;
    }
    
    btCtrl.endData += newLength;
    btCtrl.numFree -= newLength;
    
    ASSERT3((CheckContent()));
}


PFK BT_PrintFunc;


void BTREEPAGE::Print(BUFGROUP* bufGroup)
{
    int 	i;
    int		numOffsets = btCtrl.numOffsets;
    
    
    if (IsNode())   {
	for (i = 0; i < 5 - btCtrl.level; i++)  {
	    printf("\t");
	}
	printf("    <pid0 = %d>\n", btCtrl.pid0);
    }
    
    ASSERT3((CheckContent()));
    
    for (i = 0; i < numOffsets; i++)  {
	void*	key = Tuple(i).KeyValue();
	TWO		keyLen = Tuple(i).KeyLen();
	TWO		numEl;
	int 	j;
	
	for (j = 0; j < 5 - btCtrl.level; j++)  {
	    printf("\t");
	}
	printf("    <key = ");
	BT_PrintFunc(keyLen, (char*) key);
	
	if (IsNode())		{
	    SHORTPID 	pid = Tuple(i).Vector();
	    printf(", pid = %d>\n", pid);
	}
	else if (IsLeaf()) 	{
	    numEl = Tuple(i).ElCnt();
	    
	    printf(", elCnt = %d", numEl);
	    
	    for (j = 0; j < numEl; j++)  {
		char el[SM_MAXELEMLEN];
		if ( ! Tuple(i).IsOverflow())
		    BCOPY(Tuple(i).ElArray(j), el, Tuple(i).ElSize());
		else if (ov_GetElem(Tuple(i).OverflowRoot(), 
				    bufGroup, j, el)){
		    SM_ERROR(TYPE_FATAL, esmINTERNAL);
		}
		if (j == 0)
		    printf(", (%d)", * (int*) el);
	    }
	    printf(">\n");
	}
    }
}






//
//	Class		: BTREEPAGE
//	Method		: int CheckContent()
//	Description	: Checks the content of this page for inconsistency.
//				  It is fatal to be inconsistent.
//				  Checks for 1.) tuple overlapping one another
//							 2.) values of endData and numFree
//				  Returns TRUE if completed successfully.
//
int BTREEPAGE::CheckContent()
{
    ASSERT1(Level() >= 1);
    
    //
    //	Leaf page must have null first vector
    //
    ASSERT1(! (IsLeaf() && FirstVector() != NULLPID));
    
    ASSERT1(NumSlots() >= 0);
    
    int ENDDATA = 0;
    int NUMFREE = BT_DATASIZE;
#ifdef SERVER_MAKE
    // too big to put on stack
    TempBufSpace    tempSpace(BT_DATASIZE);
    char*	    map = tempSpace.location();
#else
    char map[BT_DATASIZE];
#endif /* SERVER_MAKE */

    bzero(map, BT_DATASIZE);
    
    for (int i = 0; i < NumSlots(); i++)  {
	if (slot[-i] == EMPTYSLOT)  {
	    NUMFREE -= sizeof(SLOT); 
	    continue;
	}
	BT_Tuple& tp = Tuple(i);
	int size = tp.Size();
	int endbyte = slot[-i] + size;
	for (int j = slot[-i]; j < endbyte; j++)  {
	    
	    //
	    //	Assert that tuples do not overlap
	    //
	    ASSERT1(!map[j]);
	    map[j] = 1;
	}
	if (ENDDATA < j) ENDDATA = j;
	NUMFREE -= size + sizeof(SLOT);
    }
    
    ASSERT1(ENDDATA <= btCtrl.endData);
    
    ASSERT1(FreeSpace() == NUMFREE);
    
    ASSERT1(btCtrl.endData <= BT_DATASIZE - sizeof(SLOT) * NumSlots());
    
    return TRUE;
}


