/*
 * $RCSfile: bt_FetchNext.C,v $
 * $Revision: 1.1.1.1 $
 * $Date: 1996/05/04 21:55:15 $
 */
/**********************************************************************
 * EXODUS Database Toolkit Software
 * Copyright (c) 1991 Computer Sciences Department, University of
 *                    Wisconsin -- Madison
 * All Rights Reserved.
 *
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
 * MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
 * THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
 * WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * The EXODUS Project Group requests users of this software to return 
 * any improvements or extensions that they make to:
 *
 *   EXODUS Project Group 
 *     c/o David J. DeWitt and Michael J. Carey
 *   Computer Sciences Department
 *   University of Wisconsin -- Madison
 *   Madison, WI 53706
 *
 *    or exodus@cs.wisc.edu
 *
 * In addition, the EXODUS Project Group requests that users grant the 
 * Computer Sciences Department rights to redistribute these changes.
 **********************************************************************/
#include "BTREEPAGE.h"
#include "ParentStack.h"
#include "OVPAGE.h"


//
//  Function    : int bt_FetchNext(BT_CURSOR& cursor, int groupIndex, 
//                                      KEY* bound2, SMCOND condition, 
//                                      PFC compFunc, void* ret)
//  Description : Given a cursor, look a key that satisfies
//                condition(key, bound2). The result of the search is
//                returned in retOid. 
//                NOTE: if bound2 if NULL, any key will satisfy the 
//                      the condition.
//
int bt_FetchNext(BT_CURSOR& cursor, BUFGROUP* bufGroup, void* bound2, 
                 int bound2Len, SMCOND condition, PFC compFunc, 
                 void* valuePtr, TWO* lenPtr, void* ret,
                 BOOL& eof)
{
    ASSERT1(cursor.valid);

    PageDesc    pageDesc;

    if (pageDesc.Read(bufGroup, cursor.page, SH))
        return esmFAILURE;

    BTREEPAGE* leaf = pageDesc;
    ASSERT3(leaf->IsLeaf());
    
    BT_Tuple& tp = leaf->Tuple(cursor.slotNum);

    //
    //  First, check return values
    //
    if (bound2)  {
	int diff = compFunc(bound2Len, bound2, tp.KeyLen(), tp.KeyValue());
	if ((condition & SM_EQ) &&  ! diff) ;
	else if ((condition & SM_L) && diff > 0) ;
	else if ((condition & SM_G) && diff < 0) ;
	else cursor.valid = FALSE;
	if (eof = ! cursor.valid)  {
		pageDesc.Unfix();
	    return esmNOERROR;
	}
    }
    
    //
    //  Fill return values
    //
    if (lenPtr)  {
        if (*lenPtr < tp.KeyLen())  {
            SM_ERROR(TYPE_USER, esmKEYTOOLONG);
	    pageDesc.Unfix();
            return esmFAILURE;
            }
        *lenPtr = tp.KeyLen();
        BCOPY(tp.KeyValue(), valuePtr, *lenPtr);
        }
    if ( ! tp.IsOverflow()) {
        BCOPY(tp.ElArray(cursor.index), ret, tp.ElSize());
        }
    else if (ov_GetElem(tp.OverflowRoot(), bufGroup, cursor.index, ret)) {
	pageDesc.Unfix();
        return esmFAILURE;
	}
    //
    //  Move the cursor to the next eligible unit based on 'condition'
    //
    int found = TRUE;
    switch (condition)  {
	
    case SM_EQ:
	if (++cursor.index >= leaf->Tuple(cursor.slotNum).ElCnt()) 
	    found = FALSE;
	break;
	
    case SM_L:
    case SM_LEQ:
	//
	//  Move cursor one unit to the right
	//
	if (++cursor.index >= leaf->Tuple(cursor.slotNum).ElCnt())  {
	    cursor.index = 0;
	    if (++cursor.slotNum >= leaf->NumSlots())  {
		cursor.slotNum = 0;
		
		do {
		    cursor.page.page = leaf->NextPage();
		    
		    if (cursor.page.page == NULLPID)    {
			found = FALSE;
			break;
		    }
		    
		    pageDesc.Unfix();
		    
		    if (pageDesc.Read(bufGroup, cursor.page, SH))
			return esmFAILURE;
		    
		    leaf = pageDesc;
		    ASSERT3(leaf->IsLeaf());
		    
		} while (leaf->NumSlots() <= 0);
	    }
	}
	
	break;
	
    case SM_G:
    case SM_GEQ:
	
	//
	//  Move cursor one unit to the left
	//
	if (--cursor.index < 0)  {
	    if (--cursor.slotNum < 0)  {
		do {
		    cursor.page.page = leaf->PrevPage();
		    
		    if (cursor.page.page == NULLPID)    {
			found = FALSE;
			break;
		    }
		    
		    pageDesc.Unfix();
		    
		    if (pageDesc.Read(bufGroup, cursor.page, SH))
			return esmFAILURE;
		    
		    leaf = pageDesc;
		    ASSERT3(leaf->IsLeaf());
		    
		} while (leaf->NumSlots() <= 0);
		
		cursor.slotNum = leaf->NumSlots() - 1;
	    }
	    
	    cursor.index = leaf->Tuple(cursor.slotNum).ElCnt() - 1;
	}
	
	break;
	
    default:
	//
	//  Illegal 'condition'
	//
	SM_ERROR(TYPE_FATAL, esmINTERNAL);
    }
    
    //
    //  Check if condition is still satisfied by the entry in
    //  the new cursor position.
    //
    if ( ! found) 
        cursor.valid = FALSE;
    else if (bound2)  {
       BT_Tuple& tp1 = leaf->Tuple(cursor.slotNum); // must be locally 
					// declared and initialized,  since it's a ref type

        int diff = compFunc(bound2Len, bound2, tp1.KeyLen(), tp1.KeyValue());
        if ((condition & SM_EQ) &&  ! diff) ;
        else if ((condition & SM_L) && diff > 0) ;
        else if ((condition & SM_G) && diff < 0) ;
        else cursor.valid = FALSE;
    }
    
    pageDesc.Unfix();
    
    return esmNOERROR;
}

