/*
 *   $RCSfile: bt_InsertEntry.C,v $  
 *   $Revision: 1.1.1.1 $  
 *   $Date: 1996/05/04 21:55:15 $      
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/

#include "BTREEPAGE.h"
#include "ParentStack.h"
#include "SMOPages.h"
#include "OVPAGE.h"

extern void IntPrint(int, char*);
extern void (*BT_PrintFunc)(int, char*);

int BT_NumLeaves = 0;

//
//	Function 	: int bt_InsertEntry(PID& const rootPid, int groupIndex, 
//									void* newKey, int newKeyLen,
//									int elSize, void* const elem,
//									int maxKeyLen, int unique, PFC compFunc)
//	Description	: Insert <newKey,newOid> into the btree rooted at rootPid.
//

int bt_InsertEntry(TID tid, const PID& rootPid, BUFGROUP* bufGroup, 
					const void* newKey, int newKeyLen, 
					int elSize, const void* elem, int maxKeyLen, 
					BOOL unique, PFC compFunc, LSNOFFSET undoNxtLSN)
{
	TWO				slot;
	ParentStack		pl(bufGroup);
	int				keyExists;

#ifdef DEBUG_INSERT
	BT_TRACE(("\n\n=========================\n\n"));
#endif

	//
	//	Traverse the tree
	//
	if (BT_Traverse(rootPid, newKeyLen, newKey, keyExists,
							slot, pl, BT_INSERT, maxKeyLen, compFunc))  {
		return esmFAILURE;
		}

#ifdef DEBUG_INSERT
	BT_TRACE(("After traverse, "))
	pl.Print();
#endif

	BTREEPAGE* leaf = pl.Top();
	ASSERT3(leaf->IsLeaf());
	
	SMDATATYPE keyType = leaf->KeyType();

	BT_Tuple buffer;
	int length;
	
	if (!keyExists)  {
		// 	---- Construct a brand new <key-value, oid> tuple in local buffer
		buffer.Create(newKeyLen, newKey, elem, elSize);
		length = buffer.Size();
		}
	else {
		if (unique)		{
			SM_ERROR(TYPE_USER, esmKEYALREADYEXISTS);
			return esmFAILURE;
			}
		BT_Tuple& tp = leaf->Tuple(slot);
		
		if (tp.WillOverflow())  {
			PID ovPid;
			if (ov_NewCluster(tid, rootPid.volid, bufGroup, 
											elSize, ovPid, tp))   {
				return esmFAILURE;
				}
			BT_LogLogicSetOverflow(leaf->SelfID(), pl.Top(), leaf->lrc(), 
					rootPid, newKey, newKeyLen, maxKeyLen, keyType, 
					tp.ElCnt(), tp.ElSize(), tp.ElArray(0), ovPid, 0);
			leaf->SetOverflow(slot, ovPid);
			ASSERT3(tp.IsOverflow());
			}

		if (tp.IsOverflow())  {
			if (ov_AddElem(tid, tp.OverflowRoot(), bufGroup, elem))
				return esmFAILURE;
			tp.IncrElCnt();
			BT_LogLogicIncrElCnt(leaf->SelfID(), pl.Top(), leaf->lrc(),
									rootPid, newKey, newKeyLen, 
									maxKeyLen, keyType, 0);
			pl.Top().Dirty();
			return esmNOERROR;
			}

		//	---- Inserting an element into an existing key		
		length = ALIGNSIZE(elSize);
		}


	//
	//	Split leaf recursively if there is not enough room for the
	//	new stuff.
	//

	if (((BTREEPAGE*)pl.Top())->FreeSpace() < length + sizeof(SLOT))  {
		SMOPages smo;
		
		BT_TRACE(("btree:  page split\n"));
		++BT_NumLeaves;

		//
		//	Set anchor for Compensation Log Record after split
		//
		if (BT_SetAnchor(tid))  {
			return esmFAILURE;
			}

		//
		// 	Split old page into two evenly distributed page
		//
		PageDesc	newLeafLink;
		if (BT_SplitPage(tid, bufGroup, &smo, pl, slot, length, newLeafLink))
			return esmFAILURE;

		BTREEPAGE* leaf = pl.Top();
		BTREEPAGE* newLeaf = newLeafLink;
		PID newLeafPid = newLeaf->SelfID();

		//
		// 	Determine which page the new key should go to.
		//
		int oldSlotCnt = leaf->NumSlots();
		BT_Tuple* propagateTuple = & newLeaf->Tuple(0);
		if ( ! (slot < oldSlotCnt  ||  
			(slot == oldSlotCnt && (! keyExists) &&
			(leaf->FreeSpace() > newLeaf->FreeSpace()))) ){
			//
			//	New stuff should go to new page
			//	Replace top of ParentStack with newLeaf
			//
			pl.DiscardTop();

			if (pl.ReadNPush(newLeafPid, EX)) {
				newLeafLink.Unfix();
				return esmFAILURE;
				}

			slot -= oldSlotCnt;			// slot in newLeaf for insertion

			if (slot == 0 &&  ! keyExists)  {
				//	---- New tuple will be the first on newLeaf
				//				after leaf->InsertTuple() few lines down
				propagateTuple = &buffer;
				}
			}
		newLeafLink.UnfixDirty();

		//
		//	Propagate split upwards
		//
		PageDesc acc = pl.Pop();
		if (BT_NewNodeTuple(tid, bufGroup, &smo, pl, propagateTuple->KeyLen(),
									propagateTuple->KeyValue(),
									&newLeafPid, compFunc))  {
			pl.Push(acc);
			return esmFAILURE;
			}
		pl.Push(acc);

		//
		//	Flush all pages involved in the SMO
		//	Since every update in SMO are physical operations, the server
		//	would be able to undo these updates even if the client aborts
		//	here. Flushing SMO pages are necessary because we are going
		//	to write a CLR next (to signify end of top-level action). This
		//	ensures that the splits have indeed been performed on the 
		//	server.
		//
		smo.Flush();

		//
		//	Write compensation log record to conclude SMO
		//
		if (BT_SetCLR(tid))  {
			return esmFAILURE;
			}
		}

	leaf = pl.Top();
	

	//
	//	Finally, insert into leaf page
	//
	if (keyExists)		{
		//	---- Append the new Oid to an existing key
		ASSERT3(! leaf->Tuple(slot).IsOverflow());
		ASSERT3(! leaf->Tuple(slot).WillOverflow());
		if (leaf->AddElem(slot, elem))
			return esmFAILURE;
		}
	else  {
		//	---- Add a new <key-value, oid> pair
		ASSERT3(leaf->FreeSpace() > buffer.Size());
		leaf->InsertTuple(buffer, slot);
		}

	//
	//	Log the action and mark the page dirty
	//
	ASSERT3(keyType == leaf->KeyType());
	BT_LogLogicInsert(leaf->SelfID(), pl.Top(), leaf->lrc(),
						rootPid, newKey, newKeyLen, elem, elSize, 
						maxKeyLen, unique, keyType, undoNxtLSN);

	pl.Top().Dirty();

#ifdef DEBUG_INSERT
	BT_TRACE(("After Insert, "));
	pl.Print();
#endif

	return esmNOERROR;
}
