/*
 * $RCSfile: ov_AddElem.C,v $
 * $Revision: 1.1.1.1 $
 * $Date: 1996/05/04 21:55:16 $
 */
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/

#include "BTREEPAGE.h"
#include "OVPAGE.h"
#include "sm_macro.h"


//
//	Function	: int ov_AddElem(TID tid, PID& const rootPid, 
//									BUFGROUP* bufGroup, void* const newElem)
//	Description	: Add 'newElem' to the overflow entry rooted at 'rootPid'.
//				  Split is performed if necessary.
//
int ov_AddElem(TID tid, const PID& rootPid, BUFGROUP* bufGroup, 
												const void* newElem)
{
	int idx;
	PID	pid;
	
	//
	//	Find the location to insert 'newElem'
	//
	int found;
	if (OV_Search(rootPid, bufGroup, newElem, pid, found, idx))
		return esmFAILURE;
		
	if (found)  {
		SM_ERROR(TYPE_USER, esmOIDEXISTS);
		return esmFAILURE;
		}
	
	PageDesc pageDesc;
	if (pageDesc.Read(bufGroup, pid, EX))
		return esmFAILURE;

	OVPAGE* pg = pageDesc;

	//
	//	Insert into this page if there is enough room
	//
	if (pg->ElCnt() < pg->MaxEl())  {
		pg->Insert(idx, 1, newElem);
		OV_LogInsert(pg->SelfID(), pageDesc, pg->lrc(), rootPid, pg->ElSize(),
									idx, 1, pg->Get(idx), 0);
		pageDesc.UnfixDirty();
		return esmNOERROR;
		}

	ASSERT3(pg->ElCnt() == pg->MaxEl());

	//
	//	Split this page.
	// 	First, Read in the right sibling if it exists
	//
	PageDesc rightPageDesc;
	OVPAGE*	rightPage = 0;
	if (pg->NextPage() != NULLPID)  {
		PID	rightPagePid;
		rightPagePid.volid = rootPid.volid;
		rightPagePid.page = pg->NextPage();
		if (rightPageDesc.Read(bufGroup, rightPagePid, EX))
			return esmFAILURE;

		rightPage = rightPageDesc;
		ASSERT3(rightPage->SelfID() == rightPagePid);
		ASSERT3(rightPage->PrevPage() == pid.page);
		ASSERT3(pg->NextPage() == rightPagePid.page);
		}
	

	//
	//	Allocate the new page
	//
	PID midPid;
	if (OV_AllocPage(tid, rootPid.volid, bufGroup, pg->ElSize(), midPid)) {
		rightPageDesc.Unfix();
		pageDesc.Unfix();
		return esmFAILURE;
		}

	PageDesc midPageDesc;
	if (midPageDesc.Read(bufGroup, midPid, EX)) {
		rightPageDesc.Unfix();
		pageDesc.Unfix();
		return esmFAILURE;
		}

	OVPAGE* midPage = midPageDesc;

	//
	//	Link-up all three pages
	//
	SHORTPID oldNext;
	SHORTPID oldPrev;
	
	oldNext = pg->NextPage();
	oldPrev = midPage->PrevPage();
	pg->Linkup(midPage);
	OV_LogModifyNextLink(pg->SelfID(), pageDesc, pg->lrc(), 
								rootPid, oldNext, pg->NextPage(), 0);
	OV_LogModifyPrevLink(midPage->SelfID(), midPageDesc, midPage->lrc(),
								rootPid, oldPrev, midPage->PrevPage(), 0);
	
	if ( ! rightPage)  {
		ASSERT1(midPage->NextPage() == NULLPID);
		}
	else {
		oldNext = midPage->NextPage();
		oldPrev = rightPage->PrevPage();
		midPage->Linkup(rightPage);
		OV_LogModifyNextLink(midPage->SelfID(), midPageDesc, midPage->lrc(),
				rootPid, oldNext, midPage->NextPage(), 0);
		OV_LogModifyPrevLink(rightPage->SelfID(), rightPageDesc, 
				rightPage->lrc(), rootPid, oldPrev, rightPage->PrevPage(), 0);
		rightPageDesc.UnfixDirty();
		rightPage = 0;
		}

	//
	// 	Move half of pg into midPage
	//
	BT_TRACE(("btov: split\n"));
	int start = pg->ElCnt() / 2;
	int cnt = pg->ElCnt() - start;
	OV_LogDelete(pg->SelfID(), pageDesc, pg->lrc(), 
						rootPid, pg->ElSize(), start, cnt, pg->Get(start), 0);
	pg->Shift(start, midPage);
	OV_LogInsert(midPage->SelfID(), midPageDesc, midPage->lrc(), 
								rootPid, midPage->ElSize(),
								midPage->ElCnt() - cnt, cnt, 
								midPage->Get(midPage->ElCnt() - cnt), 0);
	
	//
	//	Insert the new oid into the appropriate page
	//
	if (idx < start)   {
		pg->Insert(idx, 1, newElem);
		OV_LogInsert(pg->SelfID(), pageDesc, pg->lrc(), 
							rootPid, pg->ElSize(), idx, 1, pg->Get(idx), 0);
		}
	else  {
		midPage->Insert(idx-start, 1, newElem);
		OV_LogInsert(midPage->SelfID(), midPageDesc, midPage->lrc(),
						rootPid, midPage->ElSize(), 
						idx-start, 1, midPage->Get(idx-start), 0);
		}

	//
	//	Unfix all three pages
	//
	pageDesc.UnfixDirty();
	midPageDesc.UnfixDirty();

	ASSERT3(OV_Check(rootPid, bufGroup));

	return esmNOERROR;
}
