/*
 * $RCSfile: ov_RemoveElem.C,v $
 * $Revision: 1.1.1.1 $
 * $Date: 1996/05/04 21:55:16 $
 */
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/

#include "BTREEPAGE.h"
#include "OVPAGE.h"
#include "sm_macro.h"


//
//	Function	: int ov_RemoveElem(TID tid, PID& const rootPid, 
//									BUFGROUP* bufGroup, void* const doomedElem)
//	Description	: Remove 'doomedElem' from the overflow pages. 
//				  (Forward) merging is done if the page and the 
//				  previous page contains too few entries.
//
int ov_RemoveElem(TID tid, const PID& rootPid, BUFGROUP* bufGroup, 
											const void* doomedElem)
{
	int idx;
	PID pid;

	//
	//	Search for doomedElem
	//
	int found;
	if (OV_Search(rootPid, bufGroup, doomedElem, pid, found, idx))
		return esmFAILURE;
		
	if (!found)  {
		SM_ERROR(TYPE_USER, esmOIDNOTFOUND);
		return esmFAILURE;
		}

	//
	//	Read in the page in which doomedElem resides
	//
	PageDesc pageDesc;
	if (pageDesc.Read(bufGroup, pid, EX))
		return esmFAILURE;
	OVPAGE* pg = pageDesc;

	ASSERT3(BT_ElemDiff(pg->ElSize(), pg->Get(idx), doomedElem) == 0);

	//
	//	Remove doomedElem
	//
	OV_LogDelete(pg->SelfID(), pageDesc, pg->lrc(), 
						rootPid, pg->ElSize(), idx, 1, pg->Get(idx), 0);
	pg->Remove(idx, 1);
	ASSERT3( idx >= pg->ElCnt() || 
					BT_ElemDiff(pg->ElSize(), pg->Get(idx), doomedElem) != 0);


	//
	//	See if this page contans too few entries
	//
	PageDesc leftPageDesc;
	OVPAGE* leftPage = 0;
	PageDesc rightPageDesc;
	OVPAGE* rightPage = 0;
	int merged = 0;
	if (pg->ElCnt() < pg->MaxEl()/4 && pg->PrevPage() != NULLPID)  	{

		//
		//	Read in left page
		//
		PID leftPagePid;
		leftPagePid.volid = rootPid.volid;
		leftPagePid.page = pg->PrevPage();
		if (leftPageDesc.Read(bufGroup, leftPagePid, EX)) {
			pageDesc.Unfix();
			return esmFAILURE;
			}

		leftPage = leftPageDesc;

		//
		//	See if we need to merge
		//
		int allCnt = leftPage->ElCnt() + pg->ElCnt();
		if (allCnt <= pg->MaxEl() * 3 / 4)  {
		
			BT_TRACE(("btov: merging\n"));

			//
			//	Copy all oids in pg to leftPage
			//
			int cnt = pg->ElCnt();
			OV_LogDelete(pg->SelfID(), pageDesc, pg->lrc(), 
								rootPid, pg->ElSize(), 0, cnt, pg->Get(0), 0);
			pg->Shift(0, leftPage);
			OV_LogInsert(leftPage->SelfID(), leftPageDesc, leftPage->lrc(),
									rootPid, leftPage->ElSize(),
									leftPage->ElCnt() - cnt, cnt,
									leftPage->Get(leftPage->ElCnt()-cnt), 0);

			//
			//	Read in rightRightPage
			//
			if (pg->NextPage() != NULLPID)  {
				PID rightPagePid;
				rightPagePid.volid = rootPid.volid;
				rightPagePid.page = pg->NextPage();
				if (rightPageDesc.Read(bufGroup, rightPagePid, EX)) {
					pageDesc.Unfix();
					leftPageDesc.Unfix();
					return esmFAILURE;
					}
				rightPage = rightPageDesc;
				}

			merged = 1;
			}
		}

	if ( ! merged)   {
		pageDesc.UnfixDirty();
		if (leftPage)  leftPageDesc.Unfix();
		}
	else  {
		//
		//	Clean up and adjust links 
		//
		pageDesc.Unfix();
		if (OV_FreePage(tid, pid)) {
			pageDesc.Unfix();
			if (leftPage)  leftPageDesc.Unfix();
			if (rightPage) rightPageDesc.Unfix();
			return esmFAILURE;
			}

		SHORTPID oldNext = leftPage->NextPage();
		SHORTPID oldPrev = NULLPID;
		if (rightPage) {
			oldPrev = rightPage->PrevPage();
			}
		leftPage->Linkup(rightPage);
		OV_LogModifyNextLink(leftPage->SelfID(), leftPageDesc, leftPage->lrc(),
								rootPid, oldNext, leftPage->NextPage(), 0);


		if (rightPage)  {
			OV_LogModifyPrevLink(rightPage->SelfID(), rightPageDesc, 
								rightPage->lrc(), rootPid, 
								oldPrev, rightPage->PrevPage(), 0);
			rightPageDesc.UnfixDirty();
			}

		leftPageDesc.UnfixDirty();
		}

	ASSERT3(OV_Check(rootPid, bufGroup));

	return esmNOERROR;
}

