/*
 *   $RCSfile: CLOG_WriteLogRecord.c,v $  
 *   $Revision: 1.1.1.1 $  
 *   $Date: 1996/05/04 21:55:17 $
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/

#include "sysdefs.h"
#include "ess.h"
#include "checking.h"
#include "trace.h"
#include "error.h"
#include "list.h"
#include "pool.h"
#include "tid.h"
#include "io.h"
#include "lock.h"
#include "object.h"
#include "msgdefs.h"
#include "bf.h"
#include "chunk.h"
#include "sm_macro.h"
#include "lsn.h"
#include "loginfo.h"
#include "serverinfo.h"
#include "sm_state.h"
#include "logrecs.h"
#include "volume.h"
#include "trans.h"
#include "clog.h"
#include "util_funcs.h"
#include "bf_extfuncs.h"
#include "clog_intfuncs.h"
#include "clog_extfuncs.h"
#include "clog_globals.h"
#include "trans_globals.h"

    
/***************************************************************
 *  CLOG_writeLogRecord
 * 	
 *  PURPOSE	
 *	Put the given log information into the current log page.
 *	The log record is known to fit on the given page. 
 *  
 ***************************************************************/

FORCEMARK
CLOG_WriteLogRecord(

    TID				unused,		 /* not used */
    CLIENTLOGRECORDINFO	*recInfo,
    LRC             *returnLRC,  /* In/out parameter */
    PAGEHASH		*pageHash    /* buffer structure for affected page */
)
{
    register LOGINFO *logInfo;
    register int count;
    register LSNOFFSET  tailLSN;
    PAGEHASH	*tailBuffer;
    FORCEMARK forceMark;
    int offset, spaceLeft, total, recSize;
    LOGRECORDHDR recordHeader;
    char *writePtr, *basePtr;

    /* get register pointer to LogInfo */
    logInfo = GET_LOGINFO;

	/* validate action code */
	SM_ASSERT(LEVEL_3, recInfo->action > 0);

	/*
	 *	For now, this function is a convenient place to turn off
	 *	logging for new pages and pages on temp volumes.  
	 *  Cpu-time wise it would be better
	 *	to do it earlier, be that's a later optimization.
	 */
	if (pageHash->flags & H_TEMP) {
		SM_ASSERT(LEVEL_1, (SM_State.logLevel == LOG_NONE));
		forceMark = logInfo->logRecordCount + 1;
		goto done;
	}
	if ( !LogNewPages && pageHash->flags & H_NEW) {

		forceMark = logInfo->logRecordCount + 1;
		goto done;
	}

    /*
     * determine placement of log record(s) on the page(s)
     */
    
    INCREMENT_LRC(returnLRC);
    tailLSN = logInfo->tailLSN;
    offset = (int)(tailLSN & logInfo->pageMask);
    spaceLeft = logInfo->lastUsableByte - offset;

    /* 
     * if this is the first log record for this page, update the firstLSN and firstLRC
     *     fields in the pageHash for this page.
     */
    if ((pageHash->firstLSN.wrapCount == 0) && (pageHash->firstLSN.offset == 0)) {
	pageHash->firstLSN = logInfo->lastTailLSN;

	SM_ASSERT(LEVEL_1, (pageHash->firstLSN.offset & logInfo->pageMask)
		>= sizeof(LOGPAGEHDR));

	pageHash->firstLRC = *returnLRC;
    }

    /*
     *  increment the counter for this transaction
     */
    forceMark = ++(logInfo->logRecordCount);
    TRPRINT(TR_LOG, TR_LEVEL_2, ("forceMark:%d", forceMark));

    /* copy in the fields of the header*/
    recordHeader.magic       = LOGRECORD_MAGIC;
    recordHeader.type        = recInfo->type;
    recordHeader.action      = recInfo->action;
    recordHeader.flags       = recInfo->flags;
    recordHeader.imageCount  = recInfo->imageCount;
    recordHeader.tid         = GET_SERVERTID;
    recordHeader.previousLSN = logInfo->currentLSN;
    recordHeader.actionPid   = recInfo->actionPid;
    recordHeader.actionLRC   = *returnLRC;
    recordHeader.nextUndoLSN = 0;

    TRPRINT(TR_LOG, TR_LEVEL_1, ("LogRecord.action:%d tid:%d pid:%d.%d",
		recordHeader.action,
		recordHeader.tid,
		recordHeader.actionPid.page, recordHeader.actionPid.volid));
		
    TRPRINT(TR_LOG, TR_LEVEL_1, ("flags:0x%x LRC:%d.%d pLSN:%d",
		recordHeader.actionLRC.wrapCount, recordHeader.actionLRC.count,
		recordHeader.previousLSN));


    total = 0;
    for(count = 0; count < recInfo->imageCount; count++) {
       recordHeader.imageSize[count] = recInfo->imageSize[count];
       recordHeader.imageOffset[count] = total;
       total += (int) ALIGN(recInfo->imageSize[count],BYTESPERWORD);
    }

    /* NEED TO CHECK THIS SIZE CALC */
    recSize = total + ALIGN(sizeof(LOGRECORDHDR),BYTESPERWORD);

	/* Assert that there is enough space to write log record */
	SM_ASSERT(LEVEL_1, spaceLeft >= recSize);
	SM_ASSERT(LEVEL_1, recordHeader.imageCount <= MAX_LOGREC_IMAGES);
	SM_ASSERT(LEVEL_1, recSize <= MAX_LOGREC_LEN);

    /* write the header */
    recordHeader.length   = recSize;
    recordHeader.recordLSN.offset = tailLSN;
    recordHeader.recordLSN.wrapCount = logInfo->wrapCount;

    tailBuffer = logInfo->tailBuffer;

    /* compute ptr to start of log record in buffer*/
    basePtr = tailBuffer->pageInfo->bufFrame + offset;

    /* copy out record header */
    bcopy((char *) &recordHeader, basePtr, (int) sizeof(LOGRECORDHDR));

    /* now base is start of log record image area */
    basePtr += ALIGN(sizeof(LOGRECORDHDR),BYTESPERWORD);

    for(count = 0; count < recInfo->imageCount; count++) {
        writePtr = basePtr + recordHeader.imageOffset[count];
        bcopy((char *) recInfo->imageData[count], writePtr, 
					(int) recInfo->imageSize[count]);
    }

    /* update the first and force marks on the log page */
    if (tailBuffer->firstMark == 0) {
		tailBuffer->firstMark = forceMark;
    }
    tailBuffer->forceMark = forceMark;

    /*
     *  update the lastRecord field on the log page headers
     */
	CLOG_UpdateLogPageHeader(logInfo, (int)LOG_PAGE_OFFSET(tailLSN, logInfo));

    TRPRINT(TR_LOG, TR_LEVEL_2, ("tailLSN:%d ", tailLSN));

    /*
     *  dirty the page
     */
    bf_DirtyPage(tailBuffer);

    /*
     *  save the updated pointers
     */
    logInfo->currentLSN = tailLSN;
    logInfo->tailLSN = tailLSN + recordHeader.length;

    TRPRINT(TR_LOG, TR_LEVEL_2, ("recordLSN:%d:%d",
	recordHeader.recordLSN.wrapCount, recordHeader.recordLSN.offset));

done:
	if(forceMark < 0) {
		SM_ERROR(TYPE_FATAL, esmINTERNAL);
	}
    return forceMark;
}
