/*
 *   $RCSfile: clog_LogConvertToLarge.c,v $  
 *   $Revision: 1.1.1.1 $  
 *   $Date: 1996/05/04 21:55:17 $      
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/

#include "sysdefs.h"
#include "ess.h"
#include "checking.h"
#include "trace.h"
#include "error.h"
#include "list.h"
#include "pool.h"
#include "tid.h"
#include "io.h"
#include "lock.h"
#include "object.h"
#include "msgdefs.h"
#include "bf.h"
#include "chunk.h"
#include "lsn.h"
#include "loginfo.h"
#include "volume.h"
#include "trans.h"
#include "clog.h"
#include "logrecs.h"
#include "logaction.h"
#include "sm_state.h"
#include "util_funcs.h"
#include "bf_extfuncs.h"
#include "clog_intfuncs.h"
#include "clog_extfuncs.h"
#include "clog_globals.h"
#include "trans_globals.h"


int
clog_LogConvertToLarge(

    OPERATIONINFO	*opInfo 
)
{

    LRC 			lrc;
    int 			recordSize;
    CLIENTLOGRECORDINFO recInfo;
    register LOGINFO 	*logInfo;
    LSNOFFSET  		tailLSN;
    PAGEHASH 		*pageHash;
    int				offset;
    int 			spaceLeft;
    FORCEMARK 		forceMark;

/***************************************************************
 *  This function is used for logging the changes required to 
 *	convert a small (forwarded or not) object into a 
 *	large object slotted root or large object header.
 *	The size field of the opInfo structure must be set
 *	to the correct size of the old object header and associated data
 *
 *	FOR CONVERSION TO A LARGE ROOT:
 *
 *		This is a logical operation, which requires the page freecounts
 *		to be updated on the affected page.
 *
 *		For non-forwarded objects:
 *			In this case, it is assumed that the entire data portion of
 *		    the object has been previously deleted.  Therefore, the old
 *			data contains only the object header.  
 *
 * 			image[0] - OID
 *			image[1] - new object header and data (i.e., the slotted root)
 *			image[2] - old object header 
 *
 *		For forwarded objects:
 *			In this case, the object header and the oid of the forwarded
 *			part will be logged.
 *
 * 			image[0] - OID
 *			image[1] - new object header and data (i.e., the slotted root)
 *			image[2] - old object header and oid
 *
 *	FOR CONVERSION TO A LARGE HEADER:
 * 	    This is operation does not change the space usage of the object
 *		so the page freecounts need not be updated on the affected page.
 *
 *		For non-forwarded and non-header-on-page objects:
 *			In this case, it is assumed that the data portion of
 *		    the object EXCEPT FOR THE FIRST sizeof(SHORTPID) bytes
 *			has been previously deleted.  Therefore, the old
 *			data contains the object header and the first bytes of data.  
 *
 * 			image[0] - OID
 *			image[1] - new object header and data (i.e., the pid of the root)
 *			image[2] - old object header and pid-sized data that was overwritten
 *		
 *		For forwarded objects:
 *			In this case, the object header and the oid of the forwarded
 *			part will be logged.
 *
 * 			image[0] - OID
 *			image[1] - new object header and data (i.e., the short pid)
 *			image[2] - old object header and oid
 *
 *		For header-on-page large objects :
 *			In this case, the object header and the slotted root part of
 *			the large object will be logged.
 *
 * 			image[0] - OID
 *			image[1] - new object header and data (i.e., the short pid)
 *			image[2] - old object header and slotted root
 * 
 *	This log record is smaller than a log page.
 ***************************************************************/
     
    /* Get register pointer to log info structure */
    logInfo = GET_LOGINFO;

    /* first, make sure that we have a valid log buffer at the tail of the
       log.  Get a new log page if this isn't the case.*/

    if (logInfo->tailBuffer == NULL) {
        CLOG_GetNewLogPage();
    }

    /*
     * determine size of the log record and images (see comment at top)
     */

    if (opInfo->action == LOG_ACTION_CONVERT_TO_LARGEHDR) {

        recordSize = (int)  ( ALIGN(sizeof(LOGRECORDHDR),BYTESPERWORD) +
		ALIGN(sizeof(OID),BYTESPERWORD) +
		ALIGN(sizeof(LARGEOBJHDR),BYTESPERWORD) +
		ALIGN(opInfo->size,BYTESPERWORD)  );
    	recInfo.imageSize[1] = sizeof(LARGEOBJHDR);
    	recInfo.imageSize[2] = (UTWO) opInfo->size;
    } else {
        if (opInfo->action == LOG_ACTION_CONVERT_TO_LARGEROOT) {
            recordSize = (int) (ALIGN(sizeof(LOGRECORDHDR),BYTESPERWORD) +
				ALIGN(sizeof(OID),BYTESPERWORD) +
				ALIGN(sizeof(LARGEOBJROOT),BYTESPERWORD) +
				ALIGN(opInfo->size,BYTESPERWORD) ) ;
    	    recInfo.imageSize[1] = sizeof(LARGEOBJROOT);
    	    recInfo.imageSize[2] = (UTWO) opInfo->size;
		} else {
		    /* this is an error */
		    SM_ERROR(TYPE_FATAL,esmINTERNAL);
	    }
    }	
	    

    /* insure that the log record will fit on a log page */
    SM_ASSERT(LEVEL_1, recordSize <= logInfo->usableBytes);

    recInfo.type        = LOG_REC_TYPE_USER;
    recInfo.action      = opInfo->action;
    recInfo.flags       = NOFLAGS;
    recInfo.imageCount  = 3;
    recInfo.imageSize[0] = sizeof(OID);
    recInfo.imageData[0] = (UONE *) &(opInfo->oid);
    recInfo.imageData[1] = (UONE *) opInfo->newData;
    recInfo.imageData[2] = (UONE *) opInfo->oldData;
    recInfo.actionPid.page = opInfo->oid.diskAddr.page;
    recInfo.actionPid.volid = opInfo->oid.diskAddr.volid;

    if ((pageHash = bf_FindPage( &(recInfo.actionPid))) == NULL) {
       return(esmFAILURE);
    }
	SM_ASSERT( LEVEL_1, !(pageHash->flags & H_INVALID));
    lrc = ((SLOTTEDPAGE *) pageHash->pageInfo->bufFrame)->header.lrc;

    /*
     * determine placement of log record(s) on the page(s)
     */
    
    tailLSN = logInfo->tailLSN;
    offset = (int)(tailLSN & logInfo->pageMask);
    spaceLeft = logInfo->lastUsableByte - offset;

    /* see if we can use space on current log page. */
    if (recordSize > spaceLeft) {

        /* can't take advantage of space in current page, move on to next*/
	    CLOG_GetNewLogPage();
        spaceLeft = logInfo->usableBytes;
    }
	    
    spaceLeft -= recordSize;

    /* check the size calculation*/
    SM_ASSERT(LEVEL_1, spaceLeft >= 0);


    /* write the log record*/
    forceMark = CLOG_WriteLogRecord(TransRec.clientTid, &recInfo, &lrc, pageHash);

     pageHash->forceMark = forceMark;
     ((SLOTTEDPAGE *) pageHash->pageInfo->bufFrame)->header.lrc = lrc;
     return esmNOERROR;
}
