/*
 *   $RCSfile: clog_LogCreateObj.c,v $  
 *   $Revision: 1.1.1.1 $  
 *   $Date: 1996/05/04 21:55:17 $      
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/

#include "sysdefs.h"
#include "ess.h"
#include "checking.h"
#include "trace.h"
#include "error.h"
#include "list.h"
#include "pool.h"
#include "tid.h"
#include "io.h"
#include "lock.h"
#include "object.h"
#include "msgdefs.h"
#include "bf.h"
#include "chunk.h"
#include "lsn.h"
#include "loginfo.h"
#include "volume.h"
#include "trans.h"
#include "clog.h"
#include "logrecs.h"
#include "logaction.h"
#include "sm_state.h"
#include "util_funcs.h"
#include "bf_extfuncs.h"
#include "clog_intfuncs.h"
#include "clog_extfuncs.h"
#include "clog_globals.h"
#include "trans_globals.h"


int
clog_LogCreateObj(

    OPERATIONINFO	*opInfo 
)
{
    LRC lrc;
    int recSize;
    int fixedSize;
    int bodySize;
    CLIENTLOGRECORDINFO recInfo;
    register LOGINFO *logInfo;
    WRITEINFO writeInfo;
    LSNOFFSET  tailLSN;
    PAGEHASH *pageHash;
    int	offset;
    int spaceLeft, writtenSize, imageSize, usableBytes, leftOver;
    FORCEMARK forceMark;
    BOOL 	firstRecord;

/***************************************************************
 *  Format of a create log record is:
 * 	image[0] - OID,start location, length
 *	image[1] - new data
 *
 *  Additional log records are APPENDS 
 ***************************************************************/
     
    /* Get register pointer to log info structure */
    logInfo = GET_LOGINFO;

	/* get info on affected page */
    recInfo.actionPid.page = opInfo->oid.diskAddr.page;
    recInfo.actionPid.volid = opInfo->oid.diskAddr.volid;
    if ((pageHash = bf_FindPage( &(recInfo.actionPid))) == NULL) {
       return(esmFAILURE);
    }
	SM_ASSERT( LEVEL_1, !(pageHash->flags & H_INVALID));
    lrc = ((SLOTTEDPAGE *) pageHash->pageInfo->bufFrame)->header.lrc;

    /* first, make sure that we have a valid log buffer at the tail of the
       log.  Get a new log page if this isn't the case.*/

    if (logInfo->tailBuffer == NULL) {
        CLOG_GetNewLogPage();
    }

    /*
     * determine size of the log record 
     */

    /* fixed part of log record is aligned record header, aligned
	write info and enough space to align new image */
    fixedSize = (ALIGN(sizeof(LOGRECORDHDR),BYTESPERWORD)) +
		(ALIGN(sizeof(WRITEINFO),BYTESPERWORD)) + BYTESPERWORD;
		
    /* data portion = object hdr + object data */
    bodySize = (int) opInfo->size + sizeof(OBJHDR);

    recSize = bodySize + fixedSize;

    /*
     * set up record info fields which are common to all
     */

    recInfo.type        = LOG_REC_TYPE_USER;
    recInfo.action      = opInfo->action;
    recInfo.flags       = NOFLAGS;
    recInfo.imageCount  = 2;
    recInfo.imageSize[0] = sizeof(WRITEINFO);
    recInfo.imageData[0] = (UONE *) &writeInfo;

    /*
     * determine placement of log record(s) on the page(s)
     */
    
    tailLSN = logInfo->tailLSN;
    offset = (int)(tailLSN & logInfo->pageMask);
    spaceLeft = logInfo->lastUsableByte - offset;

    /* may need to split log record */
    usableBytes = (logInfo->usableBytes - fixedSize);

    /* now start writing out log records */
    writtenSize = 0;
    firstRecord = TRUE;
    while (bodySize > 0) {
        leftOver = (bodySize % usableBytes) + fixedSize;
	
	/* see if we can use space on current log page.  Note: make
	   sure that object header is not split across log recs*/

        if ((leftOver > spaceLeft) || 
			 (spaceLeft < (fixedSize + sizeof(OBJHDR)))) {

            /* can't take advantage of space in current page, move on to next*/
			CLOG_GetNewLogPage();
			spaceLeft = logInfo->usableBytes;
		}
	    
		/* compute amount of space available for new Data */
		imageSize = MIN((spaceLeft - fixedSize),bodySize);
		bodySize -= imageSize;

		/*adjust space left to account for the size of this record */
		spaceLeft -= (imageSize + fixedSize);

		/* check the size calculation*/
		SM_ASSERT(LEVEL_1, spaceLeft >= 0);

		/* prepare to write out this log record */
		if (firstRecord) {
			/* first record includes object header */
			writeInfo.unique = opInfo->oid.diskAddr.unique;
			writeInfo.slot = opInfo->oid.diskAddr.slot;
			writeInfo.start = 0;

			/* for create, if we can't write out the entire
			   data portion, we need to TEMPORARILY
			   adjust the size of the object header.
			   Subsequent appends will update it */

			if ( (imageSize-sizeof(OBJHDR)) < opInfo->size) {
				((OBJHDR *) opInfo->newData)->size = (imageSize-sizeof(OBJHDR));
			}
		}
		else {
			/* subsequent records are appends*/
				recInfo.action = LOG_ACTION_APPEND_SMALL_OBJ;
			writeInfo.unique = opInfo->oid.diskAddr.unique;
			writeInfo.slot = opInfo->oid.diskAddr.slot;
			writeInfo.start = writtenSize;
		}

		if (firstRecord) {
			/* first record includes the header */
				recInfo.imageData[1] = (UONE *)opInfo->newData;
			writtenSize += imageSize-sizeof(OBJHDR);
		} else {
			/* written size doesn't include the hdr, so must add */
				recInfo.imageData[1] = (UONE *)opInfo->newData + writtenSize +
									sizeof(OBJHDR);
			writtenSize += imageSize;
		}
			recInfo.imageSize[1] = imageSize;


		/* write the log record*/
		forceMark = CLOG_WriteLogRecord(TransRec.clientTid, &recInfo, &lrc, pageHash);
		
		if ((firstRecord) && ((imageSize-sizeof(OBJHDR)) < opInfo->size)) {
			/* Need to reset the object's size */
			((OBJHDR *) opInfo->newData)->size = opInfo->size;
		}

			firstRecord = FALSE;
    }

    /* mark the page dependency with the last log record written */
     pageHash->forceMark = forceMark;

    /* put the new lrc on the page */
     ((SLOTTEDPAGE *) pageHash->pageInfo->bufFrame)->header.lrc = lrc;

    /* return success */
     return(esmNOERROR);
}


int
clog_LogCreateObjND(

    OPERATIONINFO	*opInfo 
)
{

    LRC 	lrc;
    int 	recSize;
    int 	fixedSize;
    int 	bodySize;
    CLIENTLOGRECORDINFO recInfo;
    LOGINFO *logInfo;
    WRITEINFO writeInfo;
    LSNOFFSET  tailLSN;
    PAGEHASH *pageHash;
    int		offset;
    int 	spaceLeft;
    FORCEMARK forceMark;
#ifdef DEBUG
	OBJHDR	*objHdr; 
#endif /*DEBUG*/

/***************************************************************
 *  Format of a create log record is:
 * 	image[0] - OID,start location, length
 *	image[1] - new data
 *
 *  Additional log records are APPENDS 
 ***************************************************************/
    
    /* Get register pointer to log info structure */
    logInfo = GET_LOGINFO;

	/* get info on affected page */
    recInfo.actionPid.page = opInfo->oid.diskAddr.page;
    recInfo.actionPid.volid = opInfo->oid.diskAddr.volid;
    if ((pageHash = bf_FindPage( &(recInfo.actionPid))) == NULL) {
       return(esmFAILURE);
    }
	SM_ASSERT( LEVEL_1, !(pageHash->flags & H_INVALID));
    lrc = ((SLOTTEDPAGE *) pageHash->pageInfo->bufFrame)->header.lrc;

    /*
	 *	first, make sure that we have a valid log buffer at the
	 *	tail of the log.  Get a new log page if this isn't the case.
	 */
    if (logInfo->tailBuffer == NULL) {
        CLOG_GetNewLogPage();
    }
	
	/* only small and forwarded objects allowed */
#ifdef DEBUG
	objHdr = (OBJHDR*)opInfo->newData;
   	SM_ASSERT(LEVEL_3, (objHdr->properties | P_FORWARD) == P_FORWARD);
#endif /*DEBUG*/

    /*
     * determine size of the log record 
     */

    /* fixed part of log record is aligned record header, aligned
	write info and enough space to align new image */
    fixedSize = (ALIGN(sizeof(LOGRECORDHDR),BYTESPERWORD)) +
		(ALIGN(sizeof(WRITEINFO),BYTESPERWORD)) + BYTESPERWORD;
		
    /* data portion = object hdr */
    bodySize = sizeof(OBJHDR);

    recSize = bodySize + fixedSize;

    /*
     * set up record info fields which are common to all
     */

    recInfo.type        = LOG_REC_TYPE_USER;
    recInfo.action      = opInfo->action;
    recInfo.flags       = NOFLAGS;
    recInfo.imageCount  = 2;
    recInfo.imageSize[0] = sizeof(WRITEINFO);
    recInfo.imageData[0] = (UONE *) &writeInfo;

    /*
     * determine placement of log record(s) on the page(s)
     */
    
    tailLSN = logInfo->tailLSN;
    offset = (int)(tailLSN & logInfo->pageMask);
    spaceLeft = logInfo->lastUsableByte - offset;

    /* see if we can use space on current log page.  Note: make
       sure that object header is not split across log recs*/

    if ( spaceLeft < (fixedSize + sizeof(OBJHDR))) {

		/* can't take advantage of space in current page, move on to next*/
		CLOG_GetNewLogPage();
		spaceLeft = logInfo->usableBytes;
	}
	    
    /* prepare to write out this log record */
    writeInfo.unique = opInfo->oid.diskAddr.unique;
    writeInfo.slot = opInfo->oid.diskAddr.slot;
    writeInfo.start = 0;

    /* first record includes the header */
    recInfo.imageData[1] = (UONE *)opInfo->newData;
    recInfo.imageSize[1] = sizeof(OBJHDR);


    /* write the log record*/
    forceMark = CLOG_WriteLogRecord(TransRec.clientTid, 
		&recInfo, &lrc, pageHash);
    
    /* mark the page dependency with the last log record written */
     pageHash->forceMark = forceMark;

    /* put the new lrc on the page */
     ((SLOTTEDPAGE *) pageHash->pageInfo->bufFrame)->header.lrc = lrc;

     return esmNOERROR;
}
