/*
 *   $RCSfile: clog_LogDestroyObj.c,v $  
 *   $Revision: 1.1.1.1 $  
 *   $Date: 1996/05/04 21:55:17 $      
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/

#include "sysdefs.h"
#include "ess.h"
#include "checking.h"
#include "trace.h"
#include "error.h"
#include "list.h"
#include "pool.h"
#include "tid.h"
#include "io.h"
#include "lock.h"
#include "object.h"
#include "msgdefs.h"
#include "bf.h"
#include "chunk.h"
#include "lsn.h"
#include "loginfo.h"
#include "volume.h"
#include "trans.h"
#include "clog.h"
#include "logrecs.h"
#include "logaction.h"
#include "sm_state.h"
#include "util_funcs.h"
#include "bf_extfuncs.h"
#include "clog_intfuncs.h"
#include "clog_extfuncs.h"
#include "clog_globals.h"
#include "trans_globals.h"

int
clog_LogDestroyObj(

    OPERATIONINFO	*opInfo 
)
{

    LRC lrc;
    int recSize;
    int fixedSize;
    int bodySize;
    CLIENTLOGRECORDINFO recInfo;
    register LOGINFO *logInfo;
    WRITEINFO writeInfo;
    LSNOFFSET  tailLSN;
    PAGEHASH *pageHash;
    int	offset;
    int spaceLeft, writtenSize, imageSize, usableBytes, leftOver;
    FORCEMARK forceMark;

/***************************************************************
 *  Format of a destroy log record is:
 * 	image[0] - OID,start location, length
 *	image[1] - old data
 *
 *  Initial log records are DELETES 
 *  Final log reocord is DESTROY
 ***************************************************************/
     
    /* Get register pointer to log info structure */
    logInfo = GET_LOGINFO;

    /* first, make sure that we have a valid log buffer at the tail of the
       log.  Get a new log page if this isn't the case.*/

    if (logInfo->tailBuffer == NULL) {
        CLOG_GetNewLogPage();
    }

    /*
     * determine size of the log record 
     */

    /* fixed part of log record is aligned record header, aligned
	write info and enough space to align new image */
    fixedSize = (ALIGN(sizeof(LOGRECORDHDR),BYTESPERWORD)) +
		(ALIGN(sizeof(WRITEINFO),BYTESPERWORD)) + BYTESPERWORD;
		
    /* data portion = object hdr + object data */
    bodySize = (int) opInfo->size + sizeof(OBJHDR);

    recSize = bodySize + fixedSize;

    /*
     * set up record info fields which are common to all
     */

    recInfo.type        = LOG_REC_TYPE_USER;
    recInfo.action      = opInfo->action;
    recInfo.flags       = NOFLAGS;
    recInfo.imageCount  = 2;
    recInfo.imageSize[0] = sizeof(WRITEINFO);
    recInfo.imageData[0] = (UONE *) &writeInfo;
    recInfo.actionPid.page = opInfo->oid.diskAddr.page;
    recInfo.actionPid.volid = opInfo->oid.diskAddr.volid;

    if ((pageHash = bf_FindPage( &(recInfo.actionPid))) == NULL) {
       return(esmFAILURE);
    }
	SM_ASSERT( LEVEL_1, !(pageHash->flags & H_INVALID));
    lrc = ((SLOTTEDPAGE *) pageHash->pageInfo->bufFrame)->header.lrc;

    /*
     * determine placement of log record(s) on the page(s)
     */
    
    tailLSN = logInfo->tailLSN;
    offset = (int)(tailLSN & logInfo->pageMask);
    spaceLeft = logInfo->lastUsableByte - offset;

    /* may need to split log record */
    usableBytes = (logInfo->usableBytes - fixedSize);

    /* now start writing out log records */
    writtenSize = 0;
    while (bodySize > 0) {
        leftOver = (bodySize % usableBytes) + fixedSize;
	
	/* see if we can use space on current log page.  Note: make
	   sure that object header is not split across log recs*/

        if ((leftOver > spaceLeft) || 
			 (spaceLeft < (fixedSize + sizeof(OBJHDR)))) {

            /* can't take advantage of space in current page, move on to next*/
	    CLOG_GetNewLogPage();
	    spaceLeft = logInfo->usableBytes;
	    }
	    
	    /* compute amount of space available for new Data */
	    imageSize = MIN((spaceLeft - fixedSize),bodySize);
	    bodySize -= imageSize;

	    /*adjust space left to account for the size of this record */
	    spaceLeft -= (imageSize + fixedSize);

	    /* check the size calculation*/
	    SM_ASSERT(LEVEL_1, spaceLeft >= 0);

	    /* prepare to write out this log record */
	    if (bodySize == 0) {
	        /* note: last record includes object header */
    	    recInfo.action = LOG_ACTION_DESTROY_SMALL_OBJ;
	        writeInfo.unique = opInfo->oid.diskAddr.unique;
	        writeInfo.slot = opInfo->oid.diskAddr.slot;
	        writeInfo.start = 0;

			/* if we have split up log record using deletes,
			   then we must adjust the object size in the 
			   header of the object to be logged */

            if ((imageSize-sizeof(OBJHDR)) < opInfo->size) {
			   ((OBJHDR *) opInfo->oldData)->size = (imageSize-sizeof(OBJHDR)); 
		    }

            recInfo.imageData[1] = (UONE *)opInfo->oldData; 
			writtenSize += (imageSize-sizeof(OBJHDR));
	    }
	    else {
	        /* all other records are deletes*/
    	    recInfo.action = LOG_ACTION_DELETE_SMALL_OBJ;

	        /* make sure that we aren't splitting the object header */
	        if (bodySize < sizeof(OBJHDR)) {
		        imageSize -= (sizeof(OBJHDR) - bodySize);
		        bodySize = sizeof(OBJHDR);
	        }	
	        writeInfo.unique = opInfo->oid.diskAddr.unique;
	        writeInfo.slot = opInfo->oid.diskAddr.slot;
	        /* delete working backwards from the end of the object */
	        writeInfo.start = (UTWO) (opInfo->size - writtenSize - imageSize);
            recInfo.imageData[1] = (UONE *)opInfo->oldData + writeInfo.start
							+ sizeof(OBJHDR);
			writtenSize += imageSize;
	    }

        recInfo.imageSize[1] = imageSize;


	    /* write the log record*/
	    forceMark = CLOG_WriteLogRecord(TransRec.clientTid, &recInfo, &lrc, pageHash);

    }

     pageHash->forceMark = forceMark;
     ((SLOTTEDPAGE *) pageHash->pageInfo->bufFrame)->header.lrc = lrc;
     return esmNOERROR;
}


int
clog_LogDestroyObjND(
    OPERATIONINFO	*opInfo 
)
{

    LRC lrc;
    int fixedSize;
    CLIENTLOGRECORDINFO recInfo;
    register LOGINFO *logInfo;
    WRITEINFO writeInfo;
    LSNOFFSET  tailLSN;
    PAGEHASH *pageHash;
    int	offset;
    FORCEMARK forceMark;
	int		spaceLeft;
#ifdef DEBUG
	OBJHDR	*objHdr;
#endif /*DEBUG*/

/***************************************************************
 *  Format of a destroy log record is:
 * 	image[0] - OID,start location, length
 *
 ***************************************************************/

	SM_ASSERT(LEVEL_3, opInfo->oldData == NULL);

    /* Get register pointer to log info structure */
    logInfo = GET_LOGINFO;

    /* first, make sure that we have a valid log buffer at the tail of
		the log.  Get a new log page if this isn't the case.*/

    if (logInfo->tailBuffer == NULL) {
        CLOG_GetNewLogPage();
    }

	/* only small and forwarded objects allowed */
#ifdef DEBUG
	objHdr = (OBJHDR*)opInfo->oldData;
	SM_ASSERT(LEVEL_3, (objHdr->properties | P_FORWARD) == P_FORWARD);
#endif /*DEBUG*/


	/* set up the first image with info about the object */
	writeInfo.unique = opInfo->oid.diskAddr.unique;
	writeInfo.slot = opInfo->oid.diskAddr.slot;
	writeInfo.start = 0;
	SM_ASSERT(LEVEL_3, ((OBJHDR*)opInfo->oldData)->size == opInfo->size);

    /* fixed part of log record is aligned record header, aligned
	write info and enough space to align new image */
    fixedSize = (ALIGN(sizeof(LOGRECORDHDR),BYTESPERWORD)) +
				(ALIGN(sizeof(writeInfo),BYTESPERWORD) +
				ALIGN(sizeof(OBJHDR), BYTESPERWORD));
		
    /*
     * set up record info fields which are common to all
     */
    recInfo.type        = LOG_REC_TYPE_USER;
    recInfo.action      = opInfo->action;
    recInfo.flags       = NOFLAGS;
    recInfo.imageCount  = 2;
    recInfo.imageSize[0] = sizeof(writeInfo);
    recInfo.imageData[0] = (UONE *) &writeInfo;
    recInfo.imageSize[1] = sizeof(OBJHDR);
    recInfo.imageData[1] = (UONE *) &opInfo->oldData;
    recInfo.actionPid.page = opInfo->oid.diskAddr.page;
    recInfo.actionPid.volid = opInfo->oid.diskAddr.volid;

    if ((pageHash = bf_FindPage( &(recInfo.actionPid))) == NULL) {
       return(esmFAILURE);
    }
	SM_ASSERT( LEVEL_1, !(pageHash->flags & H_INVALID));
    lrc = ((SLOTTEDPAGE *) pageHash->pageInfo->bufFrame)->header.lrc;

    /*
     * determine placement of log record(s) on the page(s)
     */
    tailLSN = logInfo->tailLSN;
    offset = (int)(tailLSN & logInfo->pageMask);
    spaceLeft = (int) (logInfo->lastUsableByte - offset);

    /* see if we can use space on current log page. */
    if (fixedSize > spaceLeft) {

        /* can't take advantage of space in current page, move on to next*/
		CLOG_GetNewLogPage();
        spaceLeft = (int) logInfo->usableBytes;
    }
	    
    spaceLeft -= (int) fixedSize;

    /* check the size calculation*/
    SM_ASSERT(LEVEL_1, spaceLeft >= 0);


    /* write the log record*/
    forceMark = CLOG_WriteLogRecord(TransRec.clientTid, 
		&recInfo, &lrc, pageHash);

    /* mark the page dependency with the last log record written */
     pageHash->forceMark = forceMark;

    /* put the new lrc on the page */
     ((SLOTTEDPAGE *) pageHash->pageInfo->bufFrame)->header.lrc = lrc;

    return esmNOERROR;
}
