/*
 *   $RCSfile: clog_LogModifyObj.c,v $  
 *   $Revision: 1.1.1.1 $  
 *   $Date: 1996/05/04 21:55:17 $      
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/

#include "sysdefs.h"
#include "ess.h"
#include "checking.h"
#include "trace.h"
#include "error.h"
#include "list.h"
#include "pool.h"
#include "tid.h"
#include "io.h"
#include "lock.h"
#include "object.h"
#include "msgdefs.h"
#include "bf.h"
#include "chunk.h"
#include "lsn.h"
#include "loginfo.h"
#include "volume.h"
#include "trans.h"
#include "clog.h"
#include "logrecs.h"
#include "logaction.h"
#include "sm_state.h"
#include "util_funcs.h"
#include "bf_extfuncs.h"
#include "clog_intfuncs.h"
#include "clog_extfuncs.h"
#include "clog_globals.h"
#include "trans_globals.h"


int
clog_LogModifyObj(

    OPERATIONINFO	*opInfo 
)
{

    LRC lrc;
    int recSize;
    int fixedSize;
    int bodySize;
    CLIENTLOGRECORDINFO recInfo;
    register LOGINFO *logInfo;
    WRITEINFO writeInfo;
    LSNOFFSET  tailLSN;
    PAGEHASH *pageHash;
    int	offset;
    int spaceLeft, writtenSize, imageSize, usableBytes, leftOver;
    FORCEMARK forceMark;
    BOOL	isDelete;

/***************************************************************
 *  This function is used for logging inserts, deletes, and appends
 * 
 *
 *  For INSERTS and APPENDS, the format of a log record is:
 * 	image[0] - OID,start location, length
 *	image[1] - new data 
 *
 *  For DELETES, the format of a log record is:
 * 	image[0] - OID,start location, length
 *	image[1] - old data 
 *
 *  For INSERTS, Additional log records are INSERTS
 *  For DELETES, Additional log records are DELETES
 *  For APPENDS, Additional log records are APPENDS
 ***************************************************************/
     
    /* Get register pointer to log info structure */
    logInfo = GET_LOGINFO;

    /*
	 *	first, make sure that we have a valid log buffer at the
	 *	tail of the log.  Get a new log page if this isn't the case.
	 */
    if (logInfo->tailBuffer == NULL) {
        CLOG_GetNewLogPage();
    }

    /* delete is handled differently from the others */
    isDelete = FALSE;
    if (opInfo->action == LOG_ACTION_DELETE_SMALL_OBJ) {
       isDelete = TRUE;
    }

    /*
     * determine size of the log record 
     */

    /* fixed part of log record is aligned record header, aligned
	write info and enough space to align new image */
    fixedSize = (ALIGN(sizeof(LOGRECORDHDR),BYTESPERWORD)) +
		(ALIGN(sizeof(WRITEINFO),BYTESPERWORD)) + BYTESPERWORD;
		
    /* data portion = object data */
    bodySize = (int) opInfo->size;

    recSize = bodySize + fixedSize;

    /*
     * set up record info fields which are common to all
     */
    recInfo.type        = LOG_REC_TYPE_USER;
    recInfo.action      = opInfo->action;
    recInfo.flags       = NOFLAGS;
    recInfo.imageCount  = 2;
    recInfo.imageSize[0] = sizeof(WRITEINFO);
    recInfo.imageData[0] = (UONE *) &writeInfo;
    recInfo.actionPid.page = opInfo->oid.diskAddr.page;
    recInfo.actionPid.volid = opInfo->oid.diskAddr.volid;

    if ((pageHash = bf_FindPage( &(recInfo.actionPid))) == NULL) {
       return(esmFAILURE);
    }
	SM_ASSERT( LEVEL_1, !(pageHash->flags & H_INVALID));

    lrc = ((SLOTTEDPAGE *) pageHash->pageInfo->bufFrame)->header.lrc;

    /*
     * determine placement of log record(s) on the page(s)
     */
    tailLSN = logInfo->tailLSN;
    offset = (int)(tailLSN & logInfo->pageMask);
    spaceLeft = logInfo->lastUsableByte - offset;

    /* may need to split log record */
    usableBytes = (logInfo->usableBytes - fixedSize);

    /* now start writing out log records */
    writtenSize = 0;
    while (bodySize > 0) {
        leftOver = (bodySize % usableBytes) + fixedSize;
	
	/* see if we can use space on current log page. */

		if (leftOver > spaceLeft) {

			/* can't take advantage of space in current page, move on to next*/
			CLOG_GetNewLogPage();
			spaceLeft = logInfo->usableBytes;
		}
	    
		/* compute amount of space available for new Data */
		imageSize = MIN((spaceLeft - fixedSize),bodySize);
		bodySize -= imageSize;

		/*adjust space left to account for the size of this record */
		spaceLeft -= (imageSize + fixedSize);

		/* check the size calculation*/
		SM_ASSERT(LEVEL_1, spaceLeft >= 0);

		/* prepare to write out this log record */
		writeInfo.unique = opInfo->oid.diskAddr.unique;
		writeInfo.slot = opInfo->oid.diskAddr.slot;
		if (isDelete) {
			writeInfo.start = (UTWO) opInfo->start;
			recInfo.imageData[1] = (UONE *)opInfo->oldData + writtenSize;
		}
		else {
			writeInfo.start = (UTWO) (writtenSize + opInfo->start);
			recInfo.imageData[1] = (UONE *)opInfo->newData + writtenSize;
		}
		recInfo.imageSize[1] = imageSize;

		writtenSize += imageSize;

		/* write the log record*/
		forceMark = CLOG_WriteLogRecord(TransRec.clientTid, &recInfo, &lrc, pageHash);
    }

    /* mark the page dependency with the last log record written */
     pageHash->forceMark = forceMark;

    /* put the new lrc on the page */
     ((SLOTTEDPAGE *) pageHash->pageInfo->bufFrame)->header.lrc = lrc;

    return esmNOERROR;
}



int
clog_LogModifyObjND(
    OPERATIONINFO	*opInfo 
)
{
    LRC 		lrc;
    int 		fixedSize;
    CLIENTLOGRECORDINFO recInfo;
    register LOGINFO *logInfo;
    FULLWRITEINFO writeInfo;
    LSNOFFSET 	tailLSN;
    PAGEHASH 	*pageHash;
    int			offset;
    FORCEMARK 	forceMark;
    int 		spaceLeft;

/***************************************************************
 *  This function is used for logging inserts, deletes, and appends
 *	where no data is to be logged
 * 
 *
 *  For INSERTS and APPENDS, the format of a log record is:
 * 	image[0] - OID,start location, length
 *
 *  For DELETES, the format of a log record is:
 * 	image[0] - OID,start location, length
 *
 ***************************************************************/
 
#ifdef DEBUG
    switch (opInfo->action) {
    case LOG_ACTION_INSERT_SMALL_OBJ_ND:
    case LOG_ACTION_APPEND_SMALL_OBJ_ND:
	SM_ASSERT(LEVEL_3, opInfo->newData == NULL);
	break;
    case LOG_ACTION_DELETE_SMALL_OBJ_ND:
	SM_ASSERT(LEVEL_3, opInfo->oldData == NULL);
	break;
    default:
	SM_ERROR(TYPE_FATAL, esmINTERNAL);
	break;
    }
#endif

    /* Get register pointer to log info structure */
    logInfo = GET_LOGINFO;

    /*
	 *	first, make sure that we have a valid log buffer at the
	 *	tail of the log.  Get a new log page if this isn't the case.
	 */
    if (logInfo->tailBuffer == NULL) {
        CLOG_GetNewLogPage();
    }

    /*
     * determine size of the log record 
     * fixed part of log record is aligned record header, aligned
	 * write info 
     */
    fixedSize = (ALIGN(sizeof(LOGRECORDHDR),BYTESPERWORD)) +
		(ALIGN(sizeof(writeInfo),BYTESPERWORD));
		
    /*
     * set up record info fields which are common to all
     */
    recInfo.type        = LOG_REC_TYPE_USER;
    recInfo.action      = opInfo->action;
    recInfo.flags       = NOFLAGS;
    recInfo.imageCount  = 1;
    recInfo.imageSize[0] = sizeof(writeInfo);
    recInfo.imageData[0] = (UONE *) &writeInfo;
    recInfo.actionPid.page = opInfo->oid.diskAddr.page;
    recInfo.actionPid.volid = opInfo->oid.diskAddr.volid;

    if ((pageHash = bf_FindPage( &(recInfo.actionPid))) == NULL) {
       return(esmFAILURE);
    }
	SM_ASSERT( LEVEL_1, !(pageHash->flags & H_INVALID));

    lrc = ((SLOTTEDPAGE *) pageHash->pageInfo->bufFrame)->header.lrc;

    /*
     * determine placement of log record(s) on the page(s)
     */
    tailLSN = logInfo->tailLSN;
    offset = (int)(tailLSN & logInfo->pageMask);
    spaceLeft = logInfo->lastUsableByte - offset;

    /* see if we can use space on current log page. */
    if (fixedSize > spaceLeft) {

	/* can't take advantage of space in current page, move on to next*/
	CLOG_GetNewLogPage();
        spaceLeft = logInfo->usableBytes;
    }
	    
    spaceLeft -= fixedSize;

    /* check the size calculation*/
    SM_ASSERT(LEVEL_1, spaceLeft >= 0);

	writeInfo.common.unique = opInfo->oid.diskAddr.unique;
	writeInfo.common.slot = opInfo->oid.diskAddr.slot;
	writeInfo.common.start = 0;
	writeInfo.size = (UTWO) opInfo->size;

    /* write the log record*/
    forceMark = CLOG_WriteLogRecord(TransRec.clientTid, 
		&recInfo, &lrc, pageHash);

    /* mark the page dependency with the last log record written */
     pageHash->forceMark = forceMark;

    /* put the new lrc on the page */
     ((SLOTTEDPAGE *) pageHash->pageInfo->bufFrame)->header.lrc = lrc;

    return esmNOERROR;
}
