/*
 *   $RCSfile: clog_LogWriteLgPage.c,v $  
 *   $Revision: 1.1.1.1 $  
 *   $Date $
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/

#include "sysdefs.h"
#include "ess.h"
#include "checking.h"
#include "trace.h"
#include "error.h"
#include "list.h"
#include "pool.h"
#include "tid.h"
#include "io.h"
#include "lock.h"
#include "object.h"
#include "msgdefs.h"
#include "bf.h"
#include "chunk.h"
#include "lsn.h"
#include "loginfo.h"
#include "volume.h"
#include "trans.h"
#include "clog.h"
#include "logrecs.h"
#include "logaction.h"
#include "sm_state.h"
#include "util_funcs.h"
#include "bf_extfuncs.h"
#include "clog_intfuncs.h"
#include "clog_extfuncs.h"
#include "clog_globals.h"
#include "trans_globals.h"


int
clog_LogWriteLgPage(

    OPERATIONINFO	*opInfo 
)
{

    LRC *lrcPtr;
    int recSize;
    int fixedSize;
    int bodySize;
    CLIENTLOGRECORDINFO recInfo;
    register LOGINFO *logInfo;
    LGWRITEINFO writeInfo;
    LSNOFFSET  tailLSN;
    PAGEHASH *pageHash;
    int	offset;
    int spaceLeft, writtenSize, imageSize, usableBytes, leftOver;
    FORCEMARK forceMark;

/***************************************************************
 *  This function is used to log writes to data stored in a leaf
 *	page of a large object or slots of a large object node. 
 *  Format of a write log record is:
 * 	image[0] - PID,start location, length, rootSlot
 *	image[1] - new data
 *	image[2] - old data
 *
 *	NOTE, rootSlot will be NIL unless affected object is a root object
 *			on a slotted page.
 *
 *  Additional log records are WRITEs 
 ***************************************************************/
     
    /* Get register pointer to log info structure */
    logInfo = GET_LOGINFO;

    SM_ASSERT(LEVEL_3, (opInfo->size + opInfo->start) <= LGDATA_USABLESIZE);


#ifdef DEBUG
    /* See if old and new data are identical */
    if (bcmp(opInfo->oldData, opInfo->newData, (int) opInfo->size) == 0) {
        SM_ERROR(TYPE_LOG, esmINTERNAL);
    }
#endif

    /*
	 *	first, make sure that we have a valid log buffer at the
	 *	tail of the log.  Get a new log page if this isn't the case.
	 */
    if (logInfo->tailBuffer == NULL) {
        CLOG_GetNewLogPage();
    }

    /*
     * determine size of the log record 
     */
    /* fixed part of log record is aligned record header, aligned
	write info and enough space to align new and old images */
    fixedSize = (ALIGN(sizeof(LOGRECORDHDR),BYTESPERWORD)) +
		(ALIGN(sizeof(LGWRITEINFO),BYTESPERWORD)) + (2 * BYTESPERWORD);
		
    bodySize = 2 * (int) opInfo->size;
    recSize = bodySize + fixedSize;

    /*
     * set up record info fields which are common to all
     */

    recInfo.type        = LOG_REC_TYPE_USER;
    recInfo.action      = opInfo->action;
    recInfo.flags       = NOFLAGS;
    recInfo.imageCount  = 3;
    recInfo.imageSize[0] = sizeof(LGWRITEINFO);
    recInfo.imageData[0] = (UONE *) &writeInfo;
    recInfo.actionPid = opInfo->actionPid;

    if ((pageHash = bf_FindPage( &(recInfo.actionPid))) == NULL) {
       return(esmFAILURE);
    }
	SM_ASSERT( LEVEL_1, !(pageHash->flags & H_INVALID));

	lrcPtr = bf_lrcLoc(pageHash);

#ifdef DEBUG
    if (opInfo->action == LOG_ACTION_WRITE_LARGE_DATA) {
        /* for large datapages, the lrc is stored in the pageHash */
		SM_ASSERT(LEVEL_1, (lrcPtr == &(pageHash->lrc)));
    } else {
        if (opInfo->action == LOG_ACTION_WRITE_LARGE_SLOTS) {
            /* for large interior nodes, the lrc is stored in the header*/
		    if (opInfo->rootSlot == NIL) {
				SM_ASSERT(LEVEL_1, (lrcPtr == &(((LGNODE *) (pageHash->pageInfo->bufFrame))->header.lrc)));
			} else {
				SM_ASSERT(LEVEL_1, (lrcPtr == &(((SLOTTEDPAGE *) (pageHash->pageInfo->bufFrame))->header.lrc)));
			}
		} else {
			SM_ERROR(TYPE_FATAL, esmINTERNAL);
		}
    }
#endif DEBUG

    /*
     * determine placement of log record(s) on the page(s)
     */
    
    tailLSN = logInfo->tailLSN;
    offset = (int)(tailLSN & logInfo->pageMask);
    spaceLeft = logInfo->lastUsableByte - offset;

    /* may need to split log record */
    usableBytes = (logInfo->usableBytes - fixedSize);

    /* now start writing out log records */
    writtenSize = 0;
    while (bodySize > 0) {
        leftOver = (bodySize % usableBytes) + fixedSize;
        if (leftOver > spaceLeft ) {
            /* can't take advantage of space in current page,
							move on to next*/
			CLOG_GetNewLogPage();
			spaceLeft = logInfo->usableBytes;
		}
	    
		/* compute amount of space available for total of new and old Data */
		imageSize = MIN((spaceLeft - fixedSize),bodySize);
		bodySize -= imageSize;

		/*adjust space left to accound for the size of this record */
		spaceLeft -= (imageSize + fixedSize);

		/* check the size calculation*/
		SM_ASSERT(LEVEL_1, spaceLeft >= 0);

		/* compute the size of a single image (e.g., the oldData) */ 
		imageSize /= 2;

		/* prepare to write out this log record */
		writeInfo.start = (UTWO) (opInfo->start + writtenSize);
		writeInfo.size = (UTWO) imageSize;
		writeInfo.rootSlot = (TWO) opInfo->rootSlot;

		recInfo.imageData[1] = (UONE *)opInfo->newData + writtenSize;
		recInfo.imageSize[2] = recInfo.imageSize[1] = imageSize;
		recInfo.imageData[2] = (UONE *)opInfo->oldData + writtenSize;

		/*
		 *	write the log record
		 *  note: this will put the new lrc on the page or in
		 *		the pageHash
		 * TODO: remove lrcPtr arg from CLOG_WriteLogRecord and
		 * just have CLOG_WriteLogRecord get it from the pageHash,
		 * using bf_lrcLoc.
		 */
		forceMark = CLOG_WriteLogRecord(TransRec.clientTid, &recInfo, lrcPtr, pageHash);
		writtenSize += imageSize;
    }

    /* mark the page dependency with the last log record written */
    pageHash->forceMark = forceMark;

    return esmNOERROR;
}


int
clog_LogWriteLgPageND(
    OPERATIONINFO	*opInfo 
)
{

    LRC *lrcPtr;
    int recSize;
    int fixedSize;
    int bodySize;
    CLIENTLOGRECORDINFO recInfo;
    LOGINFO *logInfo;
    LGWRITEINFO writeInfo;
    LSNOFFSET  tailLSN;
    PAGEHASH *pageHash;
    int	offset;
    int spaceLeft, writtenSize, imageSize, usableBytes, leftOver;
    FORCEMARK forceMark;

/***************************************************************
 *  This function is used to log writes to data stored in a leaf
 *	page of a large object or slots of a large object node. 
 *  Format of a write log record is:
 * 	image[0] - PID,start location, length, rootSlot
 *	image[1] - old data
 *
 *	NOTE, rootSlot will be NIL unless affected object is a root object
 *			on a slotted page.
 *
 *  Additional log records are WRITEs 
 ***************************************************************/
    
	SM_ASSERT(LEVEL_3, opInfo->newData == NULL);

    /* Get register pointer to log info structure */
    logInfo = GET_LOGINFO;

    /*
	 *	first, make sure that we have a valid log buffer at the
	 *	tail of the log.  Get a new log page if this isn't the case.
	 */
    if (logInfo->tailBuffer == NULL) {
        CLOG_GetNewLogPage();
    }

    /*
     * determine size of the log record 
     */

    /* fixed part of log record is aligned record header, aligned
	write info and enough space to align new and old images */
    fixedSize = (ALIGN(sizeof(LOGRECORDHDR),BYTESPERWORD)) +
		(ALIGN(sizeof(LGWRITEINFO),BYTESPERWORD)) + (2 * BYTESPERWORD);
		
    bodySize = (int) opInfo->size;
    recSize = bodySize + fixedSize;

    /*
     * set up record info fields which are common to all
     */

    recInfo.type        = LOG_REC_TYPE_USER;
    recInfo.action      = opInfo->action;
    recInfo.flags       = NOFLAGS;
    recInfo.imageCount  = 2;
    recInfo.imageSize[0] = sizeof(LGWRITEINFO);
    recInfo.imageData[0] = (UONE *) &writeInfo;
    recInfo.actionPid = opInfo->actionPid;

    if ((pageHash = bf_FindPage( &(recInfo.actionPid))) == NULL) {
       return(esmFAILURE);
    }
	SM_ASSERT( LEVEL_1, !(pageHash->flags & H_INVALID));

	lrcPtr = bf_lrcLoc(pageHash);

#ifdef DEBUG
    if (opInfo->action == LOG_ACTION_WRITE_LARGE_DATA_ND) {
        /* for large datapages, the lrc is stored in the pageHash */
        SM_ASSERT(LEVEL_1, (lrcPtr == &(pageHash->lrc)));
    } else {
			SM_ERROR(TYPE_FATAL, esmINTERNAL);
    }
#endif DEBUG

    /*
     * determine placement of log record(s) on the page(s)
     */
    
    tailLSN = logInfo->tailLSN;
    offset = (int)(tailLSN & logInfo->pageMask);
    spaceLeft = logInfo->lastUsableByte - offset;

    /* may need to split log record */
    usableBytes = (logInfo->usableBytes - fixedSize);

    /* now start writing out log records */
    writtenSize = 0;
    while (bodySize > 0) {
        leftOver = (bodySize % usableBytes) + fixedSize;
        if (leftOver > spaceLeft ) {
            /* can't take advantage of space in current page,
							move on to next*/
			CLOG_GetNewLogPage();
			spaceLeft = logInfo->usableBytes;
		}
	    
		/* compute amount of space available for total of new and old Data */
		imageSize = MIN((spaceLeft - fixedSize),bodySize);
		bodySize -= imageSize;

		/*adjust space left to accound for the size of this record */
		spaceLeft -= (imageSize + fixedSize);

		/* check the size calculation*/
		SM_ASSERT(LEVEL_1, spaceLeft >= 0);

		/* prepare to write out this log record */
		writeInfo.start = (UTWO) (opInfo->start + writtenSize);
		writeInfo.size =  (UTWO) imageSize;
		writeInfo.rootSlot = (TWO) opInfo->rootSlot;

		recInfo.imageData[1] = (UONE *)opInfo->oldData + writtenSize;
		recInfo.imageSize[1] = imageSize;

		/*
		 *	write the log record
		 *  note: this will put the new lrc on the page or in the
		 *	pageHash
		 */
		forceMark = CLOG_WriteLogRecord(TransRec.clientTid, &recInfo, lrcPtr, pageHash);
		writtenSize += imageSize;
	}

    /* mark the page dependency with the last log record written */
	pageHash->forceMark = forceMark;

	return esmNOERROR;
}
