/*
 *   $RCSfile: clog_LogWriteObj.c,v $  
 *   $Revision: 1.1.1.1 $  
 *   $Date: 1996/05/04 21:55:17 $      
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/

#include "sysdefs.h"
#include "ess.h"
#include "checking.h"
#include "trace.h"
#include "error.h"
#include "list.h"
#include "pool.h"
#include "tid.h"
#include "io.h"
#include "lock.h"
#include "object.h"
#include "msgdefs.h"
#include "bf.h"
#include "chunk.h"
#include "lsn.h"
#include "loginfo.h"
#include "volume.h"
#include "trans.h"
#include "clog.h"
#include "logrecs.h"
#include "logaction.h"
#include "sm_state.h"
#include "util_funcs.h"
#include "bf_extfuncs.h"
#include "clog_intfuncs.h"
#include "clog_extfuncs.h"
#include "clog_globals.h"
#include "trans_globals.h"
#include "sm_globals.h"


int
clog_LogWriteObj(

    OPERATIONINFO	*opInfo 
)
{

    LRC 		lrc;
    int 		recSize;
    int 		fixedSize;
    int 		bodySize;
    CLIENTLOGRECORDINFO recInfo;
    LOGINFO 	*logInfo;
    WRITEINFO 	writeInfo;
    LSNOFFSET 	tailLSN;
    PAGEHASH 	*pageHash;
    int			offset, spaceLeft;
    int 		writtenSize, imageSize, usableBytes, leftOver;
    FORCEMARK 	forceMark;

/***************************************************************
 *  Format of a write log record is:
 * 	image[0] - OID,start location, length
 *	image[1] - new data
 *	image[2] - old data
 *
 *  Additional log records are WRITEs 
 ***************************************************************/
    

    /* get info on affected page */
    recInfo.actionPid.page = opInfo->oid.diskAddr.page;
    recInfo.actionPid.volid = opInfo->oid.diskAddr.volid;

	FAKEFAIL(esmBADVOLID);

    if ((pageHash = bf_FindPage( &(recInfo.actionPid))) == NULL) {
FailureReturn:
       return esmFAILURE;
    }
	SM_ASSERT( LEVEL_1, !(pageHash->flags & H_INVALID));

    lrc = ((SLOTTEDPAGE *) pageHash->pageInfo->bufFrame)->header.lrc;

    /* Get register pointer to log info structure */
    logInfo = GET_LOGINFO;

    /*
	 *	first, make sure that we have a valid log buffer at the
	 *	tail of the log.  Get a new log page if this isn't the case.
	 */
    if (logInfo->tailBuffer == NULL) {
        if(CLOG_GetNewLogPage()) {
			FAIL
		}
    }


    /*
     * determine size of the log record 
     */

    /* fixed part of log record is aligned record header, aligned
	write info and enough space to align new and old images */
    fixedSize = (ALIGN(sizeof(LOGRECORDHDR),BYTESPERWORD)) +
		(ALIGN(sizeof(WRITEINFO),BYTESPERWORD)) + (2 * BYTESPERWORD);
		
    bodySize = 2 * (int)opInfo->size;
    recSize = bodySize + fixedSize;

    /*
     * set up record info fields which are common to all
     */
    recInfo.type        = LOG_REC_TYPE_USER;
    recInfo.action      = opInfo->action;
    recInfo.flags       = NOFLAGS;
    recInfo.imageCount  = 3;
    recInfo.imageSize[0] = sizeof(writeInfo);
    recInfo.imageData[0] = (UONE *) &writeInfo;

    /*
     * determine placement of log record(s) on the page(s)
     */
    
    tailLSN = logInfo->tailLSN;
    offset = (int)(tailLSN & logInfo->pageMask);
    spaceLeft = logInfo->lastUsableByte - offset;

    /* may need to split log record */
    usableBytes = (logInfo->usableBytes - fixedSize);

    /* now start writing out log records */
    writtenSize = 0;
    while (bodySize > 0) {
        leftOver = (bodySize % usableBytes) + fixedSize;
        if (leftOver > spaceLeft ) {
            /* can't take advantage of space in current page,
							move on to next*/
			if(CLOG_GetNewLogPage()) {
				FAIL
			}
			spaceLeft = logInfo->usableBytes;
		}
	    
		/* compute amount of space available for total of new and old Data */
		imageSize = MIN((spaceLeft - fixedSize),bodySize);
		bodySize -= imageSize;

		/*adjust space left to accound for the size of this record */
		spaceLeft -= (imageSize + fixedSize);

		/* check the size calculation*/
		SM_ASSERT(LEVEL_1, spaceLeft >= 0);

		/* compute the size of a single image (e.g., the oldData) */ 
		imageSize /= 2;

		/* prepare to write out this log record */
		writeInfo.unique = opInfo->oid.diskAddr.unique;
		writeInfo.slot = opInfo->oid.diskAddr.slot;
		writeInfo.start = (int)opInfo->start + writtenSize;

		recInfo.imageData[1] = (UONE *)(opInfo->newData) + writtenSize;
		recInfo.imageSize[2] = recInfo.imageSize[1] = imageSize;
		recInfo.imageData[2] = (UONE *)(opInfo->oldData) + writtenSize;

		/* write the log record*/
		forceMark = CLOG_WriteLogRecord(TransRec.clientTid, &recInfo, &lrc, pageHash);

		writtenSize += imageSize;
    }

    /* mark the page dependency with the last log record written */
     pageHash->forceMark = forceMark;

    /* put the new lrc on the page */
     ((SLOTTEDPAGE *) pageHash->pageInfo->bufFrame)->header.lrc = lrc;

     return esmNOERROR;
}


int
clog_LogWriteObjND(
    OPERATIONINFO	*opInfo 
)
{
    LRC lrc;
    int recSize;
    int fixedSize;
    int bodySize;
    CLIENTLOGRECORDINFO recInfo;
    register LOGINFO *logInfo;
    FULLWRITEINFO writeInfo;
    LSNOFFSET  tailLSN;
    PAGEHASH *pageHash;
    int offset, spaceLeft;
    unsigned short  writtenSize, usableBytes, leftOver;
	unsigned short imageSize;
    FORCEMARK forceMark;

/***************************************************************
 *  Format of a write log record is:
 * 	image[0] - OID,start location, length
 *	image[1] - old data
 *
 *  Additional log records are WRITEs 
 ***************************************************************/
    
	SM_ASSERT(LEVEL_3, opInfo->newData == NULL);

    /* get info on affected page */
    recInfo.actionPid.page = opInfo->oid.diskAddr.page;
    recInfo.actionPid.volid = opInfo->oid.diskAddr.volid;

    if ((pageHash = bf_FindPage( &(recInfo.actionPid))) == NULL) {
FailureReturn:
       return esmFAILURE;
    }
	SM_ASSERT( LEVEL_1, !(pageHash->flags & H_INVALID));

    lrc = ((SLOTTEDPAGE *) pageHash->pageInfo->bufFrame)->header.lrc;

    /* Get register pointer to log info structure */
    logInfo = GET_LOGINFO;

    /*
	 *	first, make sure that we have a valid log buffer at the
	 *	tail of the log.  Get a new log page if this isn't the case.
	 */
    if (logInfo->tailBuffer == NULL) {
        if(CLOG_GetNewLogPage()) {
			FAIL
		}
    }


    /*
     * determine size of the log record 
     */

    /* fixed part of log record is aligned record header, aligned
	write info and enough space to align new and old images */
    fixedSize = (ALIGN(sizeof(LOGRECORDHDR),BYTESPERWORD)) +
		(ALIGN(sizeof(WRITEINFO),BYTESPERWORD)) + (2 * BYTESPERWORD);
		
    bodySize = (int)opInfo->size;
    recSize = bodySize + fixedSize;

    /*
     * set up record info fields which are common to all
     */

    recInfo.type        = LOG_REC_TYPE_USER;
    recInfo.action      = opInfo->action;
    recInfo.flags       = NOFLAGS;
    recInfo.imageCount  = 2;
    recInfo.imageSize[0] = sizeof(writeInfo);
    recInfo.imageData[0] = (UONE *) &writeInfo;

    /*
     * determine placement of log record(s) on the page(s)
     */
    
    tailLSN = logInfo->tailLSN;
    offset = (int)(tailLSN & logInfo->pageMask);
    spaceLeft = logInfo->lastUsableByte - offset;

    /* may need to split log record */
    usableBytes = (logInfo->usableBytes - fixedSize);

    /* now start writing out log records */
    writtenSize = 0;
    while (bodySize > 0) {
        leftOver = (bodySize % usableBytes) + fixedSize;
        if (leftOver > spaceLeft ) {
            /* can't take advantage of space in current page,
							move on to next*/
			if(CLOG_GetNewLogPage()) {
				FAIL
			}
			spaceLeft = logInfo->usableBytes;
		}
	    
		/* compute amount of space available for total of new and old Data */
		imageSize = MIN((((int)spaceLeft) - fixedSize),bodySize);
		bodySize -= imageSize;

		/*adjust space left to accound for the size of this record */
		spaceLeft -= (imageSize + fixedSize);

		/* check the size calculation*/
		SM_ASSERT(LEVEL_1, spaceLeft >= 0);

		/* prepare to write out this log record */
		writeInfo.common.unique = opInfo->oid.diskAddr.unique;
		writeInfo.common.slot = opInfo->oid.diskAddr.slot;
		writeInfo.common.start = (UTWO)opInfo->start + writtenSize;
		writeInfo.size = (UTWO) imageSize;

		recInfo.imageData[1] = (UONE *)opInfo->oldData + writtenSize;
		recInfo.imageSize[1] = imageSize;

		/* write the log record*/
		forceMark = CLOG_WriteLogRecord(TransRec.clientTid, 
			&recInfo, &lrc, pageHash);
		writtenSize += imageSize;
    }

    /* mark the page dependency with the last log record written */
     pageHash->forceMark = forceMark;

    /* put the new lrc on the page */
     ((SLOTTEDPAGE *) pageHash->pageInfo->bufFrame)->header.lrc = lrc;

     return esmNOERROR;
}



