/*
 *   $RCSfile: fi_FileInfoTable.C,v $  
 *   $Revision: 1.1.1.1 $  
 *   $Date: 1996/05/04 21:55:19 $      
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/

#include "sysdefs.h"
#include "ess.h"
#include "checking.h"
#include "trace.h"
#include "error.h"
#include "list.h"
#include "tid.h"
#include "io.h"
#include "lock.h"
#include "object.h"
#include "bf.h"
#include "volume.h"
#include "trans.h"
#include "file.h"
#include "sm_state.h"
#include "fi_extfuncs.h"
#include "io_extfuncs.h"


FILEINFO	*FileInfoTable = NULL;
static int	MaxFileInfoTable = 20;
static int	FileInfoCount = 0;
FLAGS		DefaultFileLogLevel = LOG_ALL;
FILEINFO	*LastFileLookup = NULL;		/* to speed lookups */

/* set to TRUE if logging is turned off for any file */
BOOL		LoggingOffForSomeFile = FALSE;  

 int
fi_InitFileInfoTable()
{
	int 	i;

	TRPRINT(TR_FI, TR_LEVEL_1, (""));

	if (FileInfoTable == NULL) {
		GET(FileInfoTable, FILEINFO, MaxFileInfoTable);
	}
	
	DefaultFileLogLevel = LOG_ALL;
	LastFileLookup = FileInfoTable;
	FileInfoCount = 0;
	for (i = 0; i < MaxFileInfoTable; i++) {
		FileInfoTable[i].fid.pid.volid = 0;
		FileInfoTable[i].logLevel = DefaultFileLogLevel;
		FileInfoTable[i].tempVolume = FALSE;
	}
	return esmNOERROR;
}


 void
fi_FreeFileInfoTable()
{
	SM_FREE_PTR(FileInfoTable);
}

  static FILEINFO*
fi_FindFreeFileInfoSlot()
{
	FILEINFO* 	f = NULL;
	static int	LastRemoved = -1;
	int			i;

	if (FileInfoCount != MaxFileInfoTable) {
		for (i = 0; i < MaxFileInfoTable; i++) {
			if (FileInfoTable[i].fid.pid.volid == 0) {
				f = &(FileInfoTable[i]);
				break;
			}
		}
		SM_ASSERT(LEVEL_3, f != NULL);
	} else {
		if (LastRemoved < 0 || LastRemoved >= MaxFileInfoTable) LastRemoved = 0;
		while (LastRemoved < MaxFileInfoTable) {
			if (FileInfoTable[LastRemoved].tempVolume || 
				FileInfoTable[LastRemoved].logLevel == DefaultFileLogLevel) {
				/* this entry can be safely reused */
				f = &(FileInfoTable[LastRemoved]);
				f->fid.pid.volid = 0;
				FileInfoCount--;
				LastRemoved++;
				break;
			}
			LastRemoved++;
		}
	}

	if (f != NULL) FileInfoCount++;
	return f;
}


 FILEINFO*
fi_FindFileInfo (

	FID			*fid,
	BOOL		add			/* add this file if not found */
)
{
	int 		i;
	FILEINFO	*fileInfo;
	VOLREC		*volRec;

	TRPRINT(TR_FI, TR_LEVEL_1, ("fid:p%d:v%d:u%d", fid->pid.page, (int)fid->pid.volid, fid->unique));

	if (FIDEQ(LastFileLookup->fid, *fid)) {
		return(LastFileLookup);
	}

	fileInfo = NULL;
	for (i = 0; i < MaxFileInfoTable; i++) {
		if (FIDEQ(FileInfoTable[i].fid, *fid)) {
			fileInfo = &(FileInfoTable[i]);
			break;
		}
	}

	if (fileInfo == NULL && add) {
		// find an empty slot
		fileInfo = fi_FindFreeFileInfoSlot();	
		if (fileInfo == NULL) {
			// there is no free space in the file info table
			SM_ERROR(TYPE_CRASH, esmINTERNAL);
			return(NULL);
		}

		fileInfo->fid = *fid;
		fileInfo->firstPageHint = NULLPID;
		fileInfo->lastPageHint = NULLPID;

		if ( (volRec = io_FindVolId(fid->pid.volid)) != NULL) {
			/* see if we are dealing with a temporary volume */
			if (volRec->properties & VOL_TEMP) {
				fileInfo->tempVolume = TRUE;
				/* not logging is performed on temporary volumes */
				fileInfo->logLevel = LOG_NONE;
				LoggingOffForSomeFile = TRUE;
			} else {
				fileInfo->tempVolume = FALSE;
				fileInfo->logLevel = DefaultFileLogLevel;
				if (DefaultFileLogLevel != LOG_ALL) {
					LoggingOffForSomeFile = TRUE;
				}
			}
		} else {
			SM_ERROR(TYPE_USER, esmBADVOLID);
			return(NULL);
		}
	}

	LastFileLookup = fileInfo;

	return(fileInfo);
}


 void
fi_RemoveFileInfo (FID * fid)
{
	int 		i;
	FILEINFO	*fileInfo = NULL;

	TRPRINT(TR_FI, TR_LEVEL_1, ("fid:p%d:v%d:u%d", fid->pid.page, (int)fid->pid.volid, fid->unique));

	if (FIDEQ(LastFileLookup->fid, *fid)) {
		fileInfo = LastFileLookup;
	} else {
		for (i = 0; i < MaxFileInfoTable; i++) {
			if (FIDEQ(FileInfoTable[i].fid, *fid)) {
				fileInfo = &(FileInfoTable[i]);
				break;
			}
		}
	}

	if (fileInfo != NULL) {
		fileInfo->fid.pid.volid = 0;
		fileInfo->logLevel = DefaultFileLogLevel;
		FileInfoCount--;
	}
	return;
}


 void 
fi_SetAllFileLogLevel( FLAGS loggingLevel)
{
	int		i;

	DefaultFileLogLevel = loggingLevel;

	if (DefaultFileLogLevel != LOG_ALL) {
		LoggingOffForSomeFile = TRUE;
	}

	for (i = 0; i < MaxFileInfoTable; i++) {
		/* set the level if it does not belong to a temporary volume */
		if ( FileInfoTable[i].fid.pid.volid != 0 &&
			 !FileInfoTable[i].tempVolume) {
			FileInfoTable[i].logLevel = loggingLevel;
		}
	}
}


/* 
 *	This function returns true if logging has been turned off
 *	for any file.
 */
 BOOL
fi_IsLoggingOff()
{
	return(LoggingOffForSomeFile);
}

/* 
 *	This function records whether logging has been turned off for
 *	any file.
 */
 void
fi_MarkLoggingOff()
{
	int		i;

	if (DefaultFileLogLevel != LOG_ALL) {
		LoggingOffForSomeFile = TRUE;
		return;
	}
	for (i = 0; i < MaxFileInfoTable; i++) {
		if (FileInfoTable[i].fid.pid.volid != 0 &&
			FileInfoTable[i].logLevel != LOG_ALL) {

			LoggingOffForSomeFile = TRUE;
			return;
		}
	}
	LoggingOffForSomeFile = FALSE;
}


/* 
 *	This function invalidates the first/last page hints
 */
 void
fi_InvalidateHints()
{
	int			i;
	FILEINFO	*fileInfo;

	for (i = 0; i < MaxFileInfoTable; i++) {
		fileInfo = FileInfoTable+i;
		if (fileInfo->fid.pid.volid != 0) {
			fileInfo->firstPageHint = NULLPID;
			fileInfo->lastPageHint = NULLPID;
			if (fileInfo->tempVolume || 
				fileInfo->logLevel == DefaultFileLogLevel) {
				/*
				 * there is no special info about this file
				 * so it can be removed from the table.
				 */
				fileInfo->fid.pid.volid = 0;
				FileInfoCount--;
			}
		}
	}
}
