/*
 *   $RCSfile: IO_ReadPageWithNeighbor.c,v $
 *   $Revision: 1.1.1.1 $
 *   $Date: 1996/05/04 21:55:21 $
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/


#include "sysdefs.h"
#include "ess.h"
#include "checking.h"
#include "trace.h"
#include "error.h"
#include "list.h"
#include "tid.h"
#include "io.h"
#include "object.h"
#include "bf.h"
#include "lock.h"
#include "msgdefs.h"
#include "fi_msg.h"
#include "volume.h"
#include "trans.h"
#include "lsn.h"
#include "loginfo.h"
#include "clog.h"
#include "chunk.h"
#include "load.h"
#include "scan.h"
#include "sm_params.h"
#include "btree.h"
#include "serverinfo.h"
#include "sm_state.h"
#include "io_extfuncs.h"
#include "io_intfuncs.h"
#include "bf_extfuncs.h"
#include "lm_extfuncs.h"
#include "msg_funcs.h"
#include "clog_globals.h"
#include "io_globals.h"
#include "msg_globals.h"
#include "trans_globals.h"
#include "sm_globals.h"

 int
IO_ReadPageWithNeighbor (
	const 	PID		*pid,
	char			*buffer,
	LOCKMODE		*lockMode,
	PAGEHASH		*pageHash,
	FLAGS			flags,
	FID				*fid,
	PID				*neighborPid
)
{
	MESSAGE			message;
	LRC				*pageLRC;
	BOOL			retryFlag = TRUE;
	VOLREC			*volRec;
	LOGINFO			*logInfo;

	TRPRINT(TR_IO, TR_LEVEL_1, ("read page:%d", pid->page));
	if(flags & REPLY_WITH_NEIGHBOR) {
		SM_ASSERT(LEVEL_1, flags & NEIGHBOR_FLAGS);
		SM_ASSERT(LEVEL_1, fid != NULL);
		SM_ASSERT(LEVEL_1, neighborPid != NULL);
	}

	/* check to see if the volume is mounted */
	if ((volRec = io_FindVolId(pid->volid)) == NULL)	{
		SM_ERROR(TYPE_USER, esmBADVOLID);
		return esmFAILURE;
	}
	logInfo = &(volRec->serverInfo->logInfo);

	SM_ASSERT(LEVEL_3, ! lm_IsLockCached(pid));

again:

	message.header.type			  = READ_PAGE;
	message.header.replyRequested = TRUE;

	/* message.header.magic is set in callServer */
	message.header.params.in.tid	= GET_SERVERTID;

	SM_ASSERT(LEVEL_1, (flags <= (FLAGS)0xff));
	message.body.page.flags = 	(UONE)(flags & 0xff);
	message.body.page.pid		  = *pid;
	if(flags & REPLY_WITH_NEIGHBOR) {
		message.body.page.Fid		  = *fid;
	}
	message.body.page.page2size   = pageHash->page2size;
	message.body.page.PidLockMode  = *lockMode;

	switch (*lockMode) {
	case SH:
		message.body.page.FidLockMode = IS;
		break;
	case EX:
		message.body.page.FidLockMode = IX;  /* MJZ: changed from IS */
		break;
	case NL:
		message.body.page.FidLockMode = NL;
		break;
	default:
		SM_ERROR(TYPE_FATAL, esmINTERNAL);
	}
	/*
	 *	We send an lrc to the server if this is a resident page.
	 *  If it is not resident, we had better be sending an impossibly
	 *  low lrc.
	 */
	message.body.page.pageType = PPAGE_TYPE(pageHash);

	pageLRC = bf_lrcLoc(pageHash);
	message.body.page.pageLRC = *pageLRC;
#ifdef DEBUG
	if (PPAGE_TYPE(pageHash) == H_CHUNK) {
		/*
		 *	we don't to inter-trans caching of large object pages, so
		 *	they should always be read, so make sure LRC is low.
		 */
		SM_ASSERT(LEVEL_3, pageLRC->wrapCount == LrcNone.wrapCount);
		SM_ASSERT(LEVEL_3, pageLRC->count == LrcNone.count);
	}
#endif /* DEBUG */

	TRPRINT(TR_IO|TR_SM, TR_LEVEL_1, 
		("reading page %d, type %d, lrc %d.%d", 
			pid->page, PPAGE_TYPE(pageHash),
			message.body.page.pageLRC.wrapCount, 
			message.body.page.pageLRC.count));

	if (callServer(volRec->serverInfo, &message, NULL, 0) < 0)	{
		TRPRINT(TR_IO|TR_SM, TR_LEVEL_1,  
				("callServer failed sm_errno %d\n", sm_errno));
		switch(sm_errno) {
		case esmPHANTOMDEADLOCK:
		case esmCAUSEDPHANTOMDEADLOCK:
			if(retryFlag && (PPAGE_TYPE(pageHash) == H_INDEX) )  {
				lm_InvalidateAllLocks();
				if (sm_errno == esmCAUSEDPHANTOMDEADLOCK)
					retryFlag = FALSE;
				goto again;
			}
			sm_errno = esmLOCKCAUSEDDEADLOCK;
			return esmFAILURE;

		case esmAFTERLASTPID:
		case esmBEFOREFIRSTPID:
			if( flags & REPLY_WITH_NEIGHBOR ) {
				/* these "errors" are ok */
				break;
			}
			/* drop through */

		default:
			return esmFAILURE;
		} /* switch */
	} 
	if(message.body.page.flags & REPLY_WITH_DATA) {
		/* server sent us a page */

		TRPRINT(TR_IO, TR_LEVEL_1, ("server sent page :%d", pid->page));

		if (receive(volRec->serverInfo, buffer, RAISETWO(pageHash->page2size)) < 0){

			return esmFAILURE;
		}

		/*
		 * If the page is sent, the lrc comes back as part of the page.
		 * If the page is a chunk, the lrc comes back in the message
		 * (as well as in the page, if the page is sent).  The lrc
		 * has to be sent separately for chunks when-and-if this 
		 * request is made to support send-part-of-page. 
		 */
		SM_ASSERT(LEVEL_1, pageLRC != NULL);
	} else  {
		/* server figures we didn't need the page -- ours is up-to-date  */ 
		TRPRINT(TR_IO, TR_LEVEL_1, ("No page sent: LRC ok"));

		/*
		 * Had better not be part of a lg object.
		 * For now, there is no inter-tx caching of lg. object 
		 * pages.  
		 */
		SM_ASSERT(LEVEL_1, PPAGE_TYPE(pageHash) != H_CHUNK);
	}
	/* Still have to update our lrc for the page */
	if (logInfo->initLRCisLSN) {
		*pageLRC = *logInfo->initLRC;
	} else if(PPAGE_TYPE(pageHash) == H_CHUNK) {
		*pageLRC = message.body.page.pageLRC;
	}
	if(PPAGE_TYPE(pageHash) == PAGE_SLOTTED) {
		sm_AuditSlotPage((SLOTTEDPAGE *)buffer);
	}

	pageHash->flags &= ~H_INVALID;
	if(volRec->properties & VOL_TEMP) {
		/*
		no log records are generated on read, so we skipt this assert; the
		bulk-load code won't meet this requirement
		SM_ASSERT( LEVEL_1, (SM_State.logLevel == LOG_NONE));
		*/
		pageHash->flags |= H_TEMP; /* page on a temp volume */
	}
	
	TRPRINT(TR_IO|TR_MSG, TR_LEVEL_1, 
		("errno:%d flags 0x%x", message.header.params.out.errno, flags));

	/* server returns a PAGEBODY instead of a FILEBODY! */
	if(flags & REPLY_WITH_NEIGHBOR) {
		SM_ASSERT(LEVEL_1, neighborPid != NULL);
		*neighborPid =  message.body.page.NeighborPid;
		TRPRINT(TR_IO|TR_MSG, TR_LEVEL_1, 
		("neighborPid:%d.%d", message.body.page.NeighborPid.page, 
			message.body.page.NeighborPid.volid));
	}

	/* return the lock mode */
	SM_ASSERT(LEVEL_3, message.body.page.PidLockMode >= *lockMode);
	*lockMode = message.body.page.PidLockMode;

	return esmNOERROR;
}
