/*
 *   $RCSfile: LG_AppendIndex.c,v $  
 *   $Revision: 1.1.1.1 $  
 *   $Date: 1996/05/04 21:55:23 $      
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/


#include "sysdefs.h"
#include "ess.h"
#include "trace.h"
#include "checking.h"
#include "error.h"
#include "io.h"
#include "page.h"
#include "list.h"
#include "object.h"
#include "tid.h"
#include "bf.h"
#include "chunk.h"
#include "lock.h"
#include "lsn.h"
#include "clog.h"
#include "logaction.h"
#include "loginfo.h"
#include "logrecs.h"
#include "bf_macro.h"
#include "lgobject.h"
#include "sm_state.h"
#include "bf_extfuncs.h"
#include "lg_intfuncs.h"
#include "lg_extfuncs.h"
#include "util_funcs.h"
#include "clog_globals.h"
#include "clog_extfuncs.h"



/*
 * Update the index of the long object following an append or insert
 * operation.
 */


 int 
LG_AppendIndex (

	int				slotsPerNode,
	int				remainSlots,
	int				replaceCount,
	int				nodeCount,
    PAGELIST		*newPageList,	/* list of pid/cnt pairs to insert			*/
    LGNODE			*origNode,		/* ptr to copy of original node contents	*/
	int				insertSlot 
)
{

	LGNODE			*currentNode;
	CHUNKPAGE		*currentPage;
	LGSLOT			*slotPtr;
	int				slotCount;
	int				offset;
	int				i;
	int				j;
	BOOL		    unFixIt;	/* true if we should unfix current page */
	OPERATIONINFO   opInfo;		/* structure for passing info to log calls*/


	TRACE(TR_LG, TR_LEVEL_1);

	unFixIt = FALSE;

	/*
	 *	Check to see if the remaining slots are
	 *	contained entirely in the last node.
	 */
	if (remainSlots <= slotsPerNode)	{

		TRPRINT(TR_LG, TR_LEVEL_2, ("remainder on last node"));

		/*
		 *	get a pointer to the current page
		 */
		currentPage = &(newPageList->pages[nodeCount - 1]);

		/*
		 *	get a pointer to the last node
		 */
		currentNode = (LGNODE *) currentPage->nodeLink->bufFrame;

		/*
		 *	copy in the remaining slots.  The counts
		 *	should be correct due to the LG_Traverse updates
		 */
		bcopy((char *) &(origNode->slot[insertSlot + replaceCount]),
			  (char *) &(currentNode->slot[slotsPerNode - remainSlots]),
			  (int) (sizeof(LGSLOT) * remainSlots) );
		TRPRINT(TR_LG, TR_LEVEL_2, ("copying %d slots from:%d to:%d", remainSlots,
				(insertSlot + replaceCount), (slotsPerNode - remainSlots)));

		/*
		 *	must fix up the counts in the last node
		 */
		for (j = 0, i = ((int)currentNode->header.numSlots - remainSlots),
			 slotPtr = &(currentNode->slot[i]);
			 i < slotsPerNode;
			 j++, i++, slotPtr++)	{

			/*
			 *	Get the slot counts from the old node
			 */
			slotPtr->count = LG_SlotCount(origNode, 
						(SLOTINDEX)((origNode->header.numSlots - remainSlots) + j));
			TRPRINT(TR_LG, TR_LEVEL_2, ("slot:%d count:%d", i, slotPtr->count));
		}

		/*
		 *	if there are no slots to be inserted later then we
		 *	must normalize the slot counts on this node.
		 */
		if (remainSlots == slotsPerNode)	{

			/*
			 *	we must set the counts to final form
			 */
			for (i = 0, slotPtr = currentNode->slot;
				 i < currentNode->header.numSlots;
				 i++, slotPtr++)	{

				/*
				 *	check to see if we are past the first node
				 */
				if (i > 0)	{

					/*
					 *	total from last slot
					 */
					slotPtr->count += (slotPtr - 1)->count;
					TRPRINT(TR_LG, TR_LEVEL_2, ("slot:%d count:%d", i, slotPtr->count));
				}
			}
			lg_PrintNode(currentNode, 0, currentPage->pid.page);

			/*
			 *	Release the node since it will not be processed
			 */
		    unFixIt = TRUE;
		} 

		if (unFixIt) {
			bf_UnfixPage(currentPage->nodeLink, BF_DEFAULT, TRUE);
		}

		/*
		 *	leave the largest count in the page struct.
		 *	this is required by LG_FixIndex
		 */
		currentPage->size = (int)currentNode->slot[slotsPerNode - 1].count;
		TRPRINT(TR_LG, TR_LEVEL_2, ("currentpage size:%d", currentPage->size));

	} else {

		TRPRINT(TR_LG, TR_LEVEL_2, ("remainder on last 2 nodes"));

		/*
		 *	(remainSlots - slotsPerNode) gives the number of
		 *	remaining slots on the 2nd last node.  We must
		 *	have slotsPerNode on the last page
		 */
		remainSlots = (remainSlots - slotsPerNode);

		/*
		 *	get a pointer to the current page entry
		 */
		currentPage = &(newPageList->pages[nodeCount - 2]);

		/*
		 *	get a pointer to the 2nd last node
		 */
		currentNode = (LGNODE *) currentPage->nodeLink->bufFrame;

		/*
		 *	Calculate the number of slots in the 2nd last node
		 *	if (remainder == (nodeCount - 1)) then the 2nd
		 *	last node has slotsPerNode + 1 slots else it
		 *	hash slotsPerNode.
		 */
		slotCount = (int)currentNode->header.numSlots;
		offset = slotCount - remainSlots;
		TRPRINT(TR_LG, TR_LEVEL_2, ("remainSlots:%d slotCount:%d offset:%d nodeCount:%d",
				remainSlots, slotCount, offset, nodeCount));

		/*
		 *	copy the data into the 2nd last node
		 */
		bcopy((char *) &(origNode->slot[insertSlot + replaceCount]),
			  (char *) &(currentNode->slot[offset]),
			  (int) (sizeof(LGSLOT) * remainSlots));
		TRPRINT(TR_LG, TR_LEVEL_2, ("copying %d slots from:%d to:%d",
				remainSlots, insertSlot + replaceCount, offset));

		/*
		 *	must fix up the counts in the last node
		 */
		for (j = 0, i = slotCount - remainSlots, 
			 slotPtr = &(currentNode->slot[i]);
			 i < slotCount; j++, i++, slotPtr++)	{

			/*
			 *	Get the slot counts from the old node
			 */
			slotPtr->count =
			 LG_SlotCount(origNode,
			    (SLOTINDEX)((origNode->header.numSlots - 
					(remainSlots + slotsPerNode)) + j));
			TRPRINT(TR_LG, TR_LEVEL_2, ("slot:%d count:%d", i, slotPtr->count));
		}

		/* get a pointer to the pagelist for last node */
		currentPage++;

		/* get a pointer to the last node */
		currentNode = (LGNODE *) currentPage->nodeLink->bufFrame;

		/* copy the data into the last node */
		bcopy((char *) &(origNode->slot[insertSlot+replaceCount + remainSlots]),
			  (char *)  &(currentNode->slot[0]),
			  (int) (sizeof(LGSLOT) * slotsPerNode));
		TRPRINT(TR_LG, TR_LEVEL_2, ("copying %d slots from slot:%d to slot:%d",
				slotsPerNode, insertSlot + replaceCount+remainSlots, 0));

		/*
		 *	must fix up the counts in the last node
		 */
		for (i = 0, slotPtr = currentNode->slot;
										i < slotsPerNode; i++, slotPtr++)	{

			/*
			 *	Get the slot count from the old node
			 */
			slotPtr->count = LG_SlotCount(origNode,
				  (SLOTINDEX)((origNode->header.numSlots - slotsPerNode) + i));

			/*
			 *	accumulate the information
			 */
			if (i > 0)	{

				/*
				 *	increment by the count before this
				 */
				slotPtr->count += (slotPtr - 1)->count;
			}
			TRPRINT(TR_LG, TR_LEVEL_2, ("slot:%d count:%d", i, slotPtr->count));
		}

		/*
		 *	save the largest byte in subtree as required by LG_FixIndex
		 */
		currentPage->size = 
				(int)currentNode->slot[currentNode->header.numSlots - 1].count;
		TRPRINT(TR_LG, TR_LEVEL_2, ("size for last node:%d",currentPage->size));
		lg_PrintNode(currentNode, 0, currentPage->pid.page);

		if (GET_LOGLEVEL(LOG_SPACE)) {
			/* log the append of the slots to the last node*/
			opInfo.action = LOG_ACTION_APPEND_LARGE_SLOTS;
			opInfo.flags = 0;
			opInfo.start = 0;
			opInfo.size = (int) (sizeof(LGSLOT) * slotsPerNode);
			opInfo.oldData = NULL;
			opInfo.newData= (char *) &(currentNode->slot[0]);
			opInfo.actionPid = currentPage->pid;
			/* this can't be a slotted root object */
			opInfo.rootSlot = NIL;
			clog_LogModifyLgPage(&opInfo);
		}

		/* Unfix the last node */
		bf_UnfixPage(currentPage->nodeLink, BF_DEFAULT, TRUE);
	}

	/* return success */
	return(esmNOERROR);
}
