/*
 *   $RCSfile: LG_BalanceData.c,v $  
 *   $Revision: 1.1.1.1 $  
 *   $Date: 1996/05/04 21:55:24 $      
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/


#include "ess.h"
#include "sysdefs.h"
#include "trace.h"
#include "checking.h"
#include "error.h"
#include "io.h"
#include "page.h"
#include "list.h"
#include "object.h"
#include "tid.h"
#include "bf.h"
#include "chunk.h"
#include "bf_macro.h"
#include "lock.h"
#include "lgobject.h"
#include "volume.h"
#include "logaction.h"
#include "clog.h"
#include "sm_state.h"
#include "io_extfuncs.h"
#include "bf_extfuncs.h"
#include "lg_intfuncs.h"
#include "lg_extfuncs.h"
#include "clog_extfuncs.h"
#include "trans_funcs.h"
#include "util_funcs.h"


/*
 * Routine to reshuffle the contents of 2 of a node's leaf-children
 * along the cut path in order to equalize their entries.  
 *
 * The leaves are read into two separate chunks for clarity and
 * buffer management flexibility.
 */


 int 
LG_BalanceData (

	FID				*fid,
	BUFGROUP		*bufGroup,		/* buffer group to use for operating on data */
	PID				*nodePid,		/* parent node page id			*/
	SLOTINDEX		nodeRootSlot,	/* non-NIL if node on slotted page*/
	LGNODE			*nodePtr,		/* ptr to buffered parent node	*/
	SLOTINDEX		*slotIndex,		/* first slot affected by the reshuffle */
	BOOL			leftShared,		/* page should be versioned		*/
	BOOL			rightShared 	/* page should be versioned 	*/
)
{
    int				totalBytes;		/* total number of leaf bytes involved */
    int				bytesMoved;		/* number of bytes that must be moved */
    PID				leftPid;		/* pid of left child node */
    PID				rightPid;		/* pid of right child node */
    int				leftBytes;		/* original number of bytes in left leaf */
    int				rightBytes;		/* original number of bytes in right leaf */
    int				newLeftBytes;	/* new number of bytes in left leaf */
    int				newRightBytes;	/* new number of bytes in right leaf */
    int				numShadows;		/* number of shadow leaves needed */
    PID				shadows[2];		/* list of allocated shadow leaves */
    PAGELIST		readList;		/* list of pid/cnt pairs for reading */
    CHUNKDESC		*leftChunkDesc;	/* chunk for left leaf accesses */
    CHUNKDESC		*rightChunkDesc;/* chunk for right leaf accesses */
    int				leftSlot;		/* slot for left kid involved */
	TWO				shadowsUsed;	
	OPERATIONINFO	opInfo;
	LGSLOT			tempSlot;
	int				e;


	TRPRINT(TR_LG, TR_LEVEL_1, ("nodePid, slotIndex:%d", nodePid->page, *slotIndex));

    /*
     *	Identify and then extract information on the pair of leaf nodes
     *	involved.  (Note:  First step is to ensure that we're able to use
     *	slotIndex's right neighbor.)  Note that this routine can
	 *	only be called if the number of slots is greater than 1.
     */
    if (*slotIndex == (nodePtr->header.numSlots - 1)) {

		(*slotIndex)--;
		TRPRINT(TR_LG, TR_LEVEL_2, ("adjusting slotIndex"));
    }
	readList.pages = NULL;

	/*
	 *	setup the paramters for the left and right slots
	 */
    leftSlot = *slotIndex;
    leftPid = nodePtr->slot[leftSlot].pid;
    rightPid = nodePtr->slot[leftSlot+1].pid;
    leftBytes = LG_SlotCount(nodePtr, leftSlot);
    rightBytes = LG_SlotCount(nodePtr, leftSlot + 1);
	TRPRINT(TR_LG, TR_LEVEL_2, ("leftSlot:%d leftPid:%d rightPid:%d leftBytes:%d rightBytes:%d",
			leftSlot, leftPid.page, rightPid.page, leftBytes, rightBytes));

    /*
     *	Determine how to redistribute the data. If all the data
	 *	will reside on a single page then delete the left page.
	 *	Note that bytesMoved can be negative which will work
	 *	OK with the remaining equations
     */
    totalBytes = leftBytes + rightBytes;
    newLeftBytes = (totalBytes <= LGMAXLEAF) ? totalBytes : (totalBytes / 2);
    newRightBytes = totalBytes - newLeftBytes;
    bytesMoved = newLeftBytes - leftBytes;
	TRPRINT(TR_LG, TR_LEVEL_2, ("totalBytes:%d newLeftBytes:%d newRightBytes:%d bytesMoved:%d",
					totalBytes, newLeftBytes, newRightBytes, bytesMoved));

    /*
     * LEFT page:	Set up a small page list, and read in page
     */
    if(LG_InitList(&readList, 2) == esmFAILURE) {
		FAIL
	}
    LG_AppendPage(&readList, &leftPid, leftBytes, FALSE, NULL, 
			(SHORTPID) ZERO, NIL, ZERO, nodePtr->slot[leftSlot].shared);

	sm_SetAbortIfError();

    if (bf_ReadChunk(fid, bufGroup, readList.numPages,
			 readList.numPages, readList.pages, 0, 0, &leftChunkDesc))	{
		FAIL
	}
	/* leftChunkDesc now points to readList.pages */
	readList.pages = NULL;

    /*
     * RIGHT page:	Set up a small page list, and read in page
     */
    if(LG_InitList(&readList, 2) == esmFAILURE) {
		FAIL
	}
    LG_AppendPage(&readList, &rightPid, rightBytes, FALSE, NULL, 
		(SHORTPID) ZERO, NIL, ZERO, nodePtr->slot[leftSlot+1].shared);
    if (bf_ReadChunk(fid, bufGroup, readList.numPages,
						 readList.numPages, readList.pages, 0, 0, &rightChunkDesc))	{
		FAIL
	}
	/* rightChunkDesc now points to readList.pages */
	readList.pages = NULL;

	bf_HackChunkPage(&leftPid);
	bf_HackChunkPage(&rightPid);

	/*
	 *	Determine how many resultant shadow pages will be needed
	 */
	numShadows = 0;
    if (newLeftBytes == totalBytes) {
		numShadows += leftShared ? 1 : 0;
	} else {
		numShadows += leftShared ? 1 : 0;
		numShadows += rightShared ? 1 : 0;
	}
	TRPRINT(TR_LG, TR_LEVEL_2, ("number of nodes needed:%d", numShadows));

	/*
	 *	allocate the shadow pages
	 */
	shadowsUsed = 0;
	if (numShadows > 0) {
		if (io_AllocPages(nodePid->volid, LGDATA_PAGE2SIZE,
						  numShadows, &leftPid, shadows, PAGE_LARGEDATA))	{
			FAIL
		}
	}

	/*
	 *	Will only the left page survive
	 */
    if (newLeftBytes == totalBytes) {

		/*
		 *	Append right page data to the left
		 */
		SM_ASSERT(LEVEL_3, bytesMoved > 0);
		SM_ASSERT(LEVEL_3, bytesMoved+leftBytes <= LGDATA_USABLESIZE);
		bcopy(rightChunkDesc->bufFrame, leftChunkDesc->bufFrame+leftBytes, bytesMoved);


        leftChunkDesc->totalLength += bytesMoved;
        leftChunkDesc->chunkPage[0].size += bytesMoved;
		SM_ASSERT(LEVEL_3, leftChunkDesc->totalLength == totalBytes);

		/*
 		 *	see if the page was shadowed
		 */
		if (shadowsUsed < numShadows)  {
		
			/*
			 *	Replace the page in the buffer with the shadow
			 */
			e = bf_ReplaceLeafChunk(leftChunkDesc, 
									leftChunkDesc->chunkPage,
									&(shadows[shadowsUsed]), NULL);
			if (e != esmNOERROR) {
				FAIL
			}
			shadowsUsed++;

			if (GET_LOGLEVEL(LOG_DATA) ) {
				/*
				 *	Log operation as an append of all data to the shadow
				 */
				opInfo.action = LOG_ACTION_APPEND_LARGE_DATA;
				opInfo.flags = 0;
				opInfo.start = 0;
				opInfo.size = leftChunkDesc->totalLength;
				opInfo.oldData = NULL;
				opInfo.newData = leftChunkDesc->bufFrame;
				opInfo.actionPid = shadows[shadowsUsed-1];
				if(clog_LogModifyLgPage(&opInfo)) {
					FAIL
				}
			}

		} else {

			if (GET_LOGLEVEL(LOG_DATA) ) {
				/*
				 *	Log operation as append of right page data to the left
				 */
				opInfo.action = LOG_ACTION_APPEND_LARGE_DATA;
				opInfo.flags = 0;
				opInfo.start = leftBytes;
				opInfo.size = bytesMoved;
				opInfo.oldData = NULL;
				opInfo.newData = rightChunkDesc->bufFrame;
				opInfo.actionPid = leftPid;
				if(clog_LogModifyLgPage(&opInfo)) {
					FAIL
				}
			}
		}
	
		/*
		 *	If right page is not shared then deallocate it
		 */
		if (!rightShared) {
			if (io_DeallocPages(LGDATA_PAGE2SIZE, 1, &rightPid, 
											FALSE, PAGE_LARGEDATA))	{
				FAIL
			}
		}

	} else {
	
		/*
		 *	Two pages involved.  See if bytes were appended to the
		 *	left page.
		 */
		if (bytesMoved > 0) { 

			/*
			 *	Append right page data to the left
			 */
			SM_ASSERT(LEVEL_3, bytesMoved > 0);
			SM_ASSERT(LEVEL_3, bytesMoved+leftBytes <= LGDATA_USABLESIZE);
			bcopy(rightChunkDesc->bufFrame, leftChunkDesc->bufFrame+leftBytes, bytesMoved);
			leftChunkDesc->totalLength += bytesMoved;
			leftChunkDesc->chunkPage[0].size += bytesMoved;
			SM_ASSERT(LEVEL_3, leftChunkDesc->totalLength == newLeftBytes);

			/*
			 *	see if the left page was shadowed
			 */
			if (leftShared)  {
			
				/*
				 *	Replace the page in the buffer with the shadow
				 */
				e = bf_ReplaceLeafChunk(leftChunkDesc, 
										leftChunkDesc->chunkPage,
										&(shadows[shadowsUsed]), NULL);
				if (e != esmNOERROR) {
					FAIL
				}
				shadowsUsed++;

				if (GET_LOGLEVEL(LOG_DATA) ) {
					/*
					 *	Log operation as an append of all data to shadow
					 */
					opInfo.action = LOG_ACTION_APPEND_LARGE_DATA;
					opInfo.flags = 0;
					opInfo.start = 0;
					opInfo.size = leftChunkDesc->totalLength;
					opInfo.oldData = NULL;
					opInfo.newData = leftChunkDesc->bufFrame;
					opInfo.actionPid = shadows[shadowsUsed-1];
					if(clog_LogModifyLgPage(&opInfo)) {
						FAIL
					}
				}
			} else {

				if (GET_LOGLEVEL(LOG_DATA) ) {
					/*
					 *	Log operation as append of right page data to left
					 */
					opInfo.action = LOG_ACTION_APPEND_LARGE_DATA;
					opInfo.flags = 0;
					opInfo.start = leftBytes;
					opInfo.size = bytesMoved;
					opInfo.oldData = NULL;
					opInfo.newData = rightChunkDesc->bufFrame;
					opInfo.actionPid = leftPid;
					if(clog_LogModifyLgPage(&opInfo)) {
						FAIL
					}
				}
			}

			/*
			 *	Log delete of data from the right page.  See if
			 *	right page is shadowed first
			 */
			if (rightShared) {
				SM_ASSERT(LEVEL_3, shadowsUsed < numShadows);

				/*
				 *	Replace the page in the buffer with the shadow
				 */
				e = bf_ReplaceLeafChunk(rightChunkDesc, 
										rightChunkDesc->chunkPage,
										&(shadows[shadowsUsed]), NULL);
				if (e != esmNOERROR) {
					FAIL
				}
				shadowsUsed++;

				if (GET_LOGLEVEL(LOG_DATA) ) {
					/*
					 *	Log operation as append of remaining data to shadow
					 */
					opInfo.action = LOG_ACTION_APPEND_LARGE_DATA;
					opInfo.flags = 0;
					opInfo.start = 0;
					opInfo.size = rightChunkDesc->totalLength-bytesMoved;
					opInfo.oldData = NULL;
					opInfo.newData = rightChunkDesc->bufFrame+bytesMoved;
					opInfo.actionPid = shadows[shadowsUsed-1];
					if(clog_LogModifyLgPage(&opInfo)) {
						FAIL
					}
				}

			} else {
				SM_ASSERT(LEVEL_3, shadowsUsed == numShadows);

				if (GET_LOGLEVEL(LOG_DATA) ) {
					/*
					 *	Log operation as delete of data in right page
					 */
					opInfo.action = LOG_ACTION_DELETE_LARGE_DATA;
					opInfo.flags = 0;
					opInfo.start = 0;
					opInfo.size = bytesMoved;
					opInfo.oldData = rightChunkDesc->bufFrame;
					/* note, new data is ptr to the previous size of the data */
					opInfo.newData = (char *) &rightBytes;
					opInfo.actionPid = rightPid; 
					if(clog_LogModifyLgPage(&opInfo)) {
						FAIL
					}
				}
			}

			/*
			 *	Delete data from right page
			 */
			bcopy(rightChunkDesc->bufFrame+bytesMoved, rightChunkDesc->bufFrame, rightChunkDesc->totalLength-bytesMoved);
			rightChunkDesc->totalLength -= bytesMoved;
			rightChunkDesc->chunkPage[0].size -= bytesMoved;
			SM_ASSERT(LEVEL_3, rightChunkDesc->totalLength == newRightBytes);

		} else {
			SM_ASSERT(LEVEL_3, newRightBytes > rightBytes);

			/*
			 *	delete data from left page
			 */
			SM_ASSERT(LEVEL_3, bytesMoved < 0);
			leftChunkDesc->totalLength += bytesMoved;
			leftChunkDesc->chunkPage[0].size += bytesMoved;
			SM_ASSERT(LEVEL_3, leftChunkDesc->totalLength == newLeftBytes);

			/*
			 *	see if the left page was shadowed
			 */
			if (leftShared)  {
			
				/*
				 *	Replace the page in the buffer with the shadow
				 */
				e = bf_ReplaceLeafChunk(leftChunkDesc, 
										leftChunkDesc->chunkPage,
										&(shadows[shadowsUsed]), NULL);
				if (e != esmNOERROR) {
					FAIL
				}
				shadowsUsed++;

				if (GET_LOGLEVEL(LOG_DATA) ) {
					/*
					 *	Log operation as an append of all data to the shadow
					 */
					opInfo.action = LOG_ACTION_APPEND_LARGE_DATA;
					opInfo.flags = 0;
					opInfo.start = 0;
					opInfo.size = leftChunkDesc->totalLength;
					opInfo.oldData = NULL;
					opInfo.newData = leftChunkDesc->bufFrame;
					opInfo.actionPid = shadows[shadowsUsed-1];
					if(clog_LogModifyLgPage(&opInfo)) {
						FAIL
					}
				}
			} else {

				if (GET_LOGLEVEL(LOG_DATA) ) {
					/*
					 *	Log operation as removal of data at end of left page
					 */
					opInfo.action = LOG_ACTION_REMOVE_LARGE_DATA;
					opInfo.flags = 0;
					opInfo.start = newLeftBytes;
					opInfo.size = -bytesMoved;
					opInfo.oldData = leftChunkDesc->bufFrame+newLeftBytes;
					opInfo.newData = NULL;
					opInfo.actionPid = leftPid;
					if(clog_LogModifyLgPage(&opInfo)) {
						FAIL
					}
				}
			}

			/*
			 *	Insert data into right page.  First shift everything
			 *	over, and then lay in the new bytes.
			 */
			bcopy(rightChunkDesc->bufFrame, rightChunkDesc->bufFrame-bytesMoved, rightBytes);
			bcopy(leftChunkDesc->bufFrame+newLeftBytes, rightChunkDesc->bufFrame, -bytesMoved);
			rightChunkDesc->totalLength -= bytesMoved;
			rightChunkDesc->chunkPage[0].size -= bytesMoved;
			SM_ASSERT(LEVEL_3, rightChunkDesc->totalLength == newRightBytes);

			/*
			 *	Log insert of data into the right page.  See if
			 *	right page is shadowed first
			 */
			if (rightShared) {
				SM_ASSERT(LEVEL_3, shadowsUsed < numShadows);

				/*
				 *	Replace the page in the buffer with the shadow
				 */
				e = bf_ReplaceLeafChunk(rightChunkDesc, 
										rightChunkDesc->chunkPage,
										&(shadows[shadowsUsed]), NULL);
				if (e != esmNOERROR) {
					FAIL
				}
				shadowsUsed++;

				if (GET_LOGLEVEL(LOG_DATA) ) {
					/*
					 *	Log operation as append of all data to shadow
					 */
					opInfo.action = LOG_ACTION_APPEND_LARGE_DATA;
					opInfo.flags = 0;
					opInfo.start = 0;
					opInfo.size = rightChunkDesc->totalLength;
					opInfo.oldData = NULL;
					opInfo.newData = rightChunkDesc->bufFrame;
					opInfo.actionPid = shadows[shadowsUsed-1];
					if(clog_LogModifyLgPage(&opInfo)) {
						FAIL
					}
				}

			} else {
				SM_ASSERT(LEVEL_3, shadowsUsed == numShadows);

				if (GET_LOGLEVEL(LOG_DATA) ) {
					/*
					 *	Log operation as insert to beginning of right page
					 */
					opInfo.action = LOG_ACTION_INSERT_LARGE_DATA;
					opInfo.flags = 0;
					opInfo.start = 0;
					opInfo.size = -bytesMoved;
					/* note, old data is ptr to the previous size of the data */
					opInfo.oldData = (char *) &rightBytes;
					opInfo.newData = rightChunkDesc->bufFrame;
					opInfo.actionPid = rightPid; 
					if(clog_LogModifyLgPage(&opInfo)) {
						FAIL
					}
				}
			}
		}
	}

	/*
	 *	Mark the left chunk pages as dirty and unfix them
	 */
	DIRTY_CHUNK(leftChunkDesc);
	/*
	 *  check to see if the page is dirty
	 */
	if (!(leftChunkDesc->chunkPage->flags & PAGE_DIRTY))   {

		/*
		 *  Dirty the page
		 */
		leftChunkDesc->chunkPage->flags |= PAGE_DIRTY;
		bf_DirtyPage(leftChunkDesc->chunkPage->pageHash);

		/*
		 *  Increment the number of dirty pages in the chunk
		 */
		leftChunkDesc->dirtyPages++;
	}
	/*
	 *	unfix the chunk
	 */
    bf_UnfixChunk(leftChunkDesc, BF_DEFAULT);

	/*
 	 *	If the right page was not deallocated, mark it dirty as well
	 */
    if (newLeftBytes < totalBytes) {
		DIRTY_CHUNK(rightChunkDesc);
		if (!(rightChunkDesc->chunkPage->flags & PAGE_DIRTY))   {

			rightChunkDesc->chunkPage->flags |= PAGE_DIRTY;
			bf_DirtyPage(rightChunkDesc->chunkPage->pageHash);

			rightChunkDesc->dirtyPages++;
		}
	}	
	bf_UnfixChunk(rightChunkDesc, BF_DEFAULT);

    /*
     * Update left parent's count and pid entries to reflect what happened.
     */
	tempSlot = nodePtr->slot[leftSlot];
    nodePtr->slot[leftSlot].count += bytesMoved;
    nodePtr->slot[leftSlot].pid = leftShared ? shadows[0] : leftPid;
    nodePtr->slot[leftSlot].shared = FALSE;
	TRPRINT(TR_LG, TR_LEVEL_2, ("leftCount:%d leftPid:%d", nodePtr->slot[leftSlot].count,
			nodePtr->slot[leftSlot].pid.page));
	
	if (GET_LOGLEVEL(LOG_SPACE)) {
		/*
		 *	Log the write to the slot
		 */
		opInfo.action = LOG_ACTION_WRITE_LARGE_SLOTS;
		opInfo.flags = 0;
		opInfo.start = leftSlot*(int)sizeof(LGSLOT);
		opInfo.size = 1 * (int)sizeof(LGSLOT);
		opInfo.oldData = (char*) &tempSlot;
		opInfo.newData = (char*) &(nodePtr->slot[leftSlot]);
		opInfo.actionPid = *nodePid;
		/* mark if it is a root object on a slotted page */
		opInfo.rootSlot = nodeRootSlot;
		if(clog_LogWriteLgPage(&opInfo)) {
			FAIL
		}
	}

	/* Check to see if we updated the right hand page (or shadowed it */
    if (numShadows == 2) {

		/*
		 *	update the shadows pid, no need to update
		 *	the count since the number of bytes up to
		 *	and including this slot must be unchanged
		 */
		tempSlot = nodePtr->slot[leftSlot+1];
		nodePtr->slot[leftSlot+1].pid = rightShared ? shadows[shadowsUsed-1] : rightPid;
		nodePtr->slot[leftSlot+1].shared = FALSE;
		TRPRINT(TR_LG, TR_LEVEL_2, ("rightCount:%d rightPid:%d",
				nodePtr->slot[leftSlot+1].count,
				nodePtr->slot[leftSlot+1].pid.page));
			
		if (GET_LOGLEVEL(LOG_SPACE)) {
			/*
			 *	Log the write to the slot
			 */
			opInfo.action = LOG_ACTION_WRITE_LARGE_SLOTS;
			opInfo.flags = 0;
			opInfo.start = (leftSlot+1)*(int)sizeof(LGSLOT);
			opInfo.size = 1 * (int)sizeof(LGSLOT);
			opInfo.oldData = (char*) &tempSlot;
			opInfo.newData = (char*) &(nodePtr->slot[leftSlot+1]);
			opInfo.actionPid = *nodePid;
			/* mark if it is a root object on a slotted page */
			opInfo.rootSlot = nodeRootSlot;
			if(clog_LogWriteLgPage(&opInfo)) {
				FAIL
			}
		}
	} else if (newRightBytes > 0) {
		/* do nothing since slot count will be the same */
    } else {
		/* we must have removed the right hand page */
		register int		copySlots;

		/*
		 *	decrement the number of slots in the node by one
		 */
        nodePtr->header.numSlots -= 1;
		TRPRINT(TR_LG, TR_LEVEL_2, ("new slot count:%d", nodePtr->header.numSlots));

		/*
		 *	Check to see if we removed the last slot in the node
		 */
		if (leftSlot < (nodePtr->header.numSlots - 1))	{

			if (GET_LOGLEVEL(LOG_SPACE)) {
				/*
				 *	Log delete of the slot
				 */
				opInfo.action = LOG_ACTION_DELETE_LARGE_SLOTS;
				opInfo.flags = 0;
				opInfo.start = (leftSlot+1)*(int)sizeof(LGSLOT);
				opInfo.size = 1 * (int)sizeof(LGSLOT);
				opInfo.oldData = (char *) &(nodePtr->slot[leftSlot+1]);
				opInfo.newData = NULL;
				opInfo.actionPid = *nodePid;
				opInfo.rootSlot = nodeRootSlot;
				if(clog_LogModifyLgPage(&opInfo)) {
					FAIL
				}
			}

			/*
			 *	calculate the number of slots to copy
			 */
			copySlots = ((int)nodePtr->header.numSlots - 1) - leftSlot;

			/*
			 *	copy in remaining slots
			 */
			bcopy((char *) &(nodePtr->slot[leftSlot+2]),
				  (char *) &(nodePtr->slot[leftSlot+1]),
				  (int) (copySlots * (int)sizeof(LGSLOT)));
			TRPRINT(TR_LG, TR_LEVEL_2, ("copying %d slots from slot:%d to slot:%d",
					copySlots, leftSlot + 2, leftSlot + 1));
        } else {
			
			if (GET_LOGLEVEL(LOG_SPACE)) {
				/*
				 *	Log removal of slot from end of page
				 */
				opInfo.action = LOG_ACTION_REMOVE_LARGE_SLOTS;
				opInfo.flags = 0;
				opInfo.start = (leftSlot+1)*(int)sizeof(LGSLOT);
				opInfo.size = 1 * (int)sizeof(LGSLOT);
				opInfo.oldData = (char *) &(nodePtr->slot[leftSlot+1]);
				opInfo.newData = NULL;
				opInfo.actionPid = *nodePid;
				opInfo.rootSlot = nodeRootSlot;
				if(clog_LogModifyLgPage(&opInfo)) {
					FAIL
				}
			}
		}
    }
Success:
#ifdef DEBUG
	lg_PrintNode(nodePtr, 0, 0);
#endif DEBUG
    return esmNOERROR;

FailureReturn:
	SM_FREE_PTR(readList.pages);
    return esmFAILURE;
}
