/*
 *   $RCSfile: LG_BuildShareStack.c,v $  
 *   $Revision: 1.1.1.1 $  
 *   $Date: 1996/05/04 21:55:24 $      
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/


#include <stdio.h>
#include "ess.h"
#include "trace.h"
#include "checking.h"
#include "io.h"
#include "page.h"
#include "list.h"
#include "object.h"
#include "tid.h"
#include "bf.h"
#include "chunk.h"
#include "lock.h"
#include "bf_macro.h"
#include "sm_macro.h"
#include "lgobject.h"
#include "pool.h"
#include "version_graph.h"
#include "bf_extfuncs.h"
#include "lg_intfuncs.h"
#include "lg_extfuncs.h"
#include "lg_versionfuncs.h"
#include "error.h"


/*
 * Traverse a tree to the indicated spot, maintaining a stack of nodes
 * with indications as to whether the node is shared by other verisons.
 * When done, the traversal stack will contain the traversal path, and 
 * the offset will indicate the offset of the first targeted byte in 
 * the path's (long) leaf page.  All pages on the path will remain 
 * fixed in the buffer pool. 
 *
 * NOTE: This code is almost an exact duplicate of LG_Traverse().
 */


 int 
LG_BuildShareStack (

	PID				*rootPid,	/* PID of tree's root page			*/
	SLOTINDEX		rootSlot,	/* slot of root, if on slotted page	*/
	PID				*stopPid,	/* PID of lowest page to traverse to*/
	int				offset,		/* byte offset from start of object	*/
	TSTACK			*stack,		/* stack for recording traversal	*/
	BUFGROUP		*bufGroup,	/* buffer group for all accesses	*/
	int				*shareCount /* number of shared pages			*/
)
{

    LGNODE			*nodePtr;		/* pointer to current node pointer		*/
	GROUPLINK		*nodeLink;
    int				height;			/* height of node we're looking at		*/
    int				travSlot;		/* slot index taken in traversal		*/
    PID				nodePid;		/* pid of the current node				*/
	BOOL			shared;			/* current node is shared	*/


	TRACE(TR_LG, TR_LEVEL_1);

    /*
     * Set up initial nodePid and stack
     */
    LG_ResetTstack(stack);
    nodePid = *rootPid;

	/*
	 *	The root node is never shared
	 */
	shared = FALSE;
	(*shareCount) = 0;

	/*
	 * Traverse until we hit the parent of the targeted leaf.
	 */
	do {

		/*
		 * Read the node in
		 */
		if ((nodeLink = bf_ReadPage(bufGroup, &nodePid,
			(rootSlot != NIL ? H_SLOTTED : H_LARGE),
			(rootSlot != NIL ? SLOTTED_PAGE2SIZE : LG_PAGE2SIZE), NL)) == NULL)	{

			SM_ERROR(TYPE_FATAL, esmINTERNAL);
		}

		/*
		 *	get the pointer to the node
		 */
		if (rootSlot != NIL)	{

			PAGESLOT		*slotPtr;
			SLOTTEDPAGE		*pagePtr;
			LARGEOBJROOT	*rootPtr;


			TRPRINT(TR_LG, TR_LEVEL_2, ("slot root pid:%d slot:%d",
							nodePid.page, rootSlot));

			/*
			 *	Get a pointer to the page, slot and object
			 */
			pagePtr = (SLOTTEDPAGE *) nodeLink->bufFrame;
			slotPtr = GETSLOTPTR(pagePtr, rootSlot);
			rootPtr = (LARGEOBJROOT *) GETOBJECTPTR(pagePtr, slotPtr);

			/*
			 *	get a pointer to the node
			 */
			nodePtr = (LGNODE *) &(rootPtr->rootPage);

			/*
			 *	check to see if the slot is the right type
			 */
			CHECK_SLOTROOT_MAGIC(nodePtr);

		} else {

			/*
			 *	the node pointer is just the page pointer
			 */
			nodePtr = (LGNODE *) nodeLink->bufFrame;
			lg_PrintNode(nodePtr, 0, nodePid.page);

			/*
			 *	check the magic number
			CHECK_LARGENODE_MAGIC(nodePtr);
			 */
		}

		/*
		 *	get the height of the current node
		 */
		height = nodePtr->header.height;

		/*
		 *	find out which slot to traverse
		 */
		travSlot = LG_BinarySearch(nodePtr, offset);

		/*
		 *	Push the node traversal information onto the stack
		 */
		LG_PushTstack(stack, travSlot, &nodePid, nodePtr,
					  FALSE, 0, nodeLink, rootSlot, shared);
		TRPRINT(TR_LG, TR_LEVEL_2, ("height:%d slot:%d rootSlot:%d",
				height, travSlot, rootSlot));

		/*
		 *	Count the number of shared nodes
		 */
		if (shared) (*shareCount)++;

		/*
		 *	Nil out the root slot for subsequent levels
		 */
		rootSlot = NIL;

		/*
		 *	Stop searching if we've reached the stopPid
		 */
		if (PIDEQ(nodePid, *stopPid)) {
			break;
		}

		/*
		 *	Set the offset relative to the subtree we're about to enter
		 */
		if (travSlot > 0) {
			offset -= (int)nodePtr->slot[travSlot - 1].count;
		}

		/*
		 *	get the pid of the next subtree to process
		 */
		nodePid = nodePtr->slot[travSlot].pid;

		/*
		 *	See if the next subtree is shared
		 */
		shared = shared | nodePtr->slot[travSlot].shared;

	} while (height > 1);

	/*
	 *	return success
	 */
	return(esmNOERROR);
}
