/*
 *   $RCSfile: LG_ComputeAppend.c,v $  
 *   $Revision: 1.1.1.1 $  
 *   $Date: 1996/05/04 21:55:23 $      
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/


#include "ess.h"
#include "sysdefs.h"
#include "trace.h"
#include "checking.h"
#include "error.h"
#include "io.h"
#include "page.h"
#include "list.h"
#include "object.h"
#include "tid.h"
#include "bf.h"
#include "chunk.h"
#include "volume.h"
#include "lock.h"
#include "io_extfuncs.h"
#include "bf_macro.h"
#include "lgobject.h"
#include "lg_intfuncs.h"
#include "lg_extfuncs.h"


/*
 * Routine to compute the effects of an append operation.  In particular,
 * it returns a list of leaf pages to be read and a list of replacement
 * pages to be used in updating them.  Some of this code could probably
 * be shared with LG_ComputeInsert().  It is separate because it allows
 * more flexibility in the algorithms;  we'll consider merging them later.
 */


 int
LG_ComputeAppend (

	int				length,			/* number of bytes to append				*/
	PAGELIST		*readList,		/* freed by caller */
	PAGELIST		*writeList,		/* freed by caller */
	TSTACK			*tstack 		/* stack for use in traversing object		*/
)
{
    int		i;				/* index for doing leaf work				*/
    int		totalBytes;		/* total byte count of interest				*/
    int		numLeaves;		/* number of leaves involved				*/
    int		lastCount;		/* byte count in (new) last leaf			*/
    int		prevCount;		/* byte count in (new) next-to-last leaf	*/
	TSTACKENT	*stackTop;
    LGNODE		*nodePtr;		/* indirect ptr to internal node in buffer	*/
	GROUPLINK	*nodeLink;
	PID		nodePid;		/* page id of the internal node */
    VOLID	volid;			/* volume id for page allocation purposes	*/
    int		slotIndex;		/* index of slot at which we're inserting	*/
    int		remainder;		/* # of remaining bytes, for balancing		*/
    int		usedBytes;		/* number of used bytes in previous leaf	*/
	int		rightBytes;		/* number of bytes in rightmost leaf		*/
    int		byteCount;		/* byte count for current leaf				*/
    PID		leafPid;		/* pid of current leaf						*/
    PID		*pidList = NULL;/* list of allocated pids					*/
	int		numShadows;


	TRPRINT(TR_LG, TR_LEVEL_1, ("appending:%d bytes", length));

    /*
     * Grab the parent of the leaf into which we're inserting.  Don't
     * pop the stack here, as we'll need it intact later.
     */
    if ((stackTop = LG_TopTstack(tstack)) == NULL)	{
		SM_ERROR(TYPE_FATAL, esmINTERNAL);
	}

	/*
	 *	get relevant fields from stack entry
	 */
	slotIndex = stackTop->slotIndex;
	nodePtr	  = stackTop->nodePtr;
	nodeLink  = stackTop->nodeLink;
	nodePid	  = stackTop->pid;
	volid	  = stackTop->pid.volid;
	TRPRINT(TR_LG, TR_LEVEL_2, ("slotIndex:%d", slotIndex));

    /*
     *	Get the total number of bytes that are currently in the
	 *	rightmost leaf page.  We need this information to see
	 *	if there is room.  Also note the relevant volume id.
     */
    rightBytes = LG_SlotCount(nodePtr, slotIndex) - length;
    TRPRINT(TR_LG, TR_LEVEL_2, ("rightBytes:%d", rightBytes));

	/*
	 *	get a list to hold the pages to read information.  Note that
	 *	at most two pages will be read.  This will only happen when
	 *	there is node overflow and the adjacent right node has free space.
	 *	If there is no space in the rightmost leaf then no pages will be read.
	 */
    if(LG_InitList(readList, 2) == esmFAILURE) {
		FAIL
	}

	/*
	 *	Check to see if there is any room in the rightmost node
	 */
	if (rightBytes < LGMAXLEAF)	{

		TRPRINT(TR_LG, TR_LEVEL_2, ("space in rightmost leaf"));

		/*
		 *	the total bytes we will be dealing with are the bytes
		 *	in the right most leaf plus the bytes to be appended
		 */
		totalBytes = rightBytes + length;

		/*
		 *	get the pid of the leaf page that will be inserted into
		 */
		leafPid = nodePtr->slot[slotIndex].pid;
		TRPRINT(TR_LG, TR_LEVEL_2, ("inserting into slot:%d pid:%d",
				slotIndex, leafPid.page));

		/*
		 *	put the page into the list to be read
		 */
		LG_AppendPage(readList, &leafPid, rightBytes, FALSE, nodeLink, nodePid.page, stackTop->rootSlot, slotIndex, stackTop->shared | nodePtr->slot[slotIndex].shared);

		/*
		 * If we're in an overflow situation, and if the previous leaf exists,
		 * then see if we can add some bytes to it.
		 */
		if ((totalBytes > LGMAXLEAF) && (slotIndex > 0)) {

			TRPRINT(TR_LG, TR_LEVEL_2, ("overflow and has left neighbor"));

			/*
			 *	get the byte count of the previous node
			 */
			usedBytes = LG_SlotCount(nodePtr, slotIndex - 1);

			/*
			 *	check to see if we have any free space
			 */
			if (usedBytes < LGMAXLEAF) {

				TRPRINT(TR_LG, TR_LEVEL_2, ("will use %d free bytes on 2nd last leaf",
						LGMAXLEAF - usedBytes));

				/*
				 * Okay, now act as though the append point was at the previous
				 * leaf, and record information needed to read that leaf as well.
				 */
				leafPid = nodePtr->slot[--slotIndex].pid;

				/*
				 *	put the neighbor page in the read list.  Insert it
				 *	into the list BEFORE the last leaf page entry.
				 */
				LG_InsertPage(readList, &leafPid, usedBytes, FALSE, 
							  NULL, nodePid.page, stackTop->rootSlot,
							  stackTop->shared | nodePtr->slot[slotIndex-1].shared);

				/*
				 *	modify the top of stack to reflect a path to the
				 *	previous leaf node
				 */
				tstack->stack[tstack->top].slotIndex--;

				/*
				 *	increment the total bytes to be read by the
				 *	bytes on the 2nd last leaf
				 */
				totalBytes += usedBytes;
				TRPRINT(TR_LG, TR_LEVEL_2, ("using left neighbor:%d total bytes:%d",
						leafPid.page, totalBytes));
			}
		}

	} else {

		TRPRINT(TR_LG, TR_LEVEL_2, ("no space in rightmost leaf"));

		/*
		 *	the total number of bytes we are dealing with are the
		 *	bytes that are being appended
		 */
		totalBytes = length;

		/*
		 *	modify the top of stack to reflect a path to the
		 *	slot to be appended, not inserted.
		 */
		tstack->stack[tstack->top].slotIndex++;

		/*
		 *	hack the count of the previous node
		 *	to account for the amount added during traversal.
		 */
		nodePtr->slot[slotIndex].count -= length;

		/*
		 *	fudge leafPid which is used for the near call
		 *	in allocating the new leaf Page
		 */
		leafPid = nodePtr->slot[slotIndex].pid;
	}
	TRPRINT(TR_LG, TR_LEVEL_2, ("totalBytes:%d", totalBytes));

	sm_SetAbortIfError();

    /*
     * Calculate how many leaves we'll need to hold totalBytes bytes.
     */
    numLeaves = (totalBytes / LGMAXLEAF) + (totalBytes % LGMAXLEAF > 0);
	TRPRINT(TR_LG, TR_LEVEL_2, ("numLeaves:%d", numLeaves));

	/*
	 *	calculate the number of shadow pages required
	 */
	numShadows = numLeaves - readList->numPages;

	/*
	 *	check to see if we have to allocate any new pages
	 */
	if (numShadows > 0)	{

		TRPRINT(TR_LG, TR_LEVEL_2, ("must allocate shadow pages"));

		/*
		 *	malloc space for the new page list
		 */
		if ((pidList = (PID *) malloc(numShadows * (int)sizeof(PID))) == NULL)	{
			SM_ERROR(TYPE_USER, esmMALLOCFAILED);
			FAIL
		}

		/*
		 * Gather up a list of resulting leaf pages, allocating new leaves
		 * (a couple of which will act as shadows) to handle the appended data.
		 * All but the last two pages should be full, with the last two pages
		 * getting prevCount and lastCount bytes, respectively.  (Note:  This
		 * must also work properly when only one or two pages are involved!)
		 */
		if (io_AllocPages(volid, LGDATA_PAGE2SIZE, numShadows,
						  &leafPid, pidList, PAGE_LARGEDATA))	{
			/* io_AllocPages does SM_ERROR */
			FAIL
		}
		TRPRINT(TR_LG, TR_LEVEL_2, ("allocated %d starting at page:%d",
				numShadows, pidList->page));
	}

	/*
	 *	initialize the write list or the resulting leaf pages
	 *	these pages will hold all the data (existing + appended)
	 */
    if(LG_InitList(writeList, numLeaves) == esmFAILURE) {
		FAIL
	}


    /*
     *	Calculate how to spread any remaining bytes over the last two leaves.
	 *	lastCount represents the bytes in the last leaf and prevCount is
	 *	the number of bytes to put in 2nd last leaf.
     */
    if (numLeaves > 1) {

		/*
		 *	calculate remaining bytes to split over the two
		 *	rightmost leaf nodes
		 */
        remainder = totalBytes - ((numLeaves - 2) * LGMAXLEAF);

		/*
		 *	calculate the number of bytes to put in 2nd last leaf
		 *	Put the extra odd byte in first page, if any
		 */
        prevCount = ALIGNSIZE((remainder - 
                    MAX(LGMINLEAF, (remainder - LGMAXLEAF))));

		/*
		 *	calculate the number of bytes to put in the last leaf
		 */
        /*lastCount = MAX(LGMINLEAF, (remainder - LGMAXLEAF));*/
		lastCount = remainder - prevCount;
		SM_ASSERT(LEVEL_3, remainder == (prevCount + lastCount));
		TRPRINT(TR_LG, TR_LEVEL_2, ("remainder:%d prevCount:%d lastCount:%d",
				remainder, prevCount, lastCount));

    } else {

		/*
		 *	all bytes will fit in the last leaf
		 */
        lastCount = totalBytes;
    }

	/*
	 *	process each leaf node that will be written.  These conditions
	 *	must work when there are 1,2 .. n leaf nodes.  We are calculating
	 *	the number of bytes each leave will hold, not their relative offset
	 */
    for (i = 0; i < numLeaves; i++) {

		/*
		 *	calculate the number bytes the node will hold
		 */
		if (i == numLeaves - 2) {

			/*
			 *	2nd last leaf
			 */
			byteCount = prevCount;

		} else if (i == numLeaves - 1) {

			/*
			 *	last leaf
			 */
			byteCount = lastCount;

		} else {
			
			/*
			 *	full node
			 */
			byteCount = LGMAXLEAF;
		}

		/*
		 *	decide where to get the pid
		 */
		if (i < readList->numPages)	{

			/*
			 *	store the data for the node.  Mark the node as being dirty.
			 */
			LG_AppendPage(writeList, &(readList->pages[i].pid),
						  byteCount, TRUE, NULL, NIL, NIL, NIL,
						  readList->pages[i].shared);
			TRPRINT(TR_LG, TR_LEVEL_2, ("putting:%d bytes in page:%d", byteCount, i));

		} else {

			/*
			 *	store the data for the node
			 */
			LG_AppendPage(writeList, &(pidList[i - readList->numPages]),
						  byteCount, TRUE, NULL, NIL, NIL, NIL,
						  FALSE);
		}
		TRPRINT(TR_LG, TR_LEVEL_2, ("putting:%d bytes in page:%d", byteCount, i));
    }
Success:
	SM_FREE_PTR(pidList);
    return esmNOERROR;

FailureReturn:
	SM_FREE_PTR(pidList);
	return esmFAILURE;
}
