/*
 *   $RCSfile: LG_ComputeInsert.c,v $  
 *   $Revision: 1.1.1.1 $  
 *   $Date: 1996/05/04 21:55:23 $      
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/


#include "ess.h"
#include "sysdefs.h"
#include "trace.h"
#include "checking.h"
#include "error.h"
#include "io.h"
#include "page.h"
#include "list.h"
#include "object.h"
#include "tid.h"
#include "bf.h"
#include "lock.h"
#include "chunk.h"
#include "volume.h"
#include "bf_macro.h"
#include "lgobject.h"
#include "io_extfuncs.h"
#include "lg_intfuncs.h"
#include "lg_extfuncs.h"


/*
 * Routine to compute the effects of an insert operation.  In particular,
 * it returns a list of leaf pages to be read and a list of replacement
 * pages to be used in updating them.  It also modifies the offset within
 * the chunk produced by the read list to indicate where the inserted data
 * should go.  (This offset may be correct as it comes into this routine,
 * but then again it may not - the latter occurs when rebalancing involves
 * the left neighbor of the node containing the insert point.)  Some of
 * this code could probably be shared with LG_ComputeAppend().
 */


 int 
LG_ComputeInsert (

	int				*offset,	/* place to insert bytes					*/
	int				length,		/* number of bytes to insert				*/
	PAGELIST		*readList,	/* list of pid/cnt pairs to be read in		*/
	PAGELIST		*writeList,	/* list of pid/cnt pairs to write back out	*/
	TSTACK			*tstack 	/* stack for use in traversing object		*/
)
{
    int			i;			/* index for doing leaf work				*/
    int			totalBytes;	/* total byte count of interest				*/
    int			numLeaves;	/* number of leaves involved				*/
    int			bytesPerLeaf;/* balanced number of bytes per leaf		*/
	TSTACKENT	*stackTop;
    LGNODE		*nodePtr;	/* indirect ptr to internal node in buffer	*/
	GROUPLINK	*nodeLink;
    PID			nodePid;	/* page id for the buffered internal node	*/
    int			slotIndex;	/* index of slot at which we're inserting	*/
    int			remainder;	/* # of remaining bytes, for balancing		*/
    int			leftBytes;	/* number of used bytes in left neighbor	*/
    int			rightBytes;	/* number of used bytes in right neighbor	*/
    int			rightFree;	/* number of free bytes in left neighbor	*/
    int			leftFree;	/* number of free bytes in right neighbor	*/
    int			byteCount;	/* byte count for current leaf				*/
    PID			leafPid;	/* pid of current leaf						*/
    PID			*pidList = NULL;/* list of allocated pids					*/
	int			overflow;
	int			numShadows;


	TRPRINT(TR_LG, TR_LEVEL_1, ("offset:%d length:%d", *offset, length));

    /*
     * Grab the parent of the leaf into which we're inserting.  Don't
     * pop the stack here, as we'll need it intact later.
     */
    if ((stackTop = LG_TopTstack(tstack)) == NULL)	{
		SM_ERROR(TYPE_FATAL, esmINTERNAL);
	}

	slotIndex = stackTop->slotIndex;
	nodePtr	  = stackTop->nodePtr;
	nodeLink  = stackTop->nodeLink;
	nodePid   = stackTop->pid;

    /*
     * Get the total number of bytes that we'll be playing around with,
     * which has been computed as a side effect of the insert-oriented
     * tree traversal.  Also note the relevant volume id.
     */
    totalBytes = LG_SlotCount(nodePtr, slotIndex);
	TRPRINT(TR_LG, TR_LEVEL_2, ("totalBytes:%d", totalBytes));

    /*
     * Record information on old leaf to be read for the insert.
     */
    if(LG_InitList(readList, 2) == esmFAILURE) {
		FAIL
	}
    leafPid = nodePtr->slot[slotIndex].pid;

	/*
	 *	calculate the original bytes in the leaf and record
	 *	the information necessary to read it
	 */
    byteCount = totalBytes - length;
    LG_AppendPage(readList, &leafPid, byteCount, FALSE, nodeLink, nodePid.page, stackTop->rootSlot, slotIndex, stackTop->shared | nodePtr->slot[slotIndex].shared);
	sm_SetAbortIfError();

    /*
     * If we're in an overflow situation, and if the previous or next leaf
     * exists, then see if we can win by adding some bytes to the one with
     * the greater amount of free space.
     */
    if (totalBytes > LGMAXLEAF)	{

		TRPRINT(TR_LG, TR_LEVEL_2, ("leaf has overflowed"));

        /*
         * Get info on left neighbor for use in doing balancing.
         */
        if (slotIndex > 0) {

	    	leftBytes = LG_SlotCount(nodePtr, slotIndex - 1);
	    	leftFree = LGMAXLEAF - leftBytes;

        } else {

	    	leftFree = 0;
	    	leftBytes = 0;
        }
		TRPRINT(TR_LG, TR_LEVEL_2, ("leftBytes:%d leftFree:%d", leftBytes, leftFree));

        /*
         * Get info on right neighbor for use in doing balancing.
         */
        if (slotIndex < nodePtr->header.numSlots - 1) {

	    	rightBytes = LG_SlotCount(nodePtr, slotIndex + 1);
	    	rightFree = LGMAXLEAF - rightBytes;

        } else {

	    	rightFree = 0;
        }
		TRPRINT(TR_LG, TR_LEVEL_2, ("rightBytes:%d rightFree:%d", rightBytes, rightFree));

		/*
		 *	calculate how many bytes have overflowed
		 */
		overflow = totalBytes - LGMAXLEAF;
		TRPRINT(TR_LG, TR_LEVEL_2, ("overflow:%d", overflow));

		/*
		 * Now decide whether or not neighbor balancing will really help.
		 * If so, arrange for balancing to take place appropriately.
		 */
		 if (leftFree > overflow) {

			TRPRINT(TR_LG, TR_LEVEL_2, ("space in left neighbor"));

			/*
			 *	The insert will occur into the left neighbor
			 */
			leafPid = nodePtr->slot[slotIndex - 1].pid;
			byteCount = leftBytes;

			/*
			 *	record the information to read that left leaf
			 */
			LG_InsertPage(readList, &leafPid, byteCount, FALSE, NULL, 
						  nodePid.page, stackTop->rootSlot, 
						  stackTop->shared | nodePtr->slot[slotIndex - 1].shared);
			/*
			 *	we are going to traverse down by on slot to the left
			 *	so we must fudge the current stack
			 */
			slotIndex--;
			tstack->stack[tstack->top].slotIndex--;

			/*
			 *	The spot that we will insert to has moved to the
			 *	right by the number of bytes in the left page
			 */
			*offset += leftBytes;

			/*
			 *	Increment the bytes to be manipulated by the
			 *	number of bytes in the left leaf
			 */
			totalBytes += byteCount;

		} else if (rightFree > overflow) {

			TRPRINT(TR_LG, TR_LEVEL_2, ("space in right neighbor"));

			/*
			 *	record the information necessary to read the right page
			 */
			leafPid = nodePtr->slot[slotIndex + 1].pid;
			byteCount = rightBytes;
			LG_AppendPage(readList, &leafPid, byteCount, FALSE, NULL, 
						  nodePid.page, stackTop->rootSlot, slotIndex, 
						  stackTop->shared | nodePtr->slot[slotIndex + 1].shared);

			/*
			 *	we must include the number of bytes in the left node
			 */
			totalBytes += byteCount;
		}
    }

    /*
     * Calculate how many leaves we'll need to hold totalBytes bytes,
     * and also how to spread the data over these leaves.
     */
    numLeaves = (totalBytes / LGMAXLEAF) + (totalBytes % LGMAXLEAF > 0);
	bytesPerLeaf = totalBytes / numLeaves;
	remainder = totalBytes % numLeaves;
	TRPRINT(TR_LG, TR_LEVEL_2, ("totalBytes:%d numLeaves:%d bytesPerLeaf:%d remainder:%d offset:%d",
			totalBytes, numLeaves, bytesPerLeaf, remainder, *offset));


	/*
	 *	calculate the number of shadows that will have to be allocated
	 */
	numShadows = numLeaves - readList->numPages;

	/*
	 *	allocate shadow pages if necessary
	 */
	if (numShadows > 0)	{

		TRPRINT(TR_LG, TR_LEVEL_2, ("allocating:%d shadows", numShadows));

		if ((pidList = (PID *) malloc(numShadows * (int)sizeof(PID))) == NULL)	{
			SM_ERROR(TYPE_USER, esmMALLOCFAILED);
			FAIL
		}

		/*
		 * Gather up a list of resulting leaf pages, allocating new leaves
		 * (a couple of which will act as shadows) to handle the appended data.
		 */
		if (io_AllocPages(nodePid.volid, LGDATA_PAGE2SIZE, numShadows,
						  &leafPid, pidList, PAGE_LARGEDATA))	{
			/* io_AllocPages does SM_ERROR */
			FAIL
		}
		TRPRINT(TR_LG, TR_LEVEL_2, ("allocated %d starting at page:%d",
				numShadows, pidList->page));
	}

	/*
	 *	initialize the write list or the resulting leaf pages
	 *	these pages will hold all the data (existing + inserted)
	 */
    if(LG_InitList(writeList, numLeaves) == esmFAILURE ) {
		FAIL
	}

	/*
	 *	Check to see if we have more than one resultant leaf
	 */
	if (numLeaves == 1)	{

		/*
		 *	just put the total bytes in the single leaf
		 */
        LG_AppendPage(writeList, &(readList->pages[0].pid), totalBytes,
					  TRUE, NULL, NIL, NIL, NIL, readList->pages[0].shared);
		TRPRINT(TR_LG, TR_LEVEL_2, ("page:%d bytes:%d", pidList->page, byteCount));

	} else {

		/*
		 *	allocate the appropriate number of bytes to each resultant leaf
		 */
		for (i = 0; i < numLeaves; i++) {

			/*
			 *	determine the number of bytes in this leaf
			 */
			byteCount = bytesPerLeaf + (remainder-- > 0);

			/*
			 *	decide where to get the pid
			 */
			if (i < readList->numPages)	{
				LG_AppendPage(writeList, &(readList->pages[i].pid),
							  byteCount, TRUE, NULL, NIL, NIL, NIL, 
							  readList->pages[0].shared);
				TRPRINT(TR_LG, TR_LEVEL_2, ("putting:%d bytes in page:%d", byteCount, i));
			} else {
				LG_AppendPage(writeList, &(pidList[i - readList->numPages]),
							  byteCount, TRUE, NULL, NIL, NIL, NIL, FALSE);
			}
			TRPRINT(TR_LG, TR_LEVEL_2, ("putting:%d bytes in page:%d", byteCount, i));
		}
    }
Success:
	SM_FREE_PTR(pidList);
    return esmNOERROR;

FailureReturn:
	SM_FREE_PTR(pidList);
    return esmFAILURE;

}
