/*
 *   $RCSfile: LG_DeleteData.c,v $  
 *   $Revision: 1.1.1.1 $  
 *   $Date: 1996/05/04 21:55:24 $      
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/


#include <stdio.h>
#include "ess.h"
#include "trace.h"
#include "checking.h"
#include "error.h"
#include "io.h"
#include "page.h"
#include "list.h"
#include "object.h"
#include "tid.h"
#include "bf.h"
#include "chunk.h"
#include "bf_macro.h"
#include "lock.h"
#include "lgobject.h"
#include "volume.h"
#include "logaction.h"
#include "clog.h"
#include "sm_state.h"
#include "io_extfuncs.h"
#include "bf_extfuncs.h"
#include "lg_intfuncs.h"
#include "lg_extfuncs.h"
#include "trans_funcs.h"
#include "clog_extfuncs.h"


/*
 * Routine to delete bytes from the last affected leaf node (where a simple
 * count change won't do the job).
 */


 int 
LG_DeleteData (

	FID				*fid,
	PID				*leafPid,	/* page id of (long) leaf node		*/
	int				leafSize,	/* number of valid bytes in leaf	*/
	int				start,		/* offset of starting pnt for deletion*/
	int				length,		/* number of bytes to be deleted	*/
	BUFGROUP		*bufGroup,	/* buffer group to be used			*/
	PID				*shadowPid,	/* page id of leaf node's shadow	*/
	BOOL			*danger,	/* return value of danger indicator	*/
	BOOL			shared 		/* if true, shadow the page 		*/
)
{

    CHUNKDESC		*chunkDesc;		/* descriptor for leaf accesses	*/
    PAGELIST		oldLeaf;		/* page list for old leaf block	*/
	LGDATAPAGE		*lgDataPage;
	OPERATIONINFO	opInfo;
	int				e;

	TRPRINT(TR_LG, TR_LEVEL_1, ("leafPid:%d leafSize:%d start:%d length:%d",
					 leafPid->page, leafSize, start, length));

    /*
     *	First read in the affected data.
     */
    if (LG_InitList(&oldLeaf, 1) == esmFAILURE)	{
		return esmFAILURE;
	}

	/*
	 *	append the page to be read
	 */
    LG_AppendPage(&oldLeaf, leafPid, leafSize, FALSE, NULL, ZERO, NIL, NIL, FALSE);

	/*
	 *	read in the page
	 */
    if (bf_ReadChunk(fid, bufGroup, 1, 1, oldLeaf.pages, ZERO, ZERO, &chunkDesc))	{

		return esmFAILURE;
	}

	/*
	 *	Get a pointer to the 
	 */
	lgDataPage = (LGDATAPAGE*)chunkDesc->bufFrame;

	/*
	 *	If the page is shared, make a copy so delete can be done on copy
	 */
	if (shared) {

		/*
		 *	allocate a new page
		 */
		SM_ERROR(TYPE_LOG, esmINTERNAL);		/* incorrect code */
		if (io_AllocPages(leafPid->volid, LGDATA_PAGE2SIZE, 1,
						  leafPid, shadowPid, PAGE_LARGEDATA))	{

			return(esmFAILURE);
		}

		/*
		 *	Replace the page with the shadow in the buffer
		 */
		e = bf_ReplaceLeafChunk(chunkDesc, chunkDesc->chunkPage,
								shadowPid, NULL);
		if (e != esmNOERROR) {
			internalAbort();
			SM_ERROR(TYPE_LOG, esmTRANSABORTED);
			return(esmFAILURE);
		}

		/*
		 *	Copy the data to overwrite the deleted portion
		 */
		bf_DeleteChunk(chunkDesc, start, length);

		if (GET_LOGLEVEL(LOG_DATA) ) {
			/*
			 *	Log the delete as an append to the new page
			 */
			opInfo.action = LOG_ACTION_APPEND_LARGE_DATA;
			opInfo.flags = 0;
			opInfo.start = 0;
			opInfo.size = chunkDesc->totalLength;
			opInfo.oldData = NULL;
			opInfo.newData = chunkDesc->bufFrame;
			opInfo.actionPid = *shadowPid;
			clog_LogModifyLgPage(&opInfo);
		}
	} else {

		*shadowPid = *leafPid;

		if (GET_LOGLEVEL(LOG_DATA) ) {
			/*
			 *	Log the delete
			 */
			opInfo.action = LOG_ACTION_DELETE_LARGE_DATA;
			opInfo.flags = 0;
			opInfo.start = start;
			opInfo.size = length;
			opInfo.oldData = chunkDesc->bufFrame+start;
			/* note, new data is ptr to the previous size of the data */
			opInfo.newData = (char *) &leafSize;
			opInfo.actionPid = *shadowPid;
			clog_LogModifyLgPage(&opInfo);
		}

		/*
		 *	delete the data from the chunk
		 */
		bf_DeleteChunk(chunkDesc, start, length);
	}

	/* Unfix the chunk and deallocate the old page */
    bf_UnfixChunk(chunkDesc, BF_DEFAULT);

	/* determine whether the leaf has underflowed */
    *danger = ((leafSize - length) < LGMINLEAF);
	TRPRINT(TR_LG, TR_LEVEL_2, ("new leaf size:%d danger:%d", leafSize - length, *danger));

	/* return success */
	return(esmNOERROR);
}
