/*
 *   $RCSfile: LG_DeletePhase.c,v $  
 *   $Revision: 1.1.1.1 $  
 *   $Date: 1996/05/04 21:55:23 $      
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/


#include "sysdefs.h"
#include "ess.h"
#include "trace.h"
#include "checking.h"
#include "error.h"
#include "io.h"
#include "page.h"
#include "list.h"
#include "object.h"
#include "tid.h"
#include "bf.h"
#include "chunk.h"
#include "logaction.h"
#include "clog.h"
#include "bf_macro.h"
#include "lock.h"
#include "sm_macro.h"
#include "lgobject.h"
#include "volume.h"
#include "sm_state.h"
#include "io_extfuncs.h"
#include "bf_extfuncs.h"
#include "lg_intfuncs.h"
#include "lg_extfuncs.h"
#include "clog_extfuncs.h"
#include "util_funcs.h"


/*
 * Routine to delete the desired range of bytes.  Unnecessary index 
 * nodes are discarded, and changed nodes (in working versions) are
 * shadowed, but the index structure is not rebalanced at this point.
 *
 * The versioned parameter is used since the "shared" flags on
 * large object index slot pointers are not guaranteed to be false
 * for non-versioned large objects.
 */


 int
LG_DeletePhase (
	FID			*fid,
	PID			*nodePid,		/* page id of current index node*/
	SLOTINDEX	nodeSlot,		/* slot for current node		*/
	PID			*rootPid,		/* page id of index root node	*/
	SLOTINDEX	rootSlot,		/* slot for root node			*/
	int			start,			/* offset of starting point for deletion*/
	int			length,			/* number of bytes to be deleted*/
	BUFGROUP	*bufGroup,		/* buffer group to be used		*/
	PID			*shadowPid,		/* new pid for index node (shadow page)	*/
	TSTACK		*lstack,		/* left cut path traversal stack*/
	TSTACK		*rstack,		/* right cut path traversal stack*/
	BOOL		*danger,		/* return value of danger indicator*/
	BOOL		versioned,		/* object is a working version	*/
	BOOL		shared 			/* this node is shared by another version*/
)
{
    LGNODE		*nodePtr;		/* indirect ptr to buffered internal node	*/
	OPERATIONINFO	opInfo;		/* operation info to log			*/
    BOOL		lastLevel;		/* true if last index level above leaves	*/
    int			totalSlots;		/* total number of slots in the node*/
    PID			tempPid;		/* pid of child or temporary pid value*/
    int			firstSlot;		/* slot index of first affected slot*/
    int			lastSlot;		/* slot index of last affected slot	*/
    int			nextSlot;		/* first later unaffected slot's index*/
    int			pathSlot;		/* slot index of slot for cut path use*/
    int			numRightDel;	/* number of rightmost bytes deleted*/
    int			numMiddleDel;	/* number of middle bytes deleted	*/
    int			numDelSlots;	/* number of completely deleted slots*/
    int			finalSlotCnt;	/* final number of slots in node	*/
    int			firstSlotByte;	/* relative offset of first byte in a slot	*/
    int			deleteSize;		/* size of delete request in bytes	*/
    int			slotIndex;		/* slot index for processing the node*/
    int			relStartByte;	/* relative number of first byte of slot	*/
    int			lastSlotByte;	/* byte number of last byte of slot	*/
    int			numSlotBytes;	/* number of bytes in slot's subtree*/
    int			slotDelBytes;	/* number of slot's bytes to delete	*/
    int			numDelBytes;	/* number of bytes deleted thus far	*/
    int			lastDelByte;	/* byte number of slot's last deleted byte	*/
    BOOL		bytesOnLeft;	/* non-deleted bytes remain on left	*/
    BOOL		bytesOnRight;	/* non-deleted bytes remain on right*/
    BOOL		leftDanger;		/* true if left path child in danger*/
    BOOL		rightDanger;	/* true if right path child in danger*/
    BOOL		*dangerPtr;		/* ptr to leftDanger or rightDanger	*/
    BOOL		leftCutPath;	/* true if we're on the left cut path*/
    int			minSlotCnt;		/* minimum slot count that may result*/
    int			dangerCnt;		/* number of kids in danger			*/
	GROUPLINK	*nodeLink;
	LGNODE		tempLgNode;		/* used to keep undo info			*/


	TRPRINT(TR_LG, TR_LEVEL_1, ("pid:%d start:%d length:%d nodeSlot:%d",
			nodePid->page, start, length, nodeSlot));

    /*
     *	Read in the node, find the first slot of interest, and then record
     *	it on the appropriate traversal path stack.  First, though, shadow
     *	the node (since it will be updated), leaving it fixed in the buffer
     *	pool.  (It will be unfixed by the "zip up index" phase.)  Finally,
     *	see if we're at the last level above the data and record the total
     *	number of slots in the node for later use.
     */
    if ((nodeLink = bf_ReadPage(bufGroup, nodePid, 
		(nodeSlot != NIL ? H_SLOTTED : H_LARGE),
		(nodeSlot != NIL ? SLOTTED_PAGE2SIZE : LG_PAGE2SIZE), NL)) == NULL)	{

		SM_ERROR(TYPE_FATAL, esmINTERNAL);
	}

	/*
	 *	Shadow the node if it is shared by another version.  Shadow 
	 *	pid is used for all further operations concerning this node.
	 */
	if (versioned && shared) {
		if (LG_Shadow(nodePid, shadowPid, rootPid, bufGroup, &nodeLink))	{

			return(esmFAILURE);
		}
	} else {
		*shadowPid = *nodePid;
		sm_SetAbortIfError();
	}

	if (nodeSlot != NIL)	{

		PAGESLOT		*slotPtr;
		SLOTTEDPAGE		*pagePtr;
		LARGEOBJROOT	*rootPtr;

		TRPRINT(TR_LG, TR_LEVEL_2, 
			("slot root pid:%d slot:%d", nodePid->page, nodeSlot));
		SM_ASSERT(LEVEL_3, PPIDEQ(shadowPid, nodePid));
		pagePtr = (SLOTTEDPAGE *) nodeLink->bufFrame;
		slotPtr = GETSLOTPTR(pagePtr, nodeSlot);
		rootPtr = (LARGEOBJROOT *) GETOBJECTPTR(pagePtr, slotPtr);

		nodePtr = (LGNODE *) &(rootPtr->rootPage);

		CHECK_SLOTROOT_MAGIC(nodePtr);
	} else {
		TRPRINT(TR_LG, TR_LEVEL_2, ("large index node"));

		nodePtr = (LGNODE *) nodeLink->bufFrame;
		CHECK_LARGENODE_MAGIC(nodePtr);
	}

	/*
	 *	lastLevel - is this the index node directly above leaves
	 *	totalSlots - total number of slots in the index node
	 *	leftCutPath - 
	 */
    lastLevel = (nodePtr->header.height == 1);
    totalSlots = (int)nodePtr->header.numSlots;
    leftCutPath = (lstack != NULL);

	/*
	 *	initial danger settings to false
	 */
    leftDanger = FALSE;
    rightDanger = FALSE;
    dangerPtr = &leftDanger;
	TRPRINT(TR_LG, TR_LEVEL_2, ("setting danger to left"));

	/*
	 *	numDelBytes - Total number of bytes deleted on this level
	 *	numRightDel - Total number of right bytes deleted on this level
	 *	numMiddleDel -Total number of middle bytes deleted on this level
	 */
    numDelBytes = 0;
    numRightDel = 0;
    numMiddleDel = 0;

	/*
	 *	lastDelByte - last byte that will be deleted in this level
	 *	deleteSize  - amount of data that will be deleted
	 */
    lastDelByte = start + length - 1;
    deleteSize = length;

	/*
	 *	firstSlot	  - leftmost slot whose range contains start
	 *	lastSlot	  - last Slot touched by range
	 *	slotIndex	  - slot currently being processed
	 *	firstSlotByte - first byte contained in the current slot
	 *	lastSlotByte  - last byte contained in current slot
	 *	numSlotBytes  - number of bytes referenced by current slot
	 *	relStartByte  - offset within slot of first byte to be deleted
	 *	length        - remaining number of bytes in this level to be deleted
	 */
    lastSlot = -1;
    firstSlot = LG_BinarySearch(nodePtr, start);
    slotIndex = firstSlot;
	TRPRINT(TR_LG, TR_LEVEL_2, ("firstSlot:%d", firstSlot));
    firstSlotByte = (slotIndex > 0) ? (int)nodePtr->slot[slotIndex - 1].count : 0;
    lastSlotByte = (int)nodePtr->slot[slotIndex].count - 1;
    numSlotBytes = lastSlotByte - firstSlotByte + 1;
    relStartByte = start - firstSlotByte;

	/*
	 *	Remember the original data for use in generating log records
	 */
	bcopy((char*) &(nodePtr->slot[firstSlot]), (char *) &(tempLgNode.slot[firstSlot]), (int)sizeof(LGSLOT)*((int)nodePtr->header.numSlots - firstSlot));

    /*
     * Step through the range of slots affected by the delete operation,
     * processing them one by one.  Update their counts along the way.
     * There are two orthogonal considerations to be accounted for during
     * delete processing.  The first is the height of the current node
     * (i.e., whether its children are leaf nodes or upper-level nodes).
     * The other is which of four deletion classes a particular slot
     * is an instance of - (i) all bytes deleted, (ii) right-hand bytes
     * deleted, (iii) left-hand bytes deleted, or (iv) only middle bytes
     * deleted.
     */
    for (;;)	{
		TRPRINT(TR_LG, TR_LEVEL_2, 
			("slotIndex:%d firstSlotByte:%d",
				slotIndex, firstSlotByte));
		TRPRINT(TR_LG, TR_LEVEL_2, 
			("lastSlotByte:%d numSlotBytes:%d relStartByte:%d",
				lastSlotByte, numSlotBytes, relStartByte));
		/*
		 *	Determine if there are bytes to the left and right the delete
		 *	range respectively that will remain after the delete
		 */
        bytesOnLeft = (start > firstSlotByte);
        bytesOnRight = (lastDelByte < lastSlotByte);
		TRPRINT(TR_LG, TR_LEVEL_2, ("bytesOnLeft:%d bytesOnRight:%d",
				bytesOnLeft, bytesOnRight));

        if (bytesOnLeft && bytesOnRight) {
			TRPRINT(TR_LG, TR_LEVEL_2, ("slot completely contains deleted bytes"));
			/*
			 * Case (i) - Slot completely contains the deleted bytes.
			 *			- the number of bytes deleted will just be length
			 *			- the loop can be iterated just this one time.
			 */
			slotDelBytes = length;
			/*
			 *	check to see if we are at the last level
			 *	or higher in the index 
			 */
			if (lastLevel) {
				/*
				 *	We must read in and delete some data from a chunk page
				 */
				if (LG_DeleteData(fid, &(nodePtr->slot[slotIndex].pid),
						  numSlotBytes, relStartByte,
						  length, bufGroup, &tempPid, dangerPtr,
						  versioned && 
							(shared || nodePtr->slot[slotIndex].shared) ))	{
					FAIL
				}
			} else {
				/* descend the tree */
				if (LG_DeletePhase(fid, &(nodePtr->slot[slotIndex].pid),
						   NIL, rootPid, rootSlot, relStartByte,
						   length, bufGroup, &tempPid, lstack, 
						   rstack, dangerPtr, versioned,
						   shared || nodePtr->slot[slotIndex].shared))	{
					FAIL
				}
			}

			/*
			 *	record the new shadow page
			 */
			nodePtr->slot[slotIndex].pid = tempPid;
			nodePtr->slot[slotIndex].shared = FALSE;
			TRPRINT(TR_LG, TR_LEVEL_2, ("putting pid:%d in slot:%d",
					tempPid.page, slotIndex));

			/*
			 *	increment the number deleted from a "middle" operation
			 */
			numMiddleDel = length;
        } else if (bytesOnLeft) {
			TRPRINT(TR_LG, TR_LEVEL_2, 
				("deleted bytes on right leaving bytes on left"));

			/*
			 * Case (ii) - Deleted bytes lie on right side of slot.
			 *			 - number of bytes deleted up to the last byte
			 */
			slotDelBytes = lastSlotByte - start + 1;

			if (lastLevel) {
				/*
				 *	We must read in and delete some data from a chunk page
				 */
				if (LG_DeleteData(fid, &(nodePtr->slot[slotIndex].pid),
						  numSlotBytes, relStartByte,
						  slotDelBytes, bufGroup, &tempPid, dangerPtr,
						  versioned && 
							(shared || nodePtr->slot[slotIndex].shared) ))	{
					FAIL
				}
			} else {
				/*
				 *	descend the tree
				 *	we are traversing the left cut path.
				 */
				if (LG_DeletePhase(fid, &(nodePtr->slot[slotIndex].pid),
						   NIL, rootPid, rootSlot, relStartByte,
						   slotDelBytes, bufGroup, &tempPid, 
						   lstack, NULL, dangerPtr, versioned,
						   shared || nodePtr->slot[slotIndex].shared))	{
					FAIL
				}

				/* 	record the newly allocated shadow pid */
				nodePtr->slot[slotIndex].pid = tempPid;
				nodePtr->slot[slotIndex].shared = FALSE;
				TRPRINT(TR_LG, TR_LEVEL_2, ("putting pid:%d in slot:%d",
						tempPid.page, slotIndex));
			}

			/*
			 *	increment the number of bytes deleted on the right
			 *	by the bytes deleted from this slot
			 */
			numRightDel = slotDelBytes;
			TRPRINT(TR_LG, TR_LEVEL_2, ("numRightDel:%d", numRightDel));
        } else if (bytesOnRight) {
			TRPRINT(TR_LG, TR_LEVEL_2, 
			("deleted bytes on left leaving bytes on right"));

			/*
			 * Case (iii) - Deleted bytes lie on left side of slot.
			 *			  - this must be the last interation of the loop
			 */
			slotDelBytes = length;

			/* see if we are at the leaf level */
			if (lastLevel) {
				/*
				 *	We must read in a delete data from the chunk page
				 */
				if (LG_DeleteData(fid, &(nodePtr->slot[slotIndex].pid),
					  numSlotBytes, relStartByte,
					  length, bufGroup, &tempPid, dangerPtr,
					  versioned && 
					  (shared || nodePtr->slot[slotIndex].shared) ))	{
					FAIL
				}
			} else {

				/*
				 *	recurse and descend the tree.  Note that we
				 *	will be traversing the right cut path
				 */
				if (LG_DeletePhase(fid, &(nodePtr->slot[slotIndex].pid),
					   NIL, rootPid, rootSlot, relStartByte,
					   length, bufGroup, &tempPid, NULL, 
					   rstack, dangerPtr, versioned, 
					   shared || nodePtr->slot[slotIndex].shared))	{
					FAIL
				}
			}

			/*
			 *	record the newly allocated shadow pid
			 */
			nodePtr->slot[slotIndex].pid = tempPid;
			nodePtr->slot[slotIndex].shared = FALSE;
			TRPRINT(TR_LG, TR_LEVEL_2, ("putting pid:%d in slot:%d",
					tempPid.page, slotIndex));
			TRPRINT(TR_LG, TR_LEVEL_2, ("exiting loop"));
        } else {
			TRPRINT(TR_LG, TR_LEVEL_2, ("deleted bytes consume this slot"));
			/*
			 * Case (iv) - Deleted bytes completely subsume this slot.
			 */
			slotDelBytes = numSlotBytes;
			if (lastLevel) {
				/*
				 *	Blow the page and any containing chunks out of the cache
				 */
				tempPid = nodePtr->slot[slotIndex].pid;
				bf_HackChunkPage(&tempPid, TRUE /*all chunks*/);
				if (versioned && (shared || nodePtr->slot[slotIndex].shared)) {
					/* do nothing */ ;
				} else {
					TRPRINT(TR_LG, TR_LEVEL_2, 
					("deallocating page:%d", tempPid.page));
					if (io_DeallocPages(LGDATA_PAGE2SIZE, 1, &tempPid, 
												FALSE, PAGE_LARGEDATA))	{
						FAIL
					}
				}
				PIDCLEAR(nodePtr->slot[slotIndex].pid);
			} else {
				/*
				 *	destroy the entire subtree rooted at this slot
				 */
				TRPRINT(TR_LG, TR_LEVEL_2, ("slotIndex:%d pid:%d",
						slotIndex, nodePtr->slot[slotIndex].pid.page));
				if (lg_DestroyObject(&(nodePtr->slot[slotIndex].pid), 
						bufGroup, nodeSlot, versioned, 
						shared || nodePtr->slot[slotIndex].shared))	{
					FAIL
				}
			}
        }

		/*
		 * Update node's byte count to reflect deleted bytes.
		 */ 
		numDelBytes += slotDelBytes;
		TRPRINT(TR_LG, TR_LEVEL_2, ("slotDelBytes:%d numDelBytes:%d",
				slotDelBytes, numDelBytes));
		nodePtr->slot[slotIndex].count -= numDelBytes;
		TRPRINT(TR_LG, TR_LEVEL_2, ("updating slot:%d to count:%d",
				slotIndex, nodePtr->slot[slotIndex].count));

		/*
		 * Flag termination condition or else prepare for next iteration.
		 */
		if (lastSlotByte >= lastDelByte) {
			lastSlot = slotIndex;
			TRPRINT(TR_LG, TR_LEVEL_2, ("lastSlot:%d", lastSlot));
			break;
		} else {
			/*
			 *	set up conditions for iterating the loop for
			 *	processing the next slot in the page
			 */
			slotIndex++;
			length -= slotDelBytes;
			firstSlotByte += numSlotBytes;
			lastSlotByte = (int)nodePtr->slot[slotIndex].count - 1;
			numSlotBytes = lastSlotByte - firstSlotByte + 1;

			/*
			 *	we must start at byte 0, the left edge of the next slot
			 */
			start = firstSlotByte;
			relStartByte = 0;
			dangerPtr = &rightDanger;
			TRPRINT(TR_LG, TR_LEVEL_2, ("setting danger to right"));
		}
    }

	TRPRINT(TR_LG, TR_LEVEL_2, ("processing node:%d", shadowPid->page));

	if (GET_LOGLEVEL(LOG_SPACE)) {
		opInfo.action = LOG_ACTION_WRITE_LARGE_SLOTS;
		opInfo.flags = 0;
		opInfo.start = firstSlot*(int)sizeof(LGSLOT);
		opInfo.size = (lastSlot-firstSlot+1) * (int)sizeof(LGSLOT);
		opInfo.oldData = (char*) &(tempLgNode.slot[firstSlot]);
		opInfo.newData = (char*) &(nodePtr->slot[firstSlot]);
		opInfo.actionPid = *shadowPid;
		/* mark if it is a root object on a slotted page */
		opInfo.rootSlot = nodeSlot;
		if(clog_LogWriteLgPage(&opInfo)) {
			FAIL
		}

		opInfo.action = LOG_ACTION_DECREMENT_SLOT_BYTECOUNTS;
		opInfo.flags = 0;
		opInfo.start = lastSlot + 1;
		opInfo.size = deleteSize;
		opInfo.oldData = NULL;
		opInfo.newData = NULL;
		opInfo.actionPid = *shadowPid;
		opInfo.rootSlot = nodeSlot;
		if(clog_LogChangeSlots(&opInfo)) {
			FAIL
		}
	}

    /*
     * Account for deleted bytes in slots following the affected area.
     */
	for (slotIndex = lastSlot + 1; slotIndex < totalSlots; slotIndex++) {

		TRPRINT(TR_LG, TR_LEVEL_2, ("decrementing:%d from slot:%d", deleteSize, slotIndex));
        nodePtr->slot[slotIndex].count -= deleteSize;
    }

	/*
	 * First, note path slot for use in remembering cut path.
	 */
	pathSlot = firstSlot;
	TRPRINT(TR_LG, TR_LEVEL_2, ("pathSlot:%d", pathSlot));

    /*
     * If numMiddleDel > 0, then there was a single (partially) affected
     * slot.  In that case, there is nothing more to do, slot-wise, and
     * otherwise there is still some cleanup work to be done in the node.
     */
    if (numMiddleDel == 0) {
		/*
		 *	if we did not start at first left byte then first
		 *	node contains data that we must keep so increment
		 *	the slot to begin deletion.
		 */
		if (numRightDel > 0) {
			firstSlot += 1;
			TRPRINT(TR_LG, TR_LEVEL_2, 
			("incrementing first slot to:%d", firstSlot));
		}

		/*
		 *	if there are bytes left on the right the lastSlot
		 *	is the first non-empty slot, else all bytes on
		 *	the right tail of page was deleted.  BytesOnRight must
		 *	remain set as this condition terminates the loop.
		 */
		if (bytesOnRight) {
			nextSlot = lastSlot;
			TRPRINT(TR_LG, TR_LEVEL_2, ("first non-empty slot:%d", nextSlot));
		} else {
			nextSlot = lastSlot + 1;
			TRPRINT(TR_LG, TR_LEVEL_2, ("first non-empty slot:%d", nextSlot));
		}

		numDelSlots = nextSlot - firstSlot;
		TRPRINT(TR_LG, TR_LEVEL_2, ("firstSlot:%d lastSlot:%d nextSlot:%d numDelSlots:%d",
				firstSlot, lastSlot, nextSlot, numDelSlots));

		/*
		 *	see if we have eliminated any slots
		 */
		if (numDelSlots > 0) {
			if (GET_LOGLEVEL(LOG_SPACE)) {
				opInfo.action = LOG_ACTION_DELETE_LARGE_SLOTS;
				opInfo.flags = 0;
				opInfo.start = firstSlot*(int)sizeof(LGSLOT);
				opInfo.size = numDelSlots * (int)sizeof(LGSLOT);
				opInfo.oldData = (char *) &(nodePtr->slot[firstSlot]);
				opInfo.newData = NULL;
				opInfo.actionPid = *shadowPid;
				opInfo.rootSlot = nodeSlot;
				if(clog_LogModifyLgPage(&opInfo)) {
					FAIL
				}
			}

			/*
			 *	Copy slots from the right to the left to
			 *	eliminate the deleted slots only if we are not
			 *	deleting the right most slot, in which case
			 *	there are no trailing slots
			 */
			if (nextSlot < nodePtr->header.numSlots)	{
				bcopy((char *) &(nodePtr->slot[nextSlot]),
					  (char *) &(nodePtr->slot[firstSlot]),
					  ((totalSlots - nextSlot) * (int)sizeof(LGSLOT)));
				TRPRINT(TR_LG, TR_LEVEL_2, ("copying slots from:%d to:%d",
						nextSlot, firstSlot));
			}
			/*
			 *	decrement the number of slots in the node
			 */
			nodePtr->header.numSlots -= numDelSlots;
			TRPRINT(TR_LG, TR_LEVEL_2, ("numSlots:%d", nodePtr->header.numSlots));
		}
    }

    /*
     *	Now figure out whether or not the current node is in danger.
     *	It is, if its current slot count is so small that losing all of
     *	its in-danger children would cause it to go below the minimum
     *	allowed number of slots for a non-leaf node.  Lastly, shadow
     *	the node, fix the path slot number in case the whole first slot
     *	went away, and then push the node on the appropriate stack.
     */
    finalSlotCnt = (int)nodePtr->header.numSlots;
    dangerCnt	 =  leftDanger + rightDanger;
    minSlotCnt	 = finalSlotCnt - dangerCnt;
	TRPRINT(TR_LG, TR_LEVEL_2, ("finalSlotCnt:%d dangerCnt:%d minSlotCnt:%d",
			finalSlotCnt, dangerCnt, minSlotCnt));

	/*
	 *	the danger level depends on whether the node is root
	 *	or an interior node
	 */
    if (PIDEQ(*rootPid, *nodePid)) {
		/*
		 *	node is root so we must check to see if this
		 *	is a single or multilevel index.  If it is a
		 *	single level index then we cannot have underflow.
		 */
        *danger = (lastLevel) ? FALSE : (minSlotCnt < 2);
		TRPRINT(TR_LG, TR_LEVEL_2, ("root level danger:%d", *danger));
    } else {
		/* node in an interior node */
        *danger = (minSlotCnt < LGMINSLOTS);
		TRPRINT(TR_LG, TR_LEVEL_2, ("interior danger:%d", *danger));
    }

	/*
	 *	Fix the path slot in case the whole slot went away
	 */
    if (pathSlot >= finalSlotCnt) {
		pathSlot = finalSlotCnt - 1;
		TRPRINT(TR_LG, TR_LEVEL_2, ("fixing pathSlot:%d", pathSlot));
    }

	/*
	 *	Check to see which cut path we are on
	 */
    if (leftCutPath) {
		TRPRINT(TR_LG, TR_LEVEL_2, ("leftCutPath"));
        if (LG_PushTstack(lstack, pathSlot, shadowPid, nodePtr,
						  *danger, dangerCnt, nodeLink, nodeSlot, 
						  FALSE))	{
			FAIL
		}

    } else {
		TRPRINT(TR_LG, TR_LEVEL_2, ("rightCutPath"));
        if (LG_PushTstack(rstack, pathSlot, shadowPid, nodePtr,
						  *danger, dangerCnt, nodeLink, nodeSlot, 
						  FALSE))	{
			FAIL
		}
    }

	bf_DirtyPage(nodeLink->pageHash);

Success:
	return esmNOERROR;

FailureReturn:
	return esmFAILURE;
}
