/*
 *   $RCSfile: LG_DestroyVersionPage.c,v $  
 *   $Revision: 1.1.1.1 $  
 *   $Date: 1996/05/04 21:55:24 $      
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/


#include <stdio.h>
#include "ess.h"
#include "trace.h"
#include "checking.h"
#include "io.h"
#include "page.h"
#include "list.h"
#include "object.h"
#include "tid.h"
#include "bf.h"
#include "chunk.h"
#include "lock.h"
#include "bf_macro.h"
#include "sm_macro.h"
#include "lgobject.h"
#include "volume.h"
#include "pool.h"
#include "version_graph.h"
#include "io_extfuncs.h"
#include "bf_extfuncs.h"
#include "lg_versionfuncs.h"
#include "error.h"


/*
 * LG_DestroyVersionPage searches for non-share pages and destroys
 * them.
 */


 int
LG_DestroyVersionPage (

	BUFGROUP	*bufGroup,
	LIST		destroyLevel[],
	LIST		adjLevel[],
	LGNODEHASHTABLE *nodeHashTable,
	POOL		*nodeListPool,	/* pool of node list elements		*/
	int			versionHeight,	/* height of version to destroy		*/
	int			height,			/* current height being destoryed 	*/
	PID			*rootPid,		/* root pid of the version			*/
	SLOTINDEX	rootSlot 		/* root slot of the version			*/
)
{
	LGNODELIST	*node;
	LGNODELIST	*matchNode;

	TRPRINT(TR_LG, TR_LEVEL_1, ("level:%d", height));

	/*
	 *	See if we've reached the height of the root node of the version
	 */
	if (height > versionHeight) {
		
		/*
		 *	Make sure we haven't reached an illegal height
		 */
		SM_ASSERT(LEVEL_3, (height > 1));

		/*
		 *	We haven't dropped to the height of the root node yet,
		 *	so we take all the nodes at this level and pass their
		 *	children to the next.  At the same time, we can free
		 *	the list elements at this level.
		 */
		while ( LIST_NOT_EMPTY(&adjLevel[height]) ) {

			/*
			 *	Get the next page, and do some error checking
			 */
			node = (LGNODELIST*) listDeq(&adjLevel[height]);
			SM_ASSERT(LEVEL_3, node->slot == NIL);

			/*
			 *	Read in the page, add each child to the next
			 *	level, and put them in the hash table if the
			 *	next level is the root of the version to be
			 *	destroyed.
			 */
			if (LG_HashChildren(bufGroup, &node->pid, node->slot, 
								&adjLevel[height-1], 
								(height-1 == versionHeight) ? 
								nodeHashTable : NULL, nodeListPool,
								FALSE )) {
				
				return(esmFAILURE);
			}

			/*
			 *	Return the list element to the pool
			 */
			poolEnq(nodeListPool, &node->list);	
		}
	
		if (LG_DestroyVersionPage(bufGroup, destroyLevel, adjLevel, 
									nodeHashTable, nodeListPool, 
									versionHeight, height-1, rootPid,
									rootSlot) ) {
			return (esmFAILURE);
		}

	} else {

		/*
		 *	For each page at the current height in the version to
		 *	destroy, remove it if it's not in adjLevel at that
		 *	height.  It can also be removed from the hash table and
		 *	from adjLevel.
		 */
		for (node = (LGNODELIST*) FIRST_LIST_ELEMENT(&destroyLevel[height]);
			 node != NULL;
			 node = (LGNODELIST*) NEXT_LIST_ELEMENT(&node->list) ) {
			
			if ((matchNode = LG_HashFindNode(&node->pid, node->slot, 
											 nodeHashTable)) != NULL) {
				/*
				 *	Remove the list node from the hash table, and the 
				 *	adjacent list, and return it to the pool.
				 */
				listRemove(&matchNode->hash);
				listRemove(&matchNode->list);
				poolEnq(nodeListPool, &matchNode->list);

				/*
				 *	Remove it from the destroy level as well
				 */
				listRemove(&node->list);
				poolEnq(nodeListPool, &node->list);

			} 
		}

		/*
		 *	See if we're still above the bottom level
		 */
		 if (height >= 1) {

			/*
			 *	For all the pages left at this height in the version to
			 *	be destroyed, add their children to 
			 *	the next level.  The pages can then be destroyed.
			 */
			while ( LIST_NOT_EMPTY(&destroyLevel[height]) ) {

				/*
				 *	Get the next page, and do some error checking
				 */
				node = (LGNODELIST*) listDeq(&destroyLevel[height]);

				/*
				 *	Read in the page, and add each child to the next
				 *	level.  The last parameter to LG_HashChildren
				 *	indicates the page should be destroyed if it is
				 *	not a slotted page.
				 */
				if (LG_HashChildren(bufGroup, &node->pid, node->slot, 
									&destroyLevel[height-1], NULL, nodeListPool,
									node->slot == NIL)) {
					
					return(esmFAILURE);
				}

				/*
				 *	Return the list element to the pool
				 */
				poolEnq(nodeListPool, &node->list);	
			}

			/*
			 *	For all the pages left at this height in adjacent 
			 *	versions, add their children to 
			 *	the next level, and put them in the hash table.
			 */
			while ( LIST_NOT_EMPTY(&adjLevel[height]) ) {

				/*
				 *	Get the next page, and do some error checking
				 */
				node = (LGNODELIST*) listDeq(&adjLevel[height]);
				SM_ASSERT(LEVEL_3, node->slot == NIL);

				/*
				 *	Read in the page, add each child to the next
				 *	level, and put them in the hash table if the
				 *	next level is the root of the version to be
				 *	destroyed.
				 */
				if (LG_HashChildren(bufGroup, &node->pid, node->slot, 
									&adjLevel[height-1], nodeHashTable, 
									nodeListPool, FALSE )) {
					
					return(esmFAILURE);
				}

				/*
				 *	Remove the list element from the hash table, and
				 *	return it to the pool
				 */
				listRemove(&node->hash);
				poolEnq(nodeListPool, &node->list);	
			}

			if (LG_DestroyVersionPage(bufGroup, destroyLevel, adjLevel,
										nodeHashTable, nodeListPool, 
										versionHeight, height-1, 
										rootPid, rootSlot) ) {
				return (esmFAILURE);
			}

		 } else {

			/*
			 *	We're at the bottom level, so destroy each page 
			 *	left in destroyLevel, and free up all LGNODELIST 
			 *	structures.  Note that the pages are data pages.
			 */

			/*
			 *	Destroy all the pages left at this height in the 
			 *	version to be destroyed, then place their
			 *	LGNODELIST structure back in the pool.
			 */
			while ( LIST_NOT_EMPTY(&destroyLevel[height]) ) {

				/*
				 *	Get the next page, and do some error checking
				 */
				node = (LGNODELIST*) listDeq(&destroyLevel[height]);
				SM_ASSERT(LEVEL_3, node->slot == NIL);

                /*
                 *  destroy any copies in the cache
                 */
                bf_HackChunkPage(&node->pid);

                /*
                 *  Deallocate the page
                 */
                if (io_DeallocPages(LGDATA_PAGE2SIZE, 1, &node->pid, FALSE, PAGE_LARGEDATA))   {

                    return(esmFAILURE);
                }
                TRPRINT(TR_LG, TR_LEVEL_2, 
						("hacking and deallocating page:%d", 
						node->pid.page));


				/*
				 *	Return the list element to the pool
				 */
				poolEnq(nodeListPool, &node->list);	
			}


			/*
			 *	For all the pages left at this height in adjacent 
			 *	versions, put their LGNODELIST structure back in 
			 *	the pool, and remove it from the hash table.
			 */
			while ( LIST_NOT_EMPTY(&adjLevel[height]) ) {

				/*
				 *	Get the next page
				 */
				node = (LGNODELIST*) listDeq(&adjLevel[height]);

				/*
				 *	Remove the list element from the hash table, and
				 *	return it to the pool
				 */
				listRemove(&node->hash);
				poolEnq(nodeListPool, &node->list);	
			}

		 }
	}

	/*
	 *	return success
	 */
	return(esmNOERROR);

}
