/*
 *   $RCSfile: LG_FixIndex.c,v $  
 *   $Revision: 1.1.1.1 $  
 *   $Date: 1996/05/04 21:55:23 $      
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/


#include "ess.h"
#include "sysdefs.h"
#include "trace.h"
#include "checking.h"
#include "error.h"
#include "io.h"
#include "page.h"
#include "list.h"
#include "object.h"
#include "tid.h"
#include "bf.h"
#include "chunk.h"
#include "lock.h"
#include "lsn.h"
#include "clog.h"
#include "logaction.h"
#include "loginfo.h"
#include "logrecs.h"
#include "bf_macro.h"
#include "lgobject.h"
#include "volume.h"
#include "sm_state.h"
#include "sm_macro.h"
#include "io_extfuncs.h"
#include "bf_extfuncs.h"
#include "lg_intfuncs.h"
#include "lg_extfuncs.h"
#include "util_funcs.h"
#include "clog_globals.h"
#include "clog_extfuncs.h"



/*
 * Update the index of the long object following an append or insert
 * operation.
 */


 int 
LG_FixIndex (
	PID				*rootPid,		/* pid of root page (for dependencies)	*/
	PAGELIST		*pageList,		/* freed by caller */
	TSTACK			*tstack,		/* stack that was used to traverse object*/
	BUFGROUP		*bufGroup,		/* buffer group to use for the operation*/
	int				replaceCount 	/* number of slots to overwrite			*/
)
{
	TSTACKENT		*stackTop;
    int				totalSlots;		/* slots in the node after insertion	*/
    int				height;			/* tree height we're currently at		*/
    int				nodeCount;		/* number of nodes after split			*/
    PID				origPid;		/* pid of buffered internal node		*/
	GROUPLINK		*origLink;		/* link of the current node				*/
    LGNODE			*origNode;		/* ptr to copy of original node contents*/
    LGNODE			*currentNode;	/* indirect ptr to buffered internal node*/
    PAGELIST		newPageList;	/* list of pid/cnt pairs to insert		*/
	int				insertSlot;
	int				remainSlots;
	int				currentSlot;
	int				currentBytes;
	CHUNKPAGE		*currentPage;
	CHUNKPAGE		*downPage;
	int				origSlotCount;
	int				slotCount;
	int				origInsertSlot;
    int				remainder;		/* slots left, for rebalancing purposes	*/
    int				slotsPerNode;	/* balanced number of slots per node	*/
    BOOL			splitRoot;		/* true if we had to split the root page*/
	BOOL			slotRoot;
	VHGNODEPTR		vhgNodePtr;		/* version history graph node for
										this obj if it's versioned 	*/
    PAGESLOT		*pageSlotPtr;	/* pointer to a slot on a slotted page */
	OPERATIONINFO	opInfo;		/* structure for passing info to log calls*/


	TRPRINT(TR_LG, TR_LEVEL_1, ("replaceCount:%d", replaceCount));

    /*
     *	Insert the cnt/pid pairs on the page list into the top node on the
     *	traversal stack.  Note that the first and perhaps the second
	 *	page in the list will REPLACE the pid link of the top of the
	 *	stack.  We are thus adding pageList->numPages - replaceCount
	 *	entries to the parent node.  When
	 *	we get to the point where we haven't added any leaves to a level
	 *	then we will have exactly 1 page to process for the next level.
	 *	This page will REPLACE the slot entry of the parent without
	 *	any possibility of splitting that node.
     */
    height = 0;
	splitRoot = FALSE;
	slotRoot = FALSE;

	/*
	 *	Get the version history graph node for the object.
	 *	If the object is not versioned, this is meaningless.
	 */
	vhgNodePtr = tstack->stack[0].nodePtr->header.vhgPtr;


	/*
	 *	Process while we have new pages to insert (not just replace)
	 *	This MAY cause a split to occur.
	 */
    while (pageList->numPages > replaceCount) {

		TRPRINT(TR_LG, TR_LEVEL_2, ("new page count:%d", pageList->numPages));

		/*
		 *	increment the height from the leaves.  We are
		 *	working up the tree from the bottom.
		 */
        height++;

		/*
		 *	Pop top node off of stack for processing.  If the
		 *	stack runs dry then we have split the root on
		 *	some previous iteration and we are adding a new level.
		 */
		if ((stackTop = LG_PopTstack(tstack)) == NULL)	{

			TRPRINT(TR_LG, TR_LEVEL_2, ("stack has run dry"));

			/*
			 *	record that the root has split
			 */
			splitRoot = TRUE;

			/*
			 *	The number of slots are just the number
			 *	being propogated up from below.  The insertion
			 *	point is the initial slot
			 */
			totalSlots = pageList->numPages;
			origInsertSlot = insertSlot = 0;

			/*
			 *	setup the original pid to use as the near 
			 *	hint for the page allocate call
			 */
			origPid = pageList->pages->pid;

		} else {

			/*
			 *	get out the appropriate stack entries
			 */
			origInsertSlot = insertSlot = stackTop->slotIndex;
			origPid		   = stackTop->pid;
			origNode	   = stackTop->nodePtr;
			origLink	   = stackTop->nodeLink;

			/*
			 *	Calculate the total number of slots that will be manipulated.
			 *	Remember that replaceCount slots will be replaced, not added.
			 */
			totalSlots = (int)origNode->header.numSlots + (pageList->numPages - replaceCount);
			TRPRINT(TR_LG, TR_LEVEL_2, ("orig slots:%d total slots:%d",
					origNode->header.numSlots, totalSlots));

		}
		TRPRINT(TR_LG, TR_LEVEL_2, ("inserting at slot:%d", insertSlot));

		/*
		 *	calculate the number of index nodes that will be required
		 *	to hold the number of slots in question
		 */
		nodeCount = (totalSlots / LGMAXSLOTS) + ((totalSlots % LGMAXSLOTS) > 0);

		/*
		 *	calculate the number of slots each node will have to hold
		 */
		slotsPerNode = totalSlots / nodeCount;

		/*
		 *	we may have some remaining slots if the number of slots
		 *	cannot be spread evenly across the node pages
		 */
		remainder = totalSlots % nodeCount;
		TRPRINT(TR_LG, TR_LEVEL_2, ("nodeCount:%d slotsPerNode:%d remainder:%d replaceCount:%d",
				nodeCount, slotsPerNode, remainder, replaceCount));

		/*
		 *	setup initial counts of slots to be inserted
		 */
		origSlotCount = pageList->numPages;
		TRPRINT(TR_LG, TR_LEVEL_2, ("origSlotCount:%d", origSlotCount));

		/*
		 *	initialize a list to hold the new index nodes that
		 *	will be created.  These cannot be more than were
		 *	created in the last level.
		 */
		newPageList.pages = NULL;
		if(LG_InitList(&newPageList, nodeCount) == esmFAILURE) {
			FAIL
		}
		newPageList.numPages = nodeCount;

		/*
		 *	Allocate enough nodes to accomodate all the possible entries.
		 *	These nodes will replace (shadow) the node that was popped
		 *	off the top of the stack.
		 */
		if (LG_AllocNodes(bufGroup, newPageList.pages, &origPid, nodeCount)){
			FAIL
		}
		sm_SetAbortIfError();

		{
			int i;
			register CHUNKPAGE	*chunkPage;

			/*
			 *	get the pointer to the first chunk entry
			 */
			chunkPage = newPageList.pages;

			/*
			 *	mark counts and heights for all the new nodes
			 */
			for (i = 0; i < nodeCount; i++, chunkPage++)	{

				/*
				 *	get a pointer to the first new node
				 */
				currentNode = (LGNODE *) chunkPage->nodeLink->bufFrame;

				/*
				 *	Initialize the new node fields
				 */
				INIT_LARGENODE_MAGIC(currentNode);
				currentNode->header.lrc = *(GET_LOGINFO->initLRC);
				currentNode->header.height = (ONE) height;
				currentNode->header.vhgPtr = vhgNodePtr;

				/* must be 0 for logging */
				currentNode->header.numSlots = 0; 

				TRPRINT(TR_LG, TR_LEVEL_2, ("height:%d numSlots:%d", height, currentNode->header.numSlots));

				if (GET_LOGLEVEL(LOG_SPACE)) {
					/* log the init */
					opInfo.action = LOG_ACTION_INIT_LARGE_NODE;
					opInfo.flags = 0;
					opInfo.start = 0;
					opInfo.size = sizeof(LGNODEHDR);
					opInfo.oldData = NULL;
					opInfo.newData = (char*) &(currentNode->header);
					opInfo.actionPid = chunkPage->pid;
					opInfo.rootSlot = NIL;
					if(clog_LogWriteLgNodeHdr(&opInfo)) {
						FAIL
					}
				}

				/* now set the number of slots */
				currentNode->header.numSlots = slotsPerNode + (remainder-- > 0);
			}
		}

		/*
		 *	setup a pointer to the first entry in the page list
		 */
		currentPage = newPageList.pages;

		/*
		 *	Get a pointer to the first node
		 */
		currentNode = (LGNODE *) currentPage->nodeLink->bufFrame;

		/*
		 *	if we have not split the root then we can copy
		 *	to the insertion point
		 */
		if (splitRoot)	{

			/*
			 *	there are no remaining slots after insertion point
			 */
			remainSlots = 0;

		} else {
			
			/*
			 *	Calculate the number of slots remaining to be copied
			 *	from the original node
			 */
			remainSlots = ((int)origNode->header.numSlots - replaceCount) - insertSlot;
			TRPRINT(TR_LG, TR_LEVEL_2, ("remainSlots:%d", remainSlots));

			/*
			 *	decide if the data up to the insertion point will
			 *	be split across two nodes.
			 */
			if (currentNode->header.numSlots < insertSlot)	{

				int		firstCount;
				int		i;


				TRPRINT(TR_LG, TR_LEVEL_2, ("split slots"));

				/*
				 *	save the number of slots in the first node
				 */
				firstCount = (int)currentNode->header.numSlots;

				/* copy data onto the first node */
				bcopy((char *) origNode->slot,
					   (char *) currentNode->slot,
					   (int) (sizeof(LGSLOT) * firstCount));

				/* set the size for current page */
				currentPage->size = 
					(int)currentNode->slot[firstCount - 1].count;

				if (GET_LOGLEVEL(LOG_SPACE)) {
					/* log the new slots*/
					opInfo.action = LOG_ACTION_APPEND_LARGE_SLOTS;
					opInfo.flags = 0;
					opInfo.start = 0;
					opInfo.size = (int) (sizeof(LGSLOT) * firstCount);
					opInfo.oldData = NULL;
					opInfo.newData = (char *) currentNode->slot;
					opInfo.actionPid = currentPage->pid;
					opInfo.rootSlot = NIL;
					if(clog_LogModifyLgPage(&opInfo)) {
						FAIL
					}
				}

				/* Unfix the node */
				bf_UnfixPage(currentPage->nodeLink, BF_DEFAULT, TRUE);

				/* Get a pointer to the nextnode */
				currentNode = (LGNODE *) (++currentPage)->nodeLink->bufFrame;

				/* set a new value of insert slot */
				insertSlot -= firstCount;
				TRPRINT(TR_LG, TR_LEVEL_2, ("insertSlot:%d firstSlot:%d",
						insertSlot, firstCount));

				/* copy the data into the second node */
				bcopy((char *) &(origNode->slot[firstCount]),
					  (char *) currentNode->slot,
					  (int) (sizeof(LGSLOT) * insertSlot));

				/*
				 *	We must update the counts in the 2nd node to match
				 *	reflect the new offset of the start of the node
				 */
				firstCount = (int)origNode->slot[firstCount - 1].count;

				for (i = 0; i < insertSlot; i++)	{

					currentNode->slot[i].count -= firstCount;
					TRPRINT(TR_LG, TR_LEVEL_2, ("slot %d count:%d",
							i, currentNode->slot[i].count));
				}

			} else {

				/*
				 *	Copy in the slots up to the insertion point
				 *	from the original node.  The byte counts and
				 *	pids will be unchanged.
				 */
				bcopy((char *) origNode->slot,
					  (char *) currentNode->slot,
					  (int) (sizeof(LGSLOT) * insertSlot));
				TRPRINT(TR_LG, TR_LEVEL_2, ("copying %d starting slots", insertSlot));
			}
		}

		/* check to make sure remainSlots is non-negative */
		SM_ASSERT(LEVEL_3, remainSlots >= 0);

		/*
		 *	Must copy in the slots that occur AFTER the insertion point
		 *	that will not be overwritten by new entries
		 */
		if (remainSlots > 0)	{

			/* process remaining slots */
			LG_AppendIndex(slotsPerNode, remainSlots, replaceCount, nodeCount,
						   &newPageList, origNode, origInsertSlot);
		}

		/*
		 *	set the current number of bytes
		 */
		if (insertSlot == 0)	{

			currentBytes = 0;

		} else {

			/*
			 *	Get the bytes from the previous slot.  The slots
			 *	up to insertion point were copied into the new
			 *	node so we can get the count from there.
			 */
			currentBytes = (int)currentNode->slot[insertSlot - 1].count;
		}
		TRPRINT(TR_LG, TR_LEVEL_2, ("currentBytes:%d", currentBytes));

		/*
		 *	get a pointer to the first entry in the previous
		 *	pageList.  This will allow us to get pids and counts out.
		 */
		downPage = pageList->pages;

		/*
		 *	Process all of the NEW slots.  Slots that were copied
		 *	from the original page are not touched by this loop
		 */
		currentSlot = insertSlot;
		slotCount = 0;

		for (;;)	{

			/*
			 *	check to see if we have to process a new node
			 */
			if (currentSlot == currentNode->header.numSlots)	{

				TRPRINT(TR_LG, TR_LEVEL_2, ("done processing node"));
				lg_PrintNode(currentNode, 0, currentPage->pid.page);

				/*
				 *	save the count of the node for the next level
				 *	this propogates up the largest count rooted
				 *	at this subtree
				 */
				currentPage->size = currentBytes;
				TRPRINT(TR_LG, TR_LEVEL_2, ("final node size:%d", currentBytes));
				lg_PrintNode(currentNode, 0, currentPage->pid.page);

				if (GET_LOGLEVEL(LOG_SPACE)) {
					/* log the new slots*/
					opInfo.action = LOG_ACTION_APPEND_LARGE_SLOTS;
					opInfo.flags = 0;
					opInfo.start = 0;
					opInfo.size =(int)(sizeof(LGSLOT) * currentSlot);
					opInfo.oldData = NULL;
					opInfo.newData = (char *) currentNode->slot;
					opInfo.actionPid = currentPage->pid;
					opInfo.rootSlot = NIL;
					if(clog_LogModifyLgPage(&opInfo)) {
						FAIL
					}
				}

				/*
				 *	unfix the node in the buffer pool
				 */
				bf_UnfixPage(currentPage->nodeLink, BF_DEFAULT, TRUE);

				/*
				 *	reset the current byte counts and 
				 *	the slot offset to zero
				 */
				currentBytes = 0;
				currentSlot = 0;

				/*
				 *	Check to see if we have processed all of the original slots
				 */
				if (slotCount == origSlotCount)	{

					/*
					 *	we are done processing
					 */
					break;
				}

				/*
				 *	increment page pointer to next page
				 */
				currentPage++;

				/*
				 *	get a pointer to the page
				 */
				currentNode = (LGNODE *) currentPage->nodeLink->bufFrame;
			}

			/*
			 *	Put the byte count and pid in the slot
			 */
		    currentNode->slot[currentSlot].count = 
												currentBytes += downPage->size;
			currentNode->slot[currentSlot].pid = downPage->pid;
			currentNode->slot[currentSlot].shared = FALSE;
			TRPRINT(TR_LG, TR_LEVEL_2, ("slot:%d pid:%d count:%d", currentSlot,
					downPage->pid.page, currentBytes));

			/*
			 *	process the next slot which is equivalent to processing
			 *	the next entry in the previous level page list
			 */
			slotCount++;
			currentSlot++;
			downPage++;

			/*
			 *	check to see if we are insert mode and finished
			 */
			if (remainSlots > 0)	{

				/*
				 *	Check to see if we have processed all of the original slots
				 */
				if (slotCount == origSlotCount)	{
					LGSLOT		*slotPtr;
					int		i;

					/*
					 *	we must set the counts to final form
					 */
					for (i = currentSlot,
							slotPtr = &(currentNode->slot[currentSlot]);
						 i < currentNode->header.numSlots;
						 i++, slotPtr++)	{

						/*
						 *	total from last slot
						 */
						slotPtr->count += (slotPtr - 1)->count;
						TRPRINT(TR_LG, TR_LEVEL_2, ("slot:%d count:%d", i, slotPtr->count));
					}

					/*
					 *	save the largest subtree count
					 */
					currentPage->size = (int)((slotPtr - 1)->count);
					TRPRINT(TR_LG, TR_LEVEL_2, ("final node size:%d", currentPage->size));
					lg_PrintNode(currentNode, 0, currentPage->pid.page);

					if (GET_LOGLEVEL(LOG_SPACE)) {
						/* log the new slots*/
						opInfo.action = LOG_ACTION_APPEND_LARGE_SLOTS;
						opInfo.flags = 0;
						opInfo.start = 0;
						opInfo.size = 
							(int) (sizeof(LGSLOT) * currentNode->header.numSlots);
						opInfo.oldData = NULL;
						opInfo.newData = (char *) currentNode->slot;
						opInfo.actionPid = currentPage->pid;
						opInfo.rootSlot = NIL;
						if(clog_LogModifyLgPage(&opInfo)) {
							FAIL
						}
					}

					/* unfix the node in the buffer pool */
					bf_UnfixPage(currentPage->nodeLink, BF_DEFAULT, TRUE);
					break;
				}
			}
        }

		/*
		 *	check to see if we had a valid node popped off the stack
		 */
		if (!splitRoot)	{

			/*
			 *	check to see if this is a slotted root page
			 */
			if (stackTop->rootSlot != NIL)	{

				TRPRINT(TR_LG, TR_LEVEL_2, ("slot root"));

				/*
				 *	check to see if there is only a single node
				 */
				if (nodeCount == 1)	{

					register GROUPLINK	*currentLink;

					/* fix the page */
					if ((currentLink = bf_ReadPage(bufGroup, 
					     &(newPageList.pages->pid),
						 H_LARGE, LG_PAGE2SIZE, NL)) == NULL)	{

						SM_ERROR(TYPE_FATAL, esmINTERNAL);
					}

					/* get a pointer to the node */
					currentNode = (LGNODE *) currentLink->bufFrame;
					SM_ASSERT(LEVEL_3, currentLink->fixedCount > 0);

					/*
					 *	check to see if the number of
					 *	slots will fit on the root page
					 */
					if (currentNode->header.numSlots > LGMAXPAGESLOTS)	{

						TRPRINT(TR_LG, TR_LEVEL_2, ("using new node"));

						/* If the root object isn't already marked as being
						   a lg object header, mark it as such. */	
						if (stackTop->nodePtr->header.numSlots != NIL) {

						    /* save the old header for logging */
						    bcopy((char*) &(stackTop->nodePtr->header),
						 		OldData, (int) (sizeof(LGNODEHDR)));

						    /*
						 	 *	Mark the header as moved
						 	 */
						     stackTop->nodePtr->header.numSlots = NIL;

							/* log the change to the slotted root object */
							pageSlotPtr = (PAGESLOT*)
							  GETSLOTPTR(((SLOTTEDPAGE*) origLink->bufFrame),
												stackTop->rootSlot);

							if (GET_LOGLEVEL(LOG_SPACE)) {
								opInfo.action = LOG_ACTION_WRITE_SMALL_OBJ;
								opInfo.flags = 0;
								opInfo.oid.diskAddr.page = origPid.page;
								opInfo.oid.diskAddr.slot = (TWO) stackTop->rootSlot;
								opInfo.oid.diskAddr.volid = origPid.volid;
								opInfo.oid.diskAddr.unique = pageSlotPtr->unique;
								opInfo.start = 0;
								opInfo.size = sizeof(LGNODEHDR);
								opInfo.oldData = OldData ;
								opInfo.newData = 
											(char *) &(stackTop->nodePtr->header);
								opInfo.actionPid = origPid;
								if(clog_LogWriteObj(&opInfo)) {
									FAIL
								}
							}

							/* unfix the page */
							bf_UnfixPage(currentLink, BF_DEFAULT, FALSE);
						}

					} else {

					   TRPRINT(TR_LG, TR_LEVEL_2, ("copying node contents in"));

			            /*note: no need to log the removal of the slots of the
						  shadow node*/

						/*
						 *	change the magic number so we can copy from current
						 *  back into the slotted root object.
						 */
						INIT_SLOTROOT_MAGIC(currentNode);

						pageSlotPtr = (PAGESLOT*) 
							GETSLOTPTR(((SLOTTEDPAGE*) origLink->bufFrame),
											stackTop->rootSlot);

						if (GET_LOGLEVEL(LOG_SPACE)) {
							/* log the update to the slotted root object */
							opInfo.action = LOG_ACTION_WRITE_SMALL_OBJ;
							opInfo.flags = 0;
							opInfo.oid.diskAddr.page = origPid.page;
							opInfo.oid.diskAddr.slot = (TWO) stackTop->rootSlot;
							opInfo.oid.diskAddr.volid = origPid.volid;
							opInfo.oid.diskAddr.unique = pageSlotPtr->unique;
							opInfo.start = 0;
							opInfo.size = sizeof(LGSLOTROOT);
							opInfo.oldData = (char *) (stackTop->nodePtr);
							opInfo.newData = (char *) currentNode;
							opInfo.actionPid = origPid;
							if(clog_LogWriteObj(&opInfo)) {
								FAIL
							}
						}

						/* copy in the node's contents */
						bcopy((char *) currentNode,
							  (char *) stackTop->nodePtr,
							  sizeof(LGSLOTROOT));

						/* Deallocate the new page */
						if (io_DeallocPages(LG_PAGE2SIZE, 1, 
								&newPageList.pages->pid, TRUE, PAGE_LARGENODE)){
							return(esmFAILURE);
						}
						TRPRINT(TR_LG, TR_LEVEL_2, ("deallocated:%d",
								newPageList.pages->pid.page));

						/* discard the page */
						bf_DiscardPage(bufGroup, newPageList.pages->nodeLink);

						/* note that we have slotted root */
						slotRoot = TRUE;

						/* print the current node */
						lg_PrintNode(stackTop->nodePtr, 0, stackTop->pid.page);
					}

				} else {

						/* save the old header for logging */
						bcopy((char*) &(stackTop->nodePtr->header),
						 		OldData, (int) (sizeof(LGNODEHDR)));

						/* Mark the header as moved */
						stackTop->nodePtr->header.numSlots = NIL;

						/* log the change to the slotted root object */
						pageSlotPtr = (PAGESLOT*)
							  GETSLOTPTR(((SLOTTEDPAGE*) origLink->bufFrame),
											stackTop->rootSlot);

						if (GET_LOGLEVEL(LOG_SPACE)) {
							opInfo.action = LOG_ACTION_WRITE_SMALL_OBJ;
							opInfo.flags = 0;
							opInfo.oid.diskAddr.page = origPid.page;
							opInfo.oid.diskAddr.slot = (TWO) stackTop->rootSlot;
							opInfo.oid.diskAddr.volid = origPid.volid;
							opInfo.oid.diskAddr.unique = pageSlotPtr->unique;
							opInfo.start = 0;
							opInfo.size = sizeof(LGNODEHDR);
							opInfo.oldData = OldData ;
							opInfo.newData = (char *) &(stackTop->nodePtr->header);
							opInfo.actionPid = origPid;
							if(clog_LogWriteObj(&opInfo)) {
								FAIL
							}
						}

					/* Mark the header as moved */
					stackTop->nodePtr->header.numSlots = NIL;
				}

				/* Unfix the page */
				bf_UnfixPage(stackTop->nodeLink, BF_DEFAULT, FALSE);

			} else {

#ifdef JUNK
	/* why log something for a page which will be dealloced below?!!! */
				if (GET_LOGLEVEL(LOG_SPACE)) {
					/*log the removal of the slots of the original node*/
					opInfo.action = LOG_ACTION_REMOVE_LARGE_SLOTS;
					opInfo.flags = 0;
					opInfo.start = 0;
					opInfo.size = (int)(sizeof(LGSLOT) * origNode->header.numSlots);
					opInfo.oldData = (char *) (origNode->slot);
					opInfo.newData = NULL;
					opInfo.actionPid = origPid;
					opInfo.rootSlot = NIL;
					if(clog_LogModifyLgPage(&opInfo)) {
						FAIL
					}
				}
#endif /* JUNK */

				/* Deallocate the node on the top of the stack */
				if (io_DeallocPages(LG_PAGE2SIZE, 1, &origPid, 
													FALSE, PAGE_LARGENODE))	{
					return(esmFAILURE);
				}
				TRPRINT(TR_LG, TR_LEVEL_2, ("deallocated page:%d", origPid.page));

				/*
				 *	Unfix the page but don't mark it as dirty
				 *	Should trash it
				 */
				SM_ASSERT(LEVEL_3, origLink->fixedCount > 0);
				bf_DiscardPage(bufGroup, origLink);
			}
		}

        /*
         *	Done with currentNode and the page list.  Move newPageList
		 *	to pgList and iterate.
		 */
		TRPRINT(TR_LG, TR_LEVEL_2, ("pageList:%d", pageList->numPages));
		TRPRINT(TR_LG, TR_LEVEL_2, ("newPageList:%d", newPageList.numPages));
		FREELIST(pageList);
		*pageList = newPageList;

		/*
		 *	set the replace count to one for the remaining
		 *	iterations of the loop.  Under current conditions,
		 *	only the leaf nodes may replace more than one slot.
		 */
		replaceCount = 1;
		TRPRINT(TR_LG, TR_LEVEL_2, ("new node count:%d  new replace count:%d",
				pageList->numPages, replaceCount));
    }

	/*
	 *	Check to see if we have a new root
	 *	We will have a new root it we split the old root or
	 *	we processed the old root and shadowed it
	 */
	if ((splitRoot || LG_EmptyTstack(tstack)) && (!slotRoot))	{

		/* return the new root pid */
		*rootPid = pageList->pages->pid;
		TRPRINT(TR_LG, TR_LEVEL_2, ("new root pid:%d", rootPid->page));
	}

    /*
	 *	We have to replace the slots in this level node.  Note
	 *	that we will NOT shadow up to the root!!
     */
	if (!LG_EmptyTstack(tstack))	{

		/* process the remainder of the nodes in the stack */
		LG_FixLast(pageList, tstack);
	}

Success:
    return esmNOERROR;

FailureReturn:
	FREELIST(&newPageList);
	return esmFAILURE;

}
