/*
 *   $RCSfile: LG_HashChildren.c,v $  
 *   $Revision: 1.1.1.1 $  
 *   $Date: 1996/05/04 21:55:24 $      
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/


#include <stdio.h>
#include "ess.h"
#include "trace.h"
#include "checking.h"
#include "io.h"
#include "page.h"
#include "list.h"
#include "object.h"
#include "tid.h"
#include "bf.h"
#include "chunk.h"
#include "lock.h"
#include "bf_macro.h"
#include "sm_macro.h"
#include "lgobject.h"
#include "volume.h"
#include "pool.h"
#include "version_graph.h"
#include "io_extfuncs.h"
#include "bf_extfuncs.h"
#include "lg_intfuncs.h"
#include "lg_extfuncs.h"
#include "lg_versionfuncs.h"
#include "error.h"


/*
 * LG_HashChildren adds each child of a node to the next level,
 * puts each child in a hash table (if provided), and deletes the
 * node if requested.
 */


 int
LG_HashChildren (

	BUFGROUP	*bufGroup,
	PID			*nodePid,		/* node to process children for		*/
	SLOTINDEX	nodeSlot,
	LIST		*levelList,
	LGNODEHASHTABLE *nodeHashTable,
	POOL		*nodeListPool,	/* pool of node list elements		*/
	BOOL		destroyNode  	/* destroy the node if TRUE			*/
)
{
	LGNODELIST	*nodeList;
    LGNODE		*nodePage;		/* ptr to buffered node			*/
	GROUPLINK	*nodeLink;
    int			i;			

	/*
	 *	check to see if this is a header on slotted pages
	 */
	if (nodeSlot != NIL)	{

		PAGESLOT		*slotPtr;
		SLOTTEDPAGE		*pagePtr;
		LARGEOBJROOT	*objectPtr;

		TRPRINT(TR_LG, TR_LEVEL_1, ("slot node"));

		/*
		 * Read in the slotted page containing the root 
		 */
		if ((nodeLink = bf_ReadPage(bufGroup, nodePid, H_SLOTTED,
									SLOTTED_PAGE2SIZE, NL)) == NULL)	{

			return(esmFAILURE);
		}

		/*
		 *	get a pointer to the page, slot and object
		 */
		pagePtr = (SLOTTEDPAGE *) nodeLink->bufFrame;
		slotPtr = GETSLOTPTR(pagePtr, nodeSlot);
		objectPtr = (LARGEOBJROOT *) GETOBJECTPTR(pagePtr, slotPtr);

		/*
		 *	get a pointer to the page
		 */
		nodePage = (LGNODE *) &(objectPtr->rootPage);

		/*
		 *	check the magic of the object
		 */
		CHECK_SLOTROOT_MAGIC(nodePage);

	} else {

		/*
		 * Read in the root page
		 */
		if ((nodeLink = bf_ReadPage(bufGroup, nodePid, H_LARGE,
									LG_PAGE2SIZE, NL)) == NULL)	{

			return(esmFAILURE);
		}

		/*
		 *	get a pointer to the node
		 */
		nodePage = (LGNODE *) nodeLink->bufFrame;
		TRPRINT(TR_LG, TR_LEVEL_2, ("at level:%d", nodePage->header.height));
		lg_PrintNode(nodePage, 0, nodePid->page);
	}
	

	/*
	 *	For each child of the node, add it to levelList, and put
	 *	it in the hash table if desired.
	 */
	for (i = 0; i < nodePage->header.numSlots; i++) {

		/*
		 *	Get an element for the node list
		 */
		nodeList = (LGNODELIST*) poolDeq(nodeListPool);
		if (nodeList == NULL) {

			SM_ERROR(TYPE_WARNING, esmNOFREELGNODELIST);
			return(esmFAILURE);
		}
		nodeList->pid = nodePage->slot[i].pid;
		nodeList->slot = NIL;

		/*
		 *	Initialize the list element, and put it in the list
		 */
		listEnq(levelList, &nodeList->list);

		/*
		 *	If the hash table is not NULL, put the child in the hash
		 * 	table.
		 */
		if (nodeHashTable != NULL) {
			
			if ( ! LG_HashPutNode(nodeHashTable, nodeList)) {

				/*
				 *	Node was already in the table, so remove it
				 */
				listRemove(&nodeList->list);
				poolEnq(nodeListPool, &nodeList->list);
			}
		}
	}

	/*
	 *	See if the node should be destroyed
	 */
	if (destroyNode) {

			/*
			 *	Node can't be a slotted page
			 */
			SM_ASSERT(LEVEL_3, nodeSlot == NIL);

            /*
             *  blow the root page out of the cache
             */
            bf_DiscardPage(bufGroup, nodeLink);

            /*
             *  Deallocate the root page
             */
            if (io_DeallocPages(LG_PAGE2SIZE, 1, nodePid, FALSE, PAGE_LARGENODE))  {

                return(esmFAILURE);
            }
            TRPRINT(TR_LG, TR_LEVEL_2, 
					("deallocating and discarding index node: %d", 
					nodePid->page));

	} else {

		/*
		 *	simply unfix the root page
		 */
		bf_UnfixPage(nodeLink, BF_DEFAULT, FALSE);
	}

	/*
	 *	return success
	 */
	return(esmNOERROR);

}
