/*
 *   $RCSfile: LG_InsertData.c,v $  
 *   $Revision: 1.1.1.1 $  
 *   $Date: 1996/05/04 21:55:23 $      
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/


#include "ess.h"
#include "sysdefs.h"
#include "trace.h"
#include "checking.h"
#include "error.h"
#include "io.h"
#include "page.h"
#include "list.h"
#include "object.h"
#include "tid.h"
#include "bf.h"
#include "chunk.h"
#include "lsn.h"
#include "clog.h"
#include "logrecs.h"
#include "loginfo.h"
#include "logaction.h"
#include "lock.h"
#include "pool.h"
#include "sm_state.h"
#include "version_graph.h"
#include "bf_macro.h"
#include "lgobject.h"
#include "bf_extfuncs.h"
#include "lg_intfuncs.h"
#include "lg_extfuncs.h"
#include "lg_versionfuncs.h"
#include "util_funcs.h"
#include "clog_globals.h"
#include "clog_extfuncs.h"

/*
 * Routine to perform the leaf portion of an insert operation.
 */


 int 
LG_InsertData (

	FID				*fid,
	int				offset,			/* offset on first page of insertion point*/
	int				length,			/* number of bytes of data to be inserted */
	char			*data,			/* ptr to the data bytes to be inserted */
	PAGELIST		*readList,		/* list of pid/cnt pairs to be read in */
	PAGELIST		*writeList,		/* list of pid/cnt pairs to write back out*/
	BUFGROUP		*bufGroup,
	BOOL			versioned,		/* true if object is versioned		*/
	BOOL			*pagesReplaced,	/* set true if pages replaced in version*/
	PID*			newParentPid 	/* new parent for the chunk 
									   involved in the insert.  NULL if
									   no new parent page created	*/
)
{

	CHUNKDESC		*chunkDesc;		/* user descriptor for leaf accesses */
	int				e;				/* error code */
	OPERATIONINFO	opInfo;
	char			*currentPtr;
	int				newBytes;
	int				i;
	char			*oldDataPtr;    /* ptr into old data to be logged */


	TRPRINT(TR_LG, TR_LEVEL_1, ("offset:%d length:%d", offset, length));

    /*
     *	Check for potential length error.
     */
    SM_ASSERT(LEVEL_3, (writeList->numBytes - readList->numBytes) == length);

    /*
     *	Read in the affected data and create a hole in the appropriate spot.
     */
    if (bf_ReadChunk(fid, bufGroup, readList->numPages, writeList->numPages,
					 readList->pages, 0, writeList->numBytes, &chunkDesc))	{

		return(esmFAILURE);
	}

	/*
	 *	If the object is versioned and the chunk contains shared
	 *	pages (the pages involved in the insert are shared) then
	 *	replace the shared pages.
	 */
	if (versioned && chunkDesc->sharedPages > 0) {
		
		/*
		 *	Replace any leaf pages that are shared, do not replace
		 *	interior nodes (since none should be shared at this
		 *	point).
		 */
		e = LG_ReplaceSharedLeaves(chunkDesc, 0, readList->numBytes, newParentPid, FALSE);
		if (e != esmNOERROR) return (e);

		/*
		 *	Since page id's in the chunk have changed, update the
		 *	writeList's page id's
		 */
		e = LG_ChangeWritePageId(chunkDesc, writeList);
		if (e != esmNOERROR) return (e);

		/*
		 *	Signal the caller that pages were replaced in the version
		 */
		*pagesReplaced = TRUE;
	} else {
		/*
		 *	Signal the caller that NO pages were replaced in the version
		 */
		*pagesReplaced = FALSE;
	}


	/* if we've rearranged data on multiple pages, we use a brute force
	   approach to logging, since there are many possible rearrangements. */
	if (writeList->numPages > 1 ) {

	    /* log removal of all data from all pages in read list */
	    /* note: this could be made more efficient if we had 
	       a good way of telling WHICH pages these are: i.e.
		    left, current or right, or some combination */
        oldDataPtr = chunkDesc->basePtr;
		if (GET_LOGLEVEL(LOG_DATA) ) {
			for (i = 0; i < readList->numPages; i++) {
				opInfo.action = LOG_ACTION_REMOVE_LARGE_DATA;
				opInfo.flags = 0;
				opInfo.start = 0;
				opInfo.size = readList->pages[i].size;
				opInfo.oldData = oldDataPtr; 
				opInfo.newData = NULL;
				opInfo.actionPid = readList->pages[i].pid;
				clog_LogModifyLgPage(&opInfo);
				oldDataPtr += readList->pages[i].size;
			}
		}
    }
	

	/*
	 *	Update the chunk information to reflect the new allocation of pages
	 */
    if (bf_InsertChunk(writeList->numPages, writeList->pages,
					   chunkDesc, offset, length))	{

		bf_UnfixChunk(chunkDesc, BF_DEFAULT);
		return(esmFAILURE);
	}


    /*
     *	Copy the data into the hole in the chunk via the user pointer.
     */
	if (data != NULL)	{
     	bcopy(data, (chunkDesc->bufFrame + offset), (int) length);
	} else {
     	bzero((chunkDesc->bufFrame + offset), (int) length);
	}

	if (GET_LOGLEVEL(LOG_DATA) ) {
		if (writeList->numPages == 1 ) {
			/* handle the simple case of an insert that affects only a single
				data page */
			opInfo.flags = 0;
			opInfo.start = offset;
			opInfo.size = length;
			/* note, old data is ptr to the previous size of the data */
			opInfo.oldData = (char *) &(readList->pages[0].size);
			opInfo.actionPid.volid = writeList->pages[0].pid.volid;
			opInfo.actionPid.page  = writeList->pages[0].pid.page;

			if (data == NULL) {
				opInfo.action = LOG_ACTION_INSERT_LARGE_DATA_ND;
				opInfo.newData = NULL;
				clog_LogModifyLgPageND(&opInfo);
			} else {
				opInfo.action = LOG_ACTION_INSERT_LARGE_DATA;
				opInfo.newData = data;
				clog_LogModifyLgPage(&opInfo);
			}
		
		} else {
			currentPtr = chunkDesc->bufFrame;
			/* log appends of data for all written pages */
			for (i = 0; i < writeList->numPages; i++) {
				newBytes = writeList->pages[i].size;
				opInfo.flags = 0;
				opInfo.start = 0;
				opInfo.size = newBytes;
				opInfo.oldData = NULL;
				opInfo.actionPid.volid = writeList->pages[i].pid.volid;
				opInfo.actionPid.page  = writeList->pages[i].pid.page;

				if (data == NULL) {
					opInfo.action = LOG_ACTION_APPEND_LARGE_DATA_ND;
					opInfo.newData = NULL;
					clog_LogModifyLgPageND(&opInfo);
				} else {
					opInfo.action = LOG_ACTION_APPEND_LARGE_DATA;
					opInfo.newData = currentPtr;
					clog_LogModifyLgPage(&opInfo);
				}

				currentPtr += newBytes;
			}
		}
	}

	/* Unfix the chunk in memory since we copied the data */
    bf_UnfixChunk(chunkDesc, BF_DEFAULT);

	/* return success */
    return(esmNOERROR);
}
