/*
 *   $RCSfile: LG_Merge.c,v $  
 *   $Revision: 1.1.1.1 $  
 *   $Date: 1996/05/04 21:55:24 $      
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/


#include "sysdefs.h"
#include "ess.h"
#include "trace.h"
#include "checking.h"
#include "error.h"
#include "io.h"
#include "page.h"
#include "list.h"
#include "object.h"
#include "tid.h"
#include "bf.h"
#include "chunk.h"
#include "logaction.h"
#include "clog.h"
#include "sm_state.h"
#include "bf_macro.h"
#include "lgobject.h"
#include "lg_intfuncs.h"
#include "lg_extfuncs.h"
#include "clog_extfuncs.h"
#include "util_funcs.h"


/*
 * Routine to merge the contents of a pair of nodes into a single node.
 */


 void
LG_Merge (

	PID				*nodePid,
	SLOTINDEX		nodeRootSlot,	/* non-NIL if root on slotted page*/
	LGNODE			*nodePtr,		/* ptr to buffered parent node	 */
	SLOTINDEX		slotIndex,		/* first slot affected by merge	 */
	PID				*leftPid,
	register LGNODE	*leftPtr,		/* ptr to buffered left child node*/
	PID				*unused,		/* rightPid */
	register LGNODE	*rightPtr,		/* ptr to buffered right child node */
	int				*slotsMoved 	/* # slots moved from right to left (+) */
)
{
    register int	source;			/* source slot index for copying  */
    register int	dest;			/* destination slot index for copying */
    int				leftBytes;		/* original # bytes in left child */
    int				rightBytes;		/* original # bytes in left child */
    int				numLeftSlots;	/* original # slots in left child */
    int				numRightSlots;	/* original # slots in right child */
	OPERATIONINFO	opInfo;
	LGSLOT			tempSlot;


	TRPRINT(TR_LG, TR_LEVEL_1, ("slotIndex:%d", slotIndex));

    /*
     * Get set to shift the right child's contents to the left.
     */
    numLeftSlots = (int)leftPtr->header.numSlots;
    numRightSlots = (int)rightPtr->header.numSlots;
    leftBytes = (int)leftPtr->slot[numLeftSlots-1].count;
    rightBytes = (int)rightPtr->slot[numRightSlots-1].count;
    *slotsMoved = numRightSlots;
	TRPRINT(TR_LG, TR_LEVEL_2, ("numLeftSlots:%d numRightSlots:%d leftBytes:%d rightBytes:%d",
					numLeftSlots, numRightSlots, leftBytes, rightBytes));

	if (GET_LOGLEVEL(LOG_SPACE)) {
		/*
		 *  Log the append that the following for loop will perform
		 */
		opInfo.action = LOG_ACTION_APPEND_LARGE_SLOTS;
		opInfo.flags = 0;
		opInfo.start = numLeftSlots * sizeof(LGSLOT);
		opInfo.size = *slotsMoved * sizeof(LGSLOT);
		opInfo.oldData = NULL;
		opInfo.newData= (char *) &(rightPtr->slot[0]);
		opInfo.actionPid = *leftPid ;
		/* this can't be a slotted root object */
		opInfo.rootSlot = NIL;
		if(clog_LogModifyLgPage(&opInfo)) {
			FAIL
		}
		/*
		 *  Log byte count changes generated by the following for loop
		 */
		opInfo.action = LOG_ACTION_INCREMENT_SLOT_BYTECOUNTS;
		opInfo.flags = 0;
		opInfo.start = numLeftSlots;
		opInfo.size = *slotsMoved;
		opInfo.oldData = NULL;
		opInfo.newData = NULL;
		opInfo.actionPid = *leftPid;
		opInfo.rootSlot = NIL;
		if(clog_LogChangeSlots(&opInfo)) {
			FAIL
		}
	}

    /*
     * Now do the shifting, then update the left child's slot count
     * to reflect its effects.
     */
    for (dest = numLeftSlots, source = 0;  source < numRightSlots;  source++, dest++) {

		leftPtr->slot[dest].count = rightPtr->slot[source].count + leftBytes;
		leftPtr->slot[dest].pid = rightPtr->slot[source].pid;
		leftPtr->slot[dest].shared = rightPtr->slot[source].shared;
    }
    leftPtr->header.numSlots += numRightSlots;

	if (GET_LOGLEVEL(LOG_SPACE)) {
		/*
		 * Now update the parent node to reflect the merge.
		 * log the update
		 */
		tempSlot = nodePtr->slot[slotIndex];
		nodePtr->slot[slotIndex].count += rightBytes;
		opInfo.action = LOG_ACTION_WRITE_LARGE_SLOTS;
		opInfo.flags = 0;
		opInfo.start = slotIndex*sizeof(LGSLOT);
		opInfo.size = 1 * sizeof(LGSLOT);
		opInfo.oldData = (char*) &tempSlot;
		opInfo.newData = (char*) &(nodePtr->slot[slotIndex]);
		opInfo.actionPid = *nodePid;
		/* mark if it is a root object on a slotted page */
		opInfo.rootSlot = nodeRootSlot;
		if(clog_LogWriteLgPage(&opInfo)) {
			FAIL
		}

		/*
		 *	delete the slot entry for the merged node 
		 *	(log the operation first
		 */
		opInfo.action = LOG_ACTION_DELETE_LARGE_SLOTS;
		opInfo.flags = 0;
		opInfo.start = (slotIndex+1)*sizeof(LGSLOT);
		opInfo.size = 1 * sizeof(LGSLOT);
		opInfo.oldData = (char *) &(nodePtr->slot[slotIndex+1]);
		opInfo.newData = NULL;
		opInfo.actionPid = *nodePid;
		opInfo.rootSlot = nodeRootSlot;
		if(clog_LogModifyLgPage(&opInfo)) {
			FAIL
		}
	}

	/* delete the slot entry for the merged node */
    nodePtr->header.numSlots -= 1;
    bcopy((char *) &(nodePtr->slot[slotIndex + 2]),
		  (char *) &(nodePtr->slot[slotIndex + 1]),
	      (int) ((nodePtr->header.numSlots - slotIndex - 1) * sizeof(LGSLOT)) );

	lg_PrintNode(nodePtr, 0, 0);
	lg_PrintNode(leftPtr, 0, 0);
	return;

FailureReturn:
	SM_ERROR(TYPE_FATAL, esmINTERNAL);
}
