/*
 *   $RCSfile: LG_ReplaceSharedLeaves.c,v $  
 *   $Revision: 1.1.1.1 $  
 *   $Date: 1996/05/04 21:55:24 $      
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/


#include "sysdefs.h"
#include "ess.h"
#include "trace.h"
#include "checking.h"
#include "error.h"
#include "io.h"
#include "page.h"
#include "list.h"
#include "object.h"
#include "tid.h"
#include "bf.h"
#include "chunk.h"
#include "lgobject.h"
#include "sm_macro.h"
#include "lock.h"
#include "volume.h"
#include "pool.h"
#include "logaction.h"
#include "clog.h"
#include "sm_state.h"
#include "version_graph.h"
#include "io_extfuncs.h"
#include "bf_extfuncs.h"
#include "lg_intfuncs.h"
#include "lg_versionfuncs.h"
#include "clog_extfuncs.h"
#include "trans.h"
#include "trans_funcs.h"

/*
 * This function replaces all the shared leaves (and their parents 
 * recursively) in a chunk which are contained in the range offset
 * to endByte.
 */

 int
LG_ReplaceSharedLeaves (

	CHUNKDESC	*chunkDesc,	/* chunk to process						*/
    int			offset,    	/* offset from beginning of chunk 		*/
	int			endByte,	/* offset of part of chunk to replace 	*/
	PID			*parentPid, /* used to provide a new parent page of 
							   entire chunk, NULL if parent is not new
							   or it replaceParent is TRUE			*/
	BOOL		replaceParent  /* true if parents of chunk pages 
								  should be checked to see if they are
								  shared and therefore need replacing*/
)
{
	CHUNKPAGE		*chunkPage;
	int				totalBytes;
	int				sharedPages;	/* # of written shared pages*/
	int				e;				/* error code				*/
	OPERATIONINFO 	opInfo;			/* log record info			*/	
	PID				*pidList = NULL;   /* array of new page id's*/


	TRPRINT(TR_LG, TR_LEVEL_1, ("offset:%d endByte:%d", offset, endByte));

	/*
	 *	Determine how many shared pages will be written
	 */
	LG_CountShared(chunkDesc, offset, endByte, &sharedPages);

	/*
	 * 	If any pages in the written part of the chunk are shared then
	 *	create new pages for them and fix up the large object tree.
	 */
	if (sharedPages > 0) {

		PID			nearPid;		/* page to allocate new pages near*/
		PID			prevParentPid;	/* previous parent page updated	*/
		PID			newParentPid;	/* pid of newly allocated parent*/
		int			newLeafIndex; 	/* index into pidList			*/
		CHUNKPAGE*	firstPage;		/* marks first written child of parent*/
		BOOL		replaced; 		/* indicates if parent page replaced*/
									

		TRPRINT(TR_LG, TR_LEVEL_2, ("shared pages:%d", sharedPages));

		/*
		 *  malloc space for the new page list
		 */
		if ((pidList = (PID *) malloc(sharedPages * (int)sizeof(PID))) == NULL)   {
			SM_ERROR(TYPE_USER, esmMALLOCFAILED);
			FAIL
		}
		/* 
		 * something has now been done that we cannot easily undo
		 * in the event of an error, so if an error occurs from now on,
		 * we must abort the tx.
		 */
		sm_SetAbortIfError();
		
		/*
		 *	In this case allocate near the last page in the chunk
		 */
		nearPid = chunkDesc->chunkPage[chunkDesc->pageCount-1].pid;

		/*
		 *	Allocate new pages for the shared pages
		 */
		e = io_AllocPages(nearPid.volid, LGDATA_PAGE2SIZE, 
			sharedPages, &nearPid, pidList, PAGE_LARGEDATA);
		if (e != esmNOERROR) {
			FAIL
		}

		/*
		 *	For each page in the chunk, if its shared and to be 
		 *	written, set up a new page.
		 */
		for (totalBytes = 0, chunkPage = chunkDesc->chunkPage,
				prevParentPid.volid = 0, 	/* no previous page 	*/
				newLeafIndex 		= 0,	/* use first new leaf	*/
				firstPage 			= NULL; 
			 newLeafIndex < sharedPages ;	/* stop after all shared */
			 totalBytes += chunkPage->size, chunkPage++) {
				
			/*
			 *	check to see if first written byte is reached
			 */
			if (totalBytes + chunkPage->size > offset)	{

				TRPRINT(TR_LG, TR_LEVEL_2, ("totalBytes:%d pid:%d",
						totalBytes, chunkPage->pid.page));
			
				/*
				 *	See if this page is the first one pointed to
				 *	by its parent.  
				 */
				if (prevParentPid.volid == 0 || 
					(chunkPage->parentPid != prevParentPid.page)) {

					/*
					 *	Remember this page
					 */
					prevParentPid.page = chunkPage->parentPid;
					prevParentPid.volid = chunkPage->pid.volid;

					/*
					 *	First to be updated, so replace the parent
					 *	if the caller requests replacement and the
					 *	parent is shared.
					 */
					if (replaceParent) {
						e = LG_ReplaceIfShared(chunkDesc->bufGroup, &chunkDesc->rootPid,
												chunkDesc->rootSlot, &prevParentPid, 
												chunkPage->offset, 
												&newParentPid, &replaced);
						if (e != esmNOERROR) {
							FAIL
						}
					} else {
						/*
						 *	Caller did not request parent replacement
						 */
						if (parentPid != NULL) {
							/* caller provided a replacement parent pid.  */
							replaced = TRUE;
							newParentPid = *parentPid;
						} else {
							/* There is no new parent page */
							replaced = FALSE;
						}
					}
					firstPage = chunkPage;
				}

				/*
				 *	Replace shared data pages
				 */
				if (chunkPage->shared) {
					
					TRPRINT(TR_LG, TR_LEVEL_2, ("shared page id:%d",
							chunkPage->pid.page));
				
					SM_ASSERT(LEVEL_3, newLeafIndex < sharedPages);	

					/*
					 *	Replace the shared leaf page with a new leaf
					 */
					e = bf_ReplaceLeafChunk(chunkDesc, chunkPage, 
							&pidList[newLeafIndex], replaced ? &newParentPid : NULL);
					if (e != esmNOERROR) {
						FAIL
					}

					if (GET_LOGLEVEL(LOG_DATA) ) {
						/*
						 *	Log the replacement as and append of data to
						 *	the new page.
						 */
						opInfo.action = LOG_ACTION_APPEND_LARGE_DATA;
						opInfo.flags = 0;
						opInfo.start = 0;
						opInfo.size = chunkPage->size;
						opInfo.oldData = NULL;
						opInfo.newData = chunkDesc->bufFrame + chunkPage->offset;
						opInfo.actionPid = pidList[newLeafIndex];
						clog_LogModifyLgPage(&opInfo);
					}

					/* get the next replacement leaf */
					newLeafIndex++;

				}

				/*
				 *	This page is the last one pointed to by its parent
				 *	if its the last one shared in the chunk or if the 
				 *	parent of the next page is different from the 
				 *	current parent.  
				 */
				if ( ( (newLeafIndex) == sharedPages) || 
					 ((chunkPage+1)->parentPid != prevParentPid.page)) {
					/*
					 *	We have processed all the children of this page
					 *	so update its parent.  The parent is updated
					 *	all at once to avoid generating lots of log
					 *	records.
					 */
					e = LG_UpdateLeafParent(chunkDesc, firstPage, chunkPage);
					if (e != esmNOERROR) {
						FAIL
					}
				}

			}
		}
		/* Deallocate the space for the page list */
	}

Success:
	SM_FREE_PTR(pidList);
	return esmNOERROR;

FailureReturn:
	SM_FREE_PTR(pidList);
	return esmFAILURE;
}
