/*
 *   $RCSfile: LG_ReplaceSharedNodes.c,v $  
 *   $Revision: 1.1.1.1 $  
 *   $Date: 1996/05/04 21:55:24 $      
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/


#include "ess.h"
#include "sysdefs.h"
#include "trace.h"
#include "checking.h"
#include "error.h"
#include "io.h"
#include "page.h"
#include "list.h"
#include "object.h"
#include "tid.h"
#include "bf.h"
#include "chunk.h"
#include "bf_macro.h"
#include "lgobject.h"
#include "sm_macro.h"
#include "lock.h"
#include "pool.h"
#include "logaction.h"
#include "clog.h"
#include "trans.h"
#include "sm_state.h"
#include "version_graph.h"
#include "bf_extfuncs.h"
#include "bf_intfuncs.h"
#include "lg_intfuncs.h"
#include "lg_versionfuncs.h"
#include "clog_extfuncs.h"


/*
 *	Routine to replace all shared nodes in a stack of large object 
 *	btree interior nodes.
 */


 int
LG_ReplaceSharedNodes (
	BUFGROUP	*bufGroup,	/* page to possibly replace			*/
	PID			*rootPid,	/* root page of object 				*/
	TSTACK		*tstack,	/* stack of interior nodes			*/
	int			sharedCount,/* number of shared nodes in stack	*/
	PID			*newTopPid 	/* new page at top of stack			*/
)
{
	int				e;			/* error code						*/
	TSTACKENT		*stackTop;
	TSTACKENT		*stackEntry;
	CHUNKPAGE		newNodes[TSTACKSIZE]; 	
	int				nextNewPage;/* index into newNodes array		*/
	int				i, j;		/* generic index variables			*/
	PID				childPid;	/* pid of new child					*/
	OPERATIONINFO	opInfo;		/* log record information			*/
	LGSLOT			tempSlot;	/* temp slot for logging old value	*/

	TRPRINT(TR_LG, TR_LEVEL_1, ("replacing %d interior nodes", 
								sharedCount));

    /*
     * Get the top page in the stack.	Don't
     * pop the stack here, as we'll need it intact later.
     */
    if ((stackTop = LG_TopTstack(tstack)) == NULL)  {

        SM_ERROR(TYPE_FATAL, esmINTERNAL);
    }

	/*
	 *	Top of stack must be shared
	 */
	SM_ASSERT(LEVEL_3, stackTop->shared == TRUE);
	SM_ASSERT(LEVEL_3, sharedCount > 0);

	/*
	 *	Allocate sharedCount nodes
	 */
	e = LG_AllocNodes(bufGroup, newNodes, rootPid, sharedCount);
	if (e != esmNOERROR) return e;

	/*
	 *	For each node marked shared, replace it and update its
	 *	parent.  Start at the top of the stack. 
	 */
	for (i = 0, stackEntry = stackTop, nextNewPage = 0, 
			childPid.volid = 0; 
		 i <= sharedCount;
		 i++, stackEntry--) {

		/*
		 *	See if page is shared
		 */
		if (stackEntry->shared) {
			LGNODE 		*newNode;
			CHUNKPAGE	*newPage;

			/*
			 *	Make sure things are consistent
			 */
			SM_ASSERT(LEVEL_3, nextNewPage < sharedCount);

			/*
			 *	Set up a new page for it
			 */
			newPage = &newNodes[nextNewPage];
			newNode = (LGNODE *) newPage->nodeLink->bufFrame;
			nextNewPage++;
			*newNode = *(stackEntry->nodePtr);
			CHECK_LARGENODE_MAGIC(newNode);
			printf("Be careful in LG_ReplaceSharedNodes to fix the lrc\n");

			/*
			 *	Set the number of slots to be zero, so that the
			 *	following log record has the correct value.
			 */
			newNode->header.numSlots = 0;

			if (GET_LOGLEVEL(LOG_SPACE)) {
				/*
				 *	Log the initialization of the node header
				 */
				opInfo.action = LOG_ACTION_INIT_LARGE_NODE;
				opInfo.flags = 0;
				opInfo.start = 0;
				opInfo.size = sizeof(LGNODEHDR);
				opInfo.oldData = NULL;
				opInfo.newData = (char*) &(newNode->header);
				opInfo.actionPid = newPage->pid;
				/* this can't be a slotted root object */
				opInfo.rootSlot = NIL;
				if(clog_LogWriteLgNodeHdr(&opInfo)){
					FAIL
				}
			}

			/*
			 *	Restore the slot count and make
			 *	all slots point to shared pages
			 */
			newNode->header.numSlots = stackEntry->nodePtr->header.numSlots; 
			for (j = 0; j < newNode->header.numSlots; j++) {
				newNode->slot[j].shared = TRUE;
			}
							
			/*
			 *	If this is not the lowest page, update the pointer
			 *	to it's child.
			 */
			if (stackEntry != stackTop) {
				int		childIndex = stackEntry->slotIndex;
				newNode->slot[childIndex].shared = FALSE;
				newNode->slot[childIndex].pid = childPid;
			}

			if (GET_LOGLEVEL(LOG_SPACE)) {
				/*
				 *	Log the copy of the slots as an append of the slots
				 */
				opInfo.action = LOG_ACTION_APPEND_LARGE_SLOTS;
				opInfo.flags = 0;
				opInfo.start = 0;
				opInfo.size = (int) (sizeof(LGSLOT) * newNode->header.numSlots);
				opInfo.oldData = NULL;
				opInfo.newData= (char *) &(newNode->slot[0]);
				opInfo.actionPid = newPage->pid;
				/* this can't be a slotted root object */
				opInfo.rootSlot = NIL;
				SM_ASSERT(LEVEL_3, stackEntry->rootSlot == NIL);
				if(clog_LogModifyLgPage(&opInfo)) {
					FAIL
				}
			}

			/* Remember this page */
			childPid = newPage->pid;

			/* Mark the new page dirty, and set up logging info */
			bf_DirtyPage(newPage->nodeLink->pageHash);

		} else {
			
			LGNODE	*node;
			int		childIndex;

			/*
			 *	Verify that loop will exit, and that we're at the
			 *	first unshared page.
			 */
			SM_ASSERT(LEVEL_3, i == sharedCount);
			SM_ASSERT(LEVEL_3, ! stackEntry->shared);

			/*
			 *	This page is the parent of the first shared page,
			 *	so update the pointer to the page 
			 */
			node = (LGNODE *) stackEntry->nodeLink->bufFrame;
			childIndex = stackEntry->slotIndex;
			tempSlot = node->slot[childIndex];
			node->slot[childIndex].shared = FALSE;
			node->slot[childIndex].pid = childPid;

			if (GET_LOGLEVEL(LOG_SPACE)) {
				/*
				 *	Log the write of the slot
				 */
				opInfo.action = LOG_ACTION_WRITE_LARGE_SLOTS;
				opInfo.flags = 0;
				opInfo.start = childIndex*sizeof(LGSLOT);
				opInfo.size = 1 * sizeof(LGSLOT);
				opInfo.oldData = (char*) &tempSlot;
				opInfo.newData = (char*) &(node->slot[childIndex]);
				opInfo.actionPid = stackEntry->pid;
				opInfo.rootSlot = stackEntry->rootSlot;
				if(clog_LogWriteLgPage(&opInfo)) {
					FAIL
				}
			}

			/* Mark the page dirty */
			bf_DirtyPage(stackEntry->nodeLink->pageHash);

		}
	}

	/* Unpin all the new pages */
	for (i = 0; i < sharedCount; i++) {
		bf_UnfixPage(newNodes[i].nodeLink, BF_DEFAULT, TRUE);
	}

	/* return the new page, noting pid was replaced */
	*newTopPid = newNodes[0].pid;
	return(esmNOERROR);
FailureReturn:
	SM_ERROR(TYPE_FATAL, esmINTERNAL);
	return esmFAILURE; /* to keep the compiler quiet */
}
