/*
 *   $RCSfile: LG_Reshuffle.c,v $  
 *   $Revision: 1.1.1.1 $  
 *   $Date: 1996/05/04 21:55:24 $      
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/


#include <stdio.h>
#include "ess.h"
#include "trace.h"
#include "checking.h"
#include "io.h"
#include "page.h"
#include "list.h"
#include "object.h"
#include "tid.h"
#include "bf.h"
#include "chunk.h"
#include "logaction.h"
#include "clog.h"
#include "sm_state.h"
#include "bf_macro.h"
#include "lgobject.h"
#include "lg_intfuncs.h"
#include "lg_extfuncs.h"
#include "clog_extfuncs.h"
#include "error.h"


/*
 * Routine to reshuffle the contents of a pair of nodes in order to
 * equalize their entries.
 */


 int
LG_Reshuffle (

	PID			*nodePid,
	SLOTINDEX	nodeRootSlot,	/* non-NIL if root on slotted page	*/
	LGNODE		*nodePtr,		/* ptr to buffered parent node 		*/
	SLOTINDEX	slotIndex,		/* first slot affected by reshuffle	*/
	PID			*leftPid,
	LGNODE		*leftPtr,		/* ptr to buffered left child node	*/
	PID			*rightPid,
	LGNODE		*rightPtr,		/* ptr to buffered right child node */
	int			*slotsMoved 	/* # slots moved from right to left (+/-) */
)
{

    int			totalSlots;		/* total # of child slots involved 	*/
    int			oldLeftSlots;	/* original # of slots in left child*/
    int			oldRightSlots;	/* original # of slots in right child*/
    int			newLeftSlots;	/* new # of slots in left child 	*/
    int			newRightSlots;	/* new # of slots in right child 	*/
    int			oldLeftBytes;	/* old # of bytes in left child 	*/
    int			newLeftBytes;	/* new # of bytes in left child 	*/
    int			movedSlots;		/* abs( # of slots to be moved over )*/
    int			movedBytes;		/* # of bytes moved as a result 	*/
    int			source;
	int			dest;			/* indices for copying slot info 	*/
	LGSLOT		tempSlot;		
	OPERATIONINFO	opInfo;


	TRACE(TR_LG, TR_LEVEL_1);

    /*
     * Figure out how to evenly distribute slots between child nodes.
     */
    oldLeftSlots = (int)leftPtr->header.numSlots;
    oldLeftBytes = (int)leftPtr->slot[oldLeftSlots-1].count;
    oldRightSlots = (int)rightPtr->header.numSlots;
    totalSlots = oldLeftSlots + oldRightSlots;
    newLeftSlots = totalSlots / 2;
    newRightSlots = totalSlots - newLeftSlots;
    *slotsMoved = newLeftSlots - oldLeftSlots;

	/* Remember the data in the relevent node slot (for logging) */
	tempSlot = nodePtr->slot[slotIndex];

    /*
     * Now do the reshuffling, either right to left or vice versa.
     */
    if (*slotsMoved > 0) {

		/*
		 * Reshuffle stuff from the right child into the left child as follows:
		 *   1. Compute the amount of stuff to move.
		 *   2. Adjust parent node count for slotIndex to reflect movement.
		 *   3. Move stuff from right node into left node, adjusting counts.
		 *   4. Move stuff from right node over (left), again adjusting counts.
		 *   5. Update overall slot counts for both child nodes.
		 */
		movedSlots = *slotsMoved;
		movedBytes =  (int)rightPtr->slot[movedSlots-1].count;
		/* logged at end-of-function */
		nodePtr->slot[slotIndex].count += movedBytes; 

		if (GET_LOGLEVEL(LOG_SPACE)) {
			/*
			 *	Log the append that the following for loop will perform
			 */
			opInfo.action = LOG_ACTION_APPEND_LARGE_SLOTS;
			opInfo.flags = 0;
			opInfo.start = oldLeftSlots * sizeof(LGSLOT);
			opInfo.size = newLeftSlots * sizeof(LGSLOT);
			opInfo.oldData = NULL;
			opInfo.newData= (char *) &(rightPtr->slot[0]);
			opInfo.actionPid = *leftPid ;
			/* this can't be a slotted root object */
			opInfo.rootSlot = NIL;
			if(clog_LogModifyLgPage(&opInfo)) {
				FAIL
			}
			/*
			 *  Log byte count changes generated by the following for loop
			 */
			opInfo.action = LOG_ACTION_INCREMENT_SLOT_BYTECOUNTS;
			opInfo.flags = 0;
			opInfo.start = oldLeftSlots;
			opInfo.size = oldLeftBytes;
			opInfo.oldData = NULL;
			opInfo.newData = NULL;
			opInfo.actionPid = *leftPid;
			opInfo.rootSlot = NIL;
			if(clog_LogChangeSlots(&opInfo)){
				FAIL
			}
		}

		/* 
 		 *	Perform the append
         */
		for (dest = oldLeftSlots, source = 0;  source < movedSlots;  source++) {

			leftPtr->slot[dest].count = rightPtr->slot[source].count + oldLeftBytes;
			leftPtr->slot[dest].pid = rightPtr->slot[source].pid;
			leftPtr->slot[dest].shared = rightPtr->slot[source].shared;
			dest++;
		}

		if (GET_LOGLEVEL(LOG_SPACE)) {
			/*
			 *  Log the delete of slots performed by the following for loop
			 */
			opInfo.action = LOG_ACTION_DELETE_LARGE_SLOTS;
			opInfo.flags = 0;
			opInfo.start = 0*sizeof(LGSLOT);
			opInfo.size = movedSlots * sizeof(LGSLOT);
			opInfo.oldData = (char *) &(rightPtr->slot[0]);
			opInfo.newData = NULL;
			opInfo.actionPid = *rightPid;
			opInfo.rootSlot = NIL;
			if(clog_LogModifyLgPage(&opInfo)) {
				FAIL
			}
			/*
			 *  Log byte count changes generated by the following for loop
			 */
			opInfo.action = LOG_ACTION_DECREMENT_SLOT_BYTECOUNTS;
			opInfo.flags = 0;
			opInfo.start = 0;
			opInfo.size = movedBytes;
			opInfo.oldData = NULL;
			opInfo.newData = NULL;
			opInfo.actionPid = *rightPid;
			opInfo.rootSlot = NIL;
			if(clog_LogChangeSlots(&opInfo)){
				FAIL
			}
		}

		/* 
 		 *	Perform the deletion of the slots
         */
		for (dest = 0, source = movedSlots;  source < oldRightSlots;  source++) {

			rightPtr->slot[dest].count = rightPtr->slot[source].count-movedBytes;
			rightPtr->slot[dest].pid = rightPtr->slot[source].pid;
			rightPtr->slot[dest].shared = rightPtr->slot[source].shared;
			dest++;
		}

		leftPtr->header.numSlots = newLeftSlots;
		rightPtr->header.numSlots = newRightSlots;

     } else {

		/*
		 * Reshuffle stuff from the left child into the right child as follows:
		 *   1. Compute the amount of stuff to move.
		 *   2. Adjust parent node count for slotIndex to reflect movement.
		 *   3. Move stuff from right node over (right), adjusting counts.
		 *   4. Move stuff from left to right node, again adjusting counts.
		 *   5. Update overall slot counts for both child nodes.
		 */
		movedSlots = -(*slotsMoved);
		newLeftBytes = (int)leftPtr->slot[newLeftSlots-1].count;
		movedBytes = oldLeftBytes - newLeftBytes;
		/* logged at end-of-function */
		nodePtr->slot[slotIndex].count -= movedBytes;

		if (GET_LOGLEVEL(LOG_SPACE)) {
			/*
			 *	Log the insert that the following for loop will perform
			 */
			opInfo.action = LOG_ACTION_INSERT_LARGE_SLOTS;
			opInfo.flags = 0;
			opInfo.start = 0 * sizeof(LGSLOT);
			opInfo.size = newRightSlots * sizeof(LGSLOT);
			opInfo.oldData = NULL;
			opInfo.newData= (char *) &(leftPtr->slot[newLeftSlots]);
			opInfo.actionPid = *rightPid ;
			/* this can't be a slotted root object */
			opInfo.rootSlot = NIL;
			if(clog_LogModifyLgPage(&opInfo)) {
				FAIL
			}
			/*
			 *  Log byte count changes generated by the second for loop
			 */
			opInfo.action = LOG_ACTION_DECREMENT_SLOT_BYTECOUNTS;
			opInfo.flags = 0;
			opInfo.start = 0;
			opInfo.size = newLeftBytes;
			opInfo.oldData = NULL;
			opInfo.newData = NULL;
			opInfo.actionPid = *rightPid;
			opInfo.rootSlot = NIL;
			clog_LogChangeSlots(&opInfo);
			if(clog_LogChangeSlots(&opInfo)){
				FAIL
			}
			/*
			 *  Log byte count changes generated by the following for loop
			 */
			opInfo.action = LOG_ACTION_INCREMENT_SLOT_BYTECOUNTS;
			opInfo.flags = 0;
			opInfo.start = newRightSlots;
			opInfo.size = movedBytes;
			opInfo.oldData = NULL;
			opInfo.newData = NULL;
			opInfo.actionPid = *rightPid;
			opInfo.rootSlot = NIL;
			if(clog_LogChangeSlots(&opInfo)){
				FAIL
			}
		}

		for (dest = newRightSlots - 1, source = oldRightSlots-1;  source >= 0;  source--) {

			rightPtr->slot[dest].count = rightPtr->slot[source].count+movedBytes;
			rightPtr->slot[dest].pid = rightPtr->slot[source].pid;
			rightPtr->slot[dest].shared = rightPtr->slot[source].shared;
			dest--;
		}

		if (GET_LOGLEVEL(LOG_SPACE)) {
			/*
			 *  Log the removal of slots performed by the following for loop
			 */
			opInfo.action = LOG_ACTION_DELETE_LARGE_SLOTS;
			opInfo.flags = 0;
			opInfo.start = newLeftSlots * sizeof(LGSLOT);
			opInfo.size = movedSlots * sizeof(LGSLOT);
			opInfo.oldData = (char *) &(leftPtr->slot[newLeftSlots]);
			opInfo.newData = NULL;
			opInfo.actionPid = *leftPid;
			opInfo.rootSlot = NIL;
			if(clog_LogModifyLgPage(&opInfo)) {
				FAIL
			}
		}
		for (source = newLeftSlots, dest = 0;  dest < movedSlots;  dest++) {

			rightPtr->slot[dest].count = leftPtr->slot[source].count-newLeftBytes;
			rightPtr->slot[dest].pid = leftPtr->slot[source].pid;
			rightPtr->slot[dest].shared = leftPtr->slot[source].shared;
			source++;
		}

		leftPtr->header.numSlots = newLeftSlots;
		rightPtr->header.numSlots = newRightSlots;
     }

	if (GET_LOGLEVEL(LOG_SPACE)) {
		/*
		 *  Log the above assignment to nodePtr->slot[slotIndex]
		 */
		opInfo.action = LOG_ACTION_WRITE_LARGE_SLOTS;
		opInfo.flags = 0;
		opInfo.start = slotIndex*sizeof(LGSLOT);
		opInfo.size = 1 * sizeof(LGSLOT);
		opInfo.oldData = (char*) &tempSlot;
		opInfo.newData = (char*) &(nodePtr->slot[slotIndex]);
		opInfo.actionPid = *nodePid;
		/* mark if it is a root object on a slotted page */
		opInfo.rootSlot = nodeRootSlot;
		if(clog_LogWriteLgPage(&opInfo)) {
			FAIL
		}
	}

	return(esmNOERROR);

FailureReturn:
	SM_ERROR(TYPE_FATAL, esmINTERNAL);
	return esmFAILURE; /* to keep the compiler quiet */
}
