/*
 *   $RCSfile: LG_ScanForLeaves.c,v $  
 *   $Revision: 1.1.1.1 $  
 *   $Date: 1996/05/04 21:55:23 $      
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/


#include <stdio.h>
#include "ess.h"
#include "trace.h"
#include "checking.h"
#include "io.h"
#include "page.h"
#include "list.h"
#include "object.h"
#include "tid.h"
#include "bf.h"
#include "chunk.h"
#include "lock.h"
#include "sm_macro.h"
#include "bf_macro.h"
#include "lgobject.h"
#include "bf_extfuncs.h"
#include "lg_intfuncs.h"
#include "lg_extfuncs.h"
#include "error.h"


/*
 * Routine to read a specified number of bytes from a large object, setting
 * up appropriate state information along the way.  The leaf contents read
 * remain fixed following this routine's invocation.  If all goes well, an
 * indirect pointer to the desired data (contiguously buffered) is returned.
 * In the event of an error, an error indication (-1) is returned.
 *
 *	curOffset and bytesLeft function as global variables.  bytesLeft only
 *	changes as leaf nodes are processed.  curOffset can only be non-zero
 *	on the initial descent of the tree.  Otherwise we are going to be
 *	always searching the leftmost subtree so it must be zero.
 */


 int 
LG_ScanForLeaves (

	PID				*pid,			/* root pid of subtree being read or written*/
	int				*curOffset,		/* current offset within subtree		*/
	int				*bytesLeft,		/* bytes left to work on				*/
	BUFGROUP		*bufGroup,		/* buffer group to use for accesses		*/
	PAGELIST		*leaves,		/* list for gathering leaf information	*/
	int				*firstOffset,	/* place to put offset within first page*/
	SLOTINDEX		rootSlot,		/* true if pid is object's root pid		*/
	BOOL			shared 			/* true if pid shared by other verions	*/
)
{

    register int		i;				/* index							*/
    register int		slot;			/* current slot						*/
    register int		leafCount;		/* leaf's byte count				*/
    register int		lastByte;		/* last byte to look at in subtree	*/
	int				lastSlot;		/* number of slots in the node		*/
    LGNODE          *nodePage;		/* pointer to current node			*/
	GROUPLINK		*nodeLink;


	TRPRINT(TR_LG, TR_LEVEL_1, ("pid:%d curoffset:%d firstoffset:%d bytesLeft:%d",
			pid->page, *curOffset, *firstOffset, *bytesLeft));

	/*
	 *	check to see if this is a header on slotted pages
	 */
	if (rootSlot != NIL)	{

		PAGESLOT		*slotPtr;
		SLOTTEDPAGE		*pagePtr;
		LARGEOBJROOT	*objectPtr;

		TRPRINT(TR_LG, TR_LEVEL_2, ("slot root"));

		/*
		 *	Read the page into a buffer
		 */
		if ((nodeLink = bf_ReadPage(bufGroup, pid, H_SLOTTED,
									SLOTTED_PAGE2SIZE, NL)) == NULL)	{

			return(esmFAILURE);
		}

		/*
		 *	get a pointer to the page, slot and object
		 */
		pagePtr	  = (SLOTTEDPAGE *) nodeLink->bufFrame;
		slotPtr	  = GETSLOTPTR(pagePtr, rootSlot);
		objectPtr = (LARGEOBJROOT *) GETOBJECTPTR(pagePtr, slotPtr);

		/*
		 *	get a pointer to the page
		 */
		nodePage = (LGNODE *) &(objectPtr->rootPage);

		/*
		 *	check the magic of the object
		 */
		CHECK_SLOTROOT_MAGIC(nodePage);

	} else {
		
		/*
		 *	read the page into a buffer
		 */
		if ((nodeLink = bf_ReadPage(bufGroup, pid, H_LARGE, LG_PAGE2SIZE, NL)) == NULL)	{

			return(esmFAILURE);
		}

		/*
		 *	get a pointer to the root page
		 */
		nodePage = (LGNODE *) nodeLink->bufFrame;

		/*
		 *	check the magic number
		CHECK_LARGENODE_MAGIC(nodePage);
		 */
	}
	lg_PrintNode(nodePage, 0, pid->page);

	/*
	 *	check to see which slot we should follow
	 */
    if (*curOffset == 0) {

		/*
		 *	take the leftmost slot
		 */
        slot = 0;

    } else {
		
		/*
		 *	Search for the slot to follow
		 */
        slot = LG_BinarySearch(nodePage, *curOffset);
    }
	TRPRINT(TR_LG, TR_LEVEL_2, ("following slot:%d", slot));

	/*
	 *	get the last slot on the page
	 */
	lastSlot = (int)nodePage->header.numSlots - 1;

    /*
     *	Calculate the last byte that needs to be accessed
	 *	in this subtree.
     */
    lastByte = *curOffset + *bytesLeft - 1;

	/*
	 *	check to see if the number of bytes left exceeds
	 *	the number in this subtree
	 */
    if (lastByte >= nodePage->slot[lastSlot].count) {

		TRPRINT(TR_LG, TR_LEVEL_2, ("last byte exceeds this subtree"));

		/*
		 *	data exceeds this subtree, set the
		 *	number of bytes to those in this subtree
		 */
        lastByte = (int)nodePage->slot[lastSlot].count - 1;
    }
	TRPRINT(TR_LG, TR_LEVEL_2, ("lastByte:%d", lastByte));

    /*
     * Adjust curOffset so that it is relative to the next subtree.
     */
    if (slot > 0) {

		/*
		 *	This can only happen on the initial descent of the tree
		 */
        *curOffset -= (int)nodePage->slot[slot-1].count;
		TRPRINT(TR_LG, TR_LEVEL_2, ("resetting curoffset:%d", *curOffset));
    }

    /*
     * If we're at the parent of a leaf, then run along above the leaves
     * to subtract byte counts from bytesLeft and add them to the list.
     */
    if (nodePage->header.height == 1) {

		register int	numSlots;


		TRPRINT(TR_LG, TR_LEVEL_2, ("at parent of a leaf"));

		/*
		 *	get the number of slots into the register
		 */
		numSlots = (int)nodePage->header.numSlots;

		/*
		 *	get another leaf to read as long as we have not yet
		 *	filled the count requested or we run out of leaves
		 */
        for (i = slot; (i < numSlots) && (*bytesLeft) > 0; i++) {

			/*
			 *	get the number of bytes contained in the ith leaf
			 */
			leafCount = LG_SlotCount(nodePage, i);

			/*
			 *	place the pid of the leaf along with the count
			 *	in the linear array of leaves to be read
			 */
            LG_AppendPage(leaves, &(nodePage->slot[i].pid), leafCount,
						  FALSE, nodeLink, pid->page, rootSlot, i, 
						  shared | nodePage->slot[i].shared);

			/*
			 *	Decrement the number of bytes required by the amount
			 *	in the leaf
			 */
            *bytesLeft -= leafCount;
			TRPRINT(TR_LG, TR_LEVEL_2, ("leafCount:%d bytesLeft:%d",
					leafCount, *bytesLeft));

            /*
             * If we're just begining the search, then curOffset will be nonzero. 
             */
	    	if (*firstOffset < 0) {

				/*
			 	 *	Record this as the offset on the first page.
				 */
				*firstOffset = *curOffset;
				TRPRINT(TR_LG, TR_LEVEL_2, ("setting first offset:%d", *firstOffset));

				/*
				 * then adjust bytesLeft so we don't shortchange the requestor.
				 */
				*bytesLeft += *curOffset;

				/*
				 *	the rest of the read will by definition have
				 *	to have current offsets of zero since
				 *	we must read contiguous space
				 */
				*curOffset = 0;
	    	}
			TRPRINT(TR_LG, TR_LEVEL_2, ("bytesLeft:%d  curOffset:%d",
					*bytesLeft, *curOffset));
        }

    } else {

		TRPRINT(TR_LG, TR_LEVEL_2, ("at node height:%d", nodePage->header.height));

		/*
         *	We recursively descend down the subtree.
         */
        for (i = slot; i < nodePage->header.numSlots; i++) {

			BOOL	ptrShared;	/* indicated that page is shared */

			TRPRINT(TR_LG, TR_LEVEL_2, ("scanning pid:%d slot:%d", pid->page, i));
	
			/*
			 *	A page is shared by another version if the shared flag is set
			 *	in the slot pointing to the page or if the current root is
			 *	shared.
			 */
			ptrShared = nodePage->slot[i].shared | shared;

            if (LG_ScanForLeaves( &(nodePage->slot[i].pid), curOffset,
	            		          bytesLeft, bufGroup, leaves, firstOffset, 
								  NIL, ptrShared))	{

				return(esmFAILURE);
			}

			/*
			 *	check to see if we have read in all the required bytes
			 */
            if (nodePage->slot[i].count > lastByte) {

				TRPRINT(TR_LG, TR_LEVEL_2, ("have read all bytes in this subtree"));
                break;
	    	}
        }
    }

    /*
	 *	unfix page
	 */
    bf_UnfixPage(nodeLink, BF_DEFAULT, FALSE);

	/*
	 *	return success
	 */
    return(esmNOERROR);
}
