/*
 *   $RCSfile: LG_Shadow.c,v $  
 *   $Revision: 1.1.1.1 $  
 *   $Date: 1996/05/04 21:55:21 $      
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/


#include "sysdefs.h"
#include "ess.h"
#include "trace.h"
#include "checking.h"
#include "error.h"
#include "io.h"
#include "page.h"
#include "list.h"
#include "object.h"
#include "tid.h"
#include "bf.h"
#include "chunk.h"
#include "lock.h"
#include "bf_macro.h"
#include "lgobject.h"
#include "volume.h"
#include "lsn.h"
#include "logaction.h"
#include "loginfo.h"
#include "clog.h"
#include "sm_state.h"
#include "io_extfuncs.h"
#include "bf_extfuncs.h"
#include "lg_intfuncs.h"
#include "lg_extfuncs.h"
#include "clog_extfuncs.h"
#include "clog_globals.h"
#include "util_funcs.h"


/*
 * Routine for shadowing internal nodes.
 *
 * It shadows the page by:
 *
 *	(1) allocating a new page
 *	(2) renaming the buffer to refer to the new page
 *	(3) setting the new page dirty
 *
 * If LG_Shadow is called on the root page itself, it simply
 * returns its original page id in newPid.  (The root
 * page is updated in place, not shadowed!!!)
 *
 * Upon return, newPid contains the pid of the new shadow page
 */


 int 
LG_Shadow (

	PID			*oldPid,		/* pid of page to be shadowed */
	PID			*newPid,		/* pid of new replacement page */
	PID			*rootPid,		/* pid of object's root page */
	BUFGROUP	*bufGroup,		/* buffer group for operation */
	GROUPLINK	**nodeLink 
)
{

	GROUPLINK		*tempLink;
	OPERATIONINFO	opInfo;
	LGNODE			*newNode;
	LGNODE			*origNode;

	TRACE(TR_LG, TR_LEVEL_1);

	/*
	 *	Check to see if we are attempting to shadow the root node
	 */
    if (PIDEQ(*oldPid, *rootPid)) {

		TRPRINT(TR_LG, TR_LEVEL_2, ("new root pid"));

		/*
		 *	copy the root pid so new pid refers to the root (same)
		 */
		*newPid = *rootPid;
		return(esmNOERROR);
	}

	/*
	 *	allocate a new page to hold the shadow
	 */
	if (io_AllocPages(oldPid->volid, LG_PAGE2SIZE, 1,
					  oldPid, newPid, PAGE_LARGENODE))	{

		FAIL
	}	
	TRPRINT(TR_LG, TR_LEVEL_2, ("allocating page:%d to shadow:%d",
			newPid->page, oldPid->page));

	/*
	 *	check the magic number of the node link
	 */
	CHECK_GROUPLINK_MAGIC(*nodeLink);

	/*
	 *  We must allocate a new buffer for the new page
	 */
	if ((tempLink = bf_GetPage(bufGroup, newPid, H_LARGE, NULL, LG_PAGE2SIZE
)) == NULL) {

		FAIL
	}

	/*
	 *  copy the root node header into the new node and initialize the lrc
	 */
	newNode = (LGNODE*)tempLink->bufFrame;
	origNode = (LGNODE*)(*nodeLink)->bufFrame;
	newNode->header = origNode->header;
	newNode->header.numSlots = 0;
	newNode->header.lrc = *(GET_LOGINFO->initLRC);
	bf_DirtyPage(tempLink->pageHash);

	if (GET_LOGLEVEL(LOG_SPACE)) {
		/*
		 *	Log header initailization of the page
		 */
		opInfo.action = LOG_ACTION_INIT_LARGE_NODE;
		opInfo.flags = 0;
		opInfo.start = 0;
		opInfo.size = (int)sizeof(LGNODEHDR);
		opInfo.oldData = NULL;
		opInfo.newData = (char*) &(newNode->header);
		opInfo.actionPid = *newPid; 
		/* this can't be a slotted root object */
		opInfo.rootSlot = NIL;
		if(clog_LogWriteLgNodeHdr(&opInfo)) {
			FAIL
		}

		/*
		 *	Log the "append" of the slots to the page
		 */
		opInfo.action = LOG_ACTION_APPEND_LARGE_SLOTS;
		opInfo.flags = 0;
		opInfo.start = 0;
		opInfo.size = (int) ((int)sizeof(LGSLOT) * origNode->header.numSlots);
		opInfo.oldData = NULL;
		opInfo.newData= (char *) &(newNode->slot[0]);
		opInfo.actionPid = *newPid;
		/* this can't be a slotted root object */
		opInfo.rootSlot = NIL;
		if(clog_LogModifyLgPage(&opInfo)){
			FAIL
		}
	}

	/*
	 *  append the slots to the new node 
	 */
	newNode->header.numSlots = origNode->header.numSlots;
	bcopy((char*) (origNode->slot+0), 
		(char*) (newNode->slot+0), 
		(int)(origNode->header.numSlots * (int)sizeof(LGSLOT)));

	/* Unfix the original version of the page */
	bf_UnfixPage(*nodeLink, BF_DEFAULT, FALSE);
		
	/* return the new new page */
	*nodeLink = tempLink;
    return(esmNOERROR);


FailureReturn:
	/* TODO: what about unfixing? */
	return esmFAILURE;
}
