/*
 *   $RCSfile: LG_SortRootNodeChildren.c,v $  
 *   $Revision: 1.1.1.1 $  
 *   $Date: 1996/05/04 21:55:24 $      
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/


#include <stdio.h>
#include "ess.h"
#include "trace.h"
#include "checking.h"
#include "io.h"
#include "page.h"
#include "list.h"
#include "object.h"
#include "tid.h"
#include "bf.h"
#include "chunk.h"
#include "lock.h"
#include "bf_macro.h"
#include "sm_macro.h"
#include "lgobject.h"
#include "volume.h"
#include "pool.h"
#include "version_graph.h"
#include "io_extfuncs.h"
#include "bf_extfuncs.h"
#include "lg_extfuncs.h"
#include "lg_versionfuncs.h"
#include "error.h"


/*
 * LG_SortRootNodeChildren finds the children of the root node for 
 * each object root node in a list and adds them to the proper levels 
 * in and array of lists for each level in the large obj btree.  They
 * are also added the the hash table for later use in detection of
 * sharing.
 */


 int
LG_SortRootNodeChildren (
	BUFGROUP	*bufGroup,
	LIST		*rootList,		/* list of root nodes				*/
	LIST		levels[],		/* nodes at each level in lg obj tree*/
	LGNODEHASHTABLE *nodeHashTable,
	POOL		*nodeListPool 	/* pool of node list elements		*/
)
{
	LGNODELIST	*node;
    LGNODE		*nodePage;		/* ptr to buffered node			*/
	GROUPLINK	*nodeLink;
    int			i;			

	TRPRINT(TR_LG, TR_LEVEL_1, ("organizing adj objects"));

	/*
	 *	Look at each root node in the root list
	 */
	for(node = (LGNODELIST*) FIRST_LIST_ELEMENT(rootList);
		node != NULL;
		node = (LGNODELIST*) NEXT_LIST_ELEMENT(&node->list)) {

		/*
		 *	check to see if this is a header on slotted pages
		 */
		if (node->slot != NIL)	{

			PAGESLOT		*slotPtr;
			SLOTTEDPAGE		*pagePtr;
			LARGEOBJROOT	*objectPtr;

			TRPRINT(TR_LG, TR_LEVEL_1, ("slot node"));

			/*
			 * Read in the slotted page containing the root 
			 */
			if ((nodeLink = bf_ReadPage(bufGroup, &node->pid, H_SLOTTED,
										SLOTTED_PAGE2SIZE, NL)) == NULL)	{

				return(esmFAILURE);
			}

			/*
			 *	get a pointer to the page, slot and object
			 */
			pagePtr = (SLOTTEDPAGE *) nodeLink->bufFrame;
			slotPtr = GETSLOTPTR(pagePtr, node->slot);
			objectPtr = (LARGEOBJROOT *) GETOBJECTPTR(pagePtr, slotPtr);

			/*
			 *	get a pointer to the page
			 */
			nodePage = (LGNODE *) &(objectPtr->rootPage);

			/*
			 *	check the magic of the object
			 */
			CHECK_SLOTROOT_MAGIC(nodePage);

		} else {

			/*
			 * Read in the root page
			 */
			if ((nodeLink = bf_ReadPage(bufGroup, &node->pid, H_LARGE,
										LG_PAGE2SIZE, NL)) == NULL)	{

				return(esmFAILURE);
			}

			/*
			 *	get a pointer to the node
			 */
			nodePage = (LGNODE *) nodeLink->bufFrame;
			TRPRINT(TR_LG, TR_LEVEL_2, ("at level:%d", nodePage->header.height));
			lg_PrintNode(nodePage, 0, node->pid.page);
		}
		

		/*
		 *	For each child of the node, add it to levelList, and put
		 *	it in the hash table.
		 */
		for (i = 0; i < nodePage->header.numSlots; i++) {

			/*
			 *	Get an element for the node list
			 */
			node = (LGNODELIST*) poolDeq(nodeListPool);
			if (node == NULL) {

				SM_ERROR(TYPE_WARNING, esmNOFREELGNODELIST);
				return(esmFAILURE);
			}
			node->pid = nodePage->slot[i].pid;
			node->slot = NIL;

			/*
			 *	Initialize the list element, and put it in the list
			 *	at the level one lower than its parent.
			 */
			SM_ASSERT(LEVEL_3, nodePage->header.height > 0);
			listEnq(&levels[nodePage->header.height-1], &node->list);

			/*
			 *	Put the page in the hash table
			 */
			if ( ! LG_HashPutNode(nodeHashTable, node)) {

				/*
				 *	Node was already in the table, so remove it
				 */
				listRemove(&node->list);
				poolEnq(nodeListPool, &node->list);
			}
		}


		/*
		 *	unfix the root page
		 */
		bf_UnfixPage(nodeLink, BF_DEFAULT, FALSE);

	}

	/*
	 *	return success
	 */
	return(esmNOERROR);

}
