/*
 *   $RCSfile: LG_Traverse.c,v $  
 *   $Revision: 1.1.1.1 $  
 *   $Date: 1996/05/04 21:55:21 $      
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/


#include <stdio.h>
#include "ess.h"
#include "trace.h"
#include "checking.h"
#include "io.h"
#include "page.h"
#include "list.h"
#include "object.h"
#include "tid.h"
#include "bf.h"
#include "chunk.h"
#include "lock.h"
#include "lsn.h"
#include "clog.h"
#include "logaction.h"
#include "loginfo.h"
#include "logrecs.h"
#include "bf_macro.h"
#include "sm_macro.h"
#include "lgobject.h"
#include "sm_state.h"
#include "bf_extfuncs.h"
#include "lg_intfuncs.h"
#include "lg_extfuncs.h"
#include "clog_globals.h"
#include "clog_extfuncs.h"
#include "error.h"


/*
 * Traverse a tree to the indicated spot.  When done, the traversal stack
 * will contain the traversal path, and the offset will indicate the offset
 * of the first targeted byte in the path's (long) leaf page.  All pages on
 * the path will remain fixed in the buffer pool.
 */


 int 
LG_Traverse (

	PID				*rootPid,		/* PID of tree's root page			*/
	register int	*offset,		/* byte offset from start of object	*/
	int			numBytes,		/* amount of data to insert or append*/
	OBJOP			opType,			/* type of traversal desired		*/
	TSTACK			*stack,			/* stack for recording traversal	*/
	BUFGROUP		*bufGroup,		/* buffer group for all accesses	*/
	SLOTINDEX		rootSlot,		/* slot of root, if on slotted page	*/
	BOOL			versioned,		/* true is object is a version	*/
	int			updateHeight,	/* begin updating counts at this 
									   height and lower */
	int			*sharedCount 	/* number of shared pages in stack 	*/
)
{

    LGNODE			*nodePtr;		/* pointer to current node pointer*/
	GROUPLINK		*nodeLink;
    int			height;			/* height of node we're looking at	*/
    int			travSlot;		/* slot index taken in traversal*/
    int			slotIndex;		/* slot number (array index)	*/
    PID				nodePid;		/* pid of the current node		*/
	BOOL			shared; 		/* current node is shared   	*/
    OPERATIONINFO	opInfo;			/* logging info */
	BOOL			newSlotRoot; 	/* are we creating a new object with a
																slotted root?*/


	TRACE(TR_LG, TR_LEVEL_1);

    /*
     * Set up initial nodePid and stack
     */
    LG_ResetTstack(stack);
    nodePid = *rootPid;

	/*
	 *	The root node is never shared
	 */
	shared = FALSE;
	*sharedCount = 0;

	/*
	 *	traverse the tree based on whether it is an insert
	 *	or an append oriented traversal
	 */
    switch(opType) {

		/*
		 *	Do an insert oriented traversal
		 */
		case INSERT:

			TRPRINT(TR_LG, TR_LEVEL_2, ("insert"));

			/*
			 * Traverse until we hit the parent of the targeted leaf.
			 */
			do {

				/*
				 * Read the node in
				 */
				if ((nodeLink = bf_ReadPage(bufGroup, &nodePid,
					(rootSlot != NIL ? H_SLOTTED : H_LARGE),
					(rootSlot != NIL ? SLOTTED_PAGE2SIZE : LG_PAGE2SIZE), NL)) == NULL)	{

					SM_ERROR(TYPE_FATAL, esmINTERNAL);
				}

				/*
				 *	get the pointer to the node
				 */
				if (rootSlot != NIL)	{

					PAGESLOT		*slotPtr;
					SLOTTEDPAGE		*pagePtr;
					LARGEOBJROOT	*rootPtr;


					TRPRINT(TR_LG, TR_LEVEL_2, ("slot root pid:%d slot:%d",
									nodePid.page, rootSlot));

					/*
					 *	Get a pointer to the page, slot and object
					 */
					pagePtr = (SLOTTEDPAGE *) nodeLink->bufFrame;
					slotPtr = GETSLOTPTR(pagePtr, rootSlot);
					rootPtr = (LARGEOBJROOT *) GETOBJECTPTR(pagePtr, slotPtr);

					/*
					 *	get a pointer to the node
					 */
					nodePtr = (LGNODE *) &(rootPtr->rootPage);

					/*
					 *	check to see if the slot is the right type
					 */
					CHECK_SLOTROOT_MAGIC(nodePtr);

				} else {

					/*
					 *	the node pointer is just the page pointer
					 */
					nodePtr = (LGNODE *) nodeLink->bufFrame;
					lg_PrintNode(nodePtr, 0, nodePid.page);

					/*
					 *	check the magic number
					 */
					CHECK_LARGENODE_MAGIC(nodePtr);
				}

				/*
				 *	get the height of the current node
				 */
				height = nodePtr->header.height;

				/*
				 *	find out which slot to traverse
				 */
				travSlot = LG_BinarySearch(nodePtr, *offset);

				/*
				 *	Push the node traversal information onto the stack
				 */
				LG_PushTstack(stack, travSlot, &nodePid, nodePtr,
							  FALSE, 0, nodeLink, rootSlot, shared);
				TRPRINT(TR_LG, TR_LEVEL_2, ("height:%d slot:%d rootSlot:%d",
						height, travSlot, rootSlot));

				/*
				 *  Count the number of shared nodes
				 */
				if (shared && versioned) (*sharedCount)++;

				/*
				 * Since we're inserting, update the slot counts on the way down.
				 * (Note that the updated pages will be fixed/locked while the
				 * operation takes place, so it's okay to do this.)
				 *
				 *	Convert to pointers - DMH
				 * 
				 *	Do not update slots on shared pages in versioned
				 *	objects or at a height > the updateHeight.
				 */
				if ( !(shared && versioned) && (numBytes > 0) && 
					 (height <= updateHeight) ) {

					if (GET_LOGLEVEL(LOG_SPACE)) {
						/* log the change to the slot byte counts*/
						opInfo.action = LOG_ACTION_INCREMENT_SLOT_BYTECOUNTS;
						opInfo.flags = 0;
						opInfo.start = travSlot;
						opInfo.size = numBytes;
						opInfo.oldData = NULL;
						opInfo.newData = NULL;
						opInfo.actionPid = nodePid;
						opInfo.rootSlot = rootSlot;
						if(clog_LogChangeSlots(&opInfo)) {
							FAIL
						}
					}	
					/*
					 *	Ok to update the slot counts
					 */
					for (slotIndex = travSlot; 
						 slotIndex < nodePtr->header.numSlots; 
						 slotIndex++) 	{

						nodePtr->slot[slotIndex].count += numBytes;
					}
				}

				/* Nil out the root slot for subsequent levels */
				rootSlot = NIL;

				/* Set the offset relative to the subtree we're about to enter.  */
				if (travSlot > 0) {
					*offset -= (int) nodePtr->slot[travSlot - 1].count;
				}

				/* get the pid of the next subtree to process */
				nodePid = nodePtr->slot[travSlot].pid;

				/* See if the next subtree is shared */
				shared = shared | nodePtr->slot[travSlot].shared;

			} while (height > 1);

			break;

		/*
		 *	Do an append oriented traversal.  There must always be
		 *	at least two levels in the tree - a root and a leaf,
		 *	even if the object is empty.
		 */
		case APPEND:

			TRPRINT(TR_LG, TR_LEVEL_2, ("append"));

			do {

				/*
				 *	read the node into a buffer
				 */
				if ((nodeLink = bf_ReadPage(bufGroup, &nodePid,
					(rootSlot != NIL ? H_SLOTTED : H_LARGE),
					(rootSlot != NIL ? SLOTTED_PAGE2SIZE : LG_PAGE2SIZE), NL)) == NULL)	{

					SM_ERROR(TYPE_FATAL, esmINTERNAL);
				}

				/* assume we are not working on a new slotted root obj 
														until proven otherwise*/
				newSlotRoot = FALSE;

				/*
				 *	get the pointer to the node
				 */
				if (rootSlot != NIL)	{

					PAGESLOT		*slotPtr;
					SLOTTEDPAGE		*pagePtr;
					LARGEOBJROOT	*rootPtr;


					TRPRINT(TR_LG, TR_LEVEL_2, ("slot root pid:%d slot:%d",
							nodePid.page, rootSlot));

					/* Get a pointer to the page, slot and object */
					pagePtr = (SLOTTEDPAGE *) nodeLink->bufFrame;
					slotPtr = GETSLOTPTR(pagePtr, rootSlot);
					rootPtr = (LARGEOBJROOT *) GETOBJECTPTR(pagePtr, slotPtr);

					/* see if we are creating a new slotted root object */
					if (rootPtr->objectHdr.properties & P_CREATEINPROGRESS) {
						newSlotRoot = TRUE;
					}


					/*
					 *	get a pointer to the node
					 */
					nodePtr = (LGNODE *) &(rootPtr->rootPage);

					/*
					 *	check to see if the slot is the right type
					 */
					CHECK_SLOTROOT_MAGIC(nodePtr);

				} else {

					/*
					 *	the node pointer is just the page pointer
					 */
					nodePtr = (LGNODE *) nodeLink->bufFrame;
					lg_PrintNode(nodePtr, 0, nodePid.page);

					/*
					 *	check the magic number
					 */
					CHECK_LARGENODE_MAGIC(nodePtr);
				}

				/*
				 *	get the height of the node
				 */
				height = nodePtr->header.height;

				/*
				 *	print the node
				 */
				lg_PrintNode(nodePtr, height, nodePid.page);

				/*
				 * take the last slot in the node for appending
				 */
				travSlot = (int)nodePtr->header.numSlots - 1;

				/*
				 *	push the traversal information onto the stack
				 */
				LG_PushTstack(stack, travSlot, &nodePid, nodePtr,
							  FALSE, 0, nodeLink, rootSlot, shared);
				TRPRINT(TR_LG, TR_LEVEL_2, ("height:%d slot:%d count:%d rootSlot:%d",
						height, travSlot,
						nodePtr->slot[travSlot].count, rootSlot));

				/* Count the number of shared nodes */
				if (shared && versioned) (*sharedCount)++;

				/*
				 *	Update the count as we go since we'll be adding new bytes.
				 *	Do not update slots on shared pages in versioned
				 *  objects.
				 */
				if ( !(shared && versioned) && (numBytes > 0) && 
					 (height <= updateHeight) ) {

					/* log the change to the slot byte counts except
					   if we are in the process of creating a new object
					   with a slotted root - in this case, logging is
					   handled by the create code */
					if (newSlotRoot == FALSE && GET_LOGLEVEL(LOG_SPACE)) {
        			    opInfo.action = LOG_ACTION_INCREMENT_SLOT_BYTECOUNTS;
        			    opInfo.flags = 0;
        			    opInfo.start = travSlot;
        		    	opInfo.size = numBytes;
        		    	opInfo.oldData = NULL;
        		    	opInfo.newData = NULL;
        		    	opInfo.actionPid = nodePid;
        		    	opInfo.rootSlot = rootSlot;
						if(clog_LogChangeSlots(&opInfo)) {
							FAIL
						}
					}

					nodePtr->slot[travSlot].count += numBytes;
				}

				/* Nil out the root slot for subsequent levels */
				rootSlot = NIL;

				/* get the pid of the next level page */
				nodePid = nodePtr->slot[travSlot].pid;

				/* See if the next subtree is shared */
				shared = shared | nodePtr->slot[travSlot].shared;

			} while (height > 1);
			break;
	
		/* this cannot happen */
		default:

			SM_ERROR(TYPE_FATAL, esmINTERNAL);
			break;
    }
	return esmNOERROR;
FailureReturn:
	SM_ERROR(TYPE_FATAL, esmINTERNAL);
	return esmFAILURE; /* to keep compiler quiet */
}
