/*
 *   $RCSfile: LG_ZipUpIndex.c,v $  
 *   $Revision: 1.1.1.1 $  
 *   $Date: 1996/05/04 21:55:24 $      
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/


#include <stdio.h>
#include "ess.h"
#include "error.h"
#include "trace.h"
#include "checking.h"
#include "io.h"
#include "page.h"
#include "list.h"
#include "object.h"
#include "tid.h"
#include "bf.h"
#include "chunk.h"
#include "lock.h"
#include "bf_macro.h"
#include "lgobject.h"
#include "volume.h"
#include "logaction.h"
#include "clog.h"
#include "sm_state.h"
#include "io_extfuncs.h"
#include "bf_extfuncs.h"
#include "lg_intfuncs.h"
#include "lg_extfuncs.h"
#include "clog_extfuncs.h"
#include "sm_macro.h"


/*
 * This routine controls the second phase of deletion - during which the
 * index structure is brought back into line with the rules regarding
 * index node occupancy.  Using the left and right cut path information
 * produced during the first phase, it effectively "zips" up the nodes
 * along the two paths as necessary to ensure at least minimum fullness.
 * (Note:  It does not set previously fixed pages dirty when it unfixes
 * them, as this will already have been done during the deletion pass.)
 * 
 * Along the way, shared pages must be versioned (shadowed).  The
 * trick behind all of this is that any page on the stack has already
 * been versioned in LG_DeletePhase.  Only pages access by obtaining
 * their pid directly from a slot need to be checked for sharing.
 */


 int
LG_ZipUpIndex (

	FID				*fid,
	PID				*rootPid,		/* pid of object to delete data	*/
	BUFGROUP		*bufGroup,		/* buffer group to be used		*/
	TSTACK			*lstack,		/* stack for left cut path		*/
	TSTACK			*rstack,		/* stack for right cut path		*/
	BOOL			versioned 		/* true if object is verioned	*/
)
{

	PID				nodePid;		/* pid for current node of interest */
	LGNODE			*nodePtr;		/* indirect ptr to buffered current node */
	SLOTINDEX		nodeRootSlot;	/* slot if root is on slotted page*/
	int				numSlots;		/* total number of slots in current node */
	SLOTINDEX		slotIndex;		/* slot index for processing current node */
	int				dangerCount;	/* number of kids in danger for node */
	int				height;			/* height of current node 		*/

	PID				kid1Pid;		/* pid for 1st child node of interest */
	LGNODE			*kid1Ptr;		/* indirect ptr to buffered kid #1 */
	int				kid1numSlots;	/* (original) number of slots in kid #1 */
	int				kid1Slot;		/* slot index for kid #1's path */
	BOOL			kid1Danger;		/* danger flag for kid #1 		*/
	int				kid1DangerCount;/* danger count for kid #1 		*/
	SLOTINDEX		kid1SlotRoot;	/* slot if root is on slotted page*/

	PID				kid2Pid;		/* pid for 2nd child node of interest */
	LGNODE			*kid2Ptr;		/* indirect ptr to buffered kid #2*/
	int				kid2numSlots;	/* (original) number of slots in kid #2 */
	int				kid2Slot;		/* slot index for kid #2's path */
	BOOL			kid2Danger;		/* danger flag for kid #2 		*/
	int				kid2DangerCount;/* danger count for kid #2 		*/
	SLOTINDEX		kid2SlotRoot;	/* slot if root is on slotted page*/

	PID				savePid;		/* pid for saved LCA's right child node */
	LGNODE			*savePtr;		/* indirect ptr to buffered saved node */
	int				saveSlot;		/* slot index for saved node's path */
	int				kid2saved;		/* true if kid2 info currently saved here */
	GROUPLINK		*kid1Link;
	GROUPLINK		*kid2Link;
	GROUPLINK		*rootLink;
	GROUPLINK		*saveLink;

	PATHTYPE		whichPath;		/* which portion of cut path node is on */
	BOOL			nodeIsLCA;		/* true if node is LCA for cut path */
	BOOL			emptyStacks;	/* true if both path stacks are empty */
	BOOL			workRemains;	/* true unless rebalancing is finished */
	int				whichKid;		/* index of appropriate neighboring kid */
	PID				shadowPid;		/* pid for shadowing purposes 	*/
	int				moveCount;		/* # child slots moved left->right */
	WHICHNODE		nextNode;		/* next node to be processed along path */
	LASTSTEP		lastOp;			/* indicates operation performed last */
	TSTACKENT		*stackTop;
	GROUPLINK		*nodeLink;
	OPERATIONINFO	opInfo;
	LGSLOT			tempSlot;
	BOOL			shared;
	SLOTTEDPAGE		*slottedPage;
	OBJECT			*objPtr;
	PAGESLOT		*slotPtr;

	TRACE(TR_LG, TR_LEVEL_1);

    /*
     * 	Get the root node off of the stack
	 */
    if ((stackTop = LG_PopTstack(lstack)) == NULL)	{

		/*
		 *	die for now
		 */
		SM_ERROR(TYPE_FATAL, esmINTERNAL);
	}

	/*
	 *	pull out the values
	 */
	slotIndex	= stackTop->slotIndex;
	nodePid		= stackTop->pid;
	dangerCount = stackTop->dangerCount;
	nodeLink	= stackTop->nodeLink;
	nodeRootSlot= stackTop->rootSlot;
	if (nodeRootSlot == NIL) {
		nodePtr		= (LGNODE *) nodeLink->bufFrame;
		CHECK_LARGENODE_MAGIC(nodePtr);
	} else {
		/*
		 *	Root on slotted page
		 */
		slottedPage = (SLOTTEDPAGE *) nodeLink->bufFrame;
		SM_ASSERT(LEVEL_3, !CHECK_SLOTTED_MAGIC(slottedPage));
		slotPtr = GETSLOTPTR(slottedPage, nodeRootSlot);
		objPtr = GETOBJECTPTR(slottedPage, slotPtr);
		nodePtr = (LGNODE*) objPtr->data;
		CHECK_SLOTROOT_MAGIC(nodePtr);
	}
    numSlots	= (int)nodePtr->header.numSlots;

	TRPRINT(TR_LG, TR_LEVEL_2, ("slotIndex:%d nodePid:%d dangerCount:%d",
			slotIndex, nodePid.page, dangerCount));

	/*
	 *	save a pointer to the root
	 */
	rootLink = nodeLink;

	/*
	 *	Check to see if the left cut path stack is empty
	 */
    if (LG_EmptyTstack(lstack)) {

		/*
		 *	Access through the right stack
		 */
		whichPath = RIGHTPATH;
		
		/*
		 *	Check to see if the right path is empty
		 */
		emptyStacks = LG_EmptyTstack(rstack);
		TRPRINT(TR_LG, TR_LEVEL_2, ("initial path right, emptyStacks:%d", emptyStacks));

	} else {

		/*
		 *	switch to the left stack
		 */
		whichPath = LEFTPATH;
		emptyStacks = FALSE;
		TRPRINT(TR_LG, TR_LEVEL_2, ("initial path left"));
	}

	/*
	 *	set kid2danger to false to prevent iteration of
	 *	some parts of this loop
	 */
	kid2Danger = FALSE;

	/*
	 *	Determine whether the node is the least common ancestor
	 *	This can be done by examining that the pid in the slot
	 *	next to the traverse slot.
	 */
	nodeIsLCA = (whichPath == LEFTPATH) && (slotIndex < numSlots - 1) &&
				LG_IsTopPid(rstack, &(nodePtr->slot[slotIndex+1].pid));
	TRPRINT(TR_LG, TR_LEVEL_2, ("nodeIsLCA:%d", nodeIsLCA));

	/*
	 *	setup initial state variables
	 */
    workRemains = TRUE;
    kid2saved = FALSE;

    /*
     * Okay, now traverse the left and right cut paths (top down, in
     * that order) to zip up the index.
     */
    while (workRemains) {

		/*
		 *	Check to see if we are at the leaf level
		 */
		if (nodePtr->header.height == 1) {

			/*
			 * Need to merge/reshuffle children that are leaves, so call
			 * the routine that rebalances at parents of leaves if needed.
			 * (Note:  In the case of the LCA, there can be two almost
			 * empty leaves, so we might rebalance twice in this case.)
			 */
			if ((numSlots > 1) && (LG_SlotCount(nodePtr, slotIndex) < LGMINLEAF)) {

				if (LG_BalanceData(fid, bufGroup, &nodePid, nodeRootSlot,
								   nodePtr, &slotIndex, 
								   (nodePtr->slot[slotIndex].shared && versioned), 
								   (nodePtr->slot[slotIndex+1].shared && versioned) ))	{

					return(esmFAILURE);
				}
			}

			/*
			 *	reestablish the number of slots because it may
			 *	have changes as a result of the above balance
			 */
			numSlots = (int)nodePtr->header.numSlots;

			/*
			 *	rebalance the leaves again if necessary
			 */
			if ((numSlots > 1) && (LG_SlotCount(nodePtr, slotIndex) < LGMINLEAF)) {

				if (LG_BalanceData(fid, bufGroup, &nodePid, nodeRootSlot,
								   nodePtr, &slotIndex, 
								   nodePtr->slot[slotIndex].shared && versioned, 
								   nodePtr->slot[slotIndex+1].shared && versioned))	{

					return(esmFAILURE);
				}
			}

			/*
			 *	set the state for the next operation
			 */
			lastOp = LEAFSTUFF;
			nextNode = NONE;

		} else {    /* (nodePtr->header.height > 1) */

			TRPRINT(TR_LG, TR_LEVEL_2, ("nodeCount:%d height:%d",
					numSlots, nodePtr->header.height));

			/*
			 *	Need to merge/reshuffle children that are internal nodes.
			 *	Start by getting the next child node from one of the stacks.
			 *	Then we must distinguish the 1-child-root case from more
			 *	general conditions, handling them (quite) differently.
			 */
			if (whichPath == LEFTPATH) {

				/*
				 *	get the next element off the left cut path
				 */
				if ((stackTop = LG_PopTstack(lstack)) == NULL)	{

					SM_ERROR(TYPE_FATAL, esmINTERNAL);
				}

				/*
				 *	get values from the top of the stack
				 */
				kid1Slot		= stackTop->slotIndex;
				kid1Pid			= stackTop->pid;
				kid1Danger		= stackTop->dangerFlag;
				kid1DangerCount = stackTop->dangerCount;
				kid1Link		= stackTop->nodeLink;
				kid1SlotRoot	= stackTop->rootSlot;
				kid1Ptr			= (LGNODE *) kid1Link->bufFrame;
				kid1numSlots	= (int)kid1Ptr->header.numSlots;
				TRPRINT(TR_LG, TR_LEVEL_2, ("popped kid1:%d", kid1Pid.page));

				/*
				 *	check to see that the pid stored matches the pid in the slot
				 */
				SM_ASSERT(LEVEL_3, PIDEQ(kid1Pid, nodePtr->slot[slotIndex].pid));

				/*
				 *	check to see if this node is the LCA
				 *	if so we must merge the two cut path siblings.
				 */
				if (nodeIsLCA) {

					/*
					 *	get the next element off the right cut path
					 */
					if ((stackTop = LG_PopTstack(rstack)) == NULL)	{

						SM_ERROR(TYPE_FATAL, esmINTERNAL);
					}

					/*
					 *	get values from the top of the stack
					 */
					kid2Slot		= stackTop->slotIndex;
					kid2Pid			= stackTop->pid;
					kid2Danger		= stackTop->dangerFlag;
					kid2DangerCount = stackTop->dangerCount;
					kid2Link		= stackTop->nodeLink;
					kid2SlotRoot	= stackTop->rootSlot;
					kid2Ptr			= (LGNODE *) kid2Link->bufFrame;
					kid2numSlots	= (int)kid2Ptr->header.numSlots;
					TRPRINT(TR_LG, TR_LEVEL_2, ("popped kid2:%d", kid2Pid.page));
				}

			} else if (!emptyStacks) {

				/*
				 *	get the next element off the right cut path
				 */
				if ((stackTop = LG_PopTstack(rstack)) == NULL)	{

					SM_ERROR(TYPE_FATAL, esmINTERNAL);
				}

				/*
				 *	get values from the top of the stack
				 */
				kid1Slot		= stackTop->slotIndex;
				kid1Pid			= stackTop->pid;
				kid1Danger		= stackTop->dangerFlag;
				kid1DangerCount = stackTop->dangerCount;
				kid1Link		= stackTop->nodeLink;
				kid1SlotRoot	= stackTop->rootSlot;
				kid1Ptr			= (LGNODE *) kid1Link->bufFrame;
				kid1numSlots	= (int)kid1Ptr->header.numSlots;
				TRPRINT(TR_LG, TR_LEVEL_2, ("popped kid1:%d", kid1Pid.page));

				/*
				 *	check to see that the pid stored matches the pid in the slot
				 */
				SM_ASSERT(LEVEL_3, PIDEQ(kid1Pid, nodePtr->slot[slotIndex].pid));

				/*
				 *	the path to right
				 */
				whichPath = RIGHTPATH;

			} else {

				/*
				 *	stacks should not be empty
				 */
				SM_ERROR(TYPE_FATAL, esmINTERNAL);
			}

			/*
			 *	check to see if we have a single slot
			 */
			if (numSlots == 1) {

				/*
				 * Okay, this is definitely a root-only case.  Handle it by
				 * making the root's only child the new root node;  do this
				 * by copying the contents of the child into the root node's
				 * buffer page.  Of course, if the child was not on one of
			 	 * the two stacks, we must start by reading it in from disk.
				 */
				if (emptyStacks) {

					TRPRINT(TR_LG, TR_LEVEL_2, ("read in new root"));

					/*
					 *	get the pid of the page to read
					 */
					kid1Pid = nodePtr->slot[0].pid;
					shared = nodePtr->slot[0].shared;
					kid1SlotRoot = NIL;

					/*
					 *	read it in from disk
					 */
					if ((kid1Link = bf_ReadPage(bufGroup, &kid1Pid,
							H_LARGE, LG_PAGE2SIZE, NL)) == NULL)	{

						return(esmFAILURE);
					}

					/*
					 *	get a pointer to the page
					 */
					kid1Ptr = (LGNODE *) kid1Link->bufFrame;

					/*
					 *	set the conditions for copying
					 */
					kid1Slot = -1;
					kid1numSlots = (int)kid1Ptr->header.numSlots;
					kid1Danger = FALSE;
					kid1DangerCount = 0;
				}

				/*
				 *	Log the copying of the child into the root.
				 *	This is logged as removing the root slots
				 *	and apeending the childs.  This cannot be
				 *	performed on a root in a slotted page.
				 */
				SM_ASSERT(LEVEL_3, nodeRootSlot == NIL && !PPIDEQ(&nodePid, rootPid));
				if (GET_LOGLEVEL(LOG_SPACE)) {
					opInfo.action = LOG_ACTION_REMOVE_LARGE_SLOTS;
					opInfo.flags = 0;
					opInfo.start = 0;
					opInfo.size = ((int)sizeof(LGSLOT) * (int)nodePtr->header.numSlots);
					opInfo.oldData = (char *) (nodePtr->slot);
					opInfo.newData = NULL;
					opInfo.actionPid = nodePid;
					opInfo.rootSlot = nodeRootSlot;
					if(clog_LogModifyLgPage(&opInfo)) {
						FAIL
					}

					/* log the append of the slots to the node*/
					opInfo.action = LOG_ACTION_APPEND_LARGE_SLOTS;
					opInfo.flags = 0;
					opInfo.start = 0;
					opInfo.size = (int)sizeof(LGSLOT) * (int)kid1Ptr->header.numSlots;
					opInfo.oldData = NULL;
					opInfo.newData= (char *) &(kid1Ptr->slot[0]);
					opInfo.actionPid = nodePid;
					/* this can't be a slotted root object */
					opInfo.rootSlot = nodeRootSlot;
					if(clog_LogModifyLgPage(&opInfo)){
						FAIL
					}
				}

				/* copy the child into the root */
				*nodePtr = *kid1Ptr;

				/*
 				 *	Discard the kid if it is not shared by another 
				 *	version.
				 */
				if (!versioned || !shared) {

					/* unfix the page and discard it */
					bf_DiscardPage(bufGroup, kid1Link);

					/* deallocate the page */
					TRPRINT(TR_LG, TR_LEVEL_2, ("discard and deallocating:%d", kid1Pid.page));
					if (io_DeallocPages(LG_PAGE2SIZE, 1, &kid1Pid, 
									FALSE, PAGE_LARGENODE))	{ 
						return(esmFAILURE);
					}
				} else {
					/* unfix the page , turn off the shared flag */
					bf_UnfixPage(kid1Link, BF_DEFAULT, FALSE);
					shared = FALSE;
				}

				/* next node is current node */
				nextNode	= CURRENT;
				slotIndex	= kid1Slot;
				numSlots	= kid1numSlots;
				dangerCount = kid1DangerCount;
				nodeLink	= kid1Link;
				nodeRootSlot= kid1SlotRoot;
		
			} else if (!emptyStacks) {
		
				/*
				 * This case applies to the root and/or other non-leaf nodes.
				 * Merge/reshuffle the node's one or two endangered children.
				 */
				if (slotIndex == numSlots - 1) {

					/* take the slot to the left */
					whichKid = slotIndex - 1;

				} else {

					/*
					 *	take the slot to the right, if this is
					 *	the LCA then we are guaranteed to
					 *	get the right cut path sibling, which we
					 *	have already gotten off the stack
					 */
					whichKid = slotIndex + 1;
				}
				TRPRINT(TR_LG, TR_LEVEL_2, ("whichKid:%d", whichKid));
		
				/*
				 *	Check to see if the child is in danger
				 *	and the node is not the LCA.  If the node
				 *	is not the LCA then we are guaranteed that
				 *	a single merge/reshuffle operation will suffice.
				 */
				if (kid1Danger && (!nodeIsLCA)) {

					TRPRINT(TR_LG, TR_LEVEL_2, ("child1 danger and no LCA"));

					/*
					 *	get the pid of the adjacent sibling by definition
					 *	we must be able to merge/shuffle with that sibling
					 *	This could be modified to check both sides (if
					 *	possible) and do a merge if possible.
					 */
					kid2Pid = nodePtr->slot[whichKid].pid;
					kid2SlotRoot = NIL;

					/*
					 *	read in the sibling
					 */
					if ((kid2Link = bf_ReadPage(bufGroup, &kid2Pid,
								H_LARGE, LG_PAGE2SIZE, NL)) == NULL)	{

						return(esmFAILURE);
					}

					/*	
					 *	Shadow the sibling if it is shared by another
					 *	version, since it will be changed
					 */
					if (versioned && nodePtr->slot[whichKid].shared) {
						if (LG_Shadow(&kid2Pid, &shadowPid, rootPid,
									  bufGroup, &kid2Link))	{
							
							return(esmFAILURE);
						}

						/* Record the new shadow pid and log the slot change */
						tempSlot = nodePtr->slot[whichKid]; 
						nodePtr->slot[whichKid].pid = shadowPid;
						nodePtr->slot[whichKid].shared = FALSE;
						if (GET_LOGLEVEL(LOG_SPACE)) {
							opInfo.action = LOG_ACTION_WRITE_LARGE_SLOTS;
							opInfo.flags = 0;
							opInfo.start = whichKid * (int)sizeof(LGSLOT);
							opInfo.size = 1 * (int)sizeof(LGSLOT);
							opInfo.oldData = (char*) &tempSlot;
							opInfo.newData = (char*) &(nodePtr->slot[whichKid]);
							opInfo.actionPid = shadowPid;
							/* mark if it is a root object on a slotted page */
							opInfo.rootSlot = NIL;
							if(clog_LogWriteLgPage(&opInfo)) {
								FAIL
							}
						}
						kid2Pid = shadowPid;
					}

					/*
					 *	initialize the kid2 stuff
					 */
					kid2Ptr = (LGNODE *) kid2Link->bufFrame;
					kid2numSlots = (int)kid2Ptr->header.numSlots;
					kid2Danger = FALSE;
					kid2DangerCount = 0;
					kid2SlotRoot = NIL;

				}
		
				/*
				 *	check to see if either of the children are in danger
				 */
				if (kid1Danger || kid2Danger) {

					TRPRINT(TR_LG, TR_LEVEL_2, ("one of the kids in danger:%d %d",
							kid1Danger, kid2Danger));

					/*
					 *	Check to see if we can merge the nodes
					 */
					if ((kid1numSlots + kid2numSlots) > LGMAXSLOTS) {

						TRPRINT(TR_LG, TR_LEVEL_2, ("cannot merge"));

						/*
						 * Reshuffle contents of kid1 and kid2 to equalize
						 * their occupancies.
						 */
						if (whichKid == slotIndex - 1) {

							/* reshuffle from kid2 to kid1 */
							LG_Reshuffle(&nodePid, nodeRootSlot, nodePtr, whichKid, &kid2Pid, kid2Ptr, &kid1Pid, kid1Ptr, &moveCount);

							/*
							 *	Note: This is definitely not the LCA case,
							 *	as slotIndex-1 would not be on the cut path.
							 */
							if (kid1Slot < moveCount) {

								/*
								 *	kid2 is next
								 */
								nextNode = KID2;
								kid2Slot = kid2numSlots + kid1Slot;
								kid2DangerCount = kid1DangerCount;
								bf_UnfixPage(kid1Link, BF_DEFAULT,  FALSE);

							} else {

								/*
								 *	kid1 is next
								 */
								nextNode = KID1;
								kid1Slot = kid1Slot - moveCount;
								bf_UnfixPage(kid2Link, BF_DEFAULT, FALSE);
							}

						} else {

							/*
							 *	reshuffle data from kid1 to kid2
							 */
							LG_Reshuffle(&nodePid, nodeRootSlot, nodePtr, slotIndex, &kid1Pid, kid1Ptr, &kid2Pid, kid2Ptr, &moveCount);

							/*
							 *	check to see which kid will be processed next
							 */
							if (kid1Slot < kid1numSlots + moveCount) {

								/* kid1 is next */
								nextNode = KID1;

								/*
								 * Current node was and is the LCA - save
								 * kid2 for later as top of right cut path.
								 */
								if ((nodeIsLCA) && (kid2Slot >= moveCount)) {

									/*
									 *	node still LCA with kid1 and kid2
									 */
									savePid		 = kid2Pid;
									savePtr		 = kid2Ptr;
									saveSlot	 = kid2Slot - moveCount;
									saveLink	 = kid2Link;
									kid2saved	 = TRUE;

								} else {

									/* we are done with kid2 */
									bf_UnfixPage(kid2Link, BF_DEFAULT, FALSE);
									nodeIsLCA = FALSE;
								}

							} else {

								/* kid2 is next */
								nextNode = KID2;
								kid2Slot = kid1Slot - kid1numSlots - moveCount;
								bf_UnfixPage(kid1Link, BF_DEFAULT, FALSE);
							}
						}

						/* record the reshuffling as the last operation */
						lastOp = RESHUFFLE;

					} else {

						TRPRINT(TR_LG, TR_LEVEL_2, ("merge nodes"));

						/*
						 * Merge the contents of kid1 and kid2 into a single
						 * node.  Note:  In the case of the LCA, this may not
						 * suffice to ensure the safety of the resulting node.
						 * If not, we have to make the next node the current
						 * node, and we also have to push the next kid back
						 * on lstack (where the next iteration will want to
						 * grab it from).
						 */
						if (nodeIsLCA && kid1Danger && kid2Danger) {

							/* we may not have eliminated the danger */
							lastOp = LCA_MERGE;
						} else {

							/* we have eliminated the danger */
							lastOp = MERGE;
						}
						TRPRINT(TR_LG, TR_LEVEL_2, ("lastOp:%d", lastOp));

						/*
						 *	check to see which child to make the
						 *	final merge target
						 */
						if (whichKid == (slotIndex - 1)) {

							TRPRINT(TR_LG, TR_LEVEL_2, ("merge kid1 into kid2"));

							/* merge the right child into the left */
							LG_Merge(&nodePid, nodeRootSlot, nodePtr, slotIndex - 1, &kid2Pid, kid2Ptr, &kid1Pid, kid1Ptr, &moveCount);
							bf_DiscardPage(bufGroup, kid1Link);

							if (io_DeallocPages(LG_PAGE2SIZE, 1, &kid1Pid, 
													FALSE, PAGE_LARGENODE))	{
								return(esmFAILURE);
							}
							TRPRINT(TR_LG, TR_LEVEL_2, ("deallocating page:%d",
									kid1Pid.page));

							/* kid2 is next */
							nextNode = KID2;
							kid2Slot = kid2numSlots + kid1Slot;
							kid2DangerCount += kid1DangerCount;

							/*
							 *	check to see if this is root
							 *	and now only has a single child
							 */
							if (nodePtr->header.numSlots == 1) {

								TRPRINT(TR_LG, TR_LEVEL_2, ("one entry node"));

								/* setup to process node again */
								nextNode = CURRENT;
								slotIndex = slotIndex - 1;

								/*
								 *	check to see which path we are traversing
								 *	and push onto that stack.
								 */
								if (whichPath == LEFTPATH) {

									TRPRINT(TR_LG, TR_LEVEL_2, ("pushing:%d onto left stack"));

									if (LG_PushTstack(lstack, kid2Slot, 
										&kid2Pid, kid2Ptr,
										kid2Danger, kid2DangerCount, 
											kid2Link, NIL, FALSE))	{

										return(esmFAILURE);
									}

								} else {

									TRPRINT(TR_LG, TR_LEVEL_2, ("pushing:%d onto right stack"));

									if (LG_PushTstack(rstack, kid2Slot, &kid2Pid, kid2Ptr,
										kid2Danger, kid2DangerCount, kid2Link, NIL, FALSE))	{

										return(esmFAILURE);
									}

								}
							}

							/* set the global LCA flag */
							nodeIsLCA = FALSE;

						} else {

							TRPRINT(TR_LG, TR_LEVEL_2, ("merge kid2 into kid1"));

							/* merge the left child into the right */
							LG_Merge(&nodePid, nodeRootSlot, nodePtr, slotIndex, &kid1Pid, kid1Ptr, &kid2Pid, kid2Ptr, &moveCount);

							bf_DiscardPage(bufGroup, kid2Link);

							if (io_DeallocPages(LG_PAGE2SIZE, 1, &kid2Pid, 
										FALSE, PAGE_LARGENODE))	{
								return(esmFAILURE);
							}
							TRPRINT(TR_LG, TR_LEVEL_2, ("deallocating page:%d",
									kid2Pid.page));

							/* kid1 is next unless danger persists */
							nextNode = KID1;
							kid1DangerCount += kid2DangerCount;

							/* check to see if danger persists */
							if (((int)nodePtr->header.numSlots == 1) ||
								(((int)kid1Ptr->header.numSlots - kid1DangerCount) < LGMINSLOTS)) {

								TRPRINT(TR_LG, TR_LEVEL_2, ("danger persists:%d",
										(int)nodePtr->header.numSlots));

								/* we must process the current node again */
								nextNode = CURRENT;

								/*
								 *	check which path we are traversing
								 *	and push onto that stack
								 */
								if (whichPath == LEFTPATH) {

									TRPRINT(TR_LG, TR_LEVEL_2, ("pushing onto left stack"));

									if (LG_PushTstack(lstack, kid1Slot, &kid1Pid, kid1Ptr,
										kid1Danger, kid1DangerCount, kid1Link, NIL, FALSE))	{
										
										return(esmFAILURE);
									}

								} else {

									TRPRINT(TR_LG, TR_LEVEL_2, ("pushing onto right stack"));

									if (LG_PushTstack(rstack, kid1Slot, &kid1Pid, kid1Ptr,
										kid1Danger, kid1DangerCount, kid1Link, NIL, FALSE))	{
										
										return(esmFAILURE);
									}

								}
							}

							/* unmark the nodeIsLCA if it was set */
							nodeIsLCA = FALSE;
						}
					}
				} else {

					/* Proceed down the current cut path, */
					nextNode = KID1;

					/* saving the LCA for use in a later iteration if we're at it now.  */
					if (nodeIsLCA) {

						TRPRINT(TR_LG, TR_LEVEL_2, ("node is LCA, saveing kid2"));

						savePid		 = kid2Pid;
						savePtr		 = kid2Ptr;
						saveSlot	 = kid2Slot;
						saveLink	 = kid2Link;
						kid2saved	 = TRUE;
					}
				}
			} else {

				/* The stacks should not be empty */
				SM_ERROR(TYPE_FATAL, esmINTERNAL);
			}
		}
		
		/*
		 * Unfix the current node, unless it's the next one of interest
		 * or else the root (in which case we'll keep it fixed until the
		 * end of the delete operation).
		 */
		if ((nextNode != CURRENT) && (!PIDEQ(nodePid, *rootPid))) {
			bf_UnfixPage(nodeLink, BF_DEFAULT, FALSE);
		}

		/*
		 * If a next node was targeted, move its info into the variables
		 * for the current node for further analysis/use.
		 */
		if (nextNode == CURRENT) {

			TRPRINT(TR_LG, TR_LEVEL_2, ("nextNode = CURRENT"));

			numSlots = (int)nodePtr->header.numSlots;
			height	 = nodePtr->header.height;

		} else if (nextNode == KID1) {

			TRPRINT(TR_LG, TR_LEVEL_2, ("nextNode == KID1"));

			nodePid		= kid1Pid;
			nodePtr		= kid1Ptr;
			numSlots	= (int)kid1Ptr->header.numSlots;
			height		= kid1Ptr->header.height;
			slotIndex	= kid1Slot;
			nodeLink	= kid1Link;
			dangerCount = kid1DangerCount;

		} else if (nextNode == KID2) {

			TRPRINT(TR_LG, TR_LEVEL_2, ("nextNode == KID1"));

			nodePid		= kid2Pid;
			nodePtr		= kid2Ptr;
			numSlots	= (int)kid2Ptr->header.numSlots;
			height		= kid2Ptr->header.height;
			slotIndex	= kid2Slot;
			nodeLink	= kid2Link;
			dangerCount = kid2DangerCount;
		}
		nodeRootSlot = NIL; /* since on next level */

		/*
		 *	check to see which path we have been following
		 */
		if (whichPath == LEFTPATH) {

			TRPRINT(TR_LG, TR_LEVEL_2, ("left path"));

			/*
			 * Left cut path case.  Start by seeing if it's time to switch
			 * to the right cut path.
			 */
			if (nextNode == NONE) {

				TRPRINT(TR_LG, TR_LEVEL_2, ("switching to right path"));

				whichPath = RIGHTPATH;

			} else if (LG_EmptyTstack(lstack) && (dangerCount == 0)) {

				TRPRINT(TR_LG, TR_LEVEL_2, ("switching to right path"));

				whichPath = RIGHTPATH;
			}

			/*
			 * If so, see if further work remains.  If so, reinstate the
			 * saved kid2 info as info about the current node;  otherwise,
			 * we'll stay along the left path, so see if the next node is
			 * the LCA.  Either way, check for the emptyStacks condition,
			 * too.
			 */
			if (whichPath == RIGHTPATH) {

				TRPRINT(TR_LG, TR_LEVEL_2, ("right path"));

				if (kid2saved) {

					TRPRINT(TR_LG, TR_LEVEL_2, ("kid2saved"));

					if (nextNode == KID1) {

						/*
						 * Got here because kid1 was found to be safe.
						 * Thus, have to unfix kid1 before switching to
						 * kid2 for processing.
						 */
						bf_UnfixPage(kid1Link, BF_DEFAULT, FALSE);
					}

					/* setup saved pid */
					nodePid	  = savePid;
					nodePtr	  = savePtr;
					numSlots  = (int)savePtr->header.numSlots;
					height	  = savePtr->header.height;
					slotIndex = saveSlot;
					nodeLink  = saveLink;

				} else {

					workRemains = FALSE;
					TRPRINT(TR_LG, TR_LEVEL_2, ("terminating loop"));
				}

			} else {

				/* check to see if node is LCA */
				nodeIsLCA = (!kid2saved) && (slotIndex < numSlots - 1) &&
						  LG_IsTopPid(rstack, &(nodePtr->slot[slotIndex+1].pid));
				TRPRINT(TR_LG, TR_LEVEL_2, ("nodeIsLCA:%d", nodeIsLCA));
			}

			/* check to see if the stacks are empty */
			emptyStacks = LG_EmptyTstack(lstack) && LG_EmptyTstack(rstack);
			TRPRINT(TR_LG, TR_LEVEL_2, ("emptyStacks:%d", emptyStacks));

		} else {  /* whichPath == RIGHTPATH */

			TRPRINT(TR_LG, TR_LEVEL_2, ("right path"));

			/*
			 * Right cut path case.  See if work remains along this path
			 * or if we're finally done at this point.  Also see if stacks
			 * are now both empty.
			 */
			if (nextNode == NONE) {

				workRemains = FALSE;
			} else if (LG_EmptyTstack(rstack)) {

				emptyStacks = TRUE;
				workRemains = ((height == 1) && (dangerCount > 0));
			}
			TRPRINT(TR_LG, TR_LEVEL_2, ("workRemains:%d", workRemains));
		}

		/*
		 *	If no work remains, unfix the next node (if one was targeted)
		 *	before finishing up.
		 */
		if ((!workRemains) && (nextNode != NONE) && (!PIDEQ(nodePid, *rootPid))) {

			bf_UnfixPage(nodeLink, BF_DEFAULT, FALSE);
		}
    }

    /* Last but not least, unfix the root page and return.  */
    bf_UnfixPage(rootLink, BF_DEFAULT, FALSE);

	return(esmNOERROR);

FailureReturn:
	SM_ERROR(TYPE_FATAL, esmINTERNAL);
	return esmFAILURE;
}
