/*
 *   $RCSfile: lg_CreateObject.c,v $  
 *   $Revision: 1.1.1.1 $  
 *   $Date: 1996/05/04 21:55:23 $      
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/


#include <stdio.h>
#include "ess.h"
#include "trace.h"
#include "checking.h"
#include "error.h"
#include "io.h"
#include "page.h"
#include "list.h"
#include "object.h"
#include "tid.h"
#include "bf.h"
#include "bf_macro.h"
#include "chunk.h"
#include "lock.h"
#include "lsn.h"
#include "clog.h"
#include "logaction.h"
#include "loginfo.h"
#include "logrecs.h"
#include "sm_macro.h"
#include "lgobject.h"
#include "volume.h"
#include "version_graph.h"
#include "load.h"
#include "scan.h"
#include "sm_state.h"
#include "io_extfuncs.h"
#include "bf_extfuncs.h"
#include "lg_intfuncs.h"
#include "lg_extfuncs.h"
#include "clog_globals.h"
#include "clog_extfuncs.h"

#ifdef DEBUG
#ifdef __cplusplus
extern "C" sm_PrintSlotPage(SLOTTEDPAGE*);
#else
extern sm_PrintSlotPage(SLOTTEDPAGE*);
#endif
#endif
/*
 * Create a new large object, complete with initial contents.
 */


 int
lg_CreateObject (

	FID				*fid,			/* volume to put new object on	*/
	PID				*nearPid,		/* put new object's header near */
	BUFGROUP		*bufGroup,		/* buffer group to be used		*/
	int				length,			/* number of initial data bytes */
	char			*data,			/* ptr to initial data bytes	*/
	PID				*rootPid,		/* root pid for new object to	*/
	SLOTINDEX		rootSlot 
)
{

    PID				leafPid;		/* pid of content-free leaf page */
	GROUPLINK		*rootLink;
	OPERATIONINFO	opInfo;			/*structure to pass info to log routines*/


	TRPRINT(TR_LG, TR_LEVEL_1, ("length:%d, rootSlot:%d", length, rootSlot));

	/* 
	 io_AllocPages() calls this right away, and we
	 call io_AllocPages right away:
		sm_SetAbortIfError();
	*/

    /*
	 *	Check to see if the object is going to fit in a slotted page
     */
	if (rootSlot != NIL)	{

		SLOTTEDPAGE		*rootPage;
		LARGEOBJROOT	*largeRoot;
		PAGESLOT		*slotPtr;


		TRPRINT(TR_LG, TR_LEVEL_2, ("slotroot"));

		/*
		 *	create a dummy page to hold no data
		 */
		if (io_AllocPages(fid->pid.volid, LGDATA_PAGE2SIZE, 1,
						  nearPid, &leafPid, PAGE_LARGEDATA))	{

			return(esmFAILURE);
		}
		TRPRINT(TR_LG, TR_LEVEL_2, ("pid of data page:%d", leafPid.page));

		/*
		 *	Get a buffer for the first page
		 */
		if ((rootLink = bf_ReadPage(bufGroup, rootPid,
					H_SLOTTED, LG_PAGE2SIZE, NL)) == NULL)	{

			return(esmFAILURE);
		}

		/*
		 *	get a pointer to the page, slot and object
		 */
		rootPage  = (SLOTTEDPAGE *) rootLink->bufFrame;
		slotPtr   = GETSLOTPTR(rootPage, rootSlot);
		largeRoot = (LARGEOBJROOT *) GETOBJECTPTR(rootPage, slotPtr);

		/*
		 *	Initialize the root page
		 */
		INIT_SLOTROOT_MAGIC(&largeRoot->rootPage);
		largeRoot->rootPage.header.height	= 1;
		largeRoot->rootPage.header.numSlots	= 1;
		VHGPTR_SET_NULL(&(largeRoot->rootPage.header.vhgPtr));
		largeRoot->rootPage.slot[0].count	= 0;
		largeRoot->rootPage.slot[0].pid		= leafPid;
#ifdef DEBUG
		sm_PrintSlotPage(rootPage);
#endif	
		
        /* logging NOTE: the update to the slotted root object is not logged
			here because at this point, the update is incomplete (the object
			may not even be in a valid state at this point).
		    The caller must insure that the slotted root object is logged
		 */

		/*
		 *	Unfix the slotted page
		 */
		bf_UnfixPage(rootLink, BF_DEFAULT, TRUE);

	} else {

		LGNODE			*rootPtr;		/* pointer to root node in buffer */


		TRPRINT(TR_LG, TR_LEVEL_2, ("seperate root"));

		/*
		 *	Create a root pid
		 */
		if (io_AllocPages(fid->pid.volid, LG_PAGE2SIZE, 1,
						  nearPid, rootPid, PAGE_LARGENODE))	{

			return(esmFAILURE);
		}
		TRPRINT(TR_LG, TR_LEVEL_2, ("pid of large root:%d", rootPid->page));

		/*
		 *	create a dummy page to hold no data near the root
		 */
		if (io_AllocPages(fid->pid.volid, LGDATA_PAGE2SIZE, 1,
						  rootPid, &leafPid, PAGE_LARGEDATA))	{

			return(esmFAILURE);
		}
		TRPRINT(TR_LG, TR_LEVEL_2, ("pid of data page:%d", leafPid.page));

		/*
		 *	Get a buffer for the first page
		 */
		if ((rootLink = bf_GetPage(bufGroup, rootPid, H_LARGE, NULL,
								   LG_PAGE2SIZE)) == NULL)	{

			return(esmFAILURE);
		}

		/*
		 *	get a pointer to the root page
		 */
		rootPtr = (LGNODE *) rootLink->bufFrame;	

		/*
		 *	Initialize the root page
		 */
		INIT_LARGENODE_MAGIC(rootPtr);
		rootPtr->header.lrc = *(GET_LOGINFO->initLRC);
		rootPtr->header.height	 = 1;
		rootPtr->header.numSlots = 1;
		VHGPTR_SET_NULL(&(rootPtr->header.vhgPtr));
		rootPtr->slot[0].count	 = 0;
		rootPtr->slot[0].pid	 = leafPid;


		if (GET_LOGLEVEL(LOG_SPACE)) {
			/* log the init */
			opInfo.action = LOG_ACTION_INIT_LARGE_NODE;
			opInfo.flags = 0;
			opInfo.start = 0;
			opInfo.size = sizeof(LGNODEHDR);
			opInfo.oldData = NULL;
			opInfo.newData = (char*) &(rootPtr->header);
			opInfo.actionPid = *rootPid;
			opInfo.rootSlot = NIL;
			if(clog_LogWriteLgNodeHdr(&opInfo)) {
				FAIL
			}
		}
		
		/* Unfix the root page to prevent interfering with the insert routines */
		bf_UnfixPage(rootLink, BF_DEFAULT, TRUE);
	}

    /* Now append any desired data to it.  */
    if (length > 0) {

		/* pass in the slot of the page or NIL */
		if (lg_AppendToObject(fid, rootPid, bufGroup, length, data, rootSlot, FALSE))	{
			return(esmFAILURE);
		}
    }

	/* return success */
    return(esmNOERROR);

FailureReturn:
	return esmFAILURE;
}

