/*
 *   $RCSfile: lg_CreateVersionRoot.c,v $  
 *   $Revision: 1.1.1.1 $  
 *   $Date: 1996/05/04 21:55:24 $      
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/


#include <stdio.h>
#include "ess.h"
#include "trace.h"
#include "checking.h"
#include "error.h"
#include "io.h"
#include "page.h"
#include "list.h"
#include "object.h"
#include "tid.h"
#include "bf.h"
#include "bf_macro.h"
#include "chunk.h"
#include "lock.h"
#include "sm_macro.h"
#include "lgobject.h"
#include "volume.h"
#include "load.h"
#include "scan.h"
#include "pool.h"
#include "lsn.h"
#include "loginfo.h"
#include "logaction.h"
#include "clog.h"
#include "sm_state.h"
#include "io_extfuncs.h"
#include "bf_extfuncs.h"
#include "version_graph.h"
#include "lg_versionfuncs.h"
#include "clog_extfuncs.h"
#include "clog_globals.h"

/*
 * Create a new version of a the root node of a large object.
 */

 int
lg_CreateVersionRoot (
	BUFGROUP	*bufGroup,
	FID			*fid,			/* file containing the version		*/
	PID			*nearPid,		/* hint for location of new root 	*/
	LARGEOBJHDR	*newVersionHdr,
	LARGEOBJHDR	*parentVersionHdr,
	VHGNODEID	nodeId 			/* VHG node for the new version		*/

)
{

    PID				rootPid;		/* pid of the new root page */
	GROUPLINK		*rootLink;
	GROUPLINK		*parentRootLink;/* parent version root page */
	LGNODE			*rootPtr;		/* pointer to root node in buffer */
	LGNODE			*parentRootPtr;	/* pointer to root node in buffer */
	int				i;				/* for loop index */
	OPERATIONINFO	opInfo;

	TRPRINT(TR_VERSION, TR_LEVEL_1, ("version node id %d", nodeId));

	/*
	 *	The Plan:
	 *	1. Get a pointer to the root node of the parent version
	 *	2. Get a new page for the root node of the new version
	 *	3. Copy the old root node to the new one and complete the
	 *	   initialization of the new root
	 *	4. Set the rootPid of the newVersionHdr
	 */

	/*
	 *	create a root pid
	 */
	rootPid.page 	= parentVersionHdr->rootPid;
	rootPid.volid	= fid->pid.volid;

	/*
	 *	Pin the parent verison root page
	 */
	if ((parentRootLink = 
		 bf_ReadPage(bufGroup, &rootPid, H_LARGE, 
					 LG_PAGE2SIZE, NL)) == NULL)  {
		return(esmFAILURE);
	}

	/*
	 *	get a pointer to the parent root page
	 */
	parentRootPtr = (LGNODE *) parentRootLink->bufFrame;
	CHECK_LARGENODE_MAGIC(parentRootPtr);

	/*
	 *	Create a new root pid
	 */
	if (io_AllocPages(fid->pid.volid, LG_PAGE2SIZE, 1,
					  nearPid, &rootPid, PAGE_LARGENODE))	{

		return(esmFAILURE);
	}

	TRPRINT(TR_LG, TR_LEVEL_2, ("pid of large root:%d", rootPid.page));

	/*
	 *	Get a buffer for the root page
	 */
	if ((rootLink = bf_GetPage(bufGroup, &rootPid, H_LARGE, NULL,
							   LG_PAGE2SIZE)) == NULL)	{

		return(esmFAILURE);
	}

	/*
	 *	get a pointer to the root page
	 */
	rootPtr = (LGNODE *) rootLink->bufFrame;	

	/*
	 *	copy the parent root page to the new root page
	 *	But, the lrc should be initialized to 1 and
	 *	a record logged.
	 */
	*rootPtr = *parentRootPtr;		
	rootPtr->header.lrc = *(GET_LOGINFO->initLRC);

	/*
	 *	set the new version's VHG node pointer to the proper one
	 */
	rootPtr->header.vhgPtr.nodeId = nodeId;

	/*
	 *	Set the number of slots to be zero, so that the
	 *	following log record has the correct value.
	 */
	rootPtr->header.numSlots = 0;

	if (GET_LOGLEVEL(LOG_SPACE)) {
		/*
		 *	Log the initialization of the node header
		 */
		opInfo.action = LOG_ACTION_INIT_LARGE_NODE;
		opInfo.flags = 0;
		opInfo.start = 0;
		opInfo.size = sizeof(LGNODEHDR);
		opInfo.oldData = NULL;
		opInfo.newData = (char*) &(rootPtr->header);
		opInfo.actionPid = rootPid;
		/* this can't be a root on a slotted page */
		opInfo.rootSlot = NIL;
		if(clog_LogWriteLgNodeHdr(&opInfo)) {
			FAIL
		}
	}

	/* Restore the slot count */
	rootPtr->header.numSlots = parentRootPtr->header.numSlots;

	/* Mark the new root's pointers as shared */
	for (i = 0; i < rootPtr->header.numSlots; i++) {
		rootPtr->slot[i].shared = TRUE;
	}

	if (GET_LOGLEVEL(LOG_SPACE)) {
		/*
		 *	Log the copy of the slots as an append of the slots
		 */
		opInfo.action = LOG_ACTION_APPEND_LARGE_SLOTS;
		opInfo.flags = 0;
		opInfo.start = 0;
		opInfo.size = (int) (sizeof(LGSLOT) * rootPtr->header.numSlots);
		opInfo.oldData = NULL;
		opInfo.newData= (char *) &(rootPtr->slot[0]);
		opInfo.actionPid = rootPid;
		/* this can't be a root on a slotted page */
		opInfo.rootSlot = NIL;
		if(clog_LogModifyLgPage(&opInfo)) {
			FAIL
		}
	}

	/* Point the new version header to the new root pid */
	newVersionHdr->rootPid = rootPid.page;

	/* Unfix both root pages */
	bf_UnfixPage(parentRootLink, BF_DEFAULT, FALSE);
	bf_UnfixPage(rootLink, BF_DEFAULT, TRUE);

	/* return success */
    return(esmNOERROR);

FailureReturn:
	/* todo : unfix? */
	SM_ERROR(TYPE_FATAL, esmINTERNAL);
	return esmFAILURE;
}

