/*
 *   $RCSfile: lg_DestroyObject.c,v $  
 *   $Revision: 1.1.1.1 $  
 *   $Date: 1996/05/04 21:55:24 $      
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/


#include <stdio.h>
#include "ess.h"
#include "trace.h"
#include "checking.h"
#include "io.h"
#include "page.h"
#include "list.h"
#include "object.h"
#include "tid.h"
#include "bf.h"
#include "chunk.h"
#include "lock.h"
#include "bf_macro.h"
#include "sm_macro.h"
#include "lgobject.h"
#include "volume.h"
#include "io_extfuncs.h"
#include "bf_extfuncs.h"
#include "lg_intfuncs.h"
#include "lg_extfuncs.h"
#include "error.h"


/*
 * Routine to delete an entire large object (or a subtree thereof).  It
 * recursively walks the tree, giving all of its nodes back as free space.
 * It is assumed that space deallocation occurs in a deferred manner (as
 * always).
 *
 * A special case is working versions of objects.  In this case no 
 * subtree should be deleted if its root page is shared.  This *can* be
 * gauranteed by only calling lg_DestroyObject recursively on non-
 * shared subtrees.  But, in order to delete chunks from the buffer
 * pool, lg_DestroyObject must be called even on shared pages.  It's
 * debatable whether this is necessary, but that's the way its 
 * implemented.
 *
 * NOTE: the current implementation stops deleting once it
 *       reaches a shared subtree of a working version
 *
 * NOTE on LOGGING: currently (before versioning has been handled, this
 * function does no logging because the server handles page deallocations
 * in a manner that does not require any additional logging by the client
 */


 int
lg_DestroyObject (

	PID			*nodePid,
	BUFGROUP	*bufGroup,
	SLOTINDEX	nodeSlot,
	BOOL		shared,			/* TRUE if subtree at node is shared*/
	BOOL		workingVersion 	/* TRUE if obj is a working version */
)
{

    LGNODE		*nodePage;		/* ptr to buffered node			*/
	GROUPLINK	*nodeLink;
    int			slotIndex;		/* slot index for deletion		*/
    PID			leafPid;		/* for leaf node page ids		*/


	TRPRINT(TR_LG, TR_LEVEL_1, ("pid:%d", nodePid->page));

	/*
	 * Enforce this:
	 * NOTE on LOGGING: currently (before versioning has been handled), this
	 * function does no logging because the server handles page deallocations
	 * in a manner that does not require any additional logging by the client
	 */
	SM_ASSERT(LEVEL_3, !workingVersion || (shared == FALSE));

	if (nodeSlot != NIL)	{
		PAGESLOT		*slotPtr;
		SLOTTEDPAGE		*pagePtr;
		LARGEOBJROOT	*objectPtr;

		TRPRINT(TR_LG, TR_LEVEL_1, ("slot node"));

		if ((nodeLink = bf_ReadPage(bufGroup, nodePid, H_SLOTTED,
									SLOTTED_PAGE2SIZE, NL)) == NULL)	{
			return(esmFAILURE);
		}

		pagePtr = (SLOTTEDPAGE *) nodeLink->bufFrame;
		slotPtr = GETSLOTPTR(pagePtr, nodeSlot);
		objectPtr = (LARGEOBJROOT *) GETOBJECTPTR(pagePtr, slotPtr);
		nodePage = (LGNODE *) &(objectPtr->rootPage);
		CHECK_SLOTROOT_MAGIC(nodePage);
	} else {
		/*
		 * Read in the node and see if we're at the last level above the data.
		 * If not, recursively delete each of the current node's subtrees.
		 * Otherwise, delete each of the current node's children (i.e., the
		 * subtree's leaves).
		 */
		if ((nodeLink = bf_ReadPage(bufGroup, nodePid, H_LARGE,
									LG_PAGE2SIZE, NL)) == NULL)	{
			return(esmFAILURE);
		}
		nodePage = (LGNODE *) nodeLink->bufFrame;
		TRPRINT(TR_LG, TR_LEVEL_2, ("at level:%d", nodePage->header.height));
		lg_PrintNode(nodePage, 0, nodePid->page);
	}
	sm_SetAbortIfError();

	/*
	 *	Check to see if we are at the leaf level
	 */
    if (nodePage->header.height > 1) {

		/*
		 * We haven't gotten to the level above the leaves yet.  We need
		 * to delete all subtrees nodeed here and then free up the node
		 * and associated buffer frame.
		 */
		for (slotIndex = 0; slotIndex < nodePage->header.numSlots; slotIndex++) {

			TRPRINT(TR_LG, TR_LEVEL_2, ("destroy subtree at slot:%d pid:%d",
					slotIndex, nodePage->slot[slotIndex].pid.page));
	
			/*
			 *	destroy the subtree, if it's not shared.
			 */
			if ( !workingVersion || !nodePage->slot[slotIndex].shared) { 
				if (lg_DestroyObject( &(nodePage->slot[slotIndex].pid), 
									  bufGroup, NIL,
									  shared || nodePage->slot[slotIndex].shared, 
									  workingVersion )) {

					return(esmFAILURE);
				}
			}
		}

    } else {

		/*
		 * We are now sitting at the level above the leaves.
		 */
		for (slotIndex = 0; slotIndex < nodePage->header.numSlots; slotIndex++) {
			
			/*
			 *	If the page is not shared, deallocate it
			 */
			if (!workingVersion || 
				!(shared || nodePage->slot[slotIndex].shared) ) {

				/*
				 *	destroy any copies in the cache
				 */
				leafPid = nodePage->slot[slotIndex].pid;
				bf_HackChunkPage(&leafPid);

				/*
				 *	Deallocate the page
				 */
				if (io_DeallocPages(LGDATA_PAGE2SIZE, 1, &leafPid, FALSE, PAGE_LARGEDATA))	{

					return(esmFAILURE);
				}
				TRPRINT(TR_LG, TR_LEVEL_2, ("hacking and deallocating page:%d", leafPid.page));
			} else {

				/*
				 *	Page is shared, so only remove chunks containing it
				 *	(and matching the deleted object's root page) from the
				 *	buffer pool.
				 */
				leafPid = nodePage->slot[slotIndex].pid;
				bf_HackChunkPage(&leafPid);
			}
		}
    }

	/*
	 *	check to see if the slot root
	 */
	if (nodeSlot == NIL)	{

		/*
		 *	handle the index page, first checking to see if 
		 *  it is shared
		 */
		if (!workingVersion || !shared) {

			/*
			 *	blow the index page out of the cache
			 */
			bf_DiscardPage(bufGroup, nodeLink);

			/*
			 *	Deallocate the index page
			 */
			if (io_DeallocPages(LG_PAGE2SIZE, 1, nodePid, FALSE, PAGE_LARGENODE))	{

				return(esmFAILURE);
			}
			TRPRINT(TR_LG, TR_LEVEL_2, ("deallocating and discarding index node:%d",
					nodePid->page));
		} else {

			/*
			 *  The page is shared, so simply unfix it
			 */
			 bf_UnfixPage(nodeLink, BF_DEFAULT, FALSE);
		}

	} else {

		/*
		 *	unfix the page
		 */
		bf_UnfixPage(nodeLink, BF_DEFAULT, FALSE);
	}

	/*
	 *	return success
	 */
	return(esmNOERROR);
}
