/*
 *   $RCSfile: lg_DestroyVersion.c,v $  
 *   $Revision: 1.1.1.1 $  
 *   $Date: 1996/05/04 21:55:24 $      
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/


#include <stdio.h>
#include "ess.h"
#include "trace.h"
#include "checking.h"
#include "io.h"
#include "page.h"
#include "list.h"
#include "object.h"
#include "tid.h"
#include "bf.h"
#include "chunk.h"
#include "lock.h"
#include "bf_macro.h"
#include "sm_macro.h"
#include "lgobject.h"
#include "volume.h"
#include "pool.h"
#include "version_graph.h"
#include "version_funcs.h"
#include "io_extfuncs.h"
#include "bf_extfuncs.h"
#include "lg_intfuncs.h"
#include "lg_extfuncs.h"
#include "lg_versionfuncs.h"
#include "lg_globals.h"
#include "error.h"


/*
 * lg_DestroyVersion controls the destruction of a versioned large
 * object.
 */


 int
lg_DestroyVersion (

	BUFGROUP	*bufGroup,
	PID			*rootPid,		/* root page of the object 			*/
	SLOTINDEX	rootSlot,		/* slot on root page				*/
	BOOL		shared,			/* TRUE if subtree at node is shared*/
	BOOL		workingVersion,	/* TRUE if obj is a working version */
	VERSIONGRAPH *vhg, 			/* version history graph			*/
	BOOL		*destroyPending,/* OUT: object not destroyed yet	*/
	BOOL		*newPending,	/* OUT:	true if obj to destroy is returned*/
	OID			*newPendingObj,	/* OUT: new obj to destroy			*/
	BOOL		*destroyVHG 	/* OUT: version graph should be destroyed */
)
{

    LGNODE		*rootPage;		/* ptr to buffered root page		*/
	GROUPLINK	*rootLink;
	VHGNODEID	rootId;
	int			rootPageHeight;
	BOOL		vhgEmpty;

	TRPRINT(TR_LG, TR_LEVEL_1, ("pid:%d", rootPid->page));

	/*
	 *	Init the out parameters
	 */
	*destroyPending	= FALSE;
	*newPending 	= FALSE;
	*destroyVHG		= FALSE;

	/*
	 *	Get the root page.
	 *	check to see if this is a header on slotted pages
	 */
	if (rootSlot != NIL)	{

		PAGESLOT		*slotPtr;
		SLOTTEDPAGE		*pagePtr;
		LARGEOBJROOT	*objectPtr;

		TRPRINT(TR_LG, TR_LEVEL_1, ("slot node"));

		if ((rootLink = bf_ReadPage(bufGroup, rootPid, H_SLOTTED,
									SLOTTED_PAGE2SIZE, NL)) == NULL)	{
			return(esmFAILURE);
		}

		pagePtr = (SLOTTEDPAGE *) rootLink->bufFrame;
		slotPtr = GETSLOTPTR(pagePtr, rootSlot);
		objectPtr = (LARGEOBJROOT *) GETOBJECTPTR(pagePtr, slotPtr);
		rootPage = (LGNODE *) &(objectPtr->rootPage);
		CHECK_SLOTROOT_MAGIC(rootPage);
	} else {
		if ((rootLink = bf_ReadPage(bufGroup, rootPid, H_LARGE,
									LG_PAGE2SIZE, NL)) == NULL)	{
			return(esmFAILURE);
		}
		rootPage = (LGNODE *) rootLink->bufFrame;
		TRPRINT(TR_LG, TR_LEVEL_2, ("at level:%d", rootPage->header.height));
		lg_PrintNode(rootPage, 0, rootPid->page);
	}
	
	/*
	 *	remember the objects version graph node id, and the height
	 *	of the large obj btree
	 */
	rootId = rootPage->header.vhgPtr.nodeId;
	rootPageHeight = rootPage->header.height;

	/*
	 *	We are done with the root page
	 */
	bf_UnfixPage(rootLink, BF_DEFAULT, FALSE);

	/* each of the following branches would do this early
	 * on, so let's just do it now (even if it's a wee bit
	 * premature)
	 */
	sm_SetAbortIfError();


	/* 
	 *	If object is not a working version and any children are
	 *	working versions then mark the corresponding node in the
	 *	version history graph as DESTROY_PENDING.  The object will
	 *	not actually be destroyed until all children are frozen.
	 */
	if (!workingVersion && workingChildren(vhg, rootId)) {

		/*
		 *	Set the destroy pending mark for the object
		 */
		markDestroyPending(vhg, rootId);
		*destroyPending = TRUE;

	} else if (workingVersion) {
		
		/*
		 *	destroy the working version
		 */
		if (lg_DestroyObject(rootPid, bufGroup, rootSlot, shared, TRUE)) {
			bf_UnfixPage(rootLink, BF_DEFAULT, FALSE);
			return(esmFAILURE);
		}

		/*
		 *  mark the version as destroyed (tombstones in the VHG will
		 *	be removed if possible at this time)
		 */
		destroyNode(vhg, rootId, &vhgEmpty);
	
		/*
		 *	A working version cannot be the last version destroyed
		 *	in the version history graph.
		 */
		if (vhgEmpty) {
			SM_ERROR(TYPE_FATAL, esmINTERNAL);
			return(esmFAILURE);
		}

		/*
		 *	see if the parent of the version is marked destroy-pending
		 *	and has all its children frozen.  If so, then remember to
		 *	destroy the parent.
		 */
		*newPending = parentDestroyPending(vhg, rootId, newPendingObj);

	} else {
		
		LIST		adjList;	/* list of all adjacent versions 	*/
		LIST		adjLevel[TSTACKSIZE];	
								/* list of pid's at each level 
								   in adjacent versions 	  		*/
		LIST		destroyLevel[TSTACKSIZE]; 	
								/* list of pid's at each level in the
								   object to be destroyed			*/
		LGNODEHASHTABLE	nodeHashTable;
		LGNODELIST 	*nodeInfo;
		int			i;
		BOOL		destroyCompleted;

		/*
		 *	The version is frozen, but there are no working children,
		 *	so we can destroy the version, but must be careful not to
		 *	delete shared pages.
		 */

		/*
		 *	Initialize each list and the hash table
		 */
		initializeList(&adjList);
		for (i = 0; i < TSTACKSIZE; i++) {
			initializeList(&adjLevel[i]);
			initializeList(&destroyLevel[i]);
		}
		LG_HashTableInit(&nodeHashTable);

		/*
		 *	Get root pages of all adjacent objects
		 */
		if (findAdjacent(bufGroup, vhg, rootId, &adjList, &NodeListPool)) {
			
			destroyCompleted = FALSE;

		} else 

			/*
			 *	Find root node children for each object root node in 
			 *	adjList and add them to the proper levels in adjLevel
			 */
			if (LG_SortRootNodeChildren(bufGroup, &adjList, adjLevel, 
										&nodeHashTable, &NodeListPool)) {

				destroyCompleted = FALSE;

		} else {
			
			/*
			 *	Return all the adjacent root node to the pool
			 */
			LG_RecycleNodeListElements(&NodeListPool, &adjList);

			/*
			 *	adjList should now be empty
			 */
			SM_ASSERT(LEVEL_3, LIST_EMPTY(&adjList));

			/*
			 *	get space to store the root pid of the version  
			 */
			if ((nodeInfo = (LGNODELIST *) poolDeq( &NodeListPool )) == NULL) {
				destroyCompleted = FALSE;

			} else {

				/*
				 *	Fill in information for the root node
				 */
				nodeInfo->pid = *rootPid;
				nodeInfo->slot = rootSlot;

				/*
				 *	Put root pid (for the version) in the top level 
				 */
				listEnq(&destroyLevel[rootPageHeight], &nodeInfo->list);

				/*
				 *	Destroy all the non shared pages of the version, 
				 *	recursively.
				 */
				if (LG_DestroyVersionPage(bufGroup, destroyLevel, adjLevel,
									&nodeHashTable, &NodeListPool,
									rootPageHeight, TSTACKSIZE-1, 
									rootPid, rootSlot)) {

					destroyCompleted = FALSE;
				} else {

					/*
					 *  mark the version as destroyed (tombstones in the VHG may
					 *	be removed or added this time)
					 */
					destroyNode(vhg, rootId, &vhgEmpty);
					destroyCompleted = TRUE;
				}
			}
		}

		/*
		 *	Free up the pool of pid's
		 */
		LG_RecycleNodeListElements(&NodeListPool, &adjList);
		for (i = 0; i < TSTACKSIZE; i++) {
			LG_RecycleNodeListElements(&NodeListPool, &adjLevel[i]);
			LG_RecycleNodeListElements(&NodeListPool, &destroyLevel[i]);
		}

		/*
		 *	If a failure occured, return now
		 */
		if ( !destroyCompleted) {
			return(esmFAILURE);	
		}

		if (vhgEmpty) {
			/*
			 *  The version histroy graph is empty, so mark it to be
			 *	destroyed
			 */
			*destroyVHG = TRUE;
		}
	}

	/*
	 *	return success
	 */
	return(esmNOERROR);

}
