/*
 *   $RCSfile: lg_InsertInObject.c,v $  
 *   $Revision: 1.1.1.1 $  
 *   $Date: 1996/05/04 21:55:23 $      
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/


#include "ess.h"
#include "sysdefs.h"
#include "trace.h"
#include "checking.h"
#include "error.h"
#include "io.h"
#include "page.h"
#include "list.h"
#include "object.h"
#include "tid.h"
#include "bf.h"
#include "lock.h"
#include "chunk.h"
#include "pool.h"
#include "version_graph.h"
#include "bf_macro.h"
#include "bf_extfuncs.h"
#include "lgobject.h"
#include "lg_intfuncs.h"
#include "lg_extfuncs.h"
#include "lg_versionfuncs.h"
#include "util_funcs.h"


/*
 * Insert routine for adding bytes to the middle of a large storage object.
 */


 int
lg_InsertInObject (

	FID			*fid,
	PID			*rootPid,		/* pid of object to insert data into	*/
	BUFGROUP	*bufGroup,		/* buffer group to be used			*/
	int			start,			/* offset of desired insertion point*/
	int			totalLength,	/* number of bytes to insert		*/
	char		*data,			/* ptr to bytes to be inserted		*/
	SLOTINDEX	rootSlot,
	BOOL		versioned 		/* true if object is versioned		*/
)
{

	int			offset;			/* offset of start within leaf block*/
	PAGELIST	readList;		/* list of leaf pid/cnt pairs to read*/
	PAGELIST	writeList;		/* list of leaf pid/cnt pairs to write*/
	TSTACK		tstack;			/* stack for use in traversing object*/
	BOOL		pagesReplaced;	/* true if insert required new pages*/
	int			sharedCount;	/* number of shared pages in stack 	*/
	int			e;				/* generic error code				*/
	PID			newTopPid;		/* new page on top of stack			*/
	int        	length;			/* length of chunk being appended   */
   	int 		chunkLen;		/* max length of a chunk			*/
    SHORTPID	oldRootPage;	/* used to detect root was changed  */

	TRPRINT(TR_LG, TR_LEVEL_1, ("start:%d length:%d", start, totalLength));

/*
	TraceFlags = TR_LG;
	TraceLevel.level[findLog2(TR_LG)] = TR_LEVEL_3;
	lg_PrintIndex(bufGroup, rootPid, 0, TRUE, rootSlot);
*/

	chunkLen = bf_BreakdownChunk(bufGroup, totalLength);

	/* from now on, abort if there's a serious error */
	sm_SetAbortIfError();

	oldRootPage = rootPid->page;

	/*
	 * Insert into the object in chunks with a max size of chunkLen
	 */
	while(totalLength > 0) {

		length = MIN(totalLength, chunkLen);

		offset = start;
		LG_ResetTstack(&tstack);

		readList.pages = NULL; 
		writeList.pages = NULL;

		/*
		 *	traverse the object index updating appropriate counts
		 */
		if (LG_Traverse(rootPid, &offset, length, INSERT, &tstack, bufGroup, rootSlot, versioned, TSTACKSIZE, &sharedCount))	{
			FAIL
		}

		/*
		 *	If the stack contains shared pages, replace them and 
		 *	then the tstack must be recreated since interior nodes of the
		 *	large object btree were updated.  LG_Traverse is called with
		 *	parameters set so as to not update byte counts for nodes which
		 *	already have had them updated.
		 */
		if (versioned && sharedCount > 0) {
			int			dummyCount;			/* dummy counter			*/
			int			dummyOffset = start;
		
			e = LG_ReplaceSharedNodes(bufGroup, rootPid, &tstack, sharedCount, &newTopPid);
			if (e != esmNOERROR) {
				FAIL
			}

			LG_UnfixTstack(&tstack, FALSE);
			/*
			 *	Reset the stack and then retraverse the tree
			 */
			LG_ResetTstack(&tstack);
			if (LG_Traverse(rootPid, &dummyOffset, length, INSERT, &tstack,
							bufGroup, rootSlot, versioned, sharedCount, 
							&dummyCount))	{
				FAIL
			}
		}

		/*
		 *	Compute the effects of the insert
		 */
		if (LG_ComputeInsert(&offset, length, &readList, &writeList, &tstack))	{
			FAIL
		}

		/*
		 *	Update the chunk information to represent the insert.  
		 */
		if (LG_InsertData(fid, offset, length, data, &readList, &writeList, bufGroup, versioned, &pagesReplaced, sharedCount == 0 ? NULL : &newTopPid))	{

			FAIL
		}

	/*
		TraceFlags = TR_LG;
		TraceLevel.level[findLog2(TR_LG)] = TR_LEVEL_3;
		lg_PrintIndex(bufGroup, rootPid, 0, TRUE, rootSlot);
		TraceFlags = 0;
		TraceLevel.level[findLog2(TR_LG)] = TR_LEVEL_1;
	*/

		/*
		 *	upate the large object index to relect the insert
		 */
		if (LG_FixIndex(rootPid, &writeList, &tstack, bufGroup, readList.numPages))	{
			FAIL
		}
        if (oldRootPage != rootPid->page) {
            if (rootSlot != NIL) {
                /*
                 * the object root is now on a separate page instead
                 * of the slotted page
                 */
                rootSlot = NIL;
            }
            oldRootPage = rootPid->page;
        }

		FREELIST(&readList);
		FREELIST(&writeList);

		totalLength -= length;
		start += length;
		if (data != NULL) data += length;
	}

Success:
    return esmNOERROR;

FailureReturn:
	LG_UnfixTstack ( &tstack, TRUE );
    FREELIST(&readList);
    FREELIST(&writeList);
	return esmFAILURE;
	
}
