/*
 *   $RCSfile: lg_ReadObject.c,v $  
 *   $Revision: 1.1.1.1 $  
 *   $Date: 1996/05/04 21:55:23 $      
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/


#include "ess.h"
#include "sysdefs.h"
#include "trace.h"
#include "checking.h"
#include "io.h"
#include "page.h"
#include "list.h"
#include "object.h"
#include "tid.h"
#include "bf.h"
#include "chunk.h"
#include "lock.h"
#include "bf_macro.h"
#include "sm_macro.h"
#include "lgobject.h"
#include "bf_extfuncs.h"
#include "lg_intfuncs.h"
#include "lg_extfuncs.h"
#include "error.h"


/*
 * Routine to read a specified number of bytes from a large object, setting
 * up appropriate state information along the way.  The leaf contents read
 * remain fixed following this routine's invocation.  If all goes well, an
 * indirect pointer to the desired data (contiguously buffered) is returned.
 * In the event of an error, an error indication (-1) is returned.
 */


 int
lg_ReadObject (

	FID				*fid,
	PID				*rootPid,		/* root pid for large object			*/
	BUFGROUP		*bufGroup,		/* buffer group for accesses			*/
	int				start,			/* desired starting offset				*/
	int				length,			/* length of desired byte run			*/
	CHUNKDESC		**chunkDesc,	/* equivalent to CUSERDESC *userDesc	*/
	SLOTINDEX		slotRoot 
)
{
    int				listSize;		/* maximum leaf page list size			*/
    PAGELIST		leaves;			/* list of long leaf pages to read		*/
	GROUPLINK		*rootLink = NULL;		/* link of root page					*/
    int				offset;			/* offset of desired data on first page	*/
    LGNODE			*rootPage;		/* pointer to object's root node		*/
    int				initOffset;		/* value of starting offset				*/
    int				numBytes;		/* number of bytes desired				*/
	int				byteCount;


	TRPRINT(TR_LG, TR_LEVEL_1, ("rootpid:%d start:%d length:%d",
			rootPid->page, start, length));

	leaves.pages = NULL;
	/*
	 *	check parameters for validity
	 */
    if (start < 0)	{
		SM_ERROR(TYPE_USER, esmBADSTART);
		FAIL
    }

	/*
	 *	check to see if this is a header on slotted pages
	 */
	if (slotRoot != NIL)	{
		PAGESLOT		*slotPtr;
		SLOTTEDPAGE		*pagePtr;
		LARGEOBJROOT	*objectPtr;

		TRPRINT(TR_LG, TR_LEVEL_2, ("slot root"));

		if ((rootLink = bf_ReadPage(bufGroup, rootPid, H_SLOTTED,
									SLOTTED_PAGE2SIZE, NL)) == NULL)	{
			FAIL
		}

		pagePtr = (SLOTTEDPAGE *) rootLink->bufFrame;
		slotPtr = GETSLOTPTR(pagePtr, slotRoot);
		objectPtr = (LARGEOBJROOT *) GETOBJECTPTR(pagePtr, slotPtr);
		rootPage = (LGNODE *) &(objectPtr->rootPage);

		CHECK_SLOTROOT_MAGIC(rootPage);
	} else {
		if ((rootLink = bf_ReadPage(bufGroup, rootPid, H_LARGE,
									LG_PAGE2SIZE, NL)) == NULL)	{
			FAIL
		}
		rootPage = (LGNODE *) rootLink->bufFrame;
#ifdef  DEBUG
		lg_PrintNode(rootPage, 0, rootPid->page);
#endif  DEBUG

		CHECK_LARGENODE_MAGIC(rootPage);
	}

	/*
	 *	get the number of bytes in the tree
	 */
	byteCount = (int)rootPage->slot[rootPage->header.numSlots - 1].count;
	TRPRINT(TR_LG, TR_LEVEL_2, ("byteCount:%d", byteCount));

	/*
	 *	check to see if we are going to return less than requested
	 */
    if ((start + length) > byteCount)	{
		/*
		 *	we should read as much as possible
		 */
		length = byteCount - start;
		TRPRINT(TR_LG, TR_LEVEL_2, ("new length:%d", length));
	}

    bf_UnfixPage(rootLink, BF_DEFAULT, FALSE);

    /*
     *	Initialize the leaf page list for the scan, allowing plenty of room.
	 *	There are an array of funny conditions that are spanned by this
	 *	simple expression.
	 */
    listSize = (length / LGMINLEAF) + 2;
    if (LG_InitList(&leaves, listSize))	{
		FAIL
	}

	/*
     * A negative initial offset is passed as a "first time called" flag.
     * Then scan for leaves and return the resulting list.  Leave root page
     * fixed after scanning (to avoid deadlock potential).
     */
    initOffset = start;
    numBytes = length;
    offset = -1;
    if (LG_ScanForLeaves(rootPid, &initOffset, &numBytes,
                         bufGroup, &leaves, &offset, slotRoot, FALSE)) {
		FAIL
	}

    /*
     * Read the desired data into a chunk in the buffer pool and record
     * necessary state information;  return an indirect pointer to the data.
     */
    if (bf_ReadChunk(fid, bufGroup, leaves.numPages, leaves.numPages,
					 	 leaves.pages, offset, 0, chunkDesc))	{
		FAIL
	}

	/*
	 *	set the number of bytes accessible to the user
	 */
	(*chunkDesc)->byteCount = length;

	/*
	 *	Remember the root page info for this chunk
	 */
	(*chunkDesc)->rootPid = *rootPid;
	(*chunkDesc)->rootSlot = slotRoot;

Success:
	SM_FREE_PTR(leaves.pages);
    return esmNOERROR;

FailureReturn:
	SM_FREE_PTR(leaves.pages);
    return esmFAILURE;

}
