/*
 *   $RCSfile: lg_WriteObject.c,v $  
 *   $Revision: 1.1.1.1 $  
 *   $Date: 1996/05/04 21:55:23 $      
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/

#include "sysdefs.h"
#include "ess.h"
#include "trace.h"
#include "checking.h"
#include "error.h"
#include "io.h"
#include "page.h"
#include "list.h"
#include "object.h"
#include "tid.h"
#include "bf.h"
#include "chunk.h"
#include "bf_macro.h"
#include "lgobject.h"
#include "sm_macro.h"
#include "lock.h"
#include "lsn.h"
#include "clog.h"
#include "logaction.h"
#include "loginfo.h"
#include "logrecs.h"
#include "volume.h"
#include "sm_state.h"
#include "pool.h"
#include "version_graph.h"
#include "io_extfuncs.h"
#include "bf_extfuncs.h"
#include "lg_intfuncs.h"
#include "lg_extfuncs.h"
#include "lg_versionfuncs.h"
#include "trans.h"
#include "trans_funcs.h"
#include "clog_extfuncs.h"
#include "clog_globals.h"

/*
 * Routine to write a specified number of bytes from a large object, using
 * the state information from a (prerequisite) read.  The updated leaf bytes
 * remain fixed following this routine's invocation.  If something goes wrong,
 * -1 is returned.  Note:  A read must precede this request, as an associated
 * state record must exist (and be pointed to by a user descriptor field).
 */


 int
lg_WriteObject (
	CHUNKDESC	*chunkDesc,		/* user descriptor from previous read*/
	char		*basePtr,		/* pointer to start of object being written*/
	int			start,			/* desired starting offset within object	*/
	int			length,			/* length of modified sequence of bytes*/
	BOOL		releaseFlag,	/* true if release should follow write*/
	char		*data 
)
{

	CHUNKPAGE			*chunkPage;
	int					totalBytes;
    int					offset;    
	int					endByte;
	int					dataLength;
	int					currentOffset;
	BOOL				first;
	int					e;				/* error code				*/
	char 				*dataStart;		/* ptr to the start of the new data*/
	OPERATIONINFO		opInfo;
	char 				*oldDataPtr;    /* ptr into the old data to be logged*/
	int					origLength;


	TRPRINT(TR_LG, TR_LEVEL_1, ("start:%d length:%d", start, length));

	dataStart = data;
	oldDataPtr = basePtr + start;
	origLength = length;

	/*
	 *	the user is passing in a reference to the relative
	 *	offset of byte basePtr so we need to calculate
	 *	the offset from the beginning of the buffered chunk
	 */
	CHECK_CHUNKDESC_MAGIC(chunkDesc);
	offset = start + chunkDesc->firstOffset;
	endByte = offset + length;
	first = TRUE;
	TRPRINT(TR_LG, TR_LEVEL_2, ("first offset:%d endByte:%d", offset, endByte));

	/*
	 *	check to make sure that the bounds are valid
	 */
	if ((offset > chunkDesc->totalLength) || (endByte > chunkDesc->totalLength))	{
		SM_ERROR(TYPE_FATAL, esmINTERNAL);
	}

	/*
	 *	Determine if there are possibly shared pages in the chunk
	 *	to be replaced.
	 *	Note that the shared pages count is only 
	 *	valid for versioned objects.
	 */
	if ((chunkDesc->userDesc->userFlags & P_VERSIONED) && 
		(chunkDesc->sharedPages > 0) ) {
		/*
		 *	Replace any shared leaf pages which will be written 
		 *	(those in the range [offset-endByte]).  Also replace
		 *	shared interior nodes (the last 2 parameters indicate this).
		 */
		e = LG_ReplaceSharedLeaves(chunkDesc, offset, endByte, NULL, TRUE);
		if (e != esmNOERROR) return e;
	}

	/*
	 *	Loop through to find the page that contains
	 *	the first target byte.
	 */
	chunkPage = chunkDesc->chunkPage;
	totalBytes = 0;

	for (;;)	{

		/*
		 *	check to see if total bytes is reached
		 */
		if (totalBytes + chunkPage->size > offset)	{

			TRPRINT(TR_LG, TR_LEVEL_2, ("totalBytes:%d pid:%d",
					totalBytes, chunkPage->pid.page));
			
			/*
			 *	calculate the length of the data on this page
			 */
			currentOffset = (first ? (offset - totalBytes) : 0);
			dataLength = MIN(length, (chunkPage->size - currentOffset));
			TRPRINT(TR_BF, TR_LEVEL_2, ("pageSize:%d offset:%d currentOffset:%d dataLength:%d",
					chunkPage->size, offset, currentOffset, dataLength));

			if (GET_LOGLEVEL(LOG_DATA) ) {
				/* log the write of the data for this page */
				opInfo.flags = 0;
				opInfo.start = currentOffset;
				opInfo.size = dataLength;
				opInfo.oldData = oldDataPtr;
				opInfo.actionPid = chunkPage->pid;
				if (data == NULL) {
					opInfo.action = LOG_ACTION_WRITE_LARGE_DATA_ND;
					opInfo.newData = NULL;
					if(clog_LogWriteLgPageND(&opInfo)) {
						FAIL
					}
				} else {
					opInfo.action = LOG_ACTION_WRITE_LARGE_DATA;
					opInfo.newData = data;
					if( clog_LogWriteLgPage(&opInfo)) {
						FAIL
					}
				}
			}
			sm_SetAbortIfError();

			/*
			 *	Copy the write to other chunks containing this page
			 */
			bf_UpdateChunkPage(chunkPage, currentOffset, dataLength, data);

			/*
			 *	increment the starting offsets
			 */
			if (data != NULL) data += dataLength;
			oldDataPtr += dataLength;
			length -= dataLength;
			first = FALSE;

			if (!(chunkPage->flags & PAGE_DIRTY))	{
				chunkPage->flags |= PAGE_DIRTY;
				bf_DirtyPage(chunkPage->pageHash);
				chunkDesc->dirtyPages++;
			}
		}

		/* increment total bytes */
		totalBytes += chunkPage->size;

		/* look at the next chunk descriptor */
		chunkPage++;

		/* check to see if we have touched all the pages */
		if (totalBytes >= endByte)	{

			/* We have updated all pages */
			break;

		}
	}

	/* now update the data in the chunk */
	if (data == NULL) {
		bzero(basePtr + start, (int) origLength);
	} else {
		bcopy(dataStart, basePtr + start, (int) origLength);
	}
	sm_SetAbortIfError();

	/* mark the chunk as dirty */
	chunkDesc->flags |= CHUNK_DIRTY;

	/* check to see if the release flag has been set */
	if (releaseFlag)	{
		/* Unfix the chunk */
		bf_UnfixChunk(chunkDesc, BF_DEFAULT);
	}
	return esmNOERROR;

FailureReturn:
	return esmFAILURE;
}



/*
 *	This function is very similar to lg_WriteObject except that it
 *	assumes the object has been written in place and that the data
 *	parameter is a copy of the original data.
 */
 int
lg_WroteObject (

	CHUNKDESC	*chunkDesc,		/* user descriptor from previous read*/
	char		*basePtr,		/* pointer to start of object being written*/
	int			start,			/* desired starting offset within object	*/
	int			length,			/* length of modified sequence of bytes*/
	BOOL		releaseFlag,	/* true if release should follow write*/
	char		*data 			/* old data for the object			*/
)
{

	CHUNKPAGE			*chunkPage;
	int					totalBytes;
    int					offset;    
	int					endByte;
	int					dataLength;
	int					currentOffset;
	BOOL				first;
	int					e;				/* error code				*/
	char 				*dataStart;		/* ptr to the start of the new data*/
	OPERATIONINFO		opInfo;
	char 				*newDataPtr;    /* ptr into the new data to be logged*/
	int					origLength;


	TRPRINT(TR_LG, TR_LEVEL_1, ("start:%d length:%d", start, length));

	dataStart = data;
	newDataPtr = basePtr + start;
	origLength = length;

	/*
	 *	the user is passing in a reference to the relative
	 *	offset of byte basePtr so we need to calculate
	 *	the offset from the beginning of the buffered chunk
	 */
	CHECK_CHUNKDESC_MAGIC(chunkDesc);
	offset = start + chunkDesc->firstOffset;
	endByte = offset + length;
	first = TRUE;
	TRPRINT(TR_LG, TR_LEVEL_2, ("first offset:%d endByte:%d", offset, endByte));

	/*
	 *	check to make sure that the bounds are valid
	 */
	if ((offset > chunkDesc->totalLength) || (endByte > chunkDesc->totalLength))	{
		SM_ERROR(TYPE_FATAL, esmINTERNAL);
	}

	/*
	 *	Determine if there are possibly shared pages in the chunk
	 *	to be replaced.
	 *	Note that the shared pages count is only 
	 *	valid for versioned objects.
	 */
	if ((chunkDesc->userDesc->userFlags & P_VERSIONED) && 
		(chunkDesc->sharedPages > 0) ) {
		/*
		 *	Replace any shared leaf pages which will be written 
		 *	(those in the range [offset-endByte]).  Also replace
		 *	shared interior nodes (the last 2 parameters indicate this).
		 */
		e = LG_ReplaceSharedLeaves(chunkDesc, offset, endByte, NULL, TRUE);
		if (e != esmNOERROR) return e;
	}

	/*
	 *	Loop through to find the page that contains
	 *	the first target byte.
	 */
	chunkPage = chunkDesc->chunkPage;
	totalBytes = 0;

	for (;;)	{

		/*
		 *	check to see if total bytes is reached
		 */
		if (totalBytes + chunkPage->size > offset)	{

			TRPRINT(TR_LG, TR_LEVEL_2, ("totalBytes:%d pid:%d",
					totalBytes, chunkPage->pid.page));


			/*
			 *	calculate the length of the data on this page
			 */
			currentOffset = (first ? (offset - totalBytes) : 0);
			dataLength = MIN(length, (chunkPage->size - currentOffset));
			TRPRINT(TR_BF, TR_LEVEL_2, ("pageSize:%d offset:%d currentOffset:%d dataLength:%d",
					chunkPage->size, offset, currentOffset, dataLength));

			if (GET_LOGLEVEL(LOG_DATA) ) {
				/* log the write of the data for this page */
				opInfo.flags = 0;
				opInfo.start = currentOffset;
				opInfo.size = dataLength;
				opInfo.oldData = data;
				opInfo.actionPid = chunkPage->pid;
				if (data == NULL) {
					SM_ASSERT(LEVEL_3, esmINTERNAL);
					opInfo.action = LOG_ACTION_WRITE_LARGE_DATA_ND;
					opInfo.newData = NULL;
					if(clog_LogWriteLgPageND(&opInfo)) {
						FAIL
					}
				} else {
					opInfo.action = LOG_ACTION_WRITE_LARGE_DATA;
					opInfo.newData = newDataPtr;
					if( clog_LogWriteLgPage(&opInfo)) {
						FAIL
					}
				}
			}

			/*
			 *	Copy the write to other chunks containing this page
			 */
			bf_UpdateChunkPage(chunkPage, currentOffset, dataLength, newDataPtr);

			/*
			 *	increment the starting offsets
			 */
			if (data != NULL) data += dataLength;
			newDataPtr += dataLength;
			length -= dataLength;
			first = FALSE;


			/*
			 *	check to see if the page is dirty
			 */
			if (!(chunkPage->flags & PAGE_DIRTY))	{

				/*
				 *	Dirty the page
				 */
				chunkPage->flags |= PAGE_DIRTY;
				bf_DirtyPage(chunkPage->pageHash);

				/*
				 *	Increment the number of dirty pages in the chunk
				 */
				chunkDesc->dirtyPages++;
			}

		}

		/* increment total bytes */
		totalBytes += chunkPage->size;

		/* look at the next chunk descriptor */
		chunkPage++;

		/* check to see if we have touched all the pages */
		if (totalBytes >= endByte)	{

			/* We have updated all pages */
			break;

		}
	}
	/* mark the chunk as dirty */
	chunkDesc->flags |= CHUNK_DIRTY;

	/* check to see if the release flag has been set */
	if (releaseFlag)	{
		/* Unfix the chunk */
		bf_UnfixChunk(chunkDesc, BF_DEFAULT);
	}
	return esmNOERROR;

FailureReturn:
	return esmFAILURE;
}
