/*
 *	$RCSfile: LHDIRPAGE.C,v $
 *	$Revision: 1.1.1.1 $
 *	$Date: 1996/05/04 21:55:25 $
 */
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/
#include "BTREEPAGE.h"
#include "LH_Hash.h"

//
//	Class		: LHDIRPAGE
//	Method		: int IndexCalc(int keyLen, void* keyVal)
//	Description	: Find the index value of input key, only root page can
//					perform such calculation.
//	Return		: Index value
//
int LHDIRPAGE::IndexCalc(int keyLen, void* keyVal)
{
	ASSERT1(lhDirCtrl.rootFlag);

	int hashVal = LH_Hash(keyLen, keyVal, lhDirCtrl.type, lhDirCtrl.I);

	if (hashVal >= lhDirCtrl.M)
		hashVal = LH_Hash(keyLen, keyVal, lhDirCtrl.type, lhDirCtrl.I-1);
	
	return hashVal;
}

//
//	Class		: LHDIRPAGE
//	Method		: int Capacity(int height)
//	Description	: Calculate the capacity of directory given height of tree
//	Return		: maximum capacity of given height
//
int LHDIRPAGE::Capacity(int height)
{
	ASSERT1(height>=0);

	int value = 1;

	for (int i=0; i<height; i++)
		value *= LH_TBLSIZE;

	return value;
}

//
//	Class		: LHDIRPAGE
//	Method		: void IndSplitUpdate(float oldFromLoad, float fromLoad, 
//										float toLoad) 
//	Description	: Update the relevant information in root directory page 
//					after index page splitting.  Only root directory page
//					can	perform such operation.
//	Return		: None
//
void LHDIRPAGE::IndSplitUpdate(float oldFromLoad, float fromLoad, float toLoad)
{
	ASSERT1(lhDirCtrl.rootFlag);
	
	//
	//	increment level when the page just splitted is the first page
	//
	if (lhDirCtrl.P == 0) lhDirCtrl.I++;

	//
	//	increment number of primary pages
	//
	lhDirCtrl.M++;
	
	//
	//	increment split pointer
	//
	lhDirCtrl.P = (lhDirCtrl.P + 1) % (INIT_PRIMARY << (lhDirCtrl.I-1)) ;

	//
	//  update load infomation:
	//
	lhDirCtrl.load += (fromLoad - oldFromLoad + toLoad - lhDirCtrl.load)
						/lhDirCtrl.M; 

}

//
//	Class		: LHDIRPAGE
//	Method		: void IndShrinkUpdate(float fromLoad, float toLoad, 
//										float newFromLoad) 
//	Description	: Update the relevant information in root directory page 
//					after index page shrinking.  Only root directory page
//					can	perform such operation.
//	Return		: None
//
void LHDIRPAGE::IndShrinkUpdate(float fromLoad, float toLoad, float newFromLoad)
{
	ASSERT1(lhDirCtrl.rootFlag);
	
	//
	//      decrement split pointer
	//
	if (lhDirCtrl.P == 0) 
		lhDirCtrl.P = (INIT_PRIMARY << (lhDirCtrl.I-1)) - 1 ;
	else 
		lhDirCtrl.P--;

	//
	//	decrement number of primary pages
	//
	lhDirCtrl.M--;

	//
	//      decrement level if necessary
	//
	if (lhDirCtrl.P == 0)
		lhDirCtrl.I--;

	//
	//  update load infomation:
	//
	lhDirCtrl.load += (lhDirCtrl.load + newFromLoad - fromLoad - toLoad)
						/lhDirCtrl.M; 
}

//
//	Class		: LHDIRPAGE
//	Method		: void LoadUpdate(float oldLoad, float curLoad) 
//	Description	: Update load information after insert or delete
//	Return		: None
//
void LHDIRPAGE::LoadUpdate(float oldLoad, float curLoad)
{
	ASSERT1(lhDirCtrl.rootFlag);

	//
	//	update the load information using the new load of updated page
	//
	lhDirCtrl.load += ((curLoad - oldLoad)/lhDirCtrl.M); 
}

//
//	Class		: LHDIRPAGE
//	Method		: void Print() 
//	Description	: Print out contend of directory page
//	Return		: None
//
void LHDIRPAGE::Print()
{

	if (IsRoot()) {
		//
		// root page
		//
		printf("\n\troot directory page: pid = %d (max # of pages = %d)\n",
					lhDirCtrl.selfID.page, Capacity(lhDirCtrl.height));
		printf("\t# of primary pages = %d:\n", (int) lhDirCtrl.M);
		printf("\t\tcurrent split ptr  = %d:\n", (int) lhDirCtrl.P);
		printf("\t\tcurrent level  = %d:\n", (int) lhDirCtrl.I);
		printf("\t\tload threshold = %f:\n", lhDirCtrl.loadThreshold);
		printf("\t\tcurrent load  = %f:\n", lhDirCtrl.load);
		}
	else {
		printf("\n\tnon-root directory page: pid = %d\n", 
			lhDirCtrl.selfID.page);
		}

	int numEntry = (int) lhDirCtrl.curPos;
	printf("\t\tcurrent height  = %d:\n", (int) lhDirCtrl.height);
	printf("\t\tcurrent # of entries  = %d:\n", numEntry);
	printf("\t\tpage table entries:\n");
	for (int i=0; i < numEntry; i++)	{
		printf("\t\tentry[%d] = %d\n", i, (int) indPid[i]);
		}
}

//
//	Class		: LHDIRPAGE
//	Method		: int CheckContent() 
//	Description	: Check content of directory page: consistancy among
//		    	  parameters
//	Return		: TRUE if O.K.
//
int LHDIRPAGE::CheckContent()
{
#ifdef DEBUG
	if (IsRoot()) {
		//
		// root page
		//
		ASSERT1(lhDirCtrl.M <= Capacity(lhDirCtrl.height));
		ASSERT1(lhDirCtrl.P < lhDirCtrl.M);
		ASSERT1(lhDirCtrl.M <= INIT_PRIMARY << (int) lhDirCtrl.I);
		}

	ASSERT1(lhDirCtrl.curPos <= LH_TBLSIZE);
	ASSERT1(lhDirCtrl.height >0);
#endif
	return TRUE;
}
