/*
 *	$RCSfile: LHINDPAGE.C,v $
 *	$Revision: 1.1.1.1 $
 *	$Date: 1996/05/04 21:55:25 $
 */
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/
#include "BTREEPAGE.h"
#include "OVPAGE.h"
#include "TempBufSpace.h"



//
//	Class		: LHINDPAGE
//	Method		: int Search(int keyLen, void* keyVal, PFC compFunc, 
//										TWO& retSlot)
//	Description	: Searches the page for a key, and returns its slot in 
//				  retSlot if found.
//	Return		: TRUE if found, FALSE otherwise
//
int LHINDPAGE::Search(int keyLen, void* keyVal, 
								PFC compFunc, TWO& retSlot)
{
	int low, high;
	register mid;
	for (low = mid = 0, high = NumSlots() - 1; low <= high;  )	{
		mid = (low + high) / 2;
		register diff = compFunc(keyLen, keyVal, Tuple(mid).KeyLen(), 
										Tuple(mid).KeyValue());
		if ( ! diff )	{
			// 	---- key found !
			retSlot = mid;
			return TRUE;
			}
		if ( diff > 0)
			low = mid + 1;
		else
			high = mid - 1;
		}

	retSlot = (low > mid) ? low : mid;

	return FALSE;
}




//
//	Class		: LHINDPAGE
//	Method		: void ShiftTuple(int slotNum, LHINDPAGE* destPage,
//												int destSlot, int numTuples)
//	Description	: Move numTuples from slotNum in this page to destSlot
//				  in destPage.
//
void LHINDPAGE::ShiftTuple(int slotNum, LHINDPAGE* destPage, 
											int destSlot, int numTuples)
{
	//
	//	Assert that we shift at least 1 tuple
	//
	ASSERT3((numTuples > 0));

	//
	//	Assert that the 'destSlot' parameter is sane
	//
	ASSERT3((destSlot <= destPage->NumSlots()));

	//
	//	Calculate total # of bytes needed to move
	//
	int byteCnt = 0;
	for (register i = slotNum; i < slotNum + numTuples; i++)  {
		byteCnt += Tuple(i).Size() + sizeof(SLOT);
		}

	//
	//	Compress destination page if necessary
	//
	if (destPage->UsableSpace() < byteCnt) 
		destPage->Compress();

	ASSERT1(destPage->UsableSpace() >= byteCnt);

	//
	//	Make room in slot[] of destination page
	//
	register SLOT* p = &(destPage->slot[-(destPage->NumSlots()-1+numTuples)]);
	SLOT* ep = &(destPage->slot[-destSlot+1]);
	for ( ; p + numTuples != ep; p++) 		*p = *(p + numTuples);

	//
	//	Move the entries
	//
	for (i = 0; i < numTuples; i++)  {
		destPage->slot[-(i+destSlot)] = destPage->lhCtrl.endData;
		BT_Tuple* tp = &Tuple(i+slotNum);
		int	entryLen = tp->Size();
		BCOPY(tp, destPage->data + destPage->lhCtrl.endData, entryLen);
		destPage->lhCtrl.endData += entryLen;
		}

	//
	//	Close gap in slot[] of source page
	//
	p = & slot[-slotNum];
	ep = & slot[-NumSlots()];
	for ( ; p - numTuples != ep; p--)		*p = * (p - numTuples);

	//
	//	Update Control Info of both pages
	//
	destPage->lhCtrl.numFree -= byteCnt;
	destPage->lhCtrl.numOffsets += numTuples;

	lhCtrl.numFree += byteCnt;
	lhCtrl.numOffsets -= numTuples;

	ASSERT3((CheckContent()));
	ASSERT3((destPage->CheckContent()));
}




//
//	Class		: LHINDPAGE
//	Method		: void InsertTuple(BT_Tuple& tuple, int slotNum)
//	Description	: Insert 'tuple' into slot[slotNum] in this page
//
void LHINDPAGE::InsertTuple(BT_Tuple& const tuple, int slotNum) 
{
	int 	length = tuple.Size();

	ASSERT3(slotNum <= NumSlots());

	//
	//	Assert right kind of tuple for right kind of page
	//
	ASSERT3(tuple.IsLeafTuple());

	//
	//	Compress the page if there is not enough room at the end
	//
	if (UsableSpace() < length + sizeof(SLOT))  {
		Compress();
		}

	ASSERT1(UsableSpace() >= length + sizeof(SLOT));
	
	//
	//	Make room in slot array for the new entry
	//
	register SLOT* p = & slot[- lhCtrl.numOffsets];
	SLOT* ep = & slot [- slotNum];
	for ( ; p != ep; p++)  *p = * (p+1);

	//
	//	Append the tuple to end of the data area and update control info
	//
	slot[-slotNum] = lhCtrl.endData;
	BCOPY((char*) &tuple, data + lhCtrl.endData, length);
	lhCtrl.endData += length;
	lhCtrl.numFree -= length + sizeof(SLOT);
	lhCtrl.numOffsets++;
	
	ASSERT3(CheckContent());
}


//
//	Class		: LHINDPAGE
//	Method		: void RemoveTuple(int slotNum)
//	Description	: Remove the tuple in slot[slotNum] in this page
//
void LHINDPAGE::RemoveTuple(int slotNum) 
{
	ASSERT3(slotNum >= 0 && slotNum < NumSlots());
	
	//
	//	Increment numFree before we remove the tuple
	//
	lhCtrl.numFree += Tuple(slotNum).Size() + sizeof(SLOT);	

	//
	// 	Close gap in the slot array 
	//
	register SLOT* p = & slot[-slotNum];
	SLOT* ep = & slot[- lhCtrl.numOffsets];
	for ( ; p != ep; p--)		*p = *(p-1);

	lhCtrl.numOffsets--;

	ASSERT3((CheckContent()));
}



//
//	Class		: LHINDPAGE
//	Method		: int AddElem(int sNum, void* const elem)
//	Description	: Add an element to the tuple in slot[sNum]. 
//				  This implementation should be improved further for efficiency.
//
int LHINDPAGE::AddElem(int sNum, void* const elem)
{
	ASSERT3(sNum < NumSlots());

	//
	//	Copy tuple into a scratch area to do the update
	//
#ifdef SERVER_MAKE
    // too big to put on stack
    TempBufSpace    tempSpace(sizeof(BT_Tuple));
    BT_Tuple&       scratch = *(BT_Tuple*)tempSpace.location();
#else
    BT_Tuple scratch;   // scratch space
#endif /* SERVER_MAKE */
	int oldLength = Tuple(sNum).Size();	
	BCOPY(&Tuple(sNum), &scratch, oldLength);
	
	ASSERT3( ! scratch.IsOverflow());
	ASSERT3( ! scratch.WillOverflow());

	if (scratch.AddElem(elem))
		return esmFAILURE;

	int newLength = scratch.Size();
	
	ASSERT3(newLength > oldLength);

	//
	//	See if we can put scratch at the end of data area
	//
	if (newLength < UsableSpace())   {
		slot[-sNum] = lhCtrl.endData;
		BCOPY(&scratch, data + slot[-sNum], newLength);

		lhCtrl.numFree += oldLength;
		}
	else {
		//
		//	Compress to free up some space for scratch
		//
		slot[-sNum] = EMPTYSLOT;		// free old tuple but not its slot
		lhCtrl.numFree += oldLength;

		Compress();

		ASSERT1(newLength <= UsableSpace());

		BCOPY(&scratch, data + lhCtrl.endData, newLength);
		slot[-sNum] = lhCtrl.endData;
		}
	lhCtrl.endData += newLength;
	lhCtrl.numFree -= newLength;

	ASSERT3((CheckContent()));
	return esmNOERROR;
}


//
//	Class		: LHINDPAGE
//	Method		: int RemoveElem(int sNum, void* const elem)
//	Description	: Delete an element from the tuple at slot[sNum]
//
int LHINDPAGE::RemoveElem(int sNum, void* const elem)
{

	ASSERT3(sNum < NumSlots());

	BT_Tuple* p = & Tuple(sNum);
	int oldLength = p->Size();
	
	ASSERT3( ! p->IsOverflow());

	if (p->RemoveElem(elem))
		return esmFAILURE;

	lhCtrl.numFree += oldLength - p->Size();

	ASSERT3((CheckContent()));

	return esmNOERROR;
}



//
//	Class		: LHINDPAGE
//	Method		: void Compress()
//	Description	: Compress the page to free up fragmented spaces
//
void LHINDPAGE::Compress()
{
	//
	//	Copy whole page onto scratch
	//
#ifdef SERVER_MAKE
    // too big to put on stack
    TempBufSpace    tempSpace(sizeof(LHINDPAGE));
    LHINDPAGE&      scratch = *(LHINDPAGE*)tempSpace.location();
#else
    LHINDPAGE scratch;   // scratch space
#endif /* SERVER_MAKE */
	BCOPY((char*) this, (char*) &scratch, sizeof(LHINDPAGE));
	
	//
	//	Loop through tuples in scratch and copy them back to btPage
	//
	char* p = data;
	for (register i = 0; i < scratch.lhCtrl.numOffsets; i++)  {
		if (scratch.slot[-i] == EMPTYSLOT)	continue;
		int length = scratch.Tuple(i).Size();
		BCOPY(&scratch.Tuple(i), p, length);	;
		slot[-i] = p-data;

//		ASSERT3((ALIGNSIZE(length) == length));

		p += length;
		}
		
	//
	// after compression, UsableSpace == FreeSpace 
	//
	lhCtrl.endData = p - data;
	lhCtrl.numFree = UsableSpace();	 

	ASSERT3((CheckContent()));
}

//
//	Class		: LHINDPAGE
//	Method		: float LoadComp()
//	Description	: Returns the new calculated load of current primary page
//
float LHINDPAGE::LoadComp()
{
	if (NumSlots()==0)
		return 0.0;
	//
	// estimate average tuple size
	//
	int avgSize = UsedSpace()/NumSlots();
	if ((avgSize= UsedSpace()/NumSlots())==0) 
		return 0.0;
	//
	// calculate the approximate capacity of current page
	//
	int capacity = LH_DATASIZE/avgSize;

	return (float)NumSlots()/(float)capacity;
}

//
//	Class		: LHINDPAGE
//	Method		: int IsSafe(LH_OPER oper, int maxKeyLen)
//	Description	: Returns TRUE if this page is "safe", FALSE otherwise.
//
int LHINDPAGE::IsSafe(LH_OPER oper, int maxKeyLen)
{
    int retVal = FALSE;
	BT_Tuple* p = 0;

	switch (oper)  {
		case LH_READ: 		
				retVal = TRUE;
				break;

		case LH_INSERT:		
				retVal = FreeSpace() >= p->PredictSize(maxKeyLen);
				break;

		case LH_DELETE:		
				retVal = UsedSpace() - p->PredictSize(maxKeyLen) >= 
													(LH_DATASIZE >> 1);
				break;
		case LH_INSERT_DELETE:
		default:
				// Illegal parameter
				ASSERT1(FALSE);
		}

	return retVal;
}




//
//	Class		: LHINDPAGE
//	Method		: void SetOverflow(int slotNum, PID& const ovPid)
//	Description	: Sets Tuple(slotNum) to overflow, account for spaces
//				  freed up by its transition.
//
void LHINDPAGE::SetOverflow(int slotNum, PID& const ovPid)
{
	BT_Tuple& tp = Tuple(slotNum);
	ASSERT3(! tp.IsOverflow());
	int oldSize = tp.Size();
	tp.SetOverflow(ovPid);
	ASSERT3(oldSize > tp.Size());
	lhCtrl.numFree += oldSize - tp.Size();
	ASSERT3(CheckContent());
}


//
//	Class		: LHINDPAGE
//	Method		: void SetUnderflow(int slotNum, int elCnt, void* elList)
//	Description	: Sets Tuple(slotNum) to underflow. Insert elList[elCnt]
//				  into the tuple.
//
void LHINDPAGE::SetUnderflow(int slotNum, int elCnt, void* const elList)
{
	ASSERT3(slotNum < NumSlots());
	
	//
	//	Copy tuple into a scratch area to do the update
	//
#ifdef SERVER_MAKE
    TempBufSpace    tempSpace(sizeof(BT_Tuple));
    BT_Tuple&       scratch = *(BT_Tuple*)tempSpace.location();
#else
    BT_Tuple        scratch;   // scratch space
#endif /* SERVER_MAKE */
	int oldLength = Tuple(slotNum).Size();	
	BCOPY(&Tuple(slotNum), &scratch, oldLength);
	
	scratch.ResetOverflow(elCnt, elList);
	
	int newLength = scratch.Size();
	
	ASSERT3(newLength > oldLength);

	//
	//	See if we can put scratch at the end of data area
	//
	if (newLength < UsableSpace())   {
		slot[-slotNum] = lhCtrl.endData;
		BCOPY(&scratch, data + slot[-slotNum], newLength);

		lhCtrl.numFree += oldLength;
		}
	else {
		//
		//	Compress to free up some space for scratch
		//
		slot[-slotNum] = EMPTYSLOT;		// free old tuple but not its slot
		lhCtrl.numFree += oldLength;

		Compress();

		ASSERT1(newLength <= UsableSpace());

		BCOPY(&scratch, data + lhCtrl.endData, newLength);
		slot[-slotNum] = lhCtrl.endData;
		}
		
	lhCtrl.endData += newLength;
	lhCtrl.numFree -= newLength;

	ASSERT3((CheckContent()));
}

extern PFK BT_PrintFunc;

void LHINDPAGE::Print(BUFGROUP* bufGroup)
{
	int	i;
	int	numOffsets = lhCtrl.numOffsets;

	ASSERT3((CheckContent()));


	if (!lhCtrl.selfOv) {
		printf("\n\n\t\tprimary index page, pid = %d\n",lhCtrl.selfID.page);
		}
	else {
		printf("\n\n\t\toverflow index page, pid = %d\n",lhCtrl.selfID.page);
		}

	printf("\t\t\tindex value = %d\n", lhCtrl.selfInd);
	printf("\t\t\thave overflow chain? = %d\n", lhCtrl.nextOv);
	printf("\t\t\tnext overflow pid = %d\n", lhCtrl.ovbId);
	printf("\t\t\tpage load = %f\n",  lhCtrl.load);
	
	for (i = 0; i < numOffsets; i++)  {
		void*	key = Tuple(i).KeyValue();
		TWO		keyLen = Tuple(i).KeyLen();
		TWO		numEl;
		int 	j;
	
		printf("\t\t\t");
		printf("    <key = ");
		BT_PrintFunc(keyLen, (char*) key);

		numEl = Tuple(i).ElCnt();

	    printf(", elCnt = %d", numEl);

		for (j = 0; j < numEl; j++)  {
			char el[SM_MAXELEMLEN];
			if ( ! Tuple(i).IsOverflow())
				BCOPY(Tuple(i).ElArray(j), el, Tuple(i).ElSize());
			else if (ov_GetElem(Tuple(i).OverflowRoot(), bufGroup, j, el)){
#ifdef SERVER_MAKE
				SM_ERROR(TYPE_FATAL, Active->errno);
#else
				SM_ERROR(TYPE_USER, sm_errno);
#endif
				}
			if (j == 0)
				printf(", (%d)", * (int*) el);
		    }
	    printf(">\n");
		}
}


//
//	Class		: LHINDPAGE
//	Method		: int CheckContent()
//	Description	: Checks the content of this page for inconsistency.
//				  It is fatal to be inconsistent.
//				  Checks for 1.) tuple overlapping one another
//							 2.) values of endData and numFree
//				  Returns TRUE if completed successfully.
//
int LHINDPAGE::CheckContent()
{
#ifdef DEBUG

	ASSERT1(NumSlots() >= 0);

	int ENDDATA = 0;
	int NUMFREE = LH_DATASIZE;
#ifdef SERVER_MAKE
    // too big to put on stack
    TempBufSpace    tempSpace(LH_DATASIZE);
    char*       map = tempSpace.location();
#else
    char map[LH_DATASIZE];
#endif /* SERVER_MAKE */

	bzero(map, LH_DATASIZE);

	for (int i = 0; i < NumSlots(); i++)  {
		if (slot[-i] == EMPTYSLOT)  {
			NUMFREE -= sizeof(SLOT); 
			continue;
			}
		BT_Tuple& tp = Tuple(i);
		int size = tp.Size();
		int endbyte = slot[-i] + size;
		for (int j = slot[-i]; j < endbyte; j++)  {

			//
			//	Assert that tuples do not overlap
			//
			ASSERT1(!map[j]);
			map[j] = 1;
			}
		if (ENDDATA < j) ENDDATA = j;
		NUMFREE -= size + sizeof(SLOT);
		}

	ASSERT1(ENDDATA <= lhCtrl.endData);

	ASSERT1(FreeSpace() == NUMFREE);

	ASSERT1(lhCtrl.endData <= LH_DATASIZE - sizeof(SLOT) * NumSlots());

#endif 
	return TRUE;
}
