/*
 *	$RCSfile: LH_DirSplitUpdate.C,v $
 *	$Revision: 1.1.1.1 $
 *	$Date: 1996/05/04 21:55:25 $
 */
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/

#include "BTREEPAGE.h"
#include "LHSTACK.h"
#include "LH_SMOPages.h"

//
//	Function	: int LH_DirSplitUpdate(TID tid, int volume, 
//						BUFGROUP* bufGroup, LHashStack& dirPl,
//						SHORTPID& const indPid, LH_SMOPages& smo,
//						float oldFromLoad, float fromLoad, float toLoad)
//  Description : Update the directory structure to reflect the split of index
//					page. One more entry is added at the end of directory to
//					hold the new index pid. And the load statistics is changed
//					accordingly.
//

int LH_DirSplitUpdate(TID tid, int volume, BUFGROUP* bufGroup, 
						LHashStack& dirPl, SHORTPID& const indPid,
						LH_SMOPages& smo,
						float oldFromLoad, float fromLoad, float toLoad)
{
	PageDesc rootPd = dirPl.Bottom();
	LHDIRPAGE* rootp = rootPd;

	ASSERT3(rootp->IsRoot());

	int newIndex = (int) rootp->Primary();
	int splitFlag = (newIndex%rootp->Capacity(1) == 0);

	PID addPid;
	addPid.volid = volume;

	BOOL log = TRUE;
#ifndef SERVER_MAKE
	log = GET_LOGLEVEL(LOG_SPACE);
#endif

	if (splitFlag) {
		//
		// need a new directory page to record the new index pid
		//
		if (LH_AllocDirPage(tid, volume, bufGroup, addPid, FALSE,
								rootp->KeyType()))
			return esmFAILURE;
		if (dirPl.ReadNPush(addPid, EX))  
			return esmFAILURE;
		PageDesc newPd = dirPl.Top();
		LHDIRPAGE* newp = newPd;
		newp->SetNextEntry(indPid);
		newPd.Dirty();
		if (log) {
			smo.Insert(addPid);
			LH_LogDirSetNextEntry(newp->SelfID(), newPd, newp->lrc(), indPid, newp->CurPos());
			}
		}
	else {
		addPid.page = indPid;
		}

	int curIndex = newIndex;
	LHDIRPAGE* dirp = rootp;
	int curHeight = dirp->Height();
	PID curPid = addPid;
	int done = FALSE;
	PageDesc curPd = rootPd;

	//
	// add the new entry properly into the directory, expand (split) if
	//	necessary
	//
	while (curHeight>1 && !done) {
		int value = dirp->Capacity(curHeight-1);
		int n = curIndex / value;
		curIndex %= value;
		if (splitFlag && curIndex==0) {
			if (dirp->CurPos() < LH_TBLSIZE) {
				//
				// simply insert the entry in current page
				//
				dirp->SetNextEntry(addPid.page);
				curPd.Dirty();
				if (log) {
					smo.Insert(dirp->SelfID());
					LH_LogDirSetNextEntry(dirp->SelfID(), curPd, dirp->lrc(),
										addPid.page, dirp->CurPos());
					}
				}
			else	{
				//
				// the current root page is splitted to increase height
				//
				if (LH_DirSplit(tid, volume, bufGroup, dirPl, smo,
						curPd, addPid.page))
					return esmFAILURE;
				}
			done = TRUE;
			break;
			}
		curPid.page = dirp->GetEntry(n);
		if (!dirPl.GetEntry(curPid, curPd)) {
			//
			// never read before
			//
			if (dirPl.ReadNPush(curPid, EX))
				return esmFAILURE;
			curPd = dirPl.Top();
			}
		dirp = curPd;
		curHeight = dirp->Height();
		}

	if (!done) {
		//
		// reach the leaf level
		//
		if (splitFlag) {
			if (LH_DirSplit(tid, volume, bufGroup, dirPl,
						smo, curPd, addPid.page))
				return esmFAILURE;
			}
		else {
			dirp->SetNextEntry(addPid.page);
			curPd.Dirty();
			if (log)	{
				smo.Insert(dirp->SelfID());
				LH_LogDirSetNextEntry(dirp->SelfID(), curPd, dirp->lrc(),
										addPid.page, dirp->CurPos());
				}
			}
		}

	//
	// record the index split: update split pointer and load statistics
	//
	rootp->IndSplitUpdate(oldFromLoad, fromLoad, toLoad);
	rootPd.Dirty();
	if (log)
		LH_LogDirUpdate(rootp->SelfID(), rootPd, rootp->lrc(),
							oldFromLoad, fromLoad, toLoad, TRUE);

	return esmNOERROR;
}

