/*
 *	$RCSfile: LH_InsertNewTuple.C,v $
 *	$Revision: 1.1.1.1 $
 *	$Date: 1996/05/04 21:55:25 $
 */
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/
#include "BTREEPAGE.h"
#include "LHSTACK.h"
#include "LH_SMOPages.h"

#define LH_LOCKWANTED(OPER)	((OPER == LH_READ) ? SH : EX)

#ifndef SERVER_MAKE
	extern int LH_BULKLOAD;
#endif

//
//	Function	: 	int LH_InsertNewTuple(TID tid, BUFGROUP* bufGroup, 
//						LHashStack& curPl, BT_Tuple& const tuple, PFC compFunc)
//
//	Description : 	Insert a new <key,Oid> tuple into the specified index page
//						chain.  A overflow page may be allocated to handle the
//						overflow case.
//				  
int LH_InsertNewTuple(TID tid, BUFGROUP* bufGroup, LHashStack& curPl, 
								PageDesc& curPd, TWO& slot, 
								BT_Tuple& const tuple, PFC compFunc)
{
	int	addLength = tuple.Size() + sizeof(SLOT);
	curPd = curPl.Bottom();
	LHINDPAGE* curp = curPd;
	PID	ovPid = curp->SelfID();
	int found;

	BOOL log = TRUE;
#ifndef SERVER_MAKE
	log = GET_LOGLEVEL(LOG_SPACE) && !LH_BULKLOAD;
#endif

	//
	// find the first page on the chain which has enough room for the
	// the new tuple
	//
	while (curp->FreeSpace() < addLength && curp->HaveOv())	{
		ovPid.page = curp->OvPage();
		if (!curPl.GetEntry(ovPid, curPd))
			return esmFAILURE;
		curp = curPd;
	}

	if (curp->FreeSpace() >= addLength)	{
		//
		// current page has enough space
		//
		found = curp->Search(tuple.KeyLen(), tuple.KeyValue(), compFunc, slot);
		ASSERT3(!found);
		curp->InsertTuple(tuple, slot);
		curPd.Dirty();
		}
	else	{
		//
		//	need a overflow page
		//
		PID newOvPid = ovPid;
		BOOL needNewPage = TRUE;

		if (curp->OvPage() != NULLPID) {
			//
			// utilizing old (hidden) empty overflow page if possible
			//
			newOvPid.page = curp->OvPage();
			SHORTPID oldOvPage = curp->OvPage();
			curPl.ReadNPush(newOvPid,EX);
			LHINDPAGE* ovp = curPl.Top();
			if (ovp->OvLabel()) {
				//
				// old overflow page can be used
				//
				curp->SetPageOverflow(newOvPid.page);
				curPd.Dirty();
				needNewPage = FALSE;
				if (log)
					LH_LogSetIndexOverflow(curp->SelfID(), curPd, curp->lrc(),
											newOvPid.page, oldOvPage);
				}
			else {
				//
				// old overflow page was promoted during split
				//
				curPl.DiscardTop();
				}
			}
		if (needNewPage)	{
			//
			// allocate a brand new overflow page
			//	it's also considered as a SMO action, not undoable once
			//	successful.
			//
			LH_SMOPages smo;
			if (log && BT_SetAnchor(tid))  {
				return esmFAILURE;
				}
			if (log)
				smo.Insert(curp->SelfID());
        	if (LH_AllocPage(tid, curp->Volume(), bufGroup, newOvPid, 
								curp->SelfInd(), curp->KeyType(), TRUE))
				return esmFAILURE;

			curp->SetPageOverflow(newOvPid.page);
			curPd.Dirty();
			if (log)
				LH_LogSetIndexOverflow(curp->SelfID(), curPd, curp->lrc(),
										newOvPid.page, NULLPID);
			curPl.ReadNPush(newOvPid,EX);

			if (log) {
				smo.Insert(newOvPid);
				smo.Flush();
				}

			if (log && BT_SetCLR(tid))  {
				return esmFAILURE;
				}
			}

		curPd = curPl.Top();
	    curp = curPd;
		slot = 0; // for new page, newly added is always the first tuple
		//
		// do the insertion
		//
		curp->InsertTuple(tuple, slot);
		curPd.Dirty();
		}

	return	esmNOERROR;
}
