/*
 *	$RCSfile: LH_NewElem.C,v $
 *	$Revision: 1.1.1.1 $
 *	$Date: 1996/05/04 21:55:25 $
 */
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/
#include "BTREEPAGE.h"
#include "LHSTACK.h"
#include "LH_SMOPages.h"

#ifndef SERVER_MAKE
	extern int LH_BULKLOAD;
#endif
//
//	Function	: 	int LH_NewElem(TID tid, BUFGROUP* bufGroup, 
//						PID& const rootPid, LHashStack& curPl, int keyLen,
//						void* keyVal, int elSize, void* const elem, 
//						int maxKeyLen, PFC compFunc)
//
//	Description : 	Insert the new element into the page which contains the 
//						tuple of newKey.  An overflow page may be allocated to
//						handle	the case that not enough room for the Element.
//				  
int LH_NewElem ( TID tid, BUFGROUP* bufGroup, PID& const rootPid, 
						LHashStack& curPl, int keyLen, void* keyVal, 
						TWO elSize, void* const elem, int maxKeyLen,
						PFC compFunc )
{
	PageDesc curPd = curPl.Bottom();
	LHINDPAGE* curp = curPd;
	PID	ovPid = curp->SelfID();
	int	found;
	TWO	slot;

	BOOL log = TRUE;
#ifndef SERVER_MAKE
	log = GET_LOGLEVEL(LOG_SPACE) && !LH_BULKLOAD;
#endif

	found = curp->Search(keyLen, keyVal, compFunc, slot);
	while (!found) {
		ovPid.page = curp->OvPage();
		if (!curPl.GetEntry(ovPid, curPd))
			return esmFAILURE;
		curp = curPd;
		found = curp->Search(keyLen, keyVal, compFunc, slot);
	}

	ASSERT1(found);

	if (curp->FreeSpace() >= elSize)	{
		//
		// we have space for the new element    
		//
		if (curp->AddElem(slot, elem))
			return esmFAILURE;
		curPd.Dirty();
		if (log)
			LH_LogLogicInsert(curp->SelfID(), curPd, curp->lrc(), rootPid,
							keyVal, keyLen, elem, elSize, maxKeyLen, FALSE,
							curp->KeyType(), !curp->OvLabel());
		return esmNOERROR;
		}

	//
	// tough case: target page is full, cannot add Elem to the current page
	//

	//
	// duplicate the old tuple 
	//
	BT_Tuple scratch;
	BCOPY(&(curp->Tuple(slot)), &scratch, curp->Tuple(slot).Size());

	ASSERT3( ! scratch.IsOverflow());
	ASSERT3( ! scratch.WillOverflow());

	//
	// the action is also considered a SMO, thus undoable upon successful
	//	completion
	//
	LH_SMOPages smo;
	if (log && BT_SetAnchor(tid))  {
		return esmFAILURE;
		}
	
	if (log)
		smo.Insert(curp->SelfID());

	//
	// delete the old tuple from the current page it resides in
	//
	if (log)
		LH_LogPhysicDelete(curp->SelfID(), curPd, curp->lrc(), slot, 1);
	curp->RemoveTuple(slot);
	curPd.Dirty();

	int addLength = scratch.Size() + elSize + sizeof(SLOT);

	//
	// find the first page on the chain which has enough room for the
	// the new tuple
	//
	while (curp->FreeSpace() < addLength && curp->HaveOv()) {
		ovPid.page = curp->OvPage();
		if (!curPl.GetEntry(ovPid, curPd))
				return esmFAILURE;
		curp = curPd;
		}

	if (curp->FreeSpace() >= addLength) {
		found = curp->Search(scratch.KeyLen(), scratch.KeyValue(), 
								compFunc, slot);
		ASSERT3(!found);
		curp->InsertTuple(scratch, slot);
		if (log)
			LH_LogPhysicInsert(curp->SelfID(), curPd, curp->lrc(), slot, 1);
		curPd.Dirty();
		}
	else	{
		//
		//  need a overflow page
		//
		PID newOvPid = ovPid;
		BOOL needNewPage = TRUE;

		if (curp->OvPage() != NULLPID) {
			//
			// utilizing old (hidden) empty overflow page if possible
			//
			newOvPid.page = curp->OvPage();
			SHORTPID oldOvPage = curp->OvPage();
			curPl.ReadNPush(newOvPid,EX);
			LHINDPAGE* ovp = curPl.Top();
			if (ovp->OvLabel()) {
				//
				// old overflow page can be used
				//
				curp->SetPageOverflow(newOvPid.page);
				curPd.Dirty();
				needNewPage = FALSE;
				if (log)
					LH_LogSetIndexOverflow(curp->SelfID(), curPd, curp->lrc(),
											newOvPid.page, oldOvPage);
				}
			else {
				//
				// old overflow page was promoted during split
				//
				curPl.DiscardTop();
				}
			}

		if (needNewPage)    {
			//
			// allocate a brand new overflow page
			//
			if (log)
				smo.Insert(curp->SelfID());
			if (LH_AllocPage(tid, curp->Volume(), bufGroup, newOvPid,
								curp->SelfInd(), curp->KeyType(), TRUE))
				return esmFAILURE;

			curp->SetPageOverflow(newOvPid.page);
			curPl.ReadNPush(newOvPid,EX);
			curPd.Dirty();
			if (log)
				LH_LogSetIndexOverflow(curp->SelfID(), curPd, curp->lrc(),
										newOvPid.page, NULLPID);
			curPl.ReadNPush(newOvPid,EX);
			}

		curPd = curPl.Top();
		curp = curPd;
		slot = 0; // for new page, newly added is always the first tuple
		curp->InsertTuple(scratch, slot);
		if (log)
			LH_LogPhysicInsert(curp->SelfID(), curPd, curp->lrc(), slot, 1);
		curPd.Dirty();
		}

	curp = curPd;

	if (log)	{
		smo.Insert(curp->SelfID());
		smo.Flush();
		}
	
	if (log && BT_SetCLR(tid))  {
		return esmFAILURE;
		}

	//
	// now append the new element
	//
	if (curp->AddElem(slot, elem))
		return esmFAILURE;
	curPd.Dirty();

	if (log)
		LH_LogLogicInsert(curp->SelfID(), curPd, curp->lrc(), rootPid, keyVal,
						keyLen, elem, elSize, maxKeyLen, FALSE,
						curp->KeyType(), !curp->OvLabel());

	return	esmNOERROR;
}
