/*
 *	$RCSfile: LH_SplitPage.C,v $
 *	$Revision: 1.1.1.1 $
 *	$Date: 1996/05/04 21:55:25 $
 */
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return
* any improvements or extensions that they make to:
*
*   EXODUS Project Group
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*    or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/

#include "BTREEPAGE.h"
#include "LHSTACK.h"
#include "LH_Hash.h"
#include "OVPAGE.h"
#include "LH_SMOPages.h"
#include "io_extfuncs.h"

extern void IntPrint(int, char*);
extern void (*BT_PrintFunc)(int, char*);

static LHINDPAGE* insTemp = 0;
static LHINDPAGE* delTemp = 0;
static LHINDPAGE** toPage = 0;
static toCount = 0;

static void CleanUp()
{
	if (insTemp) delete insTemp;
	if (delTemp) delete delTemp;

	if (toPage) {
		for (int i=0; i<=toCount; i++)
			delete toPage[i];
		delete[] toPage;
		}
}

#ifdef DEBUG
extern int  rootSplitPerformed;
#endif

//
//  Function    :   int LH_SplitPage(TID tid, LHashStack& dirPl, 
//						BUFGROUP* bufGroup, LH_SMOPages& smo, LHashStack& fromPl,
//						LHashStack& toPl, PFC compFunc)
//
//  Description :   Split the page chain in fromPl to relocate tuples between
//						fromPl and toPl according to their hash values. Also 
//						reflect the structural change in directory as well.
//						Empty overflow pages resulting from split are not
//						deallocated, they are converted to form the new page
//						chain in toPl.
//

int LH_SplitPage(TID tid, LHashStack& dirPl, BUFGROUP* bufGroup, 
					LH_SMOPages& smo, LHashStack& fromPl, LHashStack& toPl,
					PFC compFunc)
{
	LHDIRPAGE*		rootp = dirPl.Bottom();
	LHINDPAGE*		fromPage = fromPl.Bottom();
	SMDATATYPE 		keyType = fromPage->KeyType();
	BOOL log = TRUE;
				
#ifdef DEBUG
	rootSplitPerformed = (fromPl.NumEntries() >= 1);
#endif

#ifdef DEBUG_INSERT
	LH_TRACE(("\n\n===== Split ============\n\n"));
#endif

#ifdef SERVER_MAKE
	TRANSREC* transRec = (TRANSREC*) Active->transRec;
	ASSERT1(transRec->tid == tid);
#else
	log = GET_LOGLEVEL(LOG_SPACE);
#endif

	if (log)
		smo.Insert(rootp->SelfID());

	//
	//	for the hashVal affected: K mod 2*L = K mod L + L
	//
	int		shift = (rootp->SplitPtr()) ? 
						(rootp->Level() - 1): rootp->Level(); 
	FOUR 	newHashVal = fromPage->SelfInd() + (INIT_PRIMARY << shift);
	
	//
	//  the level increments
	//
	int 	newLevel = shift + 1;

	int hashVal;
	int destSlot = 0;
	int fromSlot = 0;
	PID	fromPid = fromPage->SelfID();
	PageDesc fromPd = fromPl.Bottom();

	//	toPage[] are temporary buffers to hold to-page chain  
	toPage = new (LHINDPAGE* [MAXHEIGHT]);
	toCount = 0;

	//
	//	Initialize new page
	//
	toPage[toCount] = new LHINDPAGE;
	toPage[toCount]->Init(fromPid, newHashVal, keyType, FALSE);

	//
	// rehash and split for the first page
	//
	{ for (int i = 0, stop = fromPage->NumSlots(); i < stop; i++)	{
		BT_Tuple& tp = fromPage->Tuple(fromSlot);
		hashVal = LH_Hash(tp.KeyLen(), tp.KeyValue(), keyType, newLevel);
		if (hashVal != fromPage->SelfInd())	{
			//
			// shift to the new page
			//
			fromPage->ShiftTuple(fromSlot, toPage[toCount], destSlot, 1);
			destSlot++;
			}
		else {
			fromSlot++;
			}
		}
	}

	// dummy page for storing all tuples deleted from curFromPage
	// (for generating one log record for all deletions in one page)
	delTemp = new LHINDPAGE;
	toPage[toCount]->CopyData(delTemp);
	int	delSlot = delTemp->NumSlots();	
	
	// dummy page for storing all tuples inserted to curFromPage
	// (for generating one log record for all insertions in one page)
	insTemp = new LHINDPAGE;
	insTemp->Init(fromPage->SelfID(), 0, SM_int, FALSE);
	int	insSlot = 0;	

	LHINDPAGE*  curToPage = toPage[toCount];
	LHINDPAGE*	curFromPage = fromPage;
	LHINDPAGE*  curSplitPage = fromPage;
	PageDesc	curFromPd = fromPd;
	
	//
	// log the deletion for the first page
	//
	if (delSlot > 0) {
		if (log) {
			smo.Insert(fromPid);
			LH_LogPhyLogDelete(curFromPage->SelfID(), curFromPd, delTemp,
								curFromPage->lrc(), delTemp->NumSlots());
			}
		curFromPd.Dirty();
		}
	delTemp->Init(fromPage->SelfID(),0,SM_int,FALSE);
	delSlot = 0;

	//
	// rehash and split for the overflow page
	//
	int freeCnt = 0;
	if (fromPage->HaveOv())	{
		PID 	curFromPid = curFromPage->SelfID(), curSplitPid = curFromPid;
		TWO		newSlot;
		int		newFound;
		
		do	{
			PageDesc curSplitPd;
			curSplitPid.page = curSplitPage->OvPage();
			if (!fromPl.GetEntry(curSplitPid, curSplitPd)) {
				CleanUp();
				return esmFAILURE;
				}
			curSplitPage = curSplitPd;
			fromSlot = 0;
			{ for (int i = 0, stop = curSplitPage->NumSlots(); i < stop; i++)	{
				BT_Tuple& tp = curSplitPage->Tuple(fromSlot);
				int		addLength = tp.Size() + sizeof(SLOT);
				hashVal = LH_Hash(tp.KeyLen(), tp.KeyValue(), 
									keyType, newLevel);
				if (hashVal != curSplitPage->SelfInd())	{	
					//
					//	insert to the new page
					//
					if (curToPage->FreeSpace() < addLength) {
						//
						// out of space in current page, use another page
						//
						toPage[++toCount] = new LHINDPAGE;
						toPage[toCount]->Init(fromPid, newHashVal,
												keyType, FALSE);
						curToPage = toPage[toCount];
						newSlot	= 0;
						}
					else	{
						//
						// find the proper place to insert
						//
						newFound = curToPage->Search(tp.KeyLen(), 
										tp.KeyValue(), compFunc, newSlot);
						ASSERT3(!newFound);
						}
					curSplitPage->ShiftTuple(fromSlot, curToPage, newSlot, 1);
					delTemp->InsertTuple(tp, delSlot);
					delSlot++;
					}
				else	{
					//
					//	insert to the old from-page
					//
					if ((curSplitPage != curFromPage) &&
							(curFromPage->FreeSpace() < addLength))	{
						//
						// curFrom page is full
						//
						if (insSlot > 0) {
							// log the insertion in current page
							if (log) {
								smo.Insert(curFromPage->SelfID());
								LH_LogPhyLogInsert(curFromPage->SelfID(),
											curFromPd, insTemp,
											curFromPage->lrc(), 
											insTemp->NumSlots());
								}
							curFromPd.Dirty();
							insTemp->Init(fromPage->SelfID(),0,SM_int,FALSE);
							insSlot = 0;
							}
						curFromPid.page = curFromPage->OvPage();
						if (!fromPl.GetEntry(curFromPid, curFromPd)) {
							CleanUp();
							return esmFAILURE;
							}
						curFromPage = curFromPd;
						}
					if (curSplitPage != curFromPage) {
						newFound = curFromPage->Search(tp.KeyLen(), 
											tp.KeyValue(), compFunc, newSlot);
						ASSERT3(!newFound);
					    curSplitPage->ShiftTuple(fromSlot, curFromPage, 
														newSlot, 1);
						insTemp->InsertTuple(tp, insSlot);
						insSlot++;
						delTemp->InsertTuple(tp, delSlot);
						delSlot++;
						}
					else	{
						//
						// remain in the old page, no change
						//
						fromSlot++;
						}
					}
				}
			}
			if (delSlot>0) {
				// log the deletion in last page
				if (log) {
					smo.Insert(curSplitPid);
					LH_LogPhyLogDelete(curSplitPage->SelfID(), curSplitPd, 
									delTemp, curSplitPage->lrc(), 
									delTemp->NumSlots());
					}
				curSplitPd.Dirty();
				delTemp->Init(fromPage->SelfID(),0,SM_int,FALSE);
				delSlot = 0;
				}
			}	while (curSplitPage->HaveOv());
		}

	//
	// log the insertion in last from-page
	//
	if (insSlot > 0) {
		if (log)
			LH_LogPhyLogInsert(curFromPage->SelfID(), curFromPd, insTemp,
							curFromPage->lrc(), insTemp->NumSlots());
		curFromPd.Dirty();
		}
	
	if (curSplitPage != curFromPage || curFromPage->OvPage()!= NULLPID)	{
		//
		// convert empty overflow page to page with new hash value.
		//
		PageDesc freePd;
		PID freePid = curFromPage->SelfID();
		freePid.page = curFromPage->OvPage();
		SHORTPID	firstOv = freePid.page;
		LHINDPAGE* curp;

		do {
			LHINDPAGE*	freep;
			if (!fromPl.GetEntry(freePid, freePd)) {
				//
				// it's a hidden overflow page
				//
				fromPl.ReadNPush(freePid,EX);
				freePd = fromPl.Top();
				freep = freePd;
				if (!freep->OvLabel()) {
					//
					// it's already promoted to a primary page
					//
					fromPl.DiscardTop();
					break;
					}
				}
			else	{
				freep = freePd;
				}

			SHORTPID 	ovPid = freep->OvPage();
			toPl.Push(freePd);

			if (log)
				LH_LogOvConvert(freePid, freePd, freep->lrc(),
								freep->SelfInd(), newHashVal, freeCnt==0);
			freep->OvConvert(newHashVal, freeCnt==0);
			//
			// move tuples from temporary buffer to converted page
			//
			curp = toPage[freeCnt++];
			if (curp->NumSlots()>0) {
				curp->ShiftTuple(0, freep, 0, curp->NumSlots());
				if (log)
					LH_LogPhysicInsert(freePid, freePd, freep->lrc(),
										0, freep->NumSlots());
				}
			fromPl.DiscardEntry(freePid);
			freePid.page = ovPid;
			} while (freePid.page != NULLPID && toCount >= freeCnt );
			
		//
		//	reset the last page
		//
		curFromPage->ResetPageOverflow(TRUE);
		if (!fromPl.GetEntry(curFromPage->SelfID(), fromPd)) {
			CleanUp();
			return esmFAILURE;
			}
		if (log)
			LH_LogResetIndexOverflow(curFromPage->SelfID(), fromPd, 
									curFromPage->lrc(), firstOv, TRUE);
		fromPd.Dirty();
		}
	

	if (toCount >= freeCnt) {
		//
		// new page needs to be allocated
		//
		PID     toPid;
		if (LH_AllocPage(tid, curFromPage->Volume(), bufGroup, toPid, 
							newHashVal, keyType, (freeCnt!=0))) {
			CleanUp();
			return esmFAILURE;
			}
		if (log) smo.Insert(toPid);
		if (freeCnt > 0) {
			//
			// connect to the converted page
			//
			curToPage = toPl.Top();
			curToPage->SetPageOverflow(toPid.page);
			if (log)
				LH_LogSetIndexOverflow(curToPage->SelfID(), toPl.Top(),
									curToPage->lrc(), toPid.page, NULLPID);
			}

		toPl.ReadNPush(toPid,EX);
		int j = freeCnt;
		do {
			LHINDPAGE*  curp = toPage[j++];
			curToPage = toPl.Top();
			if (curp->NumSlots()>0) {
				curp->ShiftTuple(0, curToPage, 0, curp->NumSlots());
				if (log)
					LH_LogPhysicInsert(toPid, toPl.Top(), curToPage->lrc(),
										0, curToPage->NumSlots());
				}
			if (j <= toCount)	{
				if (LH_AllocPage(tid, curFromPage->Volume(), bufGroup, toPid,
								newHashVal, keyType, TRUE)) {
					CleanUp();
					return esmFAILURE;
					}
				curToPage->SetPageOverflow(toPid.page);
				if (log)
					LH_LogSetIndexOverflow(curToPage->SelfID(), toPl.Top(),
									curToPage->lrc(), toPid.page, NULLPID);
				toPl.ReadNPush(toPid,EX);
				}
			} while (j <= toCount);
		}

	//
	// delete all temporary buffers
	//
	CleanUp();

	//
	// load information change
	//
	float oldFromLoad = ((LHINDPAGE *)fromPl.Bottom())->Load();
	float fromLoad = ((LHINDPAGE *)fromPl.Bottom())->LoadComp();
	float toLoad = ((LHINDPAGE *)toPl.Bottom())->LoadComp();

	if (oldFromLoad!=fromLoad)
		((LHINDPAGE *)fromPl.Bottom())->SetLoad(fromLoad);
	((LHINDPAGE *)toPl.Bottom())->SetLoad(toLoad);

	//
	//	Mark all pages in toPl dirty
	//
	toPl.DirtyAll();

	//
	//	Change the directory page and log the change 
	//
	if (LH_DirSplitUpdate(tid, rootp->SelfID().volid, bufGroup, dirPl,
						((LHINDPAGE*) toPl.Bottom())->SelfID().page, smo,
						oldFromLoad, fromLoad, toLoad)) 
		return esmFAILURE;

#ifdef DEBUG_INSERT
	LH_TRACE(("\n\n===== After Split========\n\n"));
#endif

	return esmNOERROR;
}
