/*
 *  $RCSfile: lh_BulkLoad.C,v $
 *  $Revision: 1.1.1.1 $
 *  $Date: 1996/05/04 21:55:25 $
 */
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*		     Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*    or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/

#include "BTREEPAGE.h"
#include "LHSTACK.h"
#include "LH_Hash.h"
#include "OVPAGE.h"

#include "sm_extfuncs.h"

extern int LH_BULKLOAD;

static LH_GetNextElement(SCANDESC* scanDesc, int& hashVal, KEY& key,
		    void** elem, SMDATATYPE keyType, 
		    int level, int primary, BOOL& eof);

static LH_AddEntryToDir(TID tid, LHashStack& dirPl, PID& const newPid);

static LH_LoadTuple(TID tid, PID& const rootPid, LHashStack& curPl, 
			void* keyVal, int keyLen, int elSize, 
			void* const elem, int maxKeyLen, int unique,
			PFC compFunc);
//
//  Function	: int lh_BulkLoad(TID tid, int volume, INDEXDESC*  ndxDesc, 
//		    FID& fid, BUFGROUP* bufGroup, int level, int primary)
//  Description : Bulk load temporary <key, oid> file into new hash index.
//

int lh_BulkLoad(TID tid, int volume, INDEXDESC*	 ndxDesc, FID& fid, 
		    BUFGROUP* bufGroup, int cnt, int level, int primary)
{
    SMDATATYPE keyType = ndxDesc->dataType;
    int elSize = ndxDesc->elSize;
    BOOL    unique = ndxDesc->unique;
    PFC compFunc = SMCOMPFUNC[keyType];

    int prevHashVal = -1;
    float loadSum = 0.0, load;
    LHashStack curPl(bufGroup), dirPl(bufGroup);
    LHINDPAGE* curp;
    PID newPid;
    BOOL eof;

    LH_BULKLOAD = TRUE;

    //
    //	    Allocate the root directory page 
    //
//     if (LH_AllocDirPage(tid, volume, bufGroup, ndxDesc->rootPid,
// 		TRUE, keyType)) {
// 	return esmFAILURE;
// 	}
    //
    // read in the root directory page
    //
    LHDIRPAGE*	rootp;
    if (dirPl.ReadNPush(ndxDesc->rootPid, EX)) {
    	LH_BULKLOAD = FALSE;
	return esmFAILURE;
	}
    rootp = dirPl.Top();

    //
    // create an undo only log on root page for control info
    //
    LH_LogDirRootCtrl(rootp->SelfID(), dirPl.Top(), rootp->lrc(), TRUE);

    //
    // allocate original primary pages
    //
    PID initPid[INIT_PRIMARY];
    for (int i = 0; i< INIT_PRIMARY; i++) {
 	if (LH_AllocPage(tid, volume, bufGroup, initPid[i],
 			    i, keyType, FALSE)) {
 	    LH_BULKLOAD = FALSE;
 	    return esmFAILURE;
 	    }
 	rootp->SetEntry(initPid[i].page, i);
	initPid[i].volid = volume;
	initPid[i].page = rootp->GetEntry(i);
	}

    //
    // open the scan of temporary file sorted on hash value
    //
    SCANDESC	*scanDesc;
    int		curHashVal;
    KEY		key;
    void*	elem;

    if (sm_OpenScan(&fid, LOGICAL_SCAN, 20, &scanDesc, NULL)) {
 	LH_BULKLOAD = FALSE;
	return esmFAILURE;
	}

    while (--cnt >= 0) {
	if (LH_GetNextElement(scanDesc, curHashVal, key, &elem, 
			      keyType, level, primary, eof)) {
 	    LH_BULKLOAD = FALSE;
	    return esmFAILURE;
	    }
	ASSERT1(! eof);
	ASSERT3(curHashVal >= prevHashVal);

	if (curHashVal != prevHashVal) {
	    //
	    // begin of a new index page
	    //
	    if (prevHashVal != -1) {
		//
		// set load for the last index page 
		//
		curp = curPl.Bottom();
		load = curp->LoadComp();
		loadSum += load;
		curp->SetLoad(load);
		//
		// finish up with the old index page
		// 
		curPl.DirtyAll();
		curPl.ReleaseAll();
		curPl.Empty();
		}
	    //
	    // for any skipped hash values, still need empty
	    // index pages for them
	    //
	    for (i = prevHashVal + 1; i <  curHashVal; i++) {
		if (i >= INIT_PRIMARY) {
		    if (LH_AllocPage(tid, volume, bufGroup, newPid, 
					curHashVal, keyType, FALSE)) {
 	    		LH_BULKLOAD = FALSE;
			return esmFAILURE;
			}
		    if (LH_AddEntryToDir(tid, dirPl, newPid)) {
 	    		LH_BULKLOAD = FALSE;
			return esmFAILURE;
			}
		    }
		}
	    //
	    // put a new index page to hash stack
	    //
	    if (curHashVal < INIT_PRIMARY) {
		//
		// can use the initial allocated page
		//
		newPid = initPid[curHashVal];
		}
	    else {
		//
		// allocate a new index page
		//
		if (LH_AllocPage(tid, volume, bufGroup, newPid, curHashVal,
							keyType, FALSE)) {
 	    	    LH_BULKLOAD = FALSE;
		    return esmFAILURE;
		    }
		//
		// add the new page id into directory
		//
		if (LH_AddEntryToDir(tid, dirPl, newPid)) {
 	    	    LH_BULKLOAD = FALSE;
		    return esmFAILURE;
		    }
		//
		// reflect the structure change in root page
		// (actual load will be set at the end)
		//
		rootp->IndSplitUpdate(0.0, 0.0, 0.0);
		}
	    //
	    // push it on the hash stack
	    //
	    curPl.ReadNPush(newPid, EX);
	    prevHashVal = curHashVal;
	    }
	//
	// add the tuple
	//
	if (LH_LoadTuple(tid, ndxDesc->rootPid, curPl, key.valuePtr,
			 key.length, elSize, elem,
			 ndxDesc->maxKeyLen,  unique, compFunc)) {
 	    LH_BULKLOAD = FALSE;
	    return esmFAILURE;
	    }
    }
    //
    // reach end of file
    //
    if (prevHashVal != -1) {
	//
	// set load for the last index page 
	//
	curp = curPl.Bottom();
	load = curp->LoadComp();
	loadSum += load;
	curp->SetLoad(load);
	//
	// finish up with the old index page
	// 
	curPl.DirtyAll();
	curPl.ReleaseAll();
	curPl.Empty();
	//
	// set entire load
	//
	prevHashVal++;
	loadSum = prevHashVal ? loadSum / prevHashVal : 0.0;
	rootp->SetLoad(loadSum);

	//
	// create a redo only log record for both control and data information
	//
    	LH_LogDirRootCtrl(rootp->SelfID(), dirPl.Bottom(), rootp->lrc(), FALSE);
	LH_LogDirSplitCopy(rootp->SelfID(), dirPl.Bottom(), rootp->lrc(), 
				rootp->Height(), rootp->CurPos()); 

	//
	// finish up with directory pages
	//
	dirPl.DirtyAll();
	dirPl.ReleaseAll();
	}

    if (sm_CloseScan(scanDesc)) {
	LH_BULKLOAD = FALSE;
	return esmFAILURE;
	}

    LH_BULKLOAD = FALSE;
    return esmNOERROR;
}

int LH_AddEntryToDir(TID tid, LHashStack& dirPl, PID& const newPid)
{
    LHDIRPAGE* rootp = dirPl.Bottom();
    PID rootPid = rootp->SelfID();
    LHDIRPAGE* dirp = dirPl.Top();
    PageDesc dirPd;
    BUFGROUP* bufGroup = dirPl.BufGroup();
    
    BOOL full = (dirp->CurPos() == LH_TBLSIZE);
    
    if (!full) {
	dirp->SetNextEntry(newPid.page);
	return esmNOERROR;
	}
    //
    // a new leaf directory page is needed
    //
    PID addPid;
    PageDesc addPd;
    LHDIRPAGE* addp;

    if (LH_AllocDirPage(tid, rootPid.volid, bufGroup, addPid, FALSE,
			    rootp->KeyType()))
	return esmFAILURE;

    if (addPd.Read(bufGroup, addPid, EX))
	return esmFAILURE;
    
    addp = addPd;
    addp->SetNextEntry(newPid.page);
    
    //
    // find the right directory page to insert the new directory entry
    //
    PID dirPid = dirp->SelfID();

    while (full && (dirPid != rootPid)) {
	//
	// release the leaf directory page
	//
	dirPd = dirPl.Pop();
	dirPd.UnfixUnlockDirty();
	//
	// check the next one on stack
	//
	dirp = dirPl.Top();
	dirPid = dirp->SelfID();
	full = (dirp->CurPos() == LH_TBLSIZE);
	}
	
    PID childPid;
    if (full && dirPid==rootPid) {
    	//
    	// a split is needed in directory rooted at dirp
    	//
    	PageDesc childPd;
    	if (LH_AllocDirPage(tid, rootPid.volid, bufGroup, childPid, FALSE,
			    dirp->KeyType()))
	    return esmFAILURE;

    	if (childPd.Read(bufGroup, childPid, EX))
	    return esmFAILURE;

    	LHDIRPAGE* childp = childPd;
    	//
    	// copy everything in current directory page to new child
    	// and release the new child page
    	//
    	dirp->CopyData(childp);
    	childPd.UnfixUnlockDirty();
    
    	//
    	// reset some of the parameters and add the new entry
    	//
    	dirp->DirSplitUpdate();
    	dirp->SetNextEntry(childPid.page);
	}
     
    int	 height = dirp->Height();
    while (height>2) {
	//
	// add intermediate nodes
	//

	if (LH_AllocDirPage(tid, rootPid.volid, bufGroup, childPid, FALSE,
			    dirp->KeyType()))
	    return esmFAILURE;

    	//
    	// push the new directory page in stack
    	//
    	dirPl.ReadNPush(childPid, EX);
    	dirp->SetNextEntry(childPid.page);

	dirp = dirPl.Top();
	dirp->SetHeight(--height);
	}
    
    //
    // now we find the right page to insert the new directory entry
    //
    dirp->SetNextEntry(addPid.page);
    dirPl.Push(addPd);
    return esmNOERROR;
}


int LH_GetNextElement(SCANDESC* scanDesc, int& hashVal, KEY& key, void** elem,
			SMDATATYPE keyType, int level, 
			int primary, BOOL& eof)
{
    USERDESC	    *userDesc;

    if (sm_ScanNextObject(scanDesc, 0, -1, &userDesc, &eof))
	return esmFAILURE;

    if (!eof)	{
	char* p = userDesc->basePtr;
	//
	// get key
	//
	key.length = * (int*) p; p += ALIGNSIZE(sizeof(int));
	key.valuePtr = p; p += ALIGNSIZE(key.length);
	//
	// get elem
	//
	*elem = p;
	//
	// get hashVal
	//
	hashVal = LH_Hash(key.length, key.valuePtr, keyType, level);
	if (hashVal >= primary)
	    hashVal = LH_Hash(key.length, key.valuePtr, keyType, level-1);
	}

    return esmNOERROR;
}


int LH_LoadTuple(TID tid, PID& const rootPid, LHashStack& curPl, void* keyVal, 
		    int keyLen, int elSize, void* const elem, 
		    int maxKeyLen, int unique, PFC compFunc)
{
    TWO slot;
    PageDesc curPd;
    LHINDPAGE* indp;
    BUFGROUP* bufGroup = curPl.BufGroup(); 

    BT_Tuple tuple;
    int length;
    int found;

    //
    // create a temporary tuple
    //
    tuple.Create(keyLen, keyVal, elem, elSize);
    length = tuple.Size();

    //
    // search the current chain for the new key
    //
    curPd = curPl.Bottom();
    indp = curPd;
    found = indp->Search(keyLen, keyVal, compFunc, slot);
    
    //
    // check the overflow page if unsuccessful
    //
    PID		ovPid = indp->SelfID();
    while (!found && indp->HaveOv()) {
	ovPid.page = indp->OvPage();
	ASSERT3( ovPid.page != NULLPID );
	if (!curPl.GetEntry(ovPid, curPd))
	    return esmFAILURE;
	indp = curPd;
	found = indp->Search(keyLen, keyVal, compFunc, slot);
	}
    //
    // load the new <key, elem>
    //
    if (found) {
	if (unique) {
	    return esmFAILURE;
	}
	
	BT_Tuple& tp = indp->Tuple(slot);
	
	if (tp.WillOverflow())	{
	    if (ov_NewCluster(tid, rootPid.volid, bufGroup, elSize,
				    ovPid, tp)) {
		return esmFAILURE;
		}
	    indp->SetOverflow(slot, ovPid);
	    ASSERT3(tp.IsOverflow());
	    }

	if (tp.IsOverflow())  {
	    if (ov_AddElem(tid, tp.OverflowRoot(), bufGroup, elem)) {
		return esmFAILURE;
		}
	    tp.IncrElCnt();
	    return esmNOERROR;
	    }
	//
	//  Inserting an element into an existing key	    
	//
	if (LH_NewElem(tid, bufGroup, rootPid, curPl, keyLen, keyVal,
			    elSize, elem, maxKeyLen, compFunc)) {
	    return esmFAILURE;
	    }
	}
    else {  
	if (LH_InsertNewTuple(tid, bufGroup, curPl, curPd, slot,
						tuple, compFunc)) {
	    return esmFAILURE;
	    }
	}
    return esmNOERROR;
}
