/*
 *	$RCSfile: lh_InsertEntry.C,v $
 *	$Revision: 1.1.1.1 $
 *	$Date: 1996/05/04 21:55:25 $
 */

/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/

#include "BTREEPAGE.h"
#include "LHSTACK.h"
#include "LH_Hash.h"
#include "OVPAGE.h"
#include "LH_SMOPages.h"

extern void IntPrint(int, char*);
extern void (*BT_PrintFunc)(int, char*);

int LH_AllocPrimePage(TID tid, int volume, BUFGROUP* bufGroup, 
						PageDesc* const rootPd, void* keyVal, int keyLen,
						SHORTPID& indPid)
{
	BOOL log = TRUE;

#ifndef SERVER_MAKE
	log = GET_LOGLEVEL(LOG_SPACE);
#endif

	LHDIRPAGE* rootp = *rootPd;
	int hashVal = rootp->IndexCalc(keyLen, keyVal);
	ASSERT1( hashVal < INIT_PRIMARY );
		
	LH_SMOPages smo;

	if (log) {
		if (BT_SetAnchor(tid))  {
			return esmFAILURE;
			}
		}

	PID	newPid;
	if (LH_AllocPage(tid, volume, bufGroup, newPid, hashVal, 
		rootp->KeyType(), FALSE)) {
		return esmFAILURE;
		}

	rootp->SetEntry(newPid.page, hashVal);
	if (log) 
		LH_LogDirSetNextEntry(rootp->SelfID(), *rootPd, rootp->lrc(), 
						newPid.page, hashVal);
	rootPd->Dirty();

	indPid = newPid.page;
	
	smo.Insert(rootp->SelfID());
	smo.Insert(newPid);

	if (log)
		smo.Flush();

	if (log)
		if (BT_SetCLR(tid))  {
			return esmFAILURE;
			}

	return esmNOERROR;
}


//
//	Function 	: int lh_InsertEntry(PID& const rootPid, int groupIndex, 
//									KEY* newKey, int newKeyLen, int elSize,
//									void* const elem, int maxKeyLen, 
//									int unique, PFC compFunc, BOOL loadCalc)
//	Description	: Insert <newKey,newOid> into the index.  loadCalc flag is
//					to control the update of global load statistics. It's 
//					always turned on for client calls, but can be turned off
//					during undo operation at server.
//

int lh_InsertEntry(TID tid, PID& const rootPid, BUFGROUP* bufGroup, 
						void* newKey, int newKeyLen, 
						int elSize, void* const elem, int maxKeyLen, 
						int unique, PFC compFunc, BOOL loadCalc)
{
	TWO			slot;
	int			keyExists;
    PageDesc    curPd, rootPd;
	LHashStack  dirPl(bufGroup), curPl(bufGroup);
	PID         indPid = rootPid;

	BOOL log = TRUE;

#ifndef SERVER_MAKE
	log = GET_LOGLEVEL(LOG_SPACE);
#endif

	//
	//  Traverse the directory first
	//
	if (LH_DirSearch(rootPid, newKeyLen, newKey, dirPl, indPid.page, LH_INSERT))
		return esmFAILURE;

	rootPd = dirPl.Bottom();

	if (indPid.page==NULLPID) {
		//
		// a new index page needs to be allocated. (when new index created,
		// only the root page is allocated, the original primary pages are
		// not allocated.  now is the time to allocate them on demand.)
		//
		if (LH_AllocPrimePage(tid, rootPid.volid, bufGroup, &rootPd, newKey, 
								newKeyLen, indPid.page))
			return esmFAILURE;
		}

	//
	//	Find the index page 
	//
	if (LH_IndSearch(indPid, newKeyLen, newKey, curPd, curPl, keyExists,
							slot, LH_INSERT, compFunc))  
		return esmFAILURE;

	LHDIRPAGE* rootp = rootPd;
	LHINDPAGE* curp = curPd;
	LHINDPAGE* indp = curPl.Bottom();

	BT_Tuple buffer;
	int length;
	
	if (!keyExists)  {
		// 	---- Construct a brand new <key-value, elem> tuple in local buffer
		buffer.Create(newKeyLen, newKey, elem, elSize);
		length = buffer.Size();
		}
	else {
		if (unique)		{
			SM_ERROR(TYPE_USER, esmKEYALREADYEXISTS);
			return esmFAILURE;
			}
		BT_Tuple& tp = curp->Tuple(slot);
		int loadChange = FALSE;
		
		if (tp.WillOverflow())  {
			PID ovPid;
			if (ov_NewCluster(tid, rootPid.volid, bufGroup, elSize,
									ovPid, tp)) {
				return esmFAILURE;
				}
			if (log)
				LH_LogLogicSetOverflow(curp->SelfID(), curPd, curp->lrc(), 
					rootPid, newKey, newKeyLen, maxKeyLen, rootp->KeyType(), 
					tp.ElCnt(), elSize, tp.ElArray(0), ovPid);
			curp->SetOverflow(slot, ovPid);
			ASSERT3(tp.IsOverflow());
			if (!curp->OvLabel()) {
				loadChange = TRUE;
				}
			}

		if (tp.IsOverflow())  {
			if (ov_AddElem(tid, tp.OverflowRoot(), bufGroup, elem)) {
				return esmFAILURE;
				}
			tp.IncrElCnt();
			if (log)
				LH_LogLogicIncrElCnt(curp->SelfID(), curPd, curp->lrc(),
							rootPid, newKey, newKeyLen, 
							maxKeyLen, rootp->KeyType());
			if (loadChange) {
				//
				// for primary page, needs to record the load change caused
				//	by the tuple overflow
				//
				float newLoad = curp->LoadComp();
				if (loadCalc)	{
					rootp->LoadUpdate(curp->Load(),newLoad);
					if (log)
						LH_LogDirLoadUpdate(rootPid, rootPd, rootp->lrc(), 
										curp->Load(), newLoad);
					rootPd.Dirty();
					}
				curp->SetLoad(newLoad);
				}
			curPd.Dirty();
			return esmNOERROR;
			}

		//	---- Inserting an element into an existing key		
		length = elSize;
		}

	//
	//  flag to record whether the new entry belong to the
	//	page to be splitted.
	//
	int splitHit = FALSE;
	
	//
	//	if there is not enough room for the	new stuff, 
	//		record a collision, split the necessary page,
	//		if not same page, insert the stuff to overflow page
	//

	if (indp->FreeSpace() < length + sizeof(SLOT) 
		&& rootp->Load() > rootp->Threshold())  {
		//
		//	collision
		//
		LHashStack	toPl(bufGroup);
		LHashStack* fromPl;

		FOUR	oldHashVal = indp->SelfInd();
		LH_SMOPages smo;

#ifdef DEBUG_INSERT
		LH_TRACE(("linear hashing:  collision\n"));
#endif
		//
		//	Set anchor for Compensation Log Record after split
		//
		if (log) {
			if (BT_SetAnchor(tid))  {
				return esmFAILURE;
				}
			}

		splitHit = ( rootp->SplitPtr() == oldHashVal );

		if (splitHit) {
			fromPl = &curPl;
			}
		else {
			//
			//  get the page id of split page
			//
			PID splitPid = rootPid;
			PageDesc	splitPd;
			LHDIRPAGE* dirp; 
			int curIndex = (int) rootp->SplitPtr();
			int curHeight = rootp->Height();

			//
			// traverse the path to find the index page entry
			//
			dirp = rootp;
			while (curHeight!=1) {
				int cap = dirp->Capacity(curHeight-1);
				int n = curIndex / cap;
				curIndex %= cap;
				splitPid.page = dirp->GetEntry(n);
				if (dirPl.GetEntry(splitPid, splitPd))
					dirp = splitPd;
				else	{
					if (dirPl.ReadNPush(splitPid, SH))
						return esmFAILURE;
					dirp = dirPl.Top();
					}
				curHeight = dirp->Height();
				}
			splitPid.page = dirp->GetEntry(curIndex);

			//
			//  read in all the split page list in EX mode
			//
			fromPl = new LHashStack(bufGroup);
			fromPl->ReadNPush(splitPid, EX);
			while	(((LHINDPAGE*) fromPl->Top())->HaveOv())	{
				splitPid.page = ((LHINDPAGE*) fromPl->Top())->OvPage();
				fromPl->ReadNPush(splitPid, EX);
				}
			}	

		//
		// split the index page
		//
		if (LH_SplitPage(tid, dirPl, bufGroup, smo, (*fromPl), toPl, 
							compFunc)) {
			if (!splitHit)
				delete fromPl;
			return esmFAILURE;
			}

		//
		// flush all pages in LH_SMO
		//
		if (log)
			smo.Flush();


	 	//
		//  Write compensation log record 
		//
		if (log)
			if (BT_SetCLR(tid))  {
				if (!splitHit)
					delete fromPl;
				return esmFAILURE;
				}

		//	
		//	decide where the new entry belongs
		//
		LHashStack* insPl;
		if (splitHit)	{
			FOUR hashVal = (FOUR) LH_Hash(maxKeyLen, newKey, rootp->KeyType(),
										rootp->Level());
			if (hashVal == oldHashVal)	{
				insPl = fromPl;
				}
			else {
				insPl = &toPl;
				}
			}
		else
			insPl = &curPl;

		if (!keyExists)	{
			if (LH_InsertNewTuple(tid, bufGroup, (*insPl), curPd, slot, 
									buffer, compFunc)) {
				if (!splitHit)
					delete fromPl;
				return esmFAILURE;
				}
			curp = curPd;
			if (log)
				LH_LogLogicInsert(curp->SelfID(), curPd, curp->lrc(), rootPid, 
						newKey, newKeyLen, elem, elSize, maxKeyLen, unique, 
						rootp->KeyType(), !curp->OvLabel());
			}
		else {
			//
			// key exists, has to find the right page to insert.
			// log is done within the function call.
			//
			if (LH_NewElem(tid, bufGroup, rootPid, (*insPl), newKeyLen, 
								newKey, elSize, elem, maxKeyLen, compFunc)) {
				if (!splitHit)
					delete fromPl;
				return esmFAILURE;
				}
			}
		
		if (!curp->OvLabel()) {
			//
			//	Update Load Information if primary page involved
			//
			float newLoad = curp->LoadComp();
			if (loadCalc)	{
				rootp->LoadUpdate(curp->Load(),newLoad);
				if (log)
					LH_LogDirLoadUpdate(rootPid, rootPd, rootp->lrc(), 
										curp->Load(), newLoad);
				rootPd.Dirty();
				}
			curp->SetLoad(newLoad);
			}
		
		//
		// in the case of not splithit, delete the new stack
		//
		if (!splitHit)
			delete fromPl;

		//
		// release directory page
		//
		return esmNOERROR;
		}

	//
	//	no split was done
	//
	if (!keyExists)	{
		if (LH_InsertNewTuple(tid, bufGroup, curPl, curPd, slot, 
									buffer, compFunc)) {
			return esmFAILURE;
			}
		curp = curPd;
		if (log)
			LH_LogLogicInsert(curp->SelfID(), curPd, curp->lrc(), rootPid, 
						newKey, newKeyLen, elem, elSize, maxKeyLen, unique, 
						rootp->KeyType(), !curp->OvLabel());
		}
	else {
		if (LH_NewElem(tid, bufGroup, rootPid, curPl, newKeyLen, 
							newKey, elSize, elem, maxKeyLen, compFunc)) {
			return esmFAILURE;
			}
		}
		
	if (!curp->OvLabel()) {
		//
		//	Update Load Information if primary page involved
		//
		float newLoad = curp->LoadComp();

		if (loadCalc)	{
			rootp->LoadUpdate(curp->Load(),newLoad);
			if (log)
				LH_LogDirLoadUpdate(rootPid, rootPd, rootp->lrc(), 
									curp->Load(), newLoad);
			rootPd.Dirty();
			}
		curp->SetLoad(newLoad);
		}

	return esmNOERROR;
}
