/*
 *	$RCSfile: lh_RemoveEntry.C,v $
 *	$Revision: 1.1.1.1 $
 *	$Date: 1996/05/04 21:55:25 $
 */
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/
#include "BTREEPAGE.h"
#include "LHSTACK.h"
#include "OVPAGE.h"
#include "io_extfuncs.h"


//
//	Function	:	int lh_RemoveEntry(TID tid, PID& const rootPid,
//						BUFGROUP* bufGroup, void* doomedKey, int doomedKeyLen,
//						int elSize, void* const doomedElem, int maxKeyLen, 
//						int unique, PFC compFunc, BOOL loadCalc)
//	Description	: Remove <doomedKey, doomedElem> from the index page
//
int lh_RemoveEntry(TID tid, PID& const rootPid, BUFGROUP* bufGroup, 
					void* doomedKey, int doomedKeyLen,
					int elSize, void* const doomedElem, int maxKeyLen, 
					int unique, PFC compFunc, BOOL loadCalc)
{
	TWO		slot;
	int		found;
    PageDesc    curPd, rootPd;
	LHashStack  dirPl(bufGroup), curPl(bufGroup);
	PID         indPid = rootPid;

	BOOL log = TRUE;

#ifdef SERVER_MAKE
	TRANSREC* transRec = (TRANSREC*) Active->transRec;
	ASSERT1(transRec->tid == tid);
#else
	log = GET_LOGLEVEL(LOG_SPACE);
#endif

    //
	//  Traverse the directory first
	//
	if (LH_DirSearch(rootPid, doomedKeyLen, doomedKey, dirPl,
						indPid.page, LH_DELETE))
		return esmFAILURE;

	rootPd = dirPl.Bottom();
	LHDIRPAGE* rootp = rootPd;

	if (indPid.page==NULLPID) {
		//
		// fall into some original unallocated page
		//
		ASSERT1( rootp->Primary()==INIT_PRIMARY );
		SM_ERROR(TYPE_USER, esmKEYNOTFOUND);
		return esmFAILURE;
		}
		
	//
	//	find the index page
	//
	if (LH_IndSearch(indPid, doomedKeyLen, doomedKey, curPd, curPl,
							found, slot, LH_DELETE, compFunc))  
		return esmFAILURE;

	if (!found)	 {
		SM_ERROR(TYPE_USER, esmKEYNOTFOUND);
		return esmFAILURE;
		}

	LHINDPAGE* curp = curPd;
	
	BT_Tuple& tp = curp->Tuple(slot);
	
	//
	//	If overflow, remove from overflow pages
	//
	if (tp.IsOverflow())  {
		if (ov_RemoveElem(tid, tp.OverflowRoot(), 
					bufGroup, doomedElem)) 
			return esmFAILURE;
		tp.DecrElCnt();
		if (log)
			LH_LogLogicDecrElCnt(curp->SelfID(), curPd, curp->lrc(),
									rootPid, doomedKey, doomedKeyLen, maxKeyLen,
									curp->KeyType());
		if (curp->Tuple(slot).ElCnt() == 0)  {
			if (ov_DestroyCluster(tid, tp.OverflowRoot(), bufGroup)) {
				return esmFAILURE;
				}
			if (log)
				LH_LogPhysicDelete(curp->SelfID(), curPd, curp->lrc(), slot, 1);
			curp->RemoveTuple(slot);
			if (!curp->OvLabel()) {
		 		//
				// for primary page, needs to record the load change 
				//
				float newLoad = curp->LoadComp();
				if (loadCalc) {
					rootp->LoadUpdate(curp->Load(), newLoad);
					if (log)
						LH_LogDirLoadUpdate(rootPid, rootPd, rootp->lrc(), 
										curp->Load(), newLoad);
					rootPd.Dirty();
					}
				curp->SetLoad(newLoad);
				}
			}
		curPd.Dirty();

		return esmNOERROR;
		}
	
	//
	//	Remove the oid
	//
	if (curp->RemoveElem(slot, doomedElem)) {
		return esmFAILURE;
		}

	//
	//	Remove the tuple if there isn't any more oid in it
	//
	if (curp->Tuple(slot).ElCnt() == 0)   {
		curp->RemoveTuple(slot);
		}

	//
	//	Log the deletion
	//
	if (log)
		LH_LogLogicDelete(curp->SelfID(), curPd, curp->lrc(), 
						rootPid, doomedKey, doomedKeyLen, doomedElem, elSize,
						maxKeyLen, unique, curp->KeyType(), !curp->OvLabel());

	//
	//  hide the resulting empty overflow page if necessary
	//
	if (curp->OvLabel() && !curp->HaveOv() && curp->NumSlots()==0) {
		PageDesc  prevPd;
		if (!curPl.GetPrevEntry(curp->SelfID(),prevPd))
			return esmFAILURE;
		LHINDPAGE* prevp = prevPd;
		prevp->ResetPageOverflow(TRUE);
		if (log)
			LH_LogResetIndexOverflow(prevp->SelfID(),prevPd, prevp->lrc(),
									curp->SelfID().page, TRUE);
		prevPd.Dirty();
		}
			
	if (!curp->OvLabel()) {
 		//
		// for primary page, needs to record the load change caused
		//  by the tuple overflow
		//
		float newLoad = curp->LoadComp();
		if (loadCalc) {
			rootp->LoadUpdate(curp->Load(), newLoad);
			if (log)
				LH_LogDirLoadUpdate(rootPid, rootPd, rootp->lrc(), 
										curp->Load(), newLoad);
			rootPd.Dirty();
			}
		curp->SetLoad(newLoad);
		}

	curPd.Dirty();

	return esmNOERROR;
}
