/*
 *   $RCSfile: dist2PC.c,v $  
 *   $Revision: 1.1.1.1 $  
 *   $Date: 1996/05/04 21:55:28 $      
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/

#include "sysdefs.h"
#include "ess.h"
#include "checking.h"
#include "error.h"
#include "trace.h"
#include "io.h"
#include "page.h"
#include "list.h"
#include "object.h"
#include "lsn.h"
#include "tid.h"
#include "bf.h"
#include "clog.h"
#include "logaction.h"
#include "loginfo.h"
#include "logrecs.h"
#include "scan.h"
#include "load.h"
#include "sm_params.h"
#include "btree.h"
#include "trans.h"
#include "lock.h"
#include "msgdefs.h"
#include "serverinfo.h"
#include "distr.h"
#include "sm_state.h"
#include "sm_intfuncs.h"
#include "sm_extfuncs.h"
#include "trans_funcs.h"
#include "msg_funcs.h"
#include "clog_extfuncs.h"
#include "trans_globals.h"
#include "msg_globals.h"
#include "io_globals.h"
#include "sm_globals.h"

static BOOL
chooseServerInfo(
	SERVERINFO      *serverInfo,
	FLAGS           *choice,
	SERVERINFO      **result
) 
{
	TRPRINT(TR_SM, TR_LEVEL_1, 
		("chooseServer checking %s,%s", 
		serverInfo->hostName, serverInfo->name));

	/*
	 *  Don't look at flag SERV_CONTACTED_IN_PAST - that's irrelevant.
	 */
	if((((serverInfo->flags)&~SERV_CONTACTED_IN_PAST) & *choice) == *choice) {
		if(*choice != SERV_NOTCONNECTED) {
			*result = serverInfo;
			return TRUE; /* found */
		}
		/* unconnected server ... try to connect to it */
		TRPRINT(TR_SM, TR_LEVEL_1, ("trying to connect"));

		if (server_connectInit(serverInfo, FALSE) == esmNOERROR) {
			*result = serverInfo;
			return TRUE; /* found */
		}
		TRPRINT(TR_SM, TR_LEVEL_1, ("cannot connect")); return FALSE;
	}
	TRPRINT(TR_SM, TR_LEVEL_1, ("wrong flags"));
	/* not found */
	return FALSE;
}


static int
getAnyServer(
	SERVERINFO		**serverInfo
)
{
	SERVERINFO		*result;
	FLAGS 			flags;

	result = NULL;

	/*
	 * try to get one that's connected, if not, get any server
	 * and connect to it.
	 */

	flags = SERV_CONNECTED;
	serverForEach(3, chooseServerInfo, &flags, &result, 0);

	TRPRINT(TR_SM, TR_LEVEL_1, ("choose Server :0x%x, %s", flags,
		result?result->hostName:"NONE"));

	if (result != NULL) {
		if (msg_initClient(result, FALSE) == esmNOERROR) {
			TRPRINT(TR_SM, TR_LEVEL_1, 
				("choose Server found %s", result->hostName));

			*serverInfo = result;
			return esmNOERROR;
		} 
	}

	flags = SERV_NOTCONNECTED;
	serverForEach(3, chooseServerInfo, &flags, &result, 0);

	if (result != NULL) {
		*serverInfo = result;
		return esmNOERROR;
	}

	return esmFAILURE;
}

 int
sm_Enter2PC (

	TID				tid,
	GTID			gtid,
	COORD_HANDLE	*handle
)
{
	TRANSREC	*transRec;
	char		*addr;
	u_short		port;

	TRACE(TR_SM|TR_TRANS, TR_LEVEL_1);

	if((transRec = findTransRec(&tid)) == NULL) {
        SM_ERROR(TYPE_USER, esmBADTRANSID);
        return esmFAILURE;
	}

    /*
     * 	Make sure there is a transaction in progress
     */
    switch (transRec->clientTransState) {
    case T_ACTIVE:
        /* this is the only case where 2PC can be entered */
		break;

    case T_ABORT:
        SM_ERROR(TYPE_USER, esmTRANSABORTED);
        return(esmFAILURE);

    case T_INACTIVE:
        SM_ERROR(TYPE_USER, esmNOACTIVETRANS);
        return(esmFAILURE);

	case T_PREPARED:
        SM_ERROR(TYPE_USER, esmTRANSPREPARED);
        return(esmFAILURE);

    default:
        SM_ERROR(TYPE_USER, esmINTERNAL);
        return(esmFAILURE);
    }

	/*
	 *	check that we have not already entered 2PC
	 */
	if (transRec->coordTid != NULL_TID) {

        SM_ERROR(TYPE_USER, esmTRANSIN2PC);
        return(esmFAILURE);
	}

	/*
	 *	pick the coordinator
	 */
	if (pickCoord(transRec)) {
			return(esmFAILURE);
	}

	if (enter2PC(transRec->coordInfo, &(transRec->coordTid), &gtid)) {

		return esmFAILURE;
	}

	/* 
	 * TODO: get a new transRec and fill it in; implement
	 * suspendTrans and resumeTrans ;
	 */

	/*
	 *	construct the handle; first fill the TID
	 */
	handle->coordTid = transRec->coordTid;

	/*
	 *	convert the address in a.b.c.d, and fill in rest of
	 *	handle
	 */
	addr = inet_ntoa(transRec->coordInfo->connectAddr.sin_addr);
	strcpy(handle->hostName, addr);
	port = ntohs(transRec->coordInfo->connectAddr.sin_port);
	sprintf(handle->serverName, "%d", port);

	return esmNOERROR;
}

 int
sm_Continue2PC (

	TID				tid,
	BOOL			willing_to_block
)
{
	TRANSREC    *transRec;

	TRACE(TR_SM|TR_TRANS, TR_LEVEL_1);

	if((transRec = findTransRec(&tid)) == NULL) {
        SM_ERROR(TYPE_USER, esmBADTRANSID);
        return esmFAILURE;
	}

    /*
     * 	Make sure there exists a transaction to continue
     */
    switch (transRec->clientTransState) {
    case T_ACTIVE:
    case T_ABORT:
	case T_PREPARED:
        /* the client can reconnect to the server in these cases */
		break;

    case T_INACTIVE:
        SM_ERROR(TYPE_USER, esmNOACTIVETRANS);
        return(esmFAILURE);

    default:
        SM_ERROR(TYPE_USER, esmINTERNAL);
        return(esmFAILURE);
    }

	/*
	 *	sm_Continue2PC() should be called only if a coord is known
	 */
	if (transRec->coordTid == NULL_TID) {

		SM_ERROR(TYPE_USER, esmCOORDUNKNOWN);
		return(esmFAILURE);
	}

	if (willing_to_block) {
		while (connectServer(transRec->coordInfo) != esmNOERROR);
	}
	else {
		/*
		 *	check if connect succeeds
		 */
		if (connectServer(transRec->coordInfo) != esmNOERROR) {
			SM_ERROR(TYPE_USER, esmWOULDBLOCK);
			return esmFAILURE;
		}
	}

init_again:
	if (msg_initClient(transRec->coordInfo, FALSE) != esmNOERROR) {
		if ((willing_to_block) && 
			((sm_errno == esmTRANSDISABLED) || 
			 (sm_errno == esmLOGDISABLED))) {
			goto init_again;

		} else {
			return esmFAILURE;
		}
	}

continue_again:
	if (continue2PC(transRec) != esmNOERROR) {
		if ((willing_to_block) && 
			((sm_errno == esmTRANSDISABLED) || 
			 (sm_errno == esmLOGDISABLED))) {
			goto continue_again;

		} else {
			return esmFAILURE;
		}
	}
    transRec->clientTransState = T_PREPARED;
	return esmNOERROR;
}

 int
sm_Recover2PC (

	COORD_HANDLE	handle,
	GTID			unused, /* gtid */
	TID				*tid,
	BOOL			willing_to_block
)
{
	SERVERINFO	*coordInfo;

	TRACE(TR_SM|TR_TRANS, TR_LEVEL_1);

    /*
     * 	Make sure no transaction is already in progress
     */
    switch (TransRec.clientTransState) {
    case T_ACTIVE:
	case T_PREPARED:
		SM_ERROR(TYPE_USER, esmTRANSINPROGRESS);
		return(esmFAILURE);

    case T_ABORT:
		SM_ERROR(TYPE_USER, esmTRANSABORTED);
		return(esmFAILURE);

    case T_INACTIVE:
		/* this is the only case where the old new transaction can be recovered */
		break;

    default:
        SM_ERROR(TYPE_USER, esmINTERNAL);
        return(esmFAILURE);
    }

	if ((coordInfo = getServerInfo(handle.hostName, handle.serverName)) == NULL) {
		SM_ERROR(TYPE_CRASH, esmBADHANDLE);
		return(esmFAILURE);
	}

	if (willing_to_block) {
		while (connectServer(coordInfo) != esmNOERROR);
	}
	else {
		/*
		 *	check if connect succeeds
		 */
		if (connectServer(coordInfo) != esmNOERROR) {
			SM_ERROR(TYPE_USER, esmWOULDBLOCK);
			return esmFAILURE;
		}
	}

init_again:
	if (msg_initClient(coordInfo, FALSE) != esmNOERROR) {
		if ((willing_to_block) && 
			((sm_errno == esmTRANSDISABLED) || 
			 (sm_errno == esmLOGDISABLED))) {
			goto init_again;

		} else {
			return esmFAILURE;
		}
	}

recover_again:
	if (recover2PC(&handle, coordInfo) != esmNOERROR) {
		if ((willing_to_block) && 
			((sm_errno == esmTRANSDISABLED) || 
			 (sm_errno == esmLOGDISABLED))) {
			goto recover_again;

		} else {
			return esmFAILURE;
		}
	}

	/*
	 * TODO: get a new transRec and fill it in; implement
	 * suspendTrans and resumeTrans
	 */
	TransRec.clientTid = *tid = newTid();
	TransRec.coordTid =  handle.coordTid;
	TransRec.coordInfo =  coordInfo;

	TransRec.clientTransState = T_PREPARED;

	/* 
	 *	initialize number of participants (servers) to zero and 
	 *	increment when a new server joins the party 
	 */
	TransRec.numParticipants = 0; 

	return esmNOERROR;
}

int 
pickCoord(
	TRANSREC	*transRec
)
{
	SERVERTRANSREC	*serverTransRec;

	/*
	 *	pick a server from the list of servers hanging off
	 *	the transRec and any others you may want to consider
	 */
	if ((serverTransRec = (SERVERTRANSREC *) FIRST_LIST_ELEMENT(&(transRec->participants))) == NULL) {
		/* pick the default server */
		if (getAnyServer(&(transRec->coordInfo))) {
			SM_ERROR(TYPE_FATAL, esmINTERNAL);
		}

	}
	else 
		transRec->coordInfo = serverTransRec->serverInfo;

	return esmNOERROR;
}
