/*
 *   $RCSfile: sm_AdminFuncs.c,v $  
 *   $Revision: 1.1.1.1 $  
 *   $Date: 1996/05/04 21:55:28 $      
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/


#include "sysdefs.h"
#include "ess.h"
#include "checking.h"
#include "io.h"
#include "list.h"
#include "tid.h"
#include "lock.h"
#include "object.h"
#include "bf.h"
#include "msgdefs.h"
#include "sm_macro.h"
#include "error.h"
#include "trace.h"
#include "load.h"
#include "sm_params.h"
#include "btree.h"
#include "scan.h"
#include "trans.h"
#include "volume.h"
#include "adminmsg.h"
#include "serverinfo.h"
#include "sm_state.h"
#include "msg_funcs.h"
#include "trans_funcs.h"
#include "sm_extfuncs.h"
#include "msg_globals.h"
#include "io_extfuncs.h"
#include "io_intfuncs.h"
#include "trans_globals.h"


/* helperfunc1: Assumes that the server Tid is already
 * in the message or the message doesn't require one.
 */
#define CONNECT_INIT 0x1 /* connect & init if not connected */
	/* CONNECT_INIT does not conflict with VOL_* flags, because
	* it's passed as a different parameter 
	*/

static  BOOL
helperfunc1 (
	SERVERINFO  *serverInfo,
	MESSAGE 	*msg,
	FLAGS		flags
)
{
	MESSAGE 	message;
	
	message = *msg; /* a copy */

	if(flags & CONNECT_INIT) {
		if(server_connectInit(serverInfo, TRUE/* is admin */) != esmNOERROR) {
			return FALSE;
		}
	}

	(void) callServer(serverInfo, &message, NULL, 0 );
	return FALSE; 
}

static BOOL
helperfunc2 (
	SERVERINFO  *serverInfo,
	MESSAGE 	*msg,
	FLAGS		flags
)
{
	if(serverInfo->flags & SERV_CONTACTED_IN_PAST) {
		helperfunc1 (serverInfo, msg, flags);
	}
	return FALSE;
}

static  BOOL
helperfunc3 (
	SERVERTRANSREC  *stl,
	MESSAGE 	*msg,
	FLAGS		flags
)
{
	/* No admin func (at present) needs a tid, so this is
	 * really superfluous, but we will leave it in for 
	 * possible extensions.
	 */
	msg->header.params.in.tid = stl->tid;
	helperfunc1 (stl->serverInfo, msg, flags);
	return FALSE; 
}



doit(
	FLAGS flags1,
	VOLID	volid,
	MESSAGE *message,
	FLAGS flags2
) 
{
	extern int sm_errno;
	VOLREC *volRec;

	sm_errno = 0;

	switch(flags1) {
	case VOL_BY_VOLID: 
		volRec = IO_FindVolId( volid, SERV_ESTAB );
		if( volRec == NULL) {
			return esmFAILURE;
		} 
		if( volRec->serverInfo == NULL) {
			SM_ERROR(TYPE_FATAL, esmINTERNAL);
		}
		(void) helperfunc1(volRec->serverInfo, message, 0);
		break;

	case VOL_ALL:
		/* stupid */
		serverForEach(3, helperfunc1, message, (void *)flags2);
		break;

	case VOL_USED_IN_TRANSACTION:
		ForEachServerInTrans(findTransRec(getLocalTid()), 
						3, helperfunc3, message, (void *)flags2);
		break;

	case VOL_MOUNTED_SINCE_INIT:
		serverForEach(3, helperfunc2, message, (void *)flags2);
		break;
	}
	if(sm_errno)
		return esmFAILURE;
	return esmNOERROR;
}

 int
sm_ChangeCheckpointFrequency (
	FLAGS			flags,
	VOLID			volid,
	int				newFreq
)
{
	MESSAGE			message;

	message.header.type		= ADMIN_CHECKPOINTFREQ; 
	message.header.replyRequested		= TRUE; 
	message.body.checkpoint.checkpointInterval = (FOUR) newFreq;

	return doit(flags, volid, &message, CONNECT_INIT);
}


 int
sm_MonDump(
	FLAGS flags,
	VOLID	volid
)
{
	MESSAGE			message;

	message.header.type		= MON_DUMP;
	message.header.replyRequested = FALSE;

	/* makes no sense to force a connection for a server we're 
	* not using 
	*/
	return doit(flags, volid, &message, 0);
}

 int
sm_TakeCheckpoint (
	FLAGS		flags,
	VOLID		volid,
	TWO			numCheckpoints  	/* number of checkpoint to take */
)
{

	MESSAGE			message;

	message.header.type		= ADMIN_CHECKPOINT; 
	message.header.replyRequested		= TRUE; 
	message.body.checkpoint.numCheckpoint = numCheckpoints;

	return doit(flags, volid, &message, CONNECT_INIT);
}



 int
sm_TestNullMessage (
	FLAGS		flags,
	VOLID		volid,
	int			numMsgs  	/* number of Null messages to send */
)
{

	MESSAGE			message;
	int				i;

	message.header.type		= TEST_NULL_RPC; 
	message.header.replyRequested	= 	TRUE; 
	/*
	 *	Send the message numMsgs times
	 *  and  expect a reply for each one.
	 *  This really should put the numMsg loop in the helper func.
	 */
	for (i = 0; i<numMsgs; i++) {
		if(doit(flags, volid, &message, CONNECT_INIT) != esmNOERROR)
			return esmFAILURE;
	}
	return esmNOERROR;
}

 int
sm_ShutdownServer (
	FLAGS		flags,
	VOLID		volid,
	FLAGS		shutdown_flags  	/* shutdown options */
)
{
	MESSAGE			message;

		/* TODO: special handling for multiple volumes/server:
			don't want to try to shut down a server twice
		*/

	message.header.type		= ADMIN_SHUTSERVER; 
	message.header.replyRequested	= TRUE; /* so we try 
		to receive and detect the disconnect */
	message.body.shutdown.shutdownFlags = shutdown_flags;

	/* possibly force a connection to kill server because
	* this is used by the admin program, shutserver
	*/
	return doit(flags, volid, &message, CONNECT_INIT);
}

#include "clog.h"
#include "logaction.h"
#include "loginfo.h"
#include "logrecs.h"
/*
 *	This function turns off logging for the current transaction.
 *	For correctness, it cannot be turned on again until a new
 *	transaction has started.  When a new transaction is  started,
 *	logging is automatically turned on.
 *	
 *	If the server is informed, then all log pages sent to 
 *	it will be marked clean so the server will not flush them
 *	to disk.
 */

 int
sm_TurnOffLogging (
	FLAGS	flags, 
	VOLID	volid,
	FLAGS	clientLogLevel
)
{
	MESSAGE		message;

    TRPRINT(TR_SM, TR_LEVEL_1, ("turning off logging"));

	if( (flags != VOL_USED_IN_TRANSACTION) ) {
		SM_ERROR(TYPE_USER, esmBADPARAMS);
		return esmFAILURE;
	}

    if (checkLocalTransState())	{
		return(esmFAILURE);
    }
	if (clientLogLevel != LOG_NONE && clientLogLevel != LOG_DATA &&
		clientLogLevel != LOG_SPACE && clientLogLevel != LOG_ALL)  {
		SM_ERROR(TYPE_USER, esmBADPARAMS);
		return(esmFAILURE);
	} else {
		SM_State.defaultLogLevel = clientLogLevel;
	}

	message.header.type = ADMIN_DEBUG;
	message.header.replyRequested = TRUE;
	message.body.admin.operation = ADMIN_DEBUG_CLIENT_LOG_OFF;
	/* used only for servers participating on this tx, so 
	* don't use CONNECT_INIT 
	*/
	return doit(flags, volid, &message, 0);
}


 int
sm_DebugSetVar (
	FLAGS	flags,
	VOLID	volid,
	int		variable,	
	int		value
)
{
	MESSAGE		message;

    TRPRINT(TR_SM, TR_LEVEL_1, ("set variable"));

    if (checkLocalTransState())	{
		return(esmFAILURE);
    }

	message.header.type = ADMIN_DEBUG;
	message.header.replyRequested = TRUE;
	message.body.admin.operation = ADMIN_DEBUG_SET_VAR;
	message.body.admin.variable = variable;
	message.body.admin.value = value;

	return doit(flags, volid, &message, CONNECT_INIT);
}

int
sm_PrintMessageStats(FILE *const f, MESSAGESTATS *const m)
{ 
	printMessageStats(f, m);
	/* this may seem flaky, but it returns an int to be consistent
	 * with all the other sm_* funcs
	 */
	return esmNOERROR;
}

int
sm_VolumeProperties(VOLID volid, int *props)
{ 
	VOLREC *volRec;

	volRec = io_Mount(volid);
	if(volRec == NULL) {
		SM_ASSERT(LEVEL_1, (sm_errno != esmNOERROR));
		/* not mounted */
		return esmFAILURE;
	}
	*props = (int) (volRec->properties & VOL_PROPERTIES);
	return esmNOERROR;
}
