/*
 *   $RCSfile: sm_FreezeVersion.c,v $  
 *   $Revision: 1.1.1.1 $  
 *   $Date: 1996/05/04 21:55:28 $      
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/


#include "ess.h"
#include "sysdefs.h"
#include "checking.h"
#include "io.h"
#include "page.h"
#include "list.h"
#include "object.h"
#include "tid.h"
#include "bf.h"
#include "chunk.h"
#include "bf_macro.h"
#include "sm_macro.h"
#include "error.h"
#include "trace.h"
#include "lock.h"
#include "file.h"
#include "pool.h"
#include "lgobject.h"
#include "version_graph.h"
#include "version_funcs.h"
#include "scan.h"
#include "load.h"
#include "sm_params.h"
#include "btree.h"
#include "clog.h"
#include "lsn.h"
#include "trans.h"
#include "loginfo.h"
#include "sm_state.h"
#include "bf_extfuncs.h"
#include "fi_extfuncs.h"
#include "sm_intfuncs.h"
#include "sm_version_intfuncs.h"
#include "sm_extfuncs.h"
#include "clog_extfuncs.h"
#include "lg_versionfuncs.h"
#include "trans_funcs.h"
#include "bf_globals.h"


 int
sm_FreezeVersion (

	int						groupIndex,
	OID						*oid 
/*
 *	This function marks an object as frozen and if the object is large
 *	it records the frozen information in the corresponding node in 
 *	the version history graph for the object.  If the object does 
 *	not already have a history graph, a new one is created for it.
 *
 *	An alternative time to create the version history graph is after
 *	a version of an object is actually created (instead of after an
 *	object is frozen).	This is not done because of the added 
 *	complexity.
 */ 
)
{
    register SLOTTEDPAGE    *homePage;	/* object's home page		*/
    register OBJECT			*objectPtr;	/* ptr to generic object	*/
	GROUPLINK				*homeLink = NULL;
	BOOL					homeLinkIsDirty = FALSE;
	BUFGROUP				*bufGroup;	/* pointer to buffer group	*/
    PID						homePid;	/* home page of object		*/
	PAGESLOT				*slotPtr;
    SLOTINDEX				objectSlot;	/* the object's slot		*/
	TWO						oldProps;											

	TRPRINT(TR_SM, TR_LEVEL_1, ("oid page:%d slot:%d unique:%d",
			 oid->diskAddr.page, oid->diskAddr.slot, oid->diskAddr.unique));

	INIT_SMSTATE(&SM_State);

	if (checkTransState(oid->diskAddr.volid))	{
		return(esmFAILURE);
	}

    /*
     *	Construct the pid of the home page
     */
    homePid.volid = oid->diskAddr.volid;
    homePid.page = oid->diskAddr.page;

	/*
	 *	get a pointer to the proper buffer group
	 */
	bufGroup = &(BufGroupPtr[groupIndex]);

	/*
	 *	check the buffer group magic number
	 */
	if (CHECK_BUFGROUP_VALID(bufGroup))	{

		SM_ERROR(TYPE_USER, esmBADGROUP);
		return(esmFAILURE);
	}

	/*
	 *	read the page into a buffer
	 */
	if ((homeLink = bf_ReadPage(bufGroup, &homePid, H_SLOTTED,
								SLOTTED_PAGE2SIZE, EX)) == NULL)	{

		FAIL
	}

	if (SM_SetLogServerState(homeLink, oid) != esmNOERROR) {
		FAIL
	}

	homePage = (SLOTTEDPAGE *) homeLink->bufFrame;
	if (CHECK_SLOTTED_MAGIC(homePage))	{
		SM_ERROR(TYPE_FATAL, esmBADSLOTTEDMAGIC);
	}

    objectSlot = oid->diskAddr.slot;
	slotPtr = GETSLOTPTR(homePage, objectSlot);

	/*
	 *	check the validity of the oid
	 */
	if (CHECKOID(objectSlot, homePage, slotPtr, oid->diskAddr.unique))	{
		SM_ERROR(TYPE_USER, esmBADOID);
		FAIL
	}

    objectPtr = GETOBJECTPTR(homePage, slotPtr);

	/*
	 * Make sure object isn't a version history graph or other 
	 * invisible object
	 */
	if (objectPtr->objectHdr.properties & INVISIBLE_OBJECT) {
		SM_ERROR(TYPE_USER, esmHISTORYGRAPHOID); 
		FAIL
	}

	/*
	 * Mark the object as versioned and frozen, and log it.	
	 */
	oldProps = objectPtr->objectHdr.properties;
    objectPtr->objectHdr.properties |= (P_VERSIONED | P_FROZEN); 

	homeLinkIsDirty = TRUE;
	sm_SetAbortIfError();

	if (GET_LOGLEVEL(LOG_SPACE)) {
		clog_LogChangeSmProps(oid, &(objectPtr->objectHdr),
							oldProps, objectPtr->objectHdr.properties);
	}

	/*
	 *	Is the object a large one
	 */
	if (objectPtr->objectHdr.properties & P_LARGEOBJ) {

		VHGNODEPTR	vhgPtr;	/* pointer to node in vhg	*/
		LARGEOBJHDR	*lgObjectPtr  = (LARGEOBJHDR*) objectPtr;

		/*
		 *	Get a "pointer" to a version graph
		 */
		if (lg_GetVHGPtr(bufGroup, lgObjectPtr, oid->diskAddr.volid, 
						 &vhgPtr)) {
			FAIL
		}

		/*
		 *	If the object doesn't have a version history graph
		 *	yet, create one for it.  
		 */
		if ( VHGPTR_IS_NULL(&vhgPtr) ) {
			FID			fid;		/* file to create VHG in		*/

			TRPRINT(TR_SM, TR_LEVEL_2, ("Need to create VHG"));

			/*
			 *	Get the file fid to create the VHG in
			 */
			fid = homePage->header.fid;

			/*
			 *	Create the VHG with oid as the root
			 */
			if (createVHG(groupIndex, &fid, oid, &vhgPtr.oid, 
						  &vhgPtr.nodeId) ) {
				FAIL
			}

			/*
			 *	Update the vhgOid and vhg root node id in the 
			 *	lgParentObj.  This should be logged when recovery 
			 *	is added.
			 */
			if (lg_SetVHGPtr(bufGroup, lgObjectPtr, &vhgPtr, oid)) {
				FAIL
			}
		} else {
			/*
			 *	Mark the version as frozen in the verison history graph
			 *	First get the version graph.
			 */
			VERSIONGRAPH*	vhg;
			OID				parentOid; 	/* parent of this version */

			TRPRINT(TR_SM, TR_LEVEL_2, ("marking VHG node frozen"));

			if (SM_GetVersionGraph(bufGroup, &vhgPtr.oid, &vhg, FALSE, 
									EX) ) {
				FAIL
			}
	
			if (markFrozen(vhg, vhgPtr.nodeId) ) {
				FAIL
			}

			/*
			 *	See if the parent version is marked destroy pending
			 *	and has all its children frozen.
			 */
			if (parentDestroyPending(vhg, vhgPtr.nodeId, &parentOid)) {
				if (sm_DestroyObject(groupIndex, &parentOid)) {
					FAIL
				}
			}
			/*
			 *	Save back the version graph
			 */
			if (SM_SetVersionGraph(bufGroup, &vhgPtr.oid, vhg) ) {
				FAIL
			}
		}
	}
Success:

	if(homeLink)
		bf_UnfixPage(homeLink, BF_DEFAULT, TRUE);
    return esmNOERROR;

FailureReturn:
	if(homeLink)
		bf_UnfixPage(homeLink, BF_DEFAULT, homeLinkIsDirty);
	SM_AbortIfRequired();
	return esmFAILURE;
}

