/*
 * $RCSfile: sm_SetClientOption.C,v $
 * $Revision: 1.1.1.1 $
 * $Date: 1996/05/04 21:55:29 $
 */

/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/

#include "sysdefs.h"
#include "ess.h"
#include "checking.h"
#include "error.h"
#include "trace.h"
#include "io.h"
#include "page.h"
#include "list.h"
#include "object.h"
#include "tid.h"
#include "bf.h"
#include "chunk.h"
#include "sm_macro.h"
#include "scan.h"
#include "load.h"
#include "sm_params.h"
#include "btree.h"
#include "volume.h"
#include "lock.h"
#include "msgdefs.h"
#include "option.h"
#include "sm_intfuncs.h"
#include "sm_extfuncs.h"
#include "trace_funcs.h"
#include "trans.h"
#include "trans_globals.h"
#include "bf_globals.h"
#include "clog_globals.h"
#include "io_globals.h"
#include "sm_globals.h"

#ifndef __cplusplus
extern char*	getenv(char*);
#endif

/*
 *	ClientOptions is defined here since space.c can be compiled with C
 */
OptionList*	ClientOptions = NULL;

/*
 * FUNCTION DESCRIPTION:
 *
 * This function is used to set an option on the client.
 * The optionValue argument points to a value for the option.
 * The valueType argument indicates what type of data optionValue
 * points to.  
 *
 * If the optionName is "optionclass" an new classification level
 * is added to the option list.  This is used primarily for
 * adding the application name to the classification list.
 */

 int
sm_SetClientOption (
	const char	*optionName,	/* name of the option 				*/
	const void	*optionValue, 	/* value for the option 			*/
	SMDATATYPE	valueType		/* type of optionValue				*/
)
{
	OptionInfo*	opt;
	char		convertValue[MaxOptLength+1]; 
	int			errCode;

	if (ClientOptions == NULL) {
		if (SM_InitClientOptions() != esmNOERROR) {
			SM_ERROR(TYPE_USER, esmMALLOCFAILED);
			return esmFAILURE;
		}
	}

	if ( (optionName == NULL) || (optionValue == NULL) ) {
		SM_ERROR(TYPE_USER, esmBADPARAMS);
		return esmFAILURE;
	}

	if (ClientOptions->lookup(optionName, opt) != OPTERR_Success) {
		SM_ERROR(TYPE_USER, esmBADOPTION);
		return esmFAILURE;
	}

	// convert optionValue to the correct type
	switch (valueType) {
	case SM_string:
		break;
	case SM_Boolean:
		switch (opt->ID()) {
		case SM_OPT_deallocpages: 
		case SM_OPT_lognewpages: 
			if (*(BOOL*)optionValue == FALSE) {
				strcpy(convertValue, "F");
			} else {
				strcpy(convertValue, "T");
			}
			break;
		default:
			SM_ERROR(TYPE_USER, esmBADPARAMS);
			return esmFAILURE;
		}
		break;
	case SM_int:
		switch (opt->ID()) {
		case SM_OPT_locktimeout: 
		case SM_OPT_bufpages: 
		case SM_OPT_groups: 
		case SM_OPT_userdesc:
		case SM_OPT_traceflags:
			sprintf(convertValue, "%d", *(int*)optionValue);
			break;
		case SM_OPT_deallocpages: 
		case SM_OPT_lognewpages: 
			if (*(int*)optionValue == 0) {
				strcpy(convertValue, "F");
			} else {
				strcpy(convertValue, "T");
			}
			break;
		default:
			SM_ERROR(TYPE_USER, esmBADPARAMS);
			return esmFAILURE;
		}
		break;
	default:
		SM_ERROR(TYPE_USER, esmBADPARAMS);
		return esmFAILURE;
	}

	// make sure the option looks set
	if (valueType == SM_string) {
		errCode = opt->setValue((char*)optionValue, TRUE);
	} else {
		errCode = opt->setValue(convertValue, TRUE);
	}

	if (errCode != 0) {
		return esmFAILURE;
	} 
    return esmNOERROR;
}

static int
str2lockmode(char *str, LOCKMODE *newval, LOCKMODE oldval) 
{
	LOCKMODE val;
	if(strcmp(str, "NL")==0) {
		val =  NL;
	} else if(strcmp(str, "IS")==0) {
		val =  IS;
	} else if(strcmp(str, "IX")==0) {
		val =  IX;
	} else if(strcmp(str, "SH")==0) {
		val =  SH;
	} else if(strcmp(str, "SIX")==0) {
		val =  SIX;
	} else if(strcmp(str, "EX")==0) {
		val =  EX;
	} else {
		*newval = oldval;
		return esmBADOPTIONVALUE;
	}
	/* for now, we only permit SH and EX */

	if((val != SH) && (val != EX)) {
		*newval = oldval;
		return esmBADOPTIONVALUE;
	} else {
		*newval = val;
		return esmNOERROR;
	}
}


 int
SM_SetOption (
	const char	*optionName,	/* name of the option 				*/
	const char	*optionValue 	/* value for the option 			*/
)
{
	OptionInfo*	opt;
	int			intValue;
	BOOL		boolValue;
	char*		optVal = (char*)optionValue;

	if (ClientOptions == NULL) {
		if (SM_InitClientOptions() != esmNOERROR) {
			return esmMALLOCFAILED;
		}
	}

	if ( (optionName == NULL) || (optionValue == NULL) ) {
		return esmBADPARAMS;
	}

#ifdef JUNK
	if (strcmp(optionName, "optionclass") == 0) {
		if (ClientOptions->addClassLevel(optVal) != OPTERR_Success) {
			return(esmBADOPTIONCLASS);
		}
		return esmNOERROR;
	}
#endif

	if (ClientOptions->lookup(optionName, opt) != OPTERR_Success) {
		return(esmBADOPTION);
	}

	// convert optionValue to the correct type
	switch (opt->ID()) {
	case SM_OPT_locktimeout: 
	case SM_OPT_bufpages: 
	case SM_OPT_groups: 
	case SM_OPT_userdesc:
	case SM_OPT_traceflags:
		intValue = (int) strtol(optVal, NULL, 0);
		break;
	case SM_OPT_filelock:
		return str2lockmode(optVal, &sm_DefaultFileLock, sm_DefaultFileLock);

	case SM_OPT_pagelock:
		return str2lockmode(optVal, &sm_DefaultPageLock, sm_DefaultPageLock);

	case SM_OPT_lognewpages:
	case SM_OPT_deallocpages:
		switch (optVal[0]) {
		case 'y': case 'Y': case 't': case 'T':
			boolValue = TRUE;
			break;
		case 'n': case 'N': case 'f': case 'F':
			boolValue = FALSE;
			break;
		default:
			return(esmBADOPTIONVALUE);
		}	
		break;
	case SM_OPT_mount:
		SM_ERROR(TYPE_FATAL, esmINTERNAL);
		break;
	default:
		SM_ERROR(TYPE_FATAL, esmINTERNAL);
		break;
	}
	
	switch (opt->ID()) {
	case SM_OPT_locktimeout: 
		if(intValue < 0) {
			return(esmBADOPTIONVALUE);
		}
		DefaultLockTimeout = intValue;
		break;
	case SM_OPT_bufpages: 
		if (intValue < 5) {
			return(esmBADOPTIONVALUE);
		}
		NumBuffers = intValue;
		break;
	case SM_OPT_groups: 
		if (intValue < 4) {
			return(esmBADOPTIONVALUE);
		}
		NumBufGroups = intValue;
		break;
	case SM_OPT_userdesc:
		if (intValue < 1) {
			return(esmBADOPTIONVALUE);
		}
		NumUserDesc = intValue;
		break;
	case SM_OPT_traceflags:
		TraceFlags = intValue;
		break;
	case SM_OPT_lognewpages:
		if(TransRec.clientTransState != T_INACTIVE) {
			return(esmTRANSINPROGRESS);
		}
		LogNewPages = boolValue;
		break;
	case SM_OPT_deallocpages:
		DeallocSmallObjPages = boolValue;
		break;
	case SM_OPT_mount:
		break;
	default:
		SM_ERROR(TYPE_FATAL, esmINTERNAL);
		break;
	}

    return esmNOERROR;
}
