/*
 * $RCSfile: SM_BulkLoad.C,v $
 * $Revision: 1.1.1.1 $
 * $Date: 1996/05/04 21:55:29 $
 */

/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/
#include "sysdefs.h"
#include "ess.h"
#include "checking.h"
#include "trace.h"
#include "error.h"
#include "io.h"
#include "page.h"
#include "list.h"
#include "object.h"
#include "tid.h"
#include "lock.h"
#include "bf.h"
#include "chunk.h"
#include "volume.h"
#include "file.h"
#include "bf_globals.h"
#include "load.h"
#include "sm_params.h"
#include "btree.h"
#include "scan.h"
#include "trans.h"
#include "clog.h"
#include "sm_params.h"
#include "lsn.h"
#include "loginfo.h"
#include "sm_state.h"
#include "sm_extfuncs.h"
#include "sm_intfuncs.h"
#include "io_extfuncs.h"
#include "bf_extfuncs.h"
#include "bt_extfuncs.h"
#include "lh_extfuncs.h"
#include "trans_funcs.h"
#include "trans_globals.h"
#include "clog_extfuncs.h"
#include "clog_globals.h"
#include "LH_Hash.h"
#include "sm_user.h"
#include "sm_globals.h"


#define MAXLOAD 5		/* size of load table */
#define RUNMAX	10		/* default run size, automatically increased */

static SMNDXLOADDESC loadTbl[MAXLOAD];
static void QuickSort(KEY key[], int cnt, PFC cmp);
static int SM_FlushLoadBuffer(int slot);
static void SM_FreeLoadDesc(SMNDXLOADDESC* p);
static SM_InternalInsert(IID* ndx, void* keyVal, int keyLen, void* elem);

sm_FindIndexLoad(IID* ndx, int* idx)
{
    *idx = -1;
    for (register i = 0; i < MAXLOAD; i++)   {
	if (loadTbl[i].volid >= 0 && loadTbl[i].iid == *ndx)  {
	    *idx = i;
	    break;
	}
    }
    
    return esmNOERROR;
}


SM_TransBulkloading(TID tid)
{
    for (register i = 0; i < MAXLOAD; i++) {
	if (loadTbl[i].volid >= 0 && loadTbl[i].tid == tid) 
	    return TRUE;
    }
    
    return FALSE;
}


int
SM_AbortAllIndexLoads(TID tid)
{
    for (register i = 0; i < MAXLOAD; i++)  {
	if (loadTbl[i].volid >= 0 && loadTbl[i].tid == tid)  {
	    if (sm_AbortIndexLoad(& loadTbl[i].iid))  {
		SM_ERROR(TYPE_FATAL, esmINTERNAL);
	    }
	}
    }
	return esmNOERROR;
}


void
sm_EndAllIndexLoads(TID tid)
{
    for (register i = 0; i < MAXLOAD; i++)  {
	if (loadTbl[i].volid >= 0 && loadTbl[i].tid == tid)  {
	    if (sm_EndIndexLoad(& loadTbl[i].iid))  {
		SM_ERROR(TYPE_FATAL, esmINTERNAL);
	    }
	}
    }
}


sm_IsIndexEmpty(IID* ndx, int groupIndex, BOOL* ret)
{
    BUFGROUP*   bufGroup = & BufGroupPtr[groupIndex];
    INDEXDESC*  np;
    GROUPLINK*  gLink;
    
    if (CHECK_BUFGROUP_VALID(bufGroup))  {
        SM_ERROR(TYPE_USER, esmBADGROUP);
        return esmFAILURE;
    }

    if (checkTransState(ndx->pid.volid))  {
        return(esmFAILURE);
    }
    
    if ((gLink = bf_ReadPage(bufGroup, & ndx->pid, H_INDEXDESC,
			     INDEXDESC_PAGE2SIZE, SH)) == NULL)
        return esmFAILURE;
    
    SM_ASSERT(LEVEL_3, ((INDEXDESCPAGE*) gLink->bufFrame)->magic == NDXMAGIC);
    
    np = & (((INDEXDESCPAGE*) gLink->bufFrame)->desc);
    switch (np->ndxType)  {
    case SM_BTREENDX:
	if (bt_IsEmpty(TransRec.clientTid, np->rootPid, bufGroup, ret))  {
	    bf_UnfixPage(gLink, BF_LRU, FALSE);
	    return esmFAILURE;
	}
	break;
	
    case SM_HASHNDX:
	if (lh_IsEmpty(TransRec.clientTid, np->rootPid, bufGroup, ret)) {
	    bf_UnfixPage(gLink, BF_LRU, FALSE);
	    return esmFAILURE;
	}
	break;

    default:
		SM_ERROR(TYPE_FATAL, esmINTERNAL);
    }
    
    bf_UnfixPage(gLink, BF_LRU, FALSE);
    
    return esmNOERROR;
}



sm_BeginIndexLoad(IID* ndx, int groupIndex, int workVolume, int runSize)
{
    static int initialized = 0;
    
    if (! initialized)  {
	for (register i = 0; i < MAXLOAD; i++)  
	    loadTbl[i].volid = -1;
	initialized = 1;
    }
    
    if (checkTransState(ndx->pid.volid))  
        return(esmFAILURE);
    
	{
		BOOL empty;
		if (sm_IsIndexEmpty(ndx, groupIndex, &empty))   {
			return esmFAILURE;
		}
		if (! empty)  {
			SM_ERROR(TYPE_USER, esmINDEXNOTEMPTY);
			return esmFAILURE;
		}
		
		int i;
		if (sm_FindIndexLoad(ndx, &i)) {
			return esmFAILURE;
		}
		if (i >= 0)  {
			SM_ERROR(TYPE_USER, esmBULKLOADING);
			return esmFAILURE;
		}
	}
    
    if (runSize < 3) runSize = 3;
    
    runSize *= BTREE_PAGESIZE;
    
    //
    //	Find & initialize a free load descriptor slot
    //	
    
    SMNDXLOADDESC* loadDesc = & loadTbl[0];
    for (int i = 0; i < MAXLOAD && loadDesc->volid >= 0; i++, loadDesc++);
    if (i >= MAXLOAD)  {
	SM_ASSERT(LEVEL_1, FALSE);
    }
    
    loadDesc->tid = TransRec.clientTid;	
    loadDesc->magic = SMNDXLOADDESCMAGIC;
    loadDesc->iid = *ndx;
    loadDesc->groupIndex = groupIndex;
    loadDesc->runSize = runSize;
    
    //
    //	Create temporary file to hold runs
    //
    if (sm_CreateFile(groupIndex, workVolume, & loadDesc->tmpFid))  {
	return esmFAILURE; 
    }
	//
	// restore SM_State for index  volume
	//
	if(checkTransState( ndx->pid.volid )) {
		SM_ERROR(TYPE_FATAL, esmINTERNAL);
    }
    
    //
    //	Read & cache index descriptor
    //
    GROUPLINK* gLink;		
    if ((gLink = bf_ReadPage(& BufGroupPtr[groupIndex],
			     & ndx->pid, H_INDEXDESC, 
			     INDEXDESC_PAGE2SIZE, EX)) == NULL)  {
	return esmFAILURE;
    }
    INDEXDESC* np = &(((INDEXDESCPAGE*)gLink->bufFrame)->desc);
    
    loadDesc->ndesc = *np;
    
    bf_UnfixPage(gLink, BF_LRU, FALSE);
    
    loadDesc->offset = loadDesc->recCount = 
	loadDesc->runCount = loadDesc->totalRec = 0;
    loadDesc->runMax = RUNMAX;
    loadDesc->numPages = 0;
    loadDesc->maxKeyLen = 0;
    loadDesc->runList = new OID[ loadDesc->runMax ];
    if (! loadDesc->runList)  {
	SM_ERROR(TYPE_USER, esmMALLOCFAILED);
	return esmFAILURE;
    }
    loadDesc->buffer = new char[runSize];
    if (! loadDesc->buffer) {
	delete[] loadDesc->runList;
	SM_ERROR(TYPE_USER, esmMALLOCFAILED);
	return esmFAILURE;
    }
    const minTupleSize = (ALIGNSIZE(sizeof(int)) + 		// for keylen
			  ALIGNSIZE(1) + 			// for key
			  ALIGNSIZE(loadDesc->ndesc.elSize));	// for elem

    // keys[] points into buffer
    loadDesc->keys = new KEY[ (runSize - 1) / minTupleSize + 1 ];
    if (! loadDesc->keys) {
	delete[] loadDesc->runList;
	delete[] loadDesc->buffer;
	SM_ERROR(TYPE_USER, esmMALLOCFAILED);
	return esmFAILURE;
    }
    
    //
    //	Commit the load descriptor slot
    //
    loadDesc->volid = ndx->pid.volid;
    
    return esmNOERROR;
}


static void
SM_FreeLoadDesc(SMNDXLOADDESC* p)
{
    int errno_save = sm_errno;
    SM_ASSERT(LEVEL_1, p->volid != -1);
    sm_DestroyFile(p->groupIndex, &p->tmpFid);
    delete[] p->runList;
    delete[] p->buffer;
    delete[] p->keys;
    p->volid = -1;
    sm_errno = errno_save;
}


static SMDATATYPE lhKeyType;
static int lhLevel;
static PFC lhCmp;
static int lhPrimary;

static 
LHashComp(int kLen1, const void* kVal1, int kLen2, const void* kVal2)
{
    int h1 = LH_Hash(kLen1, kVal1, lhKeyType, lhLevel);
    int h2 = LH_Hash(kLen2, kVal2, lhKeyType, lhLevel);
    if (h1 >= lhPrimary)
	h1 = LH_Hash(kLen1, kVal1, lhKeyType, lhLevel - 1);
    if (h2 >= lhPrimary)
	h2 = LH_Hash(kLen2, kVal2, lhKeyType, lhLevel - 1);
    if (h1 == h2)
	return lhCmp(kLen1, kVal1, kLen2, kVal2);

    return h1 - h2;
}

SM_FlushLoadBuffer(int slot)
{
    SM_ASSERT(LEVEL_1, slot >= 0 && slot < MAXLOAD);
    SMNDXLOADDESC* loadDesc = & loadTbl[slot];
    
    SM_ASSERT(LEVEL_1, loadDesc->volid != -1);
    SM_ASSERT(LEVEL_1, loadDesc->magic == SMNDXLOADDESCMAGIC);
    SM_ASSERT(LEVEL_3, loadDesc->recCount >= 0);

    PFC cmp;
    switch (loadDesc->ndesc.ndxType) {
    case SM_BTREENDX:
	cmp = SMCOMPFUNC[ loadDesc->ndesc.dataType ];
	break;
    case SM_HASHNDX:
	cmp = LHashComp;
	break;
    default:
		SM_ERROR(TYPE_FATAL, esmINTERNAL);
    }

    QuickSort(loadDesc->keys, loadDesc->recCount, cmp);
    
    OID oid;
    OID first;
    for (int i = 0; i < loadDesc->recCount; i++)  {	
	if ( sm_CreateObject(loadDesc->groupIndex, & loadDesc->tmpFid, 
			     NEAR_LAST, 0, 0, ALIGNSIZE(sizeof(int)) + 
			     ALIGNSIZE(loadDesc->keys[i].length) + 
			     loadDesc->ndesc.elSize, 
			     (char*) loadDesc->keys[i].valuePtr - ALIGNSIZE(sizeof(int)),
			     &oid)) {
	    SM_FreeLoadDesc(loadDesc);
	    return esmFAILURE;
	}
	if (oid.diskAddr.slot == 0)  ++loadDesc->numPages;
	if (i == 0) first = oid;
    }
    int marker = -1;
    if (sm_CreateObject(loadDesc->groupIndex, &loadDesc->tmpFid, 
			NEAR_LAST, 0, 0, sizeof(int), &marker, &oid)) {
	SM_FreeLoadDesc(loadDesc);
	return esmFAILURE;
    }
    
    if (loadDesc->runCount == loadDesc->runMax)  {
	OID* tmp = new OID[ loadDesc->runMax *= 2 ];
	if (! tmp)  {
	    SM_FreeLoadDesc(loadDesc);
	    SM_ERROR(TYPE_USER, esmMALLOCFAILED);
	    return esmFAILURE;
	}
	BCOPY(loadDesc->runList, tmp, loadDesc->runCount * sizeof(OID));
	delete[] loadDesc->runList;
	loadDesc->runList = tmp;
    }
/*		
  printf("Created run %d.%d.%d\n", first.diskAddr.volid, first.diskAddr.page,
  first.diskAddr.slot);
  */
    loadDesc->runList[loadDesc->runCount++] = first;
    
    loadDesc->recCount = loadDesc->offset = 0;
    
    return esmNOERROR;	
}

static 
SM_InternalInsert(IID* ndx, void* keyVal, int keyLen, void* elem)
{
    SM_ASSERT(LEVEL_1, keyLen > 0);
    int loadHandle;
    if (sm_FindIndexLoad(ndx, &loadHandle))  {
	return esmFAILURE;
    }
    SM_ASSERT(LEVEL_1, loadHandle >= 0 && loadHandle < MAXLOAD);
    
    SMNDXLOADDESC* loadDesc = & loadTbl[loadHandle];
    SM_ASSERT(LEVEL_1, loadDesc->volid != -1);
    SM_ASSERT(LEVEL_1, loadDesc->iid == *ndx);
    SM_ASSERT(LEVEL_1, loadDesc->magic == SMNDXLOADDESCMAGIC);
    
    int need = (ALIGNSIZE(sizeof(int)) + 
		ALIGNSIZE(keyLen) + 
		ALIGNSIZE(loadDesc->ndesc.elSize));
    
    if (loadDesc->runSize - loadDesc->offset < need)  {
	if (SM_FlushLoadBuffer(loadHandle))
	    return esmFAILURE;
    }
    
    KEY* k = & loadDesc->keys[loadDesc->recCount++];
    char* p = loadDesc->buffer + loadDesc->offset;
    * (int*) p = keyLen,        p += ALIGNSIZE(sizeof(int));
    BCOPY(keyVal, p, keyLen);	p += ALIGNSIZE(keyLen);
    BCOPY(elem, p, loadDesc->ndesc.elSize);
    
    k->length = keyLen;
    k->valuePtr = loadDesc->buffer + loadDesc->offset + ALIGNSIZE(sizeof(int));
    loadDesc->offset += ALIGNSIZE(need);  // so that the next record
					  // starts at an aligned boundary
    
    loadDesc->totalRec++;
    
    return esmNOERROR;
}

SM_InsertIndexLoad(IID* ndx, void* keyVal, int keyLen, void* elem)
{
    SM_ASSERT(LEVEL_1, keyLen > 0);
    int loadHandle;
    if (sm_FindIndexLoad(ndx, &loadHandle))  {
	return esmFAILURE;
    }
    SM_ASSERT(LEVEL_1, loadHandle >= 0 && loadHandle < MAXLOAD);

    SMNDXLOADDESC* loadDesc = & loadTbl[loadHandle];
    
    SM_ASSERT(LEVEL_1, loadDesc->volid != -1);
    SM_ASSERT(LEVEL_1, loadDesc->iid == *ndx);
    SM_ASSERT(LEVEL_1, loadDesc->magic == SMNDXLOADDESCMAGIC);

    if (keyLen > loadDesc->maxKeyLen) 
	loadDesc->maxKeyLen = keyLen;
 
    switch (loadDesc->ndesc.ndxType) {
    case SM_BTREENDX:
	if (SM_InternalInsert(ndx, keyVal, keyLen, elem))
	    return esmFAILURE;
	break;
    case SM_HASHNDX:
	//
	// Create a temporary object
	//
	{ 
	    char* p = loadDesc->buffer;
	    * (int*) p = keyLen; p += ALIGNSIZE(sizeof(int));
	    BCOPY(keyVal, p, keyLen); p += ALIGNSIZE(keyLen);
	    BCOPY(elem, p, loadDesc->ndesc.elSize); 
	    p += loadDesc->ndesc.elSize;
	    OID dummy;
	    if (sm_CreateObject(loadDesc->groupIndex,
				&loadDesc->tmpFid, NEAR_LAST, 0, 0, 
				p - loadDesc->buffer,
				loadDesc->buffer, &dummy)) {
		return esmFAILURE;
	    }
	    loadDesc->totalRec++;
	}
	break;
    default:
	SM_ERROR(TYPE_FATAL, esmINTERNAL);
    }

    return esmNOERROR;
}


#ifdef COMMENT
/*
 *	This QuickSort function is broken.  It sorts correctly
 *	but runs very slow.
 */
const MAXSTACKDEPTH = 30;
const LIMIT = 10;


static void 
QuickSort(KEY key[], int cnt, PFC cmp)
{
    int stack[MAXSTACKDEPTH][2];
    int sp = 0;
    int l, r;
    KEY tmp;
    register i, j;
    register KEY pivot;
    static long randx = 1;		// random number root
    
    for (l = 0, r = cnt - 1; ; ) {
	if (r - l < LIMIT) {
	    if (sp-- <= 0) break;
	    i = stack[sp][0], r = stack[sp][1];
	    continue;
	}
	randx = (randx * 1103515245 + 12345) & 0x7fffffff;
	pivot = key[1 + randx % (r - 1)];
	for (i = l, j = r; i <= j; )  {
	    while (cmp(key[i].length, key[i].valuePtr, 
		       pivot.length, pivot.valuePtr) < 0) i++;
	    while (cmp(pivot.length, pivot.valuePtr, 
		       key[j].length, key[j].valuePtr) < 0) j--;
	    if (i < j) { tmp=key[i]; key[i]=key[j]; key[j]=tmp; }
	    if (i <= j) i++, j--;
	}
	
	if (j - l < r - i) {
	    if (i < r) stack[sp][0] = i, stack[sp++][1] = r;
	    r = j;
	}
	else {
	    if (l < j) stack[sp][0] = l, stack[sp++][1] = j;
	    l = i;
	}
    }
    
    for (i = 1; i < cnt; key[j+1] = pivot, i++)
	for (j = i - 1, pivot = key[i]; j >= 0 && 
	     (cmp(pivot.length, pivot.valuePtr, 
		  key[j].length, key[j].valuePtr) < 0); 
	     key[j+1] = key[j], j--);
}
#endif /* COMMENT */


static PFC localCmp;
static int qsortCmp(const void* k1, const void* k2)
{
	const KEY* key1 = (KEY*) k1;
	const KEY* key2 = (KEY*) k2;

	    return localCmp(key1->length, key1->valuePtr, 
		       key2->length, key2->valuePtr);
}
static void 
QuickSort(KEY key[], int cnt, PFC cmp)
{
    /*
    for (int i = 0; i < cnt; i++)  {
	printf("key[%d] = %x, %d\n", i, key[i].valuePtr,
	       key[i].length);
    }
    */
	localCmp = cmp;
	qsort(key, cnt, sizeof(KEY), qsortCmp);
}


class RunScan {
  OID oid;
  USERDESC* u;
  SCANDESC* s;
  BOOL eof;
  int group;
public:
  PFC cmp;
  RunScan()		{ u = 0, s = 0, eof = FALSE; }
  ~RunScan();
  
  Init(FID& fid, OID& begin, int groupIndex, PFC cmp);
  Current(void*& base, int& klen, char*& key, void*& elem); 
  Next(BOOL& eof);
  
  IsEof()	{ return eof; }
  
  friend operator>(RunScan& s1, RunScan& s2);
  friend operator<(RunScan& s1, RunScan& s2);
};


static void CreateHeap(short heap[], int heapSize, int num_runs, RunScan sc[]);
static HeapTop(short heap[], int heapSize, int winner, RunScan sc[]);

sm_AbortIndexLoad(IID* ndx)
{
    int loadHandle;

    if (checkTransState(ndx->pid.volid))  {
        return esmFAILURE;
    }
    if (sm_FindIndexLoad(ndx, &loadHandle))  {
		return esmFAILURE;
    }
	if (loadHandle < 0) {
		SM_ERROR(TYPE_USER, esmBULKLOADING);
		return esmFAILURE;
	}
    SM_ASSERT(LEVEL_1, loadHandle >= 0 && loadHandle < MAXLOAD);
    SMNDXLOADDESC* loadDesc = & loadTbl[loadHandle];
    
    SM_FreeLoadDesc(loadDesc);
    
    return esmNOERROR;
}


static
SM_CreateLHashRuns(SMNDXLOADDESC* loadDesc)
{
    SM_ASSERT(LEVEL_1, loadDesc->volid != -1);
    SM_ASSERT(LEVEL_1, loadDesc->magic == SMNDXLOADDESCMAGIC);

    int totalRec = loadDesc->totalRec;
    int tupleSize = (ALIGNSIZE(loadDesc->maxKeyLen) +
		     loadDesc->ndesc.elSize + sizeof(TWO) + 8);
    int pageCount = totalRec * tupleSize / (BTREE_PAGESIZE - 100) + 1;
    int initCount = INIT_PRIMARY;

    if (! loadDesc->ndesc.unique)  pageCount = (int) (pageCount / 1.2);
    
    for (int level = 0; initCount < pageCount; level++)
	initCount <<= 1;

    if (pageCount < INIT_PRIMARY)  pageCount = INIT_PRIMARY;

    FID fid = loadDesc->tmpFid;
    lhLevel = level;
    lhKeyType = loadDesc->ndesc.dataType;
    lhCmp = SMCOMPFUNC[ lhKeyType ];
    lhPrimary = pageCount;

    //
    // Create a new temporary to hold sorted runs, reinitialize 
    // state variables in load desc
    //
    if (sm_CreateFile(loadDesc->groupIndex, fid.pid.volid,
		      &loadDesc->tmpFid)) {
	return esmFAILURE;
    }
    loadDesc->totalRec = 0;
    loadDesc->offset = 0;

    SCANDESC* s;
    if (sm_OpenScanWithGroup(& fid, 0,
			     loadDesc->groupIndex, &s, 0))  {
	return esmFAILURE;
    }

    USERDESC* u;
    BOOL eof = FALSE;
    while (! eof)  {
	if (sm_ScanNextObject(s, 0, READ_ALL, &u, &eof))  {
	    return esmFAILURE;
	}
	if (! eof)  {
	    int keyLen = * (int*) u->basePtr;
	    void* key = u->basePtr + ALIGNSIZE(sizeof(int));
	    void* elem = ((char*) key) + ALIGNSIZE(keyLen);
	    if (SM_InternalInsert(&loadDesc->iid, key, keyLen, elem)) {
		return esmFAILURE;
	    }
	}
    }

    sm_CloseScan(s);
    if (sm_DestroyFile(loadDesc->groupIndex, &fid))  {
	return esmFAILURE;
    }

    return esmNOERROR;
}

sm_EndIndexLoad(IID* ndx)
{
    int loadHandle;
    if (checkTransState(ndx->pid.volid))  {
        return esmFAILURE;
    }
    if (sm_FindIndexLoad(ndx, &loadHandle))  {
		return esmFAILURE;
    }
    if (loadHandle < 0) {
        SM_ERROR(TYPE_USER, esmBULKLOADING);
        return esmFAILURE;
    }

    SM_ASSERT(LEVEL_1, loadHandle >= 0 && loadHandle < MAXLOAD);
    SMNDXLOADDESC* loadDesc = & loadTbl[loadHandle];
    
    SM_ASSERT(LEVEL_1, loadDesc->volid != -1);
    SM_ASSERT(LEVEL_1, loadDesc->iid == *ndx);
    SM_ASSERT(LEVEL_1, loadDesc->magic == SMNDXLOADDESCMAGIC);

    PFC cmp;
    switch (loadDesc->ndesc.ndxType) {
    case SM_BTREENDX:
	cmp = SMCOMPFUNC[ loadDesc->ndesc.dataType ];
	break;
    case SM_HASHNDX:
        if (SM_CreateLHashRuns(loadDesc))
            return esmFAILURE;
	cmp = LHashComp;
	break;
    default:
	SM_ERROR(TYPE_FATAL, esmINTERNAL);
    }
    
    if (SM_FlushLoadBuffer(loadHandle))
	return esmFAILURE;
    
    /*
     * 	------------ MERGING -------------
     */
    int numOutPart = loadDesc->runCount;
    int numInPart = 0;
    int groupIndex = loadDesc->groupIndex;
    VOLID tmpVol = loadDesc->tmpFid.pid.volid;
    
    int numPages = loadDesc->numPages;
    int numBuffers = loadDesc->runSize / BTREE_PAGESIZE;
    int numPasses = 0;
    for (int j = numBuffers; j < numPages; j *= numBuffers, numPasses++);
    
    if (numPasses > 0)  {
	FID& outFile = loadDesc->tmpFid;
	OID* outList = loadDesc->runList;
	int outListCnt = loadDesc->runCount;
	OID* listBuf = new OID[outListCnt + 1];
	if (! listBuf)  {
	    SM_FreeLoadDesc(loadDesc);
	    SM_ERROR(TYPE_USER, esmMALLOCFAILED);
	    return esmFAILURE;
	}
	OID* inList = listBuf;
	int inListCnt = 0;
	
	for (int k = numBuffers - 1, heapSize = 1; k > 0; 
	     heapSize <<= 1, k >>= 1);
	short* heap = new short[heapSize];
	int elSize = loadDesc->ndesc.elSize;
	SM_ASSERT(LEVEL_1, heap && outList);
	
	for (int i = 0; i < numPasses; i++) {
	    { OID* tmp = inList; inList = outList, outList = tmp; }
	    inListCnt = outListCnt; 
	    outListCnt = 0;
	    FID inFile = outFile;
	    
	    int lastPass = (i == numPasses - 1);
	    
	    if (sm_CreateFile(groupIndex, tmpVol, &outFile))  {
		delete[] listBuf;
		delete[] heap;
		SM_FreeLoadDesc(loadDesc);
		return esmFAILURE;
	    }
	    numInPart = numOutPart;
	    numOutPart = (numInPart - 1) / numBuffers + 1;
/*			
  printf("Pass %d, in: %d parts, out: %d parts\n", i, 
  numInPart, numOutPart);
  */
	    int cnt = 0;
	    int b;
	    for (j = b = 0; j < numOutPart; j++, b += numBuffers)  {
				// b is the # of input runs merged so far
		int numRuns = (j == numOutPart - 1) ? numInPart - b : 
		numBuffers;
		
		RunScan* sc = new RunScan[numRuns];
		for (k = 0; k < numRuns; k++)  {
		    if (sc[k].Init(inFile, inList[b+k], groupIndex, cmp))  {
			delete[] listBuf;
			delete[] heap;
			delete[] sc;
			SM_FreeLoadDesc(loadDesc);
			SM_ERROR(TYPE_LOG, sm_errno);
			return esmFAILURE;
		    }
		}
		
		if (numRuns == 1) 
		    heap[0] = 0;
		else {
		    for (k = numRuns - 1, heapSize = 1; k > 0; 
			 heapSize <<= 1, k >>= 1);
		    CreateHeap(heap, heapSize, numRuns, sc);
		}
		
		BOOL eof;
		void* base;
		void* elem;
		int klen;
		char* key;
		OID oid;
		
		BOOL newpart = TRUE;
		
		for (int r = heap[0]; numRuns > 1; 
		     r = HeapTop(heap, heapSize, r, sc))  {
		    if (sc[r].Current(base, klen, key, elem))  {
			delete[] listBuf;
			delete[] heap;
			delete[] sc;
			SM_FreeLoadDesc(loadDesc);
			return esmFAILURE;
		    }
		    if (sm_CreateObject(groupIndex, &outFile, NEAR_LAST, 0, 0,
					ALIGNSIZE(sizeof(int)) + ALIGNSIZE(klen) + elSize,
					base, &oid))  {
			delete[] listBuf;
			delete[] heap;
			delete[] sc;
			SM_FreeLoadDesc(loadDesc);
			return esmFAILURE;
		    }
		    cnt++;
		    // printf("%d\n", * (int*) key);
		    if (newpart)  {
			outList[outListCnt++] = oid;
			newpart = FALSE;
		    }
		    
		    if (sc[r].Next(eof))   {
			delete[] listBuf;
			delete[] heap;
			delete[] sc;
			SM_FreeLoadDesc(loadDesc);
			return esmFAILURE;
		    }
		    
		    if (eof)  {
			--numRuns; continue;
		    }
		}
		
		do {
		    if (sc[r].Current(base, klen, key, elem))   {
			delete[] listBuf;
			delete[] heap;
			delete[] sc;
			SM_FreeLoadDesc(loadDesc);
			return esmFAILURE;
		    }
		    if (sm_CreateObject(groupIndex, &outFile, NEAR_LAST, 0, 0,
					ALIGNSIZE(sizeof(int)) + ALIGNSIZE(klen) + elSize,
					base, &oid))  {
			delete[] listBuf;
			delete[] heap;
			delete[] sc;
			SM_FreeLoadDesc(loadDesc);
			return esmFAILURE;
		    }
		    cnt++;
		    // printf("%d\n", * (int*) key);
		    if (newpart)  {
			outList[outListCnt++] = oid;
			newpart = FALSE;
		    }
		    
		    if (sc[r].Next(eof))   {
			delete[] listBuf;
			delete[] heap;
			delete[] sc;
			SM_FreeLoadDesc(loadDesc);
			return esmFAILURE;
		    }
		    
		} while (! eof);
		
		if (! lastPass)  {
		    int marker = -1;
		    if (sm_CreateObject(groupIndex, &outFile, NEAR_LAST, 0, 0,
					sizeof(int), &marker, &oid)) {
			delete[] listBuf;
			delete[] heap;
			delete[] sc;
			SM_FreeLoadDesc(loadDesc);
			return esmFAILURE;
		    }
		}
/*					
  printf("------------ partition %d, merged %d\n", 
  outListCnt, cnt);
  */				
		delete[] sc;
	    }
	    
	    if (sm_DestroyFile(groupIndex, &inFile))  {
		delete[] listBuf;
		delete[] heap;
		SM_FreeLoadDesc(loadDesc);
		return esmFAILURE;
	    }
	}
    }
    
	//
	// reset the log level for the volume on which the index resides
	//
	if(checkTransState( ndx->pid.volid )) {
		SM_ERROR(TYPE_FATAL, esmINTERNAL);
    }
    //
    //	Create the index 
    //
    switch (loadDesc->ndesc.ndxType) {
    case SM_BTREENDX:
	if (bt_BulkLoad(TransRec.clientTid, & BufGroupPtr[groupIndex],
			loadDesc->volid, loadDesc->tmpFid, loadDesc->totalRec, 
			loadDesc->ndesc.unique, loadDesc->ndesc.dataType, 
			loadDesc->ndesc.elSize, loadDesc->ndesc.rootPid))  {
	    SM_FreeLoadDesc(loadDesc);
	    return esmFAILURE;
	}
	break;
    case SM_HASHNDX:
	if (lh_BulkLoad(TransRec.clientTid, loadDesc->volid,
			&loadDesc->ndesc, loadDesc->tmpFid, 
			&BufGroupPtr[groupIndex], 
			loadDesc->totalRec, lhLevel, lhPrimary))    {
	    SM_FreeLoadDesc(loadDesc);
	    return esmFAILURE;
	}
	break;
    default:
	SM_ERROR(TYPE_FATAL, esmINTERNAL);
    }

    //
    //	Clean up
    //
    SM_FreeLoadDesc(loadDesc);
    
    return esmNOERROR;
}





static void CreateHeap(short heap[], int heapSize, int num_runs, RunScan sc[])
{
    register    r;      /* a run # */
    register    s, k;   /* a heap index */
    register    winner; /* the winner run */
    int         tmp;
    
    for (s = 0; s < heapSize; heap[s++] = heapSize);
    for (r = 0, k = heapSize >> 1; r < heapSize; r += 2, k++) {
	/* compare pairs of headers in runs */
	if (r < num_runs - 1) { /* two real competitors */
	    if (sc[r] > sc[r+1])  
		heap[k] = r, winner = r+1;
	    else
		heap[k] = r+1, winner = r;
	}
        else {
	    heap[k] = -1; 		/* an artifical key */
	    winner = (r>=num_runs)?-1/* 2 phonies */:r/* seed */;
	}
	
        for (s = k >> 1;; s >>= 1) { /* propagate the winner upwards */
	    if (heap[s] == heapSize) { /* no one has reach here yet */
		heap[s] = winner; break;
	    }
	    if (winner < 0) /* a dummy key */
		winner = heap[s], heap[s] = -1;
	    else if (sc[winner] > sc[heap[s]])  {
				/* a new winner */ 
		tmp = winner; winner = heap[s]; heap[s] = tmp;
	    }
	}
    } 
}



static HeapTop(short heap[], int heapSize, int winner, RunScan sc[])
{
    register    s;      /* a heap index */
    register    r;      /* a run # */
    
    for (s = (heapSize + winner) >> 1; s > 0; s >>= 1)    {
        if ((r = heap[s]) < 0) 
	    continue; 
        if (sc[winner].IsEof())
	    heap[s] = -1, winner = r;
        else if (sc[winner] > sc[r]) 
	    heap[s] = winner, winner = r;
    }
    return(winner);
}







RunScan::Init(FID& fid, OID& begin, int groupIndex, PFC c)
{
    if (s)  {
	SM_ERROR(TYPE_FATAL, esmINTERNAL);
    }
/*		
  printf("init run %d.%d.%d\n", begin.diskAddr.volid, begin.diskAddr.page,
  begin.diskAddr.slot);
  */
    group = groupIndex;
    oid = begin;
    eof = FALSE;
    cmp = c;
    
    if (sm_OpenScanWithGroup(&fid, 0, groupIndex, &s, &begin))  {
	return esmFAILURE;
    }
    u = 0;
    if (Next(eof))  {
	SM_ERROR(TYPE_FATAL, esmINTERNAL);
    }
    
    return esmNOERROR;
}

RunScan::~RunScan()
{
    if (s) {
	int errno_save = sm_errno;
	sm_CloseScan(s);
	sm_errno = errno_save;
    }
}


RunScan::Current(void*& base, int& klen, char*& key, void*& elem)
{
    SM_ASSERT(LEVEL_1, u); 
    base = u->basePtr;
    klen = * (int*) base;
    key = u->basePtr + ALIGNSIZE(sizeof(int));
    elem = key + ALIGNSIZE(klen);
    
    return esmNOERROR;
}

RunScan::Next(BOOL& _eof)
{
    _eof = FALSE;
    if (eof)  { _eof = TRUE; u = 0; return esmNOERROR; }
    
    if (sm_ScanNextObject(s, 0, READ_ALL, &u, &eof)) {
	return esmFAILURE;
    }
    
    if (eof || u->objectSize == sizeof(int))  {
	eof = _eof = TRUE;
	u = 0;
	return esmNOERROR;
    }
    
    return esmNOERROR;
}


operator>(RunScan& s1, RunScan& s2)
{
    int klen1, klen2;
    klen1 = * (int*) s1.u->basePtr;
    klen2 = * (int*) s2.u->basePtr;
    
    return s1.cmp(klen1, s1.u->basePtr + ALIGNSIZE(sizeof(int)), klen2, 
		  s2.u->basePtr + ALIGNSIZE(sizeof(int))) > 0;
}
