/*
 *   $RCSfile: SM_CheckFree.c,v $  
 *   $Revision: 1.1.1.1 $  
 *   $Date: 1996/05/04 21:55:30 $      
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/


#include "sysdefs.h"
#include "ess.h"
#include "checking.h"
#include "trace.h"
#include "io.h"
#include "page.h"
#include "list.h"
#include "object.h"
#include "tid.h"
#include "lock.h"
#ifdef SERVER_MAKE
#   include "msgdefs.h"
#   include "thread.h"
#   include "semaphore.h"
#   include "latch.h"
#endif
#include "bf.h"
#ifndef SERVER_MAKE
#   include "chunk.h"
#endif
#ifdef SERVER_MAKE
#   include "link.h"
#   include "volume.h"
#endif
#include "bf_macro.h"
#include "sm_macro.h"
#include "error.h"
#ifndef SERVER_MAKE
#   include "scan.h"
#   include "load.h"
#include "sm_params.h"
#include "btree.h"
#endif
#include "sm_extfuncs.h"
#include "sm_intfuncs.h"


 int
#ifdef SERVER_MAKE
	SM_CheckFree (
#else
	SM_CheckFree (
#endif

	SLOTTEDPAGE				*homePage,
#ifndef SERVER_MAKE
	GROUPLINK				*homeLink,
#endif
	SLOTINDEX				slot,
	SMALLOBJ				*smallObject,
	int						addedSize 
)
{

	int						totalSize;
	int						endOffset;


	TRPRINT(TR_SM, TR_LEVEL_1, ("checking page:%d slot:%d addedSize:%d",
			homePage->header.pid.page, slot, addedSize));

	/*
	 *	Check the consistency of the home page
	 */
	sm_AuditSlotPage(homePage);

	/*
	 *	check to see if there is enough space on the
	 *	neighbor page to accomodate the record
	 */
	if (homePage->header.freeBytes < ALIGNSIZE(addedSize))	{

		/*
		 * see if there is room if we add in the alignment filler
		 * from the end of the object
		 */
		if ( (smallObject->objectHdr.properties == 0) ) {

			int newSize = (int) ALIGNSIZE(smallObject->objectHdr.size + addedSize);
			if ( homePage->header.freeBytes < (newSize - ALIGNSIZE(smallObject->objectHdr.size)) ) {
				TRPRINT(TR_SM, TR_LEVEL_2, ("insufficient free space"));
				return(PAGE_NO_SPACE);
			}

			/* ok there is room */

		} else {
			TRPRINT(TR_SM, TR_LEVEL_2, ("insufficient free space"));
			return(PAGE_NO_SPACE);
		}
	}

	/*
	 *	Calculate the ending address of the object
	 */
	if (smallObject->objectHdr.properties & P_MOVED)	{

		/*
		 *	total size is header plus oid
		 */
		totalSize = ALIGNSIZE(sizeof(OBJHDR) + sizeof(OID));

	} else {
		
		/*
		 *	total size is header plus object
		 */
		totalSize = ALIGNSIZE(sizeof(OBJHDR) + ALIGNSIZE((int)smallObject->objectHdr.size));
	}

	/*
	 *	Calculate the ending address of the object
	 */
	endOffset = ((((char *) smallObject) + totalSize) - ((char *) homePage));
	TRPRINT(TR_SM, TR_LEVEL_2, ("freeStart:%d, endOffset:%d",
			homePage->header.freeStart, endOffset));

	/*
	 *	check to see if the object is the last one on the page
	 */
	if (endOffset == homePage->header.freeStart)	{

		TRPRINT(TR_SM, TR_LEVEL_2, ("slot is last on page"));

		/*
		 *	check to see if we have enough space
		 */
		if (CONTIGSPACE(homePage) >= addedSize)	{

			/*
			 *	return that we have enough space
			 */
			return(PAGE_LAST_SLOT);
		}

	} else {

		TRPRINT(TR_SM, TR_LEVEL_2, ("slot is not last on page"));

		if (CONTIGSPACE(homePage) >= totalSize + addedSize)	{

			/*
			 *	return that we have enough space
			 */
			return(PAGE_INT_SLOT);
		}
	}

#ifndef SERVER_MAKE
	/*
	 *	check to see if it is possible to compact the page
	 */
	if (SM_CheckPageUse(homeLink))	{

		/*
		 *	page cannot be compacted
		 */
		return(PAGE_NO_SPACE);
	}
#endif

	/*
	 *	check the consistency of the page
	 */
	sm_AuditSlotPage(homePage);

	/*
	 *	attempt to compact the page
	 */
	if (SM_CompactPage( homePage, slot ))	{

		TRPRINT(TR_SM, TR_LEVEL_2, ("compact failed"));

		/*
		 *	check the consistency of the page
		 */
		sm_AuditSlotPage(homePage);

		/*
		 *	return that we need a new neighbor
		 */
		return(PAGE_NO_SPACE);

	} else {

		/*
		 *	check the consistency of the page
		 */
		sm_AuditSlotPage(homePage);

		/*
		 *	return that the slot is the last on the page
		 */
		return(PAGE_LAST_SLOT);
	}
}
