/*
 *   $RCSfile: SM_CompactPage.c,v $  
 *   $Revision: 1.1.1.1 $  
 *   $Date: 1996/05/04 21:55:30 $      
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/


#include "sysdefs.h"
#include "ess.h"
#include "checking.h"
#include "io.h"
#include "page.h"
#include "list.h"
#include "object.h"
#include "tid.h"
#include "lock.h"
#ifdef SERVER_MAKE
#   include "msgdefs.h"
#   include "thread.h"
#   include "semaphore.h"
#   include "latch.h"
#endif
#include "bf.h"
#ifndef SERVER_MAKE
#   include "chunk.h"
#endif
#ifdef SERVER_MAKE
#   include "link.h"
#   include "volume.h"
#endif
#include "bf_macro.h"
#include "sm_macro.h"
#include "error.h"
#include "trace.h"
#ifndef SERVER_MAKE
#   include "scan.h"
#   include "load.h"
#include "sm_params.h"
#include "btree.h"
#endif
#include "bf_extfuncs.h"
#include "sm_intfuncs.h"
#include "sm_extfuncs.h"
#ifndef SERVER_MAKE
#	include "util_funcs.h"
#endif


/*============================================================================*/
/* 
 * SM_CompactPage:
 * 	This routine compacts a slotted page, moving all objects
 *	to the top of the page.
 *
 *	The parameter magicSlot is needed when the compaction is
 *	being done on behalf of an insert or append.
 *	If magicSlot is not equal to NIL, then the object pointed at by
 *	magicSlot is not compacted, and the value of the slot[-magicSlot]
 *	is set to the start of the page's free region after compaction.
 *	
 * RETURNS:
 *      esmNOERROR
 *
 * ERROR RETURNS:
 *	none
 *
 * SIDE EFFECTS:
 *	The objects on the page are shuffled around to compact the page.
 *	The page user descriptors for the page are updated as required.
 *	
 */
/*============================================================================*/


 int
SM_CompactPage (

	SLOTTEDPAGE		*page,
	SLOTINDEX		magicSlot 
)
{

    register SLOTINDEX	i;				/* slot index */
    register int		smallestOffset;	/* smallest offset in compaction pass */
	register PAGESLOT	*saveSlot;
	register PAGESLOT	*currentSlot;
	register PAGESLOT	*firstSlot;
	register PAGESLOT	*lastSlot;
    register int		slotCount;		/* number of slots in the page */
    int					currentOffset;	/* current offset during compaction */
	int					currentFree;
    int					bytesToMove;	/* bytes to move in compaction pass */
	int					magicBytes;
    OBJECT				*objectPtr;		/* object pointer */
	GROUPLINK			*tempBuf;
	int					freeArea;

	TRPRINT(TR_SM, TR_LEVEL_1, ("magic slot:%d", magicSlot));
#ifndef SERVER_MAKE
	sm_PrintSlotPage(page);
#endif

	/*
	 *	if magic slot is non-null then we must save
	 *	the contents of the record.
	 */
	if (magicSlot != NIL)	{

		/*
		 *	get a temporary buffer
		 */
		if ((tempBuf = bf_GetTempPage(SLOTTED_PAGE2SIZE)) == NULL)	{

			return(esmFAILURE);
		}

		/*
		 *	get a pointer to the record
		 */
		saveSlot = GETSLOTPTR(page, magicSlot);
		objectPtr = GETOBJECTPTR(page, saveSlot);

		/*
		 *	calculate the number of bytes that will
		 *	have to be moved
		 */
		if (objectPtr->objectHdr.properties & P_LARGEOBJ) {

			/*
			 *	Check to see if it is a header on slotted page
			 */
			if (objectPtr->objectHdr.properties & P_HEADERONPAGE)	{

				TRPRINT(TR_SM, TR_LEVEL_2, ("magic slotted large object"));

				/*
				 *	move the large object header
				 */
				magicBytes = ALIGNSIZE(sizeof(LARGEOBJROOT));

			} else {

				TRPRINT(TR_SM, TR_LEVEL_2, ("magic large object"));

				/*
				 *	move the large object header
				 */
				magicBytes = ALIGNSIZE(sizeof(LARGEOBJHDR));
			}

		} else {

			/*
			 *	check to see if this is a forwarded object
			 */
			if (objectPtr->objectHdr.properties & P_MOVED)	{

				TRPRINT(TR_SM, TR_LEVEL_2, ("magic moved"));

				/*
				 *	move the whole small object
				 */
				magicBytes = sizeof(OBJHDR) + ALIGNSIZE(sizeof(OID));

			} else {

				TRPRINT(TR_SM, TR_LEVEL_2, ("magic normal small"));

				/*
				 *	move the whole small object
				 */
				magicBytes = sizeof(OBJHDR) + ALIGNSIZE((int)objectPtr->objectHdr.size);
			}
		}
		TRPRINT(TR_SM, TR_LEVEL_2, ("magicBytes:%d", magicBytes));

		/*
		 *	save the record to the temporary buffer
		 */
		bcopy((char *) objectPtr, tempBuf->bufFrame, (int) magicBytes);

		/*
		 *	Nil out the slot so that it gets compacted
		 */
		saveSlot->offset = NIL;
	}

    /*
	 *	Save the number of slots in the page
     */
    slotCount = page->header.slotCount;
	TRPRINT(TR_SM, TR_LEVEL_2, ("slots in page:%d", slotCount));
    
	/*
	 *	setup pointers to the first and last slots in the page
	 */
	firstSlot = &(page->slot[0]);
	lastSlot = firstSlot - slotCount;

    /*
     *	Compact the page.
	 *
     *	We should use a heap sort here, no time
     *	put it in right now.
     */
    for (currentFree = 0, i = 0; i < slotCount; i++) {

		/*
		 *	look for the slot that has the smallest offset
		 *	that is greater than or equal to currentFree.
		 */
		smallestOffset = SLOTTED_PAGESIZE;

		/*
		 * On each pass, find the slot with the smallest offset that
		 * hasn't been scanned already.
		 */
		for (currentSlot = firstSlot; currentSlot > lastSlot; currentSlot--)	{

			/*
			 *	grab the current slot offset
			 */
			currentOffset = (int)currentSlot->offset;
			TRPRINT(TR_SM, TR_LEVEL_2, ("currentFree:%d currentOffset:%d smallestOffset:%d",
					currentFree, currentOffset, smallestOffset));	

			/*
			 *	Check to see if we have a candidate slot
			 */
			if ((currentOffset != NIL) && (currentOffset < smallestOffset))	{

				/*
				 *	check to see if the currentOffset is equal to
				 *	the currentFree.  In this case, the page
				 *	is not to be moved but currentFree must be
				 *	incremented by the size of the object
				 */
				if (currentOffset == currentFree)	{

					/*
					 *	get out of this loop as we've found the
					 *	next element that should be used
					 */
					smallestOffset = currentOffset;
					saveSlot = currentSlot;
					TRPRINT(TR_SM, TR_LEVEL_2, ("next object compacted"));
					break;


				} else if (currentOffset > currentFree)	{

					/*
					 *	record this offset and continue looking
					 */
					smallestOffset = currentOffset;
					TRPRINT(TR_SM, TR_LEVEL_2, ("new smallest offset:%d slot:%d",
							smallestOffset, currentSlot));
					saveSlot = currentSlot;
					continue;


				} else {

					TRPRINT(TR_SM, TR_LEVEL_2, ("skipping slot"));
				}

			} else {

				TRPRINT(TR_SM, TR_LEVEL_2, ("skipping slot"));
			}
		}

		/*
		 * If all the valid slots have been scanned, then 
		 * the page has been compacted already.
		 */
		if (smallestOffset == SLOTTED_PAGESIZE) {

			/*
			 * Update the page's freeStart offset.
			 */
			page->header.freeStart = currentFree;
			TRPRINT(TR_SM, TR_LEVEL_2, ("seting freestart:%d", currentFree));

			/*
			 *	get out of the loop
			 */
			TRPRINT(TR_SM, TR_LEVEL_2, ("all objects compacted"));
			break;
		}

		/*
		 *	get a pointer to the object in question
		 */
		objectPtr = GETOBJECTPTR(page, saveSlot);

		/*
		 *	calculate the number of bytes that will
		 *	have to be moved
		 */
		if (objectPtr->objectHdr.properties & P_LARGEOBJ) {

			/*
			 *	Check to see if it is a header on slotted page
			 */
			if (objectPtr->objectHdr.properties & P_HEADERONPAGE)	{

				TRPRINT(TR_SM, TR_LEVEL_2, ("large slotted header"));

				/*
				 *	move the large object header
				 */
				bytesToMove = ALIGNSIZE(sizeof(LARGEOBJROOT));

			} else {

				TRPRINT(TR_SM, TR_LEVEL_2, ("large object")); 

				/*
				 *	move the large object header
				 */
				bytesToMove = ALIGNSIZE(sizeof(LARGEOBJHDR));
			}

		} else {

			/*
			 *	check to see if this is a forwarded object
			 */
			if (objectPtr->objectHdr.properties & P_MOVED)	{

				TRPRINT(TR_SM, TR_LEVEL_2, ("moved object"));

				/*
				 *	move the whole small object
				 */
				bytesToMove = sizeof(OBJHDR) + ALIGNSIZE(sizeof(OID));

			} else {

				TRPRINT(TR_SM, TR_LEVEL_2, ("small object"));

				/*
				 *	move the whole small object
				 */
				bytesToMove = sizeof(OBJHDR) + ALIGNSIZE((int)objectPtr->objectHdr.size);
			}
		}
		TRPRINT(TR_SM, TR_LEVEL_2, ("bytes:%d", bytesToMove));

		/*
		 *	move the data bytes if necessary
		 */
		if (smallestOffset > currentFree)	{

			/*
			 *	get a pointer to the object to be moved
			 */
			objectPtr = GETOBJECTPTR(page, saveSlot);

			/*
			 *	move the object up to contiguous space
			 */
			bcopy((char *) objectPtr, ((char *) page) + currentFree, (int) bytesToMove);
			TRPRINT(TR_SM, TR_LEVEL_2, ("copying %d bytes to:%d",
					bytesToMove, currentFree));
			
			/*
			 * Update the slot's offset
			 */
			saveSlot->offset = currentFree;

			/*
			 *	update the beginning of the current free space
			 */
			currentFree += bytesToMove;

		} else {

			/*
			 *	update the current free bytes
			 */
			currentFree += bytesToMove;
		}

		/*
		 * Update the page's freeStart offset.
		 */
		page->header.freeStart = currentFree;
		TRPRINT(TR_SM, TR_LEVEL_2, ("seting freestart:%d", currentFree));
    }

	/*
	 *	If the page is empty, just set up the free space begin at the
	 *	first byte in the page
	 */
	if (slotCount == 0) {
		SM_ASSERT(LEVEL_3, magicSlot == NIL);
		currentFree = 0;
		page->header.freeStart = 0;
	}
	
#ifndef SERVER_MAKE
	sm_PrintSlotPage(page);
#endif
	/*
	 *	if magic slot is non-null then we must save
	 *	the contents of the record.
	 */
	if (magicSlot != NIL)	{

		/*
		 *	get a pointer to the record
		 */
		objectPtr = (OBJECT *) tempBuf->bufFrame;
		TRPRINT(TR_SM, TR_LEVEL_2, ("magicBytes:%d", magicBytes));

		/*
		 *	save the record to the temporary buffer
		 */
		bcopy((char *) objectPtr, ((char *) page) + page->header.freeStart, (int) magicBytes);

		/*
		 *	get a pointer to the correct slot
		 */
		saveSlot = GETSLOTPTR(page, magicSlot);

		/*
		 *	Nil out the slot so that it gets compacted
		 */
		saveSlot->offset = page->header.freeStart;

		/*
		 *	update freestart to reflect the new changes
		 */
		page->header.freeStart += magicBytes;

		/*
		 *	free up the temporary buffer
		 */
		bf_FreeTempPage(tempBuf);
	}

#ifndef SERVER_MAKE
	sm_PrintSlotPage(page);
#endif

    /*
     *  Make sure the compaction succeeded.  First calculate the size
     *  of the free area based on the end of the slot array and the
     *  beginning of the free area.  If this is not equal to the
     *  free byte count, the compaction failed.
     */
    freeArea = (int)CONTIGSPACE(page);
    if ( freeArea != page->header.freeBytes) {

        SM_ERROR(TYPE_FATAL, esmINTERNAL);
    }

	/*
	 *	return success
	 */
    return(esmNOERROR);
}
