/*
 *   $RCSfile: SM_GetVersionGraph.c,v $  
 *   $Revision: 1.1.1.1 $  
 *   $Date: 1996/05/04 21:55:29 $      
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/


#include "ess.h"
#include "sysdefs.h"
#include "checking.h"
#include "io.h"
#include "page.h"
#include "list.h"
#include "object.h"
#include "tid.h"
#include "bf.h"
#include "chunk.h"
#include "lock.h"
#include "bf_macro.h"
#include "sm_macro.h"
#include "error.h"
#include "trace.h"
#include "trans.h"
#include "file.h"
#include "scan.h"
#include "load.h"
#include "sm_params.h"
#include "btree.h"
#include "pool.h"
#include "version_graph.h"
#include "version_funcs.h"
#include "sm_intfuncs.h"
#include "sm_version_intfuncs.h"
#include "sm_extfuncs.h"
#include "sm_globals.h"


extern TRANSREC	TransRec;
extern int		sm_errno;


 int
SM_GetVersionGraph (

	BUFGROUP			*bufGroup,
	register OID		*oid,		/* oid for the vhg				*/
	VERSIONGRAPH		**vhg,
	BOOL				makeSpace,	/* insure VHG has room for new node */
	LOCKMODE			lockMode 
)
{
	USERDESC		*vhgUserDesc;
	VERSIONGRAPH	*vhgPtr;
	int				expansionAmount = 0;/* number of new nodes to add */

	TRPRINT(TR_SM, TR_LEVEL_1, ("get version: makeSpace=%d", makeSpace));
	/*
	 *	Read the entire version history graph
	 */
	if (SM_ReadObject(bufGroup, oid, 0, -1, &vhgUserDesc, NULL, TRUE, lockMode)) {
		return (esmFAILURE);	
	}

	/*
	 *	Make sure the VHG has room to create at least one more
	 *	version, by making sure the free list isn't empty.
	 */
	vhgPtr = (VERSIONGRAPH*)vhgUserDesc->basePtr;
	CHECK_VERSIONGRAPH_MAGIC(vhgPtr);
	SM_ASSERT(LEVEL_3, vhgUserDesc->userFlags & P_HISTORYGRAPH);
	if ( VHGLIST_EMPTY(vhgPtr, &vhgPtr->freeList) && makeSpace ) {
		/*
		 *	List is empty, so expand the graph space
		 */
		TRPRINT(TR_SM, TR_LEVEL_1, ("expanding graph"));

		/*
		 *	First release the user desc for the graph, since a new one
		 * 	will be needed with the larger size
		 */
		if (sm_ReleaseObject(vhgUserDesc)) {
			SM_ERROR(TYPE_FATAL, esmINTERNAL);
		}

		/*
		 *	Expand the version graph object by enough space for
		 *	5 more nodes.
		 */
		expansionAmount = 5;
		if (sm_AppendToObject(bufGroup->index, oid, 
							  expansionAmount*sizeof(VHGNODE), NULL)) {
			SM_ERROR(TYPE_WARNING, esmVHGTOOLARGE);
			return (esmFAILURE);	
		}

		/*
		 *	Read the entire version history graph
		 */
		if (SM_ReadObject(bufGroup, oid, 0, -1, &vhgUserDesc, NULL, TRUE, EX)) {
			return (esmFAILURE);	
		}

	}

	/*
	 *	Make space for a copy of the graph
	 */
	if (VersionGraphSpaceSize < vhgUserDesc->byteCount) {
		VersionGraphSpace = (char*) realloc(VersionGraphSpace, (unsigned int) vhgUserDesc->byteCount);
		if (VersionGraphSpace == NULL) {
			SM_ERROR(TYPE_SYS, esmMALLOCFAILED);
			if (sm_ReleaseObject(vhgUserDesc)) {
				SM_ERROR(TYPE_FATAL, esmINTERNAL);
			}
			return esmFAILURE;	
		}
		VersionGraphSpaceSize = (int) vhgUserDesc->byteCount;
	} 

	vhgPtr = (VERSIONGRAPH*) VersionGraphSpace;

	/*
	 *	Copy the graph
	 */
	bcopy(vhgUserDesc->basePtr, (char*) vhgPtr, (int) vhgUserDesc->byteCount);

	/*
	 *	Unpin the version graph, the copy will be used instead
	 */
	if (sm_ReleaseObject(vhgUserDesc)) {

		SM_FREE_PTR(vhgPtr);
		return (esmFAILURE);	
	}

	/*
	 *	If new nodes need to be initialized, do so.  The changes
	 *	will be written to the VHG object when a 
	 *	SM_SetVersionGraph() call is make.
	 */
	if (expansionAmount > 0) {
		if (initializeNewVHGNodes(vhgPtr, expansionAmount) ) {
			SM_ERROR(TYPE_FATAL, esmINTERNAL);
		}
	}

	*vhg = vhgPtr;
    return esmNOERROR;
}
