/*
 *   $RCSfile: SM_InitClientOptions.C,v $
 *   $Revision: 1.1.1.1 $
 *   $Date: 1996/05/04 21:55:30 $
 */

/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/

#include "sysdefs.h"
#include "ess.h"
#include "checking.h"
#include "error.h"
#include "trace.h"
#include "io.h"
#include "page.h"
#include "list.h"
#include "object.h"
#include "tid.h"
#include "bf.h"
#include "chunk.h"
#include "sm_macro.h"
#include "scan.h"
#include "load.h"
#include "sm_params.h"
#include "btree.h"
#include "volume.h"
#include "lock.h"
#include "option_macros.h"
#include "option.h"
#include "sm_intfuncs.h"
#include "sm_extfuncs.h"
#include "io_extfuncs.h"
#include "trace_funcs.h"
#include "sm_globals.h"

#ifndef __cplusplus
extern char*	getenv(char*);
#endif


int setOptionCallBack(OptionInfo* optInfo, const char* value)
{
	return(SM_SetOption(optInfo->name(), value));
}

int
parseMount( 
	const char *origstring, 
	char **errmsg, 
	int  *errid 
)
{
	char	str[MAXOPTLENGTH];
	char	*string = str;
	char	*hoststring, *portstring;
	int 	len;
	VOLID	volid;

	*errid = 0;
	*errmsg = NULL;

	(void ) strcpy(str, origstring);

	len = strspn(str, "0123456789");
	if(len < 1) {
novolid:
		*errmsg = "Missing volid; syntax is: volid port@host";
		*errid = esmOPTIONSYNTAX;
		return esmFAILURE;
	}
	str[len]='\0';
	volid = (VOLID) atoi(str);
	if(volid <=0) {
		goto novolid;
	}
	string = &str[len+1]; /* go to first char after the delimiter */

	portstring = strtok(string, "@ 	"); /* @, space, tab */

	hoststring = strtok(0, "@ 	"); /* @, space, tab */

	if (io_AddVolume(volid, hoststring, portstring) != esmNOERROR) {
		*errmsg = sm_Error(sm_errno);
		*errid = esmBADOPTIONVALUE;
		return esmFAILURE;
	}
		
	return esmNOERROR;
}

int
parseVolume( 
	const char *unused, 
	char **errmsg, 
	int  *errid 
)
{
	*errmsg = "Unknown such client option.";
	*errid = esmBADOPTION;
	return esmBADOPTION;
}

int MountOption(OptionInfo* optInfo, const char* value)
{
	int errId;
	char *errMsg;

	if (strcmp("", value) == 0) {
		// empty string, so this must be dummy initialization
	} else {
		if (parseMount(value, &errMsg, &errId) != esmNOERROR) {
			fprintf(sm_ErrorStream, "CLIENT: Error in setting option %s: %s\n", optInfo->name(), errMsg);
			return esmFAILURE;
		}
	}
	return esmNOERROR;
}

/*
 *	this global is defined here since space.c can be compiled with C
 */
extern OptionList*	ClientOptions;

 int
SM_InitClientOptions ()
{
	OptionInfo*	opt;
	char*		configEnvVar = "ESMCONFIG";
	char*		errorMsg;

	// make sure we can print things
	if (sm_ErrorStream == NULL) sm_ErrorStream = stderr;

	if (ClientOptions == NULL) {
		ClientOptions = new OptionList(NUM_OPTIONS);
		if (ClientOptions == NULL) {
			SM_ERROR(TYPE_SYS, esmMALLOCFAILED);
			return esmFAILURE;
		}
	} else {
		return esmNOERROR;
	}

	/*
	 *	Set the classification hierarchy for naming options.
     *	Initially the hierarchy is: client.optionName
	 */
	ClientOptions->addClassLevel("client");	// program type

	ClientOptions->addOpt()->init(SM_OPT_bufpages, "bufpages",
						"#>4", NULL, 
						"number of pages in the buffer pool",
						TRUE, setOptionCallBack);
	ClientOptions->addOpt()->init(SM_OPT_locktimeout, "locktimeout",
						"#>=0", "30",
						"10-second units to await a lock",
						FALSE, setOptionCallBack);
	ClientOptions->addOpt()->init(SM_OPT_groups, "groups",
						"#>3", "20", 
						"number of buffer groups", FALSE, setOptionCallBack);
	ClientOptions->addOpt()->init(SM_OPT_userdesc, "userdesc",
						"#>0", "2000", 
						"number of user descriptors",
						FALSE, setOptionCallBack);
	ClientOptions->addOpt()->init(SM_OPT_traceflags, "traceflags",
						"hex number", "0x0", 
						"sets tracing flags (for use by developers)",
						FALSE, setOptionCallBack);
	ClientOptions->addOpt()->init(SM_OPT_lognewpages, "lognewpages",
						"yes, no, ", "no", 
						"turn on/off logging for new pages on client",
						FALSE, setOptionCallBack);
	ClientOptions->addOpt()->init(SM_OPT_deallocpages, "deallocpages",
						"yes, no, ", "yes", 
						"turn on/off deallocation of small object pages",
						FALSE, setOptionCallBack);
	ClientOptions->addOpt()->init(SM_OPT_filelock, "filelock",
						"NL,IS,IX,SH,SIX,EX", "SH", 
						"default lock mode for files",
						FALSE, setOptionCallBack);
	ClientOptions->addOpt()->init(SM_OPT_pagelock, "pagelock",
						"NL,IS,IX,SH,SIX,EX", "SH", 
						"default lock mode for pages",
						FALSE, setOptionCallBack);

	ClientOptions->addOpt()->init(SM_OPT_mount, "mount",
						"volumeID  port@host", "", 
						"identify a volume and the server managing it",
						FALSE, MountOption);

	// Set the defaults on all non-required options
	ClientOptions->initScan();
	for (opt = ClientOptions->getNext(); opt != NULL; opt = ClientOptions->getNext()) {
		// mark that the option has not been set so future sets will
		// override defaults
		opt->unsetValue();
	}

	// see if the default configuration file should be read
	if (getenv(configEnvVar) != NULL) {

		if (sm_ReadConfigFile(getenv(configEnvVar), NULL, &errorMsg) != esmNOERROR) {
			if (errorMsg != NULL) {
				fprintf(sm_ErrorStream, "CLIENT: error in configuration file: %s\n", errorMsg);
			}
			return esmFAILURE;
		}
	}

    return esmNOERROR;
}
