/*
 *   $RCSfile: SM_ReturnSlot.c,v $  
 *   $Revision: 1.1.1.1 $  
 *   $Date: 1996/05/04 21:55:30 $      
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/


#	include "sysdefs.h"
#include "ess.h"
#include "checking.h"
#include "io.h"
#include "page.h"
#include "list.h"
#include "pool.h"
#include "object.h"
#include "tid.h"
#include "lock.h"
#ifdef SERVER_MAKE
#	include "msgdefs.h"
#	include "thread.h"
#	include "semaphore.h"
#	include "latch.h"
#endif
#include "bf.h"
#ifndef SERVER_MAKE
#	include "chunk.h"
#endif
#ifdef SERVER_MAKE
#	include "link.h"
#endif
#include "volume.h"
#include "bf_macro.h"
#include "sm_macro.h"
#include "error.h"
#include "trace.h"
#ifndef SERVER_MAKE
#	include "scan.h"
#	include "load.h"
#	include "file.h"
#	include "sm_params.h"
#	include "btree.h"
#	include "fi_extfuncs.h"
#	include "fi_extfuncs.h"
#	include "sm_globals.h"
#endif
#include "bf_extfuncs.h"
#include "io_extfuncs.h"
#include "sm_intfuncs.h"
#include "sm_extfuncs.h"

 static void
SM_FreeEmptySlots(SLOTTEDPAGE *page)
{
	SLOTINDEX	last, curr, prev;
	int 		numContig, numRemoved;
	PAGESLOT	*slotPtr;

	/* find contiguous empty slots at the end of the slot array */
	last = page->header.slotCount-1;
	while (last >= 0) {
		slotPtr = GETSLOTPTR(page, last);
		if (slotPtr->offset != NIL) {
			break;
		}
		last--;
	}
	last = last + 1;
	numContig = page->header.slotCount - last;

	if (numContig > 0) {

		numRemoved = 0;

		/*
		 *	generate a new free list removing the
		 *	contiguous empty slots at the end of the slot array
		 */
		prev = NIL;
		curr = page->header.freeSlot;
		SM_ASSERT(LEVEL_3, curr != NIL);
		while (curr != NIL) {
			slotPtr = GETSLOTPTR(page, curr);
			if (curr >= last) {
				/* remove curr from the free list */
				if (prev == NIL) {
					/* first on list */
					page->header.freeSlot = (SLOTINDEX) slotPtr->unique;
					curr = page->header.freeSlot;  /* start over */
				} else {
					/* link previous to next */
					GETSLOTPTR(page, prev)->unique = slotPtr->unique; 
					curr = (SLOTINDEX) slotPtr->unique;
				}
				numRemoved++;
			} else {
				/* advance to the next free slot */
				prev = curr;
				curr = (SLOTINDEX) slotPtr->unique;
			}
		}
		SM_ASSERT(LEVEL_3, numRemoved == numContig);

		/*
		 *	free the space from the contiguous slots
		 */
		page->header.freeBytes += numContig * sizeof(PAGESLOT);
		page->header.slotCount -= numContig;
	}
	SM_ASSERT(LEVEL_3, (page->header.slotCount == last));
}


 int
SM_ReturnSlot (

	GROUPLINK				*groupLink,
	SLOTINDEX				slot,
	int						freeSize 
)
{

	SLOTTEDPAGE				*page;
	PAGESLOT				*slotPtr;
#ifndef SERVER_MAKE
	FILEINFO				*fileInfo;
#endif /* SERVER_MAKE */


	TRPRINT(TR_SM, TR_LEVEL_1, ("slot:%d", slot));

	/*
	 *	get a pointer to the page
	 */
	page = (SLOTTEDPAGE *) groupLink->bufFrame;

    /*
     *	Get a pointer to the object
     */
	slotPtr = GETSLOTPTR(page, slot);

	/*
	 *	Dealing with a small object
	 */
	page->header.freeBytes += sizeof(OBJHDR) + ALIGNSIZE(freeSize);

    /*
     * Mark the object's slot as empty
     */
    slotPtr->offset = NIL;

	/*
	 *	We can garbage collect slot if we just freed the last slot entry
	 */
    if (slot == page->header.slotCount - 1) {

		TRPRINT(TR_SM, TR_LEVEL_2, ("garbage collecting last slot"));

		/*
		 *	Decrement the slot count
		 */
		page->header.slotCount--;

		/*
		 *	increment the page free space
		 */
		page->header.freeBytes += sizeof(PAGESLOT);

		SM_FreeEmptySlots(page);

    } else {

		/*
		 *	hook the slot onto the free slot list
		 */
		slotPtr->unique = page->header.freeSlot;
		TRPRINT(TR_SM, TR_LEVEL_2, ("old first free slot:%d", page->header.freeSlot));
		page->header.freeSlot = slot;
		TRPRINT(TR_SM, TR_LEVEL_2, ("new first free slot:%d", page->header.freeSlot));
	}

    /*
     *	Is the home page empty now? If so, then free it up.
     */
    if (page->header.slotCount == 0) {

		TRPRINT(TR_SM, TR_LEVEL_2, ("page is empty"));

		/*
		 *	check to see if the page is really slotted
		 */
#ifdef SERVER_MAKE
	
		/* 
		 *	No check possible since it might have been undone
		 */
#else
		SM_ASSERT(LEVEL_3, (bf_CheckBitmap( &(page->header.pid), PAGE_SLOTTED)) == esmNOERROR);
#endif


#ifndef SERVER_MAKE

		/* 
		 *	find info about the file containing the page
		 *
		 *	This could be optimized by caching the fileInfo pointer
		 *	in the global SM_State variable, but this should be fine
		 *	at least for now.
		 */
		fileInfo = fi_FindFileInfo(&page->header.fid, FALSE);
		if (fileInfo == NULL) {
			return(esmFAILURE);  /* this should not happen */
		}

		/* 
		 *	If page is not on a temporary volume, then it is ok
		 *	to deallocate the page since it cannot be in a logically
		 *	ordered file.
		 */
		if ( ! fileInfo->tempVolume  && DeallocSmallObjPages) {

			/*
			 * If there is a scan running make sure this is not
			 * the current page.  If so then unfix it from the scan.
			 */
			if (!LIST_EMPTY(&UsedScanDesc)) {
				SCANDESC *scanDesc = (SCANDESC*) FIRST_LIST_ELEMENT(&UsedScanDesc);
				while (scanDesc != NULL) {
					if (PIDEQ(page->header.pid, scanDesc->currentPid)) {
						if (scanDesc->currentLink != NULL) {
							SM_ASSERT(LEVEL_3, PIDEQ(scanDesc->currentLink->pageHash->hashList.hashPid, page->header.pid));
							bf_UnfixPage(scanDesc->currentLink, BF_DEFAULT, FALSE);
							scanDesc->currentLink = NULL;
							scanDesc->currentPage = NULL;
						}
					}
					scanDesc = (SCANDESC*) NEXT_LIST_ELEMENT( &(scanDesc->scanList) );
				}
			}

			/*
			 *	Deallocate the page and then discard it.
			 */
			if (io_DeallocSlottedPages(&(page->header.pid), fileInfo, FALSE))	{
				return(esmFAILURE);
			}
			bf_DiscardPage(groupLink->bufGroup, groupLink);

		} else {

			/*
			 *	make sure that the all free space is marked as available
			 *	contiguously
			 */
			page->header.freeStart = 0;

			/*
			 *	Mark the home page as dirty, unfix it, and return.
			 *	Don't unfix it if this is the server.
			 */
#ifdef SERVER_MAKE
#else
			bf_UnfixPage(groupLink, BF_DEFAULT, TRUE);
#endif

		}
#endif /* not SERVER_MAKE */

    } else {

		/*
		 *	There are still valid objects on the page.
		 *	Mark the home page as dirty, unfix it, and return.
		 *	Don't unfix it if this is the server.
		 */
#ifdef SERVER_MAKE
#else
		sm_AuditSlotPage(page);
		bf_UnfixPage(groupLink, BF_DEFAULT, TRUE);
#endif
	}

	/*
	 *	return success
	 */
    return(esmNOERROR);
}
