/*
 *   $RCSfile: BF_FindFreeSlot.c,v $  
 *   $Revision: 1.1.1.1 $  
 *   $Date: 1996/05/04 21:55:33 $      
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/


#include "sysdefs.h"
#include "ess.h"
#include "checking.h"
#include "trace.h"
#include "error.h"
#include "list.h"
#include "pool.h"
#include "tid.h"
#include "io.h"
#include "lock.h"
#include "object.h"
#include "msgdefs.h"
#include "thread.h"
#include "semaphore.h"
#include "latch.h"
#include "lsn.h"
#include "bf.h"
#include "bf_macro.h"
#include "bf_intfuncs.h"
#include "bf_extfuncs.h"
#include "bf_globals.h"


/*
 *============================================================================
 *
 * BF_FindFreeSlot:
 * 	Routine to initialize the fixed size buffers
 *
 * RETURNS:
 *	esmNOERROR
 *
 * ERROR RETURNS:
 *	none
 *
 * SIDE EFFECTS:
 *	All the fixed size buffers are initialized and put on the free list
 *
 * IMPORTS:
 *
 *
 * BUGS:
 *
 *============================================================================
 */


 BUFINFO
*BF_FindFreeSlot(

	int					page2size 
)
{

	register BUFINFO	*start;
	register BUFINFO	*end;
	register BUFINFO	*current;
	register BUFINFO	*limit;
	register int		length;
	register BUFINFO	*first;
	PAGEHASH			*pageHash;
	int					foundSlots;


	TRPRINT(TR_BF, TR_LEVEL_1, ("looking for blockcount:%d", BLOCKCOUNT(page2size)));

	/*
	 *	get the actual number of blocks that are needed
	 */
	length = BLOCKCOUNT(page2size);

	/*
	 *	Set the starting location to be a multiple of the maximum page
	 *	size so that the buddy system allocation method is maintained.
	 *	This assures that we start at a place that has a header.
	 */
	start = first =
	&BufInfoTable[((CacheClock >> BLOCK2COUNT(page2size)) << BLOCK2COUNT(page2size))];
	end = &BufInfoTable[NumBuffers];
	TRPRINT(TR_BF, TR_LEVEL_2, ("start:%d end:%d", start->index, NumBuffers - 1));

	/*
	 *	check to see if we are past the end of the buffer table
	 *	which can happen if we don't have a power of 2 for
	 *	the number of buffers in the system;
	 */
	if (start >= end)	{

		/*
		 *	make start and end at the beginning of the table
		 */
		start = first = BufInfoTable;
		TRPRINT(TR_BF, TR_LEVEL_2, ("setting pointer to slot 0"));
	}

	/*
	 *	look through the entire cache looking for free slots
	 */
	for (;;)	{

		/*
		 *	set the start and end of the search
		 */
		current = start;
		limit   = start + length;
		TRPRINT(TR_BF, TR_LEVEL_2, ("current:%d limit:%d", current->index, limit->index));

		/*
		 *	look at each slot to see if it is free
		 */
		while ((current < limit) && (current < end))	{


			/*
			 *	get a pointer to the pageHash
			 */
			pageHash = current->pageHash;
			CHECK_PAGEHASH_MAGIC(pageHash);

			/*
			 *	check to see if the page is in a buffer group
			 *	or is currently held by a semaphore
			 */
			if (PAGE_IS_USED(pageHash) || SEMAPHORE_IS_HELD( &(pageHash->semaphore) ))	{

				/*
				 *	can't use this run
				 */
				TRPRINT(TR_BF, TR_LEVEL_2, ("page used:%d", current->index));
				break;
			}

			/*
			 *	jump to the slot occupied by the next chunk
			 */
			TRPRINT(TR_BF, TR_LEVEL_2, ("page2size:%d", pageHash->page2size));
			current += BLOCKCOUNT(pageHash->page2size);
			TRPRINT(TR_BF, TR_LEVEL_2, ("current:%d", current->index));
		}

		/*
		 *	check to see if we have a slot of the correct length
		 */
		if ((foundSlots = (current - start)) >= length)	{

			/*
			 *	set the cache clock and return to user
			 */
			CacheClock = current - BufInfoTable;
			TRPRINT(TR_BF, TR_LEVEL_2, ("found:%d slots at:%d CacheClock:%d",
					foundSlots, start->index, CacheClock));
			return(start);

		} else {

			/*
			 *	increment the pointer
			 */
			start = (current >= end) ? BufInfoTable : start + length;
			TRPRINT(TR_BF, TR_LEVEL_2, ("start:%d", start->index));

			/*
			 *	check to see if we have searched the entire cache
			 */
			if (start == first)	{

				/*
				 *	there are no runs of sufficient length
				 */
				SM_ERROR(TYPE_USER, esmNOBUFSPACE);
				return(NULL);
			}
		}
	}	
}
