/*
 *   $RCSfile: BF_GetSmaller.c,v $  
 *   $Revision: 1.1.1.1 $  
 *   $Date: 1996/05/04 21:55:33 $      
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/


#include "sysdefs.h"
#include "ess.h"
#include "checking.h"
#include "trace.h"
#include "error.h"
#include "list.h"
#include "pool.h"
#include "tid.h"
#include "io.h"
#include "lock.h"
#include "object.h"
#include "msgdefs.h"
#include "thread.h"
#include "semaphore.h"
#include "latch.h"
#include "lsn.h"
#include "bf.h"
#include "bf_macro.h"
#include "bf_intfuncs.h"
#include "bf_extfuncs.h"
#include "bf_globals.h"


/*============================================================================*/
/* 
 * BF_GetSmaller:
 * 	Routine to search the cache for fragments that can be coalesced
 *	into a larger buffer block
 *
 * RETURNS:
 *	a GROUPLINK pointer to buffer block if found
 *	NULL if no space is found
 *
 * SIDE EFFECTS:
 *	The page is entered into the page hash.
 *
 * SPECIAL NOTE:
 *	No check is made to see whether the page is already in the hash.
 * 	This check should be made by the calling routine.
 */
/*============================================================================*/


 GROUPLINK 
*BF_GetSmaller (

	unsigned			block2size 
)
{

	register int		length;
	register PAGEHASH	*pageHash;
	register BUFINFO	*current;
	register BUFINFO	*start;
	register BUFINFO	*end;
	register GROUPLINK	*groupLink;


	TRPRINT(TR_BF, TR_LEVEL_1, ("looking for slot length:%d", BLOCKCOUNT(block2size)));

	/*
	 *	attempt to find a slot that is free
	 */
	if ((start = BF_FindFreeSlot(block2size)) == NULL)	{

		return(NULL);
	}

	/*
	 *	get a pointer to the first slot in the group
	 */
	current = start;
	length = BLOCKCOUNT(block2size);
	end = current + length;

	/*
	 *	process the imbedded buffers in this block of free buffers
	 */
	while (current < end)	{

		TRPRINT(TR_BF, TR_LEVEL_2, ("current:%d", current->index));

		/*
		 *	get a pointer to the groupLink
		 */
		pageHash = (PAGEHASH *) current->pageHash;
		CHECK_PAGEHASH_MAGIC(pageHash);

		/*
		 *	check to see if the semaphore is held
		 */
		SM_ASSERT(LEVEL_3, SEMAPHORE_IS_FREE( &(pageHash->semaphore) ));

		/*
		 *	get the semaphore
		 */
		if (waitSemaphore( &(pageHash->semaphore) ))	{

			SM_ERROR(TYPE_FATAL, esmINTERNAL);
		}

		/*
		 *	get a pointer to the groupLink
		 */
		groupLink = (GROUPLINK *) pageHash->memberList.succ->item;
		CHECK_GROUPLINK_MAGIC(groupLink);
		CHECK_PAGEHASH_MAGIC(groupLink->pageHash);

		/*
		 *	take the link off the free list
		 */
		listRemove( &(groupLink->groupList) );	

		/*
		 *	check to see if this is the first sub chunk
		 */
		if (groupLink->pageHash != start->pageHash)	{

			TRPRINT(TR_BF, TR_LEVEL_2, ("this is not first link"));
			listEnq( &BufStructPool, &(groupLink->groupList) );
		}

		/*
		 *	Advance to the next buffer
		 */
		current += BLOCKCOUNT(pageHash->page2size);
	}

	/*
	 *	get a pointer to the first slot in the group
	 */
	current = start;
	length = BLOCKCOUNT(block2size);
	end = current + length;

	/*
	 *	process the imbedded buffers in this block of free buffers
	 */
	while (current < end)	{

		TRPRINT(TR_BF, TR_LEVEL_2, ("current:%d", current->index));

		/*
		 *	get a pointer to the groupLink
		 */
		pageHash = (PAGEHASH *) current->pageHash;
		CHECK_PAGEHASH_MAGIC(pageHash);

		/*
		 *	flush the page if necessary
		 */
		BF_InvalidatePage(pageHash, NOFLAGS|BF_FORCE);

		/*
		 *	Advance to the next buffer
		 */
		current += BLOCKCOUNT(pageHash->page2size);
	}

	/*
	 *	set the pageHash info
	 */
	pageHash = start->pageHash;
	CHECK_PAGEHASH_MAGIC(pageHash);
	pageHash->page2size = block2size;

	/*
	 *	set the pageinfo information
	 */
	pageHash->blockCount = BLOCKCOUNT(block2size);

	/*
	 *	get a pointer to the groupLink
	 */
	groupLink = ((GROUPLINK *) pageHash->memberList.succ->item);
	CHECK_GROUPLINK_MAGIC(groupLink);

	/*
	 *	set the status of the slot
	 */
	BF_SetSlot(start, pageHash->blockCount, pageHash);

	/*
	 *	return the saved link
	 */
	return(groupLink);
}
