/*
 *   $RCSfile: BF_Swap.c,v $  
 *   $Revision: 1.1.1.1 $  
 *   $Date: 1996/05/04 21:55:34 $      
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/


#include "sysdefs.h"
#include "ess.h"
#include "checking.h"
#include "trace.h"
#include "error.h"
#include "list.h"
#include "pool.h"
#include "tid.h"
#include "io.h"
#include "lock.h"
#include "object.h"
#include "msgdefs.h"
#include "thread.h"
#include "semaphore.h"
#include "latch.h"
#include "lsn.h"
#include "bf.h"
#include "bf_macro.h"
#include "bf_intfuncs.h"
#include "bf_extfuncs.h"
#include "util_funcs.h"


/*============================================================================*/
/* 
 * BF_Swap:
 * 	This routine swaps out unfixed buffer regions from a group's
 *	unfixed list until the group size is decreased enough to
 *	handle a new buffer region of the specified size.
 *
 * RETURNS:
 *	esmNOERROR
 *
 * ERROR RETURNS:
 *	ebfCANTSWAP      - if there isn't enough room in the group
 *
 * SIDE EFFECTS:
 *	Unfixed regions are moved to the free list until the group's
 *	size is decreased enough to handle the new region.
 *
 *
 */
/*============================================================================*/


 int
BF_Swap (

	register BUFGROUP	*bufGroup,
	int    				charge 
)
{

    register int		eventualFixedCharges;
    register int		maxCharges;	
	register GROUPLINK	*groupLink;


	TRPRINT(TR_BF, TR_LEVEL_1, ("requested charge:%d", charge));

    /*
     * Make sure the group is open.
     */
	if (CHECK_BUFGROUP_VALID(bufGroup))	{

		SM_ERROR(TYPE_FATAL, esmBADGROUP);
	}

	/*
	 *	get the charge in blocks
	 */
	charge = BLOCKCOUNT(charge);

	/*
	 *	get maximum charges into register
	 */
	maxCharges = bufGroup->maxCharges;

	/*
	 *	calculate the charges that will result from this request
	 */
    eventualFixedCharges = bufGroup->fixedCharges + charge;
	TRPRINT(TR_BF, TR_LEVEL_2, ("fixed:%d unfixed:%d max:%d eventual fixed:%d",
			bufGroup->fixedCharges, bufGroup->unfixedCharges,
			maxCharges, eventualFixedCharges));

    /*
     * See if the group can handle the new region without a swap.
     */
    if (eventualFixedCharges + bufGroup->unfixedCharges <= maxCharges) {

		/*
		 *	Page can be accomodated without replacement
		 */
		TRPRINT(TR_BF, TR_LEVEL_2, ("can accomodate charge w/o replace"));
        return(esmNOERROR);
    }

    /*
     *	Looks like we have to do a swap.
     *	Check to see if there any hope of getting enough space.
     */
    if (eventualFixedCharges > maxCharges) {

		SM_ERROR(TYPE_USER, esmCANTSWAP);
		return(esmFAILURE);
    }

    /*
     *	At this point, we have determined that there are enough unfixed
     *	regions to swap out; we loop until we generate enough space.
     */
    while (eventualFixedCharges + bufGroup->unfixedCharges > maxCharges) {

		/*
		 *	take the next element off the list
		 */
		groupLink = (GROUPLINK *) listDeq( &(bufGroup->unfixedList) );

		BF_FreePage(bufGroup, groupLink);
    }
	BF_AuditGroup(bufGroup);

	/*
	 *	We can now get a buffer for the group
	 */
    return(esmNOERROR);
}

/*
 * The next function, BF_SwapAll, doesn't pay any attention to
 * charges; it just releases ALL unfixed buffers.
 */

 int
BF_SwapAll (
	register BUFGROUP	*bufGroup
)
{
	GROUPLINK	*groupLink;
	int 		total=0;

    /*
     * Make sure the group is open.
     */
	if (CHECK_BUFGROUP_VALID(bufGroup))	{

		SM_ERROR(TYPE_FATAL, esmBADGROUP);
	}

    while((groupLink = (GROUPLINK *) listDeq( &(bufGroup->unfixedList) )) != NULL) {
		total++;
		BF_FreePage(bufGroup, groupLink);
    }
	BF_AuditGroup(bufGroup);

    return total;
}

#ifdef DEBUG

#include "bf_globals.h"

int 
countFreeBufList(
	int i
)
{
	int			total=0;
	GROUPLINK 	*groupLink;

    groupLink = (GROUPLINK *) FIRST_LIST_ELEMENT( &FreeBufTable[i] );

    while(groupLink != NULL) {
		total++;
		groupLink = (GROUPLINK *) NEXT_LIST_ELEMENT( &(groupLink->groupList) );
    }
	return total;
}

void
printFreeBufTable()
{
	register int i, j;
	int	maxBuf2Size, maxPage2Size, mumBufferSizes;

	maxBuf2Size = findLog2(NumBuffers);
	maxPage2Size = maxBuf2Size + MIN_PAGE2SIZE;
	mumBufferSizes = OFFSET(maxPage2Size) + 1;

	for (i = 0; i < mumBufferSizes; i++)	{
		j = countFreeBufList(i);
		if(j>0) {
			fprintf(stderr, 
			"FreeBufTable[%5d] contains %4d buffers.\n", 
				RAISETWO(i+MIN_PAGE2SIZE), j);
		}
	}

}
#endif DEBUG
