/*
 *   $RCSfile: bf_ForcePage.c,v $  
 *   $Revision: 1.1.1.1 $  
 *   $Date: 1996/05/04 21:55:34 $      
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/


#include "sysdefs.h"
#include "ess.h"
#include "checking.h"
#include "list.h"
#include "pool.h"
#include "tid.h"
#include "io.h"
#include "lock.h"
#include "object.h"
#include "msgdefs.h"
#include "thread.h"
#include "semaphore.h"
#include "latch.h"
#include "lsn.h"
#include "bf.h"
#include "bf_macro.h"
#include "bf_intfuncs.h"
#include "bf_extfuncs.h"
#include "link.h"
#include "volume.h"
#include "io_extfuncs.h"
#include "trace.h"
#include "error.h"
#include "trans.h"
#include "openlog.h"
#include "logrecs.h"
#include "log_extfuncs.h"
#include "thread_globals.h"
#include "log_globals.h"

#include "file.h"

 void
bf_ForcePage (

	register PAGEHASH	*pageHash,
	FLAGS				flags 
)
{

	SHORTPID		page;
#ifdef DEBUG
	SLOTTEDPAGE 	*slottedPage;
#endif /* DEBUG */

	TRPRINT(TR_BF, TR_LEVEL_1, ("pid:%d index:%d size:%d", pageHash->hashList.hashPid.page, pageHash->bufIndex, BLOCKCOUNT(pageHash->page2size)));

	/*
	 *	check the pagehash magic
	 */
	CHECK_PAGEHASH_MAGIC(pageHash);

	/*
	 *	get the semaphore
	 */
	if (flags & BF_SEM)	{

		if (waitSemaphore( &(pageHash->semaphore) ))	{

			/*
			 *	check to see if the page disappeared
			 */
			if (Active->errno == esmPAGEGONE)	{

				SM_ERROR(TYPE_LOG, esmPAGEGONE);
				return;
			} else if (Active->errno != esmNOERROR) {
				SM_ERROR(TYPE_FATAL, Active->errno);
			}
		}
	}

	/*
	 *	check to see if the page is dirty
	 */
	if (PAGE_IS_DIRTY(pageHash) || (flags & BF_FORCE))	{

		/*
		 *	The page cannot have any pending operations 
		 */
		SM_ASSERT(LEVEL_3, pageHash->pendingOpCount == 0);

		/*
		 *	check to see if the force mark is non-zero
		 */
		if (pageHash->forceMark != 0)	{

			/*
			 *	check to see if this is a log page
			 */
			if (PAGE_IS_LOG(pageHash))	{

				/*
				 *	check to make sure WAL is correct
				 */
				SM_ASSERT(LEVEL_3, OpenLog.forceMark <= pageHash->forceMark);

				/*
				 *	assert that the lsn and lrc are null
				 */
				SM_ASSERT(LEVEL_3, pageHash->lsn.wrapCount == 0);
				SM_ASSERT(LEVEL_3, pageHash->lsn.offset == 0);
				SM_ASSERT(LEVEL_3, pageHash->lrc.wrapCount == 0);
				SM_ASSERT(LEVEL_3, pageHash->lrc.count == 0);
#ifdef DEBUG
				page = ((LOGPAGEHDR*)(pageHash->bufFrame))->pageNumber;
				SM_ASSERT(LEVEL_3, LOG_PAGE_TO_BLOCK(page, &OpenLog) == pageHash->hashList.hashPid.page);
				page = ((LOGPAGEHDR*)(pageHash->bufFrame+OpenLog.lastUsableByte))->pageNumber;
				SM_ASSERT(LEVEL_3, LOG_PAGE_TO_BLOCK(page, &OpenLog) == pageHash->hashList.hashPid.page);
#endif DEBUG

				/*
				 *	force it to disk
				 */
				if (io_WritePage( &(pageHash->hashList.hashPid),
								  pageHash->bufFrame, pageHash->page2size))	{

					SM_ERROR(TYPE_FATAL, Active->errno);
				}

				/*
				 *	clean the page
				 */
				CLEAN_PAGE(pageHash);

				/*
				 *	update the log force indicator
				 */
				OpenLog.forceMark = pageHash->forceMark;
				TRPRINT(TR_BF, TR_LEVEL_2, ("forcing log page:%d", pageHash->forceMark));

				/*
				 *	reset the force and first marks
				 */
				pageHash->firstMark = pageHash->forceMark = 0;

			} else {

#ifdef DEBUG
				/* 
				 *	The btree index code requires that fixed pages
				 *	not be forced to disk.  But, volume header pages
				 *	are always fixed, so they are an exception.
				 */
				SM_ASSERT(LEVEL_3, !bf_CheckFixed(pageHash) ||
						  pageHash->hashList.hashPid.page == HEADERADDR);

				/*
				 *	If its a slotted page, see if no objects are on
				 *	it.
				 */
				slottedPage = (SLOTTEDPAGE*) pageHash->bufFrame;
				if (slottedPage->header.magic == SLOT_PAGE_MAGIC) {
					if (slottedPage->header.slotCount == 0) {
						PRINT_PROGRESS(("writing empty slot page %d\n", pageHash->hashList.hashPid.page));
					}
				}
#endif /* DEBUG */

				/*
				 *	force the log to disk
				 */
				forceLog(pageHash->forceMark);

				/*
				 *	force it to disk
				 */
				if (io_WritePage( &(pageHash->hashList.hashPid),
								  pageHash->bufFrame, pageHash->page2size))	{

					SM_ERROR(TYPE_FATAL, Active->errno);
				}

				/*
				 *	clean the page
				 */
				bf_CleanPage(pageHash);

				/*
				 *	set the force mark, lsn, and lrc to zero
				 */
				pageHash->forceMark = pageHash->firstMark = 0;
				pageHash->lsn.wrapCount = pageHash->lsn.offset = 0;
				pageHash->lrc.wrapCount = pageHash->lrc.count = 0;
			}

		} else {

			/*
			 *	assert that this cannot be a log page
			 */
			SM_ASSERT(LEVEL_2, (!(PAGE_IS_LOG(pageHash))) && (pageHash->firstMark == 0));

			/*
			 *	force it to disk and mark it clean
			 */
			if (io_WritePage( &(pageHash->hashList.hashPid),
							  pageHash->bufFrame, pageHash->page2size))	{

				SM_ERROR(TYPE_FATAL, Active->errno);
			}
			bf_CleanPage(pageHash);

			/*
			 *	set the force mark, lsn, and lrc to zero
			 */
			pageHash->forceMark = pageHash->firstMark = 0;
			pageHash->lsn.wrapCount = pageHash->lsn.offset = 0;
			pageHash->lrc.wrapCount = pageHash->lrc.count = 0;
		}

	}

	/*
	 *	signal the semaphore
	 */
	if (flags & BF_SEM)	{

		signalSemaphore( &(pageHash->semaphore) );
	}
}
