/*
 *   $RCSfile: bf_ForcePageList.c,v $  
 *   $Revision: 1.1.1.1 $  
 *   $Date: 1996/05/04 21:55:34 $      
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/


#include "sysdefs.h"
#include "ess.h"
#include "checking.h"
#include "list.h"
#include "pool.h"
#include "tid.h"
#include "io.h"
#include "lock.h"
#include "object.h"
#include "msgdefs.h"
#include "thread.h"
#include "semaphore.h"
#include "latch.h"
#include "lsn.h"
#include "bf.h"
#include "bf_macro.h"
#include "bf_intfuncs.h"
#include "bf_extfuncs.h"
#include "link.h"
#include "volume.h"
#include "io_extfuncs.h"
#include "trace.h"
#include "error.h"
#include "trans.h"
#include "openlog.h"
#include "logrecs.h"
#include "log_extfuncs.h"
#include "thread_globals.h"
#include "log_globals.h"

#include "file.h"


/*
 *	This function forces out a list of sequential pages.
 *	The pages must be all be dirty.  If any page is a log
 *	page, they must all be log pages.
 *
 *	If the BF_SEM flag is given, semaphores will be obtained for
 *	all pages.  To make this atomic, no page can already have its
 *	semaphore held.  
 */ 

 void
bf_ForcePageList (

	PAGEHASH*	*pageHashList,
	int			pageHashListLen,
	FLAGS		flags 
)
{
	SHORTPID	page;
	PID*		pidList[MAX_IO_LIST];
	char*		bufFrameList[MAX_IO_LIST];
	PAGE2SIZE	page2sizeList[MAX_IO_LIST];
	int 		i;
	BOOL		logPages;
	FORCEMARK	forceMark;

	TRPRINT(TR_BF, TR_LEVEL_1, ("%d pages starting at pid:%d size:%d", pageHashListLen, pageHashList[0]->hashList.hashPid.page, pageHashList[0]->bufIndex));

	if (PAGE_IS_LOG(pageHashList[0])) logPages = TRUE;
	else logPages = FALSE;

	/*
	 *	get the semaphores on all pages.  None must be held already
	 *	make sure pages are ok
	 */
	for (i = 0; i < pageHashListLen ; i++) {

		/* check the pagehash magic */
		CHECK_PAGEHASH_MAGIC(pageHashList[i]);

		if (i >= MAX_IO_LIST) {
			SM_ERROR(TYPE_FATAL, esmINTERNAL);
		}

		/* make sure type of page agrees with logPages */
		if ( PAGE_IS_LOG(pageHashList[i])) {
			
			SM_ASSERT(LEVEL_3, logPages);

			/* check to make sure WAL is correct */
			SM_ASSERT(LEVEL_3, OpenLog.forceMark <= pageHashList[i]->forceMark);

			/* assert that the lsn and lrc are null */
			SM_ASSERT(LEVEL_3, pageHashList[i]->lsn.wrapCount == 0);
			SM_ASSERT(LEVEL_3, pageHashList[i]->lsn.offset == 0);
			SM_ASSERT(LEVEL_3, pageHashList[i]->lrc.wrapCount == 0);
			SM_ASSERT(LEVEL_3, pageHashList[i]->lrc.count == 0);
#ifdef DEBUG
			page = ((LOGPAGEHDR*)(pageHashList[i]->bufFrame))->pageNumber;
			SM_ASSERT(LEVEL_3, LOG_PAGE_TO_BLOCK(page, &OpenLog) == pageHashList[i]->hashList.hashPid.page);
			page = ((LOGPAGEHDR*)(pageHashList[i]->bufFrame+OpenLog.lastUsableByte))->pageNumber;
			SM_ASSERT(LEVEL_3, LOG_PAGE_TO_BLOCK(page, &OpenLog) == pageHashList[i]->hashList.hashPid.page);
#endif DEBUG

		}

		pidList[i] = &(pageHashList[i]->hashList.hashPid);
		bufFrameList[i] = pageHashList[i]->bufFrame;
		page2sizeList[i] = pageHashList[i]->page2size;

		if (flags & BF_SEM)	{

			if (pageHashList[i]->semaphore.holder == NULL) {
				if (waitSemaphore( &(pageHashList[i]->semaphore) ))	{
						SM_ERROR(TYPE_LOG, Active->errno);
						SM_ERROR(TYPE_FATAL, esmINTERNAL);
				}
			} else {

				/* the semaphore is already held */
				SM_ERROR(TYPE_FATAL, esmINTERNAL);
			}
		}	

		/* The page must be dirty and cannot have any pending operations */
		SM_ASSERT(LEVEL_3, PAGE_IS_DIRTY(pageHashList[i]) &&
						   (pageHashList[i]->pendingOpCount == 0) );

	}

	/* find out how far to force the log */
	if (!logPages) {
		forceMark = 0;
		for (i = 0; i < pageHashListLen ; i++) {
#ifdef DEBUG
			if ( pageHashList[i]->forceMark != 0 )	{
				/* 
				 *	The btree index code requires that fixed pages
				 *	not be forced to disk.  But, volume header pages
				 *	are always fixed, so they are an exception.
				 */
				SM_ASSERT(LEVEL_3, !bf_CheckFixed(pageHashList[i]) ||
						  pageHashList[i]->hashList.hashPid.page == HEADERADDR);
			}
#endif /* DEBUG */

			if (forceMark < pageHashList[i]->forceMark) {
				forceMark = pageHashList[i]->forceMark;
			}
		}
		if (forceMark) forceLog(forceMark);
	}


	/* force the list to disk */
	if (io_WritePageList(pidList, bufFrameList, page2sizeList, pageHashListLen))	{
		SM_ERROR(TYPE_FATAL, Active->errno);
	}

	for (i = 0; i < pageHashListLen ; i++) {

		/* check to see if this is a log page */
		if (PAGE_IS_LOG(pageHashList[i]))	{

			CLEAN_PAGE(pageHashList[i]);

			/* update the log force indicator */
			OpenLog.forceMark = pageHashList[i]->forceMark;
			TRPRINT(TR_BF, TR_LEVEL_2, ("forcing log page:%d", pageHashList[i]->forceMark));

			/* reset the force and first marks */
			pageHashList[i]->firstMark = pageHashList[i]->forceMark = 0;

		} else {

			bf_CleanPage(pageHashList[i]);

			/* set the force mark, lsn, and lrc to zero */
			pageHashList[i]->forceMark = pageHashList[i]->firstMark = 0;
			pageHashList[i]->lsn.wrapCount = pageHashList[i]->lsn.offset = 0;
			pageHashList[i]->lrc.wrapCount = pageHashList[i]->lrc.count = 0;
		}

		/* signal the semaphores */
		if (flags & BF_SEM)	{
				signalSemaphore( &(pageHashList[i]->semaphore) );
		}
	}
}
