/*
 *   $RCSfile: bf_GetPageBuf.c,v $  
 *   $Revision: 1.1.1.1 $  
 *   $Date: 1996/05/04 21:55:33 $      
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/


#include "sysdefs.h"
#include "ess.h"
#include "checking.h"
#include "trace.h"
#include "error.h"
#include "list.h"
#include "pool.h"
#include "tid.h"
#include "io.h"
#include "lock.h"
#include "object.h"
#include "msgdefs.h"
#include "thread.h"
#include "semaphore.h"
#include "latch.h"
#include "lsn.h"
#include "bf.h"
#include "bf_macro.h"
#include "bf_intfuncs.h"
#include "bf_extfuncs.h"
#include "bf_globals.h"


/*============================================================================*/
/* 
 * bf_GetPageBuf:
 * 	Routine to insert a page into the page hash.
 *
 * RETURNS:
 *	esmNOERROR if all went ok and address of bucket via 
 *	hashPtr parameter.
 *
 * SIDE EFFECTS:
 *	The page is entered into the page hash.
 *
 * SPECIAL NOTE:
 *	No check is made to see whether the page is already in the hash.
 * 	This check should be made by the calling routine.
 */
/*============================================================================*/


 GROUPLINK 
*bf_GetPageBuf (

	BUFGROUP			*bufGroup,
	PAGE2SIZE			page2size 
)
{

	register GROUPLINK	*groupLink;


	TRPRINT(TR_BF, TR_LEVEL_1, ("need page size:%d", BLOCKCOUNT(page2size)));

	/*
	 *	Attempt to get a buffer off of the free list
	 *	of the requested size
	 */
TryAgain:
	if (groupLink = (GROUPLINK *) listDeq( &(FreeBufTable[OFFSET(page2size)]) ))	{

		TRPRINT(TR_BF, TR_LEVEL_2, ("got right size buffer"));

		/*
		 *	check the magic number
		 */
		CHECK_GROUPLINK_MAGIC(groupLink);
		CHECK_PAGEHASH_MAGIC(groupLink->pageHash);
		TRPRINT(TR_BF, TR_LEVEL_2, ("index:%d size:%d",
				groupLink->pageHash->bufIndex, groupLink->pageHash->blockCount));

		/*
		 *	check to see that the page is the right size
		 */
		SM_ASSERT(LEVEL_3, groupLink->pageHash->page2size == page2size);

		/*
		 *	Invalidate the page
		 */
		BF_InvalidatePage(groupLink->pageHash, BF_SEM|BF_FORCE);

	} else if (groupLink = BF_GetLarger(page2size))	{

		/*
		 *	check the magic number
		 */
		CHECK_GROUPLINK_MAGIC(groupLink);
		CHECK_PAGEHASH_MAGIC(groupLink->pageHash);

	} else if (groupLink = BF_GetSmaller(page2size))	{

		/*
		 *	check the magic number
		 */
		CHECK_GROUPLINK_MAGIC(groupLink);
		CHECK_PAGEHASH_MAGIC(groupLink->pageHash);

	} else {
		/*
		 *	there is no contigous buffer space
		 *  There might actually be more space unfixed in the
		 *  buffer group, so now try to free everything.
		 */
		if (BF_SwapAll(bufGroup)  > 0) 
			goto TryAgain;

		/*
		 *	there is no contigous buffer space
		 */
		SM_ERROR(TYPE_WARNING, esmNOBUFSPACE);
		return(NULL);
	}

	SM_ASSERT(LEVEL_3, groupLink->pageHash->page2size == page2size);

	/*
	 *	check to make sure the semaphore is free
	 */
	SM_ASSERT(LEVEL_3, SEMAPHORE_IS_FREE( &(groupLink->pageHash->semaphore) ));

	/*
	 *	grab the semaphore
	 */
	waitSemaphore( &(groupLink->pageHash->semaphore) );

	/*
	 *	Hang the buffer off the fixed list
	 */
	listEnq( &(bufGroup->fixedList), &(groupLink->groupList) );

	/*
	 *	Set the group pointer in the link
	 */
	groupLink->bufGroup = bufGroup;

	/*
	 *	Set the Group count, and pending operation count
	 */
	groupLink->pageHash->memberCount = 1;
	groupLink->pageHash->pendingOpCount = 0;

	/*
	 *	Set the fixed count
	 */
	groupLink->fixedCount = 1;

	/*
	 *	Increment the fixed charges
	 */
	bufGroup->fixedCharges += BLOCKCOUNT(page2size);

	/*
	 *	set the first and force marks to zero
	 */
	groupLink->pageHash->firstMark = groupLink->pageHash->forceMark = 0;
	groupLink->pageHash->lsn.wrapCount = groupLink->pageHash->lsn.offset = 0;
	groupLink->pageHash->lrc.wrapCount = groupLink->pageHash->lrc.count = 0;

	/*
	 * Set the return value and return.
	 */
	return(groupLink);
}
