/*
 *   $RCSfile: bf_Group.c,v $  
 *   $Revision: 1.1.1.1 $  
 *   $Date: 1996/05/04 21:55:33 $      
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/

#include "sysdefs.h"
#include "ess.h"
#include "checking.h"
#include "trace.h"
#include "error.h"
#include "list.h"
#include "pool.h"
#include "tid.h"
#include "io.h"
#include "lock.h"
#include "object.h"
#include "msgdefs.h"
#include "thread.h"
#include "semaphore.h"
#include "latch.h"
#include "lsn.h"
#include "bf.h"
#include "bf_macro.h"
#include "bf_intfuncs.h"
#include "bf_extfuncs.h"
#include "bf_globals.h"
#include "thread_globals.h"
#include "threadstate.h"
#include "openlog.h"


 BUFGROUP
*bf_AllocBufGroup (

	register POOL		*head 

)
{

	register BUFGROUP	*bufGroup;


	TRACE(TR_BF, TR_LEVEL_1);

    /*
	 *	Allocate the space
     */
	if ((bufGroup = (BUFGROUP *) malloc(sizeof(BUFGROUP))) == NULL)	{

		SM_ERROR(TYPE_SYS, esmMALLOCFAILED);
		return(NULL);
	}

	/*
	 *	Initialize the list elements
	 */
	initializeList( &(bufGroup->fixedList) );
	initializeList( &(bufGroup->unfixedList) );
	initializeListElement( &(bufGroup->groupList), (char *) bufGroup );
	bufGroup->flags = FREE_GROUP;
	INIT_BUFGROUP_MAGIC(bufGroup);

	/*
	 *	increment the number of groups in the pool
	 */
	head->elements++;

	/*
	 *	return the bufgroup pointer
	 */
	return(bufGroup);
}

/*============================================================================*/
/* 
 * bf_CloseGroup:
 * 	This routine closes a group and returns it to the list of free groups.
 *
 * RETURNS:
 *	esmNOERROR
 *
 * ERROR RETURNS:
 *	esmFIXEDPAGES  - if there are still fixed pages in the group.
 *
 * SIDE EFFECTS:
 *	The BufGroup structure is put on the list of free groups.
 *	The buffer regions in the group's unfixed list are freed.
 *	If the region being freed is a slotted page, then the page's
 *	list of user descriptors is freed.
 */
/*============================================================================*/


 int
bf_CloseGroup (

	BUFGROUP	*bufGroup 
)
{

	register GROUPLINK	*groupLink;

	
	TRACE(TR_BF, TR_LEVEL_1);

	/*
	 *	check the magic number
	 */
	CHECK_BUFGROUP_MAGIC(bufGroup);

    /*
     *	Make sure there are no fixed regions in the group.
     */
    if (bufGroup->fixedCharges > 0) {

		TRPRINT(TR_BF, TR_LEVEL_2, ("fixedCharges:%d", bufGroup->fixedCharges));
		bf_PrintFixed(bufGroup);
		SM_ERROR(TYPE_USER, esmFIXEDPAGES);
		return(esmFAILURE);
    }

    /*
     *	Free up the regions on the unfixed list.
     */
	while (LIST_NOT_EMPTY( &(bufGroup->unfixedList) ))	{

		/*
		 *	remove a buffer from the list
		 */
		groupLink = (GROUPLINK *) listDeq( &(bufGroup->unfixedList) );
		CHECK_GROUPLINK_MAGIC(groupLink);

		/*
		 *	Free a buffer
		 */
		BF_FreePage(bufGroup, groupLink);
	}

	/*
	 *	add to the number of buffers available
	 */
	ReservedBufs -= bufGroup->maxCharges;

	/*
	 *	free the structure
	 */
	bufGroup->flags = FREE_GROUP;
	poolMove( &BufGroupPool, &(bufGroup->groupList) );

	/*
	 *	if there are waiters let them all go and scramble
	 *	for the newly released resources
	 */
	if (LIST_NOT_EMPTY( &BufGroupWaitList ))	{

		notify( &BufGroupWaitList, esmNOERROR, esmGROUPGONE);
	}

	/*
	 *	Return success
	 */
    return(esmNOERROR);
}

/*============================================================================*/
/* 
 *	bf_OpenGroup:
 *		This allocates and initializes a buffer group.
 *
 *	RETURNS:
 *		a pointer to the group allocated.
 *
 *	ERROR RETURNS:
 *		esmNOFREEGROUPS    - out of free buffer groups.
 *		esmNOUNRESERVED    - not enough unreserved buffer frames.
 *
 *	SIDE EFFECTS:
 *		A buffer group is pulled off the list of free groups
 *		and then initialized.
 *
 *	IMPORTS: none
 *
 *	BUGS:
 *
 *
 * The use of all the globals below should be clear except ReservedBufs.
 * ReservedBufs reflects the number of buffers that have been 
 * reserved by bf_OpenGroup() calls.
 * When a group is opened, a specified number of buffers is reserved 
 * for that group, so that when buffers are requested for the group,
 * there should be enough free buffers to satisfy the request.
 * If ReservedBufs is smaller than the number requested in an
 * bf_OpenGroup call, then bf_OpenGroup() returns an error.
 */
/*============================================================================*/


 BUFGROUP
*bf_OpenGroup (

	int					maxCharges,
	int					replacePolicy,
	FLAGS				flags 
)
{

    register BUFGROUP    *bufGroup;


	TRPRINT(TR_BF, TR_LEVEL_1, ("maxCharges:%d replacePolicy:%d\n",
			maxCharges, replacePolicy));

	/*
	 *	check to see that the replacement policy is sane
	 */
	if ((replacePolicy != BF_LRU) && (replacePolicy != BF_MRU))	{

		SM_ERROR(TYPE_USER, esmBADREPLACEPOLICY);
		return(NULL);
	}

	/*
	 *	check to see that the request can be satisfied
	 */
	if (maxCharges > NumBuffers)	{

		SM_ERROR(TYPE_USER, esmREQUESTTOOBIG);
		return(NULL);
	}

again:

    /*
	 *	Check to see if there are sufficient buffers to fill the request
     */
    if (maxCharges + ReservedBufs > NumBuffers) {

		/*
		 *	Check to see if the caller desires to wait
		 */
		if (flags & BF_WAIT)	{

			/*
			 *	place ourselves on the wait list
			 */
			if (waitList(&BufGroupWaitList, THREAD_BUFGROUP_WAIT))	{

				/*
				 *	some problem has occured
				 */
				return(NULL);
			
			} else {
				
				/*
				 *	try to get the desired amount
				 */
				goto again;
			}

		} else {
				
			SM_ERROR(TYPE_WARNING, esmNOUNRESERVED);
			return(NULL);
		}
    }

    /*
     *	Attempt to allocate the buffer group structure
     */
    if ((bufGroup = (BUFGROUP *) poolDeq( &BufGroupPool )) == NULL)	{

		SM_ERROR(TYPE_WARNING, esmNOFREEGROUPS);
		return(NULL);
	}

    /*
     * Initialize all the fields in the allocated entry.
     */
    bufGroup->flags			 = OPEN_GROUP;
    bufGroup->replacePolicy	 = replacePolicy;
    bufGroup->maxCharges	 = maxCharges;
    bufGroup->fixedCharges	 = 0;
    bufGroup->unfixedCharges = 0;

	/*
	 *	put the group on the used list
	 */
	listEnq( &UsedBufGroups, &(bufGroup->groupList) );

	/*
	 *	increment the buffers in use
	 */
    ReservedBufs += maxCharges;

	/*
	 *	return a pointer to the allocated group
	 */
    return(bufGroup);
}

#ifdef DEBUG
 void
BF_AuditGroup (

	register BUFGROUP	*bufGroup 
)
{

	register GROUPLINK	*groupLink;
	register int		charges;


	TRPRINT(TR_BF, TR_LEVEL_1, ("bufGroup:%x", bufGroup));

	CHECK_BUFGROUP_MAGIC(bufGroup);

    if ((bufGroup->flags & OPEN_GROUP) == 0) {

		SM_ERROR(TYPE_FATAL, esmINTERNAL);
    }

	/*
	 *	setup initial loop conditions
	 */
	groupLink = (GROUPLINK *) FIRST_LIST_ELEMENT( &(bufGroup->unfixedList) );
	charges = 0;

    /*
	 *	Calculate the charges on the fixed list
     */
    while (groupLink != NULL)	{

		CHECK_GROUPLINK_MAGIC(groupLink);

		charges += BLOCKCOUNT(groupLink->pageHash->page2size);

		groupLink = (GROUPLINK *) NEXT_LIST_ELEMENT( &(groupLink->groupList) );
    }

	TRPRINT(TR_BF, TR_LEVEL_2, ("unfixed count:%d indicated:%d",
			charges, bufGroup->unfixedCharges));

	if (charges != bufGroup->unfixedCharges)	{

		TRPRINT(TR_BF, TR_LEVEL_2, ("mismatch in count:%d indicated:%d",
				charges, bufGroup->unfixedCharges));
		SM_ERROR(TYPE_FATAL, esmINTERNAL);
	}

	groupLink = (GROUPLINK *) FIRST_LIST_ELEMENT( &(bufGroup->fixedList) );
	charges = 0;

    while (groupLink != NULL)	{

		/*
		 *	check the groupLink magic
		 */
		CHECK_GROUPLINK_MAGIC(groupLink);

		/*
		 *	Account for the page
		 */
		charges += BLOCKCOUNT(groupLink->pageHash->page2size);

		/*
		 *	Get a pointer to the next element on the list
		 */
		groupLink = (GROUPLINK *) NEXT_LIST_ELEMENT( &(groupLink->groupList) );
    }

	TRPRINT(TR_BF, TR_LEVEL_2, ("fixed count:%d indicated:%d",
			charges, bufGroup->fixedCharges));

	if (charges != bufGroup->fixedCharges)	{

		TRPRINT(TR_BF, TR_LEVEL_2, ("mismatch in count:%d indicated:%d",
				charges, bufGroup->fixedCharges));
		SM_ERROR(TYPE_FATAL, esmINTERNAL);
	}

#ifndef JUNK
	/* TODO : remove */
	{
		extern OPENLOG		OpenLog;
		GROUPLINK 			*gl, *gln;


		if(bufGroup == OpenLog.writeGroup) {
			gl = (GROUPLINK *)
				FIRST_LIST_ELEMENT( &(OpenLog.writeGroup->unfixedList) );
			if(gl==NULL) return;
			if(OpenLog.tailLink == NULL) return;

			if( gl->pageHash != OpenLog.tailLink->pageHash ) {
				return;
			}
			/* gl is the tail of the log */

			gln = (GROUPLINK *) NEXT_LIST_ELEMENT( &(gl->groupList) );
			if( gln == NULL) return;

			/* gl is the tail of the log but there's a page *AFTER*
			 * it in the unfixed list
			 * Print some info and barf:
			 */
				fprintf(stderr, "BF_AUDITGROUP: tail page %d\n", 
					OpenLog.tailLink->pageHash-> hashList.hashPid.page);

				while( gl != NULL ) {
					fprintf(stderr, "page %d\n",
						gl->pageHash-> hashList.hashPid.page);
					gl = (GROUPLINK *) NEXT_LIST_ELEMENT( &(gl->groupList) );
				}
				/* Reset gl for debugging purposes: */
				gl = (GROUPLINK *) FIRST_LIST_ELEMENT(
							&(OpenLog.writeGroup->unfixedList) );

			SM_ERROR(TYPE_FATAL, esmINTERNAL);
		}
	}
#endif JUNK

}
#endif DEBUG

 GROUPLINK
*BF_FindBufGroup (

	PAGEHASH			*pageHash,
	register BUFGROUP	*bufGroup 
)
{

	register GROUPLINK	*groupLink;


	TRACE(TR_BF, TR_LEVEL_1);

	/*
	 *	check the pagehash magic number
	 */
	CHECK_PAGEHASH_MAGIC(pageHash);

	/*
	 *	Check quick to see if the item is on the free list
	 */
	if (pageHash->memberCount == 0)	{

		/*
		 *	The page is on a free list
		 */
		TRPRINT(TR_BF, TR_LEVEL_2, ("page is on the free list%x", pageHash));
		Active->errno = BF_FREE;
		return((GROUPLINK *) FIRST_LIST_ELEMENT( &(pageHash->memberList) ));
	
	} else {

		/*
		 *	get a pointer to the first group link
		 */
		groupLink = (GROUPLINK *) FIRST_LIST_ELEMENT( &(pageHash->memberList) );

		/*
		 *	Traverse down the list looking for a match
		 */
		while (groupLink != NULL)	{

			/*
			 *	check magic numbers
			 */
			CHECK_GROUPLINK_MAGIC(groupLink);
			CHECK_BUFGROUP_MAGIC(groupLink->bufGroup);

			/*
			 *	Check for a match
			 */
			if (groupLink->bufGroup == bufGroup)	{

				/*
				 *	We have a match
				 */
				TRPRINT(TR_BF, TR_LEVEL_2, ("page is in the target group"));

				/*
				 *	Check to see if the page is fixed
				 */
				if (groupLink->fixedCount != 0)	{

					/*
					 *	The page is fixed
					 */
					TRPRINT(TR_BF, TR_LEVEL_2, ("page is fixed"));
					Active->errno = BF_FIXED;
					return(groupLink);
				
				} else {

					/*
					 *	The page is unfixed
					 */
					TRPRINT(TR_BF, TR_LEVEL_2, ("page is unfixed"));
					Active->errno = BF_UNFIXED;
					return(groupLink);
				}

			} else {	

				/*
				 *	Look at the next entry
				 */
				groupLink = (GROUPLINK *) NEXT_LIST_ELEMENT( &(groupLink->groupList) );
			}
		}

		/*
		 *	We did not find a match
		 */
		TRPRINT(TR_BF, TR_LEVEL_2, ("page is not in target group>"));
		Active->errno = BF_OTHER;
		return(NULL);
	}
}
