/*
 *   $RCSfile: bf_ReadPage.c,v $  
 *   $Revision: 1.1.1.1 $  
 *   $Date: 1996/05/04 21:55:33 $      
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/

#include "sysdefs.h"
#include "ess.h"
#include "checking.h"
#include "trace.h"
#include "error.h"
#include "list.h"
#include "pool.h"
#include "tid.h"
#include "io.h"
#include "lock.h"
#include "object.h"
#include "msgdefs.h"
#include "thread.h"
#include "semaphore.h"
#include "latch.h"
#include "lsn.h"
#include "link.h"
#include "bf.h"
#include "bf_macro.h"
#include "bf_intfuncs.h"
#include "bf_extfuncs.h"
#include "volume.h"
#include "io_extfuncs.h"
#include "bf_globals.h"
#include "thread_globals.h"


/*============================================================================*/
/* 
 * bf_ReadPage:
 * 	This routine will read a single page into a free buffer page unless
 *      the page is already in the buffer pool in which case it avoids an IO.
 *
 * RETURNS:
 *	esmNOERROR        - if all went well.
 *	
 * ERROR RETURNS:
 *	other errors returned by lower levels.
 *
 * SIDE EFFECTS:
 *	The requested page is read into the buffer pool if 
 *	necessary, added to the specified buffer group, and fixed
 *	The pid is entered in the page hash if it is not already there.
 *
 * 	READ-RULE:
 *		If a page is fixed, then all the pages that have a
 *		dependency on that page must be fixed, and similarly, if
 *		a page is unfixed, then all of that page's dependents
 *		must be unfixed.
 *		For hiearchical data structures this means that
 *		pages must be read and fixed  in root-to-leaf order,
 *		and must be unfixed in leaf-to-root order.
 *	
 *	The READ-RULE basically makes it impossible to run into a situation 
 *	where a fixed page needs to be flushed because of an IO dependency.
 *	It allows us to avoid ugly deadlock problems.
 *
 *	The code below has been set up to deal with the states mentioned 
 *	above, and only those states.
 *	If the the READ-RULE is violated, and a page enters a unexpected
 *	state then the code below will either panic or do something incorrect.
 *
 */
/*============================================================================*/


 GROUPLINK 
*bf_ReadPage (

	register BUFGROUP	*bufGroup,
	const	 PID     	*pid,
	PAGE2SIZE			page2size,
	FLAGS				flags 
)
{

	register PAGEHASH	*pageHash;
	register GROUPLINK	*groupLink;
	register PAGEHASH	*tempHash;
	GROUPLINK			*retLink;


	TRPRINT(TR_BF, TR_LEVEL_1, ("page:%d pagesize:%d", pid->page, RAISETWO(page2size)));

    /*
     * Check to see if the page is in the cache
     */
    if ((pageHash = BF_FindPage(bufGroup, pid, BF_SEM, &retLink)) == NULL)	{

		/*
		 *	check the errno
		 */
		if (Active->errno != esmNOERROR)	{

			return(NULL);
		}

    } else {

		/*
		 *	put the grouplink pointer in a register
		 */
		groupLink = retLink;
		SM_ASSERT(LEVEL_3, pageHash->page2size == page2size);
	}

    /*
     * Now switch on the page state and take the  
     * appropriate action.
     */
    switch (Active->errno) {

		/*
		 *	The page was not in the cache, it should be read in
		 */
		case (BF_NOT_PRESENT):

			TRPRINT(TR_BF, TR_LEVEL_1, ("page:%d not present"));
			/*
			 *	get a pageHash that will hold the pid while we
			 *	get a real buffer.  We could get blocked on an
			 *	I/O flush and we need to keep the pid locked.
			 *	In the future this should be changed to not
			 *	hold the semaphore while I/O is occurring.
			 */
			tempHash = BF_NewPageHash();
			tempHash->flags |= H_TEMP; /* has no bufFrame, etc */
			CHECK_PAGEHASH_MAGIC(tempHash);

			/*
			 *	initialize and take the semaphore
			 */
			initializeSemaphore( &(tempHash->semaphore), FALSE );
			waitSemaphore( &(tempHash->semaphore) );

			/*
			 *	initialize the force marks
			 */
			tempHash->firstMark = tempHash->forceMark = 0;
			tempHash->lsn.wrapCount = tempHash->lsn.offset = 0;
			tempHash->lrc.wrapCount = tempHash->lrc.count = 0;

			/*
			 *	put the pagehash in the table so it can be found
			 */
			BF_InsertHashEntry(tempHash, pid, NOFLAGS);

			/*
			 *	manage the buffer group space
			 */
			if (BF_Swap(bufGroup, page2size))	{

				/*
				 *	reclaim the dummy pagehash
				 */
				BF_FreeHashEntry(tempHash, esmPAGEGONE);
				return(NULL);
			}

			/*
			 *	attempt to get the buffer.  This will increment
			 *	the fixed charges for this group.  Any error
			 *	code will be set by the routine.  This routine will
			 *	return holding the page semaphore
			 */
			if ((groupLink = bf_GetPageBuf(bufGroup, page2size)) == NULL)	{

				/*
				 *	reclaim the dummy pagehash
				 */
				BF_FreeHashEntry(tempHash, esmPAGEGONE);
				return(NULL);
			}

			/*
			 *	take the waiters off the dummy pagehash if necessary
			 */
			if (SEMAPHORE_WAITERS( &(tempHash->semaphore) ))	{
				
				/*
				 *	Notify any waiters that a new page has is being
				 *	used.  bf_FindPage will detect this properly.
				 */
				notify( &(tempHash->semaphore.waitList), esmFAILURE, esmPAGEGONE);
			}

			/*
			 *	free the dummy entry
			 */
			initializeSemaphore(&(tempHash->semaphore), TRUE);
			BF_FreeHashEntry(tempHash, esmNOERROR);

			/*
			 *	attempt to insert the page into the hash table
			 */
			BF_InsertHashEntry(groupLink->pageHash, pid, NOFLAGS);

			/*
			 *	get a pointer to the pagehash entry
			 */
			pageHash = groupLink->pageHash;

			/*
			 *	check to see if the page needs to be read
			 */
			if (!(flags & BF_NOREAD))	{

				/*
				 * Read the page in from disk.
				 */
				if (io_ReadPage(pid, pageHash->bufFrame, page2size))	{

					TRPRINT(TR_BF, TR_LEVEL_2, ("got io error"));

					/*
					 *	notify waiters that may have blocked during io
					 */
					notify( &(pageHash->semaphore.waitList), esmFAILURE, Active->errno );

					/*
					 *	invalidate the page and free it
					 */
					listRemove( &(pageHash->hashList.hashList) );
					initializeLatch( &(pageHash->latch) );

					SM_ASSERT(LEVEL_3, pageHash->semaphore.holder == Active);
					initializeSemaphore( &(pageHash->semaphore), TRUE );

					/*
					 *	Put the page back on the free list
					 */
					bf_UnfixPage(groupLink, BF_DEFAULT, FALSE);

					/*
					 *	return the null pointer
					 */
					return(NULL);
				}
			}

			break;


		case (BF_FREE):
			TRPRINT(TR_BF, TR_LEVEL_1, ("page:%d is free"));

			/*
			 *	manage the buffer group space
			 */
			if (BF_Swap(bufGroup, page2size))	{

				signalSemaphore( &(pageHash->semaphore) );
				return(NULL);
			}

			/*
			 *	check the grouplink magic
			 */
			CHECK_GROUPLINK_MAGIC(groupLink);

			/*
			 *	Take it off the free list and
			 *	put the link onto the fixed list
			 */
			listMoveEnq( &(bufGroup->fixedList), &(groupLink->groupList) );
			groupLink->bufGroup = bufGroup;

			/*
			 *	Set the number of groups the page is in
			 *	and the fixed count for this group
			 */
			pageHash->memberCount = 1;
			groupLink->fixedCount = 1;

			/*
			 *	Increment the number of fixed charges for the buffer group
			 */
			bufGroup->fixedCharges += BLOCKCOUNT(page2size);

			break;

		
		case (BF_OTHER):

			TRPRINT(TR_BF, TR_LEVEL_1, ("page:%d is other"));

			/*
			 *	manage the buffer group space
			 */
			if (BF_Swap(bufGroup, page2size))	{

				signalSemaphore( &(pageHash->semaphore) );
				return(NULL);
			}

			/*
			 *	Get a free group link
			 */
			groupLink = (GROUPLINK *) poolDeq( &GroupLinkPool );

			/*
			 *	set the fields of the link
			 */
			groupLink->pageHash	  = pageHash;
			groupLink->bufGroup	  = bufGroup;
			groupLink->fixedCount = 1;
			groupLink->bufFrame	  = pageHash->bufFrame;

			/*
			 *	Put the structure on the linked lists
			 */
			listEnq( &(pageHash->memberList), &(groupLink->hashList) );
			listEnq( &(bufGroup->fixedList), &(groupLink->groupList) );

			/*
			 *	Increment the group count in the hash structure
			 */
			pageHash->memberCount++;

			/*
			 *	Increment the number of fixed charges for the buffer group
			 */
			bufGroup->fixedCharges += BLOCKCOUNT(page2size);

			break;


		case (BF_FIXED):

			TRPRINT(TR_BF, TR_LEVEL_1, ("page:%d is fixed"));
			/*
			 *	check the buffer group magic number
			 */
			CHECK_GROUPLINK_MAGIC(groupLink);

			/*
			 *	The page is already fixed in the group.
			 *	Simply increment the fix count
			 */
			groupLink->fixedCount++;

			break;

		case (BF_UNFIXED):

			TRPRINT(TR_BF, TR_LEVEL_1, ("page:%d is unfixed"));
			/*
			 *	Put it in the fixed partition
			 */
			listMoveEnq( &(bufGroup->fixedList), &(groupLink->groupList) );

			/*
			 *	Set the fixed count
			 */
			groupLink->fixedCount = 1;

			/*
			 *	Update the fixed and unfixed charges
			 */
			bufGroup->fixedCharges += BLOCKCOUNT(page2size);
			bufGroup->unfixedCharges -= BLOCKCOUNT(page2size);;

			break;
	}
	TRPRINT(TR_BF, TR_LEVEL_2, ("fixed:%d unfixed:%d",
			bufGroup->fixedCharges, bufGroup->unfixedCharges));

	/*
	 *	give back the semaphore
	 */
	if (!(flags & BF_SEM))	{

		signalSemaphore( &(pageHash->semaphore) );
	}

	/*
	 *	Return the group link pointer
	 */
	CHECK_GROUPLINK_MAGIC(groupLink);
	return(groupLink);
}
