/*
 *   $RCSfile: cl_DeallocPages.c,v $  
 *   $Revision: 1.1.1.1 $  
 *   $Date: 1996/05/04 21:55:37 $      
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/

#include "sysdefs.h"
#include "ess.h"
#include "checking.h"
#include "trace.h"
#include "error.h"
#include "list.h"
#include "pool.h"
#include "tid.h"
#include "io.h"
#include "lock.h"
#include "object.h"
#include "msgdefs.h"
#include "thread.h"
#include "semaphore.h"
#include "latch.h"
#include "link.h"
#include "bf.h"
#include "threadstate.h"
#include "volume.h"
#include "bitvec.h"
#include "lsn.h"
#include "trans.h"
#include "deallocinfo.h"
#include "file.h"
#include "consist.h"
#include "cl_funcs.h"
#include "msg_funcs.h"
#include "bf_extfuncs.h"
#include "fi_extfuncs.h"
#include "io_extfuncs.h"
#include "thread_funcs.h"
#include "trans_extfuncs.h"
#include "bf_globals.h"
#include "io_globals.h"
#include "thread_globals.h"


 void
cl_DeallocPages (

	register MESSAGE	*message,
	register LINK		*link 
)
{
	extern int			PrintProgress; /* log_globals.h */

	PID					pidList[MAX_PIDLIST_SIZE];
	TRANSREC			*transRec;
	VOLREC				*volRec;
	int					i;

	TRPRINT(TR_CL, TR_LEVEL_1, ("deallocing:%d pages", message->body.alloc.numPages));

	/*
     *  suspend the link so that it does not get selected on
     *  the basis of the data that is sent.  This must be done 
	 *	because registerTrans may block while the client is sending
	 *	the page list.	
     */
    suspendLink(link);

	/*
	 *	attempt to find the transaction record, TRUE => writing log
	 */
	if ((transRec = registerTrans(message->header.params.in.tid, TRUE)) == NULL) {

		SM_ERROR(TYPE_USER, esmBADTRANSID);
		reply(esmFAILURE, Active->errno, 1, TRUE);
	}

    /* resume the link so that the data can be read */
    resumeLink(link);

	/*
	 *	read the data into shared memory
	 */
	if (receive(link, (char *) pidList, (message->body.alloc.numPages * sizeof(PID))) < 0) {

		/*
		 *	nothing we can do here, link is closed
		 */
		threadRestart();
	}

	PRINT_PROGRESS( ("deallocing:%d, %d pages, size %d \n", pidList[0].page, message->body.alloc.numPages, message->body.alloc.page2size) )

	/*
	 *	See if the page should be deallocated immediately.
	 */
	if (message->body.alloc.flags == 1) {
		
		/*
		 *	The page should be deallocated immediately -- ie. it
		 *	can be reused by this transaction 
		 *	NOTE: This is not supported yet.
		 */
		/* SM_ERROR(TYPE_LOG, esmINTERNAL); */
	}

	if ( (volRec = io_FindVolumeIdWithLink(pidList->volid, link)) == NULL) {
		SM_ERROR(TYPE_FATAL, esmINTERNAL);
	}

	if (message->body.alloc.flags == 0 || message->body.alloc.flags == 1) {

#ifdef JUNK
		/*
		 *	The page deallocation should be delayed until end of
		 *	transaction.  NOTE: This is not supported yet.
		 */
		SM_ERROR(TYPE_FATAL, esmINTERNAL);
		if (io_DeallocPages(message->body.alloc.page2size,
							message->body.alloc.numPages, pidList, NULL,
							TRUE))	{
			reply(esmFAILURE, Active->errno, 1, TRUE);
		}
#endif

		/*
		 *	If the volume is temporary actually deallocate the page,
		 *	otherwise just record it for deallocation at commit
		 */
		if (volRec->volflags & VOL_TEMP) {

			/*
			 *  Only deallocate the slotted pages.  Other page types
			 *	are on the transactions tempPageList and will be
			 *	deallocated at EOT.
			 */
			if (message->body.alloc.pageType == PAGE_SLOTTED) {
				/*
				 *	actually deallocate each page
				 */
				if (io_DeallocPages(message->body.alloc.page2size,
									message->body.alloc.numPages,
									pidList, NULL, FALSE, TRUE)
									!= esmNOERROR) {
						reply(esmFAILURE, Active->errno, 1, TRUE);
				}
			}

			/* remove the pages from the buffer pool */
			for (i = 0; i < message->body.alloc.numPages; i++) {
				bf_InvalidatePage(pidList+i);
			}

		} else {

			if (io_FakeDeallocPages(message->body.alloc.page2size,
									message->body.alloc.numPages,
									pidList,
									&(transRec->pageDeallocList),
									message->body.alloc.pageType)
									!= esmNOERROR) {
					reply(esmFAILURE, Active->errno, 1, TRUE);
			}
		}
	} else {

		SM_ERROR(TYPE_USER, esmINTERNAL);
		reply(esmFAILURE, Active->errno, 1, TRUE);
	}


	/*
	 *	If the page is a slotted page, remove it from its file.
	 *	and mark it as no longer slotted.
	 */
	if (message->body.alloc.pageType == PAGE_SLOTTED) {

		/*
		 *	Make sure volume is not temporary since files on
		 *	temporary volumes are logically ordered.
		 */
		SM_ASSERT(LEVEL_1, !(volRec->header->properties & VOLPROP_TEMP));

		/* page is no longer slotted */
		if (io_MarkPage(volRec, pidList, PAGE_SLOTTED, FALSE) != esmNOERROR)   {
			reply(esmFAILURE, Active->errno, 1, TRUE);
		}

		/* remove the page from the file */
		if (fi_DeletePage(UserBufGroup, &(message->body.alloc.slotted.s_input.fid), pidList->page, NULL_LSN) != esmNOERROR) {
			reply(esmFAILURE, Active->errno, 1, TRUE);
		}
	}

#ifdef notdef
/* Don't do this since the pages has not been officially dealloced */
	/*
	 *	Remove the pages from the buffer
	 */
	{	register int	i;
		for (i = 0; i < message->body.alloc.numPages; i++) {
			bf_InvalidatePage(pidList+i);
		}
	}
#endif notdef
	reply(esmNOERROR, esmNOERROR, 1, TRUE);
}
