/*
 *   $RCSfile: cl_ReadPage.c,v $  
 *   $Revision: 1.1.1.1 $  
 *   $Date: 1996/05/04 21:55:37 $
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/


#include "sysdefs.h"
#include "ess.h"
#include "checking.h"
#include "trace.h"
#include "error.h"
#include "list.h"
#include "pool.h"
#include "tid.h"
#include "io.h"
#include "lock.h"
#include "object.h"
#include "msgdefs.h"
#include "thread.h"
#include "semaphore.h"
#include "latch.h"
#include "link.h"
#include "lsn.h"
#include "bf.h"
#include "volume.h"
#include "trans.h"
#include "bitvec.h"
#include "log.h"
#include "openlog.h"
#include "page.h"
#include "file.h"
#include "consist.h"
#include "sm_macro.h"
#include "bf_extfuncs.h"
#include "fi_extfuncs.h"
#include "io_extfuncs.h"
#include "sm_extfuncs.h"
#include "trans_extfuncs.h"
#include "msg_funcs.h"
#include "lm_extfuncs.h"
#include "cl_funcs.h"
#include "thread_funcs.h"
#include "bf_globals.h"
#include "cl_globals.h"
#include "thread_globals.h"
#include "log_globals.h"
#include "stat_globals.h"
#include "io_globals.h"


#define LRC_GEQ(lrcA, lrcB) \
	( ( (lrcA).wrapCount > (lrcB)->wrapCount )      || \
	  (	((lrcA).wrapCount == (lrcB)->wrapCount) &&  \
	    ((lrcA).count     >= (lrcB)->count) ) ) 


/*
 *	This is a helper function for cl_ReadPage.
 *	It looks up a page in the lrc cache and if not present
 *	it reads the page from disk.  It also returns the lrc
 *	for the page and the file it belongs to.
 */
 int
getLrcAndFid(
	PID 		*pid, 		/* IN page to get info on */
	PAGETYPE 	pageType, 	/* IN type of page to get info on */
	PAGE2SIZE	page2size,	/* IN type of page to get info on */
	BOOL		leavePinned,/* IN what to do with pinned pages */
	FID 		*fid,		/* OUT file containing this page */
	LRC 		*lrc,		/* OUT lrc for this page */
	GROUPLINK 	**retLink	/* OUT pinned page */
)
{
	LRCCACHEENTRY		*lrcce;
	GROUPLINK			*groupLink;
	SLOTTEDPAGE			*slottedPage;	
	LGNODE				*lgNode;
	LGDATAPAGE			*lgDataPage;

	lrcce = bf_FindCacheEntry(pid);
	itx_lrc_lookups++;

	TRPRINT(TR_CL, TR_LEVEL_1, 
		("look up page %d.%d, pid in lrcce is %d.%d",
		pid->page, pid->volid, 
		lrcce?lrcce->hashList.hashPid.page:0,
		lrcce?lrcce->hashList.hashPid.volid:0));

	if( (lrcce != NULL) &&
		(PPIDEQ( &(lrcce->hashList.hashPid), pid))){

		itx_lrc_finds++;
		*fid = lrcce->fid;
		*lrc = lrcce->lrc;
		*retLink = NULL;

		return esmNOERROR;
	} 

	/*
	 *	Page is not in the LRC cache, so pin it
	 */
	if ((groupLink = bf_ReadPage(UserBufGroup, pid, page2size, BF_SEM)) == NULL){
		return esmFAILURE;
	}

	switch (pageType)  {
	case PAGE_INDEX:
		*lrc = *((LRC *)(groupLink->bufFrame));
		break;

	case PAGE_INDEXDESC:
		*lrc = (((INDEXDESCPAGE*)(groupLink->bufFrame))->lrc);
		break;

	case PAGE_SLOTTED:
		slottedPage = (SLOTTEDPAGE *) groupLink->bufFrame;
		*lrc = slottedPage->header.lrc;
		if (CHECK_SLOTTED_MAGIC(slottedPage))	{
			SM_ERROR(TYPE_FATAL, esmINTERNAL);
		}
		*fid = slottedPage->header.fid;
		break;

	case PAGE_LARGENODE:
		lgNode = (LGNODE*) groupLink->bufFrame;
		*lrc = lgNode->header.lrc;
		CHECK_LARGENODE_MAGIC(lgNode);
		break;

	case PAGE_LARGEDATA:
		lgDataPage = (LGDATAPAGE*) groupLink->bufFrame;
		*lrc = lgDataPage->header.lrc;
		CHECK_LGDATAHDR_MAGIC(lgDataPage);
		break;

	default:
		SM_ERROR(TYPE_CRASH, esmINTERNAL);
		break;

	} /* switch on page type */

	SM_ASSERT( LEVEL_1, (groupLink->pageHash->pageType == pageType) ||
			(groupLink->pageHash->pageType == PAGE_GENERIC));
	groupLink->pageHash->pageType = pageType;

	if (leavePinned) {
		*retLink = groupLink;
	} else {
		signalSemaphore( &(groupLink->pageHash->semaphore) );
		bf_UnfixPage(groupLink, BF_DEFAULT, FALSE);
		*retLink = NULL;
	}

	return esmNOERROR;
}


/*
 *	Cl_ReadPage reads a page for a client.  It follows the following
 *	steps:
 *
 *	* Finds neigbor pages if necessary
 *	* Validates that the page is allocated
 *	* Checks the lrc cache and gets the page's lrc and FID
 *	* Gets locks on the file and page
 *	* Checks that the page is still correct after acquiring locks
 *    since getting a lock may block
 *  * Returns the page or a notification that the page is up-to-date
 *    to the client
 */
 void
cl_ReadPage (

	register MESSAGE		*message,
	register LINK			*unused 
)
{
	GROUPLINK		*groupLink = NULL;
	TRANSREC		*transRec;
	FID				fid;
	BOOL			bitSet;
	int				iovecParts;
	BOOL			LRCisCurrent = FALSE;
	BOOL			lockAcquired;
	LRC				lrc;
	FOUR			io_disk_reads = IO_DiskReads;
	BOOL			audit = FALSE;
	PID				resultPid;
#ifdef DEBUG
	int				__line__;
#define GOTO		__line__ = __LINE__; goto
#else
#define GOTO goto
#endif DEBUG

	TRPRINT(TR_CL, TR_LEVEL_1, ("reading page:%d pagesize:%d",
			message->body.page.pid.page, 
			RAISETWO(message->body.page.page2size)));

	/* 
	 * If this request includes getting the neighbor page, 
	 * we have to : lock the file & get the pid of the neighbor page.
	 * First, lock the file, since scans do not usually lock the pages.
	 */
	if( message->body.page.flags & REPLY_WITH_NEIGHBOR) {
		 LOCKMODE           lockMode = SH;
		
		if ((transRec = findTransRec(message->header.params.in.tid)) == NULL){
			SM_ERROR(TYPE_USER, esmBADTRANSID);
			reply(esmFAILURE, Active->errno, 1, TRUE);
		}

		Active->transRec = (char *) transRec;

		if (fi_ValidateFid(UserBufGroup, &(message->body.page.Fid)) 
			!= esmNOERROR) {
			reply(esmFAILURE, Active->errno, 1, TRUE);
		}

		if (lockFile(transRec, &(message->body.page.Fid), &lockMode, NOFLAGS))	{
			reply(esmFAILURE, Active->errno, 1, TRUE);
		}

		resultPid.volid = message->body.page.pid.volid;
		if (fi_NeighborPage(UserBufGroup, 
			&(message->body.page.Fid.pid), /* pid for file */
			 message->body.page.pid.page,	/* near-pid.page */
			 &resultPid.page,
			 message->body.page.flags & ~(REPLY_WITH_NEIGHBOR|REPLY_WITH_DATA),
			 NULL, NULL) == esmFAILURE) {

			switch(Active->errno) {
			case esmAFTERLASTPID:
			case esmBEFOREFIRSTPID:
				TRPRINT(TR_CL, TR_LEVEL_1, 
				("fi_NeighborPage returned resultPid %d", resultPid.page));
				/* ok, keep going */
				break;

			case esmNOERROR: 
				SM_ERROR(TYPE_FATAL, esmINTERNAL);

			default: 
				reply(esmFAILURE, Active->errno, 1, TRUE);
			}
		}

		if( (message->body.page.flags & REPLY_WITH_DATA) == 0) {
			/* we do not want the page - must have wanted neighborPid only */

			message->body.page.NeighborPid = resultPid;
			TRPRINT(TR_CL, TR_LEVEL_1, (" replying with errno %d resultPid %d",
				Active->errno, resultPid.page));
			reply(esmNOERROR, Active->errno, 1, TRUE); 
		}
	} /* REPLY_WITH_NEIGHBOR */

	SM_ASSERT( LEVEL_1, (message->body.page.flags & REPLY_WITH_DATA));

#ifdef DEBUG
	/* see if the page is allocated */
	if (io_CheckBitmap( &(message->body.page.pid), 0, &bitSet ))	{
		SM_ERROR(TYPE_USER, esmBADPID);
		GOTO ReplyAndExit;
	}
	if(bitSet == TRUE) { /* is free */
		SM_ERROR(TYPE_USER, esmBADPID);
		GOTO ReplyAndExit;
	}
#endif


	if (message->body.page.pageType == PAGE_SLOTTED)	{
		/* now, see if it is indeed a slotted page */

		if (io_CheckBitmap( &(message->body.page.pid), 
				message->body.page.pageType, &bitSet ))	{
			GOTO ReplyAndExit;
		}
		if (!bitSet) {
			SM_ERROR(TYPE_USER, esmBADPID);
			GOTO ReplyAndExit;
		}

		audit = TRUE;
	}

	/*
	 *	register this thread under a transaction,
	 *  so the thread can be notified (if it waits for a lock, say)
	 * if the tx aborts or such thing occurs.
	 * FALSE => not writing log
	 */
	if ((transRec = registerTrans(message->header.params.in.tid, FALSE))
			== NULL)	{

		SM_ERROR(TYPE_USER, esmBADTRANSID);
		GOTO ReplyAndExit;
	}

	/*
	 * First see if we can skip a lot of work by checking the
	 * LRC cache.  If the client gave us an up-to-date LRC,
	 * for the page, we can skip shipping the page (and so 
	 * we can skip reading it, as well).
	 * This REQUIRES that the page be in the lrc cache ONLY IF
	 * it's not in the page HT.  To assure that this is true,
	 * a page's lrc cache gets removed when the page gets inserted
	 * into the HT.
	 *
	 * Now, most of the time (one hopes), the page will be present,
	 * in which case, it would seem that the proper time to check
	 * the LrcCache is *after* we discover that the page is not present.
	 * ... but to reorganize the code to do this gets very messy, by
	 * the time we worry about getting the semaphore on the page, etc.,
	 * and we've not really saved any cycles.  So.. we check
	 * the LrcCache first.  
	 */
	if (getLrcAndFid(&(message->body.page.pid),
					 message->body.page.pageType,
					 message->body.page.page2size, FALSE, &fid, &lrc,
					 &groupLink) != esmNOERROR) {
		GOTO ReplyAndExit;
	}


	/*
	 *	Lock the page and possibly the file if the page is slotted
	 */
	lockAcquired = TRUE;
	switch (message->body.page.pageType)  {

		case PAGE_INDEX:

				TRPRINT(TR_CL, TR_LEVEL_1, ("locking index page %d",
					message->body.page.pid.page));

				if (lockPage(transRec, &(message->body.page.pid),
							 &message->body.page.PidLockMode, NOFLAGS)) {
					GOTO ReplyAndExit; 
				}
				break;
				
		case PAGE_INDEXDESC:
		
				TRPRINT(TR_CL, TR_LEVEL_1, ("locking indexdesc page %d",
					message->body.page.pid.page));
				
				if (lockPage(transRec, &(message->body.page.pid),
								&message->body.page.PidLockMode, NOFLAGS))  {
					GOTO ReplyAndExit; 
				}
				break;
				 
		case PAGE_SLOTTED: {

				TRPRINT(TR_CL, TR_LEVEL_1, ("slotted page"));

				/*
				 *	check to see if the fid is valid
				 */
				if (io_CheckBitmap( &(fid.pid), PAGE_FILEHDR, &bitSet )){
					GOTO ReplyAndExit; 
				}
				if (!bitSet) {
					SM_ERROR(TYPE_USER, esmBADFID);
					GOTO ReplyAndExit;
				}

				/*
				 *	lock the file if required
				 */
				if (message->body.page.FidLockMode != NL) {
					TRPRINT(TR_CL, TR_LEVEL_1, ("locking file for page %d", 
							message->body.page.pid.page));
					if (lockFile(transRec, &fid, 
								&message->body.page.FidLockMode, NOFLAGS)){
						GOTO ReplyAndExit; 
					}
				}

				/*
				 *	check the lock mode to see if we need to lock the page
				 */
				switch (getFileLockMode(transRec, &fid))	{

					case IX:
					case IS:

						/*
						 *	lock the page
						 */
						TRPRINT(TR_CL, TR_LEVEL_1, ("locking page %d",
							message->body.page.pid.page));
						if (lockPage(transRec, &(message->body.page.pid),
							&message->body.page.PidLockMode, NOFLAGS)) {
							GOTO ReplyAndExit; 
						}
						break;

					case SIX:

						/*
						 *	lock the page if it's exclusive
						 */
						TRPRINT(TR_CL, TR_LEVEL_1, ("locking page %d",
							message->body.page.pid.page));
						if (message->body.page.PidLockMode == EX)	{
							if(lockPage(transRec, 
								&(message->body.page.pid),
								&message->body.page.PidLockMode, NOFLAGS)) {
								GOTO ReplyAndExit; 
							}
						}
						break;

					case SH:
						/*
						 *	There is no need to lock the page if an SH is 
						 *	requested, and if EX is requested then the
						 *	file should have an SIX lock instead
						 */
						SM_ASSERT(LEVEL_3, 
								message->body.page.PidLockMode != EX);
						break;

					case EX:
						/*
						 *	No locking necessary
						 */
						break;
			
					default:
						SM_ERROR(TYPE_FATAL, esmINTERNAL);
						return;
				} /* switch (getFileLockMode(transRec, &fid) */

			} /* case PAGE_SLOTTED */
			break;

		case PAGE_LARGENODE:
		case PAGE_LARGEDATA:
			lockAcquired = FALSE;
			break;	

		default:
			SM_ERROR(TYPE_CRASH, esmINTERNAL);
			break;
	} /* switch on page type */


	/*
	 *	Verify that FID has not changed and that LRC's are current
	 */
	if (lockAcquired) {
		FID	newFid;
		if (getLrcAndFid(&(message->body.page.pid),
						 message->body.page.pageType,
						 message->body.page.page2size, TRUE, &newFid,
						 &lrc, &groupLink) != esmNOERROR) {
			GOTO UnfixReplyAndExit;
		}
		if ( (message->body.page.pageType == PAGE_SLOTTED) &&
			 !FIDEQ(fid, newFid) ) {
			/*
			 *	The page must have been deallocated and reallocated
			 */
			SM_ERROR(TYPE_USER, esmBADPID);	
			GOTO UnfixReplyAndExit;
		}
	}


	/* 
	 *	See if the version of the page cached at the client is valid
	 */
	if(AlwaysSendPage)  {
		TRPRINT(TR_CL, TR_LEVEL_1, ("AlwaysSendPage is on"));
		LRCisCurrent = FALSE;
	} else {
		LRCisCurrent = LRC_GEQ( message->body.page.pageLRC, &lrc);
	}
	if (LRCisCurrent) {
		TRPRINT(TR_CL, TR_LEVEL_1, ("LRCisCurrent - do not send page"));
		itx_client_current++;

		/*  no  page is appended  */
		iovecParts = 1;
		message->body.page.flags &= ~REPLY_WITH_DATA; 

		/* 
		 * How many disk reads did we actually do to accomplish this? 
		 * If this number is bad enough, we should start caching lists
		 * of LRC,PID so that we can respond to read requests w/o
		 * reading from the disk.
		 */
		itx_server_foiled  += (int) (IO_DiskReads - io_disk_reads);

	} else {

		/*
		 *	Client version is not current, prepare to send the page
		 */
		TRPRINT(TR_CL, TR_LEVEL_1, 
			("LRCis NOT current- send page w/ lrc %d.%d",
			lrc.wrapCount, lrc.count));
		itx_client_old++;

		/* get the page, doing a disk read if necessary */
		if (groupLink == NULL) {

			TRPRINT(TR_CL, TR_LEVEL_1, ("doing bf_ReadPage on page %d ",
				message->body.page.pid.page));

			if ((groupLink = bf_ReadPage(UserBufGroup, &(message->body.page.pid), message->body.page.page2size, BF_SEM)) == NULL){
				GOTO ReplyAndExit;
			}
		}
		SM_ASSERT( LEVEL_1, (groupLink->pageHash->pageType == message->body.page.pageType) ||
				(groupLink->pageHash->pageType == PAGE_GENERIC));
		groupLink->pageHash->pageType = message->body.page.pageType;

		/*
		 *	setup the iovec for the transfer back
		 */
		SM_ASSERT( LEVEL_1, groupLink != NULL);
		Active->iovec[1].iov_base = groupLink->bufFrame;
		Active->iovec[1].iov_len = RAISETWO(message->body.page.page2size);
		iovecParts = 2;

		/* For the H_CHUNK/PAGE_LARGEDATA case, we must send back the 
		 * lrc in the message structure.  Do not have to worry about 
		 * this when the page is up-to-date.
		 */
		if( message->body.page.pageType == PAGE_LARGEDATA )
			message->body.page.pageLRC = lrc;


		SM_ASSERT(LEVEL_1, message->body.page.flags & REPLY_WITH_DATA );

		if(audit) {
			TRPRINT(TR_CL, TR_LEVEL_1, ("Auditing p %d", 
				message->body.page.pid.page));
			sm_AuditSlotPage( (SLOTTEDPAGE *)(groupLink->bufFrame));
		}
	} 

	TRPRINT(TR_CL, TR_LEVEL_1, 
		("flags 0x%x replying with errno %d NeighborPid.page %d",
				message->body.page.flags,
				Active->errno, 
				message->body.page.NeighborPid.page));

	if(message->body.page.flags & REPLY_WITH_NEIGHBOR) {
		message->body.page.NeighborPid = resultPid;
	}

	reply(esmNOERROR, Active->errno, iovecParts, FALSE); /* do not die */

	if(groupLink) {
		signalSemaphore( &(groupLink->pageHash->semaphore) );
		bf_UnfixPage(groupLink, BF_DEFAULT, FALSE);
	}

	threadRestart();
	/* NOTREACHED */

	/* Rather than duplicate this code so many times, we jump here. */
UnfixReplyAndExit:
	if(groupLink) {
		signalSemaphore( &(groupLink->pageHash->semaphore) );
		bf_UnfixPage(groupLink, BF_DEFAULT, FALSE);
	}
	/* drop through */

ReplyAndExit:
	reply(esmFAILURE, Active->errno, 1, TRUE);
}
