/*
 *   $RCSfile: cl_WritePage.c,v $  
 *   $Revision: 1.1.1.1 $  
 *   $Date: 1996/05/04 21:55:37 $      
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/

#include "sysdefs.h"
#include "ess.h"
#include "checking.h"
#include "trace.h"
#include "error.h"
#include "list.h"
#include "pool.h"
#include "tid.h"
#include "io.h"
#include "lock.h"
#include "object.h"
#include "msgdefs.h"
#include "thread.h"
#include "semaphore.h"
#include "latch.h"
#include "link.h"
#include "threadstate.h"
#include "lsn.h"
#include "bf.h"
#include "file.h"
#include "bitvec.h"
#include "consist.h"
#include "logrecs.h"
#include "log.h"
#include "volume.h"
#include "trans.h"
#include "openlog.h"
#include "msg_funcs.h"
#include "fi_extfuncs.h"
#include "io_extfuncs.h"
#include "bf_extfuncs.h"
#include "sm_extfuncs.h"
#include "thread_funcs.h"
#include "trans_extfuncs.h"
#include "log_extfuncs.h"
#include "cl_funcs.h"
#include "server_util_funcs.h"
#include "bf_globals.h"
#include "thread_globals.h"
#include "log_globals.h"
#include "sm_macro.h"


/*
 *	This function is used to restore a new page after 
 *  an attempt to log the entire page fails.
 */
static void restorePage(GROUPLINK *groupLink, PAGETYPE pageType)
{
	SLOTTEDPAGE	*slottedPage;
	FID			fid;
	LGNODE		*lgNode;
	LGDATAPAGE	*lgDataPage;
	LRC			*lrc;

	switch (pageType) {
	case PAGE_SLOTTED:
		slottedPage = (SLOTTEDPAGE*) groupLink->bufFrame;
		fid = slottedPage->header.fid;
		INITPAGE(slottedPage, groupLink->pageHash->hashList.hashPid, fid, FALSE, 0);
		slottedPage->header.uniqueCount = 0;
		lrc = &(slottedPage->header.lrc);
        break;
		
    case PAGE_LARGENODE:
		lgNode = (LGNODE*) groupLink->bufFrame;
		INIT_LARGENODE_MAGIC(lgNode);
		lrc = &(lgNode->header.lrc);
        break;

    case PAGE_LARGEDATA:
		lgDataPage = (LGDATAPAGE*) groupLink->bufFrame;
		INIT_LGDATAHDR_MAGIC(lgDataPage);
		lrc = &(lgDataPage->header.lrc);
        break;
      
    case PAGE_INDEXDESC:
		lrc =
			&(((INDEXDESCPAGE*)groupLink->bufFrame)->lrc); /* beginning of page */
        break;

    case PAGE_INDEX:
		lrc = (LRC*) groupLink->bufFrame; /* at beginning of page */
        break;

    default:
		SM_ERROR(TYPE_FATAL, esmINTERNAL);
		break;
	}

	INITIALIZE_LRC(lrc);
}



 void
cl_WritePage (

	register MESSAGE	*message,
	register LINK		*link 
)
{

	GROUPLINK	*groupLink;
	GROUPLINK	*tempLink;
	char		*buffer;
	LRC			*lrc;
	TRANSREC 	*transRec;	
	VOLREC		*volRec;

	TRPRINT(TR_CL, TR_LEVEL_1, ("writing pid:%d pagecount:%d",
			message->body.page.pid.page, 
			BLOCKCOUNT(message->body.page.page2size)));

	/*
	 *	suspend the link so that it does not get selected on
	 *	the basis of the data that is sent
	 */
	suspendLink(link);

	/* register with transaction that sent message, FALSE => not writing log */
	if ((transRec = registerTrans(message->header.params.in.tid, FALSE)) == NULL) {
        /*
         *  make sure that if there is a tran under this id that it
         *  is being aborted
         */
        if ((transRec = findTransRec(message->header.params.in.tid)) != NULL) {
            if (transRec->transState == T_ACTIVE) {
				SM_ERROR(TYPE_FATAL, esmINTERNAL);
			}
        }

		/* read the page into a garbage location */
		if ((tempLink = bf_GetTempPage(message->body.page.page2size)) == NULL) {
			SM_ERROR(TYPE_FATAL, esmINTERNAL);
		}
		resumeLink(link);
		buffer = tempLink->bufFrame;
		if (receive(link, buffer, RAISETWO(message->body.page.page2size)) < 0) {
			/*  blow the thread away, the link is gone discard the page */
		}
		bf_FreeTempPage(tempLink);

		SM_ERROR(TYPE_USER, esmBADTRANSID);
		if (message->header.replyRequested) {
            reply(esmFAILURE, Active->errno, 1, TRUE);
        } else {
#ifdef DEBUG
			/*
			 *	print a message to remind us that clients should really
			 *	expect a response every 10 messages or so
			 */
            SM_ERROR(TYPE_LOG, esmBADTRANSID);
#endif
            threadRestart();
        }

	}

	/*
	 *	read the page
	 */
	if ((groupLink = bf_ReadPage(UserBufGroup, &(message->body.page.pid),
					  			 message->body.page.page2size,
								 BF_SEM|BF_NOREAD)) == NULL)	{
		if (message->header.replyRequested) {
            reply(esmFAILURE, Active->errno, 1, TRUE);
        } else {
            /* Error, so abort transaction */
            internalAbort(transRec, Active->errno);
			transThreadDeath(Active);
            threadRestart();
        }
	}
	SM_ASSERT( LEVEL_1,
		(groupLink->pageHash->pageType == message->body.page.pageType)
		||
		(groupLink->pageHash->pageType == PAGE_GENERIC));

	groupLink->pageHash->pageType = message->body.page.pageType;

	buffer = groupLink->bufFrame;
	resumeLink(link);

	if (receive(link, buffer, RAISETWO(message->body.page.page2size)) < 0)	{

		/* free up the page */
		signalSemaphore( &(groupLink->pageHash->semaphore) );
	
		if (PAGE_IS_DIRTY(groupLink->pageHash)) {
			/*
			 * 	Fatal error since updates were lost.  Possibly we
			 *	could do media recovery.
			 */
			/*
			 *	Here we should restore the page contents to 
			 *	its original value.
			 */
			SM_ERROR(TYPE_CRASH, esmNOTIMPLEMENTED);
		}

		/* the page has been corrupted, to remove it from buffer pool */
		bf_DiscardPage(groupLink->bufGroup, groupLink);

		SM_ERROR(TYPE_USER, esmBADSLOTTEDMAGIC);
		if (message->header.replyRequested) {
			reply(esmFAILURE, Active->errno, 1, TRUE);
		} else {
			/*
			 *	blow the thread away, the link is gone
			 *	discard the page
			 */

			transThreadDeath(Active);
			threadRestart();
		}
	}

	switch (message->body.page.pageType) {

	case PAGE_SLOTTED:

		/*
		 *	check to see if the link needs to be updated
		 */
		fi_DoLinkConsist( &(message->body.page.pid),
						  groupLink->pageHash, (SLOTTEDPAGE *) buffer );
		if (CHECK_SLOTTED_MAGIC((SLOTTEDPAGE *) buffer)) {
			/*
			 *	Bad page type
			 */
			signalSemaphore( &(groupLink->pageHash->semaphore) );
		
			if (PAGE_IS_DIRTY(groupLink->pageHash)) {
				/*
				 * 	Fatal error since updates were lost.  Possibly we
				 *	could do media recovery.
				 */
				SM_ERROR(TYPE_CRASH, esmBADSLOTTEDMAGIC);
			}

			/* the page has been corrupted, to remove it from buffer pool */
			bf_DiscardPage(groupLink->bufGroup, groupLink);

			SM_ERROR(TYPE_USER, esmBADSLOTTEDMAGIC);
			if (message->header.replyRequested) {
				reply(esmFAILURE, Active->errno, 1, TRUE);
			} else {
				/* Error, so abort transaction */
				internalAbort(transRec, Active->errno);
				transThreadDeath(Active);
				threadRestart();
			}
		}
		lrc =  & ( ((SLOTTEDPAGE *) buffer)->header.lrc );
		
		sm_AuditSlotPage ((SLOTTEDPAGE *)buffer);

		break;

	case PAGE_LARGENODE:

		if (((LGNODE *) buffer)->header.magic != LARGENODE_MAGIC) {
			/*
			 *	Bad page type
			 */
			signalSemaphore( &(groupLink->pageHash->semaphore) );
			if (PAGE_IS_DIRTY(groupLink->pageHash)) {
				/*
				 * 	Fatal error since updates were lost.  Possibly we
				 *	could do media recovery.
				 */
				SM_ERROR(TYPE_CRASH, esmINTERNAL);
			}

			/* the page has been corrupted, to remove it from buffer pool */
			bf_DiscardPage(groupLink->bufGroup, groupLink);

			SM_ERROR(TYPE_USER, esmINTERNAL);
			if (message->header.replyRequested) {
				reply(esmFAILURE, Active->errno, 1, TRUE);
			} else {
				/* Error, so abort transaction */
				internalAbort(transRec, Active->errno);
				transThreadDeath(Active);
				threadRestart();
			}
		}
		lrc =  & ( ((LGNODE *) buffer)->header.lrc );
		break;

	case PAGE_LARGEDATA:
	
		/*
		 *	The lrc for the page is in the message, so put it
		 *	on the page.  Also put the magic number on the page
		 */
		((LGDATAPAGE *) buffer)->header.lrc = message->body.page.pageLRC;
		((LGDATAPAGE *) buffer)->header.magic = LGDATAHDR_MAGIC;
		lrc =  & ( ((LGDATAPAGE *) buffer)->header.lrc );
		break;

	case PAGE_INDEX:
		
		lrc = (LRC*) buffer;
		break;
		
	case PAGE_INDEXDESC:
	
		lrc = & (((INDEXDESCPAGE*) buffer)->lrc);
		break;

	default:
		
		/*
		 *	Bad page type
		 */
		signalSemaphore( &(groupLink->pageHash->semaphore) );

		if (PAGE_IS_DIRTY(groupLink->pageHash)) {
			/*
			 * 	Fatal error since updates were lost.  Possibly we
			 *	could do media recovery.
			 */
			SM_ERROR(TYPE_CRASH, esmINTERNAL);
		}

		/* the page has been corrupted, to remove it from buffer pool */
		bf_DiscardPage(groupLink->bufGroup, groupLink);

		SM_ERROR(TYPE_USER, esmINTERNAL);
		if (message->header.replyRequested) {
			reply(esmFAILURE, Active->errno, 1, TRUE);
		} else {
			SM_ERROR(TYPE_LOG, Active->errno);
			/* Error, so abort transaction */
			internalAbort(transRec, Active->errno);
			transThreadDeath(Active);
			threadRestart();
		}
		lrc = (LRC *)NULL;
		break;
	}

	TRPRINT(TR_CL, TR_LEVEL_1, 
		("writing page %d, type %d, with LRC %d.%d",
			message->body.page.pid.page, 
			message->body.page.pageType, 
			lrc==NULL?0:lrc->wrapCount, 
			lrc==NULL?0:lrc->count));
			
	PRINT_PROGRESS( ("received client page %d, page is new:%d\n", message->body.page.pid.page, (int)(message->body.page.flags & PAGE_IS_NEW)))

	/* find out which volume this page belongs to */
	if ((volRec = io_FindVolRec(message->body.page.pid.volid)) == NULL) {
		SM_ERROR(TYPE_FATAL, esmINTERNAL);
	}

	/*
	 *  If this page was new (has no log records generated for it)
	 *  then we must generate a log record for it.
	 */
	if (message->body.page.flags & PAGE_IS_NEW) {
		/* suspendLink(link); */

		if (sm_LogEntirePage(groupLink, message->body.page.pageType) != esmNOERROR) {
			if (message->header.replyRequested) {
				reply(esmFAILURE, Active->errno, 1, FALSE);
			} else {
				/* do nothing */
			}

			/* Error, so abort transaction */

			/*	resumeLink(link); */

			/*
			 *	Here we should restore the page contents to 
			 *	its original value.  This is easy since the page
			 *	is new.
			 */
			restorePage(groupLink, message->body.page.pageType);
			signalSemaphore( &(groupLink->pageHash->semaphore) );
			bf_UnfixPage(groupLink, BF_DEFAULT, FALSE);
			internalAbort(transRec, esmNOFREELOGSPACE);
			transThreadDeath(Active);
			threadRestart();
		}
		/*	resumeLink(link); */

	} else {

		/*	
		 *	The idea here is to give the page a firstLSN-LRC (ie. the
		 *	LSN-LRC for the first log record affecting the page).  
		 *	We cannot simply assign the firstLSN-LRC from the message,
		 *	since it may not accurately reflect the firstLSN for the
		 *	transaction.  If this is the case the firstLSN for the
		 *	transaction is used instead (this should be rare).
		 */

#ifdef DEBUG
#ifdef INIT_LRC_IS_LSN
		SM_ASSERT(LEVEL_3, message->body.page.FirstLRC.wrapCount <= OpenLog.wrapCount);	
#else
		SM_ASSERT(LEVEL_3, message->body.page.FirstLRC.count < 10000);	
		SM_ASSERT(LEVEL_3, message->body.page.FirstLRC.wrapCount == 0);	
#endif /* INIT_LRC_IS_LSN */
		SM_ASSERT(LEVEL_3, message->body.page.FirstLSN.wrapCount <= OpenLog.wrapCount);	
		SM_ASSERT(LEVEL_3, message->body.page.FirstLSN.offset < OpenLog.fileBytes);	
#endif /* DEBUG */

		if (compareLSN(&(message->body.page.FirstLSN), &(transRec->firstLSN)) < 0) {

			DEPEND_LOG(groupLink->pageHash, link->lastForceMark, 
						&(transRec->firstLSN), &(message->body.page.FirstLRC));
		} else {
			DEPEND_LOG(groupLink->pageHash, link->lastForceMark, 
						&(message->body.page.FirstLSN), 
						&(message->body.page.FirstLRC));
		}

		/*
		 *	If the page is not on a temporarly volume,
		 *	link the page onto a list of all dirty pages shipped from
		 *	the client for the transaction, since the last checkpoint.
		 *
		 *	It's possible the page is already on the list if it had
		 *	been returned from the client earlier.
		 *	Since a page can only be dirtied by one client at a time,
		 *	it's ok to just not add it if it's already on a list
		 */
		SM_ASSERT(LEVEL_1, ((volRec->volflags & VOL_TEMP) ==
			(volRec->header->properties & VOLPROP_TEMP)));

		if ( !(volRec->volflags & VOL_TEMP) &&
			 (NOT_LIST_MEMBER(&(groupLink->pageHash->clientDirtyList)))) {
			listEnq(&(transRec->clientDirtyList), &(groupLink->pageHash->clientDirtyList));
		}
	}

	signalSemaphore( &(groupLink->pageHash->semaphore) );

	bf_UnfixPage(groupLink, BF_DEFAULT, TRUE);

	if (message->header.replyRequested) {
		reply(esmNOERROR, esmNOERROR, 1, TRUE);
	} else {
		transThreadDeath(Active);
		threadRestart();
	}
}
