/*
 *   $RCSfile: initializeServer.C,v $  
 *   $Revision: 1.1.1.1 $  
 *   $Date: 1996/05/04 21:55:38 $      
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/
#include "sysdefs.h"
#include "ess.h"
#include "checking.h"
#include "trace.h"
#include "error.h"

#include "list.h"
#include "tid.h"
#include "io.h"
#include "lock.h"
#include "object.h"
#include "latch.h"
#include "lsn.h"
#include "bitvec.h"
#include "msgdefs.h"
#include "consist.h"
#include "logrecs.h"
#include "pool.h"
#include "undo.h"
#include "log.h"
#include "disk.h"
#include "thread.h"
#include "semaphore.h"
#include "bf.h"
#include "link.h"
#include "msgvector.h"
#include "resource.h"
#include "bitmap.h"
#include "file.h"
#include "volume.h"
#include "trans.h"
#include "openlog.h"
#include "adminmsg.h"

#include "disk_globals.h"
#include "recover_globals.h"
#include "bf_globals.h"
#include "log_globals.h"
#include "msg_globals.h"
#include "io_globals.h"
#include "thread_globals.h"

#include "admin_funcs.h"
#include "io_extfuncs.h"
#include "bf_extfuncs.h"
#include "bm_extfuncs.h"
#include "mr_extfuncs.h"
#include "msg_funcs.h"
#include "disk_funcs.h"
#include "util_funcs.h"
#include "server_util_funcs.h"
#include "trans_extfuncs.h"
#include "fi_extfuncs.h"
#include "thread_funcs.h"
#include "init_funcs.h"
#include "trace_funcs.h"
#include "lm_extfuncs.h"
#include "log_extfuncs.h"
#include "resource_funcs.h"
#include "recover_extfuncs.h"
#include "undo_extfuncs.h"
#include "queue_consist.h"
#include "timer.h"
#include "softwareVersion.h"
#include "distr_extfuncs.h"

BOOL InitializationCompleted = FALSE;

 void
initializeServer()
{
	TRACE(TR_INIT, TR_LEVEL_1);

	if(!formatvolRunning) {
	   fprintf(sm_ErrorStream, 
				"Exodus Storage Manager Server Release %s(%x)\n",
				VERSION_NUMBER, SOFTWARE_VERSION);

		fprintf(sm_ErrorStream, "Initialization in progress ... \n");
	}


	/*
	 * get environment variables 
	 * and parameters that determine the sizes of other
	 * things initialized.
	 */
	initializeEnv();

	initializeTrace();
	TRACE(TR_INIT, TR_LEVEL_1);

	/* 
	 * we must catch signals before calling initializeSharedMemory()
	 * so ^C doesn't leave us with garbage "ipcs".
	 */
	catchSignals(); 

	/*
	 * init shared memory - relies on env variables
	 */
	initializeSharedMemory();

#ifdef DEBUG
	if(!formatvolRunning)
		TRPRINT(TR_LEVEL_1, TR_BF, ("Hybrid LRC/LSN being used"));
#endif /* DEBUG */

	admin_ClearStats();

	initVolume();

	 /*
	  *	Establish  the connection socket and allocate the MESSAGE buffers.
	  * These go in shared memory, so it must already have been initialized.
	  */
	initializeMsg( formatvolRunning==FALSE );

#ifdef MEDIA_RECOVERY
	if(!formatvolRunning) {
		int		mr_socket;

		if (mr_InitializeSM_Stub(&mr_socket, MediaRecovery) != esmNOERROR) {
			SM_ERROR(TYPE_CRASH, esmINTERNAL);
		}
		if (MediaRecovery) {
			Links[mr_socket].linkClass = CL_MEDIA_RECOVERY; 
			setLink(&(Links[mr_socket]));
		} 
	}
#endif MEDIA_RECOVERY

	if(FastPathOnly == FALSE) {
		/* 
		 * initialize the queues of DISKMSGs in shared memory 
		 */
		initializeQueues();
	}

	bf_Initialize();

	/*
	 *	initialize the bm layer
	 */
	bm_Initialize();

	initializeTrans();

	initializeRecover(); /* empty */

#ifndef INIT_LRC_IS_LSN
	/*
	 *	initialize pools used in undoing transactions during normal
	 *	operations (ie. not restart)
	 */
	initMissingUpdatePools();
#endif 

	fi_Initialize();

	initializeLocks();

	/*  Initializing the threads creates the Active tcb */
	initializeThreads(THREAD_STACKSIZE, (PFI) loop);

	/* 
	 * The I/O layer uses Active for reading volumes to get
	 * the volume ids, so it must be initialized after threads.
	 * It also uses shared memory and queues, so it must be
	 * initialized after they are.
	 */
	io_Init();

	if(formatvolRunning) 
		return; 
	/*
	 * Timers use threads (for testing purposes), so they
	 * must be initialized after threads are initialized.
	 */
	ServerTimer.Init(MAX_TIMERS);

	initializeStdin(); /* Kludge: Must be done after io_Init() so that
		io_Init, when fast-path reading the volumes, can "borrow"
		the stdin's Link structure for avoiding forking disk procs.
		*/

	/*
	 *  initialize distributed transaction stuff
	 */
	initializeDistr();

	/*
	 *	initialize resources
	 */
	initializeResources();

#ifdef JUNK
	initializeAccess();
#endif JUNK

	/*
	 *	Start up log preflush thread .  Use T_CHILD so the
	 *	thread can put itself on a waiting list.
	 */
	threadFork(T_CHILD, (PFI)preflushLogThread, 0, NULL);

	/*
	 *	Start up log checkpoint thread .  Use T_CHILD so the
	 *	thread can put itself on a waiting list.
	 */
	threadFork(T_CHILD, (PFI)checkpointThread, 0, NULL);


	/*
	 *      Start up thread that asynchrounously forces dirty pages after
	 *      a checkpoint.  Use T_CHILD so the
	 *      thread can put itself on a waiting list.
	 */
	threadFork(T_CHILD, (PFI)asyncForceDirtyPagesThread, 0, NULL);

	/*
	 * Start a threads to check for stale client links and
	 * to age lock requests.
	 * Not a real big matter whether we use T_CHILD or T_PARENT
	 */
	threadFork(T_CHILD, (PFI)timeOutClient, 0, NULL);
	threadFork(T_CHILD, (PFI)ageLocks, 0, NULL);

	/*
	 *	Start up coordinator thread .  Use T_CHILD so that the
	 *	thread can put itself on a waiting list
	 */
	threadFork(T_CHILD, (PFI)coord, 0, NULL);

	/*
	 *	open the log disk 
	 */
	if (RecoverLog) {
		fprintf(sm_ErrorStream, "Performing restart recovery ...\n");
	} else {
		fprintf(sm_ErrorStream, "Regenerating the log ...\n");
	}
	openLogDisk();
	if (ShutdownAfterRecovery) {
		fprintf(sm_ErrorStream, "Shutdown requested after log regeneration.\n");
		admin_ShutServer(SHUT_CLEAN_VOLUMES);
	}

	/*
	 *	start the log at the correct location
	 *	
	 *	This whould really be caclulated by the server after
	 *	examining all mountable volumes
	 */
	if (LogStartWrapCount != NULL_LOG_WRAP_COUNT) {
		setLogStartLocation(LogStartWrapCount);
	}
	activateLogging();

	/*
	 *	print user message
	 */

	InitializationCompleted = TRUE;
	fprintf(sm_ErrorStream, "Initialization completed\n");
	fprintf(sm_ErrorStream, "Server is ready for requests.\n");

	/*
	 *	start the threads process
	 */
#ifdef vax 
	asm("jmp _threadRestart+2"); /* +2 to skip register mask */
#else
	threadRestart();
#endif vax
}
