/*
 *   $RCSfile: server_ExternalActions.c,v $  
 *   $Revision: 1.1.1.1 $  
 *   $Date: 1996/05/04 21:55:41 $      
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/

#include "sysdefs.h"
#include "ess.h"
#include "checking.h"
#include "trace.h"
#include "error.h"
#include "list.h"
#include "pool.h"
#include "tid.h"
#include "io.h"
#include "lock.h"
#include "object.h"
#include "msgdefs.h"
#include "thread.h"
#include "semaphore.h"
#include "link.h"
#include "lsn.h"
#include "latch.h"
#include "bf.h"
#include "volume.h"
#include "trans.h"
#include "bitvec.h"
#include "trans_intfuncs.h"
#include "trans_extfuncs.h"
#include "msg_funcs.h"
#include "cl_funcs.h"
#include "thread_funcs.h"
#include "thread_globals.h"
#include "trans_globals.h"
#include "distr.h"
#include "distr_intfuncs.h"
#include "distr_extfuncs.h"
#include "distr_globals.h"
#include "server_util_funcs.h"
#include "admin_funcs.h"


/*
 *	This routine is called by a server upon receiving a command
 *	(commit or abort) from another server, the coordinator for 
 *	this transaction
 */
 void
server_ExternalCommand (

	MESSAGE				*message,
	LINK				*unused 
)
{

	register TRANSREC	*transRec;
	TID					localTid = message->body.command.localTid;
	TID					coordTid = message->body.command.coordTid;
	COMMAND				command = message->body.command.command;


	TRPRINT(TR_TRANS, TR_LEVEL_1, ("localTid:%d", localTid));

	/*
	 *	initialize the type, coordTid, and localTid of the reply message
	 */
	message->header.type = SERVER_ACK;
	message->body.ack.coordTid = coordTid;
	message->body.ack.localTid = localTid;

	/*
	 *	check the state of transaction processing
	 */
	if (TransInfo.transState != T_ACTIVE)	{

		/* do nothing */
		transThreadDeath(Active);
		threadRestart();
	}

	/*
	 *	check the state of the log
	 */
	if (TransInfo.logState != L_ACTIVE)	{

		/* do nothing */
		transThreadDeath(Active);
		threadRestart();
	}

	/*
	 *	try to find the transaction
	 */
	if ((transRec = findTransRec(localTid)) == NULL) {

		/* 
		 *	not found implies ACK was lost
		 *	send another ACK
		 */
		TRPRINT(TR_TRANS|TR_MSG|TR_DISTR, TR_LEVEL_1, 
				("server: returning SERVER_ACK : localTid:%d", localTid));
		replyDatagram(esmNOERROR, esmNOERROR, 1, TRUE);
	}

	/*
	 *	mark the active transaction
	 */
	Active->transRec = (char *) transRec;

	/*
	 *	check to see if the transaction is in the active state 
	 *	this means a log record is being written 
	 *	do nothing - another thread will reply
	 */
	if (transRec->transState == T_ACTIVE)	{

		transThreadDeath(Active);
		threadRestart();
	}

	/*
	 *	check to see if the transaction is in the prepared state 
	 */
	if (transRec->transState == T_PREPARED)	{

		/*
		 * check that the command is valid
		 */
		if ((command == C_COMMIT) || (command == C_ABORT)) {

			/*
			 *	mark the command		
			 */
			transRec->coordCommand = message->body.command.command;
		}

		/*
		 *	the thread waiting for the command will take care 
		 *	of the rest
		 */
		transThreadDeath(Active);
		threadRestart();

	} else if (transRec->transState == T_COMMIT) {

		if (command == C_COMMIT) {
			
			TRPRINT(TR_TRANS|TR_MSG|TR_DISTR, TR_LEVEL_1, 
			  ("server: returning SERVER_ACK : localTid:%d", localTid));

			replyDatagram(esmNOERROR, esmNOERROR, 1, TRUE);

		} else if (command == C_ABORT) {
			
			TRPRINT(TR_DISTR|TR_TRANS, TR_LEVEL_1, 
				("command:%d , transState:%d", command, transRec->transState));
			/* crash because we are in the wrong state */
			SM_ERROR(TYPE_CRASH, esmINTERNAL);
		}

	} else if (transRec->transState == T_ABORT) {

		if (command == C_ABORT) {

			TRPRINT(TR_TRANS|TR_MSG|TR_DISTR, TR_LEVEL_1, 
			  ("server: returning SERVER_ACK : localTid:%d", localTid));

			replyDatagram(esmNOERROR, esmNOERROR, 1, TRUE);

		} else if (command == C_COMMIT) {
			
			TRPRINT(TR_DISTR|TR_TRANS, TR_LEVEL_1, 
				("command:%d , transState:%d", command, transRec->transState));
			/* crash because we are in the wrong state */
			SM_ERROR(TYPE_CRASH, esmINTERNAL);
		}

	} else {

		TRPRINT(TR_DISTR|TR_TRANS, TR_LEVEL_1, 
			("command:%d , transState:%d", command, transRec->transState));

		/* crash because we are in the wrong state */
		SM_ERROR(TYPE_CRASH, esmINTERNAL);
	}
}



/*
 *	This routine is called by a server upon receiving a prepare
 *	message from another server, the coordinator for this transaction
 */
 void
server_ExternalPrepare (

	MESSAGE				*message,
	LINK				*unused 
)
{

	register TRANSREC	*transRec;
	TID					localTid = message->body.voteReq.localTid;
	TID					coordTid = message->body.voteReq.coordTid;
	VOTE				vote = READVOTE;


	TRPRINT(TR_TRANS, TR_LEVEL_1, ("localTid:%d", localTid));

	/*
	 *	initialize the type, coordTid, and localTid of the reply message
	 */
	message->header.type = SERVER_VOTE;
	message->body.vote.coordTid = coordTid;
	message->body.vote.localTid = localTid;

	/*
	 *	check the state of transaction processing
	 */
	if (TransInfo.transState != T_ACTIVE)	{

		/* do nothing */
		transThreadDeath(Active);
		threadRestart();
	}

	/*
	 *	check the state of the log
	 */
	if (TransInfo.logState != L_ACTIVE)	{

		/* do nothing */
		transThreadDeath(Active);
		threadRestart();
	}

	/*
	 *	try to find the transaction
	 */
	if ((transRec = findTransRec(localTid)) == NULL)	{

		/* not found implies ABORT */
		message->body.vote.vote = NOVOTE;
		TRPRINT(TR_TRANS|TR_MSG|TR_DISTR, TR_LEVEL_1, 
				("server: returning NOVOTE : localTid:%d", localTid));
		replyDatagram(esmNOERROR, esmNOERROR, 1, TRUE);
	}

	/*
	 *	mark the active transaction
	 */
	Active->transRec = (char *) transRec;

	/*
	 *	check to see if the transaction is already prepared
	 */
	if (transRec->transState == T_PREPARED)	{

		message->body.vote.vote = YESVOTE;
		TRPRINT(TR_TRANS|TR_MSG|TR_DISTR, TR_LEVEL_1, 
				("server: returning YESVOTE : localTid:%d", localTid));
		replyDatagram(esmNOERROR, esmNOERROR, 1, TRUE);
	}

	/*
	 *	check to see if the transaction is in the quiesce state 
	 *	implies some other thread is going to reply
	 *	do nothing
	 */
	if (transRec->transState == T_QUIESCE)	{

		transThreadDeath(Active);
		threadRestart();
	}

	/*
	 *	check to see if the transaction is in commit state 
	 *	implies this is a duplicate prepare message and
	 *	the coordinator has since issued a commit command
	 *	do nothing
	 */
	if (transRec->transState == T_COMMIT)	{

		transThreadDeath(Active);
		threadRestart();
	}

	/*
	 *	check to see if the transaction is not active
	 *	any other state implies abort	
	 */
	if (transRec->transState != T_ACTIVE)	{

		message->body.vote.vote = NOVOTE;
		TRPRINT(TR_TRANS|TR_MSG|TR_DISTR, TR_LEVEL_1, 
				("server: returning NOVOTE : localTid:%d", localTid));
		replyDatagram(esmNOERROR, esmNOERROR, 1, TRUE);
	}

	/*
	 *	check to see if the transaction is active but the
	 *	intent is to prepare
	 *	implies this is a duplicate prepare message and some other
	 *	thread is processing the earlier message
	 *	that thread will eventually reply, hence do nothing
	 */
	if ((transRec->transState == T_ACTIVE) && (transRec->intent == T_PREPARED))	{

		transThreadDeath(Active);
		threadRestart();
	}

	/*
	 *	store the coordAddr and coordTid in the transRec
	 */
	transRec->coordAddr = Active->replyAddr;
	transRec->coordTid = coordTid;

	/*
	 *	mark the intent to prepare 
	 */
	transRec->intent = T_PREPARED;

#ifdef JUNK
	/*
	BEWARE: the message type here is probably a PREPAREBODY not
	a TRANSBODY
	 *	if the trans has logging turned off, record this fact ??
	 */
	if (message->body.trans.loggingOn == FALSE) {
		TRPRINT(TR_TRANS, TR_LEVEL_1, ("localTid:%d has logging turned off", localTid));
		PRINT_PROGRESS(("localTid:%d has logging turned off\n", localTid));
		transRec->loggingOn = FALSE;
	}
#endif JUNK

	/*
	 *	quiesce all resources held by the transaction
	 */

	quiesceTrans(transRec, esmTRANSCOMMITTED);

#ifdef DEBUG
	if (DebugCrashLoc == 11)
		admin_ShutServer(0);
#endif DEBUG

	/*
	 *	check to see if everything went alright
	 */
#ifdef DEBUG
	if ((transRec->intent != T_PREPARED) || 
		(DebugDistrVote == NOVOTE))
#else DEBUG
	if (transRec->intent != T_PREPARED)
#endif DEBUG
	{

		/*
		 *	reply with a NO vote 
		 */
		message->body.vote.vote = NOVOTE;
		TRPRINT(TR_TRANS|TR_MSG|TR_DISTR, TR_LEVEL_1, 
				("server: returning NOVOTE : localTid:%d", localTid));
		replyDatagram(esmNOERROR, esmNOERROR, 1, FALSE);

		/*
		 *	abort the transaction
		 */
		abortTrans(transRec, esmTRANSNOTPREPARED);
		threadRestart();

	} else {

		/*
		 *	move into the prepared state after writing the prepared
		 *	log record
		 */
		if (serverPrepareTrans(transRec, &vote) < 0) {
			/*
			 *	reply with a NO vote 
			 */
			message->body.vote.vote = NOVOTE;
			TRPRINT(TR_TRANS|TR_MSG|TR_DISTR, TR_LEVEL_1, 
					("server: returning NOVOTE : localTid:%d", localTid));
			replyDatagram(esmNOERROR, esmNOERROR, 1, FALSE);
			abortTrans(transRec, esmTRANSNOTPREPARED);
			threadRestart();
		}

#ifdef DEBUG
	if (DebugCrashLoc == 12)
		admin_ShutServer(0);
#endif DEBUG

		if (vote == READVOTE) {
			
			message->body.vote.vote = READVOTE;

			TRPRINT(TR_TRANS|TR_MSG|TR_DISTR, TR_LEVEL_1, 
					("server: returning READVOTE : localTid:%d", localTid));

			replyDatagram(esmNOERROR, esmNOERROR, 1, FALSE);
			threadRestart();
		}

		/*
		 *	record the vote 
		 */
		message->body.vote.vote = vote;

		/*
		 *	mark this thread  as the master for this trans
		 */
		transRec->masterThread = Active;

		/*
		 *	keep sending the vote until reply is recd
		 */
		while (transRec->coordCommand == C_UNKNOWN)	{
			
			replyDatagram(esmNOERROR, esmNOERROR, 1, FALSE);

#ifdef DEBUG
	if (DebugCrashLoc == 13)
		admin_ShutServer(0);
#endif DEBUG

			/*
			 *	go to sleep on timeout
			 */
			Sleep(T_SERVERTIMEOUT);
		}

		/*
		 *	commit/abort based on command
		 *	reply to coord
		 */
		replyWithAck(transRec);
	}
}
