/*
 *   $RCSfile: FI_GetAfter.c,v $  
 *   $Revision: 1.1.1.1 $  
 *   $Date: 1996/05/04 21:55:41 $      
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/


#include "sysdefs.h"
#include "ess.h"
#include "checking.h"
#include "error.h"
#include "trace.h"
#include "io.h"
#include "page.h"
#include "list.h"
#include "tid.h"
#include "lock.h"
#include "object.h"
#include "msgdefs.h"
#include "thread.h"
#include "semaphore.h"
#include "latch.h"
#include "lsn.h"
#include "bf.h"
#include "file.h"
#include "fi_macro.h"
#include "link.h"
#include "volume.h"
#include "trans.h"
#include "consist.h"
#include "fi_intfuncs.h"
#include "fi_extfuncs.h"
#include "bf_extfuncs.h"


/*============================================================================*/
/* 
 * FI_GetAfter
 *	Looks for the PID immediately AFTER the one specified by the path
 *	in the pageStack; returns it in 'afterpid'
 *
 * IMPORTS:
 *	bf_UnfixPage()
 *	fi_FirstPage()
 *	FI_StackEmpty()
 *	FI_PopStack()
 *	FI_UnfixStack()
 *
 * RETURNS:
 *	esmNOERROR
 *	efiAFTERLASTPID		Can't find a page the page needed
 *				This means we are scanning after the last PID
 *				in the file.
 *	Errors returned by fi_First()
 *
 * ERROR RETURNS:
 *	None
 *
 * SIDE EFFECTS:
 *	returns the found page in 'afterPid'
 */
/*============================================================================*/


 int
FI_GetAfter (

	BUFGROUP	*bufGroup,
	VOLID		volid,
	PAGESTACK	*pageStack,
	SHORTPID	*afterPid,
	TREEDATA	*resultLeaf,
	int			*resultCount
)
{

    STACKNODE	*popNode;		/* node on path Stack					*/
    NODEPAGE	*popPage;		/* buffer page popped from path stack	*/
    NODEPAGE	*lcaPage;		/* least common ancestor buffer page	*/
    TREEDATA	*lcaArray;		/* least common ancestor page array		*/
    PID			popPid;			/* Pid of popped page from path stack	*/
    PID         lcaPid;			/* pid of least common ancestor page	*/
    PID			subTreePid;		/* the pid of the subtree to search		*/
    FOUR		popPath;		/* path/subscript taken in popped		*/
    FOUR		popNodeCount;	/* node count of popped page			*/
    FOUR		lcaPath;		/* path/subscript taken in lca			*/


	TRACE(TR_FI, TR_LEVEL_1);
#ifdef DEBUG
	if( resultLeaf == NULL) {
		SM_ASSERT(LEVEL_1, (resultCount == NULL));
	}else{
		SM_ASSERT(LEVEL_1, resultCount != NULL);
	}
#endif DEBUG

    /* 
     * Grab volume Id 
     */
    popPid.volid = lcaPid.volid = subTreePid.volid = volid;

    /*
     * Pop path stack, grab path, popPid, buffer page, node count
     */
    if ((popNode = FI_PopStack(pageStack)) == NULL)	{

		/* 
		 * If path Stack is empty, return efiAFTERLASTPID
		 * - scanning after last PID in file.
		 */
		*afterPid = NULLPID;
		SM_ERROR(TYPE_USER, esmAFTERLASTPID);
		return(esmFAILURE);
	}

	/*
	 *	Grab the values out of the node
	 */
    popPath		 = popNode->path;
    popPid.page = popNode->pageID;
    popPage		 = popNode->nodePage;
    popNodeCount = popPage->header.nodeCount;

    /*
     * Traverse UP tree(by popping stack) until reach page where the path taken
     * is NOT the last(or path stack is empty): 
     *		Unfix current popped page
     *		Pop another page from page stack
     *		Grab path, popPid, buffer page, node count
     */
    while ( (!FI_StackEmpty(pageStack)) && (popPath == (popNodeCount - 1)) )  {

		/*
		 *	Unfix the current page
		 */
		signalSemaphore(&(popNode->groupLink->pageHash->semaphore));
		bf_UnfixPage(popNode->groupLink, BF_DEFAULT, FALSE);

		/*
		 *	Pop the next element off the stack
		 */
        popNode = FI_PopStack(pageStack);

		/*
		 *	Pull the values out of the node
		 */
        popPath		 = popNode->path;
        popPid.page = popNode->pageID;
        popPage		 = popNode->nodePage;
        popNodeCount = popPage->header.nodeCount;
    }

    /*
     *	If path stack empty before found a page with
	 *	path != last then we are after last pid in file
     */
    if (popPath == (popNodeCount - 1))  {

		/*
		 *	Unfix the last page
		 */
		signalSemaphore(&(popNode->groupLink->pageHash->semaphore));
		bf_UnfixPage(popNode->groupLink, BF_DEFAULT, FALSE);

		/*
		 *	return appropriate error code
		 */
		*afterPid = NULLPID;
		SM_ERROR(TYPE_USER, esmAFTERLASTPID);
		return(esmFAILURE);

    } else {

		/*
		 * Found a page with path != last
		 *
		 * Grab least common ancestor Pid, Page, Path, array.
		 * Start search on subtree to the right of least common ancestor
		 */
		lcaPid	 = popPid;
		lcaPage	 = popNode->nodePage;
		lcaPath	 = popPath;
		lcaArray = &(lcaPage->contents);
		subTreePid.page = lcaArray->keyPtr[lcaPath+1].pagePtr;

		/* 
		 * Get the first Pid/key in the subtree
		 * Check for errors returned by search
		 */
		if (fi_FirstPhysicalPage(bufGroup, &subTreePid, afterPid,
			resultLeaf, resultCount))	{

			return(esmFAILURE);
		}

		/*
		 *	Unfix least common ancestor page
		 */
		signalSemaphore(&(popNode->groupLink->pageHash->semaphore));
		bf_UnfixPage(popNode->groupLink, BF_DEFAULT, FALSE);

		/*
		 * unfix all pages above it
		 */
		FI_UnfixStack(pageStack, FALSE);
		TRPRINT(TR_FI, TR_LEVEL_2, ("target pid:%d",*afterPid));
    }

	/*
	 *	return success to the caller
	 */
    return(esmNOERROR);
}
