/*
 *   $RCSfile: FI_GetBefore.c,v $  
 *   $Revision: 1.1.1.1 $  
 *   $Date: 1996/05/04 21:55:41 $      
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/


#include "sysdefs.h"
#include "ess.h"
#include "checking.h"
#include "error.h"
#include "trace.h"
#include "io.h"
#include "page.h"
#include "list.h"
#include "tid.h"
#include "lock.h"
#include "object.h"
#include "msgdefs.h"
#include "thread.h"
#include "semaphore.h"
#include "latch.h"
#include "lsn.h"
#include "bf.h"
#include "file.h"
#include "fi_macro.h"
#include "link.h"
#include "volume.h"
#include "trans.h"
#include "consist.h"
#include "fi_intfuncs.h"
#include "fi_extfuncs.h"
#include "bf_extfuncs.h"


/*============================================================================*/
/* 
 * FI_GetBefore
 *
 * 	Gets PID BEFORE the one specified by the path taken on the pageStack;
 *	returns it in 'beforePid'.
 *
 * IMPORTS:
 *
 *	bf_UnfixPage()
 *	fi_LastPhysicalPage()
 *	FI_StackEmpty()
 *	FI_PopStack()
 *	FI_UnfixStack()
 *
 * RETURNS:
 *
 *	esmNOERROR
 *	esmBEFOREFIRSTPID	Can't find the page
 *						This means we are scanning before the first
 *						PID in the file.
 *
 *	Errors returned by fi_LastPhysicalPage()
 *
 * ERROR RETURNS:
 *	None
 *
 * SIDE EFFECTS:
 *	Returns the found page in 'beforePid'
 */
/*============================================================================*/


 int
FI_GetBefore (

	BUFGROUP	*bufGroup,
	VOLID		volid,
	PAGESTACK	*pageStack,
	SHORTPID	*beforePid,
	TREEDATA	*resultLeaf,
	int			*resultCount
)
{

    STACKNODE	*popNode;		/* node on path stack						*/
    NODEPAGE	*lcaPage;		/* least common ancestor buffer page		*/
    TREEDATA	*lcaArray; 		/* least common ancestor key/ptr array		*/
    PID			popPid;			/* pid of popped page						*/
    PID			lcaPid;			/* pid of lca page							*/
    PID			subTreePid;		/* pid of subtree we want to search			*/
    FOUR		popPath;		/* path/subscript taken in popped page		*/
    FOUR		lcaPath;	 	/* path/subscript taken in lca page to reach*/


	TRACE(TR_FI, TR_LEVEL_1);
#ifdef DEBUG
	if( resultLeaf == NULL) {
		SM_ASSERT(LEVEL_1, (resultCount == NULL));
	}else{
		SM_ASSERT(LEVEL_1, resultCount != NULL);
	}
#endif DEBUG

    /*
     * Grab volume ID 
     */
    popPid.volid = lcaPid.volid = subTreePid.volid = volid;

    /*
     *	Pop path stack, grab buffer page, path, Pid
     */
    if ((popNode = FI_PopStack(pageStack)) == NULL)	{

		/*
		 *	If path stack empty, return esmBEFOREFIRSTPID
		 *	- scanning before first pid in file.
		 */
		*beforePid = NULLPID;
		SM_ERROR(TYPE_USER, esmBEFOREFIRSTPID);
		return(esmFAILURE);
	}

	/*
	 *	Grab information out of node
	 */
    popPath		 = popNode->path;
    popPid.page = popNode->pageID;

    /*
     * Traverse UP tree(by popping path stack) until reach page where the path 
     * taken is NOT the first(or path stack is empty):
     *		Unfix current popped page
     *		Pop another page from path stack
     *		Grab path, popPid, buffer page
     */
    while ((!FI_StackEmpty(pageStack)) && (popPath == 0))  {

		/*
		 *	Unfix the current page
		 */
		signalSemaphore(&(popNode->groupLink->pageHash->semaphore));
		bf_UnfixPage(popNode->groupLink, BF_DEFAULT, FALSE);

		/*
		 *	get the next node off the stack
		 */
        popNode = FI_PopStack(pageStack);

		/*
		 *	Pull the values out of the node
		 */
        popPid.page = popNode->pageID;
        popPath		= popNode->path;
    }

    /*
     *	If path stack empty before find a page with path != first
     *	scanning before first pid in file
     */
    if (popPath == 0)  {
		
		/*
		 *	Unfix the current page
		 */
		signalSemaphore(&(popNode->groupLink->pageHash->semaphore));
		bf_UnfixPage(popNode->groupLink, BF_DEFAULT, FALSE);

		/*
		 *	Return appropriate values
		 */
		*beforePid = NULLPID;
		SM_ERROR(TYPE_USER, esmBEFOREFIRSTPID);
		return(esmFAILURE);

    } else {

		/* 
		 *	Found a page with path != first
		 *
		 *	Grab least common ancestor pid, page, path, array.
		 *	Start search on subtree to the left of lca 
		 */
		lcaPid	 = popPid;
		lcaPage	 = popNode->nodePage;	/* = popPage */
		lcaPath	 = popPath;
		lcaArray = &(lcaPage->contents);
		subTreePid.page = lcaArray->keyPtr[lcaPath-1].pagePtr;

		/* 
		 *	Get last pid/key in the subtree
		 *	Check for errors returned by search
		 */
		if (fi_LastPhysicalPage(bufGroup, &subTreePid, 
						beforePid, resultLeaf, resultCount))	{

			return(esmFAILURE);
		}

		/*
		 *	unfix lca page
		 */
		signalSemaphore(&(popNode->groupLink->pageHash->semaphore));
		bf_UnfixPage(popNode->groupLink, BF_DEFAULT, FALSE);

		/*
		 *	unfix all the pages in the path stack.
		 */
		FI_UnfixStack(pageStack, FALSE);
    }

	/*
	 *	Return success
	 */
    return(esmNOERROR);
}
