/*
 *   $RCSfile: FI_InsertToArray.c,v $  
 *   $Revision: 1.1.1.1 $  
 *   $Date: 1996/05/04 21:55:41 $      
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/


#include "sysdefs.h"
#include "ess.h"
#include "checking.h"
#include "error.h"
#include "trace.h"
#include "io.h"
#include "page.h"
#include "list.h"
#include "tid.h"
#include "lock.h"
#include "object.h"
#include "msgdefs.h"
#include "thread.h"
#include "semaphore.h"
#include "latch.h"
#include "link.h"
#include "lsn.h"
#include "bf.h"
#include "file.h"
#include "fi_macro.h"
#include "consist.h"
#include "volume.h"
#include "util_funcs.h"
#include "fi_intfuncs.h"


/*============================================================================*/
/* 
 * FI_InsertToArray
 *    Inserts the <pid, pagePtr> pair at location 'insertLoc' in the array
 *     within the page at 'insertPage'.
 *
 * NOTE:
 *	This routine assumes that the array is NOT FULL.
 *
 *	For non leaves, the <pid, pagePtr> is inserted into the 
 *	array. For leaves, only the <pid> is inserted.
 *
 *	'insertLoc' CAN exceed the subscript of the last entry in the 
 *	array. This will cause the pair to be inserted AFTER all the 
 *	entries, i.e. at the end.
 *
 * IMPORTS:
 *	FI_StackEmpty()
 *	FI_Leaf()
 *	FI_PopStack()
 *	FI_ReleasePages()
 *	FI_StackInit()
 *
 * RETURNS:
 *	esmNOERROR
 *
 * SIDE EFFECTS:
 *	The nodeCount of the array is incremented.
 *
 */
/*============================================================================*/


 void
FI_InsertToArray (

	NODEPAGE			*insertPage,
	FOUR				insertLoc,
	SHORTPID			key,
	SHORTPID			pagePtr,
	BOOL				lobjPresent 
)
{

    register TREEDATA	*tempArray;
    FOUR				nodeCount;


	TRPRINT(TR_FI, TR_LEVEL_1, ("inserting key:%d ptr:%d at location:%d",
			key, pagePtr, insertLoc));

	/*
	 *	Check to see if the parameters are correct
	 */
	SM_ASSERT(LEVEL_3, (insertPage != NULL) && (key != NULLPID));

	/*
	 *	check for valid insert location
	 */
    SM_ASSERT(LEVEL_3, insertLoc >= 0);

    /*
     * Grab count, keys array
     */
    tempArray = &(insertPage->contents);
	nodeCount = insertPage->header.nodeCount;

    /*
     *	Separate cases for leaves and non leaves:
     * 	Move everything at and behind the insert location 1 position
     *  Insert the new keyPtr pair and increment the node count
     */
    if (FI_NotLeaf(insertPage->header.nodeType))  {

		TRPRINT(TR_FI, TR_LEVEL_2, ("inserting to non-leaf"));

		/*
		 * BCOPY does the equivalent of:
		 *    for (i = nodeCount - 1; i >= insertLoc; --i) {
		 *        tempArray->keyPtr[i+1] = tempArray->keyPtr[i];
		 *    }
		 */
		bcopy( (char *) &(tempArray->keyPtr[insertLoc].key),
			   (char *) &(tempArray->keyPtr[insertLoc + 1].key),
			   (int) ((nodeCount - insertLoc) * KEYPTRSIZE) );


		/*
		 * insert new key/ptr pair
		 */
		tempArray->keyPtr[insertLoc + 1].pagePtr = pagePtr;
		tempArray->keyPtr[insertLoc].key		 = key;

    } else {

		TRPRINT(TR_FI, TR_LEVEL_2, ("inserting to leaf"));

		/*
		 * BCOPY does the equivalent of:
		 *    for (i = nodeCount - 1; i >= insertLoc; --i) {
		 *      tempArray->mapKeys.allKeys[i+1] = tempArray->mapKeys.allKeys[i];
		 *    }
		 */
		bcopy( (char *) &(tempArray->keyPid.allKeys[insertLoc]),
			   (char *) &(tempArray->keyPid.allKeys[insertLoc + 1]),
			   (int) ((nodeCount - insertLoc) * KEYSIZE) );

		/*
		 * Is a large object present on the page
		 */
		if (lobjPresent == TRUE)  {

			/*
			 *	Set the high bit to indicate that the page has large object
			 */
			tempArray->keyPid.allKeys[insertLoc].key = (key|SETLARGE);

		} else {

			/*
			 *	put in the unmarked key
			 */
			tempArray->keyPid.allKeys[insertLoc].key = key;
		}
    }

    /*
     *	Increment the node count
     */
    insertPage->header.nodeCount += 1;
}
