/*
 *   $RCSfile: FI_SearchNode.c,v $  
 *   $Revision: 1.1.1.1 $  
 *   $Date: 1996/05/04 21:55:41 $      
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/


#include "sysdefs.h"
#include "ess.h"
#include "checking.h"
#include "error.h"
#include "trace.h"
#include "io.h"
#include "page.h"
#include "list.h"
#include "tid.h"
#include "lock.h"
#include "object.h"
#include "msgdefs.h"
#include "thread.h"
#include "semaphore.h"
#include "latch.h"
#include "link.h"
#include "lsn.h"
#include "bf.h"
#include "file.h"
#include "fi_macro.h"
#include "consist.h"
#include "volume.h"
#include "fi_intfuncs.h"
#include "fi_extfuncs.h"
#include "bf_extfuncs.h"


/*============================================================================*/
/* 
 * FI_SearchNode
 *
 * 	Searches the file B tree pointed to by fid for the key which is
 *	a SHORTPID. The path taken to get to the leaf node is kept tracked 
 *	by storing all visited nodes(their PIDs and the path from that page
 *	to the child node) in a stack. In this case, all the nodes EXCEPT the 
 *	one that contained the search key will be on the stack.
 *
 * IMPORTS:
 *      bf_ReadPage()
 *		FI_Leaf()
 *		FI_Root()
 *		FI_PushStack()
 *
 * RETURNS:
 *	esmNOERROR
 *
 * SIDE EFFECTS:
 *	pageStack	will contain the path from ROOTP to the found LEAFP
 *	foundNode	will contain the PID of the found page/node
 *	foundPage	will contain the buffer address of the found node/page
 *	foundLoc	will contain the position/subscript of the search key,
 *			or if the search key is NOT in the tree, the subscript
 *			of the key if it WAS in the tree
 */
/*============================================================================*/


 int
FI_SearchNode (

	BUFGROUP			*bufGroup,
	PID					*rootPid,
	SHORTPID			key,
	PAGESTACK			*pageStack,
	SHORTPID			*foundPid,
	GROUPLINK			**foundLink,
	FOUR				*foundLoc 
)
{

    register TREEDATA	*pageArray;		/* array of keys */
    register SHORTPID	lastKey;		/* last, middle key in page */
    register SHORTPID	midKey;
    register FOUR		mid;			/* indices for Binary search */
    register FOUR		first;
    register FOUR		last;
    NODEPAGE			*nodePage;		/* current page handled */
    FOUR				nodeCount;		/* number of keys in node to do B search on */
	GROUPLINK			*groupLink;		
    int					nodeType;		/* node Type */
    PID					readPageID;		/* PID of page to read */


	TRPRINT(TR_FI, TR_LEVEL_1, ("searching for page:%d", key));

    /*
     *	Read in first page of file and grab node type
     */
    if ((groupLink = bf_ReadPage(bufGroup, rootPid, FILE_PAGE2SIZE, BF_SEM)) == NULL)	{

		return(esmFAILURE);
    }

	/*
	 *	Get the current page out of the grouplink
	 */
	nodePage = (NODEPAGE *) groupLink->bufFrame;

	/*
	 *	Get the node type into a variable
	 */
	nodeType = nodePage->header.nodeType;

    /*
     *	Check if first page read in is ROOT PAGE
     */
    SM_ASSERT(LEVEL_3, FI_Root(nodeType));

    /*
     *	Initialize array pointer, # of keys, flag 
     */
    pageArray = &(nodePage->contents);
    nodeCount = nodePage->header.nodeCount;

	/*
	 *	Set the readpage value
	 */
	readPageID = *rootPid;

    /*
     *	Search nodes until leaf is reached
	 *
	 *	(The structure of this loop must be changed - DMH)
     */
    for (;;)	{

		/*
		 *	print the contents of the node
		 */
		fi_PrintNode(nodePage);

		/*
		 *	check the magic number of the page
		 */
		if (nodePage->header.magic != FILEPAGE_MAGIC)	{

			SM_ERROR(TYPE_FATAL, esmINTERNAL);
		}

		/*
		 *	If reached a non leaf, decrement nodeCount to do binary search 
		 *	on because non-leaf nodes have 1 less key
		 */
		if (FI_NotLeaf(nodeType))  {

			--nodeCount;
		}

		/*
		 *	Initialize Binary search indices 
		 */
		first = 0;
		last  = nodeCount - 1;
		mid   = (first+last)/2;
			
		/*
		 *	Check if any keys in node
		 */
		if (nodeCount > 0)  {

			/*
			 *	Get the middle and last keys on the page
			 */
			if (FI_Leaf(nodeType))  {

				midKey  = (pageArray->keyPid.allKeys[mid].key & LARGEMASK);
				lastKey = (pageArray->keyPid.allKeys[last].key & LARGEMASK);

			} else {

				midKey  = pageArray->keyPtr[mid].key;
				lastKey = pageArray->keyPtr[last].key;
			}

			/*
			 *	If the searchkey is larger than the last key entry in array, 
			 *	don't bother doing a binary search, use subscript to last pointer.
			 */
			if (key > lastKey)  {

				mid = nodeCount;

			} else {

				/*
				 ********************
				 * do BINARY SEARCH *
				 ********************
				 *
				 * do until first and last 'cross over' 
				 */
				while (last > first)  {

					/*
					 *	check to see where the target key resides
					 *	
					 *	(Put in equal check - DMH)
					 */
					if (key <= midKey) {

						/*
						 *	its in the first half of the search space
						 */
						last = mid;

					} else {

						/*
						 *	Its in the 2nd half of the search space
						 */
						first = mid + 1;
					}
			
					/*
					 *	Calculate the new mid point
					 */
					mid = (first+last)/2;

					/*
					 *	Grab the middle key
					 */
					if (FI_Leaf(nodeType))  {

						midKey = (pageArray->keyPid.allKeys[mid].key & LARGEMASK);

					} else {

						midKey = pageArray->keyPtr[mid].key;
					}
				}
			}
		}

		/*
		 *	we're done if we've reached a leaf node
		 */
		if (FI_Leaf(nodeType))  {

			/*
			 *	Exit the loop
			 */
			break;

		} else {

			/* 
			 *	Otherwise, push onto stack the following information:
			 *		the page #
			 *		the *page address and.....
			 *		the 'path' taken from this page, i.e. the array subscript
			 */
			FI_PushStack(pageStack, readPageID.page, nodePage, groupLink, mid, NOFLAGS);

			/*
			 *	Prepare to read the page the link points to.
			 */
			readPageID.page = pageArray->keyPtr[mid].pagePtr;

			/*
			 *	read the page in and grab the array address, # of keys, node type
			 */
    		if ((groupLink = bf_ReadPage(bufGroup, &readPageID,
										 FILE_PAGE2SIZE, BF_SEM)) == NULL)	{

				return(esmFAILURE);
			}

			/*
			 *	Setup the variables
			 */
			nodePage = (NODEPAGE *) groupLink->bufFrame;
			pageArray = &(nodePage->contents);
			nodeCount = nodePage->header.nodeCount;
			nodeType  = nodePage->header.nodeType;
		}
    }

    /*
     *	return the insert location, page number, and *page address
     */
    *foundLoc  = mid;
    *foundLink = groupLink;
    *foundPid  = readPageID.page;
	TRPRINT(TR_FI, TR_LEVEL_2, ("location:%d", *foundLoc));
    return(esmNOERROR);
}
