/*
 *   $RCSfile: FI_SplitNode.c,v $  
 *   $Revision: 1.1.1.1 $  
 *   $Date: 1996/05/04 21:55:42 $      
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/


#include "sysdefs.h"
#include "ess.h"
#include "checking.h"
#include "error.h"
#include "trace.h"
#include "io.h"
#include "page.h"
#include "list.h"
#include "pool.h"
#include "tid.h"
#include "lock.h"
#include "object.h"
#include "msgdefs.h"
#include "thread.h"
#include "semaphore.h"
#include "latch.h"
#include "lsn.h"
#include "bf.h"
#include "file.h"
#include "fi_macro.h"
#include "link.h"
#include "volume.h"
#include "trans.h"
#include "consist.h"
#include "openlog.h"
#include "logrecs.h"
#include "bitvec.h"
#include "bf_macro.h"
#include "bf_extfuncs.h"
#include "io_extfuncs.h"
#include "lm_extfuncs.h"
#include "fi_intfuncs.h"
#include "fi_extfuncs.h"
#include "log_extfuncs.h"
#include "trans_extfuncs.h"
#include "msg_funcs.h"
#include "thread_globals.h"


 int
FI_SplitNode (

	PID			*rootPid,
	BUFGROUP	*bufGroup,
	PAGESTACK	*parentStack,
	GROUPLINK	*splitLink,
	PID			*splitPid 
)
{

	GROUPLINK	*rightLink;
	GROUPLINK	*leftLink;
	GROUPLINK	*parentLink;
	NODEPAGE	*splitPage;
	NODEPAGE	*leftPage;
	NODEPAGE	*rightPage;
	NODEPAGE	*parentNode;
	TREEDATA	*rightArray;
	TREEDATA	*leftArray;
	SHORTPID	midKey;
	PID			leftPid;
	PID			rightPid;
	PID			allocPid;
	int			midPoint;
	int			parentSlot;
	int			splitType;
	LSNOFFSET	saveLSN;
	VOLREC		*volRec;


	TRACE(TR_FI, TR_LEVEL_1);

	/*
	 *	get the volume record
	 */
	if ((volRec = io_FindVolRec(rootPid->volid)) == NULL) {
		reply(esmFAILURE, Active->errno, 1, TRUE);
	}

	/*
	 *	get a pointer to the split page
	 */
	CHECK_GROUPLINK_MAGIC(splitLink);
	splitPage = (NODEPAGE *) splitLink->bufFrame;

	/*
	 *	save the current lsn
	 */
	saveLSN = ((TRANSREC *) Active->transRec)->lastLSN;
	TRPRINT(TR_FI|TR_LOG, TR_LEVEL_2, ("saveLSN:%d", saveLSN));
    
	/*
	 *	'splitPage' is gonna split into 2 new pages.
	 *	Allocate 2 new pages for children
	 */
	if (io_AllocPages(volRec, FILE_PAGE2SIZE, 1, splitPid, &allocPid)){
		return(esmFAILURE);
	}

	/*
	 * Get their allocated Pids
	 */
	rightPid = allocPid;
	TRPRINT(TR_FI, TR_LEVEL_2, ("rightPid:%d", rightPid.page));

	/*
	 *	Get a new buffer for right child 
	 *	For left child, re-use split(if it isn't ROOT) node's 
	 *	buffer (re-use it's contents too).
	 */
	if ((rightLink = bf_ReadPage(bufGroup, &rightPid,
								 FILE_PAGE2SIZE, BF_SEM|BF_NOREAD)) == NULL)	{

		/*
		 *	undo the allocation of the pages
		 */
		undoTransaction(saveLSN, NULL_LSN, NOFLAGS);

		/*
		 *	return failure
		 */
		return(esmFAILURE);
	}

	/*
	 *	get the pointer to the page
	 */
	rightPage = (NODEPAGE *) rightLink->bufFrame;
	splitType = splitPage->header.nodeType;

	/*
	 *	Log the format of the new page, basically this meens
	 *	initializing the page's lrc.
	 */
	if (FI_LogRootFormat(&rightPid, rightLink->pageHash))  {

		/*
		 *	notify waiters that the page is gone
		 */
		notify( &(rightLink->pageHash->semaphore.waitList), esmFAILURE, esmPAGEGONE );

		/*
		 *	reinitialize the semaphore
		 */
		initializeSemaphore( &(rightLink->pageHash->semaphore), TRUE );

		/*
		 *	blow the page out of the cache
		 */
		bf_DiscardPage(bufGroup, rightLink);

		/*
		 *	roll back the transaction to the save point
		 */
		undoTransaction(saveLSN, NULL_LSN, NOFLAGS);

		/*
		 *	notify the user of the error
		 */
		return(esmFAILURE);	
	}

	/* 
	 *	if *ROOT* overflows, getBuf for a 2nd time because
	 *	we want ROOT to remain the same page always so the
	 *	fileID doesn't change
	 *
	 *	if the split page is ROOT
	 *		do a 2nd GetBuf for leftPage
	 *		copy ROOT into this new page buffer
	 *	else
	 *		reuse the split page's buffer
	 *		rename that buffer to be left's buffer
	 */
	if (FI_Root(splitType))	{

		TRPRINT(TR_FI, TR_LEVEL_2, ("leaf is also root"));
    
		/*
		 *	'splitPage' is gonna split into 2 new pages.
		 *	Allocate 2 new pages for children
		 */
		if (io_AllocPages(volRec, FILE_PAGE2SIZE, 1, splitPid, &allocPid))	{

			/*
			 *	notify waiters that the page is gone
			 */
			notify( &(rightLink->pageHash->semaphore.waitList), esmFAILURE, esmPAGEGONE );

			/*
			 *	reinitialize the semaphore
			 */
			initializeSemaphore( &(rightLink->pageHash->semaphore), TRUE );

			/*
			 *	blow the page out of the cache
			 */
			bf_DiscardPage(bufGroup, rightLink);

			/*
			 *	roll back the transaction to the save point
			 */
			undoTransaction(saveLSN, NULL_LSN, NOFLAGS);

			/*
			 *	notify the user of the error
			 */
			return(esmFAILURE);	
		}

		/*
		 *	get left pid
		 */
		leftPid = allocPid;
		TRPRINT(TR_FI, TR_LEVEL_2, ("leftPid:%d", leftPid.page));

		/*
		 *	Allocate buffer
		 */
		if ((leftLink = bf_ReadPage(bufGroup, &leftPid,
									FILE_PAGE2SIZE, BF_SEM|BF_NOREAD)) == NULL)	{

			/*
			 *	notify waiters that the page is gone
			 */
			notify( &(rightLink->pageHash->semaphore.waitList), esmFAILURE, esmPAGEGONE );

			/*
			 *	reinitialize the semaphore
			 */
			initializeSemaphore( &(rightLink->pageHash->semaphore), TRUE );

			/*
			 *	blow the page out of the cache
			 */
			bf_DiscardPage(bufGroup, rightLink);

			/*
			 *	roll back the transaction to the save point
			 */
			undoTransaction(saveLSN, NULL_LSN, NOFLAGS);

			/*
			 *	notify the user of the error
			 */
			return(esmFAILURE);	
		}

		/*
		 *	Get a pointer to the actual data page
		 */
		leftPage = (NODEPAGE *) leftLink->bufFrame;

		/*
		 *	Log the format of the new page, basically this meens
		 *	initializing the page's lrc.
		 */
		if (FI_LogRootFormat(&leftPid, leftLink->pageHash))  {

			/*
			 *	notify waiters that the page is gone
			 */
			notify( &(rightLink->pageHash->semaphore.waitList), esmFAILURE, esmPAGEGONE );

			/*
			 *	reinitialize the semaphore
			 */
			initializeSemaphore( &(rightLink->pageHash->semaphore), TRUE );

			/*
			 *	blow the page out of the cache
			 */
			bf_DiscardPage(bufGroup, rightLink);

			/*
			 *	roll back the transaction to the save point
			 */
			undoTransaction(saveLSN, NULL_LSN, NOFLAGS);

			/*
			 *	notify the user of the error
			 */
			return(esmFAILURE);	
		}

		/*
		 *	Couldn't we just shift buffer pointers
		 */
		leftPage->header.nodeCount = splitPage->header.nodeCount;
		leftPage->contents		   = splitPage->contents;
		leftPage->header.magic	   = FILEPAGE_MAGIC;

		/*
		 *	setup the new node types
		 */
		if (FI_Leaf(splitType))	{

			leftPage->header.nodeType = rightPage->header.nodeType = FI_LEAFPAGE;

		} else {

			leftPage->header.nodeType = rightPage->header.nodeType = FI_INTPAGE;
		}

		/*
		 *	new root will always be internal
		 */
		splitPage->header.nodeType = FI_ROOTPAGE|FI_INTPAGE;

	} else {

		TRPRINT(TR_FI, TR_LEVEL_2, ("leaf is not root"));

		/*
		 *	Use the existing page.
		 */
		leftPage = splitPage;
		leftLink = splitLink;

		/*
		 *	setup new node type
		 */
		if (FI_Leaf(splitType))	{
		
			rightPage->header.nodeType = FI_LEAFPAGE;

		} else {

			rightPage->header.nodeType = FI_INTPAGE;
		}

		/*
		 *	the left pid will be the split pid
		 */
		leftPid = *splitPid;
	}

	/*
	 *	Grab the 2 children arrays 
	 */
	leftArray = &(leftPage->contents);
	rightArray = &(rightPage->contents);

	/*
	 *	distribute the contents
	 */
	if (FI_Leaf(splitType))	{

		/*
		 *	Midpoint of leaf page
		 */
		midPoint = FULLEAFNODE/2;
		midKey = (leftArray->keyPid.allKeys[midPoint - 1].key & LARGEMASK);
		TRPRINT(TR_FI, TR_LEVEL_2, ("midPoint:%d new mid key:%d", midPoint, midKey));

		/*
		 * copy 2nd half of left child into 1st half of right child
		 */
		bcopy( (char *) &(leftArray->keyPid.allKeys[midPoint]),
			   (char *) &(rightArray->keyPid.allKeys[0]),
			   (FULLEAFNODE - midPoint) * KEYSIZE );

		/*
		 *	Set half full for right leaf node and
		 *	the left new node will contain midPoint entries
		 *	no matter what its type
		 */
		rightPage->header.nodeCount = FULLEAFNODE - midPoint;

	} else {

		/*
		 *	Midpoint of non-leaf node
		 */
		midPoint = FULLNONLEAFNODE/2;
		midKey = leftArray->keyPtr[midPoint - 1].key;
		TRPRINT(TR_FI, TR_LEVEL_2, ("midPoint:%d new mid key:%d", midPoint, midKey));

		/*
		 * copy 2nd half of left child into 1st half of right child
		 */
		bcopy( (char *) &(leftArray->keyPtr[midPoint]),
			   (char *) &(rightArray->keyPtr[0]),
			   (FULLNONLEAFNODE - midPoint) * KEYPTRSIZE );

		/*
		 *	Set half full for non-leaf page
		 */
		rightPage->header.nodeCount = FULLNONLEAFNODE - midPoint;
	}

	/*
	 *	left node count will always be midpoint
	 */
	rightPage->header.magic = FILEPAGE_MAGIC;
	leftPage->header.nodeCount = midPoint;
	fi_PrintNode(leftPage);
	fi_PrintNode(rightPage);
	TRPRINT(TR_FI, TR_LEVEL_2, ("leftCount:%d rightCount:%d",
					leftPage->header.nodeCount, rightPage->header.nodeCount));

	/*
	 *	create a new root if the type split is a root
	 */
	if (FI_Root(splitType))	{

		TRPRINT(TR_FI, TR_LEVEL_2, ("split type is root"));

		/*
		 *	set up the header of the new root page
		 */
		splitPage->header.magic		= FILEPAGE_MAGIC;
		splitPage->header.nodeCount = 2;
		splitPage->header.nodeType	= FI_ROOTPAGE;

		/*
		 *	fill in the single entry in the root
		 */
		splitPage->contents.keyPtr[0].pagePtr = leftPid.page;
		splitPage->contents.keyPtr[0].key 	  = midKey;
		splitPage->contents.keyPtr[1].pagePtr = rightPid.page;
		splitPage->contents.keyPtr[1].key 	  = NIL;
		DIRTY_PAGE(splitLink->pageHash);
		fi_PrintNode(splitPage);

		/*
		 *	set the parent link for logging
		 */
		parentLink = splitLink;

	} else {

		TRPRINT(TR_FI, TR_LEVEL_2, ("split type not root"));

		/*
		 *	get a pointer to the parent
		 */
		parentLink = parentStack->stackNode[parentStack->stackTop].groupLink;
		parentSlot = (int) parentStack->stackNode[parentStack->stackTop].path;
		parentNode = (NODEPAGE *) parentLink->bufFrame;
		
		/*
		 *	insert the key to the above array
		 */
		FI_InsertToArray(parentNode, parentSlot, midKey, rightPid.page, FALSE);
		DIRTY_PAGE(parentLink->pageHash);
		fi_PrintNode(parentNode);
	}

	/*
	 *	log the pages
	 */
	if (FI_LogPageSplit(rootPid->volid, parentLink, leftLink, rightLink, saveLSN))	{

		SM_ERROR(TYPE_FATAL, esmINTERNAL);
	}

	/*
	 *	signal the right and left pages
	 */
	signalSemaphore( &(leftLink->pageHash->semaphore) );
	signalSemaphore( &(rightLink->pageHash->semaphore) );

	/*
	 *	unfix the right and left pages
	 */
	bf_UnfixPage(leftLink, BF_DEFAULT, TRUE);
	bf_UnfixPage(rightLink, BF_DEFAULT, TRUE);

	/*
	 *	return success
	 */
	return(esmNOERROR);
}
