/*
 *   $RCSfile: fi_CreateFile.c,v $  
 *   $Revision: 1.1.1.1 $  
 *   $Date: 1996/05/04 21:55:41 $      
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/

#include "sysdefs.h"
#include "ess.h"
#include "checking.h"
#include "error.h"
#include "trace.h"
#include "io.h"
#include "page.h"
#include "list.h"
#include "pool.h"
#include "tid.h"
#include "lock.h"
#include "object.h"
#include "msgdefs.h"
#include "thread.h"
#include "semaphore.h"
#include "latch.h"
#include "lsn.h"
#include "bf.h"
#include "file.h"
#include "fi_macro.h"
#include "link.h"
#include "volume.h"
#include "trans.h"
#include "consist.h"
#include "openlog.h"
#include "logrecs.h"
#include "deallocinfo.h"
#include "bf_extfuncs.h"
#include "io_extfuncs.h"
#include "lm_extfuncs.h"
#include "fi_intfuncs.h"
#include "fi_extfuncs.h"
#include "trans_extfuncs.h"
#include "log_extfuncs.h"
#include "io_globals.h"
#include "thread_globals.h"


/*============================================================================*/
/* 
 * fi_CreateFile
 * 	Creates a new file near "nearPid". 
 *
 * IMPORTS:
 *	bf_UnfixPage()
 *
 * RETURNS:
 *	esmNOERROR
 *
 * ERROR RETURNS:
 *	efiBADPARAMS	tid or near PID passed is NULL
 *	Errors returned by bf_GetBuf()
 *
 * SIDE EFFECTS:
 *	Returns a pointer to a new file (a PID) in 'returnFile'.
 */
/*============================================================================*/


 int
fi_CreateFile (

	VOLID				volid,
	BUFGROUP			*bufGroup,
	PID					*nearPid,
	FID					*fid,
	FLAGS				flags 
)
{

	GROUPLINK	*rootLink;		/* grouplink of root page of file	*/
    NODEPAGE	*rootPage;		/* buffer page for file				*/
	NODEHEADER	*header;
	LSNOFFSET	saveLSN;
    PID			pid;			/* page Id of file					*/
	VOLREC		*volRec;
	TRANSREC	*transRec;
	FILEDEALLOCINFO *fileDeallocInfo;
	LOCKMODE	lockMode;


	TRACE(TR_FI, TR_LEVEL_1);

	/*
	 *	save the current undo lsn
	 */
	transRec = (TRANSREC *) Active->transRec;
	saveLSN = transRec->lastLSN;
	TRPRINT(TR_FI, TR_LEVEL_2, ("saveLSN:%x", saveLSN));

	if ((volRec = io_FindVolRec(volid)) == NULL)	{
		SM_ERROR(TYPE_FATAL, esmINTERNAL);
	}
    /*
     *	Allocate 1st page of file (ROOT) page 
	 *	and keep the bitmap pages latched
     */
    if (io_AllocPages(volRec, FILE_PAGE2SIZE, 1, nearPid, &pid))	{

		return(esmFAILURE);
	}
	TRPRINT(TR_FI, TR_LEVEL_2, ("created file with root pid:%d", pid.page));

	/*
	 *	save the newly allocated fid
	 */
    fid->pid = pid;


	/* 	get the unique number for the fid */
	if (io_GetUnique( &pid, &(fid->unique), 1 ))	{
		SM_ERROR(TYPE_FATAL, esmINTERNAL);
	}

	/*
	 *	get the file lock
	 */
	lockMode = EX;
	if (lockFile((TRANSREC *) Active->transRec, fid, &lockMode, LOCK_NOWAIT))	{

		/*
		 *	roll back the transaction to the save point
		 */
		undoTransaction(saveLSN, NULL_LSN, NOFLAGS);

		/*
		 *	return error to user
		 */
		return(esmFAILURE);
	}

    /* 
     *	get a buffer for it
     */
    if ((rootLink = bf_ReadPage(bufGroup, &(fid->pid),
								 FILE_PAGE2SIZE, BF_NOREAD|BF_SEM)) == NULL)	{

		/*
		 *	roll back the transaction to the save point
		 */
		undoTransaction(saveLSN, NULL_LSN, NOFLAGS);

		/*
		 *	return failure to the caller
		 */
		return(esmFAILURE);
	}

	/*
	 *	get a pointer to the buffer
	 */
	rootPage = (NODEPAGE *) rootLink->bufFrame;

    /*
     *	Initialize page header information and set nodes to null
	 */
	header					 = &(rootPage->header);
	header->magic	  		 = FILEPAGE_MAGIC;
	header->nodeCount 		 = 0;
    header->nodeType  		 = FI_ROOTLEAFPAGE;
	header->firstLogicalNode = NULLPID;
	header->lastLogicalNode  = NULLPID;
	header->unique	  		 = fid->unique;
	header->flags	  		 = NOFLAGS;
	TRPRINT(TR_FI, TR_LEVEL_2, ("file unique:%x", fid->unique));

	SM_ASSERT(LEVEL_3, (volRec->header->properties & VOLPROP_TEMP) ==
			(volRec->volflags & VOL_TEMP));
	/*
	 *	set the type if necessary
	 */
	if (flags & LOGICAL_ORDER)	{
			
		rootPage->header.flags = LOGICAL_ORDER;
		SM_ASSERT(LEVEL_3, volRec->volflags & VOL_TEMP);
	}

	/*
	 *	dirty the page so that a checkpoint will not get a consistency error
	 */
	DIRTY_PAGE(rootLink->pageHash);

	/*
	 *	log the format of the page
	 */
	if (FI_LogRootFormat(&(fid->pid), rootLink->pageHash))	{

		/*
		 *	notify waiters that the page is gone
		 */
		notify( &(rootLink->pageHash->semaphore.waitList), esmFAILURE, esmPAGEGONE );

		/*
		 *	reinitialize the semaphore
		 */
		initializeSemaphore( &(rootLink->pageHash->semaphore), TRUE );

		/*
		 *	blow the page out of the cache
		 */
		bf_DiscardPage(bufGroup, rootLink);

		/*
		 *	roll back the transaction to the save point
		 */
		undoTransaction(saveLSN, NULL_LSN, NOFLAGS);

		/*
		 *	notify the user of the error
		 */
		return(esmFAILURE);	
	}

	/*
	 *	mark the page as a file header
	 */
	if (io_MarkPage(volRec, &(fid->pid), PAGE_FILEHDR, TRUE))	{

		/*
		 *	notify waiters that the page is gone
		 */
		notify( &(rootLink->pageHash->semaphore.waitList), esmFAILURE, esmPAGEGONE );

		/*
		 *	reinitialize the semaphore
		 */
		initializeSemaphore( &(rootLink->pageHash->semaphore), TRUE );

		/*
		 *	blow the page out of the cache
		 */
		bf_DiscardPage(bufGroup, rootLink);

		/*
		 *	roll back the transaction to the save point
		 */
		undoTransaction(saveLSN, NULL_LSN, NOFLAGS);

		/*
		 *	notify the user of the error
		 */
		return(esmFAILURE);	
	}

	/*
	 *	write a log record that will turn the create into a logical operation
	 */
	if (FI_LogFileCreateDone(fid, saveLSN))	{

		/*
		 *	notify waiters that the page is gone
		 */
		notify( &(rootLink->pageHash->semaphore.waitList), esmFAILURE, esmPAGEGONE );

		/*
		 *	reinitialize the semaphore
		 */
		initializeSemaphore( &(rootLink->pageHash->semaphore), TRUE );

		/*
		 *	blow the page out of the cache
		 */
		bf_DiscardPage(bufGroup, rootLink);

		/*
		 *	roll back the transaction to the save point
		 */
		undoTransaction(saveLSN, NULL_LSN, NOFLAGS);

		/*
		 *	notify the user of the error
		 */
		return(esmFAILURE);	
	}

	/*
	 *	signal the semaphore
	 */
	signalSemaphore( &(rootLink->pageHash->semaphore) );

    /*
     *	unfix the page
     */
    bf_UnfixPage(rootLink, BF_DEFAULT, TRUE);

	/*
	 *	If the file was allocated on a temporary volume, remember
	 *	it as one to destroy at end-of-trans.
	 */
	if (volRec->volflags & VOL_TEMP) {
		fileDeallocInfo = (FILEDEALLOCINFO*) poolDeq(&FileDeallocInfoPool);
		if (fileDeallocInfo == NULL) {
			SM_ERROR(TYPE_CRASH, esmINTERNAL);
		}
		fileDeallocInfo->fid = *fid;
		listEnq(&(transRec->tempFileList), &(fileDeallocInfo->list));
	}

    return(esmNOERROR);
}
