/*
 *   $RCSfile: fi_DeletePage.c,v $  
 *   $Revision: 1.1.1.1 $  
 *   $Date: 1996/05/04 21:55:41 $      
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/

#include "sysdefs.h"
#include "ess.h"
#include "checking.h"
#include "error.h"
#include "trace.h"
#include "io.h"
#include "page.h"
#include "list.h"
#include "tid.h"
#include "lock.h"
#include "object.h"
#include "msgdefs.h"
#include "thread.h"
#include "semaphore.h"
#include "latch.h"
#include "link.h"
#include "lsn.h"
#include "bf.h"
#include "file.h"
#include "fi_macro.h"
#include "consist.h"
#include "volume.h"
#include "trans.h"
#include "openlog.h"
#include "logrecs.h"
#include "logaction.h"
#include "log.h"
#include "util_funcs.h"
#include "fi_intfuncs.h"
#include "fi_extfuncs.h"
#include "bf_extfuncs.h"
#include "log_extfuncs.h"
#include "thread_globals.h"


/*============================================================================*/
/* 
 * fi_DeletePage
 * 	Deletes page 'pid' from the file B tree
 *
 * RETURNS:
 *	esmNOERROR
 *
 * ERROR RETURNS:
 *
 * SIDE EFFECTS:
 *	The key value 'pid' is deleted from the B tree.
 */
/*============================================================================*/


 int
fi_DeletePage (

	BUFGROUP			*bufGroup,
	FID					*fid,
	SHORTPID			pid,
	LSNOFFSET			compensationLSN
)
{

    PAGESTACK			*pathStack;		/* stack of path to delete page */
    SHORTPID			deletePid;		/* SHORTPID of page containing pid key to delete */
    PID					deleteFullPid;	/* PID of page containing pid key to delete */
    NODEPAGE			*deletePage;	/* buffer page address of delete page */
	GROUPLINK			*deleteLink;
    FOUR				deleteLoc;		/* delete location/subscript */
	FORCEMARK			forceMark;
	LSN					lsn;
	LOGRECORDINFO		recordInfo;
	LRC					*lrc;
	FILEKEYLOGINFO		fileKeyLogInfo;

	TRPRINT(TR_FI, TR_LEVEL_1, ("page:%d", pid));

	if (fi_ValidateFid(bufGroup, fid) != esmNOERROR)    {
		return(esmFAILURE);
	}

    /* Initialize path stack to NULL, set up available nodes list */
	if ((pathStack = FI_AllocStack(FALSE)) == NULL)	{
		return(esmFAILURE);
	}

    /*
     *	Search delete page using given key
     */
    if ((FI_SearchNode(bufGroup, &fid->pid, pid, pathStack, 
					   &deletePid, &deleteLink, &deleteLoc)))	{
		FI_UnfixStack(pathStack, TRUE);
		return(esmFAILURE);
	}
	deleteFullPid.page = deletePid;
	deleteFullPid.volid = fid->pid.volid;

	/*
	 *	get a pointer to the delete page
	 */
	deletePage = (NODEPAGE *) deleteLink->bufFrame;

    /*
     *	check if found page is leaf
     *	if not, unfix all pages on stack and exit
     */
    SM_ASSERT(LEVEL_3, FI_Leaf(deletePage->header.nodeType));
    
    /*
     *	check if found page to delete
     *	if not, unfix all pages on stack and exit
     */
    if (pid != (deletePage->contents.keyPid.allKeys[deleteLoc].key & LARGEMASK))  {

		/*
		 *	Unfix the leaf node
		 */
		signalSemaphore(&(deleteLink->pageHash->semaphore));
		bf_UnfixPage(deleteLink, BF_DEFAULT, FALSE);

		/*
		 *	Release the stack pages
		 */
		FI_UnfixStack(pathStack, TRUE);

		/*
		 *	return the error code
		 */
		SM_ERROR(TYPE_WARNING, esmFILEPAGENOTFOUND);
		return(esmFAILURE);
    }

	PRINT_PROGRESS(("Deleting page %d from file %d:%d\n", pid, fid->pid.page, (int) (fid->pid.volid) ));

    /*
     *	If delete page not about to underflow, do a simple delete, else
     *	delete and redistribute the keys amongst sibling pages
	 *	NOTE: balancing is a mess we do not get into at this time.
	 *
    if (deletePage->nodeCount > LEAFUNDERFLOW)  {
     */

		fileKeyLogInfo.fid = *fid;
		fileKeyLogInfo.largeObjOnPage = FALSE; /* since new page */

		/*
		 *	Log the deletion of a file btree leaf key
		 */
		lrc = &(deletePage->header.lrc);
		INCREMENT_LRC(lrc);
		/*
		 *  initialize the log information
		 */
		recordInfo.action       = LOG_ACTION_DELETE_FILE_KEY;
		if (compensationLSN == NULL_LSN) {
			recordInfo.type         = LOG_REC_TYPE_USER;
			recordInfo.nextUndoLSN  = NULL_LSN;
		} else {
			recordInfo.type         = LOG_REC_TYPE_COMPENSATION;
			recordInfo.nextUndoLSN  = compensationLSN;
		}
		recordInfo.imageCount   = 3;
		recordInfo.actionPid    = &deleteFullPid;
		recordInfo.actionLRC    = lrc;
		recordInfo.imageSize[0] = sizeof(deleteLoc);
		recordInfo.imageData[0] = (VOID *) &deleteLoc;
		recordInfo.imageSize[1] = sizeof(pid);
		recordInfo.imageData[1] = (VOID *) &pid;
		recordInfo.imageSize[2] = sizeof(fileKeyLogInfo);
		recordInfo.imageData[2] = (VOID *) &fileKeyLogInfo;
		recordInfo.flags        = NOFLAGS;

		/*
		 *  write the record to the log
		 */
		if ((forceMark = writeLogRecord((TRANSREC *) Active->transRec, &recordInfo, &lsn, NOFLAGS)) < 0)  {
			SM_ERROR(TYPE_FATAL, Active->errno);
		}
		DEPEND_LOG(deleteLink->pageHash, forceMark, &lsn, lrc);

		/*
		 *	Delete the page from the leaf
		 */
        FI_DeleteFromArray(deletePage, deleteLoc, pid);

		/*
		 *	print the contents of the node
		 */
		fi_PrintNode(deletePage);

		/*
		 *	Unfix and dirty the leaf page
		 */
		signalSemaphore(&(deleteLink->pageHash->semaphore));
		bf_UnfixPage(deleteLink, BF_DEFAULT, TRUE);

		/*
		 *	Unfix delete page and any other pages on stack
		 */
		FI_UnfixStack(pathStack, TRUE);

		/*
		 *	return success to the caller
		 */
    	return(esmNOERROR);

		/*
    } else  {

		 *	We must reshuffle the tree to maintain b-tree conditions
        FI_DeleteDistribute(bufGroup, rootPid->Pvolid, pathStack, 
                            deleteLink, deleteLoc, deletePid, pid);
        return(smNOERROR);
    }
		 */
}
