/*
 *   $RCSfile: fi_DestroyFile.c,v $  
 *   $Revision: 1.1.1.1 $  
 *   $Date: 1996/05/04 21:55:41 $      
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/


#include "sysdefs.h"
#include "ess.h"
#include "checking.h"
#include "error.h"
#include "trace.h"
#include "io.h"

#include "list.h"
#include "pool.h"
#include "tid.h"
#include "lock.h"
#include "object.h"
#include "msgdefs.h"
#include "thread.h"
#include "semaphore.h"
#include "latch.h"
#include "lsn.h"
#include "bf.h"
#include "file.h"
#include "fi_macro.h"
#include "consist.h"
#include "link.h"
#include "volume.h"
#include "fi_intfuncs.h"
#include "fi_extfuncs.h"
#include "bf_extfuncs.h"
#include "io_extfuncs.h"
#include "sm_extfuncs.h"


/*============================================================================*/
/* 
 * fi_DestroyFile
 * 	Destroys a file recursively
 *
 * RETURNS:
 *	esmNOERROR
 *
 * ERROR RETURNS:
 *	efiBADPARAMS
 *	Errors returned by bf_ReadPage()
 *
 * SIDE EFFECTS:
 *	Deallocates all the pages in the subtree and the pages pointed to
 *	by the pointers(PIDs) in the leaf pages.
 */
/*============================================================================*/


 int
fi_DestroyFile (

	BUFGROUP	*bufGroup,
	PID			*rootPid,
	LIST		*pageDeallocList,
	BOOL		shallow,			/* TRUE -> do not destroy large objects */
	BOOL		logDealloc			/* TRUE -> log the deallocations */
)
{
   
    NODEPAGE		*rootPage;
    TREEDATA		*rootArray;
	GROUPLINK		*groupLink;
	PID				childPid;
	PID				currentPid;
    FOUR			rootCount;
    int				i;


	TRPRINT(TR_FI, TR_LEVEL_1, ("rootPid:%d", rootPid->page));

    /*
     *	Grab volume Id 
     */
    currentPid.volid = childPid.volid = rootPid->volid;

    /*
     *	Read in root of subtree in.  No lock is needed since
	 *	the root has been locked by sm_DestroyFile.
     */
	if ((groupLink = bf_ReadPage(bufGroup, rootPid, FILE_PAGE2SIZE, NOFLAGS)) == NULL)	{

		return(esmFAILURE);
	}

	/*
	 *	get a pointer to the root page and check the magic number
	 */
	rootPage = (NODEPAGE *) groupLink->bufFrame;

	/*
	 *	check to see if the magic number is set
	 */
	if (rootPage->header.magic != FILEPAGE_MAGIC)	{

		SM_ERROR(TYPE_FATAL, esmINTERNAL);
	}

    /*
     *	Grab array, node count, node Type
     */
    rootArray = &(rootPage->contents);
    rootCount = rootPage->header.nodeCount;

    /*
     *	Check to see if the page is a leaf
     */
    if (FI_NotLeaf(rootPage->header.nodeType))  {

		TRPRINT(TR_FI, TR_LEVEL_2, ("non root"));

		/*
		 *	We must destroy all subtrees rooted here
		 */
		for (i = 0; i < rootCount; i++)  {

			/*
			 *	get the pid of the child
			 */
			childPid.page = rootArray->keyPtr[i].pagePtr;
				
			/*
			 *	destroy the subtree rooted at the child
			 */
			if (fi_DestroyFile(bufGroup, &childPid, pageDeallocList, shallow, logDealloc))	{
				
				return(esmFAILURE);
			}
		}

    } else {

		TRPRINT(TR_FI, TR_LEVEL_2, ("root page"));

		/* 
		 *	Deallocate pages pointed to by key values in leaf pages
		 *	First, check if there is a large object(s) on the leaf page
		 *	about to be destroyed.
		 */
		for (i = 0; i < rootCount; i++)  {

			/*
			 * Grab PID of leaf page
			 */
			currentPid.page = rootArray->keyPid.allKeys[i].key;

			/*
			 * Destroy the large objects on the page if we are
			 * not doing a shallow destroy and if a large object exists
			 * in the page 
			 */
			if ( !shallow && (currentPid.page & SETLARGE))	{

				/* take out the large bit */
				currentPid.page &= LARGEMASK;

				/*
				 *	Destroy all large objects on a page
				 */
				if (sm_DestroyPage(bufGroup, &currentPid, pageDeallocList))	{

					return(esmFAILURE);
				}

			} else {

				/* take out the large bit */
				currentPid.page &= LARGEMASK;

				if (pageDeallocList != NULL) {

					/*
					 *	Fake the deallocation -- add to list of pages
					 *	to deallocate
					 */
					if (io_FakeDeallocPages(SLOTTED_PAGE2SIZE, 1, &currentPid, pageDeallocList, PAGE_SLOTTED))	{

						return(esmFAILURE);
					}
				} else {

					/*
					 *	deallocate the slotted page
					 */
					if (io_DeallocPages(SLOTTED_PAGE2SIZE, 1, &currentPid, NULL, logDealloc, TRUE))	{

						return(esmFAILURE);
					}
					bf_InvalidatePage(&currentPid);
				}
			}
		}
    }

	if (pageDeallocList != NULL) {

		/*
		 *	deallocate the root of the subtree
		 *	Fake the deallocation -- add to list of pages
		 *	to deallocate
		 */
		bf_UnfixPage(groupLink, BF_DEFAULT, FALSE);
		if (io_FakeDeallocPages(FILE_PAGE2SIZE, 1, rootPid, pageDeallocList, PAGE_FILENODE))	{

			return(esmFAILURE);
		}

	} else {

		/*
		 *	destroy the root of the subtree
		 */
		bf_DiscardPage(bufGroup, groupLink);

		/*
		 *	deallocate the root of the subtree
		 */
		if (io_DeallocPages(FILE_PAGE2SIZE, 1, rootPid, NULL, logDealloc, TRUE))	{

			return(esmFAILURE);
		}
	}

	/*
	 *	return success to the caller
	 */
    return(esmNOERROR);
 }
