/*
 *   $RCSfile: fi_InsertPage.c,v $  
 *   $Revision: 1.1.1.1 $  
 *   $Date: 1996/05/04 21:55:41 $      
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/


#include "sysdefs.h"
#include "ess.h"
#include "checking.h"
#include "error.h"
#include "trace.h"
#include "io.h"
#include "page.h"
#include "list.h"
#include "pool.h"
#include "tid.h"
#include "lock.h"
#include "object.h"
#include "msgdefs.h"
#include "thread.h"
#include "semaphore.h"
#include "latch.h"
#include "lsn.h"
#include "bf.h"
#include "file.h"
#include "fi_macro.h"
#include "link.h"
#include "volume.h"
#include "trans.h"
#include "consist.h"
#include "openlog.h"
#include "logrecs.h"
#include "logaction.h"
#include "log.h"
#include "fi_intfuncs.h"
#include "fi_extfuncs.h"
#include "io_extfuncs.h"
#include "bf_extfuncs.h"
#include "log_extfuncs.h"
#include "thread_globals.h"


/*============================================================================*/
/* 
 * fi_InsertPage
 * 	Inserts a page into a file pointed to by fid.
 *
 * IMPORTS:
 *	bf_AddPageDependency()
 *	bf_GetBuf()
 *      bf_RenamePage()
 *	bf_UnfixPage()
 *	FI_StackEmpty()
 *	FI_Leaf()
 *	FI_PopStack()
 *	FI_ReleasePages()
 *	FI_StackInit()
 *
 * RETURNS:
 *	esmNOERROR
 *
 * ERROR RETURNS:
 *
 * SIDE EFFECTS:
 *	The <pagePtr,pid> is inserted into the file pointed to by fid
 */
/*============================================================================*/


 int
fi_InsertPage (

	BUFGROUP		*bufGroup,
	FID				*fid,
	SHORTPID		pid,
	BOOL			lobjPresent,
	SHORTPID		*predPid
)
{

    PAGESTACK		*pageStack;		/* stack to keep track of path to leaf	*/
	GROUPLINK		*insertLink;	/* pointer to link of found page		*/
	GROUPLINK		*lastLink;		/* pointer to link of found page		*/
	GROUPLINK		*rootLink;		/* pointer to link of found page		*/
	NODEPAGE		*insertPage;	/* node to insert key					*/
	NODEPAGE		*rootPage;		/* node to insert key					*/
	SLOTTEDPAGE		*slotPage;
    int		   		insertLoc;		/* subscript within insert page			*/
	LOGRECORDINFO   recordInfo;
	LRC             *rootLRC;
	LRC				tempLRC;
	LSN             lsn;
	FORCEMARK  		forceMark;
	TRANSREC		*transRec;

	TRPRINT(TR_FI, TR_LEVEL_1, ("inserting page:%d", pid));

	transRec = (TRANSREC*) Active->transRec;

	if (fi_ValidateFid(bufGroup, fid) != esmNOERROR)    {
        return(esmFAILURE);
    }

    /*
     *	Get the page stack
     */
	if ((pageStack = FI_AllocStack(FALSE)) == NULL)	{

		SM_ERROR(TYPE_WARNING, esmFAILURE);
	}

    /*
     *	Search for page to insert key into:
     *		which disk page it's on, what's the subscript of the key
     *		CHeck for error returned by search
     */
    if (FI_InsertSearch(bufGroup, &(fid->pid), pid, pageStack, &insertLink, &insertLoc))	{

		FI_UnfixStack(pageStack, TRUE);
		return(esmFAILURE);
	}

	/*
	 *	Pull page pointer out of structure
	 */
	insertPage = (NODEPAGE *) insertLink->bufFrame;

    /*
     * Check if insert page returned is LEAF PAGE
     */
    SM_ASSERT(LEVEL_3, FI_Leaf(insertPage->header.nodeType));

    /*
     *	Check to see if the insert location is currently valid
     */
	if (insertLoc < insertPage->header.nodeCount)	{

		/*
		 *	check to see if the page exists
		 */
		if (pid == (LARGEMASK & insertPage->contents.keyPid.allKeys[insertLoc].key)) {

			/*
			 *	Release the leaf page
			 */
			signalSemaphore( &(insertLink->pageHash->semaphore) );
			bf_UnfixPage(insertLink, BF_DEFAULT, FALSE);

			/*
			 *	Release stuff on the stack
			 */
			FI_UnfixStack(pageStack, TRUE);

			/*
			 *	return error code
			 */
			SM_ERROR(TYPE_WARNING, esmPAGEEXISTS);
			return(esmFAILURE);
		}
	}

    /*
	 *	Check to see if the leaf will overflow
     */
	TRPRINT(TR_FI, TR_LEVEL_2, ("nodecount:%d", insertPage->header.nodeCount));

	/*
	 *	log the key insert
	 */
	if (FI_LogKeyInsert(&(insertLink->pageHash->hashList.hashPid), insertLoc,
						pid, insertLink->pageHash, lobjPresent))	{

		/*
		 *	clean up the fixed pages
		 */
		signalSemaphore( &(insertLink->pageHash->semaphore) );
		bf_UnfixPage(insertLink, BF_DEFAULT, FALSE);
		FI_UnfixStack(pageStack, TRUE);

		/*
		 *	return failure
		 */
		return(esmFAILURE);
	}

	/*
	 *	No overflow, do simple insert
	 */
	FI_InsertToArray(insertPage, insertLoc, pid, NULLPID, lobjPresent);
	fi_PrintNode(insertPage);

	/*
	 *	mark the page as dirty for checkpoint processing
	 */
	DIRTY_PAGE(insertLink->pageHash);

	/*
	 *	get a pointer to the root page
	 */
	if (FI_Root(insertPage->header.nodeType))	{

		rootPage = insertPage;
		rootLink = insertLink;

	} else {

		rootPage = pageStack->stackNode[0].nodePage;
		rootLink = pageStack->stackNode[0].groupLink;
	}
	SM_ASSERT(LEVEL_3, rootLink->pageHash->hashList.hashPid.page == fid->pid.page);

	/*
	 *	null out the last logical pid
	 */
	if (predPid != NULL)	{

		*predPid = NULLPID;
	}

	/*
	 *	check to see if this is a logically ordered file
	 */
	if (rootPage->header.flags & LOGICAL_ORDER)	{

		BOOL	firstPage;

#ifdef DEBUG
		VOLREC	*volRec;

		volRec = io_FindVolRec(fid->pid.volid);
		if (volRec == NULL) {
			SM_ERROR(TYPE_FATAL, esmINTERNAL);
		}
		SM_ASSERT(LEVEL_1, volRec->volflags & VOL_TEMP);
#endif

		/*
 		 *	remember the lrc on the root page, and increment it
		 */
		rootLRC = &(rootPage->header.lrc);
		INCREMENT_LRC(rootLRC);

		/*
		 *	check for initial conditions
		 */
		if (rootPage->header.firstLogicalNode == NULLPID)	{

			TRPRINT(TR_FI, TR_LEVEL_2, ("setting first and last links to:%d", pid));
			firstPage = TRUE;

			/*
			 *	The file root must be updated to point
			 *	to the new last page.  
			 */
			recordInfo.action       = LOG_ACTION_FILE_ADD_SLOTTED;
			recordInfo.type         = LOG_REC_TYPE_USER;
			recordInfo.imageCount   = 2;
			recordInfo.actionPid    = &(fid->pid);
			recordInfo.actionLRC    = rootLRC;
			recordInfo.imageSize[0] = sizeof(BOOL);
			recordInfo.imageData[0] = (VOID *) &firstPage;
			recordInfo.imageSize[1] = sizeof(SHORTPID);
			recordInfo.imageData[1] = (VOID *) &pid;
			recordInfo.flags        = NOFLAGS;

			/*
			 *	write the record to the log
			 */
			if ((forceMark = writeLogRecord(transRec, &recordInfo, &lsn, NOFLAGS)) < 0)  {
				return(esmFAILURE);
			}

			/*
			 *	mark the page dependency
			 */
			DEPEND_LOG(rootLink->pageHash, forceMark, &lsn, rootLRC);

			/*
			 *	setup first node in chain
			 */
			rootPage->header.firstLogicalNode = rootPage->header.lastLogicalNode = pid;
			DIRTY_PAGE(rootLink->pageHash);

		} else {

			PID			tempPid;

			/*
			 *	this is not the first page, so get the last pid
			 */
			firstPage = FALSE;
			tempPid.page  = rootPage->header.lastLogicalNode;
			tempPid.volid = fid->pid.volid;
			TRPRINT(TR_FI, TR_LEVEL_2, ("setting next link in:%d to %d",
					tempPid.page, pid));

			/*
			 *	read in the last page
			 */
			if ((lastLink = bf_ReadPage(bufGroup, &tempPid,
										SLOTTED_PAGE2SIZE, NOFLAGS)) == NULL)	{

				SM_ERROR(TYPE_FATAL, esmINTERNAL);
			}
			slotPage = (SLOTTEDPAGE *) lastLink->bufFrame;
			
			/*
			 *	The lrc marked for the
			 *	lastLink page is faked to be 1 more than what is
			 *	on the page.  Redo must take this into account.
			 */
			tempLRC = slotPage->header.lrc;
			(void) INCREMENT_LRC(&tempLRC);

#ifdef JUNK  /* not needed for temporary volumes */
			{
				SHORTPID	logPids[2];

				/*
				 *	The current last page must be linked to the new
				 *	last page and the file root must be updated to point
				 *	to the new last page.  The log record for this linkage
				 *  does not increment the lrc on the old last page in
				 *	the file since a client may also be logging actions
				 *	for this page.  The analysis/redo/undo functions must
				 *	take this into account.  It is acceptable since the
				 *	link will only be updated once for a page.
				 */
				recordInfo.action       = LOG_ACTION_FILE_ADD_SLOTTED;
				recordInfo.type         = LOG_REC_TYPE_USER;
				recordInfo.imageCount   = 3;
				recordInfo.actionPid    = &(fid->pid);
				recordInfo.actionLRC    = rootLRC;
				recordInfo.imageSize[0] = sizeof(BOOL);
				recordInfo.imageData[0] = (VOID *) &firstPage;
				logPids[0] = pid;
				logPids[1] = tempPid.page;
				recordInfo.imageSize[1] = sizeof(logPids);
				recordInfo.imageData[1] = (VOID *) &(logPids[0]);
				recordInfo.imageSize[2] = sizeof(LRC);
				recordInfo.imageData[2] = (VOID *) &tempLRC;
				recordInfo.flags        = NOFLAGS;

				/*
				 *	write the record to the log
				 */
				if ((forceMark = writeLogRecord(transRec, &recordInfo, &lsn, NOFLAGS)) < 0)  {
					FI_UnfixStack(pageStack, TRUE);
					return(esmFAILURE);
				}

				/*
				 *	mark the page dependencies.  
				 *	The lsn for the last page is set to the first lsn
				 *	for the transaction since clients may be logging
				 *	things for the page as well.
				 */
				DEPEND_LOG(rootLink->pageHash, forceMark, &lsn, rootLRC);
				DEPEND_LOG(lastLink->pageHash, forceMark, &(transRec->firstLSN), &(tempLRC));

			}
#endif /* JUNK */

			/*
			 *	set the slotted page next pointer
			 */
			slotPage->header.nextLogicalPid = pid;

			/*
			 *	dirty and release the page
			 */
			bf_UnfixPage(lastLink, BF_DEFAULT, TRUE);

#ifdef JUNK
			/*
			 *	setup the link consistency structure
			 */
			if (fi_SetLinkConsist(LOGICAL_LINK, &tempPid, pid))	{

				SM_ERROR(TYPE_FATAL, esmINTERNAL);
			}
#endif /* JUNK */

			/*
			 *	check to see if user needs to know this pid
			 */
			if (predPid != NULL)	{

				*predPid = tempPid.page;
				TRPRINT(TR_FI, TR_LEVEL_2, ("predPid:%d", *predPid));
			}

			/*
			 *	setup tail pointer
			 */
			rootPage->header.lastLogicalNode = pid;
			DIRTY_PAGE(rootLink->pageHash);
		}
	}

	/*
	 *	release the semaphore
	 */
	signalSemaphore( &(insertLink->pageHash->semaphore) );

	/*
	 *	Unfix and Dirty the page
	 */
	bf_UnfixPage(insertLink, BF_DEFAULT, TRUE);

	/* 
	 * Unfix the pages in the stack
	 *
	 */
	FI_UnfixStack(pageStack, TRUE);

	/*
	 *	Return success
	 */
    return(esmNOERROR);
}
