/*
 *   $RCSfile: fi_LastPhysicalPage.c,v $  
 *   $Revision: 1.1.1.1 $  
 *   $Date: 1996/05/04 21:55:41 $      
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/


#include "sysdefs.h"
#include "ess.h"
#include "checking.h"
#include "error.h"
#include "trace.h"
#include "io.h"
#include "page.h"
#include "list.h"
#include "tid.h"
#include "lock.h"
#include "object.h"
#include "msgdefs.h"
#include "thread.h"
#include "semaphore.h"
#include "latch.h"
#include "link.h"
#include "lsn.h"
#include "bf.h"
#include "file.h"
#include "fi_macro.h"
#include "consist.h"
#include "volume.h"
#include "fi_intfuncs.h"
#include "fi_extfuncs.h"
#include "bf_extfuncs.h"


/*============================================================================*/
/* 
 * fi_LastPhysicalPage
 * 	returns the last page in the tree or subtree.
 *
 * IMPORTS:
 *	bf_ReadPage()
 *	FI_Leaf()
 *	FI_PushStack()
 *	FI_UnfixStack()
 *	FI_StackInit()
 *
 * RETURNS:
 *	esmNOERROR
 *	efiEMPTYFILE	when the file searched contains no pages
 *
 * ERROR RETURNS:
 *	efiBADPARAMS	bad parameters
 *	Errors returned by bf_ReadPage() 
 *
 * SIDE EFFECTS:
 *	Returns the last SHORTPID in the tree in 'lastPid'.
 *	if the tree is empty, NIL is returned in 'lastPid' and efiEMPTYFILE
 *	is return()ed.
 *	All pages are UNFIXED .
 */
/*============================================================================*/


 int
fi_LastPhysicalPage (

	BUFGROUP		*bufGroup,
	PID				*rootPid,
	SHORTPID		*lastPid ,
	TREEDATA		*resultLeaf,
	int				*resultCount
)
{
    
    PAGESTACK		*pageStack;		/* stack to remember path to leaf	*/
	GROUPLINK		*groupLink;		/* pointer to link for buffer		*/
    NODEPAGE		*curPage;		/* addr of current page buffer		*/
    TREEDATA		*curArray;		/* array of node/keyptr pairs		*/
    PID				curPid;			/* current page Id being handled	*/
    FOUR			last;			/* position of last key in Page		*/
	int				retval;			/* condition code to return			*/

	
	TRACE(TR_FI, TR_LEVEL_1);

    /*
	 *	Allocate stack for pages
     */
  	if ((pageStack = FI_AllocStack(FALSE)) == NULL)	{

		SM_ERROR(TYPE_FATAL, esmINTERNAL);
	}

	/*
	 *	Setup to access root page
	 */
    curPid = *rootPid;

	/*
	 *	Process the subtree
	 */
    for (;;)	{

		/*
		 * Read in current page, and grab it's type, array, and last key's
		 * subscript
		 */
		if ((groupLink = bf_ReadPage(bufGroup, &curPid,
									 FILE_PAGE2SIZE, BF_SEM)) == NULL)	{

			FI_FreeStack(pageStack);
			return(esmFAILURE);
		}

		/*
		 *	Get a pointer to the page
		 */
		curPage = (NODEPAGE *) groupLink->bufFrame;

		/*
		 *	check the magic number
		 */
		if (curPage->header.magic != FILEPAGE_MAGIC)	{

			SM_ERROR(TYPE_FATAL, esmINTERNAL);
		}

		/*
		 *	Get a pointer to the contents
		 */
		curArray = &(curPage->contents);

		/*
		 *	get subscript of last key
		 */
		last = curPage->header.nodeCount - 1;

		/* 
		 * If reached leaf, get last key in page
		 * Set done flag to true
		 */
		if (FI_Leaf(curPage->header.nodeType))  {

			/*
			 *	Check to see if node is empty
			 */
			if (curPage->header.nodeCount == 0)  {

				/*
				 *	set the return code
				 */
				SM_ERROR(TYPE_USER, esmEMPTYFILE);
				retval = esmEMPTYFILE;
				TRPRINT(TR_FI, TR_LEVEL_2, ("empty file"));

			} else {

				/*
				 *	Pull out last key in the page
				 */
				*lastPid = (curArray->keyPid.allKeys[last].key & LARGEMASK);
				retval = esmNOERROR;
				TRPRINT(TR_FI, TR_LEVEL_2, ("last pid:%d", *lastPid));
				if( resultLeaf != NULL ) {
					SM_ASSERT(LEVEL_1, resultCount != NULL);
					/* 
					 * Copy the rest of the leaf file page to the result leaf 
					 * and tell the caller how many keys we are copying.
					 * NB: The keys are in REVERSE ORDER.  The client
					 * must know that!
					 */

					*resultCount = (int)MIN(*resultCount, last+1);
					SM_ASSERT(LEVEL_3, sizeof(KEYS) == sizeof(SHORTPID));
					BCOPY( &(curArray->keyPid.allKeys[0]), resultLeaf,
								(sizeof(KEYS) * (*resultCount)) );
				} else {
					SM_ASSERT(LEVEL_1, resultCount == NULL);
				}
			}

			/*
			 *	Unfix the current page
			 */
			signalSemaphore(&(groupLink->pageHash->semaphore));
			bf_UnfixPage(groupLink, BF_DEFAULT, FALSE);

			/*
			 *	Release any pages on the stack
			 */
			FI_UnfixStack(pageStack, TRUE);

			/*
			 *	return the condition code
			 */
			if (retval)	{

				return(esmFAILURE);

			} else {

				return(esmNOERROR);
			}

		} else {

			/*
			 *	push current page on path stack
			 */
			FI_PushStack(pageStack, curPid.page, curPage, groupLink, last, NOFLAGS);

			/*
			 *	follow the leftmost node
			 */
			curPid.page = curArray->keyPtr[last].pagePtr;
		}
    }
}
