/*
 *   $RCSfile: fi_MarkPage.c,v $  
 *   $Revision: 1.1.1.1 $  
 *   $Date: 1996/05/04 21:55:41 $      
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/


#include "sysdefs.h"
#include "ess.h"
#include "checking.h"
#include "error.h"
#include "trace.h"
#include "io.h"
#include "page.h"
#include "list.h"
#include "pool.h"
#include "tid.h"
#include "lock.h"
#include "object.h"
#include "msgdefs.h"
#include "thread.h"
#include "semaphore.h"
#include "latch.h"
#include "link.h"
#include "lsn.h"
#include "bf.h"
#include "file.h"
#include "fi_macro.h"
#include "consist.h"
#include "volume.h"
#include "fi_intfuncs.h"
#include "fi_extfuncs.h"
#include "io_extfuncs.h"
#include "bf_extfuncs.h"


/*============================================================================*/
/* 
 * fi_MarkPage
 * 	Marks a page as containing a large object
 *
 * IMPORTS:
 *	FI_Leaf()
 *	FI_SearchNode()
 *	FI_StackInit()
 *	FI_UnfixStack()
 *
 * RETURNS:
 *	esmNOERROR
 *
 * ERROR RETURNS:
 *	efiBADPARAMS		bad parameters
 *	efiFOUNDPGNOTLEAF	page found in search is not leaf
 *	efiSEARCHPGNOTFOUND	page 'pageId' is not in file at all
 *	efiPGALREADYMARKED	page 'pageId' is already marked
 *
 * SIDE EFFECTS:
 *	Marks the bit in the bit map corresponding to the page specified
 *	- it now has a large object in it. If it was already marked, an
 *	error (efiPGALREADYMARKED) is returned.
 */
/*============================================================================*/


 int
fi_MarkPage (

	BUFGROUP	*bufGroup,
	PID			*fid,
	SHORTPID	pid 
)
{

    PAGESTACK	*pageStack;		/* stack to keep track of path to leaf */ 
	GROUPLINK	*foundLink;		/*	link structure for buffer	*/
    NODEPAGE	*foundPage;		/* buffer page of search page */
    SHORTPID	foundPid;		/* PID of page found in search */
	SHORTPID	*gotPid;
    FOUR		foundLoc;		/* location/subscript of search page in leaf */

	TRACE(TR_FI, TR_LEVEL_1);

	/*
	 *	Allocate page stack 
	 */
	if ((pageStack = FI_AllocStack(FALSE)) == NULL)	{

		SM_ERROR(TYPE_FATAL, esmINTERNAL);
	}

    /*
     * Search for page to mark in B+ tree
     * Check for error returned by search
     */
    if (FI_SearchNode(bufGroup, fid, pid, pageStack,
					  &foundPid, &foundLink, &foundLoc))	{

		FI_UnfixStack(pageStack, TRUE);
		return(esmFAILURE);
	}

	/*
	 *	Get found page out
	 */
	foundPage = (NODEPAGE *) foundLink->bufFrame;

	/*
	 *	check to make sure this is a leaf
	 */
	SM_ASSERT(LEVEL_3, FI_Leaf(foundPage->header.nodeType));

	/*
	 *	extract the pid from the page
	 */
	gotPid = &(foundPage->contents.keyPid.allKeys[foundLoc].key);
	
	/*
	 *	Check to see if we found the key
	 */
    SM_ASSERT(LEVEL_3, (*gotPid & LARGEMASK) == pid);

	/*
	 *	check to see if the page is already marked
	 */
    if (*gotPid & SETLARGE)	{

		/*
		 *	Release the target page
		 */
		signalSemaphore( &(foundLink->pageHash->semaphore) );
		bf_UnfixPage(foundLink, BF_DEFAULT, FALSE);

		/*
		 *	Unfix the stacked pages
		 */
		FI_UnfixStack(pageStack, TRUE);

		/*
		 *	return appropriate error code
	 	 */
		SM_ERROR(TYPE_USER, esmPAGEMARKED);
		return(esmNOERROR);

    } else {

		/*
		 * 	Log the bit being set to indicate a large obj on the page
		 */
		if (FI_LogLargeObjOnPage(&(foundLink->pageHash->hashList.hashPid), 
								foundLoc, pid, foundLink->pageHash)) {

			/*
			 *	Unfix the target page and stacked pages, and release 
			 *	the semaphore
			 */
			signalSemaphore( &(foundLink->pageHash->semaphore) );
			bf_UnfixPage(foundLink, BF_DEFAULT, FALSE);
			FI_UnfixStack(pageStack, TRUE);
		}

		/*
		 *	Set the bit
		 */
		*gotPid |= SETLARGE;

		/*
		 *	Unfix the target page, and release the semaphore
		 */
		signalSemaphore( &(foundLink->pageHash->semaphore) );
		bf_UnfixPage(foundLink, BF_DEFAULT, TRUE);

		/*
		 *	Unfix the stacked pages
		 */
		FI_UnfixStack(pageStack, TRUE);

		/*
		 *	Return the appropriate error code
		 */
		return(esmNOERROR);
	}
}
