/*
 *   $RCSfile: fi_NeighborPage.c,v $  
 *   $Revision: 1.1.1.1 $  
 *   $Date: 1996/05/04 21:55:41 $      
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/


#include "sysdefs.h"
#include "ess.h"
#include "checking.h"
#include "error.h"
#include "trace.h"
#include "io.h"
#include "page.h"
#include "list.h"
#include "tid.h"
#include "lock.h"
#include "object.h"
#include "msgdefs.h"
#include "thread.h"
#include "semaphore.h"
#include "latch.h"
#include "link.h"
#include "lsn.h"
#include "bf.h"
#include "file.h"
#include "fi_macro.h"
#include "consist.h"
#include "volume.h"
#include "fi_msg.h"
#include "fi_intfuncs.h"
#include "fi_extfuncs.h"
#include "bf_extfuncs.h"


/*============================================================================*/
/* 
 * fi_NeighborPage
 * 	Returns the neighbor of a given page
 *	Neighbor here depends on the 'direction' parameter:
 *		AFTER	immediately after 'nearPid'
 *		BEFORE	immediately before 'nearPid'
 *		ANY	any available neighbor page
 *
 * IMPORTS:
 *	bf_UnfixPage()
 *	FI_StackEmpty()
 *	FI_PopStack()
 *	FI_ReleasePages()
 *	FI_SearchNode()
 *	FI_StackInit()
 *
 * RETURNS:
 *	esmNOERROR
 *	efiEMPTYFILE		file is empty
 *	efiBEFOREFIRSTPID	scanning BEFORE first PID in file
 *	efiAFTERLASTPID		scanning AFTER last PID in file
 *
 * ERROR RETURNS:
 *	efiBADPARAMS		Bad input parameters
 *	efiINVALIDIRECTION	direction not AFTER, BEFORE or ANY
 *	efiNEARPGNOTFOUND	'nearPid' not found
 *
 * SIDE EFFECTS:
 *	The specified neighbor of 'nearPid' is returned in 'neighborPid'
 *	All pages on stack are unfixed
 */
/*============================================================================*/


 int
fi_NeighborPage (

	BUFGROUP	*bufGroup,
	PID			*fid,
	SHORTPID	nearPid,
	SHORTPID	*neighborPid,
	int			direction,
	TREEDATA	*resultLeaf,
	int			*resultCount
)
{
    PAGESTACK	*pageStack; 	/* stack to remember path to leaf	*/
	GROUPLINK	*nearLink;		/* pointer to buffer found			*/
    NODEPAGE	*nearPage;		/* buffer page with nearPid key		*/
    TREEDATA	*nearArray;		/* array in page with nearPid key	*/
    PID			nearPagePid;	/* pid of page with nearPid key		*/
    SHORTPID	returnPid;		/* Pid returned from search routine	*/
    VOLID		volid;			/* volume Id						*/
    FOUR		nearCount;		/* cnt of buffer Page where nearPid	*/
    int			nearType;		/* node type of nearPage lies		*/
    FOUR		nearLoc;		/* position/subscript of nearPid key*/
#ifdef DEBUG
	static 		int  nth_call=0;

	nth_call ++;
#endif DEBUG

#ifdef DEBUG
	if( resultLeaf == NULL) {
		SM_ASSERT(LEVEL_1, (resultCount == NULL));
	}else{
		SM_ASSERT(LEVEL_1, resultCount != NULL);
	}
#endif DEBUG


	TRPRINT(TR_FI, TR_LEVEL_1, ("file:%d nearPid:%d", fid->page, nearPid));


	/*
	 *	check for valid direction
	 */
    SM_ASSERT(LEVEL_3, (direction == AFTER) || (direction == BEFORE));

    /*
     *	Grab volume Id
     */
    volid = fid->volid;
    nearPagePid.volid = volid;

    /*
	 *	Allocate stack for searching
	 *	Semaphores will be held on all pages in the stack (instead
	 *	of only the top 2).  This is necessare since the code may
	 *	have to look back in the stack.  Some higher level 
	 *	semaphores could be released as an optimization, but
	 *	this is probably not worth implementing because of the
	 *	high fanout.
     */
	if ((pageStack = FI_AllocStack(TRUE)) == NULL)	{

		SM_ERROR(TYPE_FATAL, esmINTERNAL);
	}

	/*
	 *	Search for the node containing the key
	 */
	if (FI_SearchNode(bufGroup, fid, nearPid, pageStack,
		  			  &returnPid, &nearLink, &nearLoc))	{

		*neighborPid = NULLPID; /* to indicate error return */
		goto Done;
	}
	TRPRINT(TR_FI, TR_LEVEL_2, ("nearLoc:%d", nearLoc));

	/*
	 *	get a pointer to the buffer
	 */
	nearPage = (NODEPAGE *) nearLink->bufFrame;

    /*
     *	Grab page Id, array, and node count 
     */
    nearPagePid.page = returnPid;
    nearArray = &(nearPage->contents);
    nearCount = nearPage->header.nodeCount;
    nearType  = nearPage->header.nodeType;

	/*
	 *	Check if found node is LEAFP
     */
    SM_ASSERT(LEVEL_3, FI_Leaf(nearType));

    /*
     *	Check if the file is empty
     */
    if ((nearCount == 0) && FI_Root(nearType))		{

		TRPRINT(TR_FI, TR_LEVEL_2, ("file is empty"));

		/*
		 *	Free stack
		 */
		FI_FreeStack(pageStack);
		pageStack = NULL;

		/*
		 *	return NIL PID and error code
		 */
		*neighborPid = NULLPID;
		TRPRINT(TR_FI, TR_LEVEL_2, ("emptyfile"));
		SM_ERROR(TYPE_USER, esmEMPTYFILE);
		goto Done;
    }

    /*
     * if the nearPid is not in the file at all
     * MAY WANT TO TRY LATER(if want neighbour of pid not in file):
     *	    if nearPid != [nearLoc].key
     *		if nearLoc < nearCount
     *		    if AFTER or ANY return foundPid
     *		    else return left of foundPid, if any	
     *	
     *		else 
     *		    if AFTER return error
     *		    else return foundPid
     */
    if ((nearPid != (nearArray->keyPid.allKeys[nearLoc].key & LARGEMASK))
		|| (nearLoc >= nearCount))	{

		*neighborPid = NULLPID; /* to indicate error return */
		SM_ERROR(TYPE_WARNING, esmFILEPAGENOTFOUND);
		goto Done;
    }

    /*
     *	Check DIRECTION, i.e. BEFORE, AFTER, OR ANY
     */
    if (direction == BEFORE)  {

		TRPRINT(TR_FI, TR_LEVEL_2, ("looking before"));

		/*
		 *	check to see if the target is first key on leaf
		 */
		if (nearLoc == 0)  {

			TRPRINT(TR_FI, TR_LEVEL_2, ("before first on page"));

			/*
			 *	Identity of preceeding page is held in other leaf node
			 */
			if (FI_GetBefore(bufGroup, volid, pageStack, neighborPid, 
													resultLeaf, resultCount)){

				*neighborPid = NULLPID;
				/* next step is Done: */
			}

		} else {

			/*
			 * The previous page is to the left of the given key.
			 * Direction == BEFORE.
			 */
			*neighborPid = 
				(nearArray->keyPid.allKeys[nearLoc - 1].key & LARGEMASK);

			if( resultLeaf != NULL ) {
				SM_ASSERT(LEVEL_1, resultCount != NULL);
				/* 
				 * Copy the rest of the leaf file page to the result leaf 
				 * and tell the caller how many keys we are copying.
				 * NB: The keys are in REVERSE ORDER.  The client
				 * must know that!
				 */

				*resultCount = (int) MIN(nearLoc, *resultCount);
				SM_ASSERT(LEVEL_3, sizeof(KEYS) == sizeof(SHORTPID));
				BCOPY( &(nearArray->keyPid.allKeys[0]), resultLeaf,
							(sizeof(KEYS) * (*resultCount)) );
			} 
		}
		/* next step is Done: */
    } else {
		/* Direction == AFTER. */

		TRPRINT(TR_FI, TR_LEVEL_2, ("looking after"));

		/*
		 *	check to see if after key is on this leaf
		 */
		if (nearLoc == (nearCount - 1))   {

			TRPRINT(TR_FI, TR_LEVEL_2, ("after last on leaf"));

			/*
			 *	get the next key which is on the next leaf page
			 */
			if (FI_GetAfter(bufGroup, volid, pageStack, 
								neighborPid, resultLeaf, resultCount))	{
				*neighborPid = NULLPID;
				/* next step is Done: */
			}

		} else {

			/*
			 *	Following neighbor is the key to the right of given key
			 *  Direction == AFTER 
			 */
			*neighborPid = (nearArray->keyPid.allKeys[nearLoc+1].key & LARGEMASK);
			if( resultLeaf != NULL ) {
				SM_ASSERT(LEVEL_1, resultCount != NULL);
				/* 
				 * Copy the rest of the leaf file page to the result leaf 
				 * and tell the caller how many keys we are copying.
				 */

				*resultCount = (int) MIN(*resultCount, (nearCount - (nearLoc + 1)));
				SM_ASSERT(LEVEL_3, sizeof(KEYS) == sizeof(SHORTPID));
				BCOPY( &(nearArray->keyPid.allKeys[nearLoc+1]), resultLeaf,
							(sizeof(KEYS) * (*resultCount)) );
			} else {
				SM_ASSERT(LEVEL_1, 
				((resultCount == NULL) || (*resultCount == 0)));
			}
			/* next step is Done: */
		}
	}
	TRPRINT(TR_FI, TR_LEVEL_2, ("neighborPid:%d", *neighborPid));

Done:
	/*
	 *	Release leaf page.  
	 */
	signalSemaphore(&(nearLink->pageHash->semaphore));
	bf_UnfixPage(nearLink, BF_DEFAULT, FALSE);

	/*
	 *	Unfix pages on the stack.
	 *  This check for non-null pageStack is 
	 *  simply so that one case, above, when we know that
	 *  the stack is empty, we can not go through all the
	 *  empty loops of FI_UnfixStack.
	 */
	if(pageStack)
		FI_UnfixStack(pageStack, TRUE);

	if((*neighborPid == NULLPID) && (resultCount!=NULL)) 
		*resultCount = 0;

	if(*neighborPid == NULLPID)
		return esmFAILURE;

    return(esmNOERROR);
}
