/*
 *   $RCSfile: file.h,v $  
 *   $Revision: 1.1.1.1 $  
 *   $Date: 1996/05/04 21:55:44 $      
 */ 
#ifndef __FILE_H__
#define __FILE_H__


/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/

/* 
 * file.h
 *	Data structures and constants for B trees used in implementing
 *	files
 *
 * IMPORTS:
 *	page.h
 *	ess.h
 *
 * EXPORTS:	
 *	node types:
 *	    ROOTP
 *	    ROOTLEAFP
 *	    INTP
 *	    LEAFP
 *	neighbour position:
 *	    AFTER
 *	    BEFORE
 *	    ANY
 *      FULLEAFNODE 	 - max # of keys in a leaf
 *      FULLNODELEAFNODE - max # of key,ptrs in a non leaf
 *	LEAFUNDERFLOW 	 - min # of elements in a node/page(leaf) 
 *	NONLEAFUNDERFLOW - min # of elements in a node/page(non leaf) 
 *		  	 - ROOT is exceptional
 *	TREEDATA  	 - data structure for key-pointer pair.
 *      NODEPAGE  	 - a page/node in the B tree.
 *	PAGESTACK	 - stack used to traverse B tree
 *
 * BUGS:
 *      None
 *
 * DESCRIPTION:
 *	Files are implemented as B+ trees.
 *	The B+ tree like structure used is described as follows:
 *	There are 4 node types:
 *		ROOTP		Root, not leaf.
 *		ROOTLEAFP	Root and also a leaf.
 *		INTP		Internal node.
 *		LEAFP		Leaf node.
 *
 *	There are 2 kinds of node structure: leaves and non leaves.
 *			LEAVES
 *	*********************************************
 *	*                                           *
 *	*	nodeCount - int                     *
 *	*	nodeType  - int			    *
 *	*	key array - SHORTPIDs		    *
 *	*	---------------------....-----      *  	
 *	*	|key|key|key|key|key|....|key|	    *
 *	*	---------------------....-----	    *
 *	*					    *
 *	*********************************************
 *	Leaves have FULLEAFNODE keys.
 *	The nodeCount field is the count of KEYS in the page.
 *	A leaf is FULL when nodeCount = FULLEAFNODE.
 *	A leaf underflows when nodeCount = LEAFUNDERFLOW.
 *	   <key><key><key>.............<key>
 *	   <----------FULLEAFNODE)--------->
 *
 *			    NONLEAVES
 *	*****************************************************
 *	*                                                   *
 *	*	nodeCount - int                             *
 *	*	nodeType  - int			            *
 *	*	key array - SHORTPIDs		            *
 *	*	-------------------------....----------     *
 *	*	|Ptr|key|Ptr|key|Ptr|key|....|Ptr|NULL|     *
 *	*	-------------------------....----------     *
 *	*					            *
 *	*****************************************************
 *	Non leaves have FULLNONLEAFNODE <key,Ptr> pairs and
 *	the last key value is NULL.
 *	The page pointed to by Ptr[i] contains the (paths to) 
 *	keys <= key[i].
 *	The nodeCount field is the count of POINTERS in the page.
 *	A non leaf is full when nodeCount = FULLNONLEAFNODE.
 *	A non leaf underflows when nodeCount = NONLEAFUNDERFLOW.
 *	   <ptr,key><ptr,key><ptr,key>.............<ptr,null>
 *	   <---------(FULLNONLEAFNODE - 1)-------->
 *
 */

/*
 *	define the size of the file page
 */
#define FILE_PAGE2SIZE	MIN_PAGE2SIZE
#define FILE_PAGESIZE	(1 << FILE_PAGE2SIZE)

/*
 * Node/page Types
 */
#define		FI_ROOTPAGE			0x1		/* Root but not leaf */
#define		FI_INTPAGE			0x2		/* Not root, not leaf */
#define		FI_LEAFPAGE			0x4		/* leaf page */

/*
 *	Define special node that is both root and leaf
 *	for a tree of height 1
 */
#define		FI_ROOTLEAFPAGE		FI_ROOTPAGE|FI_LEAFPAGE

/*
 * constants to specify sibling/neighbour position
 */
#define		BEFORE 		NEIGHBOR_BEFORE
#define		AFTER		NEIGHBOR_AFTER
#define		ANY			NEIGHBOR_ANY	/* means any available neighbor,
										 * regardless of position
										 */

/*
 * Type def for <key,ptr> pair
 */
typedef struct {

    SHORTPID	pagePtr;
    SHORTPID	key;

} KEYPTRS;


#define		KEYPTRSIZE	(sizeof(KEYPTRS))


/*
 * Type def for <key>
 */
typedef struct {

    SHORTPID	key;

} KEYS;


#define		KEYSIZE		(sizeof(KEYS))

/* 
 * Node constants
 */

/*
 *	define the header for each node
 */
typedef struct {

	MAGIC		magic;
	LRC			lrc;
	FOUR		nodeCount;
	TWO			fileType;
	TWO			nodeType;
	FLAGS		flags;
	SHORTPID	firstLogicalNode;
	SHORTPID	lastLogicalNode;
	UFOUR		unique;

} NODEHEADER;


#ifndef TESTFILE

/*
 * Size of header in all pages:
 *	nodeType  - int
 *	nodeCount - int
 */
#define		SPACESIZE	(FILE_PAGESIZE - sizeof(NODEHEADER))

/*
 * Full leaf node
 */
#define		FULLEAFNODE			(SPACESIZE / KEYSIZE)

/*
 * Full non leaf node
 */
#define		FULLNONLEAFNODE		(SPACESIZE / KEYPTRSIZE)

#else

#define		FULLEAFNODE			3
#define		FULLNONLEAFNODE		3

#endif

/*
 * Underflow
#define		UNDERFLOW	(FULLNONLEAFNODE / 4)
 */
#define		NONLEAFUNDERFLOW	(FULLNONLEAFNODE / 4)
#define		LEAFUNDERFLOW		(FULLEAFNODE / 4)
#define		KEYSPACE			FULLEAFNODE * KEYSIZE
#define		LARGEMASK			0x7FFFFFFF
#define		SETLARGE			0x80000000

/*
 *  Type def for leaf page key array, and bit map
 */
typedef struct {

    KEYS	allKeys[FULLEAFNODE];

} KEYPIDS;


/*
 * Typedef for contents of tree
 * - different for leaf, non leaf
 */
typedef union  {

    KEYPIDS			keyPid;
    KEYPTRS			keyPtr[FULLNONLEAFNODE];

} TREEDATA;


/*
 *
 * Typedef for a node
 *
 */
typedef struct {
	
	/*
	 *	The header must always be the first thing on the page
	 */
	NODEHEADER	header;
    TREEDATA	contents;

} NODEPAGE;


/*
 *	define the file magic number
 */
#define FILEPAGE_MAGIC		0x42351cba


/*
 * Type def for stack used to traverse B tree
 */
typedef struct {

    SHORTPID		pageID;		/* PID of page on stack					*/
    NODEPAGE		*nodePage;	/* direct pointer to the page			*/
	GROUPLINK		*groupLink;	/* pointer to data structure for buffer	*/
    FOUR			path;		/* path/subscript from parent			*/
	FLAGS			flags;		/* flags for this entry					*/

} STACKNODE;


#define	STACKSIZE	20

typedef struct {

	int				stackTop;
	BOOL			holdAllSemaphores;
	STACKNODE		stackNode[STACKSIZE];
	LISTELEMENT		stackList;

} PAGESTACK;


/*
 *	define flags for file creation parameters
 */
#define LOGICAL_ORDER	0x1


/*
 * Macro to test for FID equality.
 */
#define FIDEQ(_fid1, _fid2)								\
		(((_fid1).pid.page  == (_fid2).pid.page) &&		\
		 ((_fid1).pid.volid == (_fid2).pid.volid) &&	\
		 ((_fid1).unique == (_fid2).unique))


/*
 *	data structure storing key insert/delete information in
 *	a log record image.
 */
typedef struct	{

	FID				fid;
	BOOL			largeObjOnPage;

} FILEKEYLOGINFO;

#endif __FILE_H__
