/*
 *   $RCSfile: queues.h,v $  
 *   $Revision: 1.2 $  
 *   $Date: 1996/05/04 23:51:51 $      
 */ 
#ifndef __QUEUES_H__
#define __QUEUES_H__

/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/

#include "forward.h"

/* 
 * The queue capacity is a function of the number of possible clients,
 * which is limited by 
 *	a) the number of file descriptors, although we could
 * 		use any queues size limit we wish; there's no need 
 *		to allow all client requests to fit in a queue at once.
 *	b) the number of threads
 *	c) the number of disk-buffers 
 */

extern	int				NumMsgBufs;
#define QUEUE_SIZE    	NumMsgBufs

/*
 * Each message in shared memory is
 * found by adding its offset(in bytes) to ShmAddress.
 * The offset is kept in the disk queues.
 *
 * ShmOffset is really a char * but it's called a FOUR * because
 * it must be register-size aligned.
 */
typedef FOUR * ShmOffset;  
#define ShmOffsetToAddress(x,Type) (Type *)(ShmAddress + (int)x)
#define ShmAddressToOffset(x) (((char *)x) - ShmAddress)


#ifdef TESTANDSET
/*
 * declare the test and set function.
 */

BEGIN_EXTERNC
	extern unsigned char testandset(unsigned char *, unsigned char value);
END_EXTERNC

#endif TESTANDSET

/*
 * With TESTANDSET, we don't need volatile.
 */
#if !defined(volatile) && defined(TESTANDSET)
#define volatile
#endif 

#ifdef linux
/*  Linux defines it's own MUTEX in <linux/wait.h>  */
#undef MUTEX
#endif linux

typedef struct  {
	char					magic;
	unsigned char	sync; 	/* test and set  uses sync. */
	/* Silberschatz & Peterson. algorithm 4, p. 88, "alternate edition"
	 * use flag, turn.
	 * We use one or the other of the algorithms, but compile
	 * in the data for both so that switching from one to the other
	 * doesn't require recompilation of the whole SM.
	 */
	volatile unsigned char flag[2]; 
	volatile unsigned turn:1, willBlock:1;  /* consumer will block */
} MUTEX; /* semaphore */

typedef struct 	{
	unsigned char front, /* next item in the queue */
				  back;	/* next avaliable slot */
	MUTEX		mutex;
	int			magic;
} QUEUE;

typedef struct {
	QUEUE toDisk, fromDisk;

	int			semNum;	
	VOLREC		*volRec; 
	/* **NB: the elements MUST be the
	 * last item in this structure, since it is really just
	 * the beginning of an array, whose size is TBD at run time 
	 */
	ShmOffset	elements;
} QUEUEPAIR;

/* 
 * The following structure is used
 * for synchronization between diskproc and server proc.
 * Used by disk manipulation routines (tryMutex, etc).
 */
typedef struct {
	unsigned	me:1, other:1;
} ProcessIdentity;

extern ProcessIdentity process_identity; /* in disk/space.c */
extern MUTEX 		   *Replies; 		/* in disk/space.c */


#define DISK_PROC	0
#define SERVER_PROC 1

#ifdef TESTANDSET
BEGIN_EXTERNC
	extern int atomic_op();
END_EXTERNC
#endif TESTANDSET

#define fromDiskFull(q)\
		(error because we never need to check this)
		
#define fromDiskEmpty(q)\
		(q->fromDisk.front == q->fromDisk.back)

#define toDiskFull(q)\
        (q->toDisk.back == (q->toDisk.front-1)%(QUEUE_SIZE))

#define toDiskEmpty(q)\
		(q->toDisk.front==q->toDisk.back)

#define enq_to(q, value) {\
        register ShmOffset *elements=(&(q->elements));\
	 	elements[q->toDisk.back++] = (ShmOffset) value; \
		if(q->toDisk.back >= QUEUE_SIZE)\
			q->toDisk.back = 0;\
	}

#define enq_from(q, value) {\
     register ShmOffset *elements=(&(q->elements));\
	 elements[q->fromDisk.back++] = (ShmOffset) value; \
	 if(q->fromDisk.back >= QUEUE_SIZE)\
		 q->fromDisk.back = 0;\
	}

#define deq_to(q, value) {\
     register ShmOffset *elements=(&(q->elements));\
	 value = elements[q->toDisk.front++]; \
		if(q->toDisk.front >= QUEUE_SIZE)\
			 q->toDisk.front = 0;\
	}

#define deq_from(q, value) \
	dequeueFrom(q, (ShmOffset *) &value)

#if MAGIC_CHECKING  IS_ENABLED

#define QMAGIC			0xa9045c1d
#define MUTEXMAGIC		0x53 /* DON'T SET HIGH BIT */

#define INIT_MUTEX_MAGIC(_qp) 				\
		(_qp)->magic = MUTEXMAGIC;

#define INIT_QUEUE_MAGIC(_qp) 				\
		(_qp)->magic = QMAGIC;				\
		INIT_MUTEX_MAGIC(&((_qp)->mutex));

#define CHECK_MUTEX_MAGIC(_qp) 				\
	if ((_qp)->magic != MUTEXMAGIC)	{			\
		SM_ERROR(TYPE_FATAL, esmINTERNAL);	\
	}

#define CHECK_QUEUE_MAGIC(_qp) 				\
	if ((_qp)->magic != QMAGIC)	{			\
		SM_ERROR(TYPE_FATAL, esmINTERNAL);	\
	}                                       \
	CHECK_MUTEX_MAGIC(&((_qp)->mutex));

#else MAGIC_CHECKING IS_ENABLED

#define INIT_QUEUE_MAGIC(_qp) 	/* comment */
#define CHECK_QUEUE_MAGIC(_qp) 	/* comment */

#define INIT_MUTEX_MAGIC(_qp) 	/* comment */
#define CHECK_MUTEX_MAGIC(_qp) 	/* comment */

#endif MAGIC_CHECKING IS_ENABLED

BEGIN_EXTERNC
	extern	QUEUEPAIR *diskQueue(int);
	extern	QUEUEPAIR *initDiskQueue(int, VOLREC *);
END_EXTERNC
#endif __QUEUES_H__
