/*
 *   $RCSfile: IO_MarkBitmapRange.c,v $  
 *   $Revision: 1.1.1.1 $  
 *   $Date: 1996/05/04 21:55:49 $      
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/

#include "sysdefs.h"
#include "ess.h"
#include "checking.h"
#include "trace.h"
#include "error.h"
#include "list.h"
#include "tid.h"
#include "io.h"
#include "lock.h"
#include "object.h"
#include "msgdefs.h"
#include "thread.h"
#include "latch.h"
#include "semaphore.h"
#include "link.h"
#include "lsn.h"
#include "bf.h"
#include "volume.h"
#include "disk.h"
#include "bitmap.h"
#include "io_intfuncs.h"
#include "bf_extfuncs.h"
#include "disk_funcs.h"
#include "bm_extfuncs.h"
#include "bit_funcs.h"

#ifdef DEBUG

#define VERIFY_BIT(_bit, _page)	\
	checkBit = ((_page) - firstPid->page) * BITMAP_USABLESIZE + (_bit); \
 	if (checkBit < startBit || checkBit >= numBits+startBit) {		 \
		SM_ERROR(TYPE_FATAL, esmINTERNAL);		 					 \
	}
#else

#define VERIFY_BIT(_bit, _page) /* nothing */

#endif

/*
 *	This function initializes a range of bits in a bitmap. 
 */

 int
IO_MarkBitmapRange (
	BUFGROUP		*bufGroup,
	PID				*firstPid,	/* first bitmap page		*/
	int				startBit,	/* first bit to mark		*/
	int				numBits,	/* number of bits to mark	*/
	int				mark  		/* 1 or 0 to mark each bit	*/

)
{

	GROUPLINK	*bitmapLink;
	PID			bitmapPid;
	SHORTPID	page;
	int			markedBits;
	int			firstBit;
	int			bitCount;
	int			Bit;
	int			extraBits;
	int			checkBit;	/* used in VERIFY_BIT */

	TRPRINT(TR_IO, TR_LEVEL_1, ("marking %d bits starting at %d", numBits, startBit));

	/*
	 *	Determine the first page with bits to be marked
	 */
	page = BIT_TO_PAGE(startBit);

	/*
	 *	Loop through every bitmap page in the range
	 */
	markedBits = 0;
	bitmapPid.volid = firstPid->volid;
	while (markedBits < numBits) {

		bitmapPid.page = page + firstPid->page;
		firstBit = (int) BIT_IN_PAGE(markedBits + startBit);
	
		/*
		 *	Read in the bitmap page
		 */
		bitmapLink = bf_ReadPage(bufGroup, &bitmapPid, BITMAP_PAGE2SIZE, BF_NOREAD);
		if (bitmapLink == NULL)	{

			return(esmFAILURE);
		}

		/*
		 *	See if the first bits on the first page fill only
		 *	part of a word
		 */
		extraBits = BITSPERWORD - (firstBit % BITSPERWORD);	
		if (markedBits == 0 && extraBits != BITSPERWORD) {
			
			/*
			 *	Set these bits separately, since clearmap and
			 *	setmap only work on entire bytes.
			 */
			for (Bit = 0; Bit < extraBits; Bit++) {

				VERIFY_BIT(Bit + firstBit, (int) bitmapPid.page);
				
				if (mark == 0) {

					/*
					 *	Clear the bit
					 */
					bm_ClearBit((UFOUR*)bitmapLink->bufFrame,
							    Bit + firstBit);
				} else {

					/*
					 *	mark the bits
					 */
					bm_SetBit((UFOUR*)bitmapLink->bufFrame,
							    Bit + firstBit);
				}	
			}

			markedBits += extraBits;
			firstBit += extraBits;

			/*
			 *	See if we're finished
			 */
			if (markedBits == numBits) {

				/*
				 *	We are, so unfix the bitmap page and
				 *	exit the loop
				 */
				bf_UnfixPage(bitmapLink, BF_DEFAULT, TRUE);
				break;
			}
		}


		/*
		 *	See if this is the last page with bits to mark
		 */
		if ( (markedBits + BITMAP_USABLESIZE * BITSPERBYTE) >= numBits) {
			
			/*
			 * last page, calculate the number of bits to mark
			 */
			bitCount = numBits - markedBits;
			SM_ASSERT(LEVEL_3, markedBits == 0 || firstBit == 0);

			/*
			 *	See if the last bits on the last page fill only
			 *	part of a word
			 */
			extraBits = (bitCount % BITSPERWORD);	
			if (extraBits) {

				bitCount -= extraBits;
				/*
				 *	Set these bits separately, since clearmap and
				 *	setmap only work on entire bytes.
				 */
				for (Bit = bitCount; Bit < bitCount + extraBits; Bit++) {
					
					VERIFY_BIT(Bit, (int) bitmapPid.page)

					if (mark == 0) {

						/*
						 *	Clear the bit
						 */
						bm_ClearBit((UFOUR*)bitmapLink->bufFrame,
									Bit);
					} else {

						/*
						 *	mark the bits
						 */
						bm_SetBit((UFOUR*)bitmapLink->bufFrame,
									Bit);
					}	
				}

				markedBits += extraBits;

				/*
				 *	See if we're finished
				 */
				if (markedBits == numBits) {

					/*
					 *	We are, so unfix the bitmap page and
					 *	exit the loop
					 */
					bf_UnfixPage(bitmapLink, BF_DEFAULT, TRUE);
					break;
				}
			}	

		} else {

			bitCount = BITMAP_USABLESIZE * BITSPERBYTE - firstBit;
		}

		if (mark == 0) {

			/*
			 *	Clear the bits
			 */
			clearmap((unsigned char*)bitmapLink->bufFrame+(firstBit /
					 BITSPERBYTE), bitCount);
		} else {

			/*
			 *	mark the bits
			 */
			setmap((unsigned char*)bitmapLink->bufFrame+(firstBit /
			       BITSPERBYTE), bitCount);
		}	

		/*
		 *	Unfix the bitmap page
		 */
		bf_UnfixPage(bitmapLink, BF_DEFAULT, TRUE);

		/*
		 *	advance
		 */
		markedBits += bitCount;
		page++;
	}
	
	return (esmNOERROR);
}
