/*
 *   $RCSfile: IO_MountVolume.c,v $  
 *   $Revision: 1.1.1.1 $  
 *   $Date: 1996/05/04 21:55:46 $      
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/

#include "sysdefs.h"
#include "ess.h"
#include "checking.h"
#include "trace.h"
#include "error.h"
#include "list.h"
#include "pool.h"
#include "tid.h"
#include "io.h"
#include "lock.h"
#include "object.h"
#include "msgdefs.h"
#include "thread.h"
#include "semaphore.h"
#include "link.h"
#include "latch.h"
#include "lsn.h"
#include "bf.h"
#include "bitmap.h"
#include "volume.h"
#include "bitvec.h"
#include "io_intfuncs.h"
#include "io_extfuncs.h"
#include "disk.h"
#include "disk_funcs.h"
#include "bf_extfuncs.h"
#include "msg_funcs.h"
#include "util_funcs.h"
#include "io_globals.h"
#include "log_globals.h"
#include "disk_globals.h"
#include "queue_consist.h"

/*
 * Module IOmount : routines to mount and dismount volRecs
 *
 * Mount the named device and read its volRec header information.
 *    The relevant information on the device is stored at entry "Vix"
 *    of the volRec table.
 * 
 *   Returns:
 * 	Error status
 * 	
 * 
 *  Side Effects:
 * 	Device mounted and its volRec header cached in memory
 * 	In addition the root page of the file directory associated
 * 	with the volRec is read into memory.
 */

static void
updateHighestWrapCount(VOLHDR *header)
{
	if(LogStartWrapCount < header->dismountEndOfLog.wrapCount) {
		TRPRINT(TR_RECOVER | TR_LOG, TR_LEVEL_1,
			("setting log start wrap count to %d", header->dismountEndOfLog.wrapCount));
		LogStartWrapCount = (int) header->dismountEndOfLog.wrapCount;
	}
}


/* formatvol uses this IO_MountVolume directly, with check==FALSE,
 * for display purposes.
 * We know that the volNameRec is there and is legit but
 * we don't know if the volid in it matches that in the volume header.
 */
int
IO_MountVolume (
	VOLREC		*const 	volRec,
	const 		BOOL	check
)
{
	char		*volName;
	VOLHDR		*header;
	GROUPLINK	*headerLink;
	GROUPLINK	*tempLink = NULL;


	SM_ASSERT(LEVEL_1, volRec->volNameRec != NULL);

	volName = volRec->volNameRec->volName.name;
	TRPRINT(TR_IO, TR_LEVEL_1, ("mounting volume:%s", volName));

	/*
	 *	mount the named "physical" device
	 *  Pass along a volume record structure, which is where
	 *  openDisk() stashes both the link and the disk queue info.
	 */
	SM_ASSERT(LEVEL_1, (volRec != NULL));
	SM_ASSERT(LEVEL_1, (volRec->bufGroup == NULL));
	SM_ASSERT(LEVEL_1, (volRec->volLink == NULL)); /* going into openDisk */

	if (openDisk(volName, FALSE, volRec) == FALSE )	{
		FAIL
	}

	/* after openDisk: */
	SM_ASSERT(LEVEL_1, (FastPathOnly && (volRec->volLink == NULL))
		||			(  (!FastPathOnly)&&(volRec->volLink != NULL)) );


	/*
	 *	allocate a buffer and read in the volRec header
	 *	note, we must read into a page sized buffer and then
	 *	move the selected information into the DiskTable as the entries
	 *	in DiskTable do not hold a full page - the unit of I/o
	 */
	if ((tempLink = bf_GetTempPage(HEADER_PAGE2SIZE)) == NULL)	{
		FAIL
	}

	header = (VOLHDR *) tempLink->bufFrame;
	if (IO_ReadPage(volRec, HEADERADDR, HEADER_PAGESIZE, (char *) header))	{
		FAIL
	}

	/* 
	 * Correctness checks (real possible user errors): 
	 * (The volume could have been formatted with a volid that
	 * differs from the volid in the config files, with an old
	 * formatter, with wrong size pages, etc.
	 * 
	 * We know this volid and volume name pair is in, since
	 * 	 the caller used a volNameRec table to get us here in the first
	 *   place.
	 *
	 * check: The volid in the header had better match the one in the
	 *      tables that's assoc. with this path.
	 *
	 * The server maintains the invariant that no path has
	 * more than one entry and no volid has more than one entry,
	 * so it's sufficient to do one lookup by volid and see
	 * that the path matches.
	 */
	if(check) {
		VOLNAMEREC *volNameRec = volRec->volNameRec;

		if(volNameRec->volid != header->volid) {
			if(PrintUser) fprintf(stderr, 
			"\nVolume %s: volid mismatch: header %d, config %d\n",
				volName, header->volid, volNameRec->volid);
			SM_ERROR(TYPE_LOG, esmBADVOLHEADER);
			FAIL
		}
	}

#ifdef DEBUG
	/*
	 * This is just an internal consistency check for debugging:
	 * We would LIKE to say that this shouldn't already be in the 
	 * volume table, but in fact it will be there, in VOL_INIT state.  
	 * We can't * do io_FindVolRec(), because it will wait for the 
	 * volume to be out of INIT state, and we'll deadlock.
	 * We'd like to check for flags == VOL_INIT but there are other
	 * flags, such as properties (RAW, TEMP, etc) that are legit here.
	 */
	SM_ASSERT(LEVEL_1, 
		(volRec->volflags & (VOL_OPEN|VOL_INVALID|VOL_QUIESCE|VOL_FORMAT))==0);
#endif DEBUG

	/*
	 *	check to see if the header is sane
	 */
	if(check) switch (header->magic) {
	case OLD_VOLUME_MAGIC: /* no checks */
		break;

	case VOLUME_MAGIC:
		if( (SLOTTED_PAGE2SIZE == header->slottedPage2size) &&
			(LG_PAGE2SIZE == header->lgPage2size) &&
			(LGDATA_PAGE2SIZE == header->lgDataPage2size) &&
			(BTREE_PAGE2SIZE	== header->btreePage2size) &&
			(INDEXDESC_PAGE2SIZE	== header->indexDescPage2size) &&
			(MIN_PAGE2SIZE	== header->minPage2size)  &&
			(FORMAT_VERSION	== header->version) )
				break;
		/* else fall through */
		if(PrintUser) fprintf(stderr, 
		"\nVolume %s: page size mismatch\n", volName, header->magic);
		break;

	default:
		if(PrintUser) fprintf(stderr, 
			"\nVolume %s: magic 0x%x\n",volName, header->magic);
		SM_ERROR(TYPE_LOG, esmBADVOLHEADER);
		FAIL
	}

	updateHighestWrapCount(header);

	if ((volRec->bufGroup = 
		bf_OpenGroup(BITMAP_GROUP_SIZE, BF_LRU, NOFLAGS)) == NULL)	{

		FAIL

	}

	{
		PID			headerPid;

		headerPid.page  = HEADERADDR;
		headerPid.volid = header->volid;

		/*
		 *	get a buffer to hold the header page
		 */
		if ((headerLink = bf_ReadPage(volRec->bufGroup, &headerPid,
					  HEADER_PAGE2SIZE, BF_SEM|BF_NOREAD)) == NULL)	{
			FAIL
		}
	}

	bcopy((char *) header, headerLink->bufFrame, (int) sizeof(VOLHDR));

	/*
	 *	release space occupied by disk version of volRec header
	 */
	bf_FreeTempPage(tempLink);
	tempLink = NULL; /* lest we fail later */

	volRec->dirty = FALSE;
	volRec->header = (VOLHDR *) headerLink->bufFrame;
	volRec->headerLink = headerLink;
	volRec->volflags |= (volRec->header->properties & VOLPROP_TEMP);

	SM_ASSERT(LEVEL_1, (volRec->header->properties & VOLPROP_TEMP) ==
		(volRec->volflags & VOL_TEMP));

	{
		VOLINFO		*volInfo;
		/*
		 *	Remember the volume that this link applies to
		 *	First get a VOLINFO structure then attach it to the link
		 */
		if ((volInfo = (VOLINFO *) listDeq( &VolInfoPool )) == NULL)    {
			SM_ERROR(TYPE_LOG, esmNOVOLINFO);
			FAIL
		}

		if(volRec->volLink) /* !FastPathOnly */
			listEnq( &(volRec->volLink->volList), &(volInfo->linkList) );
		volInfo->volid = volRec->header->volid;
		volInfo->volRec = volRec;
		volInfo->link = volRec->volLink;
	}

	/*
	 *	give up the semaphore on the page
	 *	but keep the page fixed
	 */
	signalSemaphore( &(headerLink->pageHash->semaphore) );
	return esmNOERROR;

FailureReturn:
	if(tempLink) {
		bf_FreeTempPage(tempLink);
	}
	if(volRec->bufGroup) {
		bf_CloseGroup(volRec->bufGroup);
	}
	if(volRec->volLink) {
		/* !FastPathOnly */
		clearLink(volRec->volLink);
		volRec->volLink = NULL; /* since it's cleared */
	}

	return esmFAILURE;
}
