/*
 *   $RCSfile: IO_OneTrackAll.c,v $  
 *   $Revision: 1.1.1.1 $  
 *   $Date: 1996/05/04 21:55:47 $      
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/


#include "sysdefs.h"
#include "ess.h"
#include "checking.h"
#include "trace.h"
#include "error.h"
#include "list.h"
#include "pool.h"
#include "tid.h"
#include "io.h"
#include "lock.h"
#include "object.h"
#include "msgdefs.h"
#include "thread.h"
#include "semaphore.h"
#include "link.h"
#include "latch.h"
#include "lsn.h"
#include "bf.h"
#include "volume.h"
#include "bitmap.h"
#include "io_intfuncs.h"
#include "io_extfuncs.h"
#include "bm_extfuncs.h"


/*
 * Allocate PagesWanted pages.
 *
 * Returns:
 *	pidList of the allocated pages
 *
 * Side Effects:
 *	page bit map updated in volume header entry.
 *
 * Errors:
 * 	eioFRAGMENTEDVOL -  vol too fragmented.
 */

#define RIGHT 0
#define LEFT  1


 int
IO_OneTrackAll (

	int				soFar,			/* Number of pages allocated so far		*/
	int				numBits,	 	/* Number of pages left to allocate		*/
	PAGESIZE		blockSize,		/* Number of blocks in each page		*/
	int				startTrack,		/* where search for free pages begins	*/
	VOLREC			*volRec,		/* points to the volume header entry	*/
	PID				pidList[],		/* Page IDs of allocated pages	*/
	LIST			*bitmapList, 	/* list of bitmap pages			*/
	LIST			*contextList	/* info about bits set 			*/
)
{

	int				bitsUsed;
    int				curTrack;		/* beginning page of the current track	*/
    int				nextTrack;		/* beginning page of the next track		*/
    int				direction;		/* direction of search					*/
	LIST			*pageList;
	FREECONTEXT		*context;


	TRPRINT(TR_IO, TR_LEVEL_1, ("startTrack:%d numBits:%d", startTrack, numBits));

	/*
	 *	check to see if a bitmap list has been passed in
	 */
	if (bitmapList != NULL)	{
		
		pageList = bitmapList;

	} else {
		SM_ERROR(TYPE_FATAL, esmINTERNAL);
	}
	SM_ASSERT(LEVEL_3, contextList != NULL);

	/*
	 *	check to see if the volume can accomodate this size request
	 */
	if (volRec->header->pagesPerTrack < numBits)	{

		TRPRINT(TR_IO, TR_LEVEL_2, ("track size too small"));
		return(esmNOERROR);
	}

	/*
	 *	setup search conditions
	 */
    direction = RIGHT;
    curTrack  = startTrack;
    nextTrack = startTrack + volRec->header->pagesPerTrack;

	/*
	 *	Right now, all callers will send only an empty list,
	 *	but they  need not be.
	 */
	SM_ASSERT(LEVEL_3, LIST_EMPTY(pageList));
	SM_ASSERT(LEVEL_3, LIST_EMPTY(contextList));

	/*
	 *	Search until found
	 */
    for (;;) {

		TRPRINT(TR_IO, TR_LEVEL_2, ("curTrack:%d nextTrack:%d", curTrack, nextTrack));

		/*
		 *	Search completes without enough allocation
		 */
		if (curTrack < 0)	{

			break;
		}

		/*
		 *	See if there are enough contiguous pages
		 *	on the track to contain a LONG page.
		 */
		context = bm_FindContigBits(volRec, numBits, curTrack, nextTrack, pageList);

		/*
		 *	check to see if we have found a page
		 */
		if (context != NULL)	{

			/*
			 *	check the context magic
			 */
			CHECK_FREECONTEXT_MAGIC(context);

			/*
			 *	hang the context on the context list
			 */
			listEnq( contextList, &(context->list) );

			/*
			 *	mark the individual pages that were allocated
			 */
			for (bitsUsed = 0; bitsUsed < numBits; bitsUsed += blockSize)	{

				/*
				 *	mark the page as used
				 */
				pidList[soFar].page  = context->bitContext[0].absoluteBitOffset + bitsUsed;
				TRPRINT(TR_IO, TR_LEVEL_2, ("alloc page:%d", pidList[soFar].page));
				pidList[soFar++].volid = volRec->header->volid;
			}

			/*
			 *	return success
			 */
			return(esmNOERROR);
		}

		/*
		 *	check to see if we should change direction
		 */
		IO_CheckDirection(&curTrack, &nextTrack, &direction, startTrack, volRec);
    }

	/*
	 *	Check the consistency of the bitmap
	 */
    SM_ASSERT(LEVEL_3, !io_Checker(volRec));

	/*
	 *	Disk didn't have any tracks that could hold the allocation
	 */
	SM_ERROR(TYPE_WARNING, esmFRAGMENTED);
    return(esmFAILURE);
}
