/*
 *   $RCSfile: io_AllocPages.c,v $  
 *   $Revision: 1.1.1.1 $  
 *   $Date: 1996/05/04 21:55:46 $      
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/


#include "sysdefs.h"
#include "ess.h"
#include "checking.h"
#include "trace.h"
#include "error.h"
#include "list.h"
#include "pool.h"
#include "tid.h"
#include "io.h"
#include "lock.h"
#include "object.h"
#include "msgdefs.h"
#include "thread.h"
#include "semaphore.h"
#include "link.h"
#include "latch.h"
#include "lsn.h"
#include "bf.h"
#include "volume.h"
#include "bitmap.h"
#include "openlog.h"
#include "io_intfuncs.h"
#include "io_extfuncs.h"
#include "io_logfuncs.h"
#include "bm_extfuncs.h"
#include "thread_globals.h"
#include "log_globals.h"

static int IO_AllocSimpleSearch(int, int, PAGESIZE, int, VOLREC*, PID[], LIST*, LIST*);

/*
 * Allocate "NumPages" pages and return their IDs in "pidList[]".
 *   If NearPID is specified, allocate pages close to it if possible.
 *
 *	   Returns:
 *		IDs of allocated pages in pidList[]
 *
 *	   Side Effects:
 *			VolTable fields are updated to indicate fewer free pages and a
 *		dirty header.
 *
 *	   Errors:
 *		eioNOSPACEONVOL - not enough space to fulfill request
 *		eioNULLPIDPTR    - null PID pointer
 *
 *	   Bugs:
 *		does not validate NearPID (too expensive to do)
 */


 int
io_AllocPages (

	VOLREC			*volRec,
	PAGE2SIZE		page2size,
	int				numPages,
	PID				*nearPid,
	PID				pidList[]
)
{

 	int			count;			/* how many SHORT page equivalents	*/
	int			error;			/* error return code				*/
	VOLHDR		*header;
	SHORTPID	nearPage;		/* page portion of nearPid			*/
	int			startTrack;		/* first page on the track in which	*/
	PAGESIZE	blockSize;
	LIST		bitmapList;		/* bitmap pages pinned during alloc	*/
	LIST		contextList;	/* info about allocated bits 		*/


	TRPRINT(TR_IO, TR_LEVEL_1, ("allocating:%d page2size:%d", numPages, page2size));

	/*
	 *	initialize the bitmap list
	 */
	initializeList( &bitmapList );
	initializeList( &contextList );

	/*
	 *	get a a semaphore on  the page
	 */
	header = volRec->header;
	if (waitSemaphore(&(volRec->headerLink->pageHash->semaphore)) != esmNOERROR)  {
		SM_ERROR(TYPE_FATAL, esmINTERNAL);
	}

	/*
	 *	Calculate the number of short pages to be allocated
	 */
	blockSize = BLOCKCOUNT(page2size);
	count = numPages * blockSize;
	TRPRINT(TR_IO, TR_LEVEL_2, ("blockSize:%d numFreePages:%d",
			blockSize, header->numFreePages));

	/*
	 *	Make sure there are enough pages on the volRec
	 */
	if (count > header->numFreePages) {

		signalSemaphore(&(volRec->headerLink->pageHash->semaphore));
		SM_ERROR(TYPE_WARNING, esmDISKFULL);
	    return(esmFAILURE);
	}

	/*
	 *	Extract the page number from nearPid or generate it randomly.
	 */
	if (nearPid != NULL)	{

		TRPRINT(TR_IO, TR_LEVEL_2, ("nearPid:%d", nearPid->page));

	    nearPage = nearPid->page;

	} else {

	    nearPage = volRec->lastAllocPage;
	}

	/*
	 *	check to make sure the near pid is valid
	 */
	if ((nearPage < 0) || (nearPage > header->numPages))	{

		signalSemaphore(&(volRec->headerLink->pageHash->semaphore));
		SM_ERROR(TYPE_USER, esmBADPID);
		return(esmFAILURE);
	}

	/*
	 *	Fist use a new (4/18/93) simple scheme to allocate the pages
	 *	based on searching the bitmap linearly from the near page.
	 *	If this fails, go to a track based technique.
	 *	Actually, this should not fail.  The track technique
	 *	(old method) is included in case there is a bug in the new one.
	 */
	error = IO_AllocSimpleSearch(0, numPages, blockSize, nearPage, volRec, pidList, &bitmapList, &contextList );

#ifdef JUNK
	if (error != esmNOERROR) {
		/*
		 *	Calculate the first page on the track where nearPage
		 *	resides.
		 */
		startTrack = (nearPage / header->pagesPerTrack) * header->pagesPerTrack;

		/* check to make sure that the startTrack is valid */
		if (startTrack >= header->numPages)	{
			signalSemaphore(&(volRec->headerLink->pageHash->semaphore));
			SM_ERROR(TYPE_WARNING, esmINTERNAL);
			return(esmFAILURE);
		}

		/*
		 * Two cases exist:
		 *     1. the number asked for can fit in one track.
		 *     2. the number asked for exceeds one track.
		 */
		if (blockSize < header->pagesPerTrack)	{

			TRPRINT(TR_IO, TR_LEVEL_2, ("blockSize less than track"));

			/* attempt to allocate all the pages on a single track */
			error = IO_OneTrackAll(0, count, blockSize, startTrack,
								   volRec, pidList, &bitmapList, &contextList);

			/* check to see if number of pages we want is > one */
			if ((error != esmNOERROR) && (numPages > 1))	{

				/*
				 *	Just try to get them so that the individual pages
				 *	dont cross track boundaries
				 */
				error = IO_OneTrackPages(0, numPages, blockSize, startTrack,
										 volRec, pidList, &bitmapList, &contextList);
			}

		} else {

			/* we do not handle this case yet */
			signalSemaphore(&(volRec->headerLink->pageHash->semaphore));
			SM_ERROR(TYPE_WARNING, esmTOOMANYBITS);
			return(esmFAILURE);
		}
	}
#endif /* JUNK */

	if (error) {

		bm_UnallocContext( &contextList );
		bm_ListUnfix( &bitmapList, FALSE );
		signalSemaphore(&(volRec->headerLink->pageHash->semaphore));
	    return(esmFAILURE);
	} else {
		SM_ASSERT(LEVEL_3, !LIST_EMPTY(&contextList));
		SM_ASSERT(LEVEL_3, !LIST_EMPTY(&bitmapList));
	}

	/*
	 *	log the page allocation
	 */
	if (io_LogPageAllocation(volRec, numPages, pidList, page2size, &bitmapList)) {
		error = Active->errno;

		/* unfix the bitmap pages */
		bm_UnallocContext( &contextList );
		bm_ListUnfix( &bitmapList, FALSE );
		signalSemaphore(&(volRec->headerLink->pageHash->semaphore));


		/* return failure to the caller */
		SM_ERROR(TYPE_USER, error);
		return(esmFAILURE);
	}
	bm_FreeContext( &contextList );
	SM_ASSERT(LEVEL_3, LIST_EMPTY(&bitmapList));
	SM_ASSERT(LEVEL_3, LIST_EMPTY(&contextList));

	/* mark the last allocated page */
	volRec->lastAllocPage = pidList[numPages - 1].page;
	TRPRINT(TR_IO, TR_LEVEL_2, ("lastAllocPage:%d", volRec->lastAllocPage));

	/* Update cached header fields and return */
	io_ChangeVolFreePages(volRec, -count);
	signalSemaphore(&(volRec->headerLink->pageHash->semaphore));

	/*
	 *	return success to the caller
	 */
	return(esmNOERROR);
}




/*
 * Allocate pages a number of pages by sequentially searching the
 * bitmap.
 *
 * Returns:
 *	pidList of the allocated pages
 *
 * Side Effects:
 *	page bit map updated in volume header entry.
 *
 * Errors:
 * 	eioFRAGMENTEDVOL -  vol too fragmented.
 */
 static int
IO_AllocSimpleSearch (

	int			soFar,			/* Number of pages allocated so far	*/
	int			leftToAlloc,	/* Number of pages left to allocate	*/
	PAGESIZE	blockSize,
	int			startPage,		/* where search for free pages begins*/
	VOLREC		*volRec,		/* points to the volume header entry*/
	PID			pidList[],		/* Page IDs of allocated pages		*/
	LIST		*bitmapList,	/* list of bitmap pages				*/
	LIST		*contextList 	/* info about bits set 				*/
)
{

    int			currPage;		/* beginning page of the current track*/
    int			lastPage;		/* beginning page of the next track	*/
	LIST		*pageList;
	FREECONTEXT	*context;
	BOOL		wrapped;
	int			numBits;
	int			bitsUsed;

	TRPRINT(TR_IO, TR_LEVEL_1, ("startPage:%d leftToAlloc:%d blockSize:%d",
			startPage, leftToAlloc, blockSize));

	/*
	 *	check to see if a bitmap list has been passed in
	 */
	if (bitmapList != NULL)	{
		pageList = bitmapList;
	} else {
		SM_ERROR(TYPE_FATAL, esmINTERNAL);
	}
	SM_ASSERT(LEVEL_3, contextList != NULL);

	/*
	 *	Right now, all callers will send only an empty list,
	 *  but they  need not be.
	 */
	SM_ASSERT(LEVEL_3, LIST_EMPTY(pageList));
	SM_ASSERT(LEVEL_3, LIST_EMPTY(contextList));


	/*
	 *	check to see if the volume can accomodate this size request
	 */
	if (volRec->header->pagesPerTrack < blockSize)	{

		TRPRINT(TR_IO, TR_LEVEL_2, ("track size too small"));
		return(esmFAILURE);
	}

	/*
	 *	The first search we do looks for contiguous pages.
	 *	We follow this with a search for individual pages.
	 */
	SM_ASSERT(LEVEL_3, startPage > 0);
    currPage  = startPage;
    lastPage = volRec->header->numPages-1;
	numBits = leftToAlloc * blockSize;
	if (currPage >= lastPage) {
		currPage = 0;
		wrapped = TRUE;
	} else {
		wrapped = FALSE;
	}


	/*
	 *	Search until found
	 */
    for (;;) {

		TRPRINT(TR_IO, TR_LEVEL_2, ("currPage:%d lastPage:%d", currPage, lastPage));

		/*
		 *	See if there are enough contiguous pages
		 *	to contain all of the  pages.
		 */
		context = bm_FindContigBits(volRec, numBits, currPage, lastPage, pageList);

		/* check to see if we have found a page */
		if (context != NULL)	{

			CHECK_FREECONTEXT_MAGIC(context);

			/* hang the context on the context list */
			listEnq( contextList, &(context->list) );

			/* mark the individual pages that were allocated */
			for (bitsUsed = 0; bitsUsed < numBits; bitsUsed += blockSize)	{
				/* mark the page as used */
				pidList[soFar].page  = context->bitContext[0].absoluteBitOffset + bitsUsed;
				TRPRINT(TR_IO, TR_LEVEL_2, ("alloc page:%d", pidList[soFar].page));
				pidList[soFar++].volid = volRec->header->volid;
			}

			/*
			 *	Check the consistency of the bitmap
			 *	return success
			 */
			SM_ASSERT(LEVEL_3, !io_Checker(volRec));
			TRPRINT(TR_IO, TR_LEVEL_2,
					("ALLOCATED %d pages (CONTIGUOUS) with fast search\n", numBits));
			return(esmNOERROR);
		}

		if (currPage == 0) {
			/* entire bitmap searched, so go to next method */
			break;
		} else {
			/* now try searching from the beginning of the disk */
			currPage = 0;
			lastPage = startPage;
		}
    }


	/*
	 *	No luck with contiguous pages, so try indiviual ones.
	 *
	 *	setup search conditions to look for individual pages
	 */
	SM_ASSERT(LEVEL_3, startPage > 0);
    currPage  = startPage;
    lastPage = volRec->header->numPages-1;
	if (currPage >= lastPage) {
		currPage = 0;
		wrapped = TRUE;
	} else {
		wrapped = FALSE;
	}

	/* Search until found */
    for (;;) {

		TRPRINT(TR_IO, TR_LEVEL_2, ("currPage:%d lastPage:%d", currPage, lastPage));

		/*
		 *	See if there are enough contiguous pages
		 *	on the track to contain a LONG page.
		 */
		context = bm_FindContigBits(volRec, blockSize, currPage, lastPage, pageList);

		/* check to see if we have found a page */
		if (context != NULL)	{

			/* hang the context on the context list */
			listEnq( contextList, &(context->list) );

			/* mark the page as used */
			pidList[soFar].page  = context->bitContext[0].absoluteBitOffset;
			TRPRINT(TR_IO, TR_LEVEL_2, ("alloc page:%d", pidList[soFar].page));
			pidList[soFar++].volid = volRec->header->volid;
			leftToAlloc--;
			TRPRINT(TR_IO, TR_LEVEL_2, ("leftToAlloc:%d", leftToAlloc));
			SM_ASSERT(LEVEL_3, context->bitContext[0].bitCount == blockSize);

			/* Check to see if we have allocated enough pages */
			if (leftToAlloc == 0)	{
				TRPRINT(TR_IO, TR_LEVEL_2, ("ALLOCATED %d pages (NON-CONTIG) with fast search\n", numBits));
				return(esmNOERROR);
			}

			currPage = pidList[soFar-1].page + 1;
		} else {
			currPage = lastPage + 1;
		}
	
		if (currPage >= volRec->header->numPages) {
			currPage = 0;
			wrapped = TRUE;
		} 

		/* Search completes without enough allocation */
		if (wrapped && currPage >= startPage)	{
			break;
		}
    }
	    
    /*
	 *	At this point we Know the volume was too fragmented to
	 *	satisfy the request, so we restore the bit map to its
	 *	previous state and return an error message.
	 */
	bm_UnallocContext( contextList );
	bm_ListUnfix(pageList, FALSE);

	/*
	 *	Check the consistency of the bitmap
	 */
    SM_ASSERT(LEVEL_3, !io_Checker(volRec));

	TRPRINT(TR_IO, TR_LEVEL_2, ("Could Not allocate %d pages with fast search\n", numBits));
    return(esmFAILURE);
}
