/*
 *   $RCSfile: io_DeallocPages.c,v $  
 *   $Revision: 1.1.1.1 $  
 *   $Date: 1996/05/04 21:55:46 $      
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/


#include "sysdefs.h"
#include "ess.h"
#include "checking.h"
#include "trace.h"
#include "error.h"
#include "list.h"
#include "pool.h"
#include "tid.h"
#include "io.h"
#include "lock.h"
#include "object.h"
#include "msgdefs.h"
#include "thread.h"
#include "latch.h"
#include "semaphore.h"
#include "link.h"
#include "lsn.h"
#include "bf.h"
#include "volume.h"
#include "bitmap.h"
#include "io_intfuncs.h"
#include "io_extfuncs.h"
#include "io_logfuncs.h"
#include "bm_extfuncs.h"
#include "thread_globals.h"


/* 
 *
 *	Module io_deallocpages: deallocate pages on a volume 
 *
 *   EXPORTS:
 *	io_DeallocPages	Deallocates pages on a volume
 *
 *
 *  Returns:
 *        The appropriate error code or "esmNOERROR"
 *
 *  Side Effects:
 *       VolHdr fields are updated to indicate more free pages and a dirty header.
 *
 */


 int
io_DeallocPages (

	PAGE2SIZE		page2size,		/* SHORT or LONG				*/
	int				numPages,		/* # of pages to deallocate		*/
	PID				pidList[],		/* array of PIDs to deallocate	*/
	LIST			*bitmapList,	/* list of bitmap pages			*/
	BOOL			logDeallocation,/* should logging be done		*/
	BOOL			doDeallocation 	/* should dealloc be done		*/
)
{

 	VOLREC		*volRec;		/* points to volume table entry */
 	int			i;				/* general purpose array index	*/
	VOLID		volid;			/* volume id of the first page	*/
	PAGESIZE	blockSize;
	LIST		tempList;


	TRPRINT(TR_IO, TR_LEVEL_1, ("deallocating:%d pages", numPages));

	/*
	 *	This implementation does not support both logging and
	 *	doing at the same time since a failure in logging is
	 *	hard to correct for.  The solution is to use
	 *	io_FakeDeallocPages().
	 */
	SM_ASSERT(LEVEL_3, !(logDeallocation && doDeallocation));

	volid = pidList[0].volid;

	if ((volRec = io_FindVolRec(volid)) == NULL)	{

		SM_ERROR(TYPE_USER, esmBADVOLID);
		return(esmFAILURE);
	}

#ifdef DEBUG
	SM_ASSERT(LEVEL_3, !io_Checker(volRec));
#endif


	/*
	 *	check to see if the user passed in a bitmap list
	 */
	if (bitmapList == NULL)	{

		/*
		 *	No list was passed in, so initialize a new one
		 */
		bitmapList = &tempList;
		initializeList(bitmapList);
	}

	/*
	 *	Get a semaphore on the volume header page
	 */
	if (waitSemaphore(&(volRec->headerLink->pageHash->semaphore)) != esmNOERROR) {
		SM_ERROR(TYPE_FATAL, esmINTERNAL);
	}

	/*
	 *	set the number of pages to zero
	 */
	blockSize = BLOCKCOUNT(page2size);

	/*
	 *	Deallocate all pages
	 */
	for (i = 0; i < numPages; i++)	{

		/*
		 *	check to see if the volume is consistent
		 */
		SM_ASSERT(LEVEL_3, volid == pidList[i].volid);

		/*
		 *	check to make sure the page is valid
		 */
		SM_ASSERT(LEVEL_1, (pidList[i].page >= ((volRec->header->bitmapPages * 4) + 1)));
		SM_ASSERT(LEVEL_1, (pidList[i].page < (volRec->header->numPages)));

		/*
		 *	Actually deallocate the page
		 */
		if (IO_DeallocPages(pidList[i].page, blockSize, volRec, bitmapList, doDeallocation))	{
		
			/*
			 *	free the bitmap pages if the list was not passed in
			 * 	and return an error
			 */
			if (bitmapList == &tempList) bm_ListUnfix(bitmapList, FALSE);
			signalSemaphore(&(volRec->headerLink->pageHash->semaphore));
			return(esmFAILURE);
		}
	}

	if (logDeallocation) {

        /*
         *  log the page deallocation, unfix the pages if the bitmap
		 *	list was not passed in
         */
        if (io_LogPageDeallocation(volRec, numPages, pidList, page2size, bitmapList, (bitmapList == &tempList))) {

			/*
			 *	unfix the bitmap pages if the list was not passed in
			 *	(execution will not reach this point)
			 */
			if (bitmapList == &tempList) bm_ListUnfix(bitmapList, FALSE);
			signalSemaphore(&(volRec->headerLink->pageHash->semaphore));

            /*
             *  return failure to the caller
             */
            return(esmFAILURE);
		}

	} 

	/*
	 *	unfix the bitmap pages if the list was not passed in
	 */
	if (bitmapList == &tempList) bm_ListUnfix(bitmapList, TRUE);

	/*
	 *	Update cached header fields and return
	 */
	if (doDeallocation) {
		io_ChangeVolFreePages(volRec, blockSize * numPages);
	} else {

		/*
		 *	Just dirty it so asserts won't complain about 
		 *	clean pages with log records
		 */
		volRec->dirty = TRUE;
		DIRTY_PAGE(volRec->headerLink->pageHash);
	}
	signalSemaphore(&(volRec->headerLink->pageHash->semaphore));

	/*
	 *	return success to the caller
	 */
#ifdef DEBUG
	if (LIST_EMPTY(bitmapList)) SM_ASSERT(LEVEL_3, !io_Checker(volRec));
#endif
	return(esmNOERROR);
}
