/*
 *   $RCSfile: io_FakeDeallocPages.c,v $  
 *   $Revision: 1.1.1.1 $  
 *   $Date: 1996/05/04 21:55:49 $      
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/


#include "sysdefs.h"
#include "ess.h"
#include "checking.h"
#include "trace.h"
#include "error.h"
#include "list.h"
#include "pool.h"
#include "tid.h"
#include "io.h"
#include "lock.h"
#include "object.h"
#include "msgdefs.h"
#include "thread.h"
#include "latch.h"
#include "semaphore.h"
#include "link.h"
#include "lsn.h"
#include "bf.h"
#include "volume.h"
#include "bitmap.h"
#include "deallocinfo.h"
#include "io_extfuncs.h"
#include "bf_extfuncs.h"
#include "bm_extfuncs.h"
#include "io_globals.h"


/*
 *	This function deallocates a set up pages.  It is assumed that
 *	all pages in the the set are on the same volume.
 */

 int
io_FakeDeallocPages (

	PAGE2SIZE		page2size,		/* SHORT or LONG				*/
	int				numPages,		/* # of pages to deallocate		*/
	PID				pidList[],		/* array of PIDs to deallocate	*/
	LIST			*pageDeallocList,	/* list of pages to dealloc	*/
	int				pageType 		/* type of pages to allocate 	*/
)
{

	PAGEDEALLOCINFO     *pageDeallocInfo;
	VOLREC				*volRec;
	int					i;
#ifdef DEBUG
	PID					bitmapPid;
	GROUPLINK			*groupLink;
#endif

	/*
	 *	If the pages are slotted pages, then we need information
	 *	about the volume
	 */
	if ((volRec = io_FindVolRec(pidList[0].volid)) == NULL) {
		SM_ERROR(TYPE_USER, esmBADVOLID);
		return(esmFAILURE);
	}

	/*
	 * 	Record the deallocation of each page in the list
	 */
	for (i = 0; i < numPages; i++) {

#ifdef DEBUG
		/*
		 *	Make sure the page is allocated
		 */
		bitmapPid.volid = pidList[i].volid;
		bitmapPid.page = BIT_TO_PAGE(pidList[i].page) + volRec->header->freeBitmapAddr;
		if ((groupLink = bf_ReadPage(volRec->bufGroup, &bitmapPid, BITMAP_PAGE2SIZE, BF_SEM)) == NULL) {
			SM_ERROR(TYPE_FATAL, esmINTERNAL);
		}
		if (bm_CheckSet((UFOUR*)groupLink->bufFrame, BIT_IN_PAGE(pidList[i].page))) {
			SM_ERROR(TYPE_FATAL, esmINTERNAL);
		}
		signalSemaphore(&(groupLink->pageHash->semaphore));
		bf_UnfixPage(groupLink, BF_DEFAULT, FALSE);
#endif

		/* 
		 *	If the page is a slotted page, mark the page as not longer 
		 *  slotted 
		 */
		if (pageType == PAGE_SLOTTED) {
			if (io_MarkPage(volRec, &(pidList[i]), pageType, FALSE) != esmNOERROR) {
				return(esmFAILURE);
			}
		}

		/*
		 *	Obtain a place to record that the page is to be 
		 *	deallocated at transaction commit.
		 */
		pageDeallocInfo = (PAGEDEALLOCINFO*) poolDeq(&PageDeallocInfoPool);
		if (pageDeallocInfo == NULL) {
			return(esmFAILURE);
		}
		CHECK_PAGEDEALLOCINFO_MAGIC(pageDeallocInfo);

		/*
		 *	Remember the page info and attach it to the transaction
		 */
		pageDeallocInfo->pid = pidList[i];
		pageDeallocInfo->page2size = page2size;
		listEnq(pageDeallocList, &(pageDeallocInfo->list));

		/*
		 *	If this is the first page in the dealloc list, record
		 *	the number of pages, else store 0.
		 */
		if (i == 0) {
			pageDeallocInfo->pageCount = numPages;
		} else {
			pageDeallocInfo->pageCount = 0;
		}

	}

	return(esmNOERROR);
}
