/*
 *   $RCSfile: io_FormatVolume.c,v $  
 *   $Revision: 1.1.1.1 $  
 *   $Date: 1996/05/04 21:55:46 $      
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/

#include "sysdefs.h"
#include "ess.h"
#include "checking.h"
#include "trace.h"
#include "error.h"
#include "list.h"
#include "tid.h"
#include "io.h"
#include "lock.h"
#include "object.h"
#include "msgdefs.h"
#include "thread.h"
#include "latch.h"
#include "semaphore.h"
#include "link.h"
#include "lsn.h"
#include "bf.h"
#include "volume.h"
#include "disk.h"
#include "bitmap.h"
#include "pool.h"
#include "logrecs.h"
#include "bf_globals.h"
#include "io_intfuncs.h"
#include "io_extfuncs.h"
#include "bf_extfuncs.h"
#include "disk_funcs.h"
#include "bm_extfuncs.h"
#include "bit_funcs.h"
#include "util_funcs.h"
#include "thread_globals.h"


#define INITIALIZE_LRC_TO_1(_lrc)		\
	 (void) ((_lrc)->wrapCount = 0, (_lrc)->count = 1, (_lrc))

 int
io_FormatVolume (
	VOLREC			*volRec,			
	VOLNAMEREC		*volNameRec,
	int				numCylinders,
	int				tracksPerCylinder,
	int				pagesPerTrack,
	int				logFileSize,
	PAGE2SIZE		logFileBlock2size,
	FLAGS			properties
)
{
	BOOL			volWasMounted;
	int				*intPtr;
	int				logFileBlockCount;
	int				logBlocks;
	int				i;
	BITMAPPAGE		*bitPage;
	int				bitmapPages;
	GROUPLINK		*headerLink;
	GROUPLINK		*groupLink;
	PID				headerPid;
	PID				rootPid;
	PID				bitmapPid;
	SHORTPID		page;
	SHORTPID		prevPage;
	ROOTPAGE		*rootPage;
	int				numBitMaps;	/* number of bitmaps 				*/
	int				availablePages;

	SM_ASSERT(LEVEL_1, (volNameRec!=NULL));

	TRPRINT(TR_IO, TR_LEVEL_1, ("formatting volName:%s volid:%d", 
		volNameRec->volName.name, volNameRec->volid));

	if (volRec == NULL) {
		volWasMounted = FALSE;

		/* this is the only case we support at present */

		/* We have to create a volRec for all this junk
		 * because lower layer routines require it.
		 */

		if (volRec = io_FindVolRec(volNameRec->volid)) {

			SM_ERROR(TYPE_USER, esmDISKMOUNTED);
			return(esmFAILURE);
		}

		if((volRec = IO_NewVolRec()) == NULL) {
			return esmFAILURE;
		}

		/*
		 *	get a buffer group
		 */
		if ((volRec->bufGroup = bf_OpenGroup(BITMAP_GROUP_SIZE, BF_LRU, NOFLAGS)) == NULL)	{
			SM_ERROR(TYPE_FATAL, esmINTERNAL);
			return(esmFAILURE);
		}
		volRec->volflags |= VOL_OPEN|VOL_FORMAT;
		volRec->volid = volNameRec->volid;
		volRec->volNameRec  =  volNameRec;
		IO_InsertVolRec(volRec); 
				
		/*
		 *	open the device for writing
		 */
		if ( openDisk(volNameRec->volName.name, TRUE,  volRec)==FALSE){
			return(esmFAILURE);
		}

	} else {
		volWasMounted = TRUE;
	}

	/*
	 *	construct the header pid
	 */
	headerPid.volid = volNameRec->volid;
	headerPid.page  = HEADERADDR;

	/*
	 *	get a buffer to hold the header page
	 */
	if ((headerLink = bf_ReadPage(UserBufGroup, &headerPid,
								  HEADER_PAGE2SIZE, BF_NOREAD)) == NULL)	{

		close(volRec->volLink->id);
		SM_ERROR(TYPE_FATAL, esmINTERNAL);
		return(esmFAILURE);
	}

	if (!volWasMounted) {
		/*
		 *	get a pointer to the header
		 */
		volRec->headerLink = headerLink;
		volRec->header = (VOLHDR *) headerLink->bufFrame;
	}

	/*
	 *	fill in the volRec header
	 */
	volRec->header->magic			  = VOLUME_MAGIC;
	volRec->header->volid			  = volRec->volid;
	volRec->header->numPages		  = numCylinders * tracksPerCylinder * pagesPerTrack;
	volRec->header->numCylinders	  = numCylinders;
	volRec->header->tracksPerCylinder = tracksPerCylinder;
	volRec->header->pagesPerTrack	  = pagesPerTrack;
	volRec->header->logFileBlock2size = logFileBlock2size;
	volRec->header->numLogFileBlocks  = logFileSize;
	volRec->header->dismountEndOfLog.wrapCount  = NULL_LOG_WRAP_COUNT;
	volRec->header->dismountEndOfLog.offset  = 0;
	INITIALIZE_LRC_TO_1(&(volRec->header->lrc));
	volRec->header->slottedPage2size = SLOTTED_PAGE2SIZE;
	volRec->header->lgPage2size = LG_PAGE2SIZE;
	volRec->header->lgDataPage2size = LGDATA_PAGE2SIZE;
	volRec->header->btreePage2size = BTREE_PAGE2SIZE;
	volRec->header->indexDescPage2size = INDEXDESC_PAGE2SIZE;
	volRec->header->minPage2size =  MIN_PAGE2SIZE;
	volRec->header->version =  FORMAT_VERSION;
	volRec->header->properties = properties;

	/*
	 *	calculate the bitmap size in bytes
	 */
	i = (int)((volRec->header->numPages - 1)/BITSPERBYTE) + 1;		/* bit map size in bytes */

	/*
	 *	calculate the bitmap size in pages
	 */ volRec->header->bitmapPages = bitmapPages = ((i - 1)/(BITMAP_USABLESIZE)) + 1;
	volRec->dirty = TRUE;

	/*
	 *	fill in the bitmap locations
	 */
	numBitMaps = 0;
	volRec->header->freeBitmapAddr = BITMAPADDR;
	numBitMaps++;
	TRPRINT(TR_IO, TR_LEVEL_2, ("free map at:%d", volRec->header->freeBitmapAddr));
	volRec->header->uniqueArrayAddr = BITMAPADDR + (numBitMaps * bitmapPages);
	numBitMaps++;
	TRPRINT(TR_IO, TR_LEVEL_2, ("unique map at:%d", volRec->header->uniqueArrayAddr));
	volRec->header->slotBitmapAddr = BITMAPADDR + (numBitMaps * bitmapPages);
	numBitMaps++;
	TRPRINT(TR_IO, TR_LEVEL_2, ("slot map at:%d", volRec->header->slotBitmapAddr));
	volRec->header->fileBitmapAddr = BITMAPADDR + (numBitMaps * bitmapPages);
	numBitMaps++;
	TRPRINT(TR_IO, TR_LEVEL_2, ("file map at:%d", volRec->header->fileBitmapAddr));

	/*
	 *	allocate the log if it is necessary
	 */
	if (logFileSize > 0)	{

		/*
		 *	allocate a block for the log control information
		 */
		volRec->header->logControlAddr = BITMAPADDR + (numBitMaps * bitmapPages);
		TRPRINT(TR_IO, TR_LEVEL_2, ("file map at:%d", volRec->header->logControlAddr));

		/*
		 *	calculate the start of the log file
		 */
		volRec->header->logFileAddr = BITMAPADDR + (numBitMaps * bitmapPages) + 1;
		TRPRINT(TR_IO, TR_LEVEL_2, ("log file at:%d", volRec->header->logFileAddr));

		/*
		 *	calculate the number of blocks in the file
		 */
		logFileBlockCount = RAISETWO(logFileBlock2size - MIN_PAGE2SIZE) * logFileSize;
		TRPRINT(TR_IO, TR_LEVEL_2, ("logFileBlockCount:%d", logFileBlockCount));

		/*
		 *	calculate the number of blocks allocated to the file
		 */
		logBlocks = logFileBlockCount + 1/*for control page*/;

	} else if (logFileSize == -1) {

		availablePages = (int)volRec->header->numPages - 
						 (numBitMaps * bitmapPages) - 
						 (2 /* vol header and root page*/ ) - 
						 1/* for logControlAddr */; 

		if (availablePages < RAISETWO(logFileBlock2size - MIN_PAGE2SIZE)) {
			SM_ERROR(TYPE_WARNING, esmBADVOLHEADER);
			return(esmFAILURE);
		}

		/*
		 *	calculate the number of blocks allocated to the file
		 */
		logFileSize = (int) (availablePages / 
			RAISETWO(logFileBlock2size - MIN_PAGE2SIZE)); 

		volRec->header->numLogFileBlocks  = logFileSize;
		logFileBlockCount = RAISETWO(logFileBlock2size - MIN_PAGE2SIZE) * logFileSize;
		logBlocks = logFileBlockCount + 1/*for control page*/;

		/*
		 *	allocate a block for the log control information
		 */
		volRec->header->logControlAddr = BITMAPADDR + (numBitMaps * bitmapPages);
		TRPRINT(TR_IO, TR_LEVEL_2, ("file map at:%d", volRec->header->logControlAddr));

		/*
		 *	calculate the start of the log file
		 */
		volRec->header->logFileAddr = BITMAPADDR + (numBitMaps * bitmapPages) + 1;
		TRPRINT(TR_IO, TR_LEVEL_2, ("log file at:%d", volRec->header->logFileAddr));

	} else {

		logBlocks = 0;
		volRec->header->logFileAddr = NULLPID;
	}

	/*
	 *	calculate the number of free bits
	 */
	volRec->header->numFreePages =
		(volRec->header->numPages - 2/*vol header and root page*/) - (numBitMaps * bitmapPages) - (logBlocks);

	/*
	 *	Read in the root page
	 */
	rootPid.volid = volNameRec->volid;
	rootPid.page  = ROOTPAGEADDR;
	if ((groupLink = bf_ReadPage(UserBufGroup, &rootPid,
								  MIN_PAGE2SIZE, BF_NOREAD)) == NULL)	{

		close(volRec->volLink->id);
		SM_ERROR(TYPE_FATAL, esmINTERNAL);
		return(esmFAILURE);
	}

	/*
	 *	Initialize the volRec root page
	 */
	rootPage = (ROOTPAGE*)groupLink->bufFrame;
	IO_InitRoot(rootPage);
	INITIALIZE_LRC_TO_1(&(rootPage->lrc));

	/*
	 *	Unfix the root page
	 */
	bf_UnfixPage(groupLink, BF_DEFAULT, TRUE);


	/*
	 *	mark all the pages as free
	 *	note that for the free bitmap, a 1 bit denotes a free page
	 */
	bitmapPid.volid = volNameRec->volid;
	bitmapPid.page  = volRec->header->freeBitmapAddr;
	if (IO_MarkBitmapRange(UserBufGroup, &bitmapPid, 0, 
						   (bitmapPages * BITMAP_USABLESIZE * BITSPERBYTE), 0)) {
		close(volRec->volLink->id);
		SM_ERROR(TYPE_FATAL, esmINTERNAL);
		return(esmFAILURE);
	}
	if (IO_MarkBitmapRange(UserBufGroup, &bitmapPid, 0, 
						   (int) volRec->header->numPages, 1)) {
		close(volRec->volLink->id);
		SM_ERROR(TYPE_FATAL, esmINTERNAL);
		return(esmFAILURE);
	}
	
	/*
	 *	Mark the pages occupied by the header, the first copy of the 
	 *	bitmap, and the root page of the directory as USED
	 */
	prevPage = NULLPID;
	groupLink = NULL;
	for (i = 0; i < (2 + (numBitMaps * bitmapPages)); i++) 	{
		
		bitmapPid.page = BIT_TO_PAGE(i) + volRec->header->freeBitmapAddr;

		if (bitmapPid.page != prevPage) {

			/*
			 *  Unfix the previous page
			 */
			if (groupLink != NULL) {
				bf_UnfixPage(groupLink, BF_DEFAULT, TRUE);
			}

			/*
			 *	Read in the appropriate bitmap page
			 */
			groupLink = bf_ReadPage(UserBufGroup, &bitmapPid, BITMAP_PAGE2SIZE, NOFLAGS);
			if (groupLink == NULL) {
				close(volRec->volLink->id);
				SM_ERROR(TYPE_FATAL, esmINTERNAL);
				return(esmFAILURE);
			}
			prevPage = bitmapPid.page;
		}

		bm_ClearBit((UFOUR *) groupLink->bufFrame, BIT_IN_PAGE(i));
	}

	/*
	 *  Unfix the previous page
	 */
	if (groupLink != NULL) {
		bf_UnfixPage(groupLink, BF_DEFAULT, TRUE);
	}

	/*
	 *	clear the other bitmaps
	 */
	bitmapPid.volid = volNameRec->volid;
	bitmapPid.page  = volRec->header->fileBitmapAddr;
	if (IO_MarkBitmapRange(UserBufGroup, &bitmapPid, 0, 
						   (bitmapPages * BITMAP_USABLESIZE * BITSPERBYTE), 0)) {
		close(volRec->volLink->id);
		SM_ERROR(TYPE_FATAL, esmINTERNAL);
		return(esmFAILURE);
	}
	bitmapPid.page  = volRec->header->slotBitmapAddr;
	if (IO_MarkBitmapRange(UserBufGroup, &bitmapPid, 0, 
						   (bitmapPages * BITMAP_USABLESIZE * BITSPERBYTE), 0)) {
		close(volRec->volLink->id);
		SM_ERROR(TYPE_FATAL, esmINTERNAL);
		return(esmFAILURE);
	}

	/*
	 *	clear the log file map if necessary
	 */
	if (logFileSize > 0)	{

		/*
		 *	Determine the correct bitmap page
		 */
		bitmapPid.page = BIT_TO_PAGE(volRec->header->logControlAddr) + 
						 volRec->header->freeBitmapAddr;
		/*
		 *	Read in the appropriate bitmap page
		 */
		groupLink = bf_ReadPage(UserBufGroup, &bitmapPid, BITMAP_PAGE2SIZE, NOFLAGS);
		if (groupLink == NULL) {
			close(volRec->volLink->id);
			SM_ERROR(TYPE_FATAL, esmINTERNAL);
			return(esmFAILURE);
		}

		/*
		 *	clear the log control information
		 */
		bm_ClearBit((UFOUR *) groupLink->bufFrame, 
					BIT_IN_PAGE(volRec->header->logControlAddr));

		/*
		 *	Unfix the bitmap page
		 */
		bf_UnfixPage(groupLink, BF_DEFAULT, TRUE);
		
		/*
		 *	mark the pages of the log as used
		 */
		prevPage = NULLPID;
		groupLink = NULL;
		for (i = 0; i < logFileBlockCount; i++) 	{

			/*
			 *	calculate the bit offset taking into account
			 *	the LSN's on the bitmap pages
			 */
		/*	bit = BIT_TO_PAGE(i) * (BITMAP_PAGESIZE * 8) + BIT_IN_PAGE(bit); */
		/*  bm_ClearBit((UFOUR *) volRec->freeBitmap, i + volRec->header->logFileAddr); */

			/*
			 *	Page to have bit marked for
			 */
			page = i + volRec->header->logFileAddr;

			bitmapPid.page = BIT_TO_PAGE(page) + 
							 volRec->header->freeBitmapAddr;

			if (bitmapPid.page != prevPage) {

				/*
				 *  Unfix the previous page
				 */
				if (groupLink != NULL) {
					bf_UnfixPage(groupLink, BF_DEFAULT, TRUE);
				}

				/*
				 *	Read in the appropriate bitmap page
				 */
				groupLink = bf_ReadPage(UserBufGroup, &bitmapPid, BITMAP_PAGE2SIZE, NOFLAGS);
				if (groupLink == NULL) {
					close(volRec->volLink->id);
					SM_ERROR(TYPE_FATAL, esmINTERNAL);
					return(esmFAILURE);
				}
				prevPage = bitmapPid.page;
			}

			bm_ClearBit((UFOUR *) groupLink->bufFrame, BIT_IN_PAGE(page));
		}

		/*
		 *  Unfix the previous page
		 */
		if (groupLink != NULL) {
			bf_UnfixPage(groupLink, BF_DEFAULT, TRUE);
		}
	}

	/*
	 *	Fill in random numbers in all the unique bitmap pages,
	 *	zero out their lrc's
	 */
	groupLink = NULL;
	for (page = volRec->header->uniqueArrayAddr; 
		 page < volRec->header->uniqueArrayAddr + bitmapPages; 
		 page++)	{

		/*
		 *	Read in the current unique page
		 */
		bitmapPid.page = page;
		groupLink = bf_ReadPage(UserBufGroup, &bitmapPid, BITMAP_PAGE2SIZE, BF_NOREAD);
		if (groupLink == NULL) {
			close(volRec->volLink->id);
			SM_ERROR(TYPE_FATAL, esmINTERNAL);
			return(esmFAILURE);
		}

		/*
		 *	initialize the unique bits to random numbers
		 */
		intPtr = (int *)groupLink->bufFrame;
		for (i = 0; i < (BITMAP_USABLESIZE/BYTESPERWORD); 
			 i++, intPtr++)	{

			*intPtr = (int) random();
		}

		/*
		 *	zero out the lrc on the unique bitmap
		 */
		bitPage = (BITMAPPAGE *) groupLink->bufFrame;
		INITIALIZE_LRC_TO_1(&(bitPage->lrc));

		/*
		 *  Unfix the unique page
		 */
		if (groupLink != NULL) {
			bf_UnfixPage(groupLink, BF_DEFAULT, TRUE);
		}

	}

	/*
	 *	Initialize the lrc's on all all free map pages
	 */
	groupLink = NULL;
	for (page = volRec->header->freeBitmapAddr; 
		 page < volRec->header->freeBitmapAddr + bitmapPages; 
		 page++)	{

		/*
		 *	Read in the current unique page
		 */
		bitmapPid.page = page;
		groupLink = bf_ReadPage(UserBufGroup, &bitmapPid, BITMAP_PAGE2SIZE, BF_NOREAD);
		if (groupLink == NULL) {
			close(volRec->volLink->id);
			SM_ERROR(TYPE_FATAL, esmINTERNAL);
			return(esmFAILURE);
		}

		/*
		 *	zero out the lrc on the unique bitmap
		 */
		bitPage = (BITMAPPAGE *) groupLink->bufFrame;
		INITIALIZE_LRC_TO_1(&(bitPage->lrc));

		/*
		 *  Unfix the unique page
		 */
		if (groupLink != NULL) {
			bf_UnfixPage(groupLink, BF_DEFAULT, TRUE);
		}

	}

	/*
	 *	Initialize the lrc's on all file bitmap pages
	 */
	groupLink = NULL;
	for (page = volRec->header->fileBitmapAddr; 
		 page < volRec->header->fileBitmapAddr + bitmapPages; 
		 page++)	{

		/*
		 *	Read in the current unique page
		 */
		bitmapPid.page = page;
		groupLink = bf_ReadPage(UserBufGroup, &bitmapPid, BITMAP_PAGE2SIZE, BF_NOREAD);
		if (groupLink == NULL) {
			close(volRec->volLink->id);
			SM_ERROR(TYPE_FATAL, esmINTERNAL);
			return(esmFAILURE);
		}

		/*
		 *	zero out the lrc on the unique bitmap
		 */
		bitPage = (BITMAPPAGE *) groupLink->bufFrame;
		INITIALIZE_LRC_TO_1(&(bitPage->lrc));

		/*
		 *  Unfix the unique page
		 */
		if (groupLink != NULL) {
			bf_UnfixPage(groupLink, BF_DEFAULT, TRUE);
		}

	}

	/*
	 *	Initialize the lrc's on all all slot bitmap pages
	 */
	groupLink = NULL;
	for (page = volRec->header->slotBitmapAddr; 
		 page < volRec->header->slotBitmapAddr + bitmapPages; 
		 page++)	{

		/*
		 *	Read in the current unique page
		 */
		bitmapPid.page = page;
		groupLink = bf_ReadPage(UserBufGroup, &bitmapPid, BITMAP_PAGE2SIZE, BF_NOREAD);
		if (groupLink == NULL) {
			close(volRec->volLink->id);
			SM_ERROR(TYPE_FATAL, esmINTERNAL);
			return(esmFAILURE);
		}

		/*
		 *	zero out the lrc on the unique bitmap
		 */
		bitPage = (BITMAPPAGE *) groupLink->bufFrame;
		INITIALIZE_LRC_TO_1(&(bitPage->lrc));

		/*
		 *  Unfix the unique page
		 */
		if (groupLink != NULL) {
			bf_UnfixPage(groupLink, BF_DEFAULT, TRUE);
		}

	}

	DIRTY_PAGE(headerLink->pageHash);

	if (!volWasMounted) {
		/* this func was called with volRec == NULL, which
		 * means that the volume wasn't already mounted, and we
		 * pseudo-mounted it here, including creating a bufGroup
		 * for it, so we have to dismount it now.
		 */

		/*
		 *	Do some defensive error checking
		 */
		if (IO_Checker(volRec)) {
			SM_ERROR(TYPE_FATAL, esmINTERNAL);
			return(esmFAILURE);
		}

		/* notify any waiters */
		notify( &(volRec->tcbList), esmNOERROR, esmNOERROR );

		/* Flush the header */
		if (IO_DismountVolume(volRec))	{
			bf_DiscardPage(UserBufGroup, headerLink);
			SM_ERROR(TYPE_FATAL, esmINTERNAL);
			return(esmFAILURE);
		}
    } else {
		/* don't dismount it, because it was mounted when we entered this func */
        /*  release and dirty the header page */
        bf_UnfixPage(headerLink, BF_DEFAULT, TRUE);
    }


	return(esmNOERROR);
}
