/*
 *   $RCSfile: io_LogPage.c,v $  
 *   $Revision: 1.1.1.1 $  
 *   $Date: 1996/05/04 21:55:49 $      
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/

#include "sysdefs.h"
#include "ess.h"
#include "checking.h"
#include "trace.h"
#include "error.h"
#include "list.h"
#include "pool.h"
#include "tid.h"
#include "io.h"
#include "lock.h"
#include "object.h"
#include "msgdefs.h"
#include "thread.h"
#include "semaphore.h"
#include "latch.h"
#include "link.h"
#include "lsn.h"
#include "bf.h"
#include "volume.h"
#include "openlog.h"
#include "trans.h"
#include "logrecs.h"
#include "logaction.h"
#include "log.h"
#include "bitmap.h"
#include "log_extfuncs.h"
#include "bf_extfuncs.h"
#include "bm_extfuncs.h"
#include "io_logfuncs.h"
#include "thread_globals.h"

 int
io_LogPageDeallocation (

	VOLREC				*volRec,
	int					numPages,
	PID					pidList[],
	PAGE2SIZE			page2size,
	LIST				*bitmapList,
	BOOL				unfixBitmap 
)
{

	register FORCEMARK	forceMark;
	LSN					lsn;
	LOGRECORDINFO		recordInfo;
	LRC                 *lrcList;
	LRC                 lrcShortList[LRC_SHORTLIST_SIZE];
	int                 lrcCount;
	LRC                 *headerLRC; /* LRC on volume header */
	PAGEALLOCINFO		pageAllocInfo;


	TRPRINT(TR_IO, TR_LEVEL_1, ("logging pid:%d", pidList->page));

    /*
     *  increment the lrc on each bitmap page, and count the
     *  number of bitmap pages
     */
    bm_ListIncrementCount(bitmapList, &lrcCount);

    /*
     *  Create the lrcList.  If its short, use the one on the stack,
     *  otherwise malloc space and remember to free it before returning.
     *  This is necessary, because the list could potentially be very
     *  large, but usually it will be small.
     */
    if (lrcCount <= LRC_SHORTLIST_SIZE) {
        lrcList = lrcShortList;
    } else {
        lrcList = (LRC*) malloc(lrcCount * sizeof(LRC));
        if (lrcList == NULL) {
            SM_ERROR(TYPE_FATAL, esmMALLOCFAILED);
        }
    }

    /*
     *  Store the lrc for each bitmap page in the lrc list
     */
    bm_ListSaveLrc(bitmapList, lrcList);

    /*
     *  Increment the lrc on the volume header
     */
    headerLRC = &(volRec->header->lrc);
    INCREMENT_LRC(headerLRC);

	/*
	 *	Set up general information on the allocation
	 */
	pageAllocInfo.numPages = numPages;
	pageAllocInfo.page2size = page2size;
	pageAllocInfo.firstBitmapPage = volRec->header->freeBitmapAddr;
	pageAllocInfo.volumeHdrPage = volRec->headerLink->pageHash->hashList.hashPid.page;
	pageAllocInfo.headerLRC = *headerLRC;

	/*
	 *	initialize the log information
	 */
	recordInfo.action		= LOG_ACTION_DEALLOCATE_PAGE;
	recordInfo.type			= LOG_REC_TYPE_USER;
	recordInfo.flags		= NOFLAGS;
	recordInfo.imageCount	= 3;
    recordInfo.actionPid    = NULL;
    recordInfo.imageSize[0] = sizeof(PAGEALLOCINFO);
    recordInfo.imageData[0] = (VOID *) &pageAllocInfo;
    recordInfo.imageSize[1] = (sizeof(PID) * numPages);
    recordInfo.imageData[1] = (VOID *) pidList;
    recordInfo.imageSize[2] = sizeof(LRC) * lrcCount;
    recordInfo.imageData[2] = (VOID *) lrcList;

	/*
	 *	write the record to the log
	 */
	if ((forceMark = writeLogRecord((TRANSREC *) Active->transRec, &recordInfo,
							  &lsn, NOFLAGS)) < 0)	{

		/*
		 *	free any malloced list space and return failure
		 */
        if (lrcList != lrcShortList) free((char*)lrcList);
		return(esmFAILURE);
	}

	if (unfixBitmap) {
		/*
		 *	release the bitmap pages and mark them
		 */
		bm_ListUnfixDepend(bitmapList, forceMark, &lsn, lrcList);
	} else {
		/*
		 *	just mark them
		 */
		bm_ListDepend(bitmapList, forceMark, &lsn, lrcList);
	}

    /*
     *  Mark the volume header dependency
     */
    DEPEND_LOG(volRec->headerLink->pageHash, forceMark, &lsn, headerLRC);

	/*
	 *	free any malloced list space and return success
	 */
    if (lrcList != lrcShortList) free((char*)lrcList);
	return(esmNOERROR);
}

 int
io_LogPageAllocation (

	VOLREC				*volRec,
	int					numPages,
	PID					pidList[],
	PAGE2SIZE			page2size,
	LIST				*bitmapList 
)
{

	register FORCEMARK	forceMark;
	LSN					lsn;
	LRC					*lrcList;
	LRC					lrcShortList[LRC_SHORTLIST_SIZE];
	int					lrcCount;
	LRC					*headerLRC;	/* LRC on volume header */
	LOGRECORDINFO		recordInfo;
	PAGEALLOCINFO		pageAllocInfo;


	TRPRINT(TR_IO, TR_LEVEL_1, ("logging pid:%d", pidList->page));

	/*
	 *	increment the lrc on each bitmap page, and count the
	 *	number of bitmap pages
	 */
	bm_ListIncrementCount(bitmapList, &lrcCount);

	/*
	 *	Create the lrcList.  If its short, use the one on the stack,
	 *	otherwise malloc space and remember to free it before returning.
	 *	This is necessary, because the list could potentially be very
	 *	large, but usually it will be small.
	 */
	if (lrcCount <= LRC_SHORTLIST_SIZE) {
		lrcList = lrcShortList;
	} else {
		lrcList = (LRC*) malloc(lrcCount * sizeof(LRC));
		if (lrcList == NULL) {
			SM_ERROR(TYPE_FATAL, esmMALLOCFAILED);
		}
	}

	/*
	 *	Store the lrc for each bitmap page in the lrc list
	 */
	bm_ListSaveLrc(bitmapList, lrcList);

	/*
	 *	Increment the lrc on the volume header
	 */
	headerLRC = &(volRec->header->lrc);
	INCREMENT_LRC(headerLRC);

	/*
	 *	Set up general information on the allocation
	 */
	pageAllocInfo.numPages = numPages;
	pageAllocInfo.page2size = page2size;
	pageAllocInfo.firstBitmapPage = volRec->header->freeBitmapAddr;
	pageAllocInfo.volumeHdrPage = volRec->headerLink->pageHash->hashList.hashPid.page;
	pageAllocInfo.headerLRC = *headerLRC;

	/*
	 *	initialize the log information
	 */
	recordInfo.action		= LOG_ACTION_ALLOCATE_PAGE;
	recordInfo.type			= LOG_REC_TYPE_USER;
	recordInfo.flags		= NOFLAGS;
	recordInfo.imageCount	= 3;
	recordInfo.actionPid	= NULL;
	recordInfo.imageSize[0] = sizeof(PAGEALLOCINFO);
	recordInfo.imageData[0] = (VOID *) &pageAllocInfo;
	recordInfo.imageSize[1] =  (sizeof(PID) * numPages);
	recordInfo.imageData[1] = (VOID *) pidList;
	recordInfo.imageSize[2] = sizeof(LRC) * lrcCount;
	recordInfo.imageData[2] = (VOID *) lrcList;

	/*
	 *	write the record to the log
	 */
	if ((forceMark = writeLogRecord((TRANSREC *) Active->transRec, &recordInfo,
							  &lsn, NOFLAGS)) < 0)	{

		/*
		 *	free any malloced list space and return failure
		 */
		if (lrcList != lrcShortList) free((char*)lrcList);
		return(esmFAILURE);
	}

	/*
	 *	release the bitmap pages and mark them
	 */
	bm_ListUnfixDepend(bitmapList, forceMark, &lsn, lrcList);

	/*
	 *	Mark the volume header dependency
	 */
	DEPEND_LOG(volRec->headerLink->pageHash, forceMark, &lsn, headerLRC);

	/*
	 *	free any malloced list space and return success
	 */
	if (lrcList != lrcShortList) free((char*)lrcList);
	return(esmNOERROR);
}
