/*
 *   $RCSfile: io_MountVolume.C,v $  
 *   $Revision: 1.1.1.1 $  
 *   $Date: 1996/05/04 21:55:46 $      
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/

#include "sysdefs.h"
#include "ess.h"
#include "checking.h"
#include "trace.h"
#include "error.h"
#include "list.h"
#include "pool.h"
#include "tid.h"
#include "io.h"
#include "lock.h"
#include "object.h"
#include "msgdefs.h"
#include "thread.h"
#include "latch.h"
#include "semaphore.h"
#include "link.h"
#include "lsn.h"
#include "bf.h"
#include "volume.h"
#include "trans.h"
#include "logrecs.h"
#include "openlog.h"
#include "io_intfuncs.h"
#include "io_extfuncs.h"
#include "io_globals.h"
#include "log_extfuncs.h"
#include "recover_extfuncs.h"
#include "thread_globals.h"
#include "trans_globals.h"
#include "log_globals.h"


VOLREC *
finishMount(
	VOLNAMEREC		*volNameRec,
	BOOL			volumeIsLog,
	BOOL			logMount
)
{

	VOLREC			*volRec;
	FORCEMARK		forceMark;
	LOGRECORDINFO	recordInfo;
	LSN				lsn;

	TRPRINT(TR_IO, TR_LEVEL_1, ("mounting volume:%d:%s", volNameRec->volid,
		volNameRec->volName.name));

	volRec = IO_NewVolRec();
	if(volRec==NULL) {
		return NULL;
	}
	volRec->volNameRec = volNameRec;

	/*
	 * We want the volume to be found in the table (and its open awaited)
	 * so that if 2 theads "simultaneously" try to open the same
	 * volume, we don't get 2 copies of everything.
	 * We put it in the table right away, and mark it as opening(VOL_INIT).
	 */
	volRec->volflags = VOL_INIT;
	volRec->volid = volNameRec->volid;
	SM_ASSERT(LEVEL_1, (volRec->volid != 0));

	IO_InsertVolRec(volRec);

	/* 
	 * mount the volume and check for header consistency
	 */
	if (IO_MountVolume(volRec, TRUE) != esmNOERROR )	{

		/*
		 *	signal any waiters
		 */
		notify( &(volRec->tcbList), esmFAILURE, Active->errno);
		IO_RemoveVolRec(volRec);
		IO_FreeVolRec(volRec);
		return NULL;
	}

	/* volRec->header is now filled in */
	volNameRec->volid = volRec->header->volid;
	SM_ASSERT(LEVEL_1, (volNameRec->volid != 0) );

	/*
	 *	Format (initialize) the volume if it is temporary
	 */
	if (volRec->volflags & VOL_TEMP) {
		if (io_FormatVolume(volRec,
			volNameRec,
			(int) volRec->header->numCylinders,
			(int) volRec->header->tracksPerCylinder,
			(int) volRec->header->pagesPerTrack,
			(int) volRec->header->numLogFileBlocks,
			volRec->header->logFileBlock2size,
			volRec->header->properties
			) != esmNOERROR) {
			SM_ERROR(TYPE_CRASH, esmINTERNAL);
		}
	}


	/*
	 *	Make sure the current end of the log is past the
	 *	last log record generated for the volume (assuming the 
	 *	volume is not the log volume).
	 */
	if ( !formatvolRunning && !volumeIsLog && 
		compareLSN(&(volRec->header->dismountEndOfLog), 
		&(OpenLog.nextValidLSN)) > 0) {
		
		/*
		 *	The volume cannot be mounted until the current end of
		 *	the log has been advanced.  So, dismount the volume.
		 */
		if (IO_DismountVolume(volRec) != esmNOERROR) {
			SM_ERROR(TYPE_CRASH, esmINTERNAL);
		}

		SM_ERROR(TYPE_USER, esmMOUNTLSNTOOHIGH); 

		{
			/* GROT: cfront can't handle newlines in a single string,
			 * so in order to keep this readable, we'll break it
			 * up into several writes.
			 */
		fprintf(sm_ErrorStream,
			"SERVER ERROR: volume %s cannot be mounted since the current\n",
			volNameRec->volName.name);
		fprintf(sm_ErrorStream,
			"\tlog wrap count, %d, is too small.  To mount this volume, shut down\n",
			OpenLog.nextValidLSN.wrapCount);

		fprintf(sm_ErrorStream,
			"\tthe server and restart it with a wrapcount of at least %d, or\n",
			(volRec->header->dismountEndOfLog.wrapCount+1));
		fprintf(sm_ErrorStream,
			"\twith this volume in the server's configuration files.\n");
		}

		/*
		 *	signal any waiters
		 */
		notify( &(volRec->tcbList), esmFAILURE, Active->errno);

		return NULL;
	}

	volRec->lastAllocPage = 0;

	/* 
	 * flags must be set before io_Checker, because io_Checker
	 * does I/O, which expects to find this volRec in the table and open.
	 */
	volRec->volflags &=  ~VOL_INIT;
	volRec->volflags |=  VOL_OPEN;

	/*
	 *	Make sure the volume is ok, assuing recovery is not in progress
	 */
	if ((TransInfo.logState == L_ACTIVE) && 
		(io_Checker(volRec) != esmNOERROR)) {
		
		/*
		 *	This should really clean up and continue
		 */
		SM_ERROR(TYPE_CRASH, esmINTERNAL);
	}

	/*
	 *	notify any waiting processes that the disk is ready
	 */
	notify( &(volRec->tcbList), esmNOERROR, esmNOERROR );

	/*
	 *	See if the mount should be logged, logging will not
	 *	be done if the volume is a temporary one.
	 */
	if (logMount && !(volRec->volflags & VOL_TEMP)) {

		/*
		 *	Generate a log record to record the mount
		 */
		TRPRINT(TR_IO, TR_LEVEL_1, ("logging mount of volume:%d", 
			volRec->header->volid));

		/*
		 *  initialize the log information
		 */
		recordInfo.action       = 0;
		recordInfo.type         = LOG_REC_TYPE_MOUNT_VOLUME;
		recordInfo.imageCount   = 2;
		recordInfo.actionPid    = NULL;
		recordInfo.actionLRC    = NULL;
		recordInfo.imageSize[0] = sizeof(VOLID);
		recordInfo.imageData[0] = (VOID *) &(volRec->header->volid);
		recordInfo.imageSize[1] = sizeof(volRec->volNameRec->volName.name);
		recordInfo.imageData[1] = (VOID *) volRec->volNameRec->volName.name;
		recordInfo.flags        = NOFLAGS;

		/*
		 *  write the record to the log
		 */
		if ((forceMark = writeNonTransLogRecord(&recordInfo, &lsn, NOFLAGS)) < 0)  {

			SM_ERROR(TYPE_FATAL, esmINTERNAL);
			return(NULL);
		}

	}

	return volRec;
}

 VOLREC
*io_MountLogVolume (
	VOLID	volid		
)
{
	VOLNAMEREC		*volNameRec;
	VOLREC			*volRec;

	TRPRINT(TR_IO, TR_LEVEL_1, ("mounting log volume:%d", volid));

	/*
	 *	Make sure the log isn't already mounted.
	 */
	if( (volRec = io_FindVolRec(volid)) != NULL ) {
		TRPRINT(TR_IO, TR_LEVEL_2, 
			("log already mounted:%d", volRec->header->volid));

		SM_ERROR(TYPE_FATAL, esmINTERNAL);
		/*
		return volRec;
		*/
	}

	/* have we ever heard of this volid? */
	volNameRec = io_FindVolNameById( volid );

	if(volNameRec==NULL) {
		fprintf(stderr, 
			"Missing format option for volume with volid %d.\n", 
			volid);
		SM_ERROR(TYPE_USER, esmNOSUCHVOLUME);
		return NULL;
	}
	
	return finishMount(volNameRec, TRUE, FALSE);
}

 VOLREC
*io_MountVolume (
	VOLID	volid,
	BOOL	logMount
)
{

	VOLREC			*volRec;
	VOLNAMEREC		*volNameRec;

	TRPRINT(TR_IO, TR_LEVEL_1, ("mounting volume:%d", volid));

	/*
	 *	Is this volume mounted ?
	 */
	if( (volRec = io_FindVolRec(volid)) != NULL ) {
		TRPRINT(TR_IO, TR_LEVEL_2, 
			("disk already mounted:%d", volRec->header->volid));

		SM_ASSERT(LEVEL_3, TransInfo.logState != L_RECOVER);

		return volRec;
	}

	/* have we ever heard of this volid? */
	volNameRec = io_FindVolNameById( volid );

	if(volNameRec==NULL) {
		SM_ERROR(TYPE_USER, esmNOSUCHVOLUME);
		return NULL;
	}

	return finishMount(volNameRec, FALSE, logMount);
}
