/*
 *   $RCSfile: io_ReallocPages.c,v $  
 *   $Revision: 1.1.1.1 $  
 *   $Date: 1996/05/04 21:55:49 $      
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/


#include "sysdefs.h"
#include "ess.h"
#include "checking.h"
#include "trace.h"
#include "error.h"
#include "list.h"
#include "pool.h"
#include "tid.h"
#include "io.h"
#include "lock.h"
#include "object.h"
#include "msgdefs.h"
#include "thread.h"
#include "latch.h"
#include "semaphore.h"
#include "link.h"
#include "lsn.h"
#include "bf.h"
#include "volume.h"
#include "bitmap.h"
#include "io_intfuncs.h"
#include "io_extfuncs.h"
#include "bm_extfuncs.h"


/* 
 *
 *	Module io_reallocpages: reallocate pages on a volume 
 *
 *   EXPORTS:
 *	io_ReallocPages		Reallocates pages on a volume
 *
 *
 *  Returns:
 *        The appropriate error code or "esmNOERROR"
 *
 */


 int
io_ReallocPages (

	PAGE2SIZE		page2size,		/* SHORT or LONG				*/
	int				numPages,		/* # of pages to deallocate		*/
	PID				pidList[],		/* array of PIDs to deallocate	*/
	LIST			*bitmapList 	/* list of bitmap pages			*/
)
{

 	register VOLREC	*volRec;		/* points to volume table entry */
 	register int	i;				/* general purpose array index	*/
	VOLID			volid;			/* volume id of the first page	*/
	PAGESIZE		blockSize;
	LIST			tempList;


	TRPRINT(TR_IO, TR_LEVEL_1, ("reallocating:%d pages ", numPages));

	volid = pidList[0].volid;

	if ((volRec = io_FindVolRec(volid)) == NULL)	{

		SM_ERROR(TYPE_USER, esmBADVOLID);
		return(esmFAILURE);
	}
#ifdef DEBUG
	SM_ASSERT(LEVEL_3, !io_Checker(volRec));
#endif

	/*
	 *	check to see if the user passed in a bitmap list
	 */
	if (bitmapList == NULL)	{

		/*
		 *	No list was passed in, so initialize a new one
		 */
		bitmapList = &tempList;
		initializeList(bitmapList);
	}

	/*
	 *	set the number of pages to zero
	 */
	blockSize = BLOCKCOUNT(page2size);

	/*
	 *	Reallocate all pages
	 */
	for (i = 0; i < numPages; i++)	{

		/*
		 *	check to see if the volume is consistent
		 */
		SM_ASSERT(LEVEL_3, volid == pidList[i].volid);

		/*
		 *	check to make sure the page is valid
		 */
		SM_ASSERT(LEVEL_1, (pidList[i].page >= ((volRec->header->bitmapPages * 4) + 1)));
		SM_ASSERT(LEVEL_1, (pidList[i].page < (volRec->header->numPages)));

		/*
		 *	Actually Reallocate the page
		 */
		if (IO_ReallocPages(pidList[i].page, blockSize, volRec, bitmapList))	{

			return(esmFAILURE);
		}
	}


	/*
	 *	unfix the bitmap pages if the list was not passed in
	 */
	if (bitmapList == &tempList) bm_ListUnfix(bitmapList, TRUE);

	/*
	 *	Update cached header fields and return
	 */
	io_ChangeVolFreePages(volRec, -(blockSize * numPages));
#ifdef JUNK
	volRec->header->numFreePages -= blockSize * numPages;
	volRec->dirty = TRUE;
	DIRTY_PAGE(volRec->headerLink->pageHash);
#endif

	/*
	 *	return success to the caller
	 */
#ifdef DEBUG
	if (LIST_EMPTY(bitmapList)) SM_ASSERT(LEVEL_3, !io_Checker(volRec));
#endif DEBUG
	return(esmNOERROR);
}
