/*
 *   $RCSfile: io_RootEntry.c,v $  
 *   $Revision: 1.1.1.1 $  
 *   $Date: 1996/05/04 21:55:46 $      
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/


#include "sysdefs.h"
#include "ess.h"
#include "checking.h"
#include "trace.h"
#include "error.h"
#include "list.h"
#include "pool.h"
#include "tid.h"
#include "io.h"
#include "lock.h"
#include "object.h"
#include "msgdefs.h"
#include "thread.h"
#include "latch.h"
#include "semaphore.h"
#include "link.h"
#include "lsn.h"
#include "bf.h"
#include "log.h"
#include "logrecs.h"
#include "logaction.h"
#include "volume.h"
#include "trans.h"
#include "openlog.h"
#include "io_extfuncs.h"
#include "io_logfuncs.h"
#include "log_extfuncs.h"
#include "bf_extfuncs.h"
#include "lm_extfuncs.h"
#include "util_funcs.h"
#include "thread_globals.h"

 int
io_RemoveRootEntry (

	VOLREC		*volRec,
	char		*name 
)
{

	ROOTENTRY	*current;
	ROOTENTRY	*end;
	GROUPLINK	*rootLink;
	ROOTPAGE	*rootPage;
	PID			pid;
	LOCKMODE	lockMode;

	TRPRINT(TR_IO, TR_LEVEL_1, ("looking for name:%s on volid:%d", name, volRec->volid));

	/*
	 *	construct the pid of the header page
	 */
	pid.volid = volRec->header->volid;
	pid.page  = ROOTPAGEADDR;

	lockMode = EX;
	if (lockPage((TRANSREC *) Active->transRec, &pid, &lockMode, NOFLAGS))	{
		return(esmFAILURE);
	}

	if ((rootLink = bf_ReadPage(volRec->bufGroup, &pid, MIN_PAGE2SIZE, BF_SEM)) == NULL) {
		return(esmFAILURE);
	}
	rootPage = (ROOTPAGE *) rootLink->bufFrame;

	/*
	 *	Initialize the search variables
	 */
	current = &(rootPage->entry[0]);
	end = &(rootPage->entry[MAX_ROOT_ENTRIES]);
	
	while (current < end)	{

		/*
		 *	check to see if the slot is free
		 */
		if (current->flags != ROOT_FREE)	{

			/*
			 *	check to see if the name matches
			 */
			if (!strcmp(current->name, name))	{

				TRPRINT(TR_IO, TR_LEVEL_2, ("found name at slot:%d", current->index));

				/*
				 *	log the changes
				 */
				if (io_LogRemoveRootEntry(current->name, current->data,
								  (int) current->dataSize, 
								  rootLink->pageHash))	{

					bf_UnfixPage(rootLink, BF_DEFAULT, FALSE);
					return(esmFAILURE);
				}
				
				/*
				 *	free the information
				 */
				current->flags = ROOT_FREE;

				/*
				 *	signal the semaphore and unfix the page
				 */
				signalSemaphore( &(rootLink->pageHash->semaphore) );
				bf_UnfixPage(rootLink, BF_DEFAULT, TRUE);
		
				/*
				 *	return success
				 */
				return(esmNOERROR);
			}
		}

		/*
		 *	look at the next entry
		 */
		current++;
	}

	/*
	 *	signal the semaphore and unfix the page
	 */
	signalSemaphore( &(rootLink->pageHash->semaphore) );
	bf_UnfixPage(rootLink, BF_DEFAULT, FALSE);

	/*
	 *	return an error
	 */
	TRPRINT(TR_IO, TR_LEVEL_2, ("name not found"));
	SM_ERROR(TYPE_USER, esmBADROOTNAME);
	return(esmFAILURE);
}



 int
io_GetRootEntry (
	VOLREC				*volRec,
	char				*name,
	char				*data,
	int					*dataSize 
)
{

	ROOTENTRY	*current;
	ROOTENTRY	*end;
	GROUPLINK	*rootLink;
	ROOTPAGE	*rootPage;
	PID			pid;
	LOCKMODE	lockMode;


	TRPRINT(TR_IO, TR_LEVEL_1, 
		("looking for name:%s on volid:%d", name, volRec->header->volid));


	/*
	 *	construct the pid of the header page
	 */
	pid.volid = volRec->header->volid;
	pid.page  = ROOTPAGEADDR;

	/*
	 *	attempt to lock the page
	 */
	lockMode = SH;
	if (lockPage((TRANSREC *) Active->transRec, &pid, &lockMode, NOFLAGS))	{

		return(esmFAILURE);
	}

	/*
	 *	read in the root page
	 */
	if ((rootLink = bf_ReadPage(volRec->bufGroup, &pid, MIN_PAGE2SIZE, BF_SEM)) == NULL) {

		return(esmFAILURE);
	}

	/*
	 *	get a pointer to the root page
	 */
	rootPage = (ROOTPAGE *) rootLink->bufFrame;

	/*
	 *	Initialize the search variables
	 */
	current = &(rootPage->entry[0]);
	end = &(rootPage->entry[MAX_ROOT_ENTRIES]);
	
	/*
	 *	look for a free spot
	 */
	while (current < end)	{

		/*
		 *	check to see if the slot is free
		 */
		if (current->flags != ROOT_FREE)	{

			/*
			 *	check to see if the name matches
			 */
			if (!strcmp(current->name, name))	{

				TRPRINT(TR_IO, TR_LEVEL_2, ("found name at slot:%d", current->index));
			
				/*
				 *	plug in the information
				 */
				bcopy(current->data, data, (int) current->dataSize);		
				*dataSize = (int)current->dataSize;

				/*
				 *	signal the semaphore and unfix the page
				 */
				signalSemaphore( &(rootLink->pageHash->semaphore) );
				bf_UnfixPage(rootLink, BF_DEFAULT, FALSE);
		
				/*
				 *	return success
				 */
				return(esmNOERROR);
			}
		}

		/*
		 *	look at the next entry
		 */
		current++;
	}

	/*
	 *	signal the semaphore and unfix the page
	 */
	signalSemaphore( &(rootLink->pageHash->semaphore) );
	bf_UnfixPage(rootLink, BF_DEFAULT, FALSE);

	/*
	 *	return an error
	 */
	TRPRINT(TR_IO, TR_LEVEL_2, ("name not found"));
	SM_ERROR(TYPE_USER, esmBADROOTNAME);
	return(esmFAILURE);
}

 int
io_LogRemoveRootEntry (

	char				*name,
	char				*data,
	int					dataSize,
	PAGEHASH			*pageHash 
)
{

	LSN					lsn;
	FORCEMARK			forceMark;
	LOGRECORDINFO		recordInfo;
	ROOTPAGE			*rootPage;
	LRC					*lrc;
	ROOTENTRYLOGINFO    entryInfo;

	TRPRINT(TR_IO, TR_LEVEL_1, ("logging name:%s", name));

	/*
	 *	get a pointer to the page
	 */
	rootPage = (ROOTPAGE *) pageHash->bufFrame;

	/*
	 *	increment the lrc on the page
	 */
	lrc = &(rootPage->lrc);
	INCREMENT_LRC(lrc);

	/* set up log image 0 */
	entryInfo.oldFlags = ROOT_FREE;
	entryInfo.nameLen = strlen(name)+1;
	bcopy((char *)name, (char *)entryInfo.name, entryInfo.nameLen);


	/*
	 *	initialize the log information
	 */
	recordInfo.action		= LOG_ACTION_REMOVE_ROOTENTRY;
	recordInfo.type			= LOG_REC_TYPE_USER;
	recordInfo.imageCount	= 2;
	recordInfo.actionPid    = &(pageHash->hashList.hashPid);
	recordInfo.actionLRC    = lrc;
	recordInfo.imageSize[0] = sizeof(entryInfo);
	recordInfo.imageData[0] = (VOID *) &entryInfo;
	recordInfo.imageSize[1] = dataSize;
	recordInfo.imageData[1] = (VOID *)data;
	recordInfo.flags		= NOFLAGS;

	/*
	 *	write the record to the log
	 */
	if ((forceMark = writeLogRecord((TRANSREC *) Active->transRec, &recordInfo,
							  &lsn, NOFLAGS)) < 0)	{
		return(esmFAILURE);
	}

	/*
	 *	record the force mark and the lsn
	 */
	DEPEND_LOG(pageHash, forceMark, &lsn, lrc);

	/*
	 *	return success
	 */
	return(esmNOERROR);
}


 int
io_LogSetRootEntry (

	char				*name,
	char				*oldData,
	int					oldDataSize,
	char				*newData,
	int					newDataSize,
	ONE					oldFlags,
	PAGEHASH			*pageHash ,
	LSNOFFSET			*nextUndoLSN
)
{

	register FORCEMARK	forceMark;
	ROOTPAGE			*rootPage;
	LSN					lsn;
	LOGRECORDINFO		recordInfo;
	LRC                 *lrc;
	ROOTENTRYLOGINFO	entryInfo;


	TRPRINT(TR_IO, TR_LEVEL_1, ("logging name:%s", name));

	/*
	 *	get a pointer to the page
	 */
	rootPage = (ROOTPAGE *) pageHash->bufFrame;

	/*
	 *	increment the lrc on the page
	 */
	lrc = &(rootPage->lrc);
	INCREMENT_LRC(lrc);

	/* set up log image 0 */
	entryInfo.oldFlags = oldFlags;
	entryInfo.nameLen = strlen(name)+1;
	bcopy((char *)name, (char *)entryInfo.name, entryInfo.nameLen);

	/*
	 *	initialize the log information
	 */
	recordInfo.action		= LOG_ACTION_SET_ROOTENTRY;
	recordInfo.imageCount	= 3;
	recordInfo.actionPid    = &(pageHash->hashList.hashPid);
	recordInfo.actionLRC    = lrc;
	recordInfo.imageSize[0] = sizeof(entryInfo);
	recordInfo.imageData[0] = (VOID *) &entryInfo;
	recordInfo.imageSize[1] = oldDataSize;
	recordInfo.imageData[1] = (VOID *)oldData;
	recordInfo.imageSize[2] = newDataSize;
	recordInfo.imageData[2] = (VOID *)newData;
	recordInfo.flags		= NOFLAGS;
	if (nextUndoLSN) {
		recordInfo.type			= LOG_REC_TYPE_COMPENSATION;
		recordInfo.nextUndoLSN  = *nextUndoLSN;
	} else {
		recordInfo.type			= LOG_REC_TYPE_USER;
	}

	/*
	 *	write the record to the log
	 */
	if ((forceMark = writeLogRecord((TRANSREC *) Active->transRec, 
									&recordInfo, &lsn, NOFLAGS)) < 0)	{
		
		return(esmFAILURE);

	}

	/*
	 *	record the force mark and the lsn
	 */
	DEPEND_LOG(pageHash, forceMark, &lsn, lrc);

	/*
	 *	return success
	 */
	return(esmNOERROR);
}

 int
io_LogUndoSetRootEntry (

	char				*name,
	char				*oldData,
	int					oldDataSize,
	ONE					oldFlags,
	LSNOFFSET			nextUndoLSN,
	PAGEHASH			*pageHash 
)
{

	LSN					lsn;
	LRC					*lrc;
	FORCEMARK			forceMark;
	LOGRECORDINFO		recordInfo;
	ROOTPAGE			*rootPage;
	ROOTENTRYLOGINFO	entryInfo;


	TRPRINT(TR_IO, TR_LEVEL_1, ("logging name:%s lsn:%d", name, nextUndoLSN));

	/*
	 *	get a pointer to the page
	 */
	rootPage = (ROOTPAGE *) pageHash->bufFrame;

	/*
	 *	increment the lrc on the page
	 */
	lrc = &(rootPage->lrc);
	INCREMENT_LRC(lrc);

	/* set up log image 0 */
    entryInfo.oldFlags = oldFlags;
    entryInfo.nameLen = strlen(name)+1;
    bcopy((char *)name, (char *)entryInfo.name, entryInfo.nameLen);

	/*
	 *	initialize the log information
	 */
	recordInfo.action		= LOG_ACTION_UNSET_ROOTENTRY;
	recordInfo.type			= LOG_REC_TYPE_COMPENSATION;
	recordInfo.imageCount	= 2;
	recordInfo.actionPid    = &(pageHash->hashList.hashPid);
	recordInfo.actionLRC    = lrc;
	recordInfo.imageSize[0] = sizeof(entryInfo);
	recordInfo.imageData[0] = (VOID *) &entryInfo;
	recordInfo.imageSize[1] = oldDataSize;
	recordInfo.imageData[1] = (VOID *)oldData;
	recordInfo.nextUndoLSN  = nextUndoLSN;
	recordInfo.flags		= NOFLAGS;

	/*
	 *	write the record to the log
	 */
	if ((forceMark = writeLogRecord((TRANSREC *) Active->transRec, &recordInfo,
							  &lsn, NOFLAGS)) < 0)	{
		
		return(esmFAILURE);
	}

	/*
	 *	record the force mark and the lsn
	 */
	DEPEND_LOG(pageHash, forceMark, &lsn, lrc);

	/*
	 *	return success
	 */
	return(esmNOERROR);
}


 int
io_SetRootEntry (

	VOLREC				*volRec,
	char				*name,
	char				*data,
	int					dataSize 
)
{

	ROOTENTRY	*current;
	ROOTENTRY	*end;
	GROUPLINK	*rootLink;
	ROOTPAGE	*rootPage;
	PID			pid;
	LOCKMODE	lockMode;


	TRPRINT(TR_IO, TR_LEVEL_1, ("setting name:%s on volid:%d", name, 
		volRec->header->volid));

	/*
	 *	check to make sure the datasize is legal
	 */
	if (dataSize > MAX_ROOTDATA_SIZE)	{

		SM_ERROR(TYPE_USER, esmBADLENGTH);
		return(esmFAILURE);
	}

	/*
	 *	construct the pid of the header page
	 */
	pid.volid = volRec->header->volid;
	pid.page  = ROOTPAGEADDR;

	/*
	 *	attempt to lock the page
	 */
	lockMode = EX;
	if (lockPage((TRANSREC *) Active->transRec, &pid, &lockMode, NOFLAGS))	{
		return(esmFAILURE);
	}

	/*
	 *	read in the root page
	 */
	if ((rootLink = bf_ReadPage(volRec->bufGroup, &pid, MIN_PAGE2SIZE, BF_SEM)) == NULL) {

		return(esmFAILURE);
	}

	/*
	 *	get a pointer to the root page
	 */
	rootPage = (ROOTPAGE *) rootLink->bufFrame;

	/*
	 *	Initialize the search variables
	 */
	current = &(rootPage->entry[0]);
	end = &(rootPage->entry[MAX_ROOT_ENTRIES]);
	
	/*
	 *	look for the string
	 */
	while (current < end)	{

		/*
		 *	check to see if the slot is free
		 */
		if (current->flags != ROOT_FREE)	{

			/*
			 *	check to see if the name matches
			 */
			if (!strcmp(current->name, name))	{

				TRPRINT(TR_IO, TR_LEVEL_2, ("found name at slot:%d", current->index));

				/*
				 *	log the changes
				 */
				if (io_LogSetRootEntry(name, current->data, 
					(int)current->dataSize, data, dataSize, current->flags, rootLink->pageHash, NULL))	{

					signalSemaphore( &(rootLink->pageHash->semaphore) );
					bf_UnfixPage(rootLink, BF_DEFAULT, FALSE);
					return(esmFAILURE);
				}
				
				/*
				 *	plug in the information
				 */
				strncpy(current->name, name, MAX_ROOTNAME_SIZE);		
				bcopy(data, current->data, (int) dataSize);
				current->dataSize = dataSize;

				/*
				 *	release the semaphore and unfix the page
				 */
				signalSemaphore( &(rootLink->pageHash->semaphore) );
				bf_UnfixPage(rootLink, BF_DEFAULT, TRUE);
			
				/*
				 *	return success
				 */
				return(esmNOERROR);
			}
		}

		/*
		 *	look at the next entry
		 */
		current++;
	}

	/*
	 *	Initialize the search variables
	 */
	current = &(rootPage->entry[0]);
	end = &(rootPage->entry[MAX_ROOT_ENTRIES]);

	/*
	 *	look for a free spot
	 */
	while (current < end)	{

		/*
		 *	check to see if the slot is free
		 */
		if (current->flags == ROOT_FREE)	{

			TRPRINT(TR_IO, TR_LEVEL_2, ("using slot:%d", current->index));

			/*
			 *	log the changes
			 */
			if (io_LogSetRootEntry( name, data, 0, data, 
								   dataSize, current->flags, 
								   rootLink->pageHash, NULL))	{

				signalSemaphore( &(rootLink->pageHash->semaphore) );
				bf_UnfixPage(rootLink, BF_DEFAULT, FALSE);
				return(esmFAILURE);
			}
			
			/*
			 *	plug in the information
			 */
			strncpy(current->name, name, MAX_ROOTNAME_SIZE);		
			bcopy(data, current->data, (int) dataSize);
			current->dataSize = dataSize;
			current->flags = ROOT_USED;

			/*
			 *	release the semaphore
			 */
			signalSemaphore( &(rootLink->pageHash->semaphore) );

			/*
			 *	unfix the page
			 */
			bf_UnfixPage(rootLink, BF_DEFAULT, TRUE);
		
			/*
			 *	return success
			 */
			return(esmNOERROR);

		} else {

			/*
			 *	look at the next entry
			 */
			current++;
		}
	}

	/*
	 *	unfix the page and return the error
	 */
	signalSemaphore( &(rootLink->pageHash->semaphore) );
	bf_UnfixPage(rootLink, BF_DEFAULT, FALSE);
	TRPRINT(TR_IO, TR_LEVEL_2, ("no free slots"));
	SM_ERROR(TYPE_WARNING, esmROOTTABLEFULL);
	return(esmFAILURE);
}
