/*
 *   $RCSfile: lg_DestroyObject.c,v $  
 *   $Revision: 1.1.1.1 $  
 *   $Date: 1996/05/04 21:55:50 $      
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/


#include "sysdefs.h"
#include "ess.h"
#include "checking.h"
#include "trace.h"
#include "error.h"
#include "list.h"
#include "pool.h"
#include "tid.h"
#include "io.h"
#include "lock.h"
#include "object.h"
#include "lgobject.h"
#include "msgdefs.h"
#include "thread.h"
#include "semaphore.h"
#include "latch.h"
#include "lsn.h"
#include "bf.h"
#include "link.h"
#include "volume.h"
#include "bf_macro.h"
#include "sm_macro.h"
#include "lg_extfuncs.h"
#include "lg_intfuncs.h"
#include "bf_extfuncs.h"
#include "io_extfuncs.h"


/*
 *	Routine to delete an entire large object (or a subtree thereof).  It
 *	recursively walks the tree, giving all of its nodes back as free space.
 *	It is assumed that space deallocation occurs in a deferred manner (as always).
 */


 int
lg_DestroyObject (

	PID			*nodePid,
	BUFGROUP	*bufGroup,
	SLOTINDEX	slotRoot,
	LIST		*pageDeallocList 
)
{

    LGNODE		*nodePage;		/* ptr to buffered node			*/
	GROUPLINK	*nodeLink;
    FOUR		slotIndex;		/* slot index for deletion		*/
    PID			leafPid;		/* for leaf node page ids		*/


	TRPRINT(TR_LG, TR_LEVEL_1, ("pid:%d", nodePid->page));

	/*
	 *	check to see if this is a header on slotted pages
	 */
	if (slotRoot != NIL)	{

		PAGESLOT		*slotPtr;
		SLOTTEDPAGE		*pagePtr;
		LARGEOBJROOT	*objectPtr;

		TRPRINT(TR_LG, TR_LEVEL_2, ("slot node"));

		/*
		 * Ensure byte range validity before proceeding with scan.
		 */
		if ((nodeLink = bf_ReadPage(bufGroup, nodePid, SLOTTED_PAGE2SIZE,
								    NOFLAGS)) == NULL)	{

			return(esmFAILURE);
		}

		/*
		 *	get a pointer to the page, slot and object
		 */
		pagePtr = (SLOTTEDPAGE *) nodeLink->bufFrame;
		slotPtr = GETSLOTPTR(pagePtr, slotRoot);
		objectPtr = (LARGEOBJROOT *) GETOBJECTPTR(pagePtr, slotPtr);

		/*
		 *	get a pointer to the page
		 */
		nodePage = (LGNODE *) &(objectPtr->rootPage);

		/*
		 *	check the magic of the object
		 */
		CHECK_SLOTROOT_MAGIC(nodePage);

	} else {

		/*
		 * Read in the node and see if we're at the last level above the data.
		 * If not, recursively delete each of the current node's subtrees.
		 * Otherwise, delete each of the current node's children (i.e., the
		 * subtree's leaves).
		 */
		if ((nodeLink = bf_ReadPage(bufGroup, nodePid, LG_PAGE2SIZE, NOFLAGS)) == NULL)	{

			return(esmFAILURE);
		}

		/*
		 *	get a pointer to the node
		 */
		nodePage = (LGNODE *) nodeLink->bufFrame;
		TRPRINT(TR_LG, TR_LEVEL_2, ("at level:%d", nodePage->header.height));
		LG_PrintNode(nodePage, 0, nodePid->page);
	}

	/*
	 *	Check to see if we are at the leaf level
	 */
    if (nodePage->header.height > 1) {

		/*
		 * We haven't gotten to the level above the leaves yet.  We need
		 * to delete all subtrees nodeed here and then free up the node
		 * and associated buffer frame.
		 */
		for (slotIndex = 0; slotIndex < nodePage->header.numSlots; slotIndex++) {

			TRPRINT(TR_LG, TR_LEVEL_2, ("destroy subtree at slot:%d pid:%d",
					slotIndex, nodePage->slot[slotIndex].pid.page));
			if (lg_DestroyObject( &(nodePage->slot[slotIndex].pid), bufGroup, NIL, pageDeallocList )) {

				return(esmFAILURE);
			}
		}

    } else {

		/*
		 * We are now sitting at the level above the leaves.
		 */
		for (slotIndex = 0; slotIndex < nodePage->header.numSlots; slotIndex++) {

			/*
			 *	get the leaf pid to deallocate
			 */
			leafPid = nodePage->slot[slotIndex].pid;

			/*
			 *	If a deallocation list is provided, add the page to
			 *	the list, otherwise deallocate it now
			 */
			if (pageDeallocList != NULL) {
				if (io_FakeDeallocPages(LGDATA_PAGE2SIZE, 1, &leafPid, pageDeallocList, PAGE_LARGEDATA))	{

					return(esmFAILURE);
				}
			} else {
				if (io_DeallocPages(LGDATA_PAGE2SIZE, 1, &leafPid, NULL, TRUE, TRUE))	{

					return(esmFAILURE);
				}

				/*
				 *	blow the data page out of the cache
				 */
				bf_InvalidatePage( &leafPid );

			}
			TRPRINT(TR_LG, TR_LEVEL_2, ("hacking and deallocating page:%d", leafPid.page));
		}
    }

	/*
	 *	check to see if the slot root
	 */
	if (slotRoot == NIL)	{

		/*
		 *	If a deallocation list is provided, add the page to
		 *	the list, otherwise deallocate it now
		 */
		if (pageDeallocList != NULL) {
			if (io_FakeDeallocPages(LG_PAGE2SIZE, 1, nodePid, pageDeallocList, PAGE_LARGENODE))	{

				return(esmFAILURE);
			}

			/*
			 *	unfix the page
			 */
			bf_UnfixPage(nodeLink, BF_DEFAULT, FALSE);

		} else {
			if (io_DeallocPages(LG_PAGE2SIZE, 1, nodePid, NULL, TRUE, TRUE))	{

				return(esmFAILURE);
			}

			/*
			 *	blow the index page out of the cache
			 */
			bf_DiscardPage(bufGroup, nodeLink);

		}
		TRPRINT(TR_LG, TR_LEVEL_2, ("deallocating and discarding index node:%d",
				nodePid->page));

	} else {

		/*
		 *	unfix the page
		 */
		bf_UnfixPage(nodeLink, BF_DEFAULT, FALSE);
	}

	/*
	 *	return success
	 */
	return(esmNOERROR);
}
