/*
 *   $RCSfile: checkPointDirtyPages.c,v $  
 *   $Revision: 1.1.1.1 $  
 *   $Date: 1996/05/04 21:55:50 $
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/

#include "sysdefs.h"
#include "ess.h"
#include "checking.h"
#include "trace.h"
#include "error.h"
#include "list.h"
#include "pool.h"
#include "tid.h"
#include "io.h"
#include "lock.h"
#include "object.h"
#include "msgdefs.h"
#include "thread.h"
#include "semaphore.h"
#include "latch.h"
#include "link.h"
#include "lsn.h"
#include "bf.h"
#include "bf_macro.h"
#include "volume.h"
#include "log.h"
#include "openlog.h"
#include "trans.h"
#include "logrecs.h"
#include "threadstate.h"
#include "undo.h"
#include "util_funcs.h"
#include "log_intfuncs.h"
#include "log_extfuncs.h"
#include "bf_globals.h"
#include "bf_extfuncs.h"
#include "io_extfuncs.h"
#include "thread_globals.h"
#include "log_globals.h"
#include "trans_globals.h"
#include "undo_globals.h"

/*
 * NB: this function  (checkPointDirtyPages()) is not
 * reentrant!!! That is OK as long as only one thread can
 * run this at a time.  There should not be more than one
 * checkpoint going on at any time.
 */

static DIRTYPAGEINFO	*dirtyInfo;	/* used by both functions below */
static int				pageCount;

/*
 *	dirtyInfoCompare is used by qsort to compare pids in the 
 *	dirty page list.
 */
static
int
dirtyInfoCompare(const void* q1, const void* q2)
{
	const DIRTYPAGEINFO* p1 = (DIRTYPAGEINFO *)q1;
	const DIRTYPAGEINFO* p2 = (DIRTYPAGEINFO *)q2;
	int diff;

	diff = p1->pid.volid - p2->pid.volid;
	if (diff == 0) {
		diff = (int) p1->pid.page - (int) p2->pid.page;
		return diff;
	}
	return(diff);
}


static
 BOOL
recordDirtyInfo (void *arg1)
{
	PAGEHASH 				*pageHash = (PAGEHASH *)arg1;
	VOLREC					*volRec;

	TRACE(TR_LOG, TR_LEVEL_1);
	SM_ASSERT(LEVEL_1, (pageHash != NULL));	
	CHECK_PAGEHASH_MAGIC(pageHash);

	/*
	 *	record info only for dirty pages that are not 
	 *	in the log volume or temp volumes.
	 */
	if ((!PAGE_IS_LOG(pageHash)) && PAGE_IS_DIRTY(pageHash))	{
		volRec = io_FindVolRec(pageHash->hashList.hashPid.volid);
		if(volRec == NULL) {
			SM_ERROR(TYPE_FATAL, esmINTERNAL);
		}
		if(volRec->volflags & VOL_TEMP) {
			SM_ASSERT(LEVEL_2, !LIST_MEMBER(&(pageHash->clientDirtyList)));
		} else {

			SM_ASSERT(LEVEL_1, (pageCount < DirtyPageTableMaxSize));

			/*
			 *	record the information
			 */
			dirtyInfo->pid =	   pageHash->hashList.hashPid;
			dirtyInfo->lsn =	   pageHash->lsn;
			dirtyInfo->lrc =	   pageHash->lrc;
			SM_ASSERT(LEVEL_3, dirtyInfo->lsn.offset > 0 && 
							   dirtyInfo->lrc.count  > 0   );
			TRPRINT(TR_LOG, TR_LEVEL_2, ("pid:%d lsn:%d:%d, lrc:%d:%d",
					dirtyInfo->pid.page, dirtyInfo->lsn.wrapCount,
					dirtyInfo->lsn.offset, dirtyInfo->lrc.wrapCount,
					dirtyInfo->lrc.count));

			/*
			 *	Remove page from list of dirty pages sent from
			 *	the client since the most recent checkpoint
			 */
			if (LIST_MEMBER(&(pageHash->clientDirtyList))) {
				listRemove(&(pageHash->clientDirtyList));
			}

			/*
			 *	increment the count and the pointer
			 */
			dirtyInfo++;
			pageCount++;

		}
	} else {
		/*
		 *	the lsn and lrc should be zero
		 */
		SM_ASSERT(LEVEL_2, (pageHash->flags & H_LRC_REPAIR) || 
						   ( (pageHash->lsn.offset == 0)    &&
							 (pageHash->lsn.wrapCount == 0) &&
							 (pageHash->lrc.count == 0)     &&
							 (pageHash->lrc.wrapCount == 0)  ));
		SM_ASSERT(LEVEL_2, !LIST_MEMBER(&(pageHash->clientDirtyList))); 
	}
	return FALSE; /* keep on going */
}

 int
checkPointDirtyPages ()

{
#ifndef INIT_LRC_IS_LSN
	int					bucket;
	MISSINGUPDATEINFO	*missingUpdateInfo;
	MISSINGUPDATE		*missingUpdate;
#endif INIT_LRC_IS_LSN


	TRACE(TR_LOG, TR_LEVEL_1);

	dirtyInfo = DirtyPageTable;
	pageCount = 0;

	bf_ForEachPageInHashTable(recordDirtyInfo, NULL);

#ifndef INIT_LRC_IS_LSN
	/*
	 *	Go through all Missing Update Tables and add each page to the
	 *	checkpoint dirty page table as well
	 */
	missingUpdateInfo = (MISSINGUPDATEINFO *) FIRST_LIST_ELEMENT( &(UsedMissingUpdateTables) ); 
	while (missingUpdateInfo != NULL) {
	
		CHECK_MISSINGUPDATEINFO_MAGIC(missingUpdateInfo);

		/*
		 *	initialize the count and get the pointer to
		 *	the first bucket on the page hash table
		 */
		for (bucket = 0; bucket < PageHashTableSize; bucket++)	{

			/*
			 *	get a pointer to the first element in the bucket
			 */
			missingUpdate = (MISSINGUPDATE *) FIRST_LIST_ELEMENT( &(missingUpdateInfo->table[bucket]) );

			/*
			 *	run down the list of pages 
			 */
			while (missingUpdate != NULL)	{

				/*
				 *	Make sure there is room in the table for this 
				 *	page.
				 */
				SM_ASSERT(LEVEL_1, (pageCount < DirtyPageTableMaxSize));

				/*
				 *	record the information
				 */
				dirtyInfo->pid =	   missingUpdate->pid;
				dirtyInfo->lsn =	   missingUpdate->firstLSN;
				dirtyInfo->lrc =	   missingUpdate->firstLRC;
				SM_ASSERT(LEVEL_3, dirtyInfo->lsn.offset > 0 && 
								   dirtyInfo->lrc.count  > 0   );
				TRPRINT(TR_LOG, TR_LEVEL_2, ("pid:%d lsn:%d:%d, lrc:%d:%d",
						dirtyInfo->pid.page, dirtyInfo->lsn.wrapCount,
						dirtyInfo->lsn.offset, dirtyInfo->lrc.wrapCount,
						dirtyInfo->lrc.count));
				PRINT_PROGRESS(
					("adding page %d undergoing LRC repair to checkpoint dirty page table\n", dirtyInfo->pid.page));
				/*
				 *	increment the count and the pointer
				 */
				dirtyInfo++;
				pageCount++;
				if(pageCount > DirtyPageTableMaxSize) {
					SM_ERROR(TYPE_FATAL, esmINTERNAL);
				}


				/*
				 *	go to the next element
				 */
				missingUpdate = (MISSINGUPDATE *) NEXT_LIST_ELEMENT( &(missingUpdate->list) );
			}

		}

		missingUpdateInfo = (MISSINGUPDATEINFO *) NEXT_LIST_ELEMENT( &(missingUpdateInfo->list) ); 
	}
#endif INIT_LRC_IS_LSN

	/* remember the number of valid pages in the table */
	DirtyPageTableCount = pageCount;

	/*
	 *	Sort the pages by volume and page ID so that they are written
	 *	more efficiently
	 */
	qsort(DirtyPageTable, pageCount, sizeof(DIRTYPAGEINFO), dirtyInfoCompare);
	/*
	 *	return the count of transactions
	 */
	return(pageCount);
}
