/*
 *   $RCSfile: combineRecord.c,v $  
 *   $Revision: 1.1.1.1 $  
 *   $Date: 1996/05/04 21:55:50 $      
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/

#include "sysdefs.h"
#include "ess.h"
#include "checking.h"
#include "trace.h"
#include "error.h"
#include "list.h"
#include "pool.h"
#include "tid.h"
#include "io.h"
#include "lock.h"
#include "object.h"
#include "msgdefs.h"
#include "thread.h"
#include "semaphore.h"
#include "latch.h"
#include "link.h"
#include "lsn.h"
#include "bf.h"
#include "volume.h"
#include "openlog.h"
#include "log.h"
#include "trans.h"
#include "logrecs.h"
#include "threadstate.h"
#include "util_funcs.h"
#include "log_intfuncs.h"
#include "log_extfuncs.h"
#include "bf_extfuncs.h"
#include "thread_funcs.h"
#include "thread_globals.h"


 int
combineRecord (

	OPENLOG				*openLog,
	LOGRECORDHDR		*recordHeader,
	int					firstSize,
	BOOL				performingUndo,
	GROUPLINK			*logRecordBufLink
)
{
	char			*undoBuffer;
	int				remainBytes;
	LOGPAGEHDR		*pageHeader;
	LOGPAGEHDR		*endHeader;
	GROUPLINK		*groupLink;
	int				dataSize;
	PID				pid;
	SHORTPID		page;
	unsigned int	wrapCount;



	/*
	 *	setup some initial variables
	 */
	undoBuffer  = logRecordBufLink->bufFrame;
	remainBytes = recordHeader->length - firstSize;
	pid.volid   = openLog->volid;
	page 		= LSN_TO_LOG_PAGE(recordHeader->recordLSN.offset, openLog);

	TRPRINT(TR_LOG|TR_RECOVER, TR_LEVEL_1, 
		("page:%d firstSize:%d", page, firstSize));
	/*
	 *	copy in the part of the record that is on the current page
	 */
	bcopy((char *) recordHeader, undoBuffer, firstSize);

	/*
	 *	increment the pointer to write data
	 */
	undoBuffer += firstSize;

	/*
	 *	setup a local copy of the log wrap count
	 */
	wrapCount = openLog->wrapCount;

	/*
	 *	while there is still data copy it in
	 */
	while (remainBytes > 0)	{

		TRPRINT(TR_LOG|TR_RECOVER, TR_LEVEL_2, ("remainBytes:%d", remainBytes));

		/*
		 *	set the page to point to the next record
		 */
		page = ((page + 1) % openLog->filePages);
		TRPRINT(TR_LOG, TR_LEVEL_2, ("page:%d", page));

		/*
		 *	check to see if the log wrapped
		 */
		if (page == 0)	{
			
			wrapCount++;
			TRPRINT(TR_LOG|TR_RECOVER, TR_LEVEL_2, ("log wrapped:%d", wrapCount));
		}

		/*
		 *	calculate the page in minimum block size
		 */
		pid.page = LOG_PAGE_TO_BLOCK(page, openLog);

		/*
		 *	read the data page
		 */
		if ((groupLink = bf_ReadPage(openLog->readGroup, &pid,
									 openLog->page2size, NOFLAGS)) == NULL)	{

			SM_ERROR(TYPE_FATAL, Active->errno);
		}

		/*
		 *	get a pointer to the page header and tailer
		 */
		pageHeader = (LOGPAGEHDR *) groupLink->pageHash->bufFrame;
		endHeader  = (LOGPAGEHDR *)(((char *) pageHeader) + openLog->lastUsableByte);
		TRPRINT(TR_LOG, TR_LEVEL_2, ("endMagic:%x", endHeader->magic));

		/*
		 *	check to make sure that the page numbers are ok
		 */
		SM_ASSERT(LEVEL_1, pageHeader->pageNumber == page);
		SM_ASSERT(LEVEL_1, endHeader->pageNumber == page);

		/*
		 *	calculate the remaining data to read off this page
		 */
		dataSize = MIN(remainBytes, openLog->usableBytes);
		TRPRINT(TR_LOG, TR_LEVEL_2, ("dataSize:%d", dataSize));

		/*
		 *	check to make sure that the page magics are correct
		 */
		if (CHECK_LOGPAGE_MAGIC(pageHeader) || CHECK_LOGPAGE_MAGIC(endHeader))	{

			SM_ERROR(TYPE_FATAL, esmBADLOGPAGEHEADER);
		}

		/*
		 *	make sure the whole page made it to disk my checking
		 *	the consistency of the header and tailer and the
		 *	current wrap count
		 *
		 *	This check is not valid during undo processing, so
		 *	if undo is being performed then it cannot be true
		 */
		if ( (pageHeader->wrapCount != wrapCount) && !performingUndo)	{

			/*
			 *	this is the log eof
			 */
			TRPRINT(TR_RECOVER, TR_LEVEL_2, ("log eof wrapCount page:%d"));
			bf_UnfixPage(groupLink, BF_DEFAULT, FALSE);
			return(LOG_EOF);
		}

		/*
		 *	check to see if the lastRecord fields match for analysis
		 */
		if (pageHeader->lastRecord != endHeader->lastRecord)	{

			SM_ERROR(TYPE_LOG, esmINTERNAL);
		}

		/*
		 *	check to see if the byte count in the tailer
		 *	holds all of this record.  Note that the TAILER
		 *	must be checked in this case to make sure that
		 *	we use only parts of the log page that actually
		 *	made it out to disk
		 */
		if (endHeader->lastRecord < FIRST_LSN)	{

			/*
			 *	this is the log eof
			 */
			TRPRINT(TR_RECOVER, TR_LEVEL_2, ("log eof lastRecord:%d",
					endHeader->lastRecord));
			bf_UnfixPage(groupLink, BF_DEFAULT, FALSE);
			return(LOG_EOF);
		}

		/*
		 *	copy the needed bytes off the page
		 */
		bcopy(groupLink->bufFrame + FIRST_LSN, undoBuffer, dataSize);

		/*
		 *	decrement the number of bytes needed
		 */
		remainBytes -= dataSize;
		TRPRINT(TR_LOG|TR_RECOVER, TR_LEVEL_2, ("remainBytes:%d", remainBytes));

		/*
		 *	make sure that remain bytes is not negative
		 */
		SM_ASSERT(LEVEL_3, remainBytes >= 0);

		/*
		 *	move the data pointer
		 */
		undoBuffer += dataSize;

		/*
		 *	release the page
		 */
		bf_UnfixPage(groupLink, BF_DEFAULT, FALSE);
	}

	/*
	 *	we read all that we were asked for
	 */
	return(esmNOERROR);
}


 GROUPLINK*
getLogRecordBuffer(int size)
{
	GROUPLINK	*groupLink;
	int 		i = 0;
	PAGE2SIZE	chunkSize;

	size = size + sizeof(LOGRECORDHDR);
	if (size >= RAISETWO(OpenLog.page2size+1)) {
		/* we only need 2 log pages */
		chunkSize = LOGREC_BUF_PAGE2SIZE;
	} else {
		chunkSize = OpenLog.page2size+1;
	}

	while ( (groupLink = bf_GetTempPage(chunkSize)) == NULL) {
		i++;
		if (i > 100) {
			fprintf(stderr, 
			"Unable to get contiguous buffers for a log record.\n");
			fprintf(stderr, 
	"You need a larger buffer pool to run a server under this kind of load.\n");
			fprintf(stderr, "Server exiting.\n");

			/* we are in a loop, so crash */
			SM_ERROR(TYPE_STOP, esmINTERNAL);
		}

		/* let another thread run which may free up space */
		Sleep(SLEEP_DONT_SLEEP);
    }

	return(groupLink);
}


 void
freeLogRecordBuffer(GROUPLINK *groupLink)
{
	SM_ASSERT(LEVEL_3, groupLink != NULL);
	bf_FreeTempPage(groupLink);
}
