/*
 *   $RCSfile: undoAction.c,v $  
 *   $Revision: 1.1.1.1 $  
 *   $Date: 1996/05/04 21:55:50 $      
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/

#include "sysdefs.h"
#include "ess.h"
#include "checking.h"
#include "trace.h"
#include "error.h"
#include "list.h"
#include "pool.h"
#include "tid.h"
#include "io.h"
#include "lock.h"
#include "object.h"
#include "msgdefs.h"
#include "thread.h"
#include "semaphore.h"
#include "latch.h"
#include "link.h"
#include "lsn.h"
#include "bf.h"
#include "volume.h"
#include "openlog.h"
#include "log.h"
#include "trans.h"
#include "logrecs.h"
#include "logaction.h"
#include "undo.h"
#include "threadstate.h"
#include "io_logfuncs.h"
#include "bf_extfuncs.h"
#include "util_funcs.h"
#include "log_intfuncs.h"
#include "log_extfuncs.h"
#include "thread_funcs.h"
#include "undo_extfuncs.h"
#include "thread_globals.h"
#include "log_globals.h"
#include "trans_globals.h"


 void
undoAction (

	register LOGRECORDHDR	*recordHeader 
)
{


	TRPRINT(TR_LOG|TR_RECOVER, TR_LEVEL_1, ("ACTION:%d", recordHeader->action));

	if(CHECK_LOGRECORD_MAGIC( recordHeader)) {
		SM_ERROR(TYPE_FATAL, esmINTERNAL);
	}

	/*
	 *	take action based on type of record
	 */
	switch (recordHeader->type)	{

		case LOG_REC_TYPE_USER:
		case LOG_REC_TYPE_USER_COMPENSATION:

			/*
			 *	check to see if this is a redo only action
			 */
			if (recordHeader->flags & REDO_ONLY)	{

				TRPRINT(TR_LOG, TR_LEVEL_2, ("redo only record"));
				return;
			}

			/*
			 *	call the undo action
			 */
			switch (recordHeader->action)	{

				case LOG_ACTION_INSERT_FILE_KEY:

					undoFileKeyInsert(recordHeader);
					TRACE(TR_LOG|TR_RECOVER, TR_LEVEL_2);
					break;

				case LOG_ACTION_DELETE_FILE_KEY:

					undoFileKeyDelete(recordHeader);
					TRACE(TR_LOG|TR_RECOVER, TR_LEVEL_2);
					break;

				case LOG_ACTION_INSERT_FILE_KEY_DONE:
					undoFileKeyInsertDone(recordHeader);
					TRACE(TR_LOG|TR_RECOVER, TR_LEVEL_2);
					break;
					
				case LOG_ACTION_FILE_ADD_SLOTTED:

					undoFileAddSlotted(recordHeader);
					TRACE(TR_LOG|TR_RECOVER, TR_LEVEL_2);
					break;

				case LOG_ACTION_SET_ROOTENTRY:

					undoSetRootEntry(recordHeader);
					TRACE(TR_LOG|TR_RECOVER, TR_LEVEL_2);
					break;

				case LOG_ACTION_REMOVE_ROOTENTRY:

					undoRemoveRootEntry(recordHeader);
					TRACE(TR_LOG|TR_RECOVER, TR_LEVEL_2);
					break;

				case LOG_ACTION_FILE_CREATE_DONE:

					TRPRINT(TR_LOG|TR_RECOVER, TR_LEVEL_2, ("LOGICAL_FILE_CREATE"));
					undoFileCreateDone(recordHeader);
					break;

				case LOG_ACTION_LOGICAL_FILE_DESTROY:

					TRPRINT(TR_LOG|TR_RECOVER, TR_LEVEL_2, ("LOGICAL_FILE_DESTROY"));
					break;

				case LOG_ACTION_ALLOCATE_PAGE:

					undoPageAllocation(recordHeader);
					break;

				case LOG_ACTION_DEALLOCATE_PAGE:

					undoPageDeallocation(recordHeader);
					break;

				case LOG_ACTION_MARK_FILE_PAGE:
				case LOG_ACTION_MARK_SLOTTED_PAGE:
				case LOG_ACTION_UNMARK_FILE_PAGE:
				case LOG_ACTION_UNMARK_SLOTTED_PAGE:

					undoPageMark(recordHeader);
					break;
				case LOG_ACTION_LARGE_OBJ_ON_PAGE:

					undoMarkLargeObjOnPage(recordHeader);
					break;


				case LOG_ACTION_WRITE_SMALL_OBJ:
				case LOG_ACTION_WRITE_SMALL_OBJ_ND:
					undoWriteSmallObj(recordHeader);
					break;

				case LOG_ACTION_CREATE_SMALL_OBJ:
				case LOG_ACTION_CREATE_SMALL_OBJ_ND:
					undoCreateSmallObj(recordHeader);
					break;

				case LOG_ACTION_DESTROY_SMALL_OBJ:
				case LOG_ACTION_DESTROY_SMALL_OBJ_ND:
					undoDestroySmallObj(recordHeader);
					break;

				case LOG_ACTION_INSERT_SMALL_OBJ:
				case LOG_ACTION_INSERT_SMALL_OBJ_ND:
					undoInsertSmallObj(recordHeader);
					break;

				case LOG_ACTION_APPEND_SMALL_OBJ:
				case LOG_ACTION_APPEND_SMALL_OBJ_ND:
					undoAppendSmallObj(recordHeader);
					break;

				case LOG_ACTION_DELETE_SMALL_OBJ:
				case LOG_ACTION_DELETE_SMALL_OBJ_ND:
					undoDeleteSmallObj(recordHeader);
					break;

				case LOG_ACTION_FORWARD_SMALL_OBJ:
					undoForwardSmallObj(recordHeader);
					break;

				case LOG_ACTION_CHANGE_SMALL_OBJHDR:
					undoChangeSmallObjHdr(recordHeader);
					break;

				case LOG_ACTION_WRITE_LARGE_DATA:
				case LOG_ACTION_WRITE_LARGE_DATA_ND:
					undoWriteLargeData(recordHeader);
					break;

				case LOG_ACTION_APPEND_LARGE_DATA:
				case LOG_ACTION_APPEND_LARGE_DATA_ND:
					undoAppendLargeData(recordHeader);
					break;

				case LOG_ACTION_REMOVE_LARGE_DATA:
					undoRemoveLargeData(recordHeader);
					break;

				case LOG_ACTION_INSERT_LARGE_DATA:
				case LOG_ACTION_INSERT_LARGE_DATA_ND:
					undoInsertLargeData(recordHeader);
					break;

				case LOG_ACTION_DELETE_LARGE_DATA:
					undoDeleteLargeData(recordHeader);
					break;

				case LOG_ACTION_INIT_LARGE_NODE:
					/* Do nothing */
					break;

				case LOG_ACTION_WRITE_LARGE_NODE:
					undoWriteLargeNode(recordHeader);
					break;

				case LOG_ACTION_WRITE_LARGE_SLOTS:
					undoWriteLargeSlots(recordHeader);
					break;

				case LOG_ACTION_APPEND_LARGE_SLOTS:
				case LOG_ACTION_INSERT_LARGE_SLOTS:
					undoAppendLargeSlots(recordHeader);
					break;

				case LOG_ACTION_REMOVE_LARGE_SLOTS:
				case LOG_ACTION_DELETE_LARGE_SLOTS:
					undoRemoveLargeSlots(recordHeader);
					break;

				case LOG_ACTION_CONVERT_TO_LARGEROOT:
					undoConvertToLargeRoot(recordHeader);
					break;

				case LOG_ACTION_CONVERT_TO_LARGEHDR:
					undoConvertToLargeHdr(recordHeader);
					break;

				case LOG_ACTION_CONVERT_TO_SMALL:
					undoConvertToSmall(recordHeader);
					break;

				case LOG_ACTION_INCREMENT_SLOT_BYTECOUNTS:
				case LOG_ACTION_DECREMENT_SLOT_BYTECOUNTS:
					undoChangeSlotByteCounts(recordHeader);
					break;

				case LOG_ACTION_BTREE_LOGIC_INSERT:
				case LOG_ACTION_BTREE_LOGIC_DELETE:
				case LOG_ACTION_BTREE_PHYSIC_INSERT:
				case LOG_ACTION_BTREE_PHYSIC_DELETE:
				case LOG_ACTION_BTREE_MODIFY_LINK:
				case LOG_ACTION_BTREE_MODIFY_VECTOR0:
				case LOG_ACTION_BTREE_MODIFY_LEVEL:
				case LOG_ACTION_BTREE_LOGIC_SET_OVERFLOW:
				case LOG_ACTION_BTREE_LOGIC_INCR_OIDCNT:
				case LOG_ACTION_BTREE_LOGIC_DECR_OIDCNT:
					undoBtree(recordHeader);
					break;
					
				case LOG_ACTION_BTOV_MODIFY_NEXT_LINK:
				case LOG_ACTION_BTOV_MODIFY_PREV_LINK:
				case LOG_ACTION_BTOV_INSERT:
				case LOG_ACTION_BTOV_DELETE:
					undoBtOv(recordHeader);
					break;
					
				case LOG_ACTION_CREATE_INDEX:
					undoCreateIndex(recordHeader);
					break;


                case LOG_ACTION_LHASH_DIRPAGE_FORMAT:
				case LOG_ACTION_LHASH_INDEXPAGE_FORMAT:
				case LOG_ACTION_LHASH_DIRPAGE_INIT:
				case LOG_ACTION_LHASH_INDEXPAGE_INIT:
				case LOG_ACTION_LHASH_PHYSIC_INSERT:
				case LOG_ACTION_LHASH_PHYSIC_DELETE:
				case LOG_ACTION_LHASH_PHYLOG_INSERT:
				case LOG_ACTION_LHASH_PHYLOG_DELETE:
				case LOG_ACTION_LHASH_LOGIC_INSERT:
				case LOG_ACTION_LHASH_LOGIC_DELETE:
				case LOG_ACTION_LHASH_SET_INDEX_OVERFLOW:
				case LOG_ACTION_LHASH_RESET_INDEX_OVERFLOW:
				case LOG_ACTION_LHASH_DIR_UPDATE:
				case LOG_ACTION_LHASH_DIR_LOAD_UPDATE:
				case LOG_ACTION_LHASH_DIR_SPLIT_RESET:
				case LOG_ACTION_LHASH_DIR_SET_NEXT_ENTRY:
				case LOG_ACTION_LHASH_DIR_ROOT_CTRL:
				case LOG_ACTION_LHASH_LOGIC_INCR_OIDCNT:
				case LOG_ACTION_LHASH_LOGIC_DECR_OIDCNT:
				case LOG_ACTION_LHASH_LOGIC_SET_OVERFLOW:
				case LOG_ACTION_LHASH_LOGIC_RESET_OVERFLOW:
				case LOG_ACTION_LHASH_OV_CONVERT:
				case LOG_ACTION_LHASH_SET_THRESHOLD:

					undoLHash(recordHeader);
					break;

				case LOG_ACTION_ENTIRE_SLOTTED:        
				case LOG_ACTION_ENTIRE_LARGE_NODE:
				case LOG_ACTION_ENTIRE_LARGE_DATA:
				case LOG_ACTION_ENTIRE_BTREE_PAGE:
				case LOG_ACTION_ENTIRE_INDEXDESC:
					undoEntirePage(recordHeader);
					break;

				case LOG_ACTION_NO_LARGE_OBJ_ON_PAGE:
				default:
printf(" log action: %d \n", recordHeader->action);
					SM_ERROR(TYPE_FATAL, esmINTERNAL);
					break;
			}
			break;

		
		case LOG_REC_TYPE_COMPENSATION:

			TRPRINT(TR_LOG, TR_LEVEL_2, ("compensation record"));
			return;

		default:

			TRPRINT(TR_RECOVER, TR_LEVEL_1, ("unexpected rec type %d", 
				recordHeader->type));	
			SM_ERROR(TYPE_FATAL, esmINTERNAL);
			break;
	}

	SANITY; /* check stacks */
}
