/*
 *   $RCSfile: writeLogRecord.c,v $  
 *   $Revision: 1.1.1.1 $  
 *   $Date: 1996/05/04 21:55:49 $      
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/

#include "sysdefs.h"
#include "ess.h"
#include "checking.h"
#include "trace.h"
#include "error.h"
#include "list.h"
#include "pool.h"
#include "tid.h"
#include "io.h"
#include "lock.h"
#include "object.h"
#include "msgdefs.h"
#include "thread.h"
#include "semaphore.h"
#include "latch.h"
#include "link.h"
#include "lsn.h"
#include "bf.h"
#include "volume.h"
#include "openlog.h"
#include "trans.h"
#include "logrecs.h"
#include "log.h"
#include "logaction.h"
#include "threadstate.h"
#include "util_funcs.h"
#include "io_extfuncs.h"
#include "log_intfuncs.h"
#include "log_extfuncs.h"
#include "thread_funcs.h"
#include "thread_globals.h"
#include "log_globals.h"
#include "trans_globals.h"
#ifdef DEBUG
#	include "io_intfuncs.h"
#endif


/*
 *	This function determines which volume a log record applies to.
 */
 VOLREC*
logRecordVolume(LOGRECORDINFO *recordInfo)
{
	VOLID	volid;

	if (recordInfo->actionPid != NULL) {
		volid = recordInfo->actionPid->volid;
	} else {
		switch (recordInfo->action) {
		case LOG_ACTION_ALLOCATE_PAGE:
		case LOG_ACTION_DEALLOCATE_PAGE:
			volid = ((PID*)(recordInfo->imageData[1]))->volid;
			break;
		case LOG_ACTION_SPLIT_FILE_NODE:
			volid = ((PID*)(recordInfo->imageData[1]))->volid;
			break;
		case LOG_ACTION_INSERT_FILE_KEY_DONE:
		case LOG_ACTION_FILE_CREATE_DONE:
			volid = ((FID*)(recordInfo->imageData[0]))->pid.volid;
			break;
		default:
			SM_ERROR(TYPE_FATAL, esmINTERNAL);
		}
	}

	return io_FindVolRec(volid);
}


/*
 *	This function writes a log record (possibly spanning many log
 *	pages) to the log.
 */

 FORCEMARK
writeLogRecord (

	TRANSREC			*transRec,
	LOGRECORDINFO		*recordInfo,
	LSN					*returnLSN,
	FLAGS				flags 
)
{

	register OPENLOG	*openLog;
	register int		count;
	register int		total;
	LOGRECORDHDR		recordHeader;
	LSN					recordLSN;
	LSNOFFSET			tempLSN;
	FORCEMARK			forceMark;
	BOOL				takeCheckPoint;
	VOLREC				*volRec;		/* volume affected by record */

	TRPRINT(TR_LOG, TR_LEVEL_1, ("ACTION:%d", recordInfo->action));

	if (recordInfo->type == LOG_REC_TYPE_USER ||
		recordInfo->type == LOG_REC_TYPE_USER_COMPENSATION	) {

		/* validation action code */
		SM_ASSERT(LEVEL_3, recordInfo->action > 0);

		/*
		 *	Get information about the volume this log record is for,
		 *	so that we can determine if logging is to be performed.
		 */
		if ( (volRec = logRecordVolume(recordInfo)) == NULL) {
			SM_ERROR(TYPE_FATAL, esmINTERNAL);
		}
		if (volRec->volflags & VOL_TEMP) {
			/* no logging to be performed */
			STORE_THREAD_SWITCH(Active);
			return NULL_LSN;
		}

	}

	/* validation type code */
	SM_ASSERT(LEVEL_3, recordInfo->type <= LOG_REC_TYPE_END);

	/*
	 *	make sure the image count is valid
	 */
	SM_ASSERT(LEVEL_2, recordInfo->imageCount <= MAX_LOGREC_IMAGES);

	/*
	 *	null out the header
	 */
	bzero( (char *) &recordHeader, (int) sizeof(LOGRECORDHDR) );

	/*
	 *	calculate the total space needed for the record
	 */
	for (count = total = 0; count < recordInfo->imageCount; count++)	{

		/*
		 *	record size of component
		 */
		recordHeader.imageSize[count] = recordInfo->imageSize[count];

		/*
		 *	record offset of record within image
		 */
		recordHeader.imageOffset[count] = total;

		/*
		 *	add in the total for the component
		 */
		total += ALIGN(recordInfo->imageSize[count], 4);
		TRPRINT(TR_LOG, TR_LEVEL_2, ("entry:%d size:%d offset:%d", count,
				recordHeader.imageSize[count], recordHeader.imageOffset[count]));
	}

	/*
	 *	get a register pointer to the openlog structure
	 */
	openLog = &OpenLog;

	/*
	 *	check the open log magic number
	 */
	CHECK_OPENLOG_MAGIC(openLog);

again:

	/*
	 *	check the transaction magic number
	 */
	CHECK_TRANSREC_MAGIC(transRec);

	/*
	 *	check the state of the logging
	 */
	switch (TransInfo.logState)	{

		case L_ACTIVE:

			/*
			 *	log is open and active
			 */
			break;

		case L_RECOVER:

			/*
			 *	check to see if the transaction is in recovery
			 */
			if (transRec->transState != T_RECOVER)	{

			/*
			 *	queue the request pending recovery completion
			 */
			if (waitList( &(openLog->recoverList) , THREAD_LOG_RECOVERY_WAIT))	{

				return esmFAILURE;
			}
			}
			break;

		case L_QUIESCE:

			/*
			 *	check to see if the transaction is in abort
			 */
			if (transRec->transState != T_ABORT)	{

				SM_ERROR(TYPE_LOG, esmLOGDISABLED);
				return esmFAILURE;
			}
			break;

		default:

			/*
			 *	reject the request
			 */
			SM_ERROR(TYPE_LOG, esmLOGDISABLED);
			return esmFAILURE;
		}

	/*
	 *	If the transaction is not in the active state
	 *	it cannot write operation log records
	 */
	/*
	if ((transRec->transState != T_ACTIVE) &&
		(recordInfo->type == LOG_REC_TYPE_USER))	{

		SM_ERROR(TYPE_USER, esmBADTRANSID);
		return esmFAILURE;
	}
	*/

	/*
	 *	get the latch on the log
	 */
	if (waitLatch( &(openLog->logLatch), SHARE_LATCH ))	{

		return esmFAILURE;
	}

	/*
	 *	check to see if the log is in use
	 */
	if (waitSemaphore( &(openLog->writeSemaphore) ))	{

		signalLatch( &(openLog->logLatch) );
		return esmFAILURE;
	}

	/*
	 *	calculate to see if there is enough space to hold the log record
	 */
	switch (calculateLogSpace(openLog, total)) {

		case esmNOERROR:

			/* 	There is space */
			takeCheckPoint = FALSE;
			break;

		case esmFAILURE:

			/* The log record is too big */
			signalSemaphore( &(openLog->writeSemaphore) );
			signalLatch( &(openLog->logLatch) );
			return esmFAILURE;

		case esmNOFREELOGSPACE:	
			
			/* 	
			 *	There is no log space and taking a checkpoint will not
			 *	help.
			 */
			signalSemaphore( &(openLog->writeSemaphore) );
			signalLatch( &(openLog->logLatch) );
			SM_ERROR(TYPE_USER, esmNOFREELOGSPACE);
			return esmFAILURE;

		case LOG_DO_CHECKPOINT:

			/* A check point is required to free up space */
			takeCheckPoint = TRUE;
			break;

		default:
			
			SM_ERROR(TYPE_FATAL, esmINTERNAL);
			return esmFAILURE;
	}

	/*
	 *	if a checkpoint is needed, perform it
	 */
	if	(takeCheckPoint) {

		/*
		 *	give back the log synchronization
		 */
		signalSemaphore( &(openLog->writeSemaphore) );
		signalLatch( &(openLog->logLatch) );

		/* start checkpoint and wait for it to complete */
		activateCheckpoint(TRUE, FALSE, TRUE);

		/*
		 *	attempt to log the record again
		 *	Note, it cannot be assumed that there is enough
		 *	space after taking the checkpoint, since other
		 *	transactions may have used up the space.
		 */
		TRPRINT(TR_LOG, TR_LEVEL_2, ("returned from checkPoint"));
		goto again;
	}

	/*
	 *	increment the counter for this transaction
	 */
	forceMark = ++(openLog->logRecordCount);
	TRPRINT(TR_LOG, TR_LEVEL_2, ("forceMark:%d", forceMark));

	/*
	 *	copy in the fields of the header
	 */
	recordHeader.magic		 = LOGRECORD_MAGIC;
	recordHeader.length		 = total + sizeof(LOGRECORDHDR);
	recordHeader.type		 = recordInfo->type;
	recordHeader.action		 = recordInfo->action;
	recordHeader.flags		 = recordInfo->flags;
	recordHeader.imageCount  = recordInfo->imageCount;
	recordHeader.tid		 = transRec->tid;
	recordHeader.previousLSN = transRec->lastLSN;
	if (recordInfo->actionPid != NULL) {

		SM_ASSERT(LEVEL_3, recordInfo->actionLRC != NULL);
		recordHeader.actionPid  = *(recordInfo->actionPid);
		recordHeader.actionLRC  = *(recordInfo->actionLRC);
	}

	/*
	 *	check to see if we are logging a compensation record
	 */
	if (recordInfo->type == LOG_REC_TYPE_COMPENSATION ||
		recordInfo->type == LOG_REC_TYPE_USER_COMPENSATION)	{

		recordHeader.nextUndoLSN = recordInfo->nextUndoLSN;
		TRPRINT(TR_LOG, TR_LEVEL_2, ("nextUndoLSN:%d", recordInfo->nextUndoLSN));

		SM_ASSERT(LEVEL_3, recordHeader.nextUndoLSN <= openLog->fileBytes);
	} else {

		recordHeader.nextUndoLSN = NULL_LSN;
	}

	/*
	 *	write the record header to the log
	 */
	recordLSN = writeLogRecordHeader(openLog, &recordHeader, &tempLSN);

	/*
	 *	check to see if this is the first record for this transaction
	 */
	if (transRec->logRecordCount == 0)	{

		/*
		 *	record this as the first log record
		 *	Don't update the transaction's firstLSN if it is
		 *	already set
		 */
		transRec->firstLogPid = LSN_TO_LOG_PAGE(recordLSN.offset, openLog);
		transRec->logUnique   = openLog->logRecordCount;
		if (transRec->firstLSN.offset == NULL_LSN) transRec->firstLSN = recordLSN;
		TRPRINT(TR_LOG, TR_LEVEL_2, ("first log record for tid:%d", recordLSN.offset));

		/* 
		 *	check to see if the log queue is empty
		 */
		if (LIST_EMPTY( &(openLog->activeList) ))	{

			/*
			 *	this is the new last page
			 */
			openLog->activePid    = transRec->firstLogPid;
			openLog->activeUnique = transRec->logUnique;
			openLog->activeLSN    = recordLSN.offset;
			TRPRINT(TR_LOG, TR_LEVEL_2, ("new first active log rec:%d", recordLSN.offset));
		}

		/*
		 *	hang the transaction record off the log ordering list
		 */
		listEnq( &(openLog->activeList), &(transRec->logActiveList) );
	}

	/*
	 *	record this record as the last record
	 *	for this transaction, and record the position to begin
	 *	backwards undo processing
	 */
	transRec->lastLSN = recordLSN.offset;
	if (recordInfo->type == LOG_REC_TYPE_COMPENSATION || 
		recordInfo->type == LOG_REC_TYPE_USER_COMPENSATION)	{
		transRec->nextUndoLSN = recordInfo->nextUndoLSN;
	} else {
		transRec->nextUndoLSN = transRec->lastLSN;
	}
	SM_ASSERT(LEVEL_3, transRec->nextUndoLSN <= openLog->fileBytes);

	/* record the log space used by this trans */
	/*
	 *	PRAKASH: for distr trans */
	if ((transRec->transState == T_ACTIVE) ||
		(transRec->transState == T_PREPARED) ||
		(transRec->transState == T_COMMIT)) {
		ActiveLogSpace += recordHeader.length;
		transRec->logSpace += recordHeader.length;
	} else {
		transRec->undoLogSpace += recordHeader.length;
	}

	/*
	 *	copy the user supplied image to the log
	 */
	for (count = 0; count < recordInfo->imageCount; count++)	{

		/*
		 *	write the image subcomponent to the log
		 */
		writeImage(openLog, recordInfo, count, &tempLSN);
	}

	/* Record the next LSN to be generated */
	openLog->nextValidLSN.wrapCount = openLog->wrapCount;
	openLog->nextValidLSN.offset = 
			FIRST_LSN_ON_PAGE(openLog->tailLSN, openLog);
	
	SM_ASSERT(LEVEL_3, auditLogPage(openLog->tailLink) == esmNOERROR);

	/*
	 *	give the log latch back
	 */
	signalLatch( &(openLog->logLatch) );
	signalSemaphore( &(openLog->writeSemaphore) );

	/*
	 *	check to see if the log needs to be forced
	 */
	if (flags & LOG_FORCE)	{

		/*
		 *	force the log
		 */
		forceLog(forceMark);
	}

	STORE_THREAD_SWITCH(Active);

	/*
	 *	increment the checkpoint count
	 */
	openLog->checkPointCount++;

	/*
	 *	increment the number of records written by the transaction
	 */
	transRec->logRecordCount++;

	/*
	 *	return the lsn if requested
	 */
	if (returnLSN != NULL)	{

		*returnLSN = recordLSN;
	}

	/*
	 *	return the log count
	 */
	return forceMark;
}
