/*
 *   $RCSfile: writeLogRecordHeader.c,v $  
 *   $Revision: 1.1.1.1 $  
 *   $Date: 1996/05/04 21:55:49 $      
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/

#include "sysdefs.h"
#include "ess.h"
#include "checking.h"
#include "trace.h"
#include "error.h"
#include "list.h"
#include "pool.h"
#include "tid.h"
#include "io.h"
#include "lock.h"
#include "object.h"
#include "msgdefs.h"
#include "thread.h"
#include "semaphore.h"
#include "latch.h"
#include "link.h"
#include "lsn.h"
#include "bf.h"
#include "volume.h"
#include "openlog.h"
#include "trans.h"
#include "logrecs.h"
#include "util_funcs.h"
#include "log_intfuncs.h"
#include "log_extfuncs.h"
#include "bf_extfuncs.h"
#include "log_globals.h"
#include "thread_globals.h"


 LSN
writeLogRecordHeader (

	OPENLOG				*openLog,
	LOGRECORDHDR		*recordHeader,
	LSNOFFSET			*tempLSN 
)
{

	register SHORTPID	tailPid;
	register LSNOFFSET	tailLSN;
	register LSNOFFSET	offset;
	PAGEHASH			*tailBuffer;
	LSN					recordLSN;
#ifdef DEBUG
	LSN					debugLSN;
#endif

	TRPRINT(TR_LOG, TR_LEVEL_1, ("tailLSN:%d tailPid:%d",
			openLog->tailLSN, openLog->tailPid));

	/*
	 *	save tail byte and pid in register
	 */
	tailPid = openLog->tailPid;
	tailLSN = openLog->tailLSN;

	/*
	 *	calculate the offset of the byte within the page
	 */
	offset = LOG_PAGE_OFFSET(tailLSN, openLog);
	TRPRINT(TR_LOG, TR_LEVEL_2, ("offset:%d", offset));

	/*
	 *	check to see if the header will fit on the page
	 */
	if ((openLog->lastUsableByte - offset) < sizeof(LOGRECORDHDR))	{

		TRPRINT(TR_LOG, TR_LEVEL_2, ("header will not fit"));

		/*
		 *	wrap to the next page
		 */
		tailPid = ((tailPid + 1) % openLog->filePages);

		/*
		 *	check to see if the tail page is zero
		 */
		if (tailPid == 0)	{

			/*
			 *	increment the wrap counter
			 */
			openLog->wrapCount++;
			/* this confuses too many users
			SM_ERROR(TYPE_LOG, esmLOGWRAPPED);
			*/
		}

		/*
		 *	recalculate the tail pointer
		 */
		tailLSN = LOG_PAGE_TO_LSN(tailPid, openLog);

		/*
		 *	write out the page header
		 */
		writeLogPageHeader(openLog, tailPid);

		/*
		 *	increment the tail pointer
		 */
		tailLSN += FIRST_LSN;

		/*
		 *	set the offset
		 */
		offset = FIRST_LSN;

	} else if (offset == 0)	{

		TRPRINT(TR_LOG, TR_LEVEL_2, ("must write header"));

		/*
		 *	write out the page header
		 */
		writeLogPageHeader(openLog, tailPid);

		/*
		 *	increment the tail pointer
		 */
		tailLSN += FIRST_LSN;

		/*
		 *	set the offset
		 */
		offset = FIRST_LSN;
	}

	/*
	 *	save the record lsn
	 */
	recordLSN.offset = tailLSN;
	recordLSN.wrapCount = openLog->wrapCount;
	*tempLSN = tailLSN;

#ifdef DEBUG
	/* make sure we don't overwrite important log records */
	debugLSN = recordLSN;
	debugLSN.wrapCount--; 
	if(compareLSN(&debugLSN, &OldestDirtyPageLSN) >= 0) {
		SM_ERROR(TYPE_FATAL, esmINTERNAL);
	}
#endif

	/*
	 *	put the lsn into the log record
	 */
	recordHeader->recordLSN = recordLSN;
	TRPRINT(TR_LOG, TR_LEVEL_3, ("lsn offset:%d wrapcount:%d",
			recordLSN.offset, recordLSN.wrapCount));

	/*
	 *	save the buffer
	 */
	tailBuffer = openLog->tailBuffer;


#ifdef LOG_CHECKING 

	/*
	 *	check to see that the point that we are going to is null
	 */
	if ( *((int *)(openLog->tailBuffer->bufFrame + offset)) != 0 )	{

		/*
		 *	check to see if this is a checkpoint record that can happen
		 *	legitimately as part of recovery
		 */
		if (recordHeader->type != LOG_REC_CHKPT)	{

			SM_ERROR(TYPE_FATAL, esmINTERNAL);
		}
	}

#endif

	SM_ASSERT(LEVEL_3, auditLogPage(openLog->tailLink) == esmNOERROR);

	/*
	 *	write out the record
	 */
	bcopy((char *) recordHeader, tailBuffer->bufFrame + offset, (int) sizeof(LOGRECORDHDR) );


	/*
	 *	check  to see if there are images to be written
	 */
	if (recordHeader->imageCount == 0)	{

		/*
		 *	update the lastRecord field as entire record is done
		 */
		updateLogPageHeader(openLog, offset);

		/*
		 *	null out the temp lsn
		 */
		*tempLSN = NULL_LSN;
	}

	/*
	 *	update the tail pointer
	 */
	tailLSN += ALIGN(sizeof(LOGRECORDHDR), 4);

	/*
	 *	make sure we have not wrapped
	 */
	SM_ASSERT(LEVEL_3, LOG_PAGE_OFFSET(tailLSN, openLog) != 0); 

	/*
	 *	check to see if we advanced to a new page
	 */
	if (LOG_PAGE_OFFSET(tailLSN, openLog) == openLog->lastUsableByte)	{

		TRPRINT(TR_LOG, TR_LEVEL_2, ("advance to new page"));

		/*
		 *	Update the amount of data on the log.  This is
		 *	done here because this is the last data that
		 *	will be written to this log page.  Other eof conditions
		 *	will detect if the rest of the record is not written.
		 */
		if (recordHeader->imageCount > 0)	{
		
			/*
			 *	header has not been already updated
			 */
			updateLogPageHeader(openLog, offset);

			/*
			 *	null out the temp lsn
			 */
			*tempLSN = NULL_LSN;
		}

		/*
		 *	wrap the page
		 */
		tailPid = ((tailPid + 1) % openLog->filePages);

		/*
		 *	recalculate the tail
		 */
		tailLSN = (tailPid << openLog->page2size);

		/*
		 *	check to see if we have wrapped
		 */
		if (tailPid == 0)	{

			/*
			 *	increment the wrap count
			 */
			openLog->wrapCount++;
			TRPRINT(TR_LOG, TR_LEVEL_2, ("wrapped:%d", openLog->wrapCount));
		}
	}

	/*
	 *	check to see if the first record needs to be set
	 */
	if (tailBuffer->firstMark == 0)	{

		tailBuffer->firstMark = openLog->logRecordCount;
		TRPRINT(TR_LOG, TR_LEVEL_3, ("firstMark:%d", tailBuffer->firstMark));
	}

	/*
	 *	set the overall force mark
	 */
	tailBuffer->forceMark = openLog->logRecordCount;
	TRPRINT(TR_LOG, TR_LEVEL_3, ("forceMark:%d", tailBuffer->forceMark));

	/*
	 *	dirty the page
	 */
	DIRTY_PAGE(tailBuffer);

	/*
	 *	save the updated pointers
	 */
	openLog->tailPid = tailPid;
	openLog->tailLSN = tailLSN;
	TRPRINT(TR_LOG, TR_LEVEL_2, ("tailLSN:%d tailPid:%d", tailLSN, tailPid));

	/*
	 *	return success to the caller
	 */
	TRPRINT(TR_LOG, TR_LEVEL_2, ("recordLSN:%d:%d", recordLSN.wrapCount, recordLSN.offset));
	return(recordLSN);
}
