/*
 *   $RCSfile: writeNonTransLogRecord.c,v $  
 *   $Revision: 1.1.1.1 $  
 *   $Date: 1996/05/04 21:55:50 $      
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/

#include "sysdefs.h"
#include "ess.h"
#include "checking.h"
#include "trace.h"
#include "error.h"
#include "list.h"
#include "pool.h"
#include "tid.h"
#include "io.h"
#include "lock.h"
#include "object.h"
#include "msgdefs.h"
#include "thread.h"
#include "semaphore.h"
#include "latch.h"
#include "link.h"
#include "lsn.h"
#include "bf.h"
#include "volume.h"
#include "openlog.h"
#include "trans.h"
#include "logrecs.h"
#include "log.h"
#include "util_funcs.h"
#include "log_intfuncs.h"
#include "log_extfuncs.h"
#include "log_globals.h"
#	include "io_extfuncs.h"

 FORCEMARK
writeNonTransLogRecord (

	LOGRECORDINFO		*recordInfo,
	LSN					*returnLSN,
	FLAGS				flags 
)
{

	register OPENLOG	*openLog;
	register unsigned short	count;
	register unsigned short	total;
	LOGRECORDHDR		recordHeader;
	LSN					recordLSN;
	LSNOFFSET			tempLSN;
	FORCEMARK			forceMark;

	TRPRINT(TR_LOG, TR_LEVEL_1, ("ACTION:%d", recordInfo->action));

	/*
	 *	make sure the image count is valid
	 */
	SM_ASSERT(LEVEL_2, recordInfo->imageCount <= MAX_LOGREC_IMAGES);

	/*
	 *	null out the header
	 */
	bzero( (char *) &recordHeader, (int) sizeof(LOGRECORDHDR) );

	/*
	 *	calculate the total space needed for the record
	 */
	for (count = total = 0; count < recordInfo->imageCount; count++)	{

		/*
		 *	record size of component
		 */
		recordHeader.imageSize[count] = recordInfo->imageSize[count];

		/*
		 *	record offset of record within image
		 */
		recordHeader.imageOffset[count] = total;

		/*
		 *	add in the total for the component
		 */
		total += ALIGN(recordInfo->imageSize[count], 4);
		TRPRINT(TR_LOG, TR_LEVEL_2, ("entry:%d size:%d offset:%d", count,
				recordHeader.imageSize[count], recordHeader.imageOffset[count]));
	}

	/*
	 *	get a register pointer to the openlog structure
	 */
	openLog = &OpenLog;

	/*
	 *	check the open log magic number
	 */
	CHECK_OPENLOG_MAGIC(openLog);


	/*
	 *	get the latch on the log
	 */
	if (waitLatch( &(openLog->logLatch), SHARE_LATCH ))	{

		return(esmFAILURE);
	}

	/*
	 *	check to see if the log is in use
	 */
	if (waitSemaphore( &(openLog->writeSemaphore) ))	{

		signalLatch( &(openLog->logLatch) );
		return(esmFAILURE);
	}

	/*
	 *	calculate to see if there is enough space to hold the log record
	 */
	switch (calculateLogSpace(openLog, total)) {

	case esmNOERROR:

		break;

	case esmFAILURE:

		/*
		 *	The log record is too big
		 */
		SM_ERROR(TYPE_FATAL, esmNOFREELOGSPACE);
		return(esmFAILURE);

	case esmNOFREELOGSPACE:	
		
		/*
		 * 	There is no log space and taking a checkpoint will not
		 *	help.
		 */
		SM_ERROR(TYPE_FATAL, esmNOFREELOGSPACE);
		return(esmFAILURE);

	case LOG_DO_CHECKPOINT:

		break;

	default:
		
		SM_ERROR(TYPE_FATAL, esmINTERNAL);
		return(esmFAILURE);
	}


	/*
	 *	increment the counter for this record
	 */
	forceMark = ++(openLog->logRecordCount);
	TRPRINT(TR_LOG, TR_LEVEL_2, ("forceMark:%d", forceMark));

	/*
	 *	copy in the fields of the header
	 */
	recordHeader.magic		 = LOGRECORD_MAGIC;
	recordHeader.length		 = total + sizeof(LOGRECORDHDR);
	recordHeader.type		 = recordInfo->type;
	recordHeader.action		 = recordInfo->action;
	recordHeader.flags		 = recordInfo->flags;
	recordHeader.imageCount  = recordInfo->imageCount;
	recordHeader.tid		 = NULL_TID;
	recordHeader.previousLSN = NULL_LSN;
	if (recordInfo->actionPid != NULL) {

		SM_ASSERT(LEVEL_3, recordInfo->actionLRC != NULL);
		recordHeader.actionPid  = *(recordInfo->actionPid);
		recordHeader.actionLRC  = *(recordInfo->actionLRC);
		/*
		 *	Make sure the page is on a mounted volume
		 */
#ifdef DEBUG
		if (io_FindVolRec(recordInfo->actionPid->volid) == NULL) {
			SM_ERROR(TYPE_FATAL, esmINTERNAL);
		}
#endif
	}

	/*
	 *	check to see if we are logging a compensation record
	 */
	if (recordInfo->type == LOG_REC_TYPE_COMPENSATION ||
		recordInfo->type == LOG_REC_TYPE_USER_COMPENSATION )	{

		recordHeader.nextUndoLSN = recordInfo->nextUndoLSN;
		TRPRINT(TR_LOG, TR_LEVEL_2, ("nextUndoLSN:%d", recordInfo->nextUndoLSN));

	} else {

		recordHeader.nextUndoLSN = NULL_LSN;
	}

	/*
	 *	write the record header to the log
	 */
	recordLSN = writeLogRecordHeader(openLog, &recordHeader, &tempLSN);


	/*
	 *	copy the user supplied image to the log
	 */
	for (count = 0; count < recordInfo->imageCount; count++)	{

		/*
		 *	write the image subcomponent to the log
		 */
		writeImage(openLog, recordInfo, count, &tempLSN);
	}

	/* Record the next LSN to be generated */
	openLog->nextValidLSN.wrapCount = openLog->wrapCount;
	openLog->nextValidLSN.offset = 
			FIRST_LSN_ON_PAGE(openLog->tailLSN, openLog);

	/*
	 *	give the log latch back
	 */
	signalLatch( &(openLog->logLatch) );
	signalSemaphore( &(openLog->writeSemaphore) );

	/*
	 *	check to see if the log needs to be forced
	 */
	if (flags & LOG_FORCE)	{

		/*
		 *	force the log
		 */
		forceLog(forceMark);
	}

	/*
	 *	increment the checkpoint count
	 */
	openLog->checkPointCount++;

	/*
	 *	return the lsn if requested
	 */
	if (returnLSN != NULL)	{

		*returnLSN = recordLSN;
	}

	/*
	 *	return the log count
	 */
	return(forceMark);
}
