/*
 *   $RCSfile: analyzeDirtyPages.c,v $  
 *   $Revision: 1.1.1.1 $  
 *   $Date: 1996/05/04 21:55:55 $      
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/


#include "sysdefs.h"
#include "ess.h"
#include "checking.h"
#include "trace.h"
#include "error.h"
#include "list.h"
#include "pool.h"
#include "tid.h"
#include "io.h"
#include "lock.h"
#include "object.h"
#include "msgdefs.h"
#include "thread.h"
#include "semaphore.h"
#include "latch.h"
#include "link.h"
#include "lsn.h"
#include "bf.h"
#include "volume.h"
#include "openlog.h"
#include "trans.h"
#include "logrecs.h"
#include "logaction.h"
#include "log.h"
#include "log_extfuncs.h"
#include "io_extfuncs.h"
#include "recover_intfuncs.h"
#include "trans_extfuncs.h"
#include "thread_globals.h"
#include "log_globals.h"


/*
 *	This function handles log records which contain lists of
 *	dirty pages.
 */


 void
analyzeDirtyPages (
	LOGRECORDHDR		*record 
)
{
	DIRTYPAGELOGINFO	*dirtyLoggedInfo;
	int					numDirtyPages;
	DIRTYPAGEINFO		*dirtyInfo;
	int					i;

	TRACE(TR_IO, TR_LEVEL_1);

	/*
	 *	Get the starting location of the list of dirty pages
	 */
	dirtyLoggedInfo = (DIRTYPAGELOGINFO*) GET_LOG_IMAGE(record, 0);
	numDirtyPages = GET_LOG_IMAGE_SIZE(record, 0) / sizeof(DIRTYPAGELOGINFO);

	/*
	 *	Add each page to the dirty page table
	 */
	for (i = 0; i < numDirtyPages; i++, dirtyLoggedInfo++) {

		TRPRINT(TR_RECOVER, TR_LEVEL_2, ("page:%d", dirtyLoggedInfo->pid.page));

		/*
		 *	Make sure the page info is reasonable
		 */
#ifdef DEBUG

#ifdef INIT_LRC_IS_LSN
		SM_ASSERT(LEVEL_3, dirtyLoggedInfo->firstLRC.wrapCount <= OpenLog.wrapCount);
#else
		SM_ASSERT(LEVEL_3, dirtyLoggedInfo->firstLRC.count < 10000);
		SM_ASSERT(LEVEL_3, dirtyLoggedInfo->firstLRC.wrapCount == 0);
#endif /* INIT_LRC_IS_LSN */
		SM_ASSERT(LEVEL_3, dirtyLoggedInfo->firstLSN.wrapCount <= OpenLog.wrapCount);
		SM_ASSERT(LEVEL_3, dirtyLoggedInfo->firstLSN.offset < OpenLog.fileBytes);
#endif DEBUG


		/*
		 *	check to see if the page is in the dirty list
		 */
		if ((dirtyInfo = searchDirtyPageTable(&(dirtyLoggedInfo->pid))) == NULL)	{

			/*
			 *	create a new entry
			 */
			if ((dirtyInfo = insertDirtyPageTable(&(dirtyLoggedInfo->pid))) == NULL)  {

				SM_ERROR(TYPE_FATAL, esmINTERNAL);
			}

			/*
			 *	update the information
			 */
			dirtyInfo->lsn = dirtyLoggedInfo->firstLSN;
			dirtyInfo->lrc = dirtyLoggedInfo->firstLRC;
			dirtyInfo->tid = record->tid;
			TRPRINT(TR_RECOVER, TR_LEVEL_2, ("lsn:%d,lrc:%d", dirtyInfo->lsn.offset, dirtyInfo->lrc.count));
			PRINT_PROGRESS(("adding page %d from commit dirty list to DPT\n", dirtyLoggedInfo->pid.page));

		} else {

			/*
			 *	The page is in the dirty list, but we should correct
			 *	both the firstLSN and the firstLRC.
			 *	The firstLSN will be incorrect if the page was not
			 *	already in the dirty page table at the time of the
			 *	checkpoint.
			 *	The firstLRC can be incorrect for any page added because
			 *	of a log record after the last checkpoint.
			 */
			dirtyInfo->lsn = dirtyLoggedInfo->firstLSN;
#ifdef DEBUG
#ifdef INIT_LRC_IS_LSN
#else
			if (dirtyInfo->lrc.count != 1 && dirtyInfo->lrc.wrapCount != 0) {
				SM_ASSERT(LEVEL_1, compareLRC(&(dirtyInfo->lrc), &(dirtyLoggedInfo->firstLRC)) == 0);
			}
#endif /* INIT_LRC_IS_LSN */
			PRINT_PROGRESS(("correcting page %d from commit dirty list to DPT\n", dirtyLoggedInfo->pid.page));
#endif DEBUG
			dirtyInfo->lrc = dirtyLoggedInfo->firstLRC;
			/* the firstLSN for the page is now fixed */
			dirtyInfo->flags &= (~DPT_FIXFIRSTLRC);
		}
	}

}
