/*
 *   $RCSfile: redoUser.c,v $  
 *   $Revision: 1.1.1.1 $  
 *   $Date: 1996/05/04 21:55:55 $      
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/

#include "sysdefs.h"
#include "ess.h"
#include "checking.h"
#include "trace.h"
#include "error.h"
#include "list.h"
#include "pool.h"
#include "tid.h"
#include "io.h"
#include "lock.h"
#include "object.h"
#include "msgdefs.h"
#include "thread.h"
#include "semaphore.h"
#include "latch.h"
#include "link.h"
#include "lsn.h"
#include "bf.h"
#include "volume.h"
#include "openlog.h"
#include "trans.h"
#include "logrecs.h"
#include "logaction.h"
#include "util_funcs.h"
#include "log_intfuncs.h"
#include "log_extfuncs.h"
#include "recover_intfuncs.h"
#include "recover_extfuncs.h"
#include "redo_extfuncs.h"
#include "bf_extfuncs.h"
#include "thread_globals.h"


 void
redoUser (

	register LOGRECORDHDR	*record,
	register LSNOFFSET		recordLSN 
)
{

	TRACE(TR_LOG|TR_RECOVER, TR_LEVEL_1);

	if (record->flags & UNDO_ONLY)  {
		/* Only one record currently uses this flag */
		SM_ASSERT(LEVEL_3, (record->action == LOG_ACTION_LHASH_DIR_LOAD_UPDATE)
						|| (record->action == LOG_ACTION_LHASH_DIR_ROOT_CTRL));
		return;
	}

	TRPRINT(TR_LOG|TR_RECOVER, TR_LEVEL_1, ("action %d", record->action));
	/*
	 *	process the record
	 */
	switch (record->action)	{

		case LOG_ACTION_ALLOCATE_PAGE:

			redoPageAllocation(record);
			break;

		case LOG_ACTION_DEALLOCATE_PAGE:

			redoPageDeallocation(record);
			break;

		case LOG_ACTION_MARK_FILE_PAGE:
		case LOG_ACTION_UNMARK_FILE_PAGE:
		case LOG_ACTION_MARK_SLOTTED_PAGE:
		case LOG_ACTION_UNMARK_SLOTTED_PAGE:

			redoPageMark(record);
			break;

		case LOG_ACTION_SET_ROOTENTRY:

			redoSetRootEntry(record);
			break;

		case LOG_ACTION_UNSET_ROOTENTRY:

			redoUnSetRootEntry(record);
			break;

		case LOG_ACTION_REMOVE_ROOTENTRY:

			redoRemoveRootEntry(record);
			break;

		case LOG_ACTION_INCREMENT_UNIQUE:

			redoGetUnique(record);
			break;

		case LOG_ACTION_SLOTTED_PAGE_FORMAT:

			redoSlottedFormat(record);
			break;

		case LOG_ACTION_FILE_ROOT_FORMAT:

			redoFileRootFormat(record);
			break;

		case LOG_ACTION_INSERT_FILE_KEY:

			redoFileKeyInsert(record);
			break;

		case LOG_ACTION_DELETE_FILE_KEY:

			redoFileKeyDelete(record);
			break;

		case LOG_ACTION_SPLIT_FILE_NODE:

			redoSplitFileNode(record);
			break;

		case LOG_ACTION_FILE_ADD_SLOTTED:
			redoFileAddSlotted(record);
			break;

		case LOG_ACTION_FILE_REMOVE_SLOTTED:
			redoFileRemoveSlotted(record);
			break;

		case LOG_ACTION_LARGE_OBJ_ON_PAGE:

			redoMarkLargeObjOnPage(record);
			break;

		case LOG_ACTION_NO_LARGE_OBJ_ON_PAGE:
			
			redoUnMarkLargeObjOnPage(record);
			break;

		case LOG_ACTION_WRITE_SMALL_OBJ:
		case LOG_ACTION_WRITE_SMALL_OBJ_ND:
			redoWriteSmallObj(record);
			break;

		case LOG_ACTION_CREATE_SMALL_OBJ:
		case LOG_ACTION_CREATE_SMALL_OBJ_ND:
			redoCreateSmallObj(record);
			break;

		case LOG_ACTION_DESTROY_SMALL_OBJ:
		case LOG_ACTION_DESTROY_SMALL_OBJ_ND:
			redoDestroySmallObj(record);
			break;

		case LOG_ACTION_INSERT_SMALL_OBJ:
		case LOG_ACTION_INSERT_SMALL_OBJ_ND:
			redoInsertSmallObj(record);
			break;

		case LOG_ACTION_APPEND_SMALL_OBJ:
		case LOG_ACTION_APPEND_SMALL_OBJ_ND:
			redoAppendSmallObj(record);
			break;

		case LOG_ACTION_DELETE_SMALL_OBJ:
		case LOG_ACTION_DELETE_SMALL_OBJ_ND:
			redoDeleteSmallObj(record);
			break;

		case LOG_ACTION_FORWARD_SMALL_OBJ:
			redoForwardSmallObj(record);
			break;

		case LOG_ACTION_CHANGE_SMALL_OBJHDR:
			redoChangeSmallObjHdr(record);
			break;

		case LOG_ACTION_WRITE_LARGE_DATA:
		case LOG_ACTION_APPEND_LARGE_DATA:
		case LOG_ACTION_WRITE_LARGE_DATA_ND:
		case LOG_ACTION_APPEND_LARGE_DATA_ND:
			redoWriteLargeData(record);
			break;

		case LOG_ACTION_INSERT_LARGE_DATA:
		case LOG_ACTION_DELETE_LARGE_DATA:
		case LOG_ACTION_INSERT_LARGE_DATA_ND:
			redoInsertLargeData(record);
			break;

		case LOG_ACTION_REMOVE_LARGE_DATA:
			/* do nothing */
			break;

		case LOG_ACTION_INIT_LARGE_NODE:
			redoInitLargeNode(record);
			break;

		case LOG_ACTION_WRITE_LARGE_NODE:
			redoWriteLargeNode(record);
			break;

		case LOG_ACTION_WRITE_LARGE_SLOTS:
		case LOG_ACTION_APPEND_LARGE_SLOTS:
		case LOG_ACTION_INSERT_LARGE_SLOTS:
			redoWriteLargeSlots(record);
			break;

		case LOG_ACTION_REMOVE_LARGE_SLOTS:
		case LOG_ACTION_DELETE_LARGE_SLOTS:
			redoRemoveLargeSlots(record);
			break;

		case LOG_ACTION_CONVERT_TO_LARGEROOT:
			redoConvertToLargeRoot(record);
			break;

		case LOG_ACTION_CONVERT_TO_LARGEHDR:
			redoConvertToLargeHdr(record);
			break;

		case LOG_ACTION_CONVERT_TO_SMALL:
			redoConvertToSmall(record);
			break;

		case LOG_ACTION_UNDO_CONVERT_TO_LARGEROOT:
			redoUndoConvertToLargeRoot(record);
			break;

		case LOG_ACTION_UNDO_CONVERT_TO_LARGEHDR:
			redoUndoConvertToLargeHdr(record);
			break;

		case LOG_ACTION_INCREMENT_SLOT_BYTECOUNTS:
		case LOG_ACTION_DECREMENT_SLOT_BYTECOUNTS:
			redoChangeSlotByteCounts(record);
			break;

		case LOG_ACTION_FILE_CREATE_DONE:
		case LOG_ACTION_INSERT_FILE_KEY_DONE:
		case LOG_ACTION_TRANS_DIRTY_PAGES:
		case LOG_ACTION_LOGICAL_FILE_DESTROY:
		case LOG_ACTION_PURE_CLR:

			break;

		case LOG_ACTION_LG_NODE_PAGE_FORMAT:

			redoLgNodePageFormat(record);
			break;

		case LOG_ACTION_LG_DATA_PAGE_FORMAT:

			redoLgDataPageFormat(record);
			break;

		case LOG_ACTION_BTREE_PAGE_FORMAT:

			redoBtreePageFormat(record);
			break;

		case LOG_ACTION_BTREE_PAGE_INIT:
		case LOG_ACTION_BTREE_MODIFY_LINK:
		case LOG_ACTION_BTREE_MODIFY_VECTOR0:
		case LOG_ACTION_BTREE_MODIFY_LEVEL:
		case LOG_ACTION_BTREE_PHYSIC_INSERT:
		case LOG_ACTION_BTREE_PHYSIC_DELETE:
		case LOG_ACTION_BTREE_LOGIC_INSERT:
		case LOG_ACTION_BTREE_LOGIC_DELETE:
		case LOG_ACTION_BTREE_LOGIC_SET_OVERFLOW:
		case LOG_ACTION_BTREE_LOGIC_RESET_OVERFLOW:
		case LOG_ACTION_BTREE_LOGIC_INCR_OIDCNT:
		case LOG_ACTION_BTREE_LOGIC_DECR_OIDCNT:
			
			redoBtree(record);
			break;
			
		case LOG_ACTION_BTOV_PAGE_INIT:
		case LOG_ACTION_BTOV_MODIFY_NEXT_LINK:
		case LOG_ACTION_BTOV_MODIFY_PREV_LINK:
		case LOG_ACTION_BTOV_INSERT:
		case LOG_ACTION_BTOV_DELETE:
			
			redoBtOv(record);
			break;

		case LOG_ACTION_LHASH_DIRPAGE_FORMAT:
		case LOG_ACTION_LHASH_INDEXPAGE_FORMAT:
		case LOG_ACTION_LHASH_DIRPAGE_INIT:
		case LOG_ACTION_LHASH_INDEXPAGE_INIT:
		case LOG_ACTION_LHASH_PHYSIC_INSERT:
		case LOG_ACTION_LHASH_PHYSIC_DELETE:
		case LOG_ACTION_LHASH_PHYLOG_INSERT:
		case LOG_ACTION_LHASH_PHYLOG_DELETE:
		case LOG_ACTION_LHASH_LOGIC_INSERT:
		case LOG_ACTION_LHASH_LOGIC_DELETE:
		case LOG_ACTION_LHASH_SET_INDEX_OVERFLOW:
		case LOG_ACTION_LHASH_RESET_INDEX_OVERFLOW:
		case LOG_ACTION_LHASH_DIR_UPDATE:
		case LOG_ACTION_LHASH_DIR_LOAD_UPDATE:
		case LOG_ACTION_LHASH_DIR_SET_NEXT_ENTRY:
		case LOG_ACTION_LHASH_DIR_SPLIT_COPY:
		case LOG_ACTION_LHASH_DIR_SPLIT_RESET:
		case LOG_ACTION_LHASH_LOGIC_SET_OVERFLOW:
		case LOG_ACTION_LHASH_LOGIC_RESET_OVERFLOW:
		case LOG_ACTION_LHASH_LOGIC_INCR_OIDCNT:
		case LOG_ACTION_LHASH_LOGIC_DECR_OIDCNT:
		case LOG_ACTION_LHASH_OV_CONVERT:
		case LOG_ACTION_LHASH_SET_THRESHOLD:
		case LOG_ACTION_LHASH_DIR_ROOT_CTRL:
		
			redoLHash(record);
			break;


		case LOG_ACTION_NDXDESC_PAGE_FORMAT:

			redoNdxDescPageFormat(record);
			break;
			
		case LOG_ACTION_CREATE_INDEX:
		
			redoCreateIndex(record);
			break;
			
		case LOG_ACTION_ENTIRE_SLOTTED:        
		case LOG_ACTION_ENTIRE_LARGE_NODE:
		case LOG_ACTION_ENTIRE_LARGE_DATA:
		case LOG_ACTION_ENTIRE_BTREE_PAGE:
		case LOG_ACTION_ENTIRE_INDEXDESC:
			redoEntirePage(record);
			break;

		case LOG_ACTION_BTOV_PAGE_FORMAT:
		default:

			TRPRINT(TR_RECOVER, TR_LEVEL_1,
				("unknown action %d", record->action));
			SM_ERROR(TYPE_FATAL, esmINTERNAL);
			break;
	}
}
