/*
 *   $RCSfile: sm_AppendToSmallObj.c,v $  
 *   $Revision: 1.1.1.1 $  
 *   $Date: 1996/05/04 21:55:59 $      
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/


#include "sysdefs.h"
#include "ess.h"
#include "checking.h"
#include "trace.h"
#include "error.h"
#include "list.h"
#include "tid.h"
#include "io.h"
#include "lock.h"
#include "object.h"
#include "msgdefs.h"
#include "thread.h"
#include "latch.h"
#include "semaphore.h"
#include "link.h"
#include "lsn.h"
#include "bf.h"
#include "log.h"
#include "pool.h"
#include "volume.h"
#include "logrecs.h"
#include "sm_intfuncs.h"
#include "sm_extfuncs.h"
#include "sm_macro.h"


 int
sm_AppendToSmallObj (

    SLOTTEDPAGE     *slottedPage,       /* object's home page       */
    char            *data,
    OID             *oid,
    FOUR            length 

)
{

    SLOTINDEX		objectSlot;			/* the object's slot		*/
    FOUR			newSize;			/* size of the object after */
	FOUR			origSize;
	FOUR			origAlignLength;
	FOUR			newAlignLength;
    SMALLOBJ		*objectPtr;			/* ptr to generic object	*/
	PAGESLOT		*slotPtr;
	BOOL			changeHdrSize;
	FOUR			spaceNeeded;		/* space needed to appedn 	*/


	TRPRINT(TR_SM, TR_LEVEL_1, ("length:%d", length));

	/*
	 *	check the slotted magic number
	 */
	if (CHECK_SLOTTED_MAGIC(slottedPage))	{

		SM_ERROR(TYPE_FATAL, esmBADSLOTTEDMAGIC);
		return(esmFAILURE);
	}

	/*
	 *	get the slot position and a pointer to the slot
	 */
	objectSlot = oid->diskAddr.slot;
	slotPtr = GETSLOTPTR(slottedPage, objectSlot);

    /*
     * Make sure we have a valid oid.
     */
    if (CHECKOID(objectSlot, slottedPage, slotPtr, oid->diskAddr.unique))	{

		SM_ERROR(TYPE_WARNING, esmBADOID);
		return(esmFAILURE);
	}
	TRPRINT(TR_SM, TR_LEVEL_2, ("valid oid"));

    /*
     *	Get a pointer to the existing object
     */
    objectPtr = (SMALLOBJ *) GETOBJECTPTR(slottedPage, slotPtr);

    /*
     * Make sure object is either not versioned or not frozen
     */
    if ( (objectPtr->objectHdr.properties & (P_VERSIONED | P_FROZEN)) ==
         (P_VERSIONED | P_FROZEN) ) {
        SM_ERROR(TYPE_USER, esmFROZEN);
        return(esmFAILURE);
    }

    /*
     * Determine whether we are dealing with a large or small object.
     */
    if (objectPtr->objectHdr.properties & P_LARGEOBJ) {

		/*
		 *	This case can arise from undoConvertToLarge*.c and
		 *	redoUndoConvertToLarge*.c
		 */
		if (objectPtr->objectHdr.properties & P_HEADERONPAGE) {

			SM_ASSERT(LEVEL_3, length == sizeof(OID));
			newSize = length + sizeof(LGSLOTROOT);
			origSize = sizeof(LGSLOTROOT);

		} else {

			newSize = length + sizeof(SHORTPID);
			origSize = sizeof(SHORTPID);
		}
		origAlignLength = ALIGNSIZE(origSize);
		spaceNeeded = length;

		/* do not adjust objectHdr.size */
		changeHdrSize = FALSE;

	} else if (objectPtr->objectHdr.properties & P_MOVED)   {

		/*
		 *	RedoConvertLargeSlotRoot will call this
		 */
		SM_ASSERT(LEVEL_3, length == sizeof(LGSLOTROOT));
		newSize = length; /* no data in obj yet */
		origSize = sizeof(OID);
		origAlignLength = ALIGNSIZE(origSize);

		/* 
		 *	space is already available for and OID, so just calculate
		 *	the difference
		 */
		spaceNeeded = MAX(0, ALIGNSIZE(sizeof(LGSLOTROOT)) - sizeof(OID));
		/* do not adjust objectHdr.size */
		changeHdrSize = FALSE;
	
	} else if (objectPtr->objectHdr.properties & P_FORWARD)    {

		/*
		 *	Calculate its total final size
		 *	calculate original size
		 */
		newSize = objectPtr->objectHdr.size + length;
		origSize = objectPtr->objectHdr.size;
		origAlignLength = ALIGNSIZE(origSize);
		/* remember to adjust object's header */
		changeHdrSize = TRUE;
		spaceNeeded = length;
		TRPRINT(TR_SM, TR_LEVEL_2, ("origAlignLength:%d", origAlignLength));
		
	} else {

		/*
		 *	Calculate its total final size
		 *	calculate original size
		 */
		newSize = objectPtr->objectHdr.size + length;
		origSize = objectPtr->objectHdr.size;
		origAlignLength = ALIGNSIZE(origSize);
		/* remember to adjust object's header */
		changeHdrSize = TRUE;
		spaceNeeded = length;
		TRPRINT(TR_SM, TR_LEVEL_2, ("origAlignLength:%d", origAlignLength));
	}

	TRPRINT(TR_SM, TR_LEVEL_2, ("properties:%x newSize:%d",
			objectPtr->objectHdr.properties, newSize));

	/*
	 *	round the size up to the nearest alignment value
	 */
	newAlignLength = ALIGNSIZE(newSize);
	TRPRINT(TR_SM, TR_LEVEL_2, ("newAlignLength:%d", newAlignLength));

	/*
	 *	check to see if the object can fit on a small page
	 */
	if (newSize > LARGETHRESHOLD)	{

        /*
         *  For now this is an internal error
         */
        SM_ERROR(TYPE_FATAL, esmINTERNAL);
        return(esmFAILURE);
	}

	/*
	 *	See if the object is simple or if this is the case
	 *	of converting a forwarded object to a LARGESLOTROOT
	 */
	if ( !(objectPtr->objectHdr.properties & P_MOVED) )   {

		/*
		 * Dealing with a simple small object.
		 * First make sure there is enough free space for the new data.
		 */
		switch (SM_CheckFree(slottedPage, objectSlot, objectPtr, spaceNeeded))	{

			case PAGE_NO_SPACE:

				TRPRINT(TR_SM, TR_LEVEL_2, ("PAGE_NO_SPACE"));
		
				SM_ERROR(TYPE_FATAL, esmINTERNAL);
				return(esmFAILURE);

				break;

			case PAGE_LAST_SLOT:

				TRPRINT(TR_SM, TR_LEVEL_2, ("PAGE_LAST_SLOT"));

				/*
				 *	Reestablish pointer as it may have moved
				 */
				objectPtr = (SMALLOBJ *) GETOBJECTPTR(slottedPage, slotPtr);

				/*
				 *	we can just copy in the data
				 */
				if (data != NULL)	{
					bcopy(data, ((char *) objectPtr->data + origSize), (int) length);
				} else {
					bzero(((char *) objectPtr->data + origSize), (int) length);
				}

				/*
				 *	adjust the free start and the free count
				 */
				slottedPage->header.freeBytes -= (newAlignLength - origAlignLength);
				slottedPage->header.freeStart += (newAlignLength - origAlignLength);

				break;

			
			case PAGE_INT_SLOT:

				TRPRINT(TR_SM, TR_LEVEL_2, ("PAGE_INT_SLOT"));

				/*
				 *	Reestablish pointer as it may have moved
				 */
				objectPtr = (SMALLOBJ *) GETOBJECTPTR(slottedPage, slotPtr);

				/*
				 *	copy the exisiting data to the end of the page
				 */
				bcopy((char *) objectPtr, ((char *) slottedPage) + slottedPage->header.freeStart,
					  (int) (sizeof(OBJHDR) + origSize));

				/*
				 *	adjust the offset in the slot
				 */
				slotPtr->offset = slottedPage->header.freeStart;

				/* 
				 *	Get a pointer to the location of the new object,
				 *	Update the object's header, and adjust the 
				 *	user pointers into the object.
				 */
				objectPtr = (SMALLOBJ *) GETOBJECTPTR(slottedPage, slotPtr);

				/*
				 *	tack the data onto the end
				 */
				if (data != NULL)	{
					bcopy(data, ((char *) objectPtr->data + origSize), (int) length);
				} else {
					bzero(((char *) objectPtr->data + origSize), (int) length);
				}

				/*
				 *	adjust the free start to account for the extra data
				 */
				slottedPage->header.freeStart += sizeof(OBJHDR) + newAlignLength;

				/*
				 *	adjust the free space to account for the extra data
				 */
				slottedPage->header.freeBytes -= (newAlignLength - origAlignLength);
				TRPRINT(TR_SM, TR_LEVEL_2, ("freeBytes:%d freeStart:%d",
						slottedPage->header.freeBytes, slottedPage->header.freeStart));

				break;

			
			default:

				/*
				 *	this cannot happen
				 */
				SM_ERROR(TYPE_FATAL, esmINTERNAL);
				break;
		}

	} else {

		/*
		 *	Object is being converted to LARGESLOTROOT
		 * 	First make sure there is enough free space for the new data.
		 */
		switch (SM_CheckFree(slottedPage, objectSlot, objectPtr, spaceNeeded))	{

			case PAGE_NO_SPACE:

				TRPRINT(TR_SM, TR_LEVEL_2, ("PAGE_NO_SPACE"));
		
				SM_ERROR(TYPE_FATAL, esmINTERNAL);
				return(esmFAILURE);

				break;

			case PAGE_LAST_SLOT:

				TRPRINT(TR_SM, TR_LEVEL_2, ("PAGE_LAST_SLOT"));

				/*
				 *	Reestablish pointer as it may have moved
				 */
				objectPtr = (SMALLOBJ *) GETOBJECTPTR(slottedPage, slotPtr);

				/*
				 *	we can just copy in the data
				 */
				if (data != NULL)	{
					bcopy(data, ((char *) objectPtr->data), (int) length);
				} else {
					/* this cannot happen */
					SM_ERROR(TYPE_FATAL, esmINTERNAL);
					bzero(((char *) objectPtr->data), (int) length);
				}

				/*
				 *	adjust the free start and the free count
				 */
				slottedPage->header.freeBytes -= (newAlignLength - origAlignLength);
				slottedPage->header.freeStart += (newAlignLength - origAlignLength);

				break;

			
			case PAGE_INT_SLOT:

				TRPRINT(TR_SM, TR_LEVEL_2, ("PAGE_INT_SLOT"));

				/*
				 *	Reestablish pointer as it may have moved
				 */
				objectPtr = (SMALLOBJ *) GETOBJECTPTR(slottedPage, slotPtr);

				/*
				 *	adjust the offset in the slot
				 */
				slotPtr->offset = slottedPage->header.freeStart;

				/* 
				 *	Get a pointer to the location of the new object,
				 *	Update the object's header, and adjust the 
				 *	user pointers into the object.
				 */
				objectPtr = (SMALLOBJ *) GETOBJECTPTR(slottedPage, slotPtr);

				/*
				 *	tack the data onto the end
				 */
				if (data != NULL)	{
					bcopy(data, ((char *) objectPtr->data), (int) length);
				} else {
					/* this cannot happen */
					SM_ERROR(TYPE_FATAL, esmINTERNAL);
					bzero(((char *) objectPtr->data), (int) length);
				}

				/*
				 *	adjust the free start to account for the extra data
				 */
				slottedPage->header.freeStart += sizeof(OBJHDR) + newAlignLength;

				/*
				 *	adjust the free space to account for the extra data
				 */
				slottedPage->header.freeBytes -= (newAlignLength - origAlignLength);
				TRPRINT(TR_SM, TR_LEVEL_2, ("freeBytes:%d freeStart:%d",
						slottedPage->header.freeBytes, slottedPage->header.freeStart));

				break;

			
			default:

				/*
				 *	this cannot happen
				 */
				SM_ERROR(TYPE_FATAL, esmINTERNAL);
				break;
		}
	}

	/*
	 *	update the length in the header, if necessary
	 */
	if (changeHdrSize) objectPtr->objectHdr.size = newSize;

	/*
	 *	End the action and return
	 */
    return(esmNOERROR);
}
