/*
 *   $RCSfile: sm_DeleteFromSmallObj.c,v $  
 *   $Revision: 1.1.1.1 $  
 *   $Date: 1996/05/04 21:55:59 $      
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/


#include "sysdefs.h"
#include "ess.h"
#include "checking.h"
#include "trace.h"
#include "error.h"
#include "list.h"
#include "tid.h"
#include "io.h"
#include "lock.h"
#include "object.h"
#include "msgdefs.h"
#include "thread.h"
#include "latch.h"
#include "semaphore.h"
#include "link.h"
#include "lsn.h"
#include "bf.h"
#include "log.h"
#include "pool.h"
#include "volume.h"
#include "logrecs.h"
#include "sm_extfuncs.h"
#include "sm_macro.h"



 int
sm_DeleteFromSmallObj (

    SLOTTEDPAGE		*slottedPage,	/* object's home page			*/
	OID				*oid,
	FOUR			start,
	FOUR			length 
)
{

    SLOTINDEX		objectSlot;		/* the object's slot			*/
    OBJECT			*objectPtr;		/* ptr to generic object		*/
    SMALLOBJ		*smallObject;	/* pointer to the object		*/
    FOUR			newSize;		/* size of the object after delete	*/
	FOUR			origAlignLength;
	FOUR			newAlignLength;
    FOUR			bytesToMove;	/* bytes to move for delete		*/
    char			*source;		/* the source for the byte move	*/
    char			*deletePoint;	/* the deletion point			*/
	PAGESLOT		*slotPtr;


	TRPRINT(TR_SM, TR_LEVEL_1, ("start:%d length:%d", start, length));

	/* Make sure the page is valid */
	if (CHECK_SLOTTED_MAGIC(slottedPage))   {

		SM_ERROR(TYPE_FATAL, esmBADSLOTTEDMAGIC);
		return(esmFAILURE);
	}

	/* Get the slot index and the pointer to the slot */
    objectSlot = oid->diskAddr.slot;
	slotPtr = GETSLOTPTR(slottedPage, objectSlot);

    /* Make sure we have a valid oid.  */
    if (CHECKOID(objectSlot, slottedPage, slotPtr, oid->diskAddr.unique))	{
		SM_ERROR(TYPE_FATAL, esmINTERNAL);
		return(esmFAILURE);
	}

    /* Get a pointer to the object */
    objectPtr = GETOBJECTPTR(slottedPage, slotPtr);

    /*
     * Make sure object is either not versioned or not frozen
     */
    if ( (objectPtr->objectHdr.properties & (P_VERSIONED | P_FROZEN)) ==
         (P_VERSIONED | P_FROZEN) ) {
        SM_ERROR(TYPE_FATAL, esmINTERNAL);
        return(esmFAILURE);
    }

    /*
     * Determine whether we are dealing with a large or small object.
     */
    if (objectPtr->objectHdr.properties & P_LARGEOBJ) {

		TRPRINT(TR_SM, TR_LEVEL_2, ("large object"));	

		/* For now this is an internal error */
		SM_ERROR(TYPE_FATAL, esmINTERNAL);
		return(esmFAILURE);

    } else if (objectPtr->objectHdr.properties & P_MOVED)	{
		
		SM_ASSERT(LEVEL_3, length == sizeof(OID) && start == 0);
		origAlignLength = ALIGNSIZE(sizeof(OID));
		bytesToMove		= (sizeof(OID) - start) - length;
		/* no need to change: objectPtr->objectHdr.size */
		/*
		 * Calculate the data size of the object after the delete.
		 */
		newSize = 0;

    } else if (objectPtr->objectHdr.properties & P_FORWARD)	{
		/*
		 *	Object is forwarded part of a forwarded object
		 */

		/*
		 *	Check start and length parameters
		 */
		if (start > objectPtr->objectHdr.size)	{
			SM_ERROR(TYPE_CRASH, esmINTERNAL);
			return(esmFAILURE);
		}
		if (length < 0) {
			SM_ERROR(TYPE_FATAL, esmINTERNAL);
		} else {

			/*
			 *	Check to make sure the caller is not deleting
			 *	more data than the object holds
			 */
			SM_ASSERT(LEVEL_3, start + length <= objectPtr->objectHdr.size);	
		}

		/*
		 * Calculate the data size of the object after the delete.
		 */
		newSize = objectPtr->objectHdr.size - length;

		/*
		 *	calculate movement parameters
		 */
		origAlignLength = ALIGNSIZE(objectPtr->objectHdr.size);
		bytesToMove	= (objectPtr->objectHdr.size - start) - length;
		objectPtr->objectHdr.size = newSize;

	} else {

		/* Object is a normal small object */

		/*
		 *	Check start and length parameters
		 */
		if (start > objectPtr->objectHdr.size)	{
			SM_ERROR(TYPE_CRASH, esmINTERNAL);
			return(esmFAILURE);
		}
		if (length < 0) {
			SM_ERROR(TYPE_FATAL, esmINTERNAL);
		} else {

			/*
			 *	Check to make sure the caller is not deleting
			 *	more data than the object holds
			 */
			SM_ASSERT(LEVEL_3, start + length <= objectPtr->objectHdr.size);	
		}


		/*
		 * Calculate the data size of the object after the delete.
		 */
		newSize = objectPtr->objectHdr.size - length;

		/*
		 *	calculate movement parameters
		 */
		origAlignLength = ALIGNSIZE(objectPtr->objectHdr.size);
		TRPRINT(TR_SM, TR_LEVEL_2, ("origAlignLength:%d", origAlignLength));
		bytesToMove	= (objectPtr->objectHdr.size - start) - length;
		objectPtr->objectHdr.size = newSize;
	}

    /*
     * Dealing with a small object.
     * Delete the bytes using bcopy, then update object's header info.
     */
	smallObject		= (SMALLOBJ *) objectPtr;
    deletePoint		= smallObject->data + start;
    source			= deletePoint + length;
    bcopy(source, deletePoint, (int) bytesToMove);

	/*
	 *	calculate the new aligned length
	 */
	newAlignLength = ALIGNSIZE(newSize);
	TRPRINT(TR_SM, TR_LEVEL_2, ("newAlignLength:%d", newAlignLength));

    /* Update the page info */
    slottedPage->header.freeBytes += (origAlignLength - newAlignLength);

    return(esmNOERROR);
}
