/*
 *   $RCSfile: sm_DestroyPage.c,v $  
 *   $Revision: 1.1.1.1 $  
 *   $Date: 1996/05/04 21:55:59 $      
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/


#include "sysdefs.h"
#include "ess.h"
#include "checking.h"
#include "trace.h"
#include "error.h"
#include "list.h"
#include "pool.h"
#include "tid.h"
#include "io.h"
#include "lock.h"
#include "object.h"
#include "msgdefs.h"
#include "thread.h"
#include "semaphore.h"
#include "latch.h"
#include "lsn.h"
#include "link.h"
#include "bf.h"
#include "bf_macro.h"
#include "sm_macro.h"
#include "volume.h"
#include "io_extfuncs.h"
#include "bf_extfuncs.h"
#include "lg_extfuncs.h"
#include "sm_extfuncs.h"


/*============================================================================*/
/* 
 * sm_DestroyPage:
 * 	Routine to destroy all the objects on a page.
 *	It is called by the fi-level whenever the fi-level
 *	goes to delete a 'marked' page.
 *
 * RETURNS:
 *      eNOERROR if all went well. 
 *
 * ERROR RETURNS:
 *	esmBADPARAMS    - if there are bad input parameters.
 *	errors returned by lower levels.
 *
 * SIDE EFFECTS:
 *	If the page is unmarked, it is simply deallocated.
 *	If the page is marked, then all the large objects on
 *	the page are destroyed, then the page is deallocated.
 */
/*============================================================================*/


 int
sm_DestroyPage(

	BUFGROUP		*bufGroup,
	PID				*pid,
	LIST			*pageDeallocList 
)
{

    OBJECT			*objectPtr;		/* ptr to generic object */
	GROUPLINK		*pageLink;
    SLOTTEDPAGE		*pagePtr;		/* object's home page */
	PAGESLOT		*slotPtr;
    PID				rootPid;		/* pid of large object root */
    int				i;				/* generic index */


	TRPRINT(TR_SM, TR_LEVEL_1, ("pid:%d", pid->page));

    /*
     * Looks like there are large objects on the page.
     * Read in the page.
     */
	if ((pageLink = bf_ReadPage(bufGroup, pid, SLOTTED_PAGE2SIZE, NOFLAGS)) == NULL)	{

		SM_ERROR(TYPE_FATAL, esmINTERNAL);
	}

	/*
	 *	get a pointer to the page
	 */
	pagePtr = (SLOTTEDPAGE *) pageLink->bufFrame;

	/*
	 *	check the magic number of the slotted page
	 */
	if (CHECK_SLOTTED_MAGIC(pagePtr))	{

		SM_ERROR(TYPE_FATAL, esmBADSLOTTEDMAGIC);
		return(esmFAILURE);
	}

	/*
	 *	check to see if the large count is greater than zero
	 */
	if (pagePtr->header.largeCount > 0)	{

		/*
		 * Run through the objects on the page.
		 * Call large object destroy whenever a large object is hit.
		 */
		for (i = 0; i < pagePtr->header.slotCount; i++) {

			/*
			 *	Check to see if the slot is used
			 */
			if (pagePtr->slot[-i].offset == NIL) {

				continue;
			}

			/*
			 *	get a pointer to the object in the slot
			 */
			slotPtr = GETSLOTPTR(pagePtr, i);
			objectPtr = GETOBJECTPTR(pagePtr, slotPtr);

			/*
			 *	check to see if we have a large object
			 */
			if (objectPtr->objectHdr.properties & P_LARGEOBJ) {

				TRPRINT(TR_SM, TR_LEVEL_2, ("large object in slot:%d", i));

				/*
				 *	Check to see if this is a slotted root
				 */
				if (objectPtr->objectHdr.properties & P_HEADERONPAGE)	{

					TRPRINT(TR_SM, TR_LEVEL_2, ("slotted root in slot:%d", i));

					/*
					 *	Delete the bytes
					 */
					if (lg_DestroyObject(pid, bufGroup, i, pageDeallocList))	{

						return(esmFAILURE);
					}

				} else {

					TRPRINT(TR_SM, TR_LEVEL_2, ("page root"));

					/*
					 *	Dealing with a large object.  Dig out the root pid
					 */
					rootPid.volid = pid->volid;
					rootPid.page = ((LARGEOBJHDR *) objectPtr)->rootPid;

					/*
					 *	destroy the object
					 */
					if (lg_DestroyObject(&rootPid, bufGroup, NIL, pageDeallocList))	{

						return(esmFAILURE);
					}
				}

				/*
				 *	Blast the slot pointer just to play it safe.
				 *	Note that if the page somehow doesn't get deallocated,
				 *	then this will screw up the free space count on the page.
				 *	At least we won't get any dangling pointers to large
				 *	objects, though.
				 */
				pagePtr->slot[-i].offset = NIL;
			}
		}
	}

	/*
	 *	invalidate the page
	 */
	pagePtr->header.magic = ZERO;

	/*
	 *	Attach the page to the list of pages to be deallocated by
	 *	this transaction, if the list is provided.
	 */
	if (pageDeallocList != NULL) {

		/*
		 *	record the page deallocation
		 */
		bf_UnfixPage(pageLink, BF_DEFAULT, FALSE);
		if (io_FakeDeallocPages(SLOTTED_PAGE2SIZE, 1, pid, pageDeallocList, PAGE_SLOTTED))	{

			SM_ERROR(TYPE_FATAL, esmINTERNAL);
		}

	} else {

		/*
		 *	deallocate the page
		 */
		bf_DiscardPage(bufGroup, pageLink);
		if (io_DeallocPages(SLOTTED_PAGE2SIZE, 1, pid, NULL, TRUE, TRUE))	{

			SM_ERROR(TYPE_FATAL, esmINTERNAL);
		}

	}

	/*
	 *	return success to caller
	 */
    return(esmNOERROR);
}
