/*
 *   $RCSfile: sm_RecreateSmallObj.c,v $  
 *   $Revision: 1.1.1.1 $  
 *   $Date: 1996/05/04 21:55:59 $      
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/


#include "sysdefs.h"
#include "ess.h"
#include "checking.h"
#include "trace.h"
#include "error.h"
#include "list.h"
#include "tid.h"
#include "io.h"
#include "lock.h"
#include "object.h"
#include "msgdefs.h"
#include "thread.h"
#include "latch.h"
#include "semaphore.h"
#include "link.h"
#include "lsn.h"
#include "bf.h"
#include "log.h"
#include "pool.h"
#include "volume.h"
#include "logrecs.h"
#include "trans.h"
#include "bitmap.h"
#include "logaction.h"
#include "trans_extfuncs.h"
#include "sm_extfuncs.h"
#include "sm_intfuncs.h"
#include "sm_macro.h"


 int
sm_RecreateSmallObj (

	UONE				type, 
	SLOTTEDPAGE			*slottedPage,
	SMALLOBJ			*logObj,
	SLOTINDEX			slot,
	UNIQUE				unique,
	UFOUR				size,
	BOOL				noData
)
{ 
	int					alignLength;
	int					spaceNeeded;
	PAGESLOT			*slotPtr;
	PAGESLOT			*prevSlotPtr;
	SMALLOBJ			*objPtr;
	SLOTINDEX			emptySlot;
	SLOTINDEX			prevSlot;
	SLOTINDEX			allocateSlot;

	TRPRINT(TR_SM|TR_LOG, TR_LEVEL_1, (""));

	/*
 	 *	Calculate the space needed for the new object
	 */
	alignLength = ALIGNSIZE(size);
	spaceNeeded = alignLength + sizeof(OBJHDR);
	if (slot >= slottedPage->header.slotCount) {
		/* we will need to allocate a slot */
		spaceNeeded += sizeof(PAGESLOT);
	}

	/*
 	 *	Make sure the page has enough space
	 */
	SM_ASSERT(LEVEL_3, slottedPage->header.freeBytes >= spaceNeeded);
	if (CONTIGSPACE(slottedPage) < spaceNeeded) {
		if (SM_CompactPage(slottedPage, NIL) != esmNOERROR)  {
			SM_ERROR(TYPE_FATAL, esmINTERNAL);
		} else {

			/*
			 *  check the consistency of the page
			 *	The consistency cannot be checked if the object is
			 *	marked large, because redo will not be far enough along
			 *	to be consistent.
			 */
			if ( !(logObj->objectHdr.properties & P_LARGEOBJ)) {
				sm_AuditSlotPage(slottedPage);
			}
			SM_ASSERT(LEVEL_3, CONTIGSPACE(slottedPage) >= spaceNeeded);
		}
	}

	/*
	 *	See if the slot we need is already available for allocation
	 */
	if (slot < slottedPage->header.slotCount) {

		/*
		 *	Search for the slot in the free list
		 */
		prevSlot = NIL;
		emptySlot = slottedPage->header.freeSlot;
		while (emptySlot != NIL && emptySlot != slot) {
			slotPtr = GETSLOTPTR(slottedPage, emptySlot);
			prevSlot = emptySlot;
			emptySlot = slotPtr->unique;
		}
		/* was it found */
		if (emptySlot == slot) {
			
			/*
			 *	remove slot from free list
			 */
			slotPtr = GETSLOTPTR(slottedPage, emptySlot);
			if (prevSlot == NIL) {
				slottedPage->header.freeSlot = slotPtr->unique;
			} else {
				prevSlotPtr = GETSLOTPTR(slottedPage, prevSlot);
				prevSlotPtr->unique = slotPtr->unique;	
			}
		} else {
			SM_ERROR(TYPE_CRASH, esmINTERNAL);
		}

		/*
		 *	No new slots allocated
		 */
		allocateSlot = 0;

	} else {
		
		/*
		 *	Allocate new slots sufficient to support the slot
		 */
		allocateSlot = slot - slottedPage->header.slotCount + 1;

		/*
		 *	Add new slots to free list
		 */
		for (emptySlot = slottedPage->header.slotCount; emptySlot < slot; emptySlot++) {
			slotPtr = GETSLOTPTR(slottedPage, emptySlot);
			slotPtr->unique = slottedPage->header.freeSlot;
			slotPtr->offset = NIL;
			slottedPage->header.freeSlot = emptySlot;
		}
		emptySlot = slot;
	}
	slottedPage->header.slotCount += allocateSlot;
	SM_ASSERT(LEVEL_3, slottedPage->header.slotCount > emptySlot)
	slotPtr = GETSLOTPTR(slottedPage, emptySlot);
	
	/*
	 *	Make sure we have the correct slot
	 */
	SM_ASSERT(LEVEL_3, emptySlot == slot);

	/*
	 *	Get a pointer to the space to hold the object
	 */
	objPtr = (SMALLOBJ *) (((char *) slottedPage) + slottedPage->header.freeStart);

	/*
 	 *	Copy the new header and data
	 */
	objPtr->objectHdr = logObj->objectHdr;
	if (noData) {
		bzero(objPtr->data, size);
	} else {
		bcopy(logObj->data, objPtr->data, size);
	}

	/*
 	 *	Set up the slot contents
	 */
	slotPtr->offset = slottedPage->header.freeStart;
	slotPtr->unique = unique;
	SM_ASSERT(LEVEL_1, slotPtr->offset % ALIGNON == 0);

	/*
	 *	If the log record is a user record, then update
	 *	the unique counts.  Otherwise, its a compensation record
	 *	and there is no need to update the uniques.
	 */
	if (type == LOG_REC_TYPE_USER) {
		slottedPage->header.unique++;
		slottedPage->header.uniqueCount--;
	}

    /*
     *  Update the home page stats to account for the
     *  space allocated for the object
     */
    slottedPage->header.freeStart += sizeof(OBJHDR) + alignLength;
    slottedPage->header.freeBytes -= sizeof(OBJHDR) + alignLength;

    /*
     *  Update free space if we allocated a new slot
     */
	slottedPage->header.freeBytes -= allocateSlot * sizeof(PAGESLOT);
    TRPRINT(TR_SM, TR_LEVEL_2, ("freeStart:%d freeBytes:%d",
            slottedPage->header.freeStart, slottedPage->header.freeBytes));

	/* check the consistency of the page.  */
	sm_AuditSlotPage(slottedPage);

	return(esmNOERROR);
}
