/*
 *   $RCSfile: abortTrans.c,v $  
 *   $Revision: 1.1.1.1 $  
 *   $Date: 1996/05/04 21:56:02 $      
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/

#include "sysdefs.h"
#include "ess.h"
#include "checking.h"
#include "trace.h"
#include "error.h"
#include "list.h"
#include "pool.h"
#include "tid.h"
#include "io.h"
#include "lock.h"
#include "object.h"
#include "msgdefs.h"
#include "thread.h"
#include "semaphore.h"
#include "link.h"
#include "lsn.h"
#include "latch.h"
#include "bf.h"
#include "volume.h"
#include "trans.h"
#include "openlog.h"
#include "logrecs.h"
#include "bitvec.h"
#include "adminmsg.h"
#include "trans_intfuncs.h"
#include "trans_extfuncs.h"
#include "msg_funcs.h"
#include "log_extfuncs.h"
#include "lm_extfuncs.h"
#include "io_extfuncs.h"
#include "trans_globals.h"
#include "log_globals.h"
#include "thread_globals.h"
#include "admin_globals.h"
#include "bf_macro.h"
#include "distr.h"
#include "distr_intfuncs.h"

/*
 *	This function should be called when a transaction is in
 *	any state except T_ACTIVE.  If the transaction is active,
 *	then internalAbort() should be called.
 */

 void
abortTrans (

	TRANSREC	*transRec,
	int			reason
)
{

	PAGEHASH		*pageHash;
	LIST			pendingBitmapList;



	TRPRINT(TR_DISTR|TR_TRANS, TR_LEVEL_1, ("tid:%d", GETTID(transRec)));

	/*
	 *	set the transaction pointer in the thread
	 */
	Active->transRec = (char *) transRec;

	transRec->transState = T_ABORT;

	/* deallocate all temporary files and pages */
   	deallocTempFilesPages(transRec);

	initializeList(&(pendingBitmapList));

	/* 
	 *	Free all information on files and pages
	 *	deallocated by this transaction
	 */
	freeFileDeallocInfo(transRec);
	freePageDeallocInfo(transRec);
	freeServerTidInfo(transRec);

	/*
	 *	check to see if there are any log records
	 *	that are written for this transaction
	 */
	if (transRec->logRecordCount > 0)	{

		/*
		 *	undo the transaction actions
		 */
		undoTransaction(NULL_LSN, NULL_LSN, NOFLAGS);

		/*
		 *  add all the pages in the all the files to be destroyed
		 *  to the list of pages to deallocate
		 */
		if (transDeallocFilePages(transRec) != esmNOERROR) {
			/* clean up dealloc info */
			SM_ERROR(TYPE_CRASH, esmINTERNAL);
		}

		/*
		 *  Log all pending page deallocations
		 */
		if (logTransDeallocPages(transRec, &pendingBitmapList) != esmNOERROR) {
			/*
			 *  For now, just crash, but we should really clean things up
			 *  and abort the transaction.  Clean-up is complicated by
			 *  the bitmap and volume header pages which are pinned with
			 *  pending operations.
			 */
			SM_ERROR(TYPE_CRASH, esmINTERNAL);
		}

		/*
		 *	Unlink all pages in the list of pages written by this
		 *	transaction since the last checkpoint. 
		 */
		while ( (pageHash = (PAGEHASH*) listDeq(&(transRec->clientDirtyList))) != NULL) {
			CHECK_PAGEHASH_MAGIC(pageHash);
		}

		/*
		 *	we must log the abort record, there is no reason to force
		 *	the record to disk
		 */
		if (writeTransStateRecord(transRec, LOG_REC_TYPE_ABORT, ZERO, T_ABORT, FALSE) < esmNOERROR)	{

			SM_ERROR(TYPE_FATAL, Active->errno);
		}

		TRPRINT(TR_DISTR|TR_TRANS, TR_LEVEL_1, ("server :wrote T_ABORT : tid:%d", GETTID(transRec)));

        /*
         *  Perform all pending page deallocations
         */
        if (deallocTransPages(transRec, &pendingBitmapList) != esmNOERROR) {
            SM_ERROR(TYPE_FATAL, esmINTERNAL);
        }

		/*
		 *	take the record off the log queue
		 */
		removeLogList(transRec);
	}

	/*
	 *	Make sure all bitmaps are consistent after any
	 *	deallocation/allocation
	 */
	SM_ASSERT(LEVEL_3, io_CheckAllVolumes() == esmNOERROR);

	/*
	 *	free all the locks
	 */
	unlockAll(transRec, esmTRANSABORTED);

	/* Store the reason for the abort on this link */
	if (transRec->masterLink != NULL) {
		transRec->masterLink->linkTransError = reason;
	}

	/*
	 *	free the structures
	 */
	freeTransVolRecs(transRec);

	TotalAborts++;

	/*
	 *	give back the transaction record
	 */
	freeTransRec(transRec);

	return;
}
