/*
 *   $RCSfile: recordMissingUpdate.c,v $  
 *   $Revision: 1.1.1.1 $  
 *   $Date: 1996/05/04 21:56:03 $      
 */ 
/**********************************************************************
* EXODUS Database Toolkit Software
* Copyright (c) 1991 Computer Sciences Department, University of
*                    Wisconsin -- Madison
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY OF WISCONSIN --
* MADISON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  
* THE DEPARTMENT DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The EXODUS Project Group requests users of this software to return 
* any improvements or extensions that they make to:
*
*   EXODUS Project Group 
*     c/o David J. DeWitt and Michael J. Carey
*   Computer Sciences Department
*   University of Wisconsin -- Madison
*   Madison, WI 53706
*
*	 or exodus@cs.wisc.edu
*
* In addition, the EXODUS Project Group requests that users grant the 
* Computer Sciences Department rights to redistribute these changes.
**********************************************************************/


#include "sysdefs.h"
#include "ess.h"
#include "checking.h"
#include "trace.h"
#include "error.h"
#include "list.h"
#include "tid.h"
#include "io.h"
#include "lock.h"
#include "object.h"
#include "msgdefs.h"
#include "thread.h"
#include "latch.h"
#include "semaphore.h"
#include "link.h"
#include "lsn.h"
#include "bf.h"
#include "pool.h"
#include "log.h"
#include "volume.h"
#include "trans.h"
#include "logrecs.h"
#include "undo.h"
#include "openlog.h"
#include "logaction.h"
#include "io_logfuncs.h"
#include "io_extfuncs.h"
#include "undo_extfuncs.h"
#include "bf_extfuncs.h"
#include "log_extfuncs.h"
#include "util_funcs.h"
#include "thread_globals.h"
#include "bf_globals.h"
#include "trans_globals.h"
#include "sm_macro.h"


 void
recordMissingUpdate (
	BOOL				actionDone,	/* Indicates whether the action
									   was ever performed on the page */
	LRC					**returnLRCptr,
#ifndef INIT_LRC_IS_LSN
	MISSINGUPDATEINFO 	*missingUpdateInfo,
#endif INIT_LRC_IS_LSN
	PID					*pid,
	LRC					*pageLRC
#ifndef INIT_LRC_IS_LSN
	,
	PAGEHASH			*pageHash,
	LOGRECORDHDR		*recordHeader,
	PAGETYPE			pageType
#endif INIT_LRC_IS_LSN

)
{
	

#ifdef INIT_LRC_IS_LSN
	
	TRPRINT(TR_RECOVER, TR_LEVEL_1, ("pid %d:%d", pid->page, pid->volid));

	/*
	 *	No patching of LRCs is needed.
	 */
	if (actionDone) {
		/*	
		 *	Do nothing but return a pointer to the lrc in the page
		 */
		*returnLRCptr = pageLRC;
	} else {
		/*	
		 *	return the pointer to the temporary LRC passed in,
		 *	i.e. don't change returnLRCptr;
		 */
	}

#else
	MISSINGUPDATE		*missingUpdate;
	LRC					*logLRC = &(recordHeader->actionLRC);
	LSN					*firstLSN = NULL;
	LRC					*firstLRC = NULL;
	LRC					tempLRC;

	TRPRINT(TR_RECOVER, TR_LEVEL_1, ("pid %d:%d", pid->page, pid->volid));

  	SM_ASSERT(LEVEL_3, (missingUpdateInfo != NULL));
  	SM_ASSERT(LEVEL_3, (missingUpdateInfo->table != NULL));
	/*
	 *	See if the page is in the missing update table
	 */
	if ( (missingUpdate = lookupUndoPage(missingUpdateInfo->table, pid)) != NULL) {
		/*
		 *	The page is in the table, was the action performed
		 */
		if (actionDone) {
			
			/*
			 *	Record the updated LRC onto the page.
			 *	Return a pointer to the lrc on the page.
			 *	Update the firstLRC/LSN for the page to
			 *	  reflect the first operation not done
			 *	Remove the page from the table
			 */
			*pageLRC = missingUpdate->lrc;
			*returnLRCptr = pageLRC;
			DEPEND_LOG(pageHash, pageHash->forceMark, &(missingUpdate->firstLSN), &(missingUpdate->firstLRC));
			removeUndoPage(missingUpdateInfo->table, pid);

			/*
			 *	Mark the page as having been repaired (ie. it is
			 *	now dirty
			 */
			pageHash->flags &= ~(H_LRC_REPAIR);
			DIRTY_PAGE(pageHash);

		} else {

			/*
			 *	Return a pointer to the lrc in the table, first
			 */
			*returnLRCptr = &(missingUpdate->lrc);
		}

	} else {

		/*
		 *	The page was not in the table, was the action performed
		 */
		if (actionDone) {

			/*
			 *	Do nothing but return a pointer to the lrc in the page
			 */
			*returnLRCptr = pageLRC;

		} else {
			
			/*
			 *	Generate the firstLSN/LRC for the page.  First,
			 *	see if the page already has a firstLSN/LRC.
			 */
			if (pageHash->lsn.offset != NULL_LSN) {

				/*
				 *	The page already has them, so use those
				 */
				firstLSN = &(pageHash->lsn);
				firstLRC = &(pageHash->lrc);
			} else {

				/*
				 *	The page has not been dirtied, so use the
				 *	lrc on the page +1 as the firstLRC and the 
				 *	transaction's first LSN as the firstLSN
				 */
				firstLSN = &(((TRANSREC*) (Active->transRec))->firstLSN);
				SM_ASSERT(LEVEL_3, compareLSN(&(recordHeader->recordLSN), firstLSN) >= 0);
				tempLRC = *pageLRC;
				INCREMENT_LRC(&tempLRC);
				firstLRC = &tempLRC;
			}

			/*
			 *	Put the page in the table and return a pointer to the
			 *	missing update info in the table.
			 */
			missingUpdate = addUndoPage(missingUpdateInfo->table, pid, logLRC, firstLSN, firstLRC, pageType);
			*returnLRCptr = &(missingUpdate->lrc);
			PRINT_PROGRESS(("Patching lrc for page %d\n", pid->page));

			/*
			 *	Mark the page as being repaired instead of dirty
			 */
			CLEAN_PAGE(pageHash);
			pageHash->flags |= H_LRC_REPAIR;
		}
	}

#endif /* INIT_LRC_IS_LSN */

	/*
	 *	return success
	 */
	return;
}
